Encoding: UTF-8
Package: dplR
Type: Package
Title: Dendrochronology Program Library in R
Version: 1.7.9
Authors@R: c(person("Andy", "Bunn", role = c("aut", "cph","cre", "trl"), email = "bunna@wwu.edu"), 
        person("Mikko","Korpela", role = c("aut", "cph", "trl")), 
        person("Franco", "Biondi", role = c("aut", "cph")), 
        person("Filipe", "Campelo", role = c("aut", "cph")), 
        person("Stefan","Klesse", role = c("aut", "cph")),
        person("Pierre", "Mérian", role = c("aut","cph")), 
        person("Fares", "Qeadan", role = c("aut", "cph")),
        person("Christian", "Zang", role = c("aut", "cph")),
        person("Allan", "Buras", role = "ctb"), 
        person("Alice","Cecile", role = "ctb"), 
        person("Manfred", "Mudelsee", role ="ctb"), 
        person("Michael", "Schulz", role = "ctb"),
        person("David", "Frank",role = "ctb"), 
        person("Ronald","Visser", role = "ctb"),
        person("Ed","Cook", role = "ctb"),
        person("Kevin","Anchukaitis", role = "ctb"))
Copyright: Authors and file inst/COPYRIGHTS
Depends: R (>= 3.5.0)
Imports: graphics, grDevices, grid, stats, utils, lattice (>= 0.13-6),
        Matrix (>= 1.0-3), digest (>= 0.2.3), matrixStats (>= 0.50.2),
        png (>= 0.1-2), R.utils (>= 1.32.1), stringi (>= 0.2-3),
        stringr (>= 0.4), XML (>= 2.1-0), signal, boot, lme4, lifecycle
Suggests: Cairo (>= 1.5-0), dichromat (>= 1.2-3), foreach, forecast (>=
        3.6), gmp (>= 0.5-5), iterators, knitr, RColorBrewer,
        rmarkdown, testthat (>= 0.8), tikzDevice, waveslim
Description: Perform tree-ring analyses such as detrending, chronology
        building, and cross dating.  Read and write standard file formats
        used in dendrochronology.
LazyData: no
License: GPL (>= 2)
URL: https://github.com/OpenDendro/dplR
NeedsCompilation: yes
Packaged: 2026-05-20 21:26:30 UTC; andybunn
Author: Andy Bunn [aut, cph, cre, trl],
  Mikko Korpela [aut, cph, trl],
  Franco Biondi [aut, cph],
  Filipe Campelo [aut, cph],
  Stefan Klesse [aut, cph],
  Pierre Mérian [aut, cph],
  Fares Qeadan [aut, cph],
  Christian Zang [aut, cph],
  Allan Buras [ctb],
  Alice Cecile [ctb],
  Manfred Mudelsee [ctb],
  Michael Schulz [ctb],
  David Frank [ctb],
  Ronald Visser [ctb],
  Ed Cook [ctb],
  Kevin Anchukaitis [ctb]
Maintainer: Andy Bunn <bunna@wwu.edu>
Repository: CRAN
Date/Publication: 2026-05-21 05:10:20 UTC

useDynLib(dplR, .registration = TRUE, .fixes = "dplR.")


###############################################################
#
## Imports
#
###############################################################

import(graphics, stats)

importFrom(digest, digest)

importFrom(grDevices, dev.hold, dev.flush, rainbow, dev.capabilities,
           dev.cur, png, dev.off, dev.set, devAskNewPage, dev.capture)

importFrom(grid, gpar, grid.lines, grid.newpage, grid.polygon,
           grid.segments, grid.text, pushViewport, seekViewport, unit,
           viewport, vpList, vpTree, plotViewport, grid.grill, upViewport,
           grid.points, popViewport, grid.rect, textGrob, grid.draw,
           segmentsGrob, linesGrob, grobTree)

importFrom(lattice, panel.abline, panel.dotplot, panel.segments,
           trellis.par.set, xyplot)

importFrom(matrixStats, colAlls, colAnys, colMedians, colSds, rowAlls,
           rowAnys, rowSds)

importFrom(png, readPNG)

importFrom(R.utils, captureOutput)

importFrom(stringi, stri_trans_nfc, stri_trans_nfd)

importFrom(stringr, str_pad, str_trim)

importFrom(utils, head, packageDescription, read.fwf, tail,
           packageVersion, write.table, read.table, write.csv)

import(Matrix)

importFrom(XML, xmlEventParse)

importFrom(plyr, alply)

importFrom(signal, butter, filtfilt)

importFrom(boot, boot, boot.ci)

###############################################################
#
## Methods
#
###############################################################

S3method(plot, crn)
S3method(plot, crs)
S3method(plot, rwl)
S3method(print, redfit)
S3method(print, rwl.report)
S3method(summary, rwl)
S3method(time, rwl)
S3method(time, crn)
S3method("time<-", rwl)
S3method("time<-", crn)

###############################################################
#
## Exports
#
###############################################################

export("ads",
       "as.rwl",
       "autoread.ids",
       "bai.in",
       "bai.out",
       "caps",
       "ccf.series.rwl",
       "chron",
       "chron.ars",
       "chron.ci",
       "chron.stabilized",
       "cms",
       "combine.rwl",
       "common.interval",
       "corr.rwl.seg",
       "corr.series.seg",
       "crn.plot",
       "csv2rwl",
       "delete.ring",
       "detrend",
       "detrend.series",
       "ffcsaps",
       "fill.internal.NA",
       "gini.coef",
       "glk.legacy",
       "glk",
       "hanning",
       "i.detrend",
       "i.detrend.series",
       "insert.ring",
       "interseries.cor",
       "latexDate",
       "latexify",
       "morlet",
       "net",
       "pass.filt",
       "plot.crn",
       "plot.crs",
       "plot.rwl",
       "plotRings",
       "po.to.wc",
       "pointer",
       "powt",
       "powt.series",
       "print.redfit",
       "print.rwl.report",
       "rasterPlot",
       "rcs",
       "read.compact",
       "read.crn",
       "read.fh",
       "read.ids",
       "read.rwl",
       "read.tridas",
       "read.tucson",
       "redfit",
       "runcrit",
       "rwi.stats",
       "rwi.stats.legacy",
       "rwi.stats.running",
       "rwl.report",
       "rwl.stats",
       "sea",
       "seg.plot",
       "sens1",
       "sens2",
       "series.rwl.plot",
       "sgc",
       "skel.plot",
       "spag.plot",
       "ssf",
       "sss",
       "strip.rwl",
       "summary.rwl",
       "tbrm",
       "time.rwl",
       "time<-",
       "treeMean",
       "tridas.vocabulary",
       "uuid.gen",
       "wavelet.plot",
       "wc.to.po",
       "write.compact",
       "write.crn",
       "write.rwl",
       "write.tridas",
       "write.tucson",
       "xdate.floater",
       "xskel.ccf.plot",
       "xskel.plot")

rwl.report <- function(rwl, small.thresh = NA, big.thresh = NA){
  oldw <- getOption("warn")
  options(warn = -1)
  
  # make sure input is class(rwl)
  if (!inherits(rwl, "rwl")) {
    stop('use only with "rwl" objects')
  }
  
  res <- list()
  res$small.thresh <- small.thresh
  res$big.thresh <- big.thresh
  
  
  # start with a summary
  tmp.sum <- summary.rwl(rwl)
  res$nSeries <- ncol(rwl)
  res$n <- length(rwl[!is.na(rwl)])
  res$meanSegLength <- mean(tmp.sum$year)
  res$firstYear <- min(tmp.sum$first)
  res$lastYear <- max(tmp.sum$last)
  res$meanAR1 <- mean(tmp.sum$ar1)
  res$sdAR1 <- sd(tmp.sum$ar1)
  
  # unconnected spans
  naRowSum <- apply(rwl,1,function(x) { sum(is.na(x))})
  unconnectedFlag <- naRowSum == res$nSeries
  res$unconnected <- any(unconnectedFlag)
  res$unconnectedYrs <- time(rwl)[which(unconnectedFlag)]
  
  # missing rings
  zedsLogical <- rwl == 0
  res$nZeros <- table(zedsLogical)["TRUE"] 
  zeds <- apply(zedsLogical,2,which)
  zeds <- sapply(zeds, function(x) {as.numeric(names(x))} )
  zeds <- zeds[lapply(zeds,length)>0]
  if(length(zeds)<1) res$zeros <- numeric(0)
  else res$zeros <- zeds
  
  # any years with all zeros?
  samps <- rowSums(!is.na(rwl))
  pctSeriesZero <- rowSums(zedsLogical,na.rm = TRUE)/samps
  res$allZeroYears <- which(pctSeriesZero==1)
  
  # Any places with >1 consecutive zeros?
  # Find runs of consecutive zeros
  consecutiveZerosVec <- function(x){
    # Extract lengths of runs and values using the ever confusing rle()
    runs <- rle(x == 0)
    run_lengths <- runs$lengths
    run_values <- runs$values
    
    # Identify positions where there are >1 consecutive zeros
    consecutive_zeros_indices <- which(run_values & run_lengths > 1)
    
    # Create a logical vector of length x indicating where there are >1 consecutive zeros
    consecutive_zeros_logical <- logical(length(x))
    for (i in consecutive_zeros_indices) {
      consecutive_zeros_logical[(sum(run_lengths[1:(i-1)]) + 1):(sum(run_lengths[1:i]))] <- TRUE
    }
    consecutive_zeros_logical
  }
  
  consecutiveZerosLogical <- apply(rwl,2,consecutiveZerosVec)
  rownames(consecutiveZerosLogical) <- time(rwl)
  # make a list where every series is an element
  consecutiveZerosLogicalList <- apply(consecutiveZerosLogical,2,which)
  # get years from names instead of indices
  consecutiveZerosLogicalList <- sapply(consecutiveZerosLogicalList, 
                                       function(x) {as.numeric(names(x))})
  # drop series without consec zeroes 
  mask <- lapply(consecutiveZerosLogicalList,length)>0
  consecutiveZerosLogicalList <- consecutiveZerosLogicalList[mask]
  # clean up for output
  if(length(consecutiveZerosLogicalList)<1) res$consecutiveZeros <- numeric(0)
  else res$consecutiveZeros <- consecutiveZerosLogicalList
  
  
  # check overlap of all series
  
  # interseries correlation
  res$meanInterSeriesCor <- mean(interseries.cor(rwl)[,1])
  res$sdInterSeriesCor <- sd(interseries.cor(rwl)[,1])
  
  
  # internal NA
  # this used alply from plyr. cutting it out. feb 2024
  # internalNAs <- alply(rwl, 2, find.internal.na) # like apply but forces a list
  # names(internalNAs) <- names(rwl)
  # internalNAs <- sapply(internalNAs, function(x) {as.numeric(rownames(rwl)[x])} )
  # internalNAs <- internalNAs[lapply(internalNAs,length)>0]
  internalNAs <- as.list(apply(rwl, 2, find.internal.na))
  names(internalNAs) <- names(rwl)
  internalNAs <- sapply(internalNAs, function(x) {as.numeric(names(x))} )
  internalNAs <- internalNAs[lapply(internalNAs,length)>0]
  
  if(length(internalNAs)<1) res$internalNAs <- numeric(0)
  else res$internalNAs <- internalNAs

  # small rings
  if(is.na(small.thresh)) res$smallRings <- numeric(0)
  else {
    smallRings <- rwl > 0 & rwl < small.thresh
    smallRings <- apply(smallRings,2,which)
    smallRings <- sapply(smallRings, function(x) {as.numeric(names(x))} )
    smallRings <- smallRings[lapply(smallRings,length)>0]
    if(length(smallRings)<1) res$smallRings <- numeric(0)
    else res$smallRings <- smallRings
  }

  # big rings
  if(is.na(big.thresh)) res$bigRings <- numeric(0)
  else {
    bigRings <- rwl > big.thresh
    bigRings <- apply(bigRings,2,which)
    bigRings <- sapply(bigRings, function(x) {as.numeric(names(x))} )
    bigRings <- bigRings[lapply(bigRings,length)>0]
    if(length(bigRings)<1) res$bigRings <- numeric(0)
    else res$bigRings <- bigRings
  }
  
  options(warn = oldw)
  class(res) <- "rwl.report"
  res
}

print.rwl.report <- function(x, ...){
  cat("Number of dated series:",x$nSeries,"\n")
  cat("Number of measurements:",x$n,"\n")
  cat("Number of missing (0) rings: ", x$nZeros, 
      " (", round(x$nZeros/x$n * 100, 3),"%)\n",sep="")    
  cat("Avg series length:",x$meanSegLength,"\n")
  cat("Range: ", x$lastYear - x$firstYear + 1, "\n")
  cat("Span: ",x$firstYear, "-", x$lastYear, "\n")
  cat("Mean (Std dev) series intercorrelation: ",x$meanInterSeriesCor, " (", 
      x$sdInterSeriesCor,")\n",sep="")
  cat("Mean (Std dev) AR1: ",x$meanAR1, " (", 
      x$sdAR1,")\n",sep="")
  
  cat("-------------\n")
  cat("Years where all rings are NA\n")
  if(!x$unconnected) cat("    None \n")
  else{
    cat("Warning: Having years with all NA is not standard practice and can break dplR.\n")
    cat(x$unconnectedYrs,"\n")
  }
  
  cat("-------------\n")
  cat("Years where all rings are zero\n")
  if(length(x$allZeroYears)==0) cat("    None \n")
  else{
    cat("Warning: Having years with all zeros is atypical (but not unheard of).\n")
    cat(x$allZeroYears,"\n")
  }
  
  cat("-------------\n")
  cat("Years with missing (0) rings listed by series \n")
  if(length(x$zeros)==0) cat("    None \n")
  else{
    for(i in 1:length(x$zeros)){
      tmp = x$zeros[[i]]
      if(length(tmp)==0) next()
      cat("    Series", names(x$zeros)[i],"--",tmp,"\n",  
          sep = " ")
    }
  }
  cat("-------------\n")
  cat("Years with more than one consecutive missing (0) rings listed by series \n")
  if(length(x$consecutiveZeros)==0) cat("    None \n")
  else{
    for(i in 1:length(x$consecutiveZeros)){
      tmp = x$consecutiveZeros[[i]]
      if(length(tmp)==0) next()
      cat("    Series", names(x$consecutiveZeros)[i],"--",tmp,"\n",  
          sep = " ")
    }
  }
  cat("-------------\n")
  cat("Years with internal NA values listed by series \n")
  if(length(x$internalNAs)==0) cat("    None \n")
  else{
    cat("Warning: Using internal NA values is not standard practice and can break dplR \n")
    for(i in 1:length(x$internalNAs)){
      tmp = x$internalNAs[[i]]
      if(length(tmp)==0) next()
      cat("    Series", names(x$internalNAs)[i],"--",tmp,"\n",  
          sep = " ")
    }
  }
  if(!is.na(x$small.thresh)){
    cat("-------------\n")
    cat("Years with values <", x$small.thresh, "listed by series \n")
    if(length(x$smallRings)==0) cat("    None \n")
    else{
      for(i in 1:length(x$smallRings)){
        tmp = x$smallRings[[i]]
        if(length(tmp)==0) next()
        cat("   Series", names(x$smallRings)[i],"--",tmp,"\n",  
            sep = " ")
      }
    }
  }
  if(!is.na(x$big.thresh)){
    cat("-------------\n")
    cat("Years with values >", x$big.thresh, " listed by series \n")
    if(length(x$bigRings)==0) cat("    None \n")
    else{
      for(i in 1:length(x$bigRings)){
        tmp = x$bigRings[[i]]
        if(length(tmp)==0) next()
        cat("   Series", names(x$bigRings)[i],"--",tmp,"\n",  
            sep = " ")
      }
    }
  }
}

`chron.ars` <- function(x, biweight=TRUE, maxLag=10,
                        firstAICmin=TRUE, verbose = TRUE,
                        prewhitenMethod=c("ar.yw","arima.CSS-ML")){
  # helpers
  # this is a time killer. needs to be vectorized.
  pooledAR <- function(x, maxLag=maxLag, firstAICmin=TRUE){
    
    nYrs <- dim(x)[1] # num years
    nSeries <- dim(x)[2] # num series
    
    # remove the mean of each series (yes this does columns, I checked.)
    x <- scale(x, center = TRUE, scale = FALSE)
    
    # Init the product sum structure. Doesn't need to be a matrix really
    productSumMat <- matrix(0,maxLag+1,1)
    
    # start at lag 0 (index 1 is lag0)
    # i renamed the loops since (in my mind) lags are always k
    # should do a c implementation of this for speed. 
    # Takes forever! Should be possible to vectorize. Need
    # to build a single object of all the pairs with lags.
    for(k in 0:maxLag){
      for(j in 1:nSeries){
        for(i in 1:nSeries){
          series <- cbind(x[,j],x[,i])
          # Common data interval between series
          mask <- which(rowSums(!is.na(series))==2)
          # if no overlap, bail out
          if(length(mask)==0) { next }
          # Subset series to mask
          series1 <- x[mask,j]
          series2 <- x[mask,i]
          # Lag the i'th series (series2) by k lags
          # check overlap to avoid negative subscripts if
          # lagged series2 has would have length 0
          if(length(series2)<=k) { next } 
          series2 <- series2[1:(length(series2)-k)] ## check overlap
          series2 <- c(rep(NA,k),series2)
          # Find period where series overlap again.
          mask2 <- which(rowSums(!is.na(cbind(series1,series2)))==2)
          # if no overlap, bail out
          if(length(mask2)==0) { next }
          # And add the pairwise product sum to the matrix.
          productSumMat[k+1,1] = productSumMat[k+1,1] +
            t(series1[mask2]) %*% series2[mask2]
        }
      }
    }
    productSumMat <- t(productSumMat)
    r0 <- productSumMat[1,1]
    xAcf <- productSumMat / r0
    A <- acf2AR(xAcf) * -1
    vp_shortcut <- c(r0,c(r0 - -1*A %*% matrix(t(productSumMat)[-1,])))
    aic2 <- nYrs * log(vp_shortcut) + 2.0 * 1:(maxLag+1)
    
    # names
    names(aic2) <- paste0("ar(",0:maxLag,")")
    xAcf <- xAcf[1,]
    names(xAcf) <- paste0("ar(",0:maxLag,")")
    
    getFirstMin <- function(y){
      if(y[1] < y[2]) out <- 1
      else {
        increasing <- which(diff(y) > 0)
        out <- if(length(increasing) == 0) Inf else min(increasing)
      }
      return(out)
    }
    if(firstAICmin){
      orderOut <- getFirstMin(aic2) - 1 #for lag 0
      if(!is.finite(orderOut)){
        stop("The pooled AIC did not reach a minimum within maxLag (",
             maxLag, ") lags. The selected AR order cannot be determined.\n",
             "  Consider increasing maxLag.\n",
             "  If the problem persists, the data may not be well-suited to ",
             "the pooled autoregressive approach. Highly persistent or ",
             "non-stationary series can prevent the AIC from reaching a ",
             "minimum within a reasonable lag range. See Cook (1985) for ",
             "guidance on appropriate data preparation.",
             call. = FALSE)
      }
    } else {
      orderOut <- which.min(aic2)  - 1
    }
    
    res <-list(ACF = xAcf,ARcoefs = A,aic=aic2, orderOut = orderOut)
    return(res)
  }
  
  # AGB Nov 2018. Porting KJA's postredden func.
  # This needs vectorizing but isn't the time killer that
  # the AR pooling func is.
  
  postAR <- function(x,phi){
    # STEPS:
    #  1. the original, (pre)white(ned) series is flipped around
    #  2. the working series is created as a combination of null initial values
    #     and the original series
    #  3. A double nest applies the autoregressive model as a function of value
    #     (e.g. per year) and autoregressive order (e.g. per coefficient)
    #  4. The first estimate of the post-reddened series drops the initial
    #     values
    #  5. A backcasting step develops a better initial value estimate by using
    #     the first estimate from step 4
    #  6. New initial values are taken from the backcast estimate
    #  7. Post-reddening is again applied again to the initial series, now with the
    #     improved initial value estimates
    
    mask <- is.na(x)
    x0 <- x # make a copy
    # AR(0): no persistence to reintroduce, return the series unchanged.
    if(length(phi) == 0) return(x0)
    x <- x[!mask]
    xMean <- mean(x)
    x <- x - xMean
    
    nx <- length(x)
    nPhi <- length(phi)
    # step 1
    xRev <- rev(x)
    # step 2
    xRevPad <- c(rep(0, nPhi), xRev)
    
    # step 3 -- vectorize this and the other loops.
    for(i in 1:nx) {
      for(j in 1:nPhi) {
        xRevPad[i + nPhi] <-  xRevPad[i + nPhi] + phi[j] * xRevPad[i + nPhi - j]
      }
    }
    # step 4 -- xRevInit is the first pass on reddening xRev
    xRevInit <- xRevPad[-c(1:nPhi)]
    
    # step 5 -- do it again. This backcast gets better initial values
    for(i in 1:nPhi) {
      cntr <- nPhi - i + 1
      xRevInit[cntr] <- 0 # get first few obs.
      for(j in 1:nPhi) {
        xRevInit[cntr] <- xRevInit[cntr] + phi[j] * xRevInit[cntr + j]
      }
    }
    
    # step 6&7 -- Use new init values from xRevInit. Then staple x on and apply the
    # AR model. With the new init values we can now roll forward now.
    xAR <- c(xRevInit[1:nPhi],x)
    
    for(i in 1:length(xRevInit)) {
      for(j in 1:nPhi) {
        xAR[i + nPhi] <-  xAR[i + nPhi] + phi[j] * xAR[i + nPhi - j]
      }
    }
    # chop off the init values (they were estimates) and return
    xAR <- xAR[-c(1:nPhi)]
    x0[!mask] <- xAR + xMean
    return(x0)
  }
  
  # Uses ar and defaults to yule-walker
  prewhitenAR <- function(x,p){
    # AR(0): residuals of a mean-only model are the series itself.
    if(p == 0) return(x)
    mask <- is.na(x)
    x2 <- x[!mask]
    out <- ar(x2,aic = FALSE, order.max = p,demean=TRUE)
    x[!mask] <- out$resid + mean(x2)
    return(x)
  }
  
  # Uses arima and CSS-ML. this is slower. requires optim
  prewhitenARIMA <- function(x,p){
    # AR(0): residuals of a mean-only model are the series itself.
    if(p == 0) return(x)
    mask <- is.na(x)
    x2 <- x[!mask]
    out <- arima(x2,order = c(p,0,0),method = "CSS-ML", #"ML",
                 optim.control = list(maxit = 1000))
    x[!mask] <- out$residuals + mean(x2)
    return(x)
  }
  # end helpers
  
  known.prewhitenMethods <- c("ar.yw","arima.CSS-ML")
  prewhitenMethod2 <- match.arg(arg = prewhitenMethod,
                                choices = known.prewhitenMethods,
                                several.ok = FALSE)
  
  
  samps <- rowSums(!is.na(x))
  
  # calc std chronology
  if(!biweight) {
    stdCrn <- rowMeans(x, na.rm=TRUE)
  } else {
    stdCrn <- apply(x, 1, tbrm, C=9)
  }
  
  ### Get the pooled ACF and AR coefs from the RWI
  outAR <- pooledAR(x,firstAICmin = firstAICmin, maxLag = maxLag)
  # model order
  p <- outAR$orderOut
  # do some verbose output here
  if(verbose){
    cat("Pooled AR Summary\n")
    cat("ACF\n")
    print(outAR$ACF)
    cat("AR Coefs\n")
    print(outAR$ARcoefs)
    cat("AIC\n")
    print(outAR$aic,digits=4)
    cat("Selected Order\n")
    print(outAR$orderOut)
  }
  
  ###  Prewhiten each RWI series individually using the model order
  if(prewhitenMethod2 == "ar.yw") {
    RWIclean <- apply(x,2,prewhitenAR,p=p)  
  }
  if(prewhitenMethod2 == "arima.CSS-ML") {
    RWIclean <- apply(x,2,prewhitenARIMA,p=p)  
  }
  
  
  ### Make a chron using the prewhitened series (RES)
  # Note that the final RES chron needs to be prewhitened again out to `p`.
  # See pp 153-154 in Ed's dissertation.
  if(!biweight) {
    if(prewhitenMethod2 == "ar.yw") {
      resCrn <- prewhitenAR(rowMeans(RWIclean, na.rm=TRUE),p=p)
    }
    if(prewhitenMethod2 == "arima.CSS-ML") {
      resCrn <- prewhitenARIMA(rowMeans(RWIclean, na.rm=TRUE),p=p)  
    }
    
  } else {
    if(prewhitenMethod2 == "ar.yw") {
      resCrn <- prewhitenAR(apply(RWIclean,1,tbrm, C=9),p=p)
    }
    if(prewhitenMethod2 == "arima.CSS-ML") {
      resCrn <- prewhitenARIMA(apply(RWIclean,1,tbrm, C=9),p=p)
    }
    
  }
  
  ### Post-redden the prewhitened series
  # When p == 0, ARcoefs[0, ][1:0] does not return numeric(0) cleanly in R
  # (1:0 is c(1,0) and indexing a zero-row matrix with that gives NA).
  # Guard explicitly so postAR receives a proper zero-length vector.
  phi <- if(p > 0) outAR$ARcoefs[p,][1:p] else numeric(0)
  RWIar <- apply(RWIclean,2,postAR,-phi)
  
  
  ### 10. Make the ARSTAN chronology (ARS)
  if(!biweight) {
    arsCrn <- rowMeans(RWIar, na.rm=TRUE)
  } else {
    arsCrn <- apply(RWIar, 1, tbrm, C=9)
  }
  
  out <- data.frame(std=stdCrn,res=resCrn,
                    ars=arsCrn,samp.depth=samps)
  row.names(out) <- row.names(x)
  class(out) <- c("crn", "data.frame")
  return(out)
}

`skel.plot` <-
    function(rw.vec, yr.vec = NULL, sname = "", filt.weight = 9,
             dat.out = FALSE, master=FALSE, plot=TRUE)
{
    if (length(sname) == 0) {
        sname2 <- ""
    } else {
        sname2 <- as.character(sname)[1]
        if (is.na(sname2)) {
            sname2 <- "NA"
        } else if (Encoding(sname2) == "bytes" ||
                   nchar(sname2, type = "width") > 7) {
            stop("'sname' must be a character string whose width is less than 8")
        }
    }

    ## what about NA. Internal NA?
    na.mask <- is.na(rw.vec)
    rw.vec2 <- rw.vec[!na.mask]

    n.val <- length(rw.vec2)
    if(n.val > 840) {
        cat(gettextf("input series has length of %d\n", n.val))
        stop("long series (> 840) must be split into multiple plots")
    }
    stopifnot(filt.weight >= 3)
    if(n.val < filt.weight) {
        cat(gettextf("input series has length of %d", n.val),
            gettextf("'filt.weight' is %f\n", filt.weight), sep=", ")
        stop("'filt.weight' must not be larger than length of input series")
    }

    ## should wrap this into a function called skel.calc that returns the
    ## dates and skel

    ## if no yr then....
    if(is.null(yr.vec))
        yr.vec2 <- 0:(n.val - 1)
    else
        yr.vec2 <- yr.vec[!na.mask]
    ## pad down to the nearest 10 if not already there
    pad0 <- floor(min(yr.vec2) / 10) * 10
    if(pad0 != min(yr.vec2)){
        pad.length <- min(yr.vec2) - pad0
        rw.df <- data.frame(rw = rep(NA, pad.length),
                            yr = pad0:(pad0 + pad.length - 1))
        rw.df <- rbind(rw.df, data.frame(rw = rw.vec2, yr = yr.vec2))
    }
    else {
        pad.length <- 0
        rw.df <- data.frame(rw = rw.vec2, yr = yr.vec2)
    }

    ## detrend and pad
    rwRw <- rw.df[["rw"]]
    rw.dt <- hanning(rwRw, filt.weight)
    skel <- rep(NA, length(rwRw))
    ## calc rel growth
    n.diff <- length(rwRw) - 1
    idx <- 2:n.diff
    temp.diff <- diff(rwRw)
    skel[idx] <- rowMeans(cbind(temp.diff[-n.diff],
                                -temp.diff[-1])) / rw.dt[idx]
    skel[skel > 0] <- NA
    ## rescale from 0 to 10
    na.flag <- is.na(skel)
    if(all(na.flag))
        skel.range <- c(NA, NA)
    else
        skel.range <- range(skel[!na.flag])
    newrange <- c(10, 1)
    mult.scalar <-
        (newrange[2] - newrange[1]) / (skel.range[2] - skel.range[1])
    skel <- newrange[1] + (skel - skel.range[1]) * mult.scalar
    skel[skel < 3] <- NA
    skel <- ceiling(skel)

    ## Variables for plotting
    ## page width
    pw <- 278
    ## page height
    ph <- 178
    ## row height
    rh <- 22
    ## row width
    rw <- 240
    ## spacer for text and dashed cutting lines
    spcr <- 5

    ## break series into sections of 120 years with an index
    yrs.col <- rw / 2 # n years per row
    n <- length(skel)
    n.rows <- ceiling(n / yrs.col)
    m <- seq_len(n.rows)
    skel.df <- data.frame(yr=rw.df[["yr"]], skel)
    if(plot){
        ## master page
        grid.newpage()
        vps <- vector(mode = "list", length = n.rows)
        y <- ph
        for (i in m) {
            y <- y - (rh + spcr)
            vps[[i]] <-
                viewport(x=unit(19, "mm"), y=unit(y, "mm"),
                         width=unit(246, "mm"), height=unit(rh, "mm"),
                         just=c("left", "bottom"), name=LETTERS[i])
        }

        ## seq for 0 to plot width by 2mm
        tmp.1 <- seq(from=0, to=rw, by=2)
        tmp.2 <- seq(from=0, to=rh, by=2)
        ticks <- seq(from=0, to=rw, by=20)
        vSegments <-
            segmentsGrob(x0 = tmp.1, y0 = 0, x1 = tmp.1, y1 = rh,
                         default.units = "mm",
                         gp = gpar(col="green", lineend = "square",
                         linejoin = "round"))
        hSegments <-
            segmentsGrob(x0 = 0, y0 = tmp.2, x1 = rw, y1 = tmp.2,
                         default.units = "mm",
                         gp = gpar(col="green", lineend = "square",
                         linejoin = "round"))
        ## decadal lines
        decades <-
            segmentsGrob(x0 = ticks, y0 = 0, x1 = ticks, y1 = rh,
                         default.units = "mm",
                         gp = gpar(col = "black", lwd = 1.5, lty = "dashed",
                         lineend = "square", linejoin = "round"))
        ## lines on top and bottom of plot
        topLine <-
            linesGrob(x = c(0, rw), y = c(rh, rh),
                      default.units = "mm",
                      gp = gpar(lwd = 2, lineend = "square",
                      linejoin = "round"))
        bottomLine <-
            linesGrob(x = c(0, rw), y = c(0, 0),
                      default.units = "mm",
                      gp = gpar(lwd = 2, lineend = "square",
                      linejoin = "round"))
        rowTree <- grobTree(vSegments, hSegments, decades, topLine, bottomLine)
        if(!master){
            yy1 <- c(0, 6, 6)
            yy2 <- rh - 1
            sjust <- c("right", "bottom")
            yrjust <- c("center", "bottom")
            yry <- rh + 0.5
        }
        else{
            yy1 <- c(rh, 16, 16)
            yy2 <- 1
            sjust <- c("left", "bottom")
            yrjust <- c("center", "top")
            yry <- rh - 22.5
        }
        ## set up page with the right number of rows
        dev.hold()
        on.exit(dev.flush())
        pushViewport(vpTree(viewport(width=unit(pw, "mm"), y=1, just="top",
                                     height=unit(ph, "mm"), name="page",
                                     clip="off"),
                            do.call(vpList, vps)))
        row.last <- 0
        for (i in m) {

            seekViewport(LETTERS[i])
            ## working code goes here - e.g., skelplot!
            grid.draw(rowTree)

            ## plot x axis
            ## get this row's data
            row.first <- row.last + 1
            row.last <- min(row.first + (yrs.col - 1), n)
            skel.sub <- skel.df[row.first:row.last, ]
            skelYr <- skel.sub[["yr"]]
            skel2 <- skel.sub[["skel"]]
            end.yr <- length(skelYr)
            init.lab <- min(skelYr)
            x.labs <- seq(from=init.lab, length.out = length(ticks), by=10)
            grid.text(label = x.labs, x = ticks, y = yry,
                      default.units = "mm", just = yrjust,
                      gp = gpar(fontsize=10))
            ## plot data
            notNA <- which(!is.na(skel2))
            if (length(notNA) > 0) {
                xx <- (notNA - 1) * 2
                if (!master) {
                    y0 <- 0
                    y1 <- 2 * skel2[notNA]
                } else {
                    y0 <- 22
                    y1 <- 22 - 2 * skel2[notNA]
                }
                grid.segments(x0 = xx, x1 = xx, y0 = y0, y1 = y1,
                              default.units = "mm",
                              gp = gpar(col = "black", lwd = 2,
                              lineend = "square", linejoin = "round"))
            }
        }
        ## end arrow
        end.mm <- (end.yr - 1) * 2
        grid.lines(x=unit(c(end.mm, end.mm), "mm"), y=unit(c(rh, 0), "mm"),
                   gp = gpar(lwd = 2, lineend = "square", linejoin = "round"))
        grid.polygon(x = c(end.mm, end.mm, end.mm + 2), y = yy1,
                     gp = gpar(fill = "black", lineend = "square",
                     linejoin = "round"), default.units = "mm")
        ## start arrow and sample id
        seekViewport(LETTERS[1])
        start.mm <- pad.length * 2
        grid.lines(x=unit(c(start.mm, start.mm), "mm"),
                   y=unit(c(rh, 0), "mm"),
                   gp = gpar(lwd = 2, lineend = "square", linejoin = "round"))
        fontsize.sname <-
            if (nchar(sname2, type = "width") > 6) {
                9
            } else {
                10
            }
        grid.polygon(x = c(start.mm, start.mm, start.mm - 2),
                     y = yy1, default.units = "mm",
                     gp=gpar(fill = "black", lineend = "square",
                     linejoin = "round"))
        grid.text(label = sname2, x = start.mm - 1, y = yy2,
                  just = sjust, rot = 90, default.units = "mm",
                  gp = gpar(fontsize=fontsize.sname))
        popViewport()
        for (i in seq(from = 2, by = 1, length.out = n.rows - 1)) {
            seekViewport(LETTERS[i])
            popViewport()
        }
        popViewport()
    }
    if (dat.out) {
        skel.df
    }
}

`csv2rwl` <- function(fname,...)
{
  # check for unique names
  nameCheck <- scan(fname,nlines=1,what = "char",sep=",")
  if(any(duplicated(nameCheck))){
    warn.fmt <- gettext("Duplicated series ID detected: %s",domain="R-dplR")
    ids.dup <- paste(nameCheck[which(duplicated(nameCheck))],
                     collapse = "; ")
    stop(sprintf(warn.fmt, ids.dup), domain=NA) 
  }
  else{
    dat <- read.table(fname, header=TRUE, sep=",",check.names=TRUE,...)
    rownames(dat) <- as.character(dat[,1])
    dat <- dat[,-1]
    class(dat) <- c("rwl","data.frame")
    return(dat)
  }
}

insert.ring <- function(rw.vec, rw.vec.yrs=as.numeric(names(rw.vec)),
                        year, ring.value=mean(rw.vec,na.rm=TRUE),
                        fix.last=TRUE, fix.length=TRUE) {
    n <- length(rw.vec)
    stopifnot(is.numeric(ring.value), length(ring.value) == 1,
              is.finite(ring.value), ring.value >= 0,
              is.numeric(year), length(year) == 1, is.finite(year),
              n > 0, length(rw.vec.yrs) == n,
              identical(fix.last, TRUE) || identical(fix.last, FALSE))
    first.yr <- rw.vec.yrs[1]
    last.yr <- rw.vec.yrs[n]
    if (!is.finite(first.yr) || !is.finite(last.yr) ||
        round(first.yr) != first.yr || last.yr - first.yr != n - 1) {
        ## Basic sanity check, _not_ a full test of consecutive years
        stop("input data must have consecutive years in increasing order")
    }
    if (year == first.yr - 1) {
        year.index <- 0
    } else {
        year.index <- which(rw.vec.yrs == year)
    }
    if (length(year.index) == 1) {
        rw.vec2 <- c(rw.vec[seq_len(year.index)],
                     ring.value,
                     rw.vec[seq(from = year.index+1, by = 1,
                                length.out = n - year.index)])
        if (fix.last) {
            names(rw.vec2) <- (first.yr-1):last.yr
        } else {
            names(rw.vec2) <- first.yr:(last.yr+1)
        }
        if(fix.last & fix.length){
          rw.vec2 <- rw.vec2[-1]
        }
        
        if(!fix.last & fix.length){
          rw.vec2 <- rw.vec2[-length(rw.vec2)]
        }
        rw.vec2
    } else {
        stop("invalid 'year': skipping years not allowed")
    }
}

delete.ring <- function(rw.vec, rw.vec.yrs=as.numeric(names(rw.vec)),
                        year, fix.last=TRUE, fix.length=TRUE) {
    n <- length(rw.vec)
    stopifnot(is.numeric(year), length(year) == 1, is.finite(year),
              n > 0, length(rw.vec.yrs) == n,
              identical(fix.last, TRUE) || identical(fix.last, FALSE))
    first.yr <- rw.vec.yrs[1]
    last.yr <- rw.vec.yrs[n]
    if (!is.finite(first.yr) || !is.finite(last.yr) ||
        round(first.yr) != first.yr || last.yr - first.yr != n - 1) {
        ## Basic sanity check, _not_ a full test of consecutive years
        stop("input data must have consecutive years in increasing order")
    }
    year.index <- which(rw.vec.yrs == year)
    if (length(year.index) == 1) {
        rw.vec2 <- rw.vec[-year.index]
        if (n > 1) {
            if (fix.last) {
                names(rw.vec2) <- (first.yr+1):last.yr
            } else {
                names(rw.vec2) <- first.yr:(last.yr-1)
            }
        }
        if(fix.last & fix.length){
          rw.vec2 <- c(NA,rw.vec2)
        }
        
        if(!fix.last & fix.length){
          rw.vec2 <- c(rw.vec2,NA)
        }
        rw.vec2
    } else {
        stop("'year' not present in 'rw.vec.yrs'")
    }
}

`crn.plot` <- function(x,add.spline=FALSE, 
                       nyrs=NULL,
                       ...) { 
  warning("crn.plot has been deprecated to plot.crn for simplicity and consistency.")
  plot.crn(x=x,...) 
}

`plot.crn` <- function(x,
                       add.spline=FALSE, 
                       nyrs=NULL,
                       ...) {
  crn <- x
  if("ssfLong" %in% comment(x)){
    x <- x$ssfCrn
  }
  
  if(!("crn" %in% class(crn))) stop("'x' must be class crn")
  
  ## Check to see if the crn has ci (from chron.ci)
  ci.exist <- "ci" %in% comment(crn)
  
  if(ci.exist){
    warning("Objects from chron.ci are plotted without confidence intervals.\n  See help file for chron.ci for more flexible plotting options.")
    crn$lowerCI <- NULL
    crn$upperCI <- NULL
  }
  
  crn.names <- names(crn)
  ## Check to see if the crn has sample depth
  sd.exist <- "samp.depth" %in% crn.names
  ## Check to see if the crn has running rbar (from chron.stablized)
  rbar.exist <- "running.rbar" %in% crn.names
  
  yr.vec <- as.numeric(row.names(crn))
  crn.names <- names(crn)
  known.crns <- c(std="Standard", res="Residual", ars="Arstan",sfc = "Simple Signal Free",
                  vsc = "Variance Stabilized",
                  samp.depth = "Sample Depth",running.rbar="Running rbar")
  
  # get better names for the columns
  crn.names.long <- known.crns[crn.names]
  crn.names.long[is.na(crn.names.long)] <- crn.names[is.na(crn.names.long)]
  names(crn.names.long) <- NULL
  # add units -- could do with attr?
  known.units <- c(std="RWI", res="RWI", ars="RWI",sfc = "RWI",
                   vsc = "RWI",
                   samp.depth = "n",running.rbar="r")
  ylabs <- known.units[crn.names]
  ylabs[is.na(ylabs)] <- NULL
  names(ylabs) <- NULL
  
  # calc ylims -- make same for all RWI?
  ranges <- apply(crn,MARGIN = 2,range,na.rm=TRUE)
  if("samp.depth" %in% colnames(ranges)){
    ranges[1,which(colnames(ranges) == "samp.depth")] <- 0
  }
  if("running.rbar" %in% colnames(ranges)){
    ranges[,which(colnames(ranges) == "running.rbar")] <- c(0,1)
  }
  # make all the RWI ranges the same
  ranges[,ylabs=="RWI"] <- range(ranges[,ylabs=="RWI"])
  
  nCrn <- ncol(crn)
  
  # args0 <- list(...)
  # args1 <- args0
  # args1[c("x", "y", "type", "axes", "xlab", "col")] <-
  #   list(yr.vec, as.name("y"), "l", FALSE, "", "grey50")
  # print(args1)
  args0 <- list(x=yr.vec, y=as.name("y"), type="l", axes=FALSE, 
                xlab="", col="grey50")  
  
  op <- par(no.readonly=TRUE) # Save par
  on.exit(par(op))            # Reset par on exit
  layout(matrix(seq_len(nCrn), nrow=nCrn, ncol=1))
  par(mgp=c(1.1, 0.1, 0),tcl=0.5, xaxs='i')
  
  nyrs2 <- nyrs  
  for(i in seq_len(nCrn)){
    
    if(i==1) {
      par(mar=c(0.1, 3, 3, 3))
    }
    else if(i==nCrn) {
      par(mar=c(3, 3, 0.1, 3))
    }
    else{
      par(mar=c(0.1, 3, 0.1, 3))
    }
    
    y <- crn[[i]]
    argsTmp <- args0
    argsTmp$ylab <- ylabs[i]
    argsTmp$ylim <- ranges[,i]
    do.call("plot", argsTmp)
    legend("topleft",legend=crn.names.long[i],bty = "n")
    # add abline for all RWI
    
    tmp <- na.omit(y)
    if(add.spline & ylabs[i] == "RWI") {
      ## Only possibly NULL in the first round of the for loop
      if(is.null(nyrs2)) nyrs2 <- length(tmp)*0.33
      y[!is.na(y)] <- caps(y=tmp, nyrs=nyrs2)
      lines(yr.vec, y, col="darkred", lwd=1.75)
    }
    if(ylabs[i]=="RWI"){
      abline(h = 1)
    }
    axis(2)
    if(i==nCrn) axis(1)
    if(i==1) axis(3)
    box()
  }
  
}

qa.xdate <- function(rwl, seg.length, n, bin.floor){
    if(!is.data.frame(rwl))
        stop("'rwl' must be a data.frame")
    ## seg.length
    if(seg.length*2 > nrow(rwl))
        stop("'seg.length' can be at most 1/2 the number of years in 'rwl'")
    if(as.logical(seg.length %% 2)) stop("'seg.length' must be even")
    if(!is.int(seg.length)) stop("'seg.length' and 'seg.lag' must be integers")
    if(seg.length <= 0) stop("'seg.length' must be positive")
    ## n
    if(!is.null(n)){
        if(!is.int(n)) stop("'n' must be an integer")
        if(!as.logical(n %% 2)) stop("'n' must be odd")
        if(n <= 3) stop("'n' must be larger than 3")
    }
    ## bin.floor
    if(!is.null(bin.floor) && (!is.int(bin.floor) || bin.floor < 0))
        stop("'bin.floor' must be a non-negative integer")
}

morlet <- function(y1, x1=seq_along(y1), p2=NULL, dj=0.25, siglvl=0.95){

    morlet.func <- function(k0=6, Scale, k) {
        n <- length(k)
        expnt <- -(Scale * k - k0) ^ 2 / 2 * as.numeric(k > 0)
        Dt <- 2 * pi / (n * k[2])
        norm <- sqrt(2 * pi * Scale / Dt) * (pi ^ (-0.25)) #  total energy=N   [Eqn(7)]

        morlet <- norm * exp(ifelse(expnt > -100, expnt, 100))
        morlet <- morlet * (as.numeric(expnt > -100))  # Avoid underflow errors
        morlet <- morlet * (as.numeric(k > 0))  # Heaviside step function (Morlet is complex)
        fourier_factor <-
            (4 * pi) / (k0 + sqrt(2 + k0 ^ 2)) # Scale-->Fourier [Sec.3h]
        period <- Scale * fourier_factor
        coi <- fourier_factor / sqrt(2)   # Cone-of-influence [Sec.3g]
        ## dofmin = 2   # Degrees of freedom with no smoothing
        ## Cdelta = -1
        ## if(k0 == 6) Cdelta = 0.776 # Reconstruction factor
        ## psi0 = pi^(-0.25)
        list(psi_fft = morlet, period = period, coi = coi)
    }

    ## Construct optional inputs, these could be passed in as args
    Dt <- 1
    s0 <- 1
    pad <- TRUE
    lag1 <- 0
    do_daughter <- TRUE
    fft_theor <- NULL
    ## mother="morlet"
    ## do_coi=TRUE
    ## do_signif=TRUE
    ## dof=NULL
    ## global=NULL

    ## r = 0
    n <- length(y1)
    stopifnot(is.numeric(dj), is.numeric(siglvl), length(dj) == 1,
              length(siglvl) == 1, is.numeric(x1), is.numeric(y1),
              is.null(p2) || (is.numeric(p2) && length(p2) == 1),
              n > 0)
    if(length(x1) != length(y1)) stop("'x1' and 'y1' lengths differ")
    n1 <- n
    base2 <- trunc(log2(n) + 0.4999)   # power of 2 nearest to N
    if(is.null(p2)) J <- trunc(log2(n * Dt / s0) / dj) # [Eqn(10)]
    else J <- p2 / dj

    if(is.null(lag1)){
        ## Estimate lag-1 autocorrelation, for red-noise significance
        ## tests.  Note that we can actually use the global wavelet
        ## spectrum (GWS) for the significance tests, but if you
        ## wanted to use red noise, here is how you could calculate
        ## it...
        lag1 <- acf(y1, lag.max = 2, plot = FALSE)$acf[2]
    }
    else lag1 <- lag1[1]

    ## Construct time series to analyze, pad if necessary
    ypad <- y1 - mean(y1)  # Remove mean
    if(pad){ # Pad with extra zeroes, up to power of 2
        ypad <- c(ypad, rep(0, 2 ^ (base2 + 1) - n))
        n <- length(ypad)
    }
    ## Construct SCALE array & empty PERIOD & WAVE arrays
    na <- J + 1            # Num of scales
    Scale <- seq(from=0, to=na - 1) * dj    # Array of j-values
    Scale <- 2 ^ (Scale) * s0       # Array of scales  2^j   [Eqn(9)]
    period <- rep(0, na)        # Empty period array (filled in below)
    wave <- matrix(complex(), n, na) # Empty wavelet array
    if(do_daughter) daughter <- wave # Empty daughter array
    ## Construct wavenumber array used in transform [Eqn(5)]
    k <- 2 * pi / (n * Dt) * seq_len(n / 2)
    k <- c(0, k, -rev(k[-length(k)]))
    ## Compute FFT of the (padded) time series
    yfft <- fft(ypad) / length(ypad) # [Eqn(3)]
    if(length(fft_theor) == n)
        fft_theor_k <- fft_theor
    else
        fft_theor_k <-
            (1 - lag1 ^ 2) / (1 - 2 * lag1 * cos(k * Dt) + lag1 ^ 2) # [Eqn(16)]
    fft_theor <- rep(0, na)
    for(a1 in seq_len(na)) { # Scale
        morlet.out <- morlet.func(Scale=Scale[a1], k=k) #,period1,coi,dofmin,Cdelta,psi0)
        psi_fft <- morlet.out$psi_fft
        coi <- morlet.out$coi # One value per scale
        wave[, a1] <- fft(yfft * psi_fft, inverse=TRUE)
        if(do_daughter) daughter[, a1] <- fft(psi_fft, inverse=TRUE)
        period[a1] <- morlet.out$period   # Save period
        fft_theor[a1] <- sum((abs(psi_fft) ^ 2) * fft_theor_k) / n
    }
    time.scalar <- c(seq_len(floor(n1 + 1) / 2),
                     seq.int(from=floor(n1 / 2), to=1, by=-1)) * Dt
    coi <- coi * time.scalar
    if(do_daughter) { # Shift so DAUGHTERs are in middle of array
        daughter <-
            rbind(daughter[(n - n1 / 2):nrow(daughter), , drop=FALSE],
                  daughter[seq_len(n1 / 2 - 1), , drop=FALSE])
    }
    ## Significance levels [Sec.4]
    Var <- var(y1) # Variance (T&C call this sdev in their code)
    fft_theor <- Var * fft_theor  # Include time-series variance
    dof <- 2
    Signif <- fft_theor * qchisq(siglvl, dof) / dof   # [Eqn(18)]

    wave2 <- wave[seq_len(n1), , drop=FALSE]
    Power <- abs(wave2)
    Power <- Power * Power  # Compute wavelet power spectrum

    ## Done
    list(y=y1, x=x1, wave = wave2, coi = coi,
         period = period, Scale = Scale, Signif = Signif, Power = Power,
         siglvl = siglvl)
}

write.rwl <-
    function(rwl.df, fname, format=c("tucson", "compact", "tridas"), ...)
{
    ## NOTE: This function is documented to return fname.  Therefore,
    ## each branch of the switch must return fname.
    switch(match.arg(format),
           tucson = write.tucson(rwl.df, fname, ...),
           compact = write.compact(rwl.df, fname, ...),
           tridas = write.tridas(rwl.df, fname, ...))
}

write.compact <- function(rwl.df, fname, append=FALSE, prec=0.01,
                          mapping.fname="", mapping.append=FALSE, ...) {
    line.term <- "\x0D\x0A" # CR+LF, ASCII carriage return and line feed
    if (!is.data.frame(rwl.df)) {
        stop("'rwl.df' must be a data.frame")
    }
    if (!is.numeric(prec) || length(prec) != 1 || is.na(prec) ||
        !(prec == 0.01 || prec == 0.001)) {
        stop("'prec' must equal 0.01 or 0.001")
    }
    if (append && !file.exists(fname)) {
        stop(gettextf("file %s does not exist, cannot append", fname))
    }

    ## Loop through series and write each one
    nseries <- ncol(rwl.df)
    yrs.all <- row.names(rwl.df)

    line.width <- 80 # max line width
    prec.rproc <- if (prec == 0.01) 100 else 1000 # reciprocal of precision
    max.field.width.width <-
        nchar(nchar(round(max(rwl.df, na.rm=TRUE) * prec.rproc)))
    max.n.width <- nchar(nrow(rwl.df)) # conservative
    max.i.width <- max(nchar(yrs.all)) # conservative
    ## Conservative length limit for the name of each series
    name.width <-
        line.width - max.field.width.width - max.n.width - max.i.width - 17

    col.names <- names(rwl.df)
    stopifnot(is.character(col.names), !is.na(col.names),
              Encoding(col.names) != "bytes")
    col.names <- fix.names(x=col.names, limit=name.width,
                           mapping.fname=mapping.fname,
                           mapping.append=mapping.append, basic.charset=TRUE)

    ## Sort years using increasing order, reorder rwl.df accordingly
    yrs.all <- as.numeric(yrs.all)
    yrs.order <- sort.list(yrs.all)
    yrs.all <- yrs.all[yrs.order]
    rwl.df2 <- rwl.df[yrs.order, , drop=FALSE]

    if (append) {
        rwl.out <- file(fname, "a")
    } else {
        rwl.out <- file(fname, "w")
    }
    on.exit(close(rwl.out))
    missing.str <- 0

    for (l in seq_len(nseries)) {
        series <- rwl.df2[[l]]
        idx <- !is.na(series)
        yrs <- yrs.all[idx]
        series <- round(prec.rproc * series[idx])

        min.year <- min(yrs)
        max.year <- max(yrs)
        nyrs <- max.year - min.year + 1

        rwl.df.name <- col.names[l]

        ## Find missing data.
        missing.years <- setdiff(min.year:max.year, yrs)
        ## Mark missing data.
        if (length(missing.years) > 0) {
            yrs <- c(yrs, missing.years)
            series <- c(series, rep(missing.str, times=length(missing.years)))
            series.order <- sort.list(yrs)
            yrs <- yrs[series.order]
            series <- series[series.order]
        }
        ## Find negative values and mark as missing data
        series[series < 0] <- missing.str

        series <- as.character(series)
        field.width <- max(nchar(series))
        n.fields <- floor(line.width / field.width)
        n.lines <- floor(nyrs / n.fields)
        remainder <- nyrs - n.lines * n.fields

        ## Write header
        head1 <- paste0(nyrs, "=N", " ", min.year, "=I", " ")
        head2 <- paste0(if (prec == 0.01) "-2" else "-3", "(",
                        n.fields, "F", field.width, ".0)~")
        n.space <- line.width - nchar(head1) - nchar(head2) - nchar(rwl.df.name)
        if (n.space < 1) {
            ## since names are cut to length, this should not happen
            stop(gettextf("series %s: header line would be too long",
                          rwl.df.name))
        }
        cat(head1, rwl.df.name, rep(" ", n.space), head2, line.term,
            file=rwl.out, sep="")

        ## Write full lines
        for (i in seq_len(n.lines)) {
            end.idx <- i * n.fields
            line.str <- sprintf("%*s", field.width,
                                series[(end.idx - n.fields + 1) : end.idx])
            cat(line.str, line.term, file=rwl.out, sep="")
        }
        ## Write possibly remaining shorter line
        if (remainder > 0) {
            end.idx <- length(series)
            line.str <- sprintf("%*s", field.width,
                                series[(end.idx - remainder + 1) : end.idx])
            cat(line.str, line.term, file=rwl.out, sep="")
        }
    }
    fname
}

spag.plot <- function(rwl, zfac=1, useRaster = FALSE, res = 150, ...){
    stopifnot(is.logical(useRaster), length(useRaster) == 1)
    nseries <- ncol(rwl)
    if (nseries == 0) {
        stop("empty 'rwl' given, nothing to draw")
    }
    rwl2 <- scale(rwl * zfac, center = TRUE, scale = FALSE) # result is a matrix
    rwl2 <- as.rwl(rwl2)
    yr <- time(rwl2)
    first.year <- as.matrix(apply(rwl2, 2, yr.range, yr.vec=yr))[1, ]
    neworder <- order(first.year, decreasing=FALSE)
    rwl2 <- rwl2[, neworder, drop=FALSE]
    op <- par(no.readonly=TRUE)
    on.exit(par(op))
    par(mar=c(2, 5, 2, 5) + 0.1, mgp=c(1.1, 0.1, 0), tcl=0.5,
        xaxs="i",yaxs="i")
    ## Set vertical offset for plotting each series
    for (i in 1:nseries) {
        rwl2[, i] <- rwl2[, i] + i
    }
    plot(yr, rwl2[, 1], type="n", ylim=c(0, max(rwl2, na.rm=TRUE)),
         axes=FALSE, ylab="", xlab=gettext("Year", domain="R-dplR"))
    abline(h=1:nseries, col="grey")
    grid(ny = NA)
    if (is.na(useRaster)) {
        useRaster2 <- names(dev.cur()) %in% c("pdf", "postscript")
    } else {
        useRaster2 <- useRaster
    }
    cl <- quote(for (i in 1:nseries) {
        lines(yr, rwl2[, i], ...)
    })
    if (useRaster2) {
        tryCatch(rasterPlot(cl, res = res),
                 error = function(e) {
                     message(as.character(e), appendLF = FALSE)
                     message("reverting to useRaster=FALSE")
                     eval(cl)
                 })
    } else {
        eval(cl)
    }
    tmp.seq <- seq(from=1, to=nseries, by=2)
    axis(2, at=tmp.seq,
         labels=colnames(rwl2)[tmp.seq], srt=45,
         tick=FALSE, las=2)
    if (nseries > 1) {
        tmp.seq <- seq(from=2, to=nseries, by=2)
        axis(4, at=tmp.seq,
             labels=colnames(rwl2)[tmp.seq], srt=45,
             tick=FALSE, las=2)
    }
    axis(1)
    axis(3)
    box()
}

.onUnload <- function (libpath) {
    library.dynam.unload("dplR", libpath)
}
.onAttach <- function(libname, pkgname) {
  packageStartupMessage("This is dplR version ", 
                        packageVersion("dplR"),
                        ".\n",
                        "dplR is part of openDendro https://opendendro.org",
                        ".\n",
                        "New users can visit https://opendendro.github.io/dplR-workshop/ to get started.")
}

`rwi.stats.legacy` <-
    function(rwi, ids=NULL, period=c("max", "common"))
{
    ## Run correlations over period common to all cores is "common"
    ## over maximum individual pair overlaps is "max"

    period2 <- match.arg(period)
    if(period2 == "max")
        period2 <- "pairwise.complete.obs"
    else if(period2 == "common")
        period2 <- "complete.obs"
    n.cores <- ncol(rwi)

    ## If tree.id is NULL then assume one core per tree
    if(is.null(ids)){
        ids2 <- data.frame(tree=seq_len(n.cores), core=rep(1, n.cores))
    } else{
        ## Make error checks here
        if(nrow(ids) != n.cores)
            stop("dimension problem: ", "'ncol(rwi)' != 'nrow(ids)'")
        if(!all(vapply(ids, is.numeric, TRUE)))
            stop("'ids' must have numeric columns")
        ids2 <- ids
    }

    n.trees <- length(unique(ids2$tree))
    n.cores.tree <- data.frame(table(ids2$tree))$Freq

    r.mat <- cor(rwi, use=period2)
    ## If r.mat is all NA and period is common then it is likely that
    ## there is no overlap bewixt the cores. Warn the user.
    if(all(is.na(r.mat)) && period2 == "complete.obs"){
        warning("Correlations are all NA. No overlap in series?", call. = FALSE)
    }

    ## See p 138 in C&K
    ## Mean of all correlations among different cores (within and between trees)
    n.tot <- 0.5 * n.cores * (n.cores - 1) # length(r.mat[upper.tri(r.mat)])
    rbar.tot <- mean(r.mat[upper.tri(r.mat)], na.rm=TRUE) # 1/n.tot * sum(r.mat[upper.tri(r.mat)])
    ## Within-tree signal
    n.wt <- 0.5 * sum(n.cores.tree * (n.cores.tree - 1), na.rm=TRUE)

    r.wt.func <- function(x, x.id){
        samps <- unique(x.id)
        r.vec <- rep(0, length(samps))
        for(i in samps){
            x.part <- x[, x.id == i]
            if(!is.null(ncol(x.part))){
                r <- cor(x.part, use=period2)
                r.vec[i] <- sum(r[upper.tri(r)])
            }
        }
        r.vec
    }

    n <- n.cores

    ## Modify output and eps if one core per tree (e.g., ids were null)
    if(all(n.cores.tree == 1)){
        rbar.wt <- 0
        rbar.bt <- rbar.tot
        rbar.eff <- 0
        eps <- (n*rbar.bt) / ((n*rbar.bt) + (1-rbar.bt))
        n.bt <- n.tot
        c.eff <- 1
    } else{
        r.wt <- r.wt.func(rwi, ids2$tree)
        rbar.wt <- 1/n.wt * sum(r.wt, na.rm=TRUE)

        n.bt <- n.tot - n.wt
        rbar.bt <- 1/n.bt * (rbar.tot*n.tot - rbar.wt*n.wt)

        c.eff <- (1/n.trees * sum(1/n.cores.tree))^-1

        rbar.eff <- rbar.bt / (rbar.wt + (1-rbar.wt) / c.eff)

        eps <- (n*rbar.eff) / ((n*rbar.eff) + (1-rbar.eff))
    }

    compos.stats <- data.frame(n.tot, n.wt, n.bt, rbar.tot, rbar.wt,
                               rbar.bt, c.eff, rbar.eff, eps)
    round(compos.stats, 3)
}

sss <- function(rwi, ids=NULL) {
  # rwi.stats is robust enough to have a single call regardless of
  # whether ids is passed in because if no ids are passed in
  # rbar.eff == rbar.bt and n.cores=n.trees
  # Thus, if ids is NULL assume one core / tree use and 
  # we use rbar.bt (which is the same as rbar.eff). If there are ids
  # we use rbar.eff which is not the same as rbar.bt. Ditto for getting
  # N. Satisfy yourself via:
  #ca533.rwi <- detrend(ca533,method="Spline")
  #ca533.ids <- autoread.ids(ca533)
  #rwi.stats(ca533.rwi)
  #rwi.stats(ca533.rwi,ca533.ids)
  
  rwiVars <- rwi.stats(rwi, ids=ids)
  rbar <- rwiVars$rbar.eff
  N <- rwiVars$n.trees

  
  if(is.null(ids)){
    # n is is the number of cores in subsample
    n <- rowSums(!is.na(rwi))
  }
  else {
    # n is the number of trees in the subsample.
    # calculating n is kind of tedious:
    # we need n trees, not n cores in a year
    colnames.rwi <- colnames(rwi)
    n <- rep(NA_integer_, nrow(rwi))
    for(i in seq_len(nrow(rwi))){
      cols.with.data <- c(!is.na(rwi[i, ]))
      trees.this.year <- ids$tree[rownames(ids) %in% colnames.rwi[cols.with.data]]
      n[i] <- length(unique(trees.this.year))
    }
  }
  res <- (n*(1+(N-1)*rbar)) / (N*(1+(n-1)*rbar))
  res
}

`i.detrend.series` <- function(y, y.name=NULL, nyrs = NULL, f = 0.5,
                               pos.slope = FALSE)
{
    fits <- detrend.series(y, y.name, make.plot=TRUE, nyrs = nyrs, f = f,
                           pos.slope = pos.slope)
    ## Remove the nec resids if all na
    fits <- fits[, !colAlls(is.na(fits)), drop=FALSE]
    col.names <- names(fits)
    cat(gettextf("\nChoose a detrending method for this series %s.\n",
                 y.name, domain="R-dplR"))
    cat(gettext("Methods are: \n", domain="R-dplR"))
    for(i in seq_along(col.names))
        cat(i, ": ", col.names[i], "\n", sep="")
    ans <- as.integer(readline(gettext("Enter a number ", domain="R-dplR")))
    while(ans < 1 || ans > i || is.na(ans)){
        message("number out of range or not an integer\n")
        ans <- as.integer(readline(gettext("Enter a number ", domain="R-dplR")))
    }
    method <- col.names[ans]
    res <- fits[, method]
    names(res) <- names(y)
    res
}

common.interval <- function(rwl, type=c("series", "years", "both"),
                            make.plot=TRUE) {

    if (!is.data.frame(rwl)) {
        stop("'rwl' must be a data.frame")
    }

    if (!all(vapply(rwl, is.numeric, FALSE, USE.NAMES=FALSE))) {
        stop("'rwl' must have numeric columns")
    }
    rnames <- row.names(rwl)
    if (is.null(rnames)) {
        stop("'rwl' must have row names")
    }
    yrs <- as.numeric(rnames)
    if (!is.numeric(yrs) || any(is.na(yrs)) || any(round(yrs) != yrs)) {
        stop("row names of 'rwl' must be interpretable as years")
    }

    check.flags(make.plot)
    type2 <- match.arg(type, c("series", "years", "both"))

    ## rm.short is a function to remove short series and keep the
    ## series with overlaps
    rm.short <- function(rwl, yrs, rwlNotNA, row.idx, flag=FALSE) {
        n <- 0
        anyNotNA <- colAnys(rwlNotNA)
        which.good <- which(anyNotNA)
        nCol.orig <- length(which.good)
        series.range <- matrix(NA_real_, 2, nCol.orig)
        for (k in seq_len(nCol.orig)) {
            series.range[, k] <- yr.range(rwl[[which.good[k]]][row.idx],
                                          yr.vec = yrs)
        }
        span.order <-
            which.good[sort.list(series.range[2, ] - series.range[1, ])]
        nRow.orig <- nrow(rwlNotNA)
        keep.col <- logical(length(rwl))
        keep.col[which.good] <- TRUE
        keep.col.output <- keep.col
        dontkeep.row <- rep.int(TRUE, nRow.orig)
        keep.row.output <- rep.int(FALSE, nRow.orig)
        nRow <- 0
        nRow.output <- 0
        nCol.output <- nCol.orig
        nCol <- nCol.orig

        for (i in seq(0, max(0, nCol.orig - 2))) {
            if (i > 0) {
                keep.col[span.order[i]] <- FALSE
                nCol <- nCol - 1
                if (nCol * nRow.orig < n) {
                    ## to break if it is not possible to improve the
                    ## common interval
                    break
                }
            }
            tmp <- rowAlls(rwlNotNA, rows = dontkeep.row, cols = keep.col)
            dontkeep.row[dontkeep.row] <- !tmp
            nRow <- nRow + sum(tmp)
            n.years <- nCol * nRow
            ## to keep the rwl if has more years
            if (n.years > n) {
                n <- n.years
                keep.col.output <- keep.col
                keep.row.output <- !dontkeep.row
                nCol.output <- nCol
                nRow.output <- nRow
                if (flag) {
                    ## to give the common interval with the highest
                    ## sample depth for the case of
                    ## common.interval(rwl, type="series")
                    break
                }
            }
        }
        list(nRow.output, nCol.output, keep.row.output, keep.col.output)
    }

###########
    nCol.rwl <- length(rwl)
    nRow.rwl <- nrow(rwl)
    yrs.ordered <- all(diff(yrs) >= 0)
    if (!yrs.ordered) {
        order.yrs <- sort.list(yrs)
    }
    output <- 0
    opt <- 0
    keep.row.output <- numeric(0)
    keep.col.output <- logical(nCol.rwl)
    nCol.output <- 0
    nRow.output <- 0
    nCol <- 0
    nRow <- 0
    rwlNotNA <- !is.na(rwl)

    ## to get sample depth
    if (nCol.rwl > 0) {
        samp.depth <- rowSums(rwlNotNA)
    } else {
        ## Workaround for R bug number 14959.  Fixed in R >= 2.15.2.
        samp.depth <- 0
    }

    type.series <- type2 == "series"
    type.years <- type2 == "years"
    for (i in dec(max(samp.depth), 2)) { # dec() forces a decreasing sequence
        if (yrs.ordered) {
            tmp <- which(samp.depth >= i)
            row.idx <- tmp[1]:tmp[length(tmp)]
        } else {
            common.range <- range(yrs[samp.depth >= i])
            row.idx <- which(yrs >= common.range[1] & yrs <= common.range[2])
        }
        nRow <- length(row.idx)
        if (i * nRow < output) {
            break
        }
        if (type.series) {
            tmp <- rm.short(rwl, yrs[row.idx],
                            rwlNotNA[row.idx, , drop = FALSE], row.idx,
                            flag = TRUE)
            nRow.output <- tmp[[1]]
            nCol.output <- tmp[[2]]
            keep.row.output <- row.idx[tmp[[3]]]
            keep.col.output <- tmp[[4]]
            break
        } else if (type.years) {
            tmp <- rm.short(rwl, yrs[row.idx],
                            rwlNotNA[row.idx, , drop = FALSE], row.idx)
            nRow <- tmp[[1]]
            nCol <- tmp[[2]]
            keep.row <- tmp[[3]]
            keep.col <- tmp[[4]]
        } else { # type2 == "both"
            keep.col <- colAlls(rwlNotNA, rows = row.idx)
            nCol <- sum(keep.col)
        }
        opt <- nRow * nCol
        if (opt > output) {
            output <- opt
            nRow.output <- nRow
            nCol.output <- nCol
            if (type.years) {
                keep.row.output <- row.idx[keep.row]
            } else {
                keep.row.output <- row.idx
            }
            keep.col.output <- keep.col
        }
    }

    if (make.plot) {
        op <- par(no.readonly = TRUE)
        on.exit(par(op))
        par(mar = c(5, 5, 2, 2) + 0.1, mgp = c(1.25, 0.25, 0), tcl = 0.25)
        if (nRow.rwl > 0 && nCol.rwl > 0) {
            ## original rwl
            series.range <- vapply(rwl, yr.range, numeric(2), yr.vec = yrs)
            ## ensure that series.range is a matrix
            dim(series.range) <- c(2, length(rwl))
            first.year <- series.range[1, ]

            neworder <- sort.list(first.year, na.last = TRUE)
            if (yrs.ordered) {
                rwl.first <- yrs[1]
                rwl.last <- yrs[nRow.rwl]
            } else {
                rwl.first <- min(yrs)
                rwl.last <- max(yrs)
            }
            plot.first <- first.year[neworder[1]]
            if (is.na(plot.first)) {
                plot.first <- rwl.first
                plot.last <- rwl.last
            } else {
                plot.last <- max(series.range[2, ], na.rm = TRUE)
            }
            plot(1, 1, type = "n", xlim = c(plot.first, plot.last + 1),
                 ylim = c(1, nCol.rwl), axes = FALSE, ylab = "",
                 xlab = gettext("Year", domain = "R-dplR"))
            rwl.seq <- seq(from = rwl.first, to = rwl.last + 1, by = 0.5)
            n.rwl.seq <- length(rwl.seq)
            rwl.everyother <- seq(from = 2, by = 2, length.out = nRow.rwl)
        } else {
            plot(1, 1, type = "n", axes = FALSE, ylab = "", xlab = "")
        }
        sub.str1 <- gettextf("Original: %d series, %d years",
                             nCol.rwl, nRow.rwl, domain="R-dplR")
        sub.str2 <-
            gettextf("Common Interval (type='%s'): %d series x %d years = %d",
                     type2, nCol.output, nRow.output,
                     nCol.output * nRow.output, domain="R-dplR")
        sub.str <- paste(sub.str1, sub.str2, sep="\n")
        mtext(text = sub.str, side = 1, line = 3)
        ## common.rwl
        yrs2 <- yrs[keep.row.output]
        any.common <- length(yrs2) > 0
        if (any.common) {
            common.first <- min(yrs2)
            common.last <- max(yrs2)
            common.seq <- seq(from = common.first,
                              to = common.last + 1, by = 0.5)
            n.common.seq <- length(common.seq)
            common.everyother <- seq(from = 2, by = 2, length.out = nRow.output)
        }
        if (!yrs.ordered) {
            order.yrs <- sort.list(yrs)
            order.yrs2 <- sort.list(yrs2)
        }
        for (i in seq_len(nCol.rwl)) {
            this.col <- neworder[i]
            seg <- rwl[[this.col]]
            seg[rwlNotNA[, this.col]] <- i
            if (yrs.ordered) {
                seg.ordered <- seg
            } else {
                seg.ordered <- seg[order.yrs]
            }
            seg.fill <- rep.int(i, n.rwl.seq)
            seg.fill[rwl.everyother] <- seg.ordered
            lines(rwl.seq, seg.fill, lwd = 2, col = "grey")
            if (keep.col.output[this.col]) {
                seg2 <- seg[keep.row.output]
                if (!yrs.ordered) {
                    seg2 <- seg2[order.yrs2]
                }
                seg2.fill <- rep.int(i, n.common.seq)
                seg2.fill[common.everyother] <- seg2
                lines(common.seq, seg2.fill, lwd = 2, col = "black")
            }
        }
        if (nCol.rwl > 0) {
            axis(2, at = seq_len(nCol.rwl), labels = names(rwl)[neworder],
                 srt = 45, tick = FALSE, las = 2)
        }
        if (nRow.rwl > 0) {
            axis(1)
        }
        if (any.common) {
            common.at <- c(common.first, common.last + 1)
            common.labels <- as.character(c(common.first, common.last))
            abline(v = common.at, lty = "dashed")
            axis(3, at = common.at, labels = common.labels, tcl = -0.25)
        }
        box()
    }

    if (nRow.output < nRow.rwl || nCol.output < nCol.rwl) {
        rwl[keep.row.output, keep.col.output, drop = FALSE]
    } else {
        rwl
    }
}

pointer <- function(rwl, rgv.thresh=10, nseries.thresh=75,
                    round.decimals=2) {
    stopifnot(is.numeric(rgv.thresh), length(rgv.thresh) == 1,
              is.finite(rgv.thresh))
    if (rgv.thresh < 0) {
        stop("'rgv.thresh' must be > 0")
    }
    if (rgv.thresh >= 100) {
        warning("'rgv.thresh' > 100 is unusual.")
    }
    stopifnot(is.numeric(nseries.thresh), length(nseries.thresh) == 1,
              is.finite(nseries.thresh))
    if (nseries.thresh < 0 || nseries.thresh > 100) {
        stop("'nseries.thresh' must range from 0 to 100")
    }
    rwl2 <- as.matrix(rwl)
    if (!is.matrix(rwl2)) {
        stop("'rwl' must be coercible to a matrix")
    }
    rnames <- rownames(rwl2)
    if (is.null(rnames)) {
        stop("'rwl' must have explicit row names")
    }
    yrs <- as.numeric(rnames)
    nyrs <- length(yrs)
    if (nyrs < 2) {
        stop("'rwl' must have at least 2 rows")
    }
    gv <- rwl2[-1, , drop=FALSE] / rwl2[-nyrs, , drop=FALSE]
    out <- matrix(NA_real_, nrow=nyrs - 1, ncol=7)
    colnames(out) <- c("Year", "Nb.series", "Perc.pos", "Perc.neg",
                       "Nature", "RGV_mean", "RGV_sd")
    out[, 1] <- yrs[-1]
    out[, 2] <- rowSums(!is.na(gv))
    pos.thresh <- rgv.thresh / 100 + 1
    neg.thresh <- -pos.thresh + 2
    out[, 3] <- rowSums(gv >= pos.thresh, na.rm=TRUE) / out[, 2] * 100
    out[, 4] <- rowSums(gv <= neg.thresh, na.rm=TRUE) / out[, 2] * 100
    nat.y.1 <- pmax(0, out[, 3] - (nseries.thresh - 0.0000001))
    nat.y.2 <- pmax(0, out[, 4] - (nseries.thresh - 0.0000001))
    out[, 5] <- sign(nat.y.1 - nat.y.2)
    out[, 6] <- (rowMeans(gv, na.rm=TRUE) - 1) * 100
    out[, 7] <- rowSds(gv, na.rm=TRUE) * 100
    if (is.numeric(round.decimals) && length(round.decimals) > 0 &&
        is.finite(round.decimals[1]) && round.decimals[1] >= 0) {
        for (i in c(3, 4, 6, 7)) {
            out[, i] <- round(out[, i], round.decimals[1])
        }
    }
    as.data.frame(out)
}

`detrend.series` <-
  function(y, y.name = "", make.plot = TRUE,
           method = c("Spline", "ModNegExp", "Mean", "Ar", "Friedman",
                      "ModHugershoff", "AgeDepSpline"),
           nyrs = NULL, f = 0.5, pos.slope = FALSE,
           constrain.nls = c("never", "when.fail", "always"),
           verbose = FALSE, return.info = FALSE,
           wt, span = "cv", bass = 0, difference = FALSE)
  {
    check.flags(make.plot, pos.slope, verbose, return.info)
    dirtyDog <- FALSE # this will be used to warn the user in any fits are <0
    
    if (length(y.name) == 0) {
      y.name2 <- ""
    } else {
      y.name2 <- as.character(y.name)[1]
      stopifnot(Encoding(y.name2) != "bytes")
    }
    known.methods <- c("Spline", "ModNegExp", "Mean", "Ar", "Friedman",
                       "ModHugershoff", "AgeDepSpline")
    constrain2 <- match.arg(constrain.nls)
    method2 <- match.arg(arg = method,
                         choices = known.methods,
                         several.ok = TRUE)
    wt.missing <- missing(wt)
    wt.description <- NULL
    if (verbose) {
      widthOpt <- getOption("width")
      indentSize <- 1
      indent <- function(x) {
        paste0(paste0(rep.int(" ", indentSize), collapse = ""), x)
      }
      sepLine <-
        indent(paste0(rep.int("~", max(1, widthOpt - 2 * indentSize)),
                      collapse = ""))
      cat(sepLine,
          gettext("Verbose output: ", domain="R-dplR"), y.name2,
          sep = "\n")
      wt.description <- if (wt.missing) "default" else deparse(wt)
      opts <- c("make.plot" = make.plot,
                "method(s)" = deparse(method2),
                "nyrs" = if (is.null(nyrs)) "NULL" else nyrs,
                "f" = f,
                "pos.slope" = pos.slope,
                "constrain.nls" = constrain2,
                "verbose" = verbose,
                "return.info" = return.info,
                "wt" = wt.description,
                "span" = span,
                "bass" = bass,
                "difference" = difference)
      optNames <- names(opts)
      optChar <- c(gettext("Options", domain="R-dplR"),
                   paste(str_pad(optNames,
                                 width = max(nchar(optNames)),
                                 side = "right"),
                         opts, sep = "  "))
      cat(sepLine, indent(optChar), sep = "\n")
    }
    
    ## Remove NA from the data (they will be reinserted later)
    good.y <- which(!is.na(y))
    if(length(good.y) == 0) {
      stop("all values are 'NA'")
    } else if(any(diff(good.y) != 1)) {
      stop("'NA's are not allowed in the middle of the series")
    }
    y2 <- y[good.y]
    nY2 <- length(y2)
    
    ## Recode any zero values to 0.001
    if (verbose || return.info) {
      years <- names(y2)
      if (is.null(years)) {
        years <- good.y
      }
      zeroFun <- function(x) list(zero.years = years[is.finite(x) & x == 0])
      nFun <- function(x) list(n.zeros = length(x[[1]]))
      zero.years.data <- zeroFun(y2)
      n.zeros.data <- nFun(zero.years.data)
      dataStats <- c(n.zeros.data, zero.years.data)
      if (verbose) {
        cat("", sepLine, sep = "\n")
        if (n.zeros.data[[1]] > 0){
          if (is.character(years)) {
            cat(indent(gettext("Zero years in input series:\n",
                               domain="R-dplR")))
          } else {
            cat(indent(gettext("Zero indices in input series:\n",
                               domain="R-dplR")))
          }
          cat(indent(paste(zero.years.data[[1]], collapse = " ")),
              "\n", sep = "")
        } else {
          cat(indent(gettext("No zeros in input series.\n",
                             domain="R-dplR")))
        }
      }
    }
    y2[y2 == 0] <- 0.001
    
    resids <- list()
    curves <- list()
    modelStats <- list()
    ################################################################################    
    ################################################################################    
    # Ok. Let's start the methods
    
    ################################################################################    
    if("ModNegExp" %in% method2){
      ## Nec or lm
      nec.func <- function(Y, constrain) {
        nY <- length(Y)
        a <- mean(Y[seq_len(max(1, floor(nY * 0.1)))])
        b <- -0.01
        k <- mean(Y[floor(nY * 0.9):nY])
        nlsForm <- Y ~ I(a * exp(b * seq_along(Y)) + k)
        nlsStart <- list(a=a, b=b, k=k)
        checked <- FALSE
        constrained <- FALSE
        ## Note: nls() may signal an error
        if (constrain == "never") {
          nec <- nls(formula = nlsForm, start = nlsStart)
        } else if (constrain == "always") {
          nec <- nls(formula = nlsForm, start = nlsStart,
                     lower = c(a=0, b=-Inf, k=0),
                     upper = c(a=Inf, b=0, k=Inf),
                     algorithm = "port")
          constrained <- TRUE
        } else {
          nec <- nls(formula = nlsForm, start = nlsStart)
          coefs <- coef(nec)
          if (coefs[1] <= 0 || coefs[2] >= 0) {
            stop()
          }
          fits <- predict(nec)
          if (fits[nY] > 0) {
            checked <- TRUE
          } else {
            nec <- nls(formula = nlsForm, start = nlsStart,
                       lower = c(a=0, b=-Inf, k=0),
                       upper = c(a=Inf, b=0, k=Inf),
                       algorithm = "port")
            constrained <- TRUE
          }
        }
        if (!checked) {
          coefs <- coef(nec)
          if (coefs[1] <= 0 || coefs[2] >= 0) {
            stop()
          }
          fits <- predict(nec)
          if (fits[nY] <= 0) {
            ## This error is a special case that needs to be
            ## detected (if only for giving a warning).  Any
            ## smarter way to implement this?
            return(NULL)
          }
        }
        tmpFormula <- nlsForm
        formEnv <- new.env(parent = environment(detrend.series))
        formEnv[["Y"]] <- Y
        formEnv[["a"]] <- coefs["a"]
        formEnv[["b"]] <- coefs["b"]
        formEnv[["k"]] <- coefs["k"]
        environment(tmpFormula) <- formEnv
        structure(fits, constrained = constrained,
                  formula = tmpFormula, summary = summary(nec))
      }
      ModNegExp <- try(nec.func(y2, constrain2), silent=TRUE)
      mneNotPositive <- is.null(ModNegExp)
      
      if (verbose) {
        cat("", sepLine, sep = "\n")
        cat(indent(gettext("Detrend by ModNegExp.\n", domain = "R-dplR")))
        cat(indent(gettext("Trying to fit nls model...\n",
                           domain = "R-dplR")))
      }
      if (mneNotPositive || inherits(ModNegExp,"try-error")) {
        if (verbose) {
          cat(indent(gettext("nls failed... fitting linear model...",
                             domain = "R-dplR")))
        }
        ## Straight line via linear regression
        if (mneNotPositive) {
          dirtyDog <- TRUE
          msg <- gettext("Fits from method==\'ModNegExp\' are not all positive. \n  See constrain.nls argument in detrend.series. \n  ARSTAN would tell you to plot that dirty dog at this point.\n  Proceed with caution.",
                         domain = "R-dplR")
          if(y.name2==""){
            msg2 <- gettext(msg, domain = "R-dplR")
          }
          else {
            msg2 <- c(gettextf("In raw series %s: ", y.name2, domain = "R-dplR"),
                      gettext(msg, domain = "R-dplR"))
          }
          warning(msg2)
          if (verbose) {
            cat(sepLine, indent(msg), sepLine, sep = "\n")
          }
        }
        x <- seq_len(nY2)
        lm1 <- lm(y2 ~ x)
        coefs <- coef(lm1)
        xIdx <- names(coefs) == "x"
        coefs <- c(coefs[!xIdx], coefs[xIdx])
        if (verbose) {
          cat(indent(c(gettext("Linear model fit", domain = "R-dplR"),
                       gettextf("Intercept: %s", format(coefs[1]),
                                domain = "R-dplR"),
                       gettextf("Slope: %s", format(coefs[2]),
                                domain = "R-dplR"))),
              sep = "\n")
        }
        if (all(is.finite(coefs)) && (coefs[2] <= 0 || pos.slope)) {
          tm <- cbind(1, x)
          ModNegExp <- drop(tm %*% coefs)
          useMean <- !isTRUE(ModNegExp[1] > 0 &&
                               ModNegExp[nY2] > 0)
          if (useMean) {
            dirtyDog <- TRUE
            msg <- gettext("Linear fit (backup of method==\'ModNegExp\') is not all positive. \n  Proceed with caution. \n  ARSTAN would tell you to plot that dirty dog at this point.",
                           domain = "R-dplR")
            if(y.name2==""){
              msg2 <- gettext(msg, domain = "R-dplR")
            }
            else {
              msg2 <- c(gettextf("In raw series %s: ", y.name2, domain = "R-dplR"),
                        gettext(msg, domain = "R-dplR"))
            }
            warning(msg2)
            if (verbose) {
              cat(sepLine, indent(msg), sepLine, sep = "\n")
            }
          }
        } else {
          useMean <- TRUE
        }
        if (useMean) {
          theMean <- mean(y2)
          if (verbose) {
            cat(indent(c(gettext("lm has a positive slope",
                                 "pos.slope = FALSE",
                                 "Detrend by mean.",
                                 domain = "R-dplR"),
                         gettextf("Mean = %s", format(theMean),
                                  domain = "R-dplR"))),
                sep = "\n")
          }
          ModNegExp <- rep.int(theMean, nY2)
          mneStats <- list(method = "Mean", mean = theMean)
        } else {
          mneStats <- list(method = "Line", coefs = coef(summary(lm1)))
        }
      } else if (verbose || return.info) {
        mneSummary <- attr(ModNegExp, "summary")
        mneCoefs <- mneSummary[["coefficients"]]
        mneCoefsE <- mneCoefs[, 1]
        if (verbose) {
          cat(indent(c(gettext("nls coefs", domain = "R-dplR"),
                       paste0(names(mneCoefsE), ": ",
                              format(mneCoefsE)))),
              sep = "\n")
        }
        mneStats <- list(method = "NegativeExponential",
                         is.constrained = attr(ModNegExp, "constrained"),
                         formula = attr(ModNegExp, "formula"),
                         coefs = mneCoefs)
      } else {
        mneStats <- NULL
      }
      if(difference){ resids$ModNegExp <- y2 - ModNegExp }
      else{ resids$ModNegExp <- y2 / ModNegExp }
      curves$ModNegExp <- ModNegExp
      modelStats$ModNegExp <- mneStats
      do.mne <- TRUE
    } else {
      do.mne <- FALSE
    }
    ################################################################################    
    if("ModHugershoff" %in% method2){
      ## hug or lm
      hug.func <- function(Y, constrain) {
        nY <- length(Y)
        a <- mean(Y[floor(nY * 0.9):nY])
        b <- 1
        g <- 0.1
        d <- mean(Y[floor(nY * 0.9):nY])
        nlsForm <- Y ~ I(a*seq_along(Y)^b*exp(-g*seq_along(Y))+d)
        nlsStart <- list(a=a, b=b, g=g, d=d)
        checked <- FALSE
        constrained <- FALSE
        ## Note: nls() may signal an error
        if (constrain == "never") {
          hug <- nls(formula = nlsForm, start = nlsStart)
        } else if (constrain == "always") {
          hug <- nls(formula = nlsForm, start = nlsStart,
                     lower = c(a=0, b=-Inf, g=0, d=0),
                     upper = c(a=Inf, b=0, g=Inf, d=Inf),
                     algorithm = "port")
          constrained <- TRUE
        } else {
          hug <- nls(formula = nlsForm, start = nlsStart)
          coefs <- coef(hug)
          if (coefs[1] <= 0 || coefs[2] <= 0) {
            stop()
          }
          fits <- predict(hug)
          if (fits[nY] > 0) {
            checked <- TRUE
          } else {
            hug <- nls(formula = nlsForm, start = nlsStart,
                       lower = c(a=0, b=-Inf, g=0, d=0),
                       upper = c(a=Inf, b=0, g=Inf, d=Inf),
                       algorithm = "port")
            constrained <- TRUE
          }
        }
        if (!checked) {
          coefs <- coef(hug)
          if (coefs[1] <= 0 || coefs[2] <= 0) {
            stop()
          }
          fits <- predict(hug)
          if (fits[nY] <= 0) {
            ## This error is a special case that needs to be
            ## detected (if only for giving a warning).  Any
            ## smarter way to implement this?
            return(NULL)
          }
        }
        tmpFormula <- nlsForm
        formEnv <- new.env(parent = environment(detrend.series))
        formEnv[["Y"]] <- Y
        formEnv[["a"]] <- coefs["a"]
        formEnv[["b"]] <- coefs["b"]
        formEnv[["g"]] <- coefs["g"]
        formEnv[["d"]] <- coefs["d"]
        environment(tmpFormula) <- formEnv
        structure(fits, constrained = constrained,
                  formula = tmpFormula, summary = summary(hug))
      }
      ModHugershoff <- try(hug.func(y2, constrain2), silent=TRUE)
      hugNotPositive <- is.null(ModHugershoff)
      
      if (verbose) {
        cat("", sepLine, sep = "\n")
        cat(indent(gettext("Detrend by ModHugershoff.\n", domain = "R-dplR")))
        cat(indent(gettext("Trying to fit nls model...\n",
                           domain = "R-dplR")))
      }
      if (hugNotPositive || inherits(ModHugershoff,"try-error")) {
        if (verbose) {
          cat(indent(gettext("nls failed... fitting linear model...",
                             domain = "R-dplR")))
        }
        ## Straight line via linear regression
        if (hugNotPositive) {
          dirtyDog <- TRUE
          msg <- gettext("Fits from method==\'ModHugershoff\' are not all positive. \n  See constrain.nls argument in detrend.series. \n  ARSTAN would tell you to plot that dirty dog at this point.\n  Proceed with caution.",
                         domain = "R-dplR")
          if(y.name2==""){
            msg2 <- gettext(msg, domain = "R-dplR")
          }
          else {
            msg2 <- c(gettextf("In raw series %s: ", y.name2, domain = "R-dplR"),
                      gettext(msg, domain = "R-dplR"))
          }
          warning(msg2)
          if (verbose) {
            cat(sepLine, indent(msg), sepLine, sep = "\n")
          }
        }
        x <- seq_len(nY2)
        lm1 <- lm(y2 ~ x)
        coefs <- coef(lm1)
        xIdx <- names(coefs) == "x"
        coefs <- c(coefs[!xIdx], coefs[xIdx])
        if (verbose) {
          cat(indent(c(gettext("Linear model fit", domain = "R-dplR"),
                       gettextf("Intercept: %s", format(coefs[1]),
                                domain = "R-dplR"),
                       gettextf("Slope: %s", format(coefs[2]),
                                domain = "R-dplR"))),
              sep = "\n")
        }
        if (all(is.finite(coefs)) && (coefs[2] <= 0 || pos.slope)) {
          tm <- cbind(1, x)
          ModHugershoff <- drop(tm %*% coefs)
          useMean <- !isTRUE(ModHugershoff[1] > 0 &&
                               ModHugershoff[nY2] > 0)
          if (useMean) {
            dirtyDog <- TRUE
            msg <- gettext("Linear fit (backup of method==\'ModHugershoff\') is not all positive. \n  ARSTAN would tell you to plot that dirty dog at this point. \n  Proceed with caution.",
                           domain = "R-dplR")
            if(y.name2==""){
              msg2 <- gettext(msg, domain = "R-dplR")
            }
            else {
              msg2 <- c(gettextf("In raw series %s: ", y.name2, domain = "R-dplR"),
                        gettext(msg, domain = "R-dplR"))
            }
            warning(msg2)
            if (verbose) {
              cat(sepLine, indent(msg), sepLine, sep = "\n")
            }
            
          }
        } else {
          useMean <- TRUE
        }
        if (useMean) {
          theMean <- mean(y2)
          if (verbose) {
            cat(indent(c(gettext("lm has a positive slope",
                                 "pos.slope = FALSE",
                                 "Detrend by mean.",
                                 domain = "R-dplR"),
                         gettextf("Mean = %s", format(theMean),
                                  domain = "R-dplR"))),
                sep = "\n")
          }
          ModHugershoff <- rep.int(theMean, nY2)
          hugStats <- list(method = "Mean", mean = theMean)
        } else {
          hugStats <- list(method = "Line", coefs = coef(summary(lm1)))
        }
      } else if (verbose || return.info) {
        hugSummary <- attr(ModHugershoff, "summary")
        hugCoefs <- hugSummary[["coefficients"]]
        hugCoefsE <- hugCoefs[, 1]
        if (verbose) {
          cat(indent(c(gettext("nls coefs", domain = "R-dplR"),
                       paste0(names(hugCoefsE), ": ",
                              format(hugCoefsE)))),
              sep = "\n")
        }
        hugStats <- list(method = "Hugershoff",
                         is.constrained = attr(ModHugershoff, "constrained"),
                         formula = attr(ModHugershoff, "formula"),
                         coefs = hugCoefs)
      } else {
        hugStats <- NULL
      }
      if(difference){ resids$ModHugershoff <- y2 - ModHugershoff }
      else{ resids$ModHugershoff <- y2 / ModHugershoff }
      curves$ModHugershoff <- ModHugershoff
      modelStats$ModHugershoff <- hugStats
      do.hug <- TRUE
    } else {
      do.hug <- FALSE
    }
    ################################################################################    
    if("AgeDepSpline" %in% method2){
      ## Age dep smoothing spline with nyrs (50 default) as the init stiffness
      ## are NULL
      if(is.null(nyrs))
        nyrs2 <- 50
      else
        nyrs2 <- nyrs
      if (verbose) {
        cat("", sepLine, sep = "\n")
        cat(indent(c(gettext(c("Detrend by age-dependent spline.",
                               "Spline parameters"), domain = "R-dplR"),
                     paste0("nyrs = ", nyrs2, ", pos.slope = ", pos.slope))),
            sep = "\n")
      }
      AgeDepSpline <- ads(y=y2, nyrs0=nyrs2, pos.slope = pos.slope)
      if (any(AgeDepSpline <= 0)) {
        dirtyDog <- TRUE
        msg <- "Fits from method==\'AgeDepSpline\' are not all positive. \n  This is extremely rare. Series will be detrended with method==\'Mean\'. \n  This might not be what you want. \n  ARSTAN would tell you to plot that dirty dog at this point. \n  Proceed with caution."
        if(y.name2==""){
          msg2 <- gettext(msg, domain = "R-dplR")
        }
        else {
          msg2 <- c(gettextf("In raw series %s: ", y.name2, domain = "R-dplR"),
                    gettext(msg, domain = "R-dplR"))
        }
        warning(msg2)
        if (verbose) {
          cat(sepLine, indent(msg), sepLine, sep = "\n")
        }
        theMean <- mean(y2)
        AgeDepSpline <- rep.int(theMean, nY2)
        AgeDepSplineStats <- list(method = "Mean", mean = theMean)
      } else {
        AgeDepSplineStats <- list(method = "Age-Dep Spline", nyrs = nyrs2, pos.slope=pos.slope)
      }
      if(difference){ resids$AgeDepSpline <- y2 - AgeDepSpline }
      else{ resids$AgeDepSpline <- y2 / AgeDepSpline }
      curves$AgeDepSpline <- AgeDepSpline
      modelStats$AgeDepSpline <- AgeDepSplineStats
      
      do.ads <- TRUE
    } else {
      do.ads <- FALSE
    }
    ################################################################################    
    if("Spline" %in% method2){
      ## Smoothing spline
      ## "n-year spline" as the spline whose frequency response is
      ## 50%, or 0.50, at a wavelength of 67%n years if nyrs and f
      ## are NULL
      if(is.null(nyrs))
        nyrs2 <- floor(nY2 * 0.67)
      else
        nyrs2 <- nyrs
      if (verbose) {
        cat("", sepLine, sep = "\n")
        cat(indent(c(gettext(c("Detrend by spline.",
                               "Spline parameters"), domain = "R-dplR"),
                     paste0("nyrs = ", nyrs2, ", f = ", f))),
            sep = "\n")
      }
      #Spline <- ffcsaps(y=y2, x=seq_len(nY2), nyrs=nyrs2, f=f)
      Spline <- caps(y=y2, nyrs=nyrs2, f=f)
      if (any(Spline <= 0)) {
        dirtyDog <- TRUE
        msg <- "Fits from method==\'Spline\' are not all positive. \n  Series will be detrended with method==\'Mean\'. \n  This might not be what you want. \n  ARSTAN would tell you to plot that dirty dog at this point. \n  Proceed with caution."
        if(y.name2==""){
          msg2 <- gettext(msg, domain = "R-dplR")
        }
        else {
          msg2 <- c(gettextf("In raw series %s: ", y.name2, domain = "R-dplR"),
                    gettext(msg, domain = "R-dplR"))
        }
        warning(msg2)
        if (verbose) {
          cat(sepLine, indent(msg), sepLine, sep = "\n")
        }
        theMean <- mean(y2)
        Spline <- rep.int(theMean, nY2)
        splineStats <- list(method = "Mean", mean = theMean)
      } else {
        splineStats <- list(method = "Spline", nyrs = nyrs2, f = f)
      }
      if(difference){ resids$Spline <- y2 - Spline }
      else{ resids$Spline <- y2 / Spline }
      curves$Spline <- Spline
      modelStats$Spline <- splineStats
      
      do.spline <- TRUE
    } else {
      do.spline <- FALSE
    }
    ################################################################################    
    if("Mean" %in% method2){
      ## Fit a horiz line
      theMean <- mean(y2)
      Mean <- rep.int(theMean, nY2)
      if (verbose) {
        cat("", sepLine, sep = "\n")
        cat(indent(c(gettext("Detrend by mean.", domain = "R-dplR"),
                     paste("Mean = ", format(theMean)))),
            sep = "\n")
      }
      meanStats <- list(method = "Mean", mean = theMean)
      if(difference){ resids$Mean <- y2 - Mean }
      else{ resids$Mean <- y2 / Mean }
      curves$Mean <- Mean
      modelStats$Mean <- meanStats
      do.mean <- TRUE
    } else {
      do.mean <- FALSE
    }
    ################################################################################    
    if("Ar" %in% method2){
      ## Fit an ar model - aka prewhiten
      Ar <- ar.func(y2, model = TRUE)
      arModel <- attr(Ar, "model")
      if (verbose) {
        cat("", sepLine, sep = "\n")
        cat(indent(gettext("Detrend by prewhitening.", domain = "R-dplR")))
        print(arModel)
      }
      arStats <- list(method = "Ar", order = arModel[["order"]],
                      ar = arModel[["ar"]])
      # This will propagate NA to rwi as a result of detrending.
      # Other methods don't. Problem when interacting with other
      # methods?
      # Also, this can (and does!) produce negative RWI values.
      # See example using CAM011. Thus:
      if (any(Ar <= 0, na.rm = TRUE)) {
        dirtyDog <- TRUE
        msg <- "Fits from method==\'Ar\' are not all positive. \n  Setting values <0 to 0 before rescaling.  \n  This might not be what you want. \n  ARSTAN would tell you to plot that dirty dog at this point. \n  Proceed with caution."
        if(y.name2==""){
          msg2 <- gettext(msg, domain = "R-dplR")
        }
        else {
          msg2 <- c(gettextf("In raw series %s: ", y.name2, domain = "R-dplR"),
                    gettext(msg, domain = "R-dplR"))
        }
        warning(msg2)
        if (verbose) {
          cat(sepLine, indent(msg), sepLine, sep = "\n")
        }
        
        Ar[Ar<0] <- 0
      }
      if(difference){ 
        # changed per issue 22
        # https://github.com/OpenDendro/dplR/issues/22
        #resids$Ar <- Ar - mean(Ar,na.rm=TRUE) 
        resids$Ar <- Ar - mean(y2, na.rm = TRUE) 
        }
      else{ 
        # changed per issue 22
        # https://github.com/OpenDendro/dplR/issues/22
        #resids$Ar <- Ar / mean(Ar,na.rm=TRUE) 
        resids$Ar <- Ar/mean(y2, na.rm = TRUE)
        }
      curves$Ar <- mean(Ar,na.rm=TRUE)
      modelStats$Ar <- arStats
      do.ar <- TRUE
    } else {
      do.ar <- FALSE
    }
    ################################################################################    
    if ("Friedman" %in% method2) {
      if (is.null(wt.description)) {
        wt.description <- if (wt.missing) "default" else deparse(wt)
      }
      if (verbose) {
        cat("", sepLine, sep = "\n")
        cat(indent(c(gettext(c("Detrend by Friedman's super smoother.",
                               "Smoother parameters"), domain = "R-dplR"),
                     paste0("span = ", span, ", bass = ", bass),
                     paste0("wt = ", wt.description))),
            sep = "\n")
      }
      if (wt.missing) {
        Friedman <- supsmu(x = seq_len(nY2), y = y2, span = span,
                           periodic = FALSE, bass = bass)[["y"]]
      } else {
        Friedman <- supsmu(x = seq_len(nY2), y = y2, wt = wt, span = span,
                           periodic = FALSE, bass = bass)[["y"]]
      }
      if (any(Friedman <= 0)) {
        dirtyDog <- TRUE
        msg <- "Fits from method==\'Friedman\' are not all positive. \n  Series will be detrended with method==\'Mean\'. \n  This might not be what you want. \n  ARSTAN would tell you to plot that dirty dog at this point. \n  Proceed with caution."
        if(y.name2==""){
          msg2 <- gettext(msg, domain = "R-dplR")
        }
        else {
          msg2 <- c(gettextf("In raw series %s: ", y.name2, domain = "R-dplR"),
                    gettext(msg, domain = "R-dplR"))
        }
        warning(msg2)
        if (verbose) {
          cat(sepLine, indent(msg), sepLine, sep = "\n")
        }
        
        
        theMean <- mean(y2)
        Friedman <- rep.int(theMean, nY2)
        friedmanStats <- list(method = "Mean", mean = theMean)
      } else {
        friedmanStats <- list(method = "Friedman", wt = wt.description, span = span, bass = bass)
      }
      
      if(difference){ resids$Friedman <- y2 - Friedman }
      else{ resids$Friedman <- y2 / Friedman }
      curves$Friedman <- Friedman
      modelStats$Friedman <-
        list(method = "Friedman",
             wt = if (wt.missing) "default" else wt,
             span = span, bass = bass)
      do.friedman <- TRUE
    } else {
      do.friedman <- FALSE
    }
    ################################################################################    
    ################################################################################    
    
    resids <- data.frame(resids)
    curves <- data.frame(curves)
    if (verbose || return.info) {
      zero.years <- lapply(resids, zeroFun)
      n.zeros <- lapply(zero.years, nFun)
      modelStats <- mapply(c, modelStats, n.zeros, zero.years,
                           SIMPLIFY = FALSE)
      if (verbose) {
        n.zeros2 <- unlist(n.zeros, use.names = FALSE)
        zeroFlag <- n.zeros2 > 0
        methodNames <- names(modelStats)
        if (any(zeroFlag)) {
          cat("", sepLine, sep = "\n")
          for (i in which(zeroFlag)) {
            if (is.character(years)) {
              cat(indent(gettextf("Zero years in %s series:\n",
                                  methodNames[i], domain="R-dplR")))
            } else {
              cat(indent(gettextf("Zero indices in %s series:\n",
                                  methodNames[i], domain="R-dplR")))
            }
            cat(indent(paste(zero.years[[i]][[1]], collapse = " ")),
                "\n", sep = "")
          }
        }
      }
    }
    
    if(make.plot){
      cols <- c("#24492e","#015b58","#2c6184","#59629b","#89689d","#ba7999","#e69b99")
      op <- par(no.readonly=TRUE)
      on.exit(par(op))
      n.methods <- ncol(resids)
      par(mar=c(2.1, 2.1, 2.1, 2.1), mgp=c(1.1, 0.1, 0),
          tcl=0.5, xaxs="i")
      if (n.methods > 4) {
        par(cex.main = min(1, par("cex.main")))
      }
      mat <- switch(n.methods,
                    matrix(c(1,2), nrow=2, ncol=1, byrow=TRUE),
                    matrix(c(1,1,2,3), nrow=2, ncol=2, byrow=TRUE),
                    matrix(c(1,2,3,4), nrow=2, ncol=2, byrow=TRUE),
                    matrix(c(1,1,2,3,4,5), nrow=3, ncol=2, byrow=TRUE),
                    matrix(c(1,1,1,2,3,4,5,6,0), nrow=3, ncol=3, byrow=TRUE),
                    matrix(c(1,1,1,2,3,4,5,6,7), nrow=3, ncol=3, byrow=TRUE),
                    matrix(c(1,2,3,4,5,6,7,8), nrow=4, ncol=2, byrow=TRUE))
      
      layout(mat,
             widths=rep.int(0.5, ncol(mat)),
             heights=rep.int(1, nrow(mat)))
      
      # 1
      plot(y2, type="l", ylab="mm", col = "grey",
           xlab=gettext("Age (Yrs)", domain="R-dplR"),
           main=gettextf("Raw Series %s", y.name2, domain="R-dplR"))
      if(do.spline) lines(Spline, col=cols[1], lwd=2)
      if(do.mne) lines(ModNegExp, col=cols[2], lwd=2)
      if(do.mean) lines(Mean, col=cols[3], lwd=2)
      if(do.friedman) lines(Friedman, col=cols[5], lwd=2)
      if(do.hug) lines(ModHugershoff, col=cols[6], lwd=2)
      if(do.ads) lines(AgeDepSpline, col=cols[7], lwd=2)
      
      # 1
      if(do.spline){
        plot(resids$Spline, type="l", col=cols[1],
             main=gettext("Spline", domain="R-dplR"),
             xlab=gettext("Age (Yrs)", domain="R-dplR"),
             ylab=gettext("RWI", domain="R-dplR"))
        if(difference){ abline(h=0) }
        else{ abline(h=1) }
      }
      # 2
      if(do.mne){
        plot(resids$ModNegExp, type="l", col=cols[2],
             main=gettext("Neg. Exp. Curve or Straight Line",
                          domain="R-dplR"),
             xlab=gettext("Age (Yrs)", domain="R-dplR"),
             ylab=gettext("RWI", domain="R-dplR"))
        if(difference){ abline(h=0) }
        else{ abline(h=1) }
        
      }
      # 3
      if(do.mean){
        plot(resids$Mean, type="l", col=cols[3],
             main=gettext("Horizontal Line (Mean)", domain="R-dplR"),
             xlab=gettext("Age (Yrs)", domain="R-dplR"),
             ylab=gettext("RWI", domain="R-dplR"))
        if(difference){ abline(h=0) }
        else{ abline(h=1) }
        
      }
      # 4
      if(do.ar){
        plot(resids$Ar, type="l", col=cols[4],
             main=gettextf("Ar", domain="R-dplR"),
             xlab=gettext("Age (Yrs)", domain="R-dplR"),
             ylab=gettext("RWI", domain="R-dplR"))
        if(difference){ abline(h=0) }
        else{ abline(h=1) }
        mtext(text="(Not plotted with raw series)",side=3,line=-1,cex=0.75)
      }
      # 5
      if (do.friedman) {
        plot(resids$Friedman, type="l", col=cols[5],
             main=gettext("Friedman's Super Smoother", domain="R-dplR"),
             xlab=gettext("Age (Yrs)", domain="R-dplR"),
             ylab=gettext("RWI", domain="R-dplR"))
        if(difference){ abline(h=0) }
        else{ abline(h=1) }
        
      }
      # 6
      if(do.hug){
        plot(resids$ModHugershoff, type="l", col=cols[6],
             main=gettext("Hugershoff or Straight Line",
                          domain="R-dplR"),
             xlab=gettext("Age (Yrs)", domain="R-dplR"),
             ylab=gettext("RWI", domain="R-dplR"))
        if(difference){ abline(h=0) }
        else{ abline(h=1) }
        
      }
      # 7
      if(do.ads){
        plot(resids$AgeDepSpline, type="l", col=cols[7],
             main=gettext("Age Dep Spline",
                          domain="R-dplR"),
             xlab=gettext("Age (Yrs)", domain="R-dplR"),
             ylab=gettext("RWI", domain="R-dplR"))
        if(difference){ abline(h=0) }
        else{ abline(h=1) }
        
      }
    }
      # Done get output
      resids2 <- matrix(NA, ncol=ncol(resids), nrow=length(y))
      resids2 <- data.frame(resids2)
      names(resids2) <- names(resids)
      if(!is.null(names(y))) row.names(resids2) <- names(y)
      resids2[good.y, ] <- resids
      
      curves2 <- matrix(NA, ncol=ncol(curves), nrow=length(y))
      curves2 <- data.frame(curves2)
      names(curves2) <- names(curves)
      if(!is.null(names(y))) row.names(curves2) <- names(y)
      curves2[good.y, ] <- curves
      ## Reorder columns of output to match the order of the argument
      ## "method".
      resids2 <- resids2[, method2]
      curves2 <- curves2[, method2]
      ## Make sure names (years) are included if there is only one method
      if(!is.data.frame(resids2)) names(resids2) <- names(y)
      if (return.info) {
        list(series = resids2,
             curves = curves2,
             model.info = modelStats[method2],
             data.info = dataStats,
             dirtyDog = dirtyDog)
      } else {
        resids2
      }
    }
    

`glk` <- 
    function(x, overlap = 50, prob = TRUE) 
{
  if(!("rwl" %in% class(x))) {
    warning("'x' is not class rwl")
  }
  if(any(length(overlap)!=1 | !is.numeric(overlap) | 
         overlap%%1!=0 | overlap < 3)){
    stop("'overlap' should be a single integer >=3")
  }
  if(overlap < 50) {
    warning("The minimum number of overlap is lower than 50. This might lead to statistically insignificant matches.")
  }
  if (!is.logical(prob)) {
    stop("'prob' must be either TRUE (the default) or FALSE")
  }
  # function starts here
  n <- dim(x)[2]
  GLK_mat <- matrix(NA_real_, nrow = n, ncol = n)
  Overlap_n <- matrix(NA_real_, nrow = n, ncol = n)
  rownames(GLK_mat) <- colnames(GLK_mat) <- names(x)
  rownames(Overlap_n) <- colnames(Overlap_n) <- names(x)
  treering_sign <- apply(x, 2, diff)
  treering_sign <- sign(treering_sign)
  for (i in 1:n) {
    treering_GC <- abs(treering_sign[,i]-treering_sign)
    # overlap is the number of overlapping growth changes
    treering_GCol <- colSums(!is.na(treering_GC))
    treering_GCol[treering_GCol==0] <- NA
    treering_GCol[treering_GCol<overlap] <- NA
    GLK_values <- 1-(colSums(treering_GC,na.rm=TRUE)/(2*treering_GCol))
    GLK_mat[i,] <- GLK_values
    Overlap_n[i,] <- treering_GCol
  }
  diag(GLK_mat) <- 1
  if (prob == TRUE) {
    s_df <- 1 /(2*sqrt(Overlap_n))
    z_df <- (GLK_mat-0.5)/s_df
    z_normcdf <- apply(z_df, 2, function(z) pnorm(z, mean=0, sd=1))
    GLK_p <- 2*(1-z_normcdf)   
    glk_list <- list(glk_mat = GLK_mat,
                     overlap = Overlap_n,
                     p_mat = GLK_p)
  } else {
    glk_list <- list(glk_mat = GLK_mat,
                     overlap = Overlap_n)}
  glk_list
    }

`glk.legacy` <- 
  function(x) 
{
    ## 20-Oct-2020 AGB - changing this to glk.legacy as Visser's faster glk is implemented.
    n <- dim(x)[2]
    G <- matrix(NA, nrow = n, ncol = n)
    rownames(G) <- colnames(G) <- names(x)
    for (i in inc(1, n - 1)) {
      col1 <- x[, i]
      not.na.1 <- which(!is.na(col1))
      if (length(not.na.1) >= 3) {
        for (k in (i + 1):n) {
          col2 <- x[, k]
          not.na.2 <- which(!is.na(col2))
          ## check if common interval is longer than 3 years
          not.na.both <- sort(intersect(not.na.1, not.na.2))
          m <- length(not.na.both)
          if (m >= 3) {
            if (not.na.both[m] - not.na.both[1] + 1 != m) {
              ## Should not happen; missing value marker is zero
              warning(gettextf("Intersection of series %d and %d is not contiguous. NA returned.",
                               i, k))
            } else {
              dif1 <- sign(diff(col1[not.na.both]))
              dif2 <- sign(diff(col2[not.na.both]))
              G[i, k] <- 1 - sum(abs(dif1 - dif2))/(2 * m - 2)
            }
          }
        }
      }
    }
    G
}

rasterPlot <- function(expr, res = 150, region=c("plot", "figure"), antialias,
                       bg = "transparent", interpolate = TRUE, draw = TRUE,
                       Cairo = FALSE, ...) {
    draw2 <- isTRUE(as.logical(draw)[1L])
    Cairo2 <- isTRUE(as.logical(Cairo)[1L])
    ## Plotting commands 'expr' will be evaluated in the environment
    ## of the caller of rasterPlot()
    pf <- parent.frame()
    fallback <- FALSE
    if (draw2 &&
        identical(dev.capabilities("rasterImage")[["rasterImage"]], "no")) {
        message("device does not support raster images")
        fallback <- TRUE
    }
    cairoRaster <- FALSE
    fallback <- TRUE
    for (k in 1:2) {
        if (Cairo2) {
            if (requireVersion("Cairo", "1.4-8")) {
                caps <- Cairo::Cairo.capabilities()
                if (isTRUE(as.vector(caps["raster"]))) {
                    fallback <- FALSE
                    cairoRaster <- TRUE
                } else if (isTRUE(as.vector(caps["png"]))) {
                    fallback <- FALSE
                } else {
                    message("png and raster unsupported in this Cairo library")
                }
                if (!fallback) {
                    if (k == 2) {
                        message("using Cairo device")
                    }
                    break
                }
            } else {
                message("Cairo device (>= 1.4-8) unavailable")
            }
            Cairo2 <- FALSE
        } else {
            if (sum(capabilities(c("cairo", "png", "aqua")), na.rm=TRUE) > 0) {
                fallback <- FALSE
                if (k == 2) {
                    message("using png device")
                }
                break
            } else {
                message("png device unavailable")
            }
            Cairo2 <- TRUE
        }
    }
    if (fallback && !draw2) {
        return(NULL)
    }
    region2 <- match.arg(region)
    plotRegion <- region2 == "plot"
    ## Start new plot if one does not exist
    parnew <- tryCatch(par(new = TRUE), warning = function(...) NULL)
    op <- NULL
    marzero <- FALSE
    if (is.null(parnew)) {
        if (!plotRegion && !fallback) {
            plot.new()
            op <- par(no.readonly = TRUE)
            par(mar = c(0, 0, 0, 0))
            marzero <- TRUE
        }
        plot.new()
        parnew <- list(new = FALSE)
    } else if (!parnew[[1L]]) {
        par(new = FALSE)
    }
    usr <- par("usr")
    ## Limits of the plot region in user coordinates
    usrLeft <- usr[1]
    usrRight <- usr[2]
    usrBottom <- usr[3]
    usrTop <- usr[4]
    figCoord <- function() {
        usrWidth <- usrRight - usrLeft
        usrHeight <- usrTop - usrBottom
        plt <- par("plt")
        ## Limits of the plot region proportional to the figure region, 0..1
        pltLeft <- plt[1]
        pltRight <- plt[2]
        pltWidth <- pltRight - pltLeft
        pltBottom <- plt[3]
        pltTop <- plt[4]
        pltHeight <- pltTop - pltBottom
        ## Limits of the figure region in user coordinates
        figLeft <- usrLeft - pltLeft / pltWidth * usrWidth
        figRight <- usrRight + (1 - pltRight) / pltWidth * usrWidth
        figBottom <- usrBottom - pltBottom / pltHeight * usrHeight
        figTop <- usrTop + (1 - pltTop) / pltHeight * usrHeight
        return(c(figLeft, figBottom, figRight, figTop))
    }
    if (fallback) {
        message("using fallback: regular plotting")
        on.exit(par(parnew))
        parxpd <- par(xpd = !plotRegion)
        on.exit(par(parxpd), add = TRUE)
        if (length(bg) != 1 || !is.character(bg) || bg != "transparent") {
            if (plotRegion) {
                rect(usrLeft, usrBottom, usrRight, usrTop,
                     col = bg, border = NA)
            } else {
                fc <- figCoord()
                rect(fc[1], fc[2], fc[3], fc[4], col = bg, border = NA)
            }
        }
        par(new = TRUE)
        eval(expr, pf)
        return(invisible(NULL))
    }
    ## Record number of current device so it can be reactivated later
    curDev <- dev.cur()
    ## Record graphical parameters of the device
    if (is.null(op)) {
        op <- par(no.readonly = TRUE)
    }
    pngWidthHeight <- op[[c(figure="fin", plot="pin")[region2]]]
    op <- op[!(names(op) %in%
               c("ask", "bg", "fig", "fin", "mar", "mfcol", "mfg", "mfrow",
                 "new", "oma", "omd", "omi", "pin", "plt",
                 if (plotRegion) "mai"))]
    ## Open a png device (raster image) using a temporary file.  Width
    ## and height are set to match the dimensions of the figure region
    ## in the original device.  Resolution (points per inch) is the
    ## argument 'res'.
    fname <- tempfile(fileext = ".png")
    if (Cairo2) {
        if (cairoRaster) {
            cairoType <- "raster"
            cairoFile <- ""
        } else {
            check.tempdir()
            cairoType <- "png"
            cairoFile <- fname
        }
        Cairo::Cairo(width = pngWidthHeight[1], height = pngWidthHeight[2],
                     units = "in", dpi = res, bg = bg,
                     type = cairoType, file = cairoFile, ...)
    } else if (missing(antialias)) {
        check.tempdir()
        png(fname, width = pngWidthHeight[1], height = pngWidthHeight[2],
            units = "in", res = res, bg = bg, ...)
    } else {
        check.tempdir()
        png(fname, width = pngWidthHeight[1], height = pngWidthHeight[2],
            units = "in", res = res, bg = bg, antialias = antialias, ...)
    }
    ## Record things to do on exit (will be removed from list one-by-one)
    on.exit(dev.off())
    on.exit(dev.set(curDev), add=TRUE)
    if (!cairoRaster) {
        on.exit(unlink(fname), add=TRUE)
    }
    devAskNewPage(FALSE)
    par(mfcol=c(1,1))
    par(omi=rep(0, 4))
    if (plotRegion) {
        par(mai=rep(0, 4))
    }
    ## Initialize and copy graphical parameters from original device
    plot.new()
    par(op)
    eval(expr, pf)
    if (cairoRaster) {
        ## Capture raster data from device before closing
        rasterData <- dev.capture(native = TRUE)
        on.exit(dev.set(curDev))
        dev.off()
        on.exit()
    } else {
        on.exit(dev.set(curDev))
        on.exit(unlink(fname), add=TRUE)
        dev.off()
        on.exit(unlink(fname))
    }
    ## Return to the original plot (device)
    dev.set(curDev)
    if (!cairoRaster) {
        ## Read the png image to memory
        rasterData <- readPNG(fname, native=TRUE)
        on.exit()
        ## Remove the temporary .png file
        unlink(fname)
    }
    if (!draw2) {
        return(rasterData)
    }
    if (plotRegion || marzero) {
        ## Add a raster image to the plot region of the original plot
        rasterImage(rasterData, xleft = usrLeft, ybottom = usrBottom,
                    xright = usrRight, ytop = usrTop,
                    interpolate = interpolate)
    } else {
        ## Set clipping to figure region, restore at exit
        par(xpd = TRUE)
        on.exit(par(xpd = op[["xpd"]]))
        ## Add a raster image to the figure region of the original plot
        fc <- figCoord()
        rasterImage(rasterData, xleft = fc[1], ybottom = fc[2],
                    xright = fc[3], ytop = fc[4],
                    interpolate = interpolate)
    }
    invisible(NULL)
}

time.rwl <- function(x,...){
  as.numeric(rownames(x))
}

time.crn <- function(x,...){
  as.numeric(rownames(x))
}


# creating the method for `time.rwl<-` and `time.crn<-`
`time<-` <- function (x, value) {
  UseMethod("time<-")
}

`time<-.rwl` <- function(x, value){
  row.names(x) <- value
  x
}

`time<-.crn` <- function(x, value){
  row.names(x) <- value
  x
}

`chron.stabilized` <-
  function(x, winLength, biweight = TRUE, running.rbar = FALSE)
  {
    if(!is.int(winLength)) stop("'winLength' must be an integer.")
    if(winLength > nrow(x)) stop("'winLength' must be (considerably) shorter than the chronology length.")
    if(winLength <= 30) warning("'winLength' < 30 is not recommended.\n  Consider a longer window.")
    if(winLength/nrow(x) > 0.5) warning("'winLength' > 50% of chronology length is not recommended.\n  Consider a shorter window.")
    
    # get rbar for some window length
    rbarWinLength <-function (x, WL=winLength) {
      corMat <- cor(x, use="pairwise.complete.obs")
      diag(corMat) <- NA
      presenceMatrix <- ifelse(is.na(x),0,1)
      overlapMatrix <- t(presenceMatrix)%*%presenceMatrix
      # must have more than 1/3 of WL to count to rbar
      corMat[overlapMatrix < (WL/3)] <- NA  
      res <- mean(corMat, na.rm=TRUE)
      res
    }
    #pre-process the data to have a mean of zero
    mean.x <-mean(rowMeans(x,na.rm=TRUE,dims=1)) 
    x0 <- x-mean.x  
    nSamps <- rowSums(!is.na(x0))
    xCrn <- rowMeans(x0,na.rm=T)
    if(biweight) {
      xCrn <- apply(x0,1,tbrm)
    }
    
    xCorrelMat <- cor(x0,use="pairwise.complete.obs")
    diag(xCorrelMat) <- NA
    rbar <- mean(xCorrelMat, na.rm =TRUE) #rbar
    
    movingRbarVec <- rep(NA,nrow(x0))
    
    # if winLength is odd
    
    if(winLength%%2 == 1){
      for(i in 1:(nrow(x0)-winLength+1)){
        movingRbarVec[i+(winLength-1)/2] <- rbarWinLength(x0[i:(i+winLength-1),])
      } 
    }
    # if winLength is even
    else{
      for(i in 1:(nrow(x0)-winLength+1)){
        movingRbarVec[i+(winLength)/2] <- rbarWinLength(x0[i:(i+winLength-1),])
      } 
    }
    # The 1st winLength/2 values of movingRbarVec are NA as are the 
    # last winLength/2 (depending on odd or even winLength). 
    # Pad with with first and last real values. This replaces the original call
    # to na.locf from zoo.
    # find the indices of the first and last NA values
    idxNA <- which(!is.na(movingRbarVec))
    padLow <- min(idxNA)
    padHigh <- max(idxNA)
    # pad the low end
    movingRbarVec[1:padLow] <- movingRbarVec[padLow]
    # and the high end
    movingRbarVec[padHigh:length(movingRbarVec)] <- movingRbarVec[padHigh]
    # keep NA where there no samples
    movingRbarVec[nSamps==0] <- NA
    
    nSampsEff <- nSamps/(1+(nSamps-1)*movingRbarVec)
    nSampsEff <- pmin(nSampsEff,nSamps,na.rm=TRUE) 

    # res
    xCrnAdjusted <- xCrn*(nSampsEff*rbar)^.5
    # add back the mean to the data
    xCrnAdjusted <- scale(xCrnAdjusted,
                          center=-mean.x, scale=FALSE)[,1]
    # make output
    if(running.rbar){
      res <- data.frame(vsc = xCrnAdjusted,
                        running.rbar = movingRbarVec,
                        samp.depth = nSamps)
    }
    else{
      res <- data.frame(vsc = xCrnAdjusted,
                        samp.depth = nSamps)
    }  
    rownames(res)<-rownames(x0)
    class(res) <- c("crn", "data.frame")
    return(res)
  }

fill.internal.NA <- function(x, fill=c("Mean", "Spline", "Linear")){
    fillInternalNA.series <- function(x, fill=0){
        x.na <- is.na(x)
        x.ok <- which(!x.na)
        n.ok <- length(x.ok)
        if (n.ok <= 1) {
            return(x)
        }
        ## find first and last
        first.ok <- x.ok[1]
        last.ok <- x.ok[n.ok]
        ## fill internal NA
        if (last.ok - first.ok + 1 > n.ok) {
            first.to.last <- first.ok:last.ok
            x2 <- x[first.to.last]
            x2.na <- x.na[first.to.last]
            if (fill == "Mean") {
                ## fill internal NA with series mean
                x2[x2.na] <- mean(x2[!x2.na])
            } else if (is.numeric(fill)) {
                ## fill internal NA with user supplied value
                x2[x2.na] <- fill
            } else {
                good.x <- which(!x2.na)
                good.y <- x2[good.x]
                bad.x <- which(x2.na)
                if (fill == "Spline") {
                    ## fill internal NA with spline
                    x2.aprx <- spline(x=good.x, y=good.y, xout=bad.x)
                } else {
                    ## fill internal NA with linear interpolation
                    x2.aprx <- approx(x=good.x, y=good.y, xout=bad.x)
                }
                x2[bad.x] <- x2.aprx$y
            }
            ## repad x
            x3 <- x
            x3[first.to.last] <- x2
            x3
        } else {
            x
        }
    }
    if (!is.data.frame(x)) {
        stop("'x' must be a data.frame")
    }
    if (!all(vapply(x, is.numeric, FALSE, USE.NAMES=FALSE))) {
        stop("'x' must have numeric columns")
    }
    if (is.numeric(fill)) {
        if (length(fill) == 1) {
            fill2 <- fill[1]
        } else {
            stop("'fill' must be a single number or character string")
        }
    } else {
        fill2 <- match.arg(fill)
    }
    y <- vapply(x, fillInternalNA.series, numeric(nrow(x)), fill=fill2)
    dim(y) <- dim(x)
    y <- as.data.frame(y)
    row.names(y) <- row.names(x)
    names(y) <- names(x)
    y
}

xskel.ccf.plot <- function(rwl,series,series.yrs = as.numeric(names(series)),
         win.start, win.width=50, n = NULL, prewhiten = TRUE,
         biweight = TRUE, series.x = FALSE) {
  
  ## check to see that win.width is even
  if(as.logical(win.width %% 2)) stop("'win.width' must be even")
  if (win.width > 100) {
    warning("win.width should be < 100 unless your plotting is very wide!")
  }

  ## Handle different types of 'series'
  tmp <- pick.rwl.series(rwl, series, series.yrs)
  rwl2 <- tmp[[1]]
  series2 <- tmp[[2]]

  master.yrs <- as.numeric(rownames(rwl2))
  series.yrs2 <- as.numeric(names(series2))
  yrs <- seq(from=win.start,to=win.start+win.width)
  ## nyrs <- length(yrs)
  cen.win <- win.width/2

  ## check window overlap with master and series yrs
  if (!all(yrs %in% series.yrs2)) {
    cat("Window Years: ", min(yrs), "-", max(yrs)," & ",
        "Series Years: ", min(series.yrs2), "-", max(series.yrs2),
        "\n", sep="")
    stop("Fix window overlap")
  }
  if (!all(yrs %in% master.yrs)) {
    cat("Window Years: ", min(yrs), "-", max(yrs)," & ",
        "Master Years: ", min(master.yrs), "-", max(master.yrs),
        "\n", sep="")
    stop("Fix window overlap")
  }

  ## normalize.
  names(series2) <- series.yrs2
  tmp <- normalize.xdate(rwl2, series2, n, prewhiten, biweight)

  ## master
  master <- tmp$master
  master.yrs <- as.numeric(names(master))
  master <- master[master.yrs%in%yrs]
  master.yrs <- as.numeric(names(master))
  ## series
  series2 <- tmp$series
  series.yrs2 <- as.numeric(names(series2))
  series2 <- series2[series.yrs2%in%yrs]
  series.yrs2 <- as.numeric(names(series2))


  ## skeleton
  master.skel <- cbind(master.yrs,xskel.calc(master))
  master.skel <- master.skel[master.skel[,1]%in%yrs,]
  master.yrs.sig <- master.skel[!is.na(master.skel[,2]),1]
  series.skel <- cbind(series.yrs2,xskel.calc(series2))
  series.skel <- series.skel[series.skel[,1]%in%yrs,]
  series.yrs.sig <- series.skel[!is.na(series.skel[,2]),1]

  ## divide in half
  first.half <- 1:cen.win
  second.half <- (cen.win + 1):win.width
  first.yrs <- yrs[first.half]
  second.yrs <- yrs[second.half]
  master.early <- master[first.half]
  series.early <- series2[first.half]
  master.late <- master[second.half]
  series.late <- series2[second.half]

  ## subset skel data
  early.series.skel <- series.skel[series.skel[,1]%in%first.yrs,]
  early.series.yrs.sig <- early.series.skel[!is.na(early.series.skel[,2]),1]

  early.master.skel <- master.skel[master.skel[,1]%in%first.yrs,]
  early.master.yrs.sig <- early.master.skel[!is.na(early.master.skel[,2]),1]

  late.series.skel <- series.skel[series.skel[,1]%in%second.yrs,]
  late.series.yrs.sig <- late.series.skel[!is.na(late.series.skel[,2]),1]

  late.master.skel <- master.skel[master.skel[,1]%in%second.yrs,]
  late.master.yrs.sig <- late.master.skel[!is.na(late.master.skel[,2]),1]


  ## ccf
  if(series.x){
    ccf.early <- as.vector(ccf(x=series.early,y=master.early,lag.max=5,plot=FALSE)$acf)
    ccf.late <- as.vector(ccf(x=series.late,y=master.late,lag.max=5,plot=FALSE)$acf)
  }
  else{
    ccf.early <- as.vector(ccf(x=master.early,y=series.early,lag.max=5,plot=FALSE)$acf)
    ccf.late <- as.vector(ccf(x=master.late,y=series.late,lag.max=5,plot=FALSE)$acf)
  }
  pcrit=0.05
  sig <- qnorm(1 - pcrit / 2) / sqrt(length(master.early))
  sig <- c(-sig, sig)

  ## cor and skel agreement
  overall.r <- round(cor(series2,master),3)
  early.r <- round(cor(series.early,master.early),3)
  late.r <- round(cor(series.late,master.late),3)

  ## aggreement btwn series skel and master skel
  overall.agree <- sum(series.yrs.sig%in%master.yrs.sig)/length(master.yrs.sig)
  overall.agree <- round(overall.agree*100,1)

  early.agree <- sum(early.series.yrs.sig%in%early.master.yrs.sig)/length(early.master.yrs.sig)
  early.agree <- round(early.agree*100,1)

  late.agree <- sum(late.series.yrs.sig%in%late.master.yrs.sig)/length(late.master.yrs.sig)
  late.agree <- round(late.agree*100,1)

  ## build page for plotting
  grid.newpage()
  fontsize <- 12       # fontsize for all text
  pointsize <- 12      # fontsize for grid.points()
  textJust <- "center" # justification for horizontal text elements
  col1light <- "lightgreen"
  col1dark <- "darkgreen"
  col2light <- "lightblue"
  col2dark <- "darkblue"
  ## 1.0 a bounding box for margins
  bnd.vp <- plotViewport(margins=rep(0.5,4), # 1/2 line margin
                         name = "bnd.vp",
                         gp = gpar(fontsize = fontsize))
  ## go from bottom up.

  ## 2.1 bounding box for ccf early: 30% of area height inside bnd.vp
  ccf.early.bnd.vp <- viewport(x = 0, y = 0, width = 0.5, height = 0.3,
                               just = c("left", "bottom"),
                               name = "ccf.early.bnd.vp")
  ## 2.12 plotting region for ccf early. 1 line margin bottom. 2 lines left
  ccf.early.region.vp <- plotViewport(margins=c(1,2,0,0),
                                      xscale=c(0,12),
                                      yscale=c(-1,1),
                                      name = "ccf.early.region.vp")
  ## 2.2 bounding box for ccf late: 30% of area height inside bnd.vp
  ccf.late.bnd.vp <- viewport(x = 0.5, y = 0, width = 0.5, height = 0.3,
                              just = c("left", "bottom"),
                              name = "ccf.late.bnd.vp")
  ## 2.22 plotting region for ccf late. 1 line margin bottom. 2 lines right
  ccf.late.region.vp <- plotViewport(margins=c(1, 0, 0, 2),
                                     xscale=c(0,12),
                                     yscale=c(-1,1),
                                     name = "ccf.late.region.vp")

  ## 3.0 box for text comparing early and late periods. 10% area height
  text.bnd.vp <- viewport(x = 0, y = 0.3, width = 1, height = 0.1,
                          just = c("left", "bottom"), name = "text.bnd.vp")

  ## 4.1 bounding box for skeleton plot. 55% of area
  skel.bnd.vp <- viewport(x = 0, y = 0.4, width = 1, height = 0.55,
                          just = c("left", "bottom"), name = "skel.bnd.vp")
  ## 4.2 plotting region for skeleton plot. 2 lines left and right.
  ## 3 lines on top and bottom
  skel.region.vp <- plotViewport(margins=c(3,2,3,2),
                                 xscale=c(min(yrs)-0.5,max(yrs)+0.5),
                                 yscale=c(-10,10),
                                 name = "skel.region.vp")
  ## 5.0 a box for overall text. 5%
  overall.txt.vp <- viewport(x = 0, y = 0.95, width = 1, height = 0.05,
                             just = c("left", "bottom"),
                             name = "overall.txt.vp")



  ## actual plotting
  dev.hold()
  on.exit(dev.flush())
  pushViewport(bnd.vp) # inside margins
  pushViewport(skel.bnd.vp) # inside skel
  pushViewport(skel.region.vp) # inside margins
  grid.rect(gp = gpar(col=col1light, lwd=1))
  grid.grill(h = unit(seq(-10, 10, by=1), "native"),
             v = unit(yrs-0.5, "native"),
             gp = gpar(col=col1light, lineend = "square",
                       linejoin = "round"))
  ## rw plot
  grid.rect(x = yrs, y = 0, width = 1, height = 2 * master,
            hjust = 0.5, vjust = 1, default.units = "native",
            gp=gpar(fill=col1light,col=col1dark))
  grid.rect(x = yrs, y = 0, width = 1, height = 2 * series2,
            hjust = 0.5, vjust = 0, default.units = "native",
            gp=gpar(fill=col1light,col=col1dark))

  ## master
  grid.segments(x0=master.yrs.sig,y0=0,
                x1=master.yrs.sig,y1=-10,
                default.units="native",
                gp=gpar(lwd=1,col='black',lineend="butt"))
  grid.segments(x0=master.skel[,1],y0=0,
                x1=master.skel[,1],y1=master.skel[,2]*-1,
                default.units="native",
                gp=gpar(lwd=5,col='black',lineend="butt"))
  ## series
  grid.segments(x0=series.yrs.sig,y0=0,
                x1=series.yrs.sig,y1=10,
                default.units="native",
                gp=gpar(lwd=1,col='black',lineend="butt"))
  grid.segments(x0=series.skel[,1],y0=0,
                x1=series.skel[,1],y1=series.skel[,2],
                default.units="native",
                gp=gpar(lwd=5,col='black',lineend="butt"))

  ## text
  grid.text(master.yrs.sig, x=unit(master.yrs.sig,"native"),
            y = unit(0, "npc"), rot = 90,just="right")
  grid.text(series.yrs.sig, x=unit(series.yrs.sig,"native"),
            y = unit(1, "npc"), rot = 90,just="left")
  grid.text(gettext("Master", domain="R-dplR"),x=unit(0,"npc"),
            y=unit(0,"npc"),hjust = 0,vjust = 0,rot=90)
  grid.text(gettext("Series", domain="R-dplR"),x=unit(0,"npc"),
            y=unit(1,"npc"),hjust=1,vjust=0,rot=90)

  popViewport(2) # back to bnd

  negText <- textGrob(gettext("(Negative)", domain="R-dplR"),
                      y=unit(0.25,"lines"),x=unit(3,"native"),
                      just = textJust)
  posText <- textGrob(gettext("(Positive)", domain="R-dplR"),
                      y=unit(0.25,"lines"),x=unit(9,"native"),
                      just = textJust)
  

  for (period in c("early", "late")) {
      if (period == "early") {
          vp1 <- ccf.early.bnd.vp
          vp2 <- ccf.early.region.vp
          ccf.period <- ccf.early
      } else {
          vp1 <- ccf.late.bnd.vp
          vp2 <- ccf.late.region.vp
          ccf.period <- ccf.late
      }
      pushViewport(vp1) # into ccf
      pushViewport(vp2) # inside margins
      grid.grill(v = unit(seq(1, 11, by=1), "native"),
                 h=NA,
                 gp = gpar(col=col2light, lineend = "square",
                 linejoin = "round"))
      grid.segments(x0=unit(c(0, 0), "native"),y0=unit(sig, "native"),
                    x1=unit(c(12, 12), "native"),y1=unit(sig, "native"),
                    gp=gpar(col=col2dark, lty="dashed",lwd=2))
      grid.segments(x0=unit(c(0, 6), "native"),y0=unit(c(0, -1), "native"),
                    x1=unit(c(12, 6), "native"),y1=unit(c(0, 1), "native"),
                    gp=gpar(col="black", lty="solid",lwd=1))
      grid.segments(x0=1:11,y0=0,x1=1:11,y1=ccf.period,
                    default.units="native",
                    gp=gpar(lwd=2,lend="butt", col=col2dark))
      grid.points(x=1:11, y=ccf.period, pch=21,
                  default.units="native",
                  gp=gpar(fill=col2light, col=col2dark,
                  fontsize=pointsize))
      grid.draw(negText)
      grid.draw(posText)
      popViewport(2) # back to bnd
  }

  periodPattern <- gettext("Period: %d-%d", domain = "R-dplR")
  agreePattern <- gettext("Skeleton Agreement %s%%", domain = "R-dplR")

  grid.segments(x0=0.5,y0=0.05,x1=0.5,y1=0.95,
                default.units="npc",
                gp=gpar(lwd=2,lend="butt", col="black"))
  pushViewport(text.bnd.vp) # description
  tmp.txt <- substitute(period * ", " * r[lag0] == corr,
                        list(period = sprintf(periodPattern,
                             min(first.yrs), max(first.yrs)),
                             corr = early.r))
  
  grid.text(tmp.txt,y=unit(0.65,"npc"),x=unit(0.25,"npc"),
            just = textJust)

  grid.text(sprintf(agreePattern, early.agree),
            y=unit(0.35,"npc"), x=unit(0.25,"npc"),
            just = textJust)


  tmp.txt <- substitute(period * ", " * r[lag0] == corr,
                        list(period = sprintf(periodPattern,
                             min(second.yrs), max(second.yrs)),
                             corr = late.r))
  grid.text(tmp.txt,y=unit(0.65,"npc"),x=unit(0.75,"npc"),
            just = textJust)

  grid.text(sprintf(agreePattern, late.agree),
            y=unit(0.35,"npc"), x=unit(0.75,"npc"),
            just = textJust)

  popViewport(1) # back to bnd

  pushViewport(overall.txt.vp) # description
  tmp.txt <- substitute(period * ", " * r[lag0] == corr * ", " * agree,
                        list(period = sprintf(periodPattern,
                             min(yrs), max(yrs)),
                             corr = overall.r,
                             agree = sprintf(agreePattern, overall.agree)))
  grid.text(tmp.txt,y=unit(0.5,"npc"),x=unit(0.5,"npc"),
            just = textJust)
  
  
  popViewport(2)
  series.x.txt <- ifelse(series.x,
                         "NB: With series.x = TRUE, postive lags indicate missing rings in series",
                         "NB: With series.x = FALSE (default), negative lags indicate missing rings in series")
  grid.text(series.x.txt,y=unit(0.015,"npc"),x=unit(0.5,"npc"),
            just = textJust)
  
}

`write.crn` <- function(crn, fname, header=NULL, append=FALSE)
{
    stopifnot(is.data.frame(crn))
    if (ncol(crn) != 2) {
        stop("'crn' must have 2 columns")
    }
    cnames <- names(crn)
    stopifnot(is.character(cnames), !is.na(cnames),
              Encoding(cnames) != "bytes")
    crn2 <- crn

    if (any(is.na(crn2))) {
        na.flag <- is.na(crn2[, 1])
        crn2[na.flag, 2] <- 0
        crn2[na.flag, 1] <- 9.99
        print(head(crn2))
    }

    if (append) {
        if (!file.exists(fname)) {
            stop(gettextf("file %s does not exist, cannot append", fname))
        }
        if (length(header) > 0) {
            stop("bad idea to append with 'header'")
        }
    }
    if (length(header) > 0){
        if (!is.list(header)) {
            stop("header must be a list")
        }
        header.names <-
            c("site.id", "site.name", "spp.code", "state.country",
              "spp", "elev", "lat", "long", "first.yr", "last.yr",
              "lead.invs", "comp.date")
        if (!all(header.names %in% names(header))) {
            stop("'header' must be a list with the following names: ",
                 paste(dQuote(header.names), collapse = ", "))
        }
        ## Record #1: 1-6 Site ID, 10-61 Site Name, 62-65 Species Code,
        ## optional ID#s
        ## Record #2: 1-6 Site ID, 10-22 State/Country, 23-40 Species,
        ## 41-45 Elevation, 48-57 Lat-Long, 68-76 1st & last Year
        ## Note: lat-lons are in degrees and minutes, ddmm or dddmm
        ## Record #3: 1-6 Site ID, 10-72 Lead Investigator, 73-80
        ## comp. date
        header2 <- vapply(lapply(header, as.character), "[", character(1), 1)
        stopifnot(!is.na(header2), Encoding(header2) != "bytes")
        header2["lat.long"] <- if (nchar(header2["long"]) > 5) {
            paste0(header2["lat"], header2["long"])
        } else {
            paste(header2["lat"], header2["long"], sep=" ")
        }
        header2["yrs"] <-
            paste(header2["first.yr"], header2["last.yr"], sep=" ")

        field.name <-
            c("site.id", "site.name", "spp.code", "state.country", "spp",
              "elev", "lat.long", "yrs", "lead.invs", "comp.date")
        field.width <- c(6, 52, 4, 13, 18, 5, 10, 9, 63, 8)
        for (i in seq_along(field.name)) {
            this.name <- field.name[i]
            this.width <- field.width[i]
            this.var <- header2[this.name]
            this.nchar <- nchar(this.var)
            if (this.nchar > this.width) {
                header2[this.name] <- substr(this.var, 1, this.width)
            } else if (this.nchar < this.width) {
                header2[this.name] <-
                    encodeString(this.var, width = this.width)
            }
        }

        hdr1 <- paste0(header2["site.id"], "   ", header2["site.name"],
                       header2["spp.code"])
        hdr2 <- paste0(header2["site.id"], "   ", header2["state.country"],
                       header2["spp"], header2["elev"], "  ",
                       header2["lat.long"], "          ", header2["yrs"])
        hdr3 <- paste0(header2["site.id"], "   ", header2["lead.invs"],
                       header2["comp.date"])
        hdr <- c(hdr1, hdr2, hdr3)
    }

    yrs <- as.numeric(row.names(crn2))
    decades.vec <- yrs%/%10 * 10
    decades <- unique(decades.vec)
    n.decades <- length(decades)
    ## 1-6
    crn.name <- cnames[1]
    crn.width <- nchar(crn.name)
    ## If crn.width > 6, truncate
    if (crn.width > 6) {
        crn.name <- substr(crn.name, 1, 6)
    } else if (crn.width < 6) {# Pad to nchar 6 (no leading zero)
        crn.name <- formatC(crn.name, width = 6, format = "f")
    }

    dec.str <- character(n.decades)
    for (i in seq_len(n.decades)) {
        ## 7-10 decade column
        dec <- decades[i]
        dec.idx <- which(decades.vec %in% dec)
        n.yrs <- length(dec.idx)
        dec.yrs <- yrs[dec.idx]
        ## Pad to nchar 4 (no leading zero)
        dec.yrs <- formatC(dec.yrs, digits = 0, width = 4, format = "f")

        dec.rwi <- crn2[dec.idx, 1]
        ## Pad to nchar 4 (no leading zero)
        dec.rwi <- round(dec.rwi, 3) * 1000
        dec.rwi <- formatC(dec.rwi, digits = 0, width = 4, format = "f")

        ## Pad to nchar 3 (no leading zero)
        dec.samp.depth <- crn2[dec.idx, 2]
        dec.samp.depth <- formatC(dec.samp.depth,
                                  digits = 0, width = 3, format = "f")
        ## Pad front end
        if (i == 1) {
            tmp <- paste(rep("9990  0", 10-n.yrs), collapse="")
        } else {
            tmp <- ""
        }
        ## Put it all together
        dec.str[i] <-
            paste0(crn.name, dec.yrs[1], tmp,
                   paste0(dec.rwi, dec.samp.depth, collapse=""))
    }
    ## Finish last decade with 9990 as NA and 0 as samp depth.
    dec.str[i] <- paste0(dec.str[i],
                         paste(rep("9990  0", 10-n.yrs), collapse=""))
    if (length(header) > 0) {
        dec.str <- c(hdr, dec.str)
    }
    cat(dec.str, file = fname, sep = "\n", append=append)
    fname
}

powt <- function (rwl, method = "universal", rescale = FALSE,
                  return.power=FALSE) 
{
  # add a check for negative nums
  if(any(rwl < 0, na.rm = TRUE)) {
    stop("'rwl' values cannot be negative")
  }
  # check to see that rwl is either a data.frame or a numeric vector
  if(!(is.data.frame(rwl) || (is.numeric(rwl) && is.vector(rwl))) )
    stop("'rwl' must be a data.frame or a numeric vector")
  if (!is.logical(rescale))
    stop("'rescale' must be either FALSE (the default) or TRUE")
  if (!is.logical(return.power))
    stop("'return.power' must be either FALSE (the default) or TRUE")
  # get method
  if( !(is.character(method) && length(method) == 1) )
    stop("'method' must be a character vector of length 1")
  
  known.methods <- c("universal", "cook")
  
  method2 <- match.arg(arg = method,
                       choices = known.methods,
                       several.ok = FALSE)
  
  if (is.vector(rwl) && method2 != "cook") {
    stop("If rwl is a vector, method must be 'cook'")
  }
  
  # call function
  if (is.vector(rwl) && method2 == "cook") {
    res <- capPOWTseries(series = rwl, rescale = rescale, 
                         return.power = return.power)
  }
  
  if (is.data.frame(rwl) && method2 == "cook") {
    res <- capPOWT(rwl = rwl, rescale = rescale, 
                   return.power = return.power)
  }
  
  if (is.data.frame(rwl) && method2 == "universal") {
    res <- universalPOWT(rwl = rwl, rescale = rescale, 
                         return.power = return.power)
  }
  
  return(res)
}

corr.rwl.seg <- function(rwl, seg.length = 50, bin.floor = 100, n = NULL,
                         prewhiten = TRUE, pcrit=0.05, biweight = TRUE,
                         method = c("spearman", "pearson", "kendall"),
                         make.plot = TRUE, label.cex=1,
                         floor.plus1 = FALSE, master = NULL,
                         master.yrs = as.numeric(if (is.null(dim(master))) {
                           names(master)
                         } else {
                           rownames(master)
                         }),
                         ...) {
  method2 <- match.arg(method)
  ## run error checks
  qa.xdate(rwl, seg.length, n, bin.floor)
  
  ## turn off warnings for this function
  ## The sig test for spearman's rho often produces warnings.
  w <- options(warn = -1)
  on.exit(options(w))
  
  nseries <- length(rwl)
  if (is.null(master) && nseries < 2) {
    stop("At least 2 series are needed in 'rwl'")
  } else if (nseries < 1) {
    stop("At least 1 series is needed in 'rwl'")
  }
  
  cnames <- names(rwl)
  yrs <- as.numeric(row.names(rwl))
  min.yr <- min(yrs)
  max.yr <- max(yrs)
  rwl2 <- rwl
  
  ## Ensure that rwl has consecutive years in increasing order
  if (!all(diff(yrs) == 1)) {
    yrs <- min.yr : max.yr
    rwl2 <- matrix(NA_real_,
                   nrow = max.yr - min.yr + 1,
                   ncol = nseries,
                   dimnames = list(as.character(yrs), cnames))
    rwl2[row.names(rwl), ] <- as.matrix(rwl)
    rwl2 <- as.data.frame(rwl2)
  }
  
  ## Pad rwl and master (if present) to same number of years
  if (!is.null(master)) {
    master.dim <- dim(master)
    min.master.yr <- min(master.yrs)
    max.master.yr <- max(master.yrs)
    
    if (!is.null(master.dim) && length(master.dim) == 2 &&
        master.dim[2] > 1) {
      ## A. master is a data.frame or a matrix.  Normalize and
      ## compute master chronology as a mean of series
      ## (columns).
      
      ## Ensure that master has consecutive years in increasing order
      if (!all(diff(master.yrs) == 1)) {
        char.yrs <- as.character(min.master.yr : max.master.yr)
        master.inc <- matrix(NA_real_,
                             nrow = max.master.yr - min.master.yr + 1,
                             ncol = master.dim[2],
                             dimnames = list(char.yrs,
                                             colnames(master)))
        master.inc[rownames(master), ] <- as.matrix(master)
      } else {
        master.inc <- master
      }
      
      ## normalize all series (columns in master matrix)
      tmp <- normalize1(master.inc, n, prewhiten)
      master.norm <- tmp$rwi.mat[, tmp$idx.good, drop=FALSE]
      
      ## compute master series by normal mean or robust mean
      if (!biweight) {
        master2 <- apply(master.norm, 1, exactmean)
      } else {
        master2 <- apply(master.norm, 1, tbrm, C=9)
      }
    } else {
      ## B. master is a vector
      master2 <- rep(NA_real_, max.master.yr - min.master.yr + 1)
      names(master2) <- as.character(min.master.yr : max.master.yr)
      master2[as.character(master.yrs)] <- master
    }
    
    if (min.master.yr < min.yr) {
      n.pad <- min.yr - min.master.yr
      padding <- matrix(NA_real_, n.pad, nseries)
      colnames(padding) <- cnames
      rownames(padding) <- min.master.yr : (min.yr - 1)
      rwl2 <- rbind(padding, rwl2)
      min.yr <- min.master.yr
    } else if (min.master.yr > min.yr) {
      n.pad <- min.master.yr - min.yr
      padding <- rep(NA_real_, n.pad)
      names(padding) <- min.yr : (min.master.yr - 1)
      master2 <- c(padding, master2)
    }
    if (max.master.yr < max.yr) {
      n.pad <- max.yr - max.master.yr
      padding <- rep(NA_real_, n.pad)
      names(padding) <- (max.master.yr + 1) : max.yr
      master2 <- c(master2, padding)
    } else if (max.master.yr > max.yr) {
      n.pad <- max.master.yr - max.yr
      padding <- matrix(NA_real_, n.pad, nseries)
      colnames(padding) <- cnames
      rownames(padding) <- (max.yr + 1) : max.master.yr
      rwl2 <- rbind(rwl2, padding)
      max.yr <- max.master.yr
    }
    yrs <- min.yr : max.yr
  }
  
  seg.lag <- seg.length / 2
  nyrs <- length(yrs)
  if (is.null(bin.floor) || bin.floor == 0) {
    min.bin <- min.yr
  } else if (floor.plus1) {
    min.bin <- ceiling((min.yr - 1) / bin.floor) * bin.floor + 1
  } else {
    min.bin <- ceiling(min.yr / bin.floor) * bin.floor
  }
  max.bin <- max.yr - seg.length + 1
  if (max.bin < min.bin) {
    stop("shorten 'seg.length' or adjust 'bin.floor'")
  }
  bins <- seq(from=min.bin, to=max.bin, by=seg.lag)
  bins <- cbind(bins, bins + (seg.length - 1), deparse.level=0)
  nbins <- nrow(bins)
  bin.names <- paste0(bins[, 1], ".", bins[, 2])
  ## structures for results
  res.cor <- matrix(NA, nseries, nbins)
  rownames(res.cor) <- cnames
  colnames(res.cor) <- bin.names
  
  res.pval <- matrix(NA, nseries, nbins)
  rownames(res.pval) <- cnames
  colnames(res.pval) <- bin.names
  
  overall.cor <- matrix(NA, nseries, 2)
  rownames(overall.cor) <- cnames
  colnames(overall.cor) <- c("rho", "p-val")
  
  ## normalize all series
  norm.one <- normalize1(rwl2, n, prewhiten)
  ## rwi for segments altered by normalizing
  rwi <- norm.one$rwi.mat # is a matrix
  idx.good <- norm.one$idx.good
  
  ## loop through series
  seq.series <- seq_len(nseries)
  for (i in seq.series) {
    if (is.null(master)) {
      idx.noti <- rep(TRUE, nseries)
      idx.noti[i] <- FALSE
      master.norm <- rwi[, idx.good & idx.noti, drop=FALSE]
      
      ## compute master series by normal mean or robust mean
      if (!biweight) {
        master2 <- apply(master.norm, 1, exactmean)
      } else {
        master2 <- apply(master.norm, 1, tbrm, C=9)
      }
    }
    series <- rwi[, i]
    ## loop through bins
    for (j in seq_len(nbins)) {
      mask <- yrs %in% seq(from=bins[j, 1], to=bins[j, 2])
      ## cor is NA if there is not complete overlap
      if (!any(mask) ||
          any(is.na(series[mask])) ||
          any(is.na(master2[mask]))) {
        bin.cor <- NA
        bin.pval <- NA
      } else {
        tmp <- cor.test(series[mask], master2[mask],
                        method = method2, alternative = "greater")
        bin.cor <- tmp$estimate
        bin.pval <- tmp$p.val
      }
      res.cor[i, j] <- bin.cor
      res.pval[i, j] <- bin.pval
    }
    ## overall correlation
    tmp <- cor.test(series, master2,
                    method = method2, alternative = "greater")
    overall.cor[i, 1] <- tmp$estimate
    overall.cor[i, 2] <- tmp$p.val
  }
  ## avg seg correlation
  segavg.cor <- colMeans(res.cor, na.rm=TRUE)
  
  ## make a list of problem segments
  seg.flags <- rep(NA, nseries)
  names(seg.flags) <- cnames
  flag.logical <- res.pval >= pcrit
  flag.logical[is.na(flag.logical)] <- FALSE
  for (i in seq_along(seg.flags)) {
    seg.flags[i] <- paste(bin.names[flag.logical[i, ]], collapse = ", ")
  }
  seg.flags <- seg.flags[seg.flags != ""]
  
  res <- list(spearman.rho = res.cor, p.val = res.pval, overall = overall.cor,
              avg.seg.rho = segavg.cor, flags = seg.flags, bins = bins, 
              rwi = rwi, seg.lag = seg.lag, seg.length = seg.length, 
              pcrit = pcrit, label.cex = label.cex)
  class(res) <- c("list","crs")
  
  
  ## plot
  if (make.plot) {
    plot(res)
  }
  res
}

capPOWTseries <- function (series, rescale = FALSE, return.power=FALSE) 
{
  
  # add a check for negative nums
  if(any(series < 0,na.rm = TRUE)) {
    stop("'series' values cannot be negative")
  }
  
  #  check rescale
  if (!is.logical(rescale))
    stop("'rescale' must be either FALSE (the default) or TRUE")
  
  # helpers
  # used to set min numb to zeros.
  getprec <- function(series) {
    series.num <- as.numeric(series)
    series.num <- series.num[!is.na(series.num) & series.num != 0]
    if (length(series.num) == 0) {
      NA_real_
    }
    else {
      series.char <- format(series.num, scientific = FALSE)
      if (grepl(".", series.char[1], fixed = TRUE)) {
        maxdig <- nchar(sub("^[^.]*\\.", "", series.char[1]))
      }
      else {
        rm.trail.zeros <- sub("0*$", "", series.char)
        n.trail.zeros <- nchar(series.char) - nchar(rm.trail.zeros)
        maxdig <- -min(n.trail.zeros)
      }
      10^-maxdig
    }
  }
  
  # to get p
  fit.lm <- function(series) {
    n <- length(series)
    drop.1 <- series[-1]
    drop.n <- series[-n]
    runn.M <- (drop.1 + drop.n)/2
    runn.S <- abs(drop.1 - drop.n)
    runn.S[runn.S == 0] <- prec
    runn.M[runn.M == 0] <- prec
    mod <- lm.fit(cbind(1, log(runn.M)), log(runn.S))
    b <- mod[["coefficients"]][2]
    1 - b
  }
  
  # do the trans
  transf <- function(x) {
    Xt <- x
    X.nna <- which(!is.na(x))
    X <- na.omit(x)
    p <- abs(fit.lm(X))
    X2 <- X^p
    Xt[X.nna] <- X2
    res <- list(Xt=Xt,p=p) 
    return(res)
  }
  
  prec <- getprec(series)
  
  xt <- transf(series)[['Xt']]
  p <-  transf(series)[['p']]
  if(rescale){
    xtNames <- names(xt)
    xt <- c(scale(xt) * sd(series,na.rm = TRUE) + mean(series,na.rm = TRUE))
    names(xt) <- xtNames
  }
  res <- xt
  
  if(return.power==TRUE){ #optional output of the power coefficient, default is FALSE
    names(p) <- NULL
    res <- list(transformed.data=res,power=p)
  }
  
  return(res)
}

bai.in <- function(rwl, d2pith = NULL) {
    
    if(!is.data.frame(rwl))
        stop("'rwl' must be a data.frame")
    
    if(!is.null(d2pith)) {
        if(ncol(rwl) != nrow(d2pith))
            stop("dimension problem: ", "'ncol(rw)' != 'nrow(d2pith)'")
        if(!all(d2pith[, 1] == names(rwl))){
            print(data.frame(rwlNames=names(rwl),seriesID=d2pith[,1],test=d2pith[, 1] == names(rwl)))
            stop("series ids in 'd2pith' and 'rwl' do not match exactly.")
        }
        if(!is.data.frame(d2pith)){
            stop("'d2pith' must be a data.frame")
        }
        d2pith.vec <- d2pith[, 2]
    } else {
        ## distance offset if not given
        d2pith.vec <- rep(0, ncol(rwl))
    }
    
    out <- rwl
    ## vector of years
    n.vec <- seq_len(nrow(rwl))
    for(i in seq_len(ncol(rwl))){
        ## series to work with
        dat <- rwl[[i]]
        ## strip out data from NA
        dat2 <- na.omit(dat)
        ## get ring area
        bai <- pi*dat2*(dat2+2*(cumsum(dat2) + d2pith.vec[i] - dat2))
        ## find NA / not NA locations
        na <- attributes(dat2)$na.action
        no.na <- n.vec[!n.vec %in% na]
        ## write result
        out[no.na, i] <- bai
    }
    ## return result
    out
}

net <- function(x, weights = c(v=1, g=1)) {
    stopifnot(is.numeric(weights), is.finite(weights))
    weights2 <- vecMatched(weights, c("v", "g"))
    dimX <- dim(x)
    if (is.null(dimX) || length(dimX) != 2) {
        stop("'x' must be a matrix-like object")
    }
    if (!isTRUE(all(dimX >= 2))) {
        stop("'x' must have at least 2 rows and 2 columns")
    }
    x2 <- as.matrix(x)
    if (!is.numeric(x2)) {
        stop("'x' must contain numeric data")
    }
    ## Standard deviation standardized by mean
    variability <- function(mat) {
        Sd <- rowSds(mat, na.rm = TRUE)
        Mean <- rowMeans(mat, na.rm = TRUE)
        Sd / Mean
    }
    ## Gleichlaufigkeit as in the NET paper by Esper et al.
    gleichlauf <- function(mat) {
        delta <- diff(mat)
        isNA <- is.na(delta)
        N <- ncol(mat) - rowSums(isNA)
        delta[isNA] <- 0
        pos <- rowSums(delta > 0)
        neg <- rowSums(delta < 0)
        res <- c(NA_real_, pmax(pos, neg) / N)
        names(res) <- rownames(mat)
        res
    }
    w1 <- weights2[1]
    w2 <- weights2[2]
    do1 <- w1 != 0
    do2 <- w2 != 0
    NetJ <- if (do1 && do2) {
        w1 * variability(x2) + w2 * (1 - gleichlauf(x2))
    } else if (do1) {
        w1 * variability(x2)
    } else if (do2) {
        w2 * (1 - gleichlauf(x2))
    } else {
        structure(rep.int(NA_real_, dimX[1]), names = rownames(x2))
    }
    Net <- mean(NetJ, na.rm = TRUE)
    list(all = NetJ, average = Net)
}

`read.tucson` <- function(fname, header = NULL, long = FALSE,
                          encoding = getOption("encoding"),
                          edge.zeros = TRUE,verbose=TRUE)
{
  ## Checks that the input is good. The input variables are vectors
  ## ('series', 'decade.yr') or matrices ('x') containing most of
  ## the data acquired from the input file 'fname'.
  input.ok <- function(series, decade.yr, x) {
    if (length(series) == 0) {
      return(FALSE)
    }
    
    ## Number of values allowed per row depends on first year modulo 10
    n.per.row <-
      apply(x, 1,
            function(x) {
              notna <- which(!is.na(x))
              n.notna <- length(notna)
              if (n.notna == 0) {
                0
              } else {
                notna[n.notna]
              }
            })
    full.per.row <- 10 - decade.yr %% 10
    ## One extra column per row is allowed:
    ## a. enough space will be allocated (max.year is larger than
    ## last year of any series)
    ## b. the extra col may contain a stop marker (non-standard location)
    idx.bad <- which(n.per.row > full.per.row + 1)
    n.bad <- length(idx.bad)
    if (n.bad > 0) {
      warn.fmt <-
        ngettext(n.bad,
                 "%d row has too many values (ID, decade %s)",
                 "%d rows have too many values (IDs, decades %s)",
                 domain="R-dplR")
      if (n.bad > 5) {
        idx.bad <- sample(idx.bad, 5)
        ids.decades <- paste(paste(series[idx.bad], decade.yr[idx.bad],
                                   sep=", ", collapse="; "),
                             "...", sep="; ")
      } else {
        ids.decades <- paste(series[idx.bad], decade.yr[idx.bad],
                             sep="-", collapse=", ")
      }
      warning(sprintf(warn.fmt, n.bad, ids.decades), domain=NA)
      return(FALSE)
    }
    series.ids <- unique(series)
    nseries <- length(series.ids)
    series.index <- match(series, series.ids)
    last.row.of.series <- logical(length(series))
    for (i in seq_len(nseries)) {
      idx.these <- which(series.index == i)
      last.row.of.series[idx.these[which.max(decade.yr[idx.these])]] <-
        TRUE
    }
    flag.bad2 <- n.per.row < full.per.row
    if (!all(last.row.of.series) && all(flag.bad2[!last.row.of.series])) {
      warning("all rows (last rows excluded) have too few values")
      return(FALSE)
    }
    min.year <- min(decade.yr)
    max.year <- ((max(decade.yr)+10) %/% 10) * 10
    if (max.year > as.numeric(format(Sys.Date(), "%Y")) + 100) {
      ## Must do something to stop R from trying to build huge
      ## data structures if the maximum year is not detected
      ## correctly.  Not too strict (allow about 100 years past
      ## today).
      warning("file format problems (or data from the future)")
      return(FALSE)
    }
    # look for duplicate IDs -- common problem with Tucson RWL files
    span <- max.year - min.year + 1
    val.count <- matrix(0, span, nseries)
    for (i in seq_along(series)) {
      this.col <- series.index[i]
      these.rows <- seq(from = decade.yr[i] - min.year + 1, by = 1,
                        length.out = n.per.row[i])
      val.count[these.rows, this.col] <-
        val.count[these.rows, this.col] + 1
    }
    extra.vals <- which(val.count > 1, arr.ind=TRUE)
    n.extra <- nrow(extra.vals)
    # if (n.extra > 0) {
    #   warn.fmt <-
    #     ngettext(n.bad,
    #              "Duplicated series ID detected with overlap in %d pair of ID, year: %s",
    #              "Duplicated series ID detected with overlaps in %d pairs of ID, year: %s",
    #              domain="R-dplR")
    #   if (n.extra > 5) {
    #     extra.vals <- extra.vals[sample(n.extra, 5), ]
    #     ids.years <- paste(paste(series.ids[extra.vals[, 2]],
    #                              min.year - 1 + extra.vals[, 1],
    #                              sep=", ", collapse="; "),
    #                        "...", sep="; ")
    #   } else {
    #     ids.years <- paste(series.ids[extra.vals[, 2]],
    #                        min.year - 1 + extra.vals[, 1],
    #                        sep=", ", collapse="; ")
    #   }
    #   warning(sprintf(warn.fmt, n.extra, ids.years), domain=NA)
    #   FALSE
    # }
    ####################
    # simplifying error message to user to make it clearer that there are 
    # duplicate IDs
    if (n.extra > 0) {
      warn.fmt <- gettext("Duplicated series ID detected: %s",domain="R-dplR")
      ids.dup <- paste(unique(series.ids[extra.vals[, 2]]),
                       sep = ", ",collapse = "; ")
      warning(sprintf(warn.fmt, ids.dup), domain=NA)
      return(FALSE)
    }
    else {
      return(TRUE)
    }
  } # end input.ok func
  
  ## Read data file into memory
  con <- file(fname, encoding = encoding)
  on.exit(close(con))
  goodLines <- readLines(con)
  close(con)
  on.exit()
  ## Strip empty lines (caused by CR CR LF endings etc.)
  goodLines <- goodLines[nzchar(goodLines)]
  ## Remove comment lines (print them?)
  foo <- regexpr("#", goodLines, fixed=TRUE)
  commentFlag <- foo >= 1 & foo <= 78
  goodLines <- goodLines[!commentFlag]
  ## Temporary file for 'goodLines'. Reading from this file is
  ## faster than making a textConnection to 'goodLines'.
  tf <- tempfile()
  check.tempdir()
  tfcon <- file(tf, encoding="UTF-8")
  on.exit(close(tfcon))
  on.exit(unlink(tf), add=TRUE)
  writeLines(goodLines, tf)
  ## New connection for reading from the temp file
  close(tfcon)
  tfcon <- file(tf, encoding="UTF-8")
  if (is.null(header)) {
    ## Try to determine if the file has a header. This is failable.
    ## 3 lines in file
    hdr1 <- readLines(tfcon, n=1)
    if (length(hdr1) == 0) {
      stop("file is empty")
    }
    if (nchar(hdr1) < 12) {
      stop("first line in rwl file ends before col 12")
    }
    is.head <- FALSE
    yrcheck <- suppressWarnings(as.numeric(substr(hdr1, 9, 12)))
    if (is.null(yrcheck) || length(yrcheck) != 1 || is.na(yrcheck) ||
        yrcheck < -1e04 || yrcheck > 1e04 ||
        round(yrcheck) != yrcheck) {
      is.head <- TRUE
    }
    if (!is.head) {
      datacheck <- substring(hdr1,
                             seq(from=13, by=6, length=10),
                             seq(from=18, by=6, length=10))
      datacheck <- sub("^[[:blank:]]+", "", datacheck)
      idx.good <- which(nzchar(datacheck))
      n.good <- length(idx.good)
      if (n.good == 0) {
        is.head <- TRUE
      } else {
        datacheck <- datacheck[seq_len(idx.good[n.good])]
        if (any(grepl("[[:alpha:]]", datacheck))) {
          is.head <- TRUE
        } else {
          datacheck <- suppressWarnings(as.numeric(datacheck))
          if (is.null(datacheck) ||
              any(!is.na(datacheck) &
                  round(datacheck) != datacheck)) {
            is.head <- TRUE
          }
        }
      }
    }
    if (is.head) {
      hdr1.split <- strsplit(str_trim(hdr1, side="both"),
                             split="[[:space:]]+")[[1]]
      n.parts <- length(hdr1.split)
      if (n.parts >= 3 && n.parts <= 13) {
        hdr1.split <- hdr1.split[2:n.parts]
        if (!any(grepl("[[:alpha:]]", hdr1.split))) {
          yrdatacheck <- suppressWarnings(as.numeric(hdr1.split))
          if (!(is.null(yrdatacheck) ||
                any(!is.na(yrdatacheck) &
                    round(yrdatacheck) != yrdatacheck))) {
            is.head <- FALSE
          }
        }
      }
    }
    if (is.head) {
      if(verbose){
      cat(gettext("There appears to be a header in the rwl file\n",
                  domain="R-dplR"))
      }
    } else {
      if(verbose){
      cat(gettext("There does not appear to be a header in the rwl file\n",
                  domain="R-dplR"))
      }
    }
  } else if (!is.logical(header) || length(header) != 1 || is.na(header)) {
    stop("'header' must be NULL, TRUE or FALSE")
  } else {
    is.head <- header
  }
  
  skip.lines <- if (is.head) 3 else 0
  data1 <- readLines(tfcon, n=skip.lines + 1)
  if (length(data1) < skip.lines + 1) {
    stop("file has no data")
  }
  on.exit(unlink(tf))
  ## Test for presence of tabs
  if (!grepl("\t", data1[length(data1)])) {
    ## Using a connection instead of a file name in read.fwf and
    ## read.table allows the function to support different encodings.
    if (isTRUE(long)) {
      ## Reading 11 years per decade allows nonstandard use of stop
      ## marker at the end of a line that already has 10
      ## measurements.  Such files exist in ITRDB.
      fields <- c(7, 5, rep(6, 11))
    } else {
      fields <- c(8, 4, rep(6, 11))
    }
    ## First, try fixed width columns as in Tucson "standard"
    dat <-
      tryCatch(read.fwf(tfcon, widths=fields, skip=skip.lines,
                        comment.char="", strip.white=TRUE,
                        blank.lines.skip=FALSE,
                        colClasses=c("character", rep("integer", 11),
                                     "character")),
               error = function(...) {
                 ## If predefined column classes fail
                 ## (e.g. missing values marked with "."), convert
                 ## types manually
                 tfcon <- file(tf, encoding="UTF-8")
                 tmp <-
                   read.fwf(tfcon, widths=fields, skip=skip.lines,
                            strip.white=TRUE, blank.lines.skip=FALSE,
                            colClasses="character", comment.char="")
                 for (idx in 2:12) {
                   asnum <- as.numeric(tmp[[idx]])
                   if (!identical(round(asnum), asnum)) {
                     stop("non-integral numbers found")
                   }
                   tmp[[idx]] <- as.integer(asnum)
                 }
                 tmp
               })
    dat <- dat[!is.na(dat[[2]]), , drop=FALSE] # requires non-NA year
    series <- dat[[1]]
    decade.yr <- dat[[2]]
    series.fixed <- series
    decade.fixed <- decade.yr
    x <- as.matrix(dat[3:12])
    ## Convert values <= 0 or < 0 (not -9999) to NA
    if (isTRUE(edge.zeros)) {
      x[x < 0 & x != -9999] <- NA
    } else {
      x[x <= 0 & x != -9999] <- NA
    }
    x.fixed <- x
    fixed.ok <- input.ok(series, decade.yr, x)
  } else {
    warning("tabs used, assuming non-standard, tab-delimited file")
    fixed.ok <- FALSE
  }
  ## If that fails, try columns separated by white space (non-standard)
  if (!fixed.ok) {
    warning("fixed width failed, trying to reread with variable width columns")
    tfcon <- file(tf, encoding="UTF-8")
    ## Number of columns is decided by length(col.names)
    dat <-
      tryCatch(read.table(tfcon, skip=skip.lines, blank.lines.skip=FALSE,
                          comment.char="", col.names=letters[1:13],
                          colClasses=c("character", rep("integer", 11),
                                       "character"), fill=TRUE, quote=""),
               error = function(...) {
                 ## In case predefined column classes fail
                 tfcon <- file(tf, encoding="UTF-8")
                 tmp <- read.table(tfcon, skip=skip.lines,
                                   blank.lines.skip=FALSE, quote="",
                                   comment.char="", fill=TRUE,
                                   col.names=letters[1:13],
                                   colClasses="character")
                 tmp[[1]] <- as.character(tmp[[1]])
                 for (idx in 2:12) {
                   asnum <- as.numeric(tmp[[idx]])
                   if (!identical(round(asnum), asnum)) {
                     stop("non-integral numbers found")
                   }
                   tmp[[idx]] <- as.integer(asnum)
                 }
                 tmp
               })
    dat <- dat[!is.na(dat[[2]]), , drop=FALSE] # requires non-NA year
    series <- dat[[1]]
    decade.yr <- dat[[2]]
    x <- as.matrix(dat[3:12])
    if (isTRUE(edge.zeros)) {
      x[x < 0 & x != -9999] <- NA
    } else {
      x[x <= 0 & x != -9999] <- NA
    }
    if (!input.ok(series, decade.yr, x)) {
      if (exists("series.fixed", inherits=FALSE) &&
          exists("decade.fixed", inherits=FALSE) &&
          exists("x.fixed", inherits=FALSE) &&
          (any(is.na(x) != is.na(x.fixed)) ||
           any(x != x.fixed, na.rm=TRUE))) {
        series <- series.fixed
        decade.yr <- decade.fixed
        warning("trying fixed width names, years, variable width data")
        if (!input.ok(series, decade.yr, x)) {
          stop("failed to read rwl file")
        }
      } else {
        stop("failed to read rwl file")
      }
    }
  }
  series.ids <- unique(series)
  nseries <- length(series.ids)
  ## At this time match does not support long vectors in the second
  ## argument and always returns integers, but let's check the
  ## result anyway.
  series.index <- tryCatch(as.integer(match(series, series.ids)),
                           warning = conditionMessage,
                           error = conditionMessage)
  if (!is.integer(series.index)) {
    stop(gettextf("series.index must be integer: %s",
                  paste(as.character(series.index), collapse = ", "),
                  domain = "R-dplR"))
  }
  extra.col <- dat[[13]]
  
  res <- .Call(dplR.readloop, series.index, decade.yr, x)
  rw.mat <- res[[1]]
  min.year <- res[[2]]
  prec.rproc <- res[[3]]
  span <- nrow(rw.mat)
  if (span == 0) {
    rw.df <- as.data.frame(rw.mat)
    names(rw.df) <- as.character(series.ids)
    return(rw.df)
  }
  max.year <- min.year + (span - 1)
  rownames(rw.mat) <- min.year:max.year
  
  ## The operations in the loop depend on the precision of each series.
  ## It's not exactly clear whether the Tucson format allows mixed
  ## precisions in the same file, but we can support that in any case.
  prec.unknown <- logical(nseries)
  for (i in seq_len(nseries)) {
    if (!(prec.rproc[i] %in% c(100, 1000))) {
      these.rows <- which(series.index == i)
      these.decades <- decade.yr[these.rows]
      has.stop <- which(extra.col[these.rows] %in% c("999", "-9999"))
      if (length(has.stop) == 1 &&
          which.max(these.decades) == has.stop) {
        warning(gettextf("bad location of stop marker in series %s",
                         series.ids[i], domain="R-dplR"),
                domain=NA)
        if (extra.col[these.rows[has.stop]] == "999") {
          prec.rproc[i] <- 100
        } else {
          prec.rproc[i] <- 1000
        }
      }
    }
    this.prec.rproc <- prec.rproc[i]
    if (this.prec.rproc == 100) {
      ## Convert stop marker (and any other) 999 to NA (precision 0.01)
      rw.mat[rw.mat[, i] == 999, i] <- NA
    } else if (this.prec.rproc == 1000) {
      ## Ditto, -9999 to NA (precision 0.001)
      rw.mat[rw.mat[, i] == -9999, i] <- NA
    } else {
      prec.unknown[i] <- TRUE
    }
    ## Convert to mm
    rw.mat[, i] <- rw.mat[, i] / this.prec.rproc
  }
  if (all(prec.unknown)) {
    stop("precision unknown in all series")
  }
  ## Accommodate mid-series upper and lower case differences: If a
  ## series doesn't end with a stop marker, see if the series ID of
  ## the next row in the file matches when case differences are
  ## ignored.
  if (any(prec.unknown)) {
    upper.ids <- toupper(series.ids)
    new.united <- TRUE
    series.united <- 1:ncol(rw.mat)
    while (new.united) {
      new.united <- FALSE
      for (this.series in which(prec.unknown)) {
        these.rows <- which(series.index == this.series)
        last.row <- these.rows[length(these.rows)]
        next.series <- series.united[series.index[last.row + 1]]
        if (last.row == length(series) ||
            upper.ids[this.series] != upper.ids[next.series]) {
          new.united <- FALSE
          break
        }
        last.decade <- decade.yr[last.row]
        next.decade <- decade.yr[last.row + 1]
        if (!prec.unknown[next.series] &&
            next.decade > last.decade &&
            next.decade <= last.decade + 10) {
          val.count <- numeric(span)
          this.col <- rw.mat[, this.series]
          next.col <- rw.mat[, next.series]
          flag.this <- !is.na(this.col) & this.col != 0
          val.count[flag.this] <- 1
          flag.next <- !is.na(next.col) & next.col != 0
          val.count[flag.next] <- val.count[flag.next] + 1
          if (any(val.count > 1)) {
            new.united <- FALSE
            break
          }
          this.prec.rproc <- prec.rproc[next.series]
          if (this.prec.rproc == 100) {
            this.col[this.col == 999] <- NA
          } else if (this.prec.rproc == 1000) {
            this.col[this.col == -9999] <- NA
          }
          this.col <- this.col / this.prec.rproc
          rw.mat[flag.this, next.series] <- this.col[flag.this]
          series.united[this.series] <- next.series
          new.united <- TRUE
          prec.unknown[this.series] <- FALSE
          warning(gettextf("combining series %s and %s",
                           series.ids[this.series],
                           series.ids[next.series],
                           domain="R-dplR"), domain=NA)
        }
      }
    }
    prec.unknown <- which(prec.unknown)
    n.unknown <- length(prec.unknown)
    if (n.unknown > 0) {
      stop(sprintf(ngettext(n.unknown,
                            "precision unknown in series %s",
                            "precision unknown in series %s",
                            domain="R-dplR"),
                   paste0(series.ids[prec.unknown], collapse=", ")),
           domain=NA)
    } else {
      to.keep <- which(series.united == 1:ncol(rw.mat))
      rw.mat <- rw.mat[, to.keep, drop=FALSE]
      nseries <- length(to.keep)
      series.ids <- series.ids[to.keep]
      prec.rproc <- prec.rproc[to.keep]
    }
  }
  
  the.range <-
    as.matrix(apply(rw.mat, 2, yr.range, yr.vec=min.year:max.year))
  series.min <- the.range[1, ]
  series.max <- the.range[2, ]
  series.min.char <- format(series.min, scientific=FALSE, trim=TRUE)
  series.max.char <- format(series.max, scientific=FALSE, trim=TRUE)
  seq.series.char <- format(seq_len(nseries), scientific=FALSE, trim=TRUE)
  if(verbose){
    cat(sprintf(ngettext(nseries,
                         "There is %d series\n",
                         "There are %d series\n",
                         domain="R-dplR"),
                nseries))
    cat(paste0(format(seq.series.char, width=5), "\t",
               format(series.ids, width=8), "\t",
               format(series.min.char, width=5, justify="right"), "\t",
               format(series.max.char, width=5, justify="right"), "\t",
               format(1/prec.rproc, scientific=FALSE,drop0trailing=TRUE),"\n"),
        sep="")
  }
  ## trim the front and back of the output to remove blank rows
  good.series <- !is.na(series.min)
  if (!any(good.series)) {
    stop("file has no good data")
  }
  incl.rows <- seq.int(min(series.min[good.series])-min.year+1,
                       max(series.max[good.series])-min.year+1)
  ## trim
  rw.mat <- rw.mat[incl.rows, , drop=FALSE]
  ## Fix internal NAs. These are coded as 0 in the DPL programs
  fix.internal.na <- function(x) {
    na.flag <- is.na(x)
    good.idx <- which(!na.flag)
    y <- x
    if (length(good.idx) >= 2) {
      min.good <- min(good.idx)
      max.good <- max(good.idx)
      fix.flag <- na.flag & c(rep(FALSE, min.good),
                              rep(TRUE, max.good-min.good-1),
                              rep(FALSE, length(x)-max.good+1))
      y[fix.flag] <- 0
    }
    y
  }
  rw.df <- as.data.frame(apply(rw.mat, 2, fix.internal.na))
  names(rw.df) <- as.character(series.ids)
  class(rw.df) <- c("rwl", "data.frame")
  rw.df
}

context("input / output functions")
test.read.tucson <- function() {
    MISSINGVAL <- 0

    ## Invalid file
    tf <- tempfile()
    fh <- file(tf, "wt")
    on.exit(unlink(tf))
    writeLines("TEST1A  1734  1230   456   789    12    34    56     7     6",
               fh)
    close(fh)
    test_that("read.tucson catches lines that are too long", {
        expect_error(read.tucson(tf), "failed to read")
    })

    ## Precision 0.01
    tf2 <- tempfile()
    fh2 <- file(tf2, "wt")
    on.exit(unlink(tf2), add=TRUE)
    writeLines("TEST2A  1734  1230   456   789    12    34   999", fh2)
    close(fh2)
    test_that("read.tucson can handle data with precision 0.01", {
        res.tf2 <- read.tucson(tf2)
        expect_true(is.data.frame(res.tf2))
        expect_named(res.tf2, "TEST2A")
        expect_equal(row.names(res.tf2), as.character(1734:1738))
        expect_equal(res.tf2[[1]], c(12.3, 4.56, 7.89, 0.12, 0.34))
    })

    ## Precision 0.001
    tf3 <- tempfile()
    fh3 <- file(tf3, "wt")
    on.exit(unlink(tf3), add=TRUE)
    writeLines("TEST3A  1734  1230   456   789    12    34 -9999", fh3)
    close(fh3)
    test_that("read.tucson can handle data with precision 0.001", {
        res.tf3 <- read.tucson(tf3)
        expect_true(is.data.frame(res.tf3))
        expect_named(res.tf3, "TEST3A")
        expect_equal(row.names(res.tf3), as.character(1734:1738))
        expect_equal(res.tf3[[1]], c(1.23, 0.456, 0.789, 0.012, 0.034))
    })

    ## Unusual line separator
    tf4 <- tempfile()
    fh4 <- file(tf4, "wt")
    on.exit(unlink(tf4), add=TRUE)
    writeLines(c("TEST4A  1734  1230   456   789    12    34     5",
                 "TEST4A  1740   678   999"), fh4, sep="\r\r\n")
    close(fh4)
    test_that("read.tucson works with unusual line separators", {
        res.tf4 <- read.tucson(tf4)
        expect_true(is.data.frame(res.tf4))
        expect_named(res.tf4, "TEST4A")
        expect_equal(row.names(res.tf4), as.character(1734:1740))
        expect_equal(res.tf4[[1]], c(12.3, 4.56, 7.89, 0.12, 0.34, 0.05, 6.78))
    })

    ## Tab-delimited file
    tf5 <- tempfile()
    fh5 <- file(tf5, "wt")
    on.exit(unlink(tf5), add=TRUE)
    writeLines("TEST5A\t1734\t1230\t456\t789\t12\t34\t999", fh5)
              close(fh5)
    test_that("read.tucson works with tab delimited data", {
        res.tf5 <- read.tucson(tf5)
        expect_true(is.data.frame(res.tf5))
        expect_named(res.tf5, "TEST5A")
        expect_equal(row.names(res.tf5), as.character(1734:1738))
        expect_equal(res.tf5[[1]], c(12.3, 4.56, 7.89, 0.12, 0.34))
    })

    ## Stop marker is 13th column (non-standard)
              tf6 <- tempfile()
              fh6 <- file(tf6, "wt")
    on.exit(unlink(tf6), add=TRUE)
    writeLines(c("TEST6A  1734   123   123   123   123   123   123",
                 "TEST6A  1740   123   123   123   123   123   123   123   123   123   123 -9999"), fh6)
    close(fh6)
    test_that("read.tucson accepts stop marker in extra column", {
        res.tf6 <- read.tucson(tf6)
        expect_true(is.data.frame(res.tf6))
        expect_named(res.tf6, "TEST6A")
        expect_equal(row.names(res.tf6), as.character(1734:1749))
        expect_equal(res.tf6[[1]], rep.int(0.123, 16))
    })

    ## Non-standard missing data marker
    tf7 <- tempfile()
    fh7 <- file(tf7, "wt")
    on.exit(unlink(tf7), add=TRUE)
    writeLines("TEST7A  1734  1230   456     .    12    34   999", fh7)
    close(fh7)
    test_that("read.tucson accepts dot as missing data marker", {
        res.tf7 <- read.tucson(tf7)
        expect_true(is.data.frame(res.tf7))
        expect_named(res.tf7, "TEST7A")
        expect_equal(row.names(res.tf7), as.character(1734:1738))
        expect_equal(res.tf7[[1]], c(12.3, 4.56, MISSINGVAL, 0.12, 0.34))
    })

    ## Overlapping data is an error
    tf8 <- tempfile()
    fh8 <- file(tf8, "wt")
    on.exit(unlink(tf8), add=TRUE)
    writeLines(c("TEST8A  1734  1230   456   789    12    34   999",
                 "TEST8A  1730  1230   456   789    12    34   999"), fh8)
    close(fh8)
    test_that("read.tucson stops on overlapping data", {
        expect_error(read.tucson(tf8), "failed to read")
    })

    ## Non-standard file with missing decade
    tf9 <- tempfile()
    fh9 <- file(tf9, "wt")
    on.exit(unlink(tf9), add=TRUE)
    writeLines(c("TEST9A  1734   123   123   123   123   123   123",
                 "TEST9A  1750   123   123   123   123   123   123   123   123   123 -9999"), fh9)
    close(fh9)
    test_that("read.tucson marks missing decades", {
        res.tf9 <- read.tucson(tf9)
        expect_true(is.data.frame(res.tf9))
        expect_named(res.tf9, "TEST9A")
        expect_equal(row.names(res.tf9), as.character(1734:1758))
        expect_equal(res.tf9[[1]],
                     c(rep.int(0.123, 6), rep.int(MISSINGVAL, 10),
                       rep.int(0.123, 9)))
    })

    ## Two series
    tf10 <- tempfile()
    fh10 <- file(tf10, "wt")
    on.exit(unlink(tf10), add=TRUE)
    writeLines(c("TST10A  1734  1230  1230  1230  1230  1230 -9999",
                 "TST10B  1732   123   123   123   123   999"), fh10)
    close(fh10)
    test_that("read.tucson supports mixed precisions", {
        res.tf10 <- read.tucson(tf10)
        expect_true(is.data.frame(res.tf10))
        expect_named(res.tf10, c("TST10A", "TST10B"))
        expect_equal(row.names(res.tf10), as.character(1732:1738))
        expect_equal(res.tf10[[1]], c(rep.int(NA_real_, 2), rep.int(1.23, 5)))
        expect_equal(res.tf10[[2]], c(rep.int(1.23, 4), rep.int(NA_real_, 3)))
    })

    ## Need 5 characters for year, effect of parameter 'long'
    tf11 <- tempfile()
    fh11 <- file(tf11, "wt")
    on.exit(unlink(tf11), add=TRUE)
    writeLines("TST11A -1734  1230   456   789   999", fh11)
    close(fh11)
    test_that("read.tucson argument 'long' works", {
        res.tf11a <- read.tucson(tf11)
        expect_true(is.data.frame(res.tf11a))
        expect_named(res.tf11a, "TST11A -")
        expect_equal(row.names(res.tf11a), as.character(1734:1736))
        expect_equal(res.tf11a[[1]], c(12.3, 4.56, 7.89))
        res.tf11b <- read.tucson(tf11, long=TRUE)
        expect_true(is.data.frame(res.tf11b))
        expect_named(res.tf11b, "TST11A")
        expect_equal(row.names(res.tf11b), as.character(-1734:-1732))
        expect_equal(res.tf11b[[1]], c(12.3, 4.56, 7.89))
    })

    ## Mixed case ("Tst12A" does not have a stop marker)
    tf12 <- tempfile()
    fh12 <- file(tf12, "wt")
    on.exit(unlink(tf12), add=TRUE)
    writeLines(c("Tst12A  1734  1230   456   789    12    34     5",
                 "TST12A  1740   678   999"), fh12)
    close(fh12)
    test_that("read.tucson corrects mixed case typos", {
        res.tf12 <- read.tucson(tf12)
        expect_true(is.data.frame(res.tf12))
        expect_named(res.tf12, "TST12A")
        expect_equal(row.names(res.tf12), as.character(1734:1740))
        expect_equal(res.tf12[[1]],
                     c(12.3, 4.56, 7.89, 0.12, 0.34, 0.05, 6.78))
    })

    ## File has no data (invalid file)
    tf13 <- tempfile()
    fh13 <- file(tf13, "wt")
    on.exit(unlink(tf13), add=TRUE)
    writeLines("TST13A  1734", fh13)
    close(fh13)
    test_that("read.tucson gives empty result when appropriate", {
        expect_equal(0, nrow(read.tucson(tf13, header = FALSE)))
    })

    tf14 <- tempfile()
    fh14 <- file(tf14, "wt")
    on.exit(unlink(tf14), add=TRUE)
    writeLines(c("TST14A  1906     0     0   100   200",
                 "TST14A  1910   300   200   100   200   300   999",
                 "TST14B  1905   300   200   100   200   300",
                 "TST14B  1910   200   100     0     0   999",
                 "TST14C  1906     0   200   100   200",
                 "TST14C  1910   300   200   100     0   999"), fh14)
    close(fh14)
    test_that("read.tucson (by default) preserves edge zeros", {
        res.tf14 <- read.tucson(tf14)
        expect_true(is.data.frame(res.tf14))
        expect_named(res.tf14, c("TST14A", "TST14B", "TST14C"))
        expect_equal(row.names(res.tf14), as.character(1905:1914))
        expect_equal(res.tf14[[1]],
                     c(NA_real_, 0, 0, 1, 2, 3, 2, 1, 2, 3))
        expect_equal(res.tf14[[2]],
                     c(3, 2, 1, 2, 3, 2, 1, 0, 0, NA_real_))
        expect_equal(res.tf14[[3]],
                     c(NA_real_, 0, 2, 1, 2, 3, 2, 1, 0, NA_real_))
        res.tf14B <- read.tucson(tf14, edge.zeros=FALSE)
        expect_true(is.data.frame(res.tf14B))
        expect_named(res.tf14B, c("TST14A", "TST14B", "TST14C"))
        expect_equal(row.names(res.tf14B), as.character(1905:1914))
        NA2 <- rep.int(NA_real_, 2)
        NA3 <- rep.int(NA_real_, 3)
        expect_equal(res.tf14B[[1]],
                     c(NA3, 1, 2, 3, 2, 1, 2, 3))
        expect_equal(res.tf14B[[2]],
                     c(3, 2, 1, 2, 3, 2, 1, NA3))
        expect_equal(res.tf14B[[3]],
                     c(NA2, 2, 1, 2, 3, 2, 1, NA2))
    })

}
test.read.tucson()
### We should write tests for other I/O functions, also

### Written by Mikko Korpela.
###
### Creates a generator of Universally Unique IDentifiers (UUIDs).
### Returns a parameterless function (with some randomish internal state),
### a call to which returns a vector of length one containing
### the character representation of a Version 4 UUID (RFC 4122).
uuid.gen <- function(more.state="") {
    ## We know that the PRNG algorithms of R are reasonably good, but there
    ## is no guarantee about the quality of an arbitrary user supplied PRNG.
    if (RNGkind()[1] == "user-supplied") {
        warning("a user-coded (pseudo) random number generator is in use")
    }
    ## Pastes together system and software information and current time.
    op <- options(digits.secs=6)
    prefix <- paste(c(Sys.info(), unlist(R.version),
                      unlist(.Platform), getwd(), Sys.getpid(),
                      format(Sys.time(), "%Y%m%d%H%M%S", usetz=TRUE),
                      more.state),
                    collapse="")
    options(op)
    ## Lookup table from hex to hex. Corresponds to setting
    ## * the most significant bit to 1 and
    ## * the second most significant bit to 0.
    ## Linear search with string keys seems to be faster than
    ## alternatives using a) a hashed environment or
    ## b) the hex converted to integer +1 as an index to the table.
    uuid.17.lookup <-
        c("0" = "8", "1" = "9", "2" = "a", "3" = "b",
          "4" = "8", "5" = "9", "6" = "a", "7" = "b", "8" = "8", "9" = "9",
          "a" = "a", "b" = "b", "c" = "8", "d" = "9", "e" = "a", "f" = "b",
          "A" = "A", "B" = "B", "C" = "8", "D" = "9", "E" = "A", "F" = "B")

    function() {
        ## We extract "enough" pseudo randomness, even preparing for a true
        ## heavy user of UUIDs: 5 numbers correspond to 150-160 varying bits,
        ## depending on the random number generator used (NO SAFETY GUARD
        ## against the stupid use of a bad custom PRNG).
        dgst <- digest(paste0(prefix, paste0(runif(5), collapse="")),
                       algo = "md5",
                       serialize = FALSE)
        ## The UUID is formatted with hyphen-minus characters.
        ## Some bits are set to fixed values according to UUID Version 4.
        paste0(substr(dgst, 1, 8), "-",
               substr(dgst, 9, 12), "-",
               "4", substr(dgst, 14, 16), "-",
               uuid.17.lookup[substr(dgst, 17, 17)], substr(dgst, 18, 20), "-",
               substr(dgst, 21, 32))
    }
}

autoread.ids <- function(rwl, ignore.site.case = TRUE, ignore.case = "auto",
                         fix.typos = TRUE, typo.ratio = 5, use.cor = TRUE) {
    read.ids(rwl, "auto", ignore.site.case, ignore.case,
             fix.typos, typo.ratio, use.cor)
}

read.ids <- function(rwl, stc=c(3, 2, 3), ignore.site.case = FALSE,
                     ignore.case = FALSE, fix.typos = FALSE,
                     typo.ratio = 5, use.cor = TRUE) {

### Check arguments
    stopifnot(is.data.frame(rwl))
    ids <- names(rwl)
    if (is.null(ids) || any(is.na(ids))) {
        stop("'rwl' must have non-NA names")
    }
    stopifnot(Encoding(ids) != "bytes")
    check.flags(fix.typos, use.cor, ignore.site.case)
    if (fix.typos) {
        stopifnot(is.numeric(typo.ratio), length(typo.ratio) == 1,
                  is.finite(typo.ratio), typo.ratio > 1)
    }
    if (identical(stc, "auto") || identical(is.na(stc), TRUE) ||
        is.null(stc)) {
        auto.mode <- TRUE
    } else {
        auto.mode <- FALSE
        if(!all(is.int(stc))) {
            stop("Site-Tree-Core mask must only contain integral values")
        }
        if(length(stc) != 3) {
            stop("length of Site-Tree-Core mask must be 3")
        }
    }
    if (identical(ignore.case, "auto") ||
        identical(is.na(ignore.case), TRUE) || is.null(ignore.case)) {
        auto.case <- TRUE
    } else {
        check.flags(ignore.case)
        auto.case <- FALSE
    }

### Regular expressions for matching
    DIGIT.MISC <- "^[_\\?\\*]*\\d[_\\?\\*[:digit:]]*$"
    ALPHA.MISC <- "^[_\\?\\*]*[[:alpha:]][_\\?\\*[:alpha:]]*$"
    LOWER.MISC <- "^[^[:upper:]]*[[:lower:]][^[:upper:]]*$"
    UPPER.MISC <- "^[^[:lower:]]*[[:upper:]][^[:lower:]]*$"
    START.NO.DIGIT <- "^\\D"
    START.NO.ALPHA <- "^[^[:alpha:]]"

### Helper functions
    warn.changed <- function(ids, ids.new) {
        diff.ids <- which(ids != ids.new)
        warn.fmt <- gettext('changed tree-core code "%s" to "%s"',
                            domain = "R-dplR")
        for (i in diff.ids) {
            warning(sprintf(warn.fmt, ids[i], ids.new[i]), domain = NA)
        }
    }

    ## Heuristics to avoid snapping characters from the (presumed)
    ## tree part to the site part.  Parameters: part1 is all
    ## alphabetic, and what was previously thought to be the site
    ## part. part2 is what comes after part1
    site.tree.split <- function(part1, part2) {
        res <- part1
        digit.part2 <- grepl("^\\d+$", part2)
        remaining <- which(digit.part2)
        nchar.rem <- nchar(part1[remaining])
        unique.nchar <- sort(unique(nchar.rem), decreasing=TRUE)
        for (nc in unique.nchar[unique.nchar > 1]) {
            flag.nc <- nchar.rem == nc
            idx.nc <- remaining[flag.nc]
            if (length(idx.nc) == 0) {
                next
            }
            p12.tmp <- part1[idx.nc]
            p1 <- substr(p12.tmp, 1, nc - 1)
            p2 <- substr(p12.tmp, nc, nc)
            split.ok <- FALSE
            if (any((grepl("^[[:lower:]]+$", p1) &
                     grepl("[^[:lower:]]", p2)))) {
                split.ok <- TRUE
            } else if (any((grepl("^[[:upper:]]+$", p1) &
                            grepl("[^[:upper:]]", p2)))) {
                split.ok <- TRUE
            }
            if (split.ok) {
                p1.rem <- substr(part1[remaining], 1, nc - 1)
                foo <- which(tolower(p1.rem) %in% unique(tolower(p1)))
                res[remaining[foo]] <- p1.rem[foo]
                remaining <- remaining[-foo]
                nchar.rem <- nchar.rem[-foo]
            } else {
                foo <- which(flag.nc)
                remaining <- remaining[-foo]
                nchar.rem <- nchar.rem[-foo]
            }
        }
        highconf <- res != part1
        remaining <- which(!highconf)
        nchar.rem <- nchar(part1[remaining])
        highconf <- which(highconf)
        nchar.highconf <- nchar(part1[highconf])
        for (nc in intersect(unique(nchar.rem), unique(nchar.highconf))) {
            tmp.nc <- unique(nchar(res[highconf[nchar.highconf == nc]]))
            if (length(tmp.nc) == 1) {
                foo <- which(nchar.rem == nc)
                rem.match <- remaining[foo]
                res[rem.match] <- substr(part1[rem.match], 1, tmp.nc)
                remaining <- remaining[-foo]
                nchar.rem <- nchar.rem[-foo]
            }
        }
        remaining <- res == part1
        highconf <- which(!remaining)
        remaining <- which(remaining)
        n.remaining <- length(remaining)
        n.highconf <- length(highconf)
        if (n.remaining > 0 && n.highconf > 0) {
            nchar1 <- nchar(part1[highconf])
            nchar2 <- nchar(res[highconf])
            trim.amount <- nchar1[1] - nchar2[1]
            part1.rem <- part1[remaining]
            nchar.rem <- nchar(part1.rem)
            if (all(nchar1 - nchar2 == trim.amount) &&
                all(nchar.rem >= trim.amount)) {
                res[remaining] <- substring(part1.rem,
                                            1, nchar.rem - trim.amount)
            }
        }
        res
    }

    ## Decrements a generalized variable length integer counter by
    ## one. The counter is a vector where counter == c(c, b, a) with
    ## counterMax == c(C, B, A) represents the non-negative integer
    ## value (a-1) + (b-1) * A + (c-1) * B * A.  Minimum value of the
    ## individual digit, i.e. counter[i], is 1, max value is
    ## counterMax[i]. When the counter is decremented from all ones,
    ## corresponding to the value 0, all member of the returned vector
    ## are set to zero. The length of the return value is same as the
    ## length of the input counter, i.e. possible leading zeros
    ## (represented by 1 in the counter vector) are kept.
    decrement <- function(counter, counterMax) {
        k <- length(counter)
        res <- counter
        for (i in k:1) {
            if (counter[i] > 1) {
                res[i] <- counter[i] - 1
                break
            } else if (i == 1) {
                res[] <- 0
            } else {
                res[i] <- counterMax[i]
            }
        }
        res
    }

    ## Finds approximate matches of pattern in x. All strings must
    ## have the same length.
    amatches <- function(pattern, x, max.distance=1, ignore.case=FALSE) {
        if (ignore.case) {
            psplit <- strsplit(pattern, "")[[1]]
            xsplit <- strsplit(x, "")
        } else {
            psplit <- strsplit(tolower(pattern), "")[[1]]
            xsplit <- strsplit(tolower(x), "")
        }
        dists <- numeric(length(x))
        for (i in seq_along(dists)) {
            dists[i] <- sum(psplit != xsplit[[i]])
        }
        x[dists <= max.distance]
    }

    ## In character strings x and y, see if one vector can be
    ## converted to the other by transforming each differing character
    ## to its possible look-alike character of the opposite type
    ## (alphabets <-> digits).
    match.helper <- function(x, y) {
        x2 <- strsplit(x, "")[[1]]
        y2 <- strsplit(y, "")[[1]]
        diff.idx <- which(x2 != y2)
        sub.tonum <-
            c(G="6", S="5", s="5", I="1", i="1", O="0", o="0")
        sub.tolower <- c("5"="s", "1"="i", "0"="o")
        sub.toupper <- c("6"="G", "5"="S", "1"="I", "0"="O")
        for (i in diff.idx) {
            this.x <- x2[i]
            this.y <- y2[i]
            if (!(identical(unname(sub.tonum[this.x]), this.y) ||
                  identical(unname(sub.tolower[this.x]), this.y) ||
                  identical(unname(sub.toupper[this.x]), this.y))) {
                return(FALSE)
            }
        }
        TRUE
    }

    ## One kind of typo fixer. For each character position, check if
    ## the proportion of alphabets or digits exceeds a threshold. If
    ## so, try to find look-alike characters of the opposite type
    ## (alphabets <-> digits) and convert them to the majority type.
    ## Changes to ids that differ from the majority (over the
    ## threshold) type in more than max.distance character places are
    ## prohibited. Finally, if keep.monotonic is TRUE, the function
    ## keeps the original ids if changing them would violate a
    ## possible monotonic order of numeric prefixes in ids.
    typofix <- function(ids, keep.monotonic=TRUE, max.distance=1) {
        nchar.ids <- nchar(ids)
        max.nchar.ids <- max(nchar.ids)
        n.ids <- length(ids)
        remaining <- seq_len(n.ids)
        ids.new <- character(n.ids)
        min.ratio <- typo.ratio / (typo.ratio + 1)
        minorityCols <- numeric(n.ids)
        for (i in seq_len(max.nchar.ids)) {
            remaining <- remaining[nchar.ids[remaining] >= i]
            these.substrs <- substr(ids[remaining], i, i)
            n.tot <- length(these.substrs)
            digit.flag <- grepl("\\d", these.substrs)
            alpha.flag <- grepl("[[:alpha:]]", these.substrs)
            n.123 <- sum(digit.flag)
            n.abc <- sum(alpha.flag)
            if (n.123 >= min.ratio * n.tot && n.123 < n.tot) {
                minorityCols <- minorityCols + !digit.flag
                ids.new[remaining] <-
                    paste0(ids.new[remaining],
                           sub("G", "6",
                               sub("S", "5",
                                   sub("I", "1",
                                       sub("O", "0",
                                           these.substrs,
                                           ignore.case = TRUE),
                                       ignore.case = TRUE),
                                   ignore.case = TRUE),
                               fixed = TRUE))
            } else if (n.abc >= min.ratio * n.tot && n.abc < n.tot) {
                minorityCols <- minorityCols + !alpha.flag
                alpha.strs <- these.substrs[alpha.flag]
                n.lower <- length(grep("[[:lower:]]", alpha.strs))
                n.upper <- length(grep("[[:upper:]]", alpha.strs))
                if (n.lower > n.upper) {
                    ids.new[remaining] <-
                        paste0(ids.new[remaining],
                               sub("5", "s",
                                   sub("1", "i",
                                       sub("0", "o",
                                           these.substrs, fixed = TRUE),
                                       fixed = TRUE),
                                   fixed = TRUE))
                } else {
                    ids.new[remaining] <-
                        paste0(ids.new[remaining],
                               sub("6", "G",
                                   sub("5", "S",
                                       sub("1", "I",
                                           sub("0", "O",
                                               these.substrs,
                                               fixed = TRUE),
                                           fixed = TRUE),
                                       fixed = TRUE),
                                   fixed = TRUE))
                }
            } else {
                ids.new[remaining] <- paste0(ids.new[remaining], these.substrs)
            }
        }
        ids.new <- ifelse(minorityCols <= max.distance, ids.new, ids)
        if (keep.monotonic) {
            num.ids <- as.numeric(sub("^(\\d*).*", "\\1", ids))
            num.ids <- num.ids[!is.na(num.ids)]
            if (length(num.ids) >= 2) {
                num.ids.new <-
                    as.numeric(sub("^(\\d*).*", "\\1", ids.new))
                num.ids.new <- num.ids.new[!is.na(num.ids.new)]
                if (length(num.ids.new) >= 2) {
                    diff.ids <- diff(num.ids)
                    diff.ids.new <- diff(num.ids.new)
                    if ((all(diff.ids >= 0) && any(diff.ids.new < 0)) ||
                        (all(diff.ids <= 0) && any(diff.ids.new > 0))) {
                        return(ids)
                    }
                }
            }
        }
        warn.changed(ids, ids.new)
        ids.new
    }

    ## Trim away unnecessary prefix of 'ids'.  Returns a number
    ## 'tree.start' between 1 and 'min(stc.t.vec)' so that the tree /
    ## core mapping is not affected if the first 'tree.start - 1'
    ## characters of each member of 'ids' are trimmed away.
    trim.from.start <- function(ids, stc.t.vec) {
        tree.start <- 1
        orig.tree.strs <- substring(ids, 1, stc.t.vec)
        unique.orig.trees <- unique(orig.tree.strs)
        orig.match <- match(orig.tree.strs, unique.orig.trees)
        for (new.start in dec(min(stc.t.vec), 2)) {
            new.tree.strs <- substring(ids, new.start, stc.t.vec)
            new.match <- match(new.tree.strs, unique(new.tree.strs))
            if (all(new.match == orig.match &
                    !grepl("^(\\s*0*)*$", new.tree.strs))) {
                tree.start <- new.start
                break
            }
        }
        tree.start
    }

    default.stc.t <- function(ids, stc.t.vec) {
        n.tot <- length(stc.t.vec)
        na.flag <- is.na(stc.t.vec)
        good.idx <- which(!na.flag)
        n.good <- length(good.idx)
        threshold <- typo.ratio / (typo.ratio + 1) * n.tot
        if (n.good >= threshold) {
            stc.t.freq <- sort(table(stc.t.vec[good.idx]), decreasing=TRUE)
            if (stc.t.freq[1] >= threshold) {
                as.numeric(names(stc.t.freq)[1])
            } else {
                max(nchar(ids))
            }
        } else {
            max(nchar(ids))
        }
    }

    ## Computes agreement scores of different stc.t values as in
    ## correlation clustering.  Here, the set of available clusterings
    ## (assignment of series to trees) is limited by the series names
    ## (minus site part) and its chosen split into tree and core
    ## parts.
    cor.clust <- function(rwl, ids) {
        n.cases <- length(ids)
        rwl.cor <- cor(rwl, use="pairwise.complete.obs")
        up.tri <- upper.tri(rwl.cor, diag=FALSE)
        med.cor <- median(rwl.cor[up.tri], na.rm=TRUE)
        ## rwl series are similar if their correlation is above median
        ## NOTE: Other definitions would be possible.
        sim.mat <- !is.na(rwl.cor) & rwl.cor > med.cor
        max.nchar.ids <- max(nchar(ids))
        agreement <- rep(0, max.nchar.ids)
        ## Compute the number of agreements for each clustering
        for (nc in seq_len(max.nchar.ids)) {
            firstpart <- substr(ids, 1, nc)
            clnum <- match(firstpart, unique(firstpart))
            same.clust <-
                matrix(rep(clnum, times = n.cases), n.cases, n.cases) ==
                    matrix(rep(clnum, each = n.cases), n.cases, n.cases)
            agreement[nc] <- sum(sim.mat[same.clust & up.tri]) +
                sum(!sim.mat[!same.clust & up.tri])
        }
        ## In case of ties, prefer larger stc.t
        max.nchar.ids - which.max(rev(agreement)) + 1
    }

    ## Convert alphabets to their look-alike numbers.
    ## If one.only is TRUE, only fixes strings that occur once in 'ids'.
    ## If max.distance is finite, only fixes strings that require changes
    ## to no more than max.distance characters.
    typos.to.numeric <- function(ids, one.only = TRUE,
                                 max.distance = Inf) {
        ids.new <- ids
        if (one.only) {
            id.freq <- table(ids)
            fix.idx <- which(ids %in% names(id.freq)[id.freq == 1])
        } else {
            fix.idx <- seq_along(ids)
        }
        ids.new[fix.idx] <-
            gsub("G", "6",
                 gsub("S", "5",
                      gsub("I", "1",
                           gsub("O", "0", ids[fix.idx], ignore.case = TRUE),
                           ignore.case = TRUE),
                      ignore.case = TRUE),
                 fixed = TRUE)

        if (is.finite(max.distance)) {
            dists <- numeric(length(ids))
            oldsplit <- strsplit(ids, "")
            newsplit <- strsplit(ids.new, "")
            for (i in seq_along(dists)) {
                dists[i] <- sum(oldsplit[[i]] != newsplit[[i]])
            }
            ifelse(dists <= max.distance, ids.new, ids)
        } else {
            ids.new
        }
    }

    ## Convert numbers to their look-alike upper case alphabets.
    ## NOTE: By design, this function must perform a superset of the
    ## substitutions done by typos.to.lower(). Also, if both functions
    ## substitute an alphabet for a given digit, it must be the same
    ## letter, only differing in case.
    typos.to.upper <- function(ids) {
        gsub("6", "G",
             gsub("5", "S", gsub("1", "I", gsub("0", "O", ids, fixed = TRUE),
                                 fixed = TRUE),
                  fixed = TRUE),
             fixed = TRUE)
    }

    ## Convert numbers to their look-alike lower case alphabets
    typos.to.lower <- function(ids) {
        gsub("5", "s", gsub("1", "i", gsub("0", "o", ids, fixed = TRUE),
                            fixed = TRUE),
             fixed = TRUE)
    }

    ## Map each unique negative number in tree.vec to an unused
    ## positive integral value.
    fix.negative <- function(tree.vec) {
        neg.idx <- which(tree.vec < 0)
        n.cases <- length(neg.idx)
        if (n.cases > 0) {
            unique.neg <- sort(unique(tree.vec[neg.idx]), decreasing=TRUE)
            max.num <- max(0, max(tree.vec)) # no NAs allowed
            n.unique.neg <- length(unique.neg)
            if (max.num > 1) {
                free.ids <- as.numeric(which(!(1:(max.num - 1) %in% tree.vec)))
                n.free <- length(free.ids)
                if (n.free < n.unique.neg) {
                    free.ids <- c(free.ids,
                                  seq(from = max.num + 1, by = 1,
                                      length.out = n.unique.neg - n.free))
                }
            } else {
                free.ids <- (max.num + 1):(max.num + n.unique.neg)
            }
            tree.vec2 <- tree.vec
            tree.vec2[neg.idx] <- free.ids[match(tree.vec2[neg.idx],
                                                 unique.neg)]
            tree.vec2
        } else {
            tree.vec
        }
    }

    ## Make it so that each number in the output corresponds to one
    ## unique string in 'strs'.  Keep 'nums' where possible; use
    ## negative integral values in the output when more than one
    ## string in 'strs' share the same number in 'nums'.  This is used
    ## for preserving initial zeros, spaces etc.
    preserve.zeros <- function(nums, strs) {
        nums2 <- nums
        n.cases <- length(nums)
        unique.nums <- unique(nums)
        min.num <- min(0, min(unique.nums))
        free.ids <- (min.num - 1):(min.num - n.cases)
        ids.taken <- 0
        for (this.num in unique.nums) {
            idx <- which(nums == this.num)
            these.strs <- strs[idx]
            unique.strs <- sort(unique(these.strs))
            n.unique <- length(unique.strs)
            if (n.unique > 1) {
                unique.m1 <- unique.strs[-1]
                foo <- these.strs %in% unique.m1
                nums2[idx[foo]] <-
                    free.ids[ids.taken + match(these.strs[foo], unique.m1)]
                ids.taken <- ids.taken + n.unique - 1
            }
        }
        nums2
    }

    ## Tree and core IDs from tree.strs and core.strs.  Respects
    ## original IDs in strings that can be converted to integral
    ## numbers, but takes into account any reserved tree IDs in
    ## prev.tree.vec.  If collisions occur, uses unreserved negative
    ## numbers for the IDs in question.
    tree.and.core.ids <- function(tree.strs,
                                  core.strs,
                                  prev.tree.vec = NULL,
                                  auto.case = FALSE,
                                  auto.trim = FALSE,
                                  .drop0leading = !auto.trim) {
        auto.case2 <- auto.case && any(grepl("[[:upper:]]", tree.strs))
        auto.case2 <- auto.case2 && any(grepl("[[:lower:]]", tree.strs))
        if (auto.trim && auto.case2) {
            tree.strs2 <- sub("^\\s+", "", tree.strs)
            tree.strs2 <- sub("^0+([1-9].*)", "\\1", tree.strs2)
            core.strs2 <- sub("^\\s+", "", core.strs)
            core.strs2 <- sub("^0+([1-9].*)", "\\1", core.strs2)
            res1 <- tree.and.core.ids(tolower(tree.strs2), core.strs2,
                                      prev.tree.vec, FALSE, FALSE, TRUE)
            dupl1 <- res1$dupl
            if (dupl1 > 0) {
                res4 <- tree.and.core.ids(tree.strs, core.strs,
                                          prev.tree.vec, FALSE, FALSE, FALSE)
                dupl4 <- res4$dupl
                if (dupl1 == dupl4) {
                    return(res1)
                } else {
                    res2 <- tree.and.core.ids(tree.strs2, core.strs2,
                                              prev.tree.vec,
                                              FALSE, FALSE, TRUE)
                    dupl2 <- res2$dupl
                    if (dupl2 == dupl4) {
                        return(res2)
                    } else {
                        res3 <- tree.and.core.ids(tolower(tree.strs),
                                                  core.strs, prev.tree.vec,
                                                  FALSE, FALSE, FALSE)
                        dupl3 <- res3$dupl
                        if (dupl3 == dupl4) {
                            return(res3)
                        } else {
                            return(res4)
                        }
                    }
                }
            } else {
                return(res1)
            }
        } else if (auto.trim) {
            tree.strs2 <- sub("^\\s+", "", tree.strs)
            tree.strs2 <- sub("^0+([1-9].*)", "\\1", tree.strs2)
            core.strs2 <- sub("^\\s+", "", core.strs)
            core.strs2 <- sub("^0+([1-9].*)", "\\1", core.strs2)
            res1 <- tree.and.core.ids(tree.strs2, core.strs2,
                                      prev.tree.vec, FALSE, FALSE, TRUE)
            dupl1 <- res1$dupl
            if (dupl1 > 0) {
                res2 <- tree.and.core.ids(tree.strs, core.strs,
                                          prev.tree.vec, FALSE, FALSE, FALSE)
                dupl2 <- res2$dupl
                if (dupl2 < dupl1) {
                    return(res2)
                } else {
                    return(res1)
                }
            } else {
                return(res1)
            }
        } else if(auto.case2) {
            res1 <- tree.and.core.ids(tolower(tree.strs), core.strs,
                                      prev.tree.vec, FALSE, FALSE,
                                      .drop0leading)
            dupl1 <- res1$dupl
            if (dupl1 > 0) {
                res2 <- tree.and.core.ids(tree.strs, core.strs, prev.tree.vec,
                                          FALSE, FALSE, .drop0leading)
                dupl2 <- res2$dupl
                if (dupl2 < dupl1) {
                    return(res2)
                } else {
                    return(res1)
                }
            } else {
                return(res1)
            }
        }
        n.cases <- length(tree.strs)
        tree.vec <- suppressWarnings(as.numeric(tree.strs))
        tree.vec[!is.na(tree.vec) & round(tree.vec) != tree.vec] <- NA
        tree.vec <- abs(tree.vec)
        na.flag <- is.na(tree.vec)
        idx.check <- which(!na.flag)
        n.check <- length(idx.check)
        tree.collisions <- FALSE
        if (!is.null(prev.tree.vec)) {
            notna.tree.vec <- prev.tree.vec[!is.na(prev.tree.vec)]
        }
        if (!is.null(prev.tree.vec) && n.check > 0) {
            unique.site.trees <- unique(tree.vec[idx.check])
            unique.tree.vec <- unique(notna.tree.vec)
            collisions <- unique.site.trees %in% unique.tree.vec
            if (any(collisions)) {
                wrecks <- sort(unique.site.trees[collisions])
                n.wrecks <- length(wrecks)
                tree.collisions <- TRUE
                max.existing <- max(c(unique.site.trees, unique.tree.vec),
                                    na.rm=TRUE)
                if (max.existing > 1) {
                    free.ids <- which(!(1:(max.existing - 1) %in%
                                        c(unique.site.trees, unique.tree.vec)))
                    n.free <- length(free.ids)
                    if (n.free < n.wrecks) {
                        free.ids <- c(free.ids,
                                      seq(from = max.existing + 1, by = 1,
                                          length.out = n.wrecks - n.free))
                    }
                } else {
                    free.ids <- (max.existing + 1):(max.existing + n.wrecks)
                }
                wreck.idx <- idx.check[tree.vec[idx.check] %in% wrecks]
                all.wrecks <- tree.vec[wreck.idx]
                matching.wreck <- match(all.wrecks, wrecks)
                tree.vec[wreck.idx] <- as.numeric(free.ids[matching.wreck])
            }
        }
        ## Make leading zeros significant
        if (!.drop0leading && n.check > 1 &&
            any(grepl("0", tree.strs[idx.check]))) {
            tree.vec[idx.check] <-
                preserve.zeros(tree.vec[idx.check], tree.strs[idx.check])
        }
        na.idx <- which(na.flag)
        n.na <- length(na.idx)
        if (n.na > 0) {
            if (is.null(prev.tree.vec) || length(notna.tree.vec) == 0) {
                min.num <- 0
            } else {
                min.num <- min(0, min(notna.tree.vec))
            }
            free.ids <- (min.num - 1):(min.num - n.cases)
            ids.taken <- 0
            unique.na.trees <- unique(tree.strs[na.idx])
            ## Series with empty tree strings are interpreted as separate trees
            if ("" %in% unique.na.trees) {
                empty.flag <- tree.strs[na.idx] == ""
                idx <- na.idx[empty.flag]
                n.empty <- length(idx)
                tree.vec[idx] <- free.ids[(ids.taken+1):(ids.taken+n.empty)]
                ids.taken <- ids.taken + n.empty
                unique.na.trees <- setdiff(unique.na.trees, "")
                na.idx <- na.idx[!empty.flag]
            }
            tree.vec[na.idx] <- free.ids[match(tree.strs[na.idx],
                                               sort(unique.na.trees)) +
                                         ids.taken]
        }
        core.vec <- rep(NA_real_, n.cases)
        unique.tree.vec <- unique(tree.vec)
        n.duplicated <- 0
        for (treeCounter in seq_along(unique.tree.vec)) {
            tree.idx <- which(tree.vec == unique.tree.vec[treeCounter])
            these.cores <- core.strs[tree.idx]
            ## The same numbering scheme is used for cores,
            ## i.e. respect identifiers that are already integer
            cores.as.int <- suppressWarnings(as.numeric(these.cores))
            cores.as.int[!is.na(cores.as.int) &
                         round(cores.as.int) != cores.as.int] <- NA
            cores.as.int <- abs(cores.as.int)
            na.flag <- is.na(cores.as.int)
            idx.check <- which(!na.flag)
            n.check <- length(idx.check)
            if (n.check > 0) {
                check.strs <- these.cores[idx.check]
                ## Make leading zeros significant
                if (!.drop0leading && n.check > 1 &&
                    any(nchar(check.strs) > 1) &&
                    any(grepl("0", check.strs))) {
                    cores.as.int[idx.check] <-
                        preserve.zeros(cores.as.int[idx.check],
                                       these.cores[idx.check])
                }
            }
            na.idx <- which(na.flag)
            n.na <- length(na.idx)
            if (n.na > 0) {
                if (n.check > 0) {
                    max.num <- max(cores.as.int[idx.check])
                } else {
                    max.num <- 0
                }
                free.ids <- as.numeric((max.num+1):(max.num+n.cases))
                if (max.num > 1) {
                    free.ids <-
                        c(as.numeric(which(!(1:(max.num-1) %in%
                                             cores.as.int[idx.check]))),
                          free.ids)
                }
                ids.taken <- 0
                unique.cores <- sort(unique(these.cores[na.idx]))
                for (uc in unique.cores) {
                    core.idx <- na.idx[these.cores[na.idx] == uc]
                    n.cores <- length(core.idx)
                    if (n.cores == 1) {
                        cores.as.int[core.idx] <- free.ids[ids.taken + 1]
                        ids.taken <- ids.taken + 1
                    } else if (grepl("[-_ \\?\\*]", uc)) {
                        ## Duplicated core strings containing any of
                        ## "-", "_", " ", "?" or "*" are given separate
                        ## core IDs.
                        cores.as.int[core.idx] <-
                            free.ids[(ids.taken + 1):(ids.taken + n.cores)]
                        ids.taken <- ids.taken + n.cores
                    } else {
                        cores.as.int[core.idx] <- free.ids[ids.taken + 1]
                        ids.taken <- ids.taken + 1
                    }
                }
            }
            cores.as.int <- preserve.zeros(cores.as.int, these.cores)
            n.duplicated <- n.duplicated + sum(duplicated(cores.as.int))
            core.vec[tree.idx] <- cores.as.int
        }
        list(tree=tree.vec, core=core.vec, coll=tree.collisions,
             dupl=n.duplicated)
    }
### Actual body of the main function
    n.cases <- length(ids)
    if (n.cases < 2) {
        return(data.frame(tree = seq_len(n.cases), core = rep(1, n.cases),
                          row.names = ids))
    }
### Automatic boundaries of site, tree, core parts
    if (auto.mode) {
        ## 1. Preprocessing
        ids <- sub("^\\s+", "", ids)
        ids <- sub("\\s+$", "", ids)
        nchar.ids <- nchar(ids)
        min.nchar.ids <- min(nchar.ids)
        max.nchar.ids <- max(nchar.ids)
        if (min.nchar.ids < 1) {
            stop("series names must be at least 1 character long")
        }
        ## In case of constant-width ids, remove possible filler zeros
        ## from the end of ids. Only done to all-zero columns.
        if (min.nchar.ids > 1 && min.nchar.ids == max.nchar.ids) {
            last.good <- max.nchar.ids
            for (k in max.nchar.ids:2) {
                if (all(substr(ids, k, k) == "0")) {
                    last.good <- k - 1
                } else {
                    break
                }
            }
            if (last.good < max.nchar.ids) {
                ids <- substr(ids, 1, last.good)
                nchar.ids[] <- max.nchar.ids <- min.nchar.ids <- last.good
            }
        }
        ## Simple case, early return
        if (max.nchar.ids == 1) {
            res <- tree.and.core.ids(substr(ids, 1, 1), rep("1", n.cases))
            return(data.frame(tree=res$tree, core=res$core,
                              row.names=names(rwl)))
        }
        ## 2. Find sites. Different naming schemes are supported, but
        ##    the results can be somewhat uncertain.
        site.vec <- rep(NA_real_, n.cases)
        ## 2.1 Name consists of digits only
        class.flag1 <- grepl("^\\d{2,}$", ids)
        class.idx1 <- which(class.flag1)
        n.class <- length(class.idx1)
        site.str1 <- character(0)
        if (n.class > 0) {
            max.sitelength <- min(nchar.ids[class.idx1]) - 1
            stc.s <- 1
            while (stc.s <= max.sitelength &&
                   length(unique(substr(ids[class.idx1], 1, stc.s))) == 1) {
                stc.s <- stc.s + 1
            }
            stc.s <- stc.s - 1
            site.str1 <- substr(ids[class.idx1[1]], 1, stc.s) # may be empty
        }
        ## 2.2 Alpha(numeric) site ID ending with alphabet, followed by
        ##     number(s) and alphabet(s) (+ possibly punctuation)
        class.idx15 <- which(!class.flag1)
        if (length(class.idx15) > 0) {
            class.def <- "^\\w*[[:alpha:]]\\d+[[:alpha:]]+[[:punct:]]*$"
            class.idx15 <- class.idx15[grepl(class.def, ids[class.idx15])]
            class.sub <- "^(\\w*[[:alpha:]])\\d+[[:alpha:]]+[[:punct:]]*$"
            site.strs15 <- sub(class.sub, "\\1", ids[class.idx15])
            ## Remove cases where the possible alphabetic prefix of
            ## sites also containing digits is found somewhere else
            others.flag <- rep(TRUE, n.cases)
            others.flag[class.idx1] <- FALSE
            others.flag[class.idx15] <- FALSE
            others.idx <- which(others.flag)
            unique.alphasites <-
                unique(sub("^([[:alpha:]]*).*", "\\1",
                           grep("\\d", site.strs15, value=TRUE)))
            unique.alphasites <-
                unique.alphasites[nzchar(unique.alphasites)]
            for (ua in unique.alphasites) {
                nchar.this <- nchar(ua)
                sub.others <- substr(ids[others.idx], 1, nchar.this)
                if ((!ignore.site.case && any(sub.others == ua)) ||
                    (ignore.site.case &&
                     any(tolower(sub.others) == tolower(ua)))) {
                    idx <- which(substr(site.strs15, 1, nchar.this) == ua)
                    site.strs15 <- site.strs15[-idx]
                    class.idx15 <- class.idx15[-idx]
                }
            }
        } else {
            site.strs15 <- character(0)
        }
        class.flag15 <- rep(FALSE, n.cases)
        class.flag15[class.idx15] <- TRUE
        ## 2.3 Alphabetic site ID (part 1, high confidence)
        class.idx2 <- which(!(class.flag1 | class.flag15))
        if (length(class.idx2) > 0) {
            class.def <- paste0("^[[:alpha:]]+[[:space:][:punct:]]+",
                                "[^[:space:][:punct:]]")
            class.idx2 <- class.idx2[grepl(class.def, ids[class.idx2])]
            site.strs2 <- sub("^([[:alpha:]]+).*", "\\1", ids[class.idx2])
        } else {
            site.strs2 <- character(0)
        }
        class.flag2 <- rep(FALSE, n.cases)
        class.flag2[class.idx2] <- TRUE
        ## 2.4 Numeric site ID (NOTE: disabled, hard to define)
        class.idx3 <- integer(0)
        site.strs3 <- character(0)
        class.flag3 <- rep(FALSE, n.cases)
        ## 2.5 Site ID separated with spaces and/or punctuation (also
        ##     second, separate space / punctuation sequence required)
        class.idx4 <- which(!(class.flag1 | class.flag15 |
                              class.flag2 | class.flag3))
        if (length(class.idx4) > 0) {
            class.def <- paste0("^[^[:space:][:punct:]]+[[:space:][:punct:]]+",
                                "[^[:space:][:punct:]]+[[:space:][:punct:]]+",
                                "[^[:space:][:punct:]]")
            class.idx4 <- class.idx4[grepl(class.def, ids[class.idx4])]
            site.strs4 <- sub("^([^[:space:][:punct:]]+).*", "\\1",
                              ids[class.idx4])
        } else {
            site.strs4 <- character(0)
        }
        class.flag4 <- rep(FALSE, n.cases)
        class.flag4[class.idx4] <- TRUE
        ## 2.6 Alphabetic site ID (part 2, low confidence)
        class.idx5 <- which(!(class.flag1 | class.flag15 | class.flag2 |
                              class.flag3 | class.flag4))
        if (length(class.idx5) > 0) {
            class.def <- "^[[:alpha:]]+[^[:alpha:][:space:][:punct:]]"
            class.idx5 <- class.idx5[grepl(class.def, ids[class.idx5])]
            site.strs5 <- sub("^([[:alpha:]]+).*", "\\1", ids[class.idx5])
            n.lowconf <- length(class.idx5)
            unique.sites <- unique(site.strs5)
            nchar.sites <- nchar(unique.sites)
            if (n.lowconf > 0) {
                unique.altsites.high <- grep("^[[:alpha:]]+$",
                                             unique(c(site.strs15, site.strs2,
                                                      site.strs4)),
                                             value = TRUE)
                nchar.altsites.high <- nchar(unique.altsites.high)
            }
            ## Indicates elements of site.strs5 in which we have
            ## low confidence. Low confidence status is reversed
            ## if the site string matches a previous high
            ## confidence site string.
            lowconf <- rep(TRUE, n.lowconf)
            low.flag <- rep(TRUE, length(unique.sites))
            for (siteC in seq_along(unique.sites)) {
                this.site <- unique.sites[siteC]
                idx.nc <- which(nchar.altsites.high == nchar.sites[siteC])
                alt.nc <- unique.altsites.high[idx.nc]
                if ((!ignore.site.case && this.site %in% alt.nc) ||
                    (ignore.site.case &&
                     tolower(this.site) %in% tolower(alt.nc))) {
                    lowconf[site.strs5 == this.site] <- FALSE
                    low.flag[siteC] <- FALSE
                }
            }
            ## Rename remaining low confidence sites if a prefix of
            ## the site name was also identified as a low-confidence
            ## site name. _But_: Demand there is a case change upper
            ## <-> lower at the border. Truly a heuristic...
            if (auto.case) {
                unique.sites <- unique.sites[low.flag]
                nchar.sites <- nchar.sites[low.flag]
                ii <- order(nchar.sites, decreasing = FALSE)
                unique.sites <- unique.sites[ii]
                nchar.sites <- nchar.sites[ii]
                for (siteC in seq_along(unique.sites)) {
                    this.site <- unique.sites[siteC]
                    nchar.this <- nchar.sites[siteC]
                    idx.temp <- which(nchar.sites < nchar.this)
                    nchar.temp <- nchar.sites[idx.temp]
                    for (nc in unique(nchar.temp)) {
                        sub.this <- substr(this.site, 1, nc)
                        idx.nc <- idx.temp[nchar.temp == nc]
                        if (((ignore.site.case &&
                             tolower(sub.this) %in%
                             tolower(unique.sites[idx.nc])) ||
                             (!ignore.site.case &&
                              sub.this %in% unique.sites[idx.nc])) &&
                            grepl(paste0("[[:lower:]][[:upper:]]|",
                                         "[[:upper:]][[:lower:]]"),
                                  substr(this.site, nc, nc + 1))) {
                            site.strs5[site.strs5 == this.site] <- sub.this
                            break
                        }
                    }
                }
            }
            if (auto.case && any(lowconf)) {
                old.strs <- site.strs5[lowconf]
                new.strs <- site.tree.split(old.strs,
                                            substring(ids[class.idx5[lowconf]],
                                                      nchar(old.strs) + 1,
                                                      max.nchar.ids))
                site.strs5[lowconf] <- new.strs
            }
        } else {
            site.strs5 <- character(0)
        }
        class.idx2 <- c(class.idx15, class.idx2, class.idx3, class.idx4,
                        class.idx5)
        if (length(class.idx2) > 0) {
            combi.order <- order(class.idx2)
            class.idx2 <- class.idx2[combi.order]
            site.strs <- c(site.strs15, site.strs2,
                           site.strs3, site.strs4, site.strs5)[combi.order]
            class.ids <- ids[class.idx2]
            ## 2.7 Try to fix rare site strings (possible typos)
            if (fix.typos) {
                site.freq <- table(c(site.strs,
                                     rep(site.str1, length(class.idx1))))
                unique.sites <- names(site.freq)
                if (ignore.site.case) {
                    site.match <- tolower(site.strs)
                    lower.sites <- tolower(unique.sites)
                    unique.lower <- unique(lower.sites)
                    tmp <- match(lower.sites, unique.lower)
                    ## Sum of number of occurrences when case is ignored
                    site.freq.lower <- numeric(length(lower.sites))
                    names(site.freq.lower) <- unique.sites
                    for (i in seq_along(unique.lower)) {
                        tmp2 <- tmp == i
                        site.freq.lower[tmp2] <- sum(site.freq[tmp2])
                    }
                } else {
                    site.match <- site.strs
                }
                ## Sites that have a chance of being fixed to a >= typo.ratio
                ## times more frequent alternative
                if (ignore.site.case) {
                    max.freq <- max(site.freq.lower)
                    foo <-
                        which(site.freq.lower <= floor(max.freq / typo.ratio))
                    ii <- order(site.freq.lower[foo], decreasing = TRUE)
                    sites.to.fix <- unique.sites[foo[ii]]
                } else {
                    max.freq <- max(site.freq)
                    foo <- which(site.freq <= floor(max.freq / typo.ratio))
                    ii <- order(site.freq[foo], decreasing = TRUE)
                    sites.to.fix <- unique.sites[foo]
                }
                ## The "all digits" site string will not be changed
                ## (even if it occurs elsewhere, too)
                sites.to.fix <- setdiff(sites.to.fix, site.str1)
                if (ignore.site.case) {
                    to.fix.lower <- tolower(sites.to.fix)
                }
                warn.times <- gettext("%d times: ", domain = "R-dplR")
                warn.fmt <- gettext('changed site code "%s" to "%s"',
                                    domain = "R-dplR")
            } else {
                sites.to.fix <- character(0)
            }
            already.fixed <- character(0)
            for (this.site in sites.to.fix) {
                if (this.site %in% already.fixed) {
                    next
                }
                if (ignore.site.case) {
                    frequent.sites <-
                        unique.sites[site.freq.lower >=
                                     typo.ratio * site.freq.lower[this.site]]
                    frequent.lower <- unique(tolower(frequent.sites))
                } else {
                    frequent.sites <-
                        unique.sites[site.freq >=
                                     typo.ratio * site.freq[this.site]]
                }
                if (length(frequent.sites) == 0) {
                    next
                }
                nchar.frequent <- nchar(frequent.sites)
                if (ignore.site.case) {
                    these.sites <-
                        sites.to.fix[to.fix.lower == tolower(this.site)]
                    fix.idx <- which(site.strs %in% these.sites)
                } else {
                    fix.idx <- which(site.strs == this.site)
                }
                ids.to.fix <- class.ids[fix.idx]
                n.fix <- length(ids.to.fix)
                nchar.this <- nchar(this.site)
                ## First, fix typos of one character and require
                ## equal length (i.e. substitution only)
                if (nchar.this > 1) {
                    close.match1 <-
                        amatches(this.site,
                                 frequent.sites[nchar.frequent == nchar.this],
                                 max.distance = 1,
                                 ignore.case = ignore.site.case)
                    if (ignore.site.case) {
                        close.match1 <- close.match1[tolower(close.match1) !=
                                                     tolower(this.site)]
                    }
                } else {
                    close.match1 <- character(0)
                }
                n.matches1 <- length(close.match1)
                if (n.matches1 > 1) {
                    ## If there is more than one close match, see if only
                    ## one of those is "look-alike", and choose that
                    lookalikes <-
                        close.match1[vapply(close.match1, match.helper, FALSE,
                                            y=this.site)]
                    if (ignore.site.case) {
                        lookalikes <- unique(tolower(lookalikes))
                    }
                    if (length(lookalikes) == 1) {
                        close.match1 <- lookalikes
                        n.matches1 <- 1
                    } else {
                        next
                    }
                } else if (n.matches1 == 1 && ignore.site.case) {
                    close.match1 <- tolower(close.match1)
                }
                if (n.matches1 == 1) {
                    matches <- class.ids[site.match == close.match1]
                    suffix1 <- substr(ids.to.fix,
                                      nchar.this + 1, max.nchar.ids)
                    sfxes <- substr(matches, nchar.this + 1, max.nchar.ids)
                    if ((!auto.case && ignore.case &&
                         any(tolower(suffix1) %in% tolower(sfxes))) ||
                        ((auto.case || !ignore.case) &&
                         any(suffix1 %in% sfxes)) ||
                        !all(unique(nchar(suffix1)) %in%
                             unique(nchar(sfxes)))) {
                        next
                    }
                } else if (n.matches1 > 1) {
                    next
                }
                ## AND see if the site string could be
                ## shortened or made longer by one
                ## character by interpreting a letter as a
                ## resembling digit or vice versa
                n.matches2 <- 0
                n.matches3 <- 0
                if (this.site %in% c(site.strs2, site.strs15, site.strs5)) {
                    ## this.site ends with alphabet
                    ## add fixed last character to tree/core part
                    if (nchar.this > 1) {
                        add2 <- typos.to.numeric(substr(this.site, nchar.this,
                                                        nchar.this),
                                                 one.only = FALSE)
                    }
                    if (nchar.this > 1 && grepl("\\d", add2)) {
                        tmp <- substr(this.site, 1, nchar.this - 1)
                        if (ignore.site.case) {
                            tmp <- tolower(tmp)
                        }
                        if ((ignore.site.case && tmp %in% frequent.lower) ||
                            (!ignore.site.case && tmp %in% frequent.sites)) {
                            if (n.matches1 > 0) {
                                next
                            }
                            matches <- class.ids[site.match == tmp]
                            suffix.these <- substr(ids.to.fix, nchar.this + 1,
                                                   max.nchar.ids)
                            suffix2 <- paste0(add2, suffix.these)
                            sfxes <- substr(matches, nchar.this, max.nchar.ids)
                            if (!((!auto.case && ignore.case &&
                                   any(tolower(suffix2) %in% tolower(sfxes))) ||
                                  ((auto.case || !ignore.case) &&
                                   any(suffix2 %in% sfxes)) ||
                                  any(!(unique(nchar(suffix2)) %in%
                                        unique(nchar(sfxes)))))) {
                                close.match2 <- tmp
                                n.matches2 <- 1
                            } else {
                                next
                            }
                        }
                    }
                    ## drop fixed first character of tree/core
                    ## part, glue it to site part
                    charsN1 <- substr(ids.to.fix,
                                      nchar.this + 1, nchar.this + 1)
                    drop3.u <- typos.to.upper(charsN1)
                    isalpha.u <- grepl("[[:alpha:]]", drop3.u)
                    drop3.l <- typos.to.lower(charsN1)
                    isalpha.l <- grepl("[[:alpha:]]", drop3.l)
                    close.match3 <- character(n.fix)
                    for (idC in seq_len(n.fix)) {
                        if (isalpha.u[idC]) {
                            match.u <- match.l <- FALSE
                            if (ignore.site.case) {
                                tmp.l <-
                                    tolower(paste0(this.site,
                                                   drop3.u[idC])) # yes, .u
                                match.l <- tmp.l %in% frequent.lower
                            } else {
                                tmp.u <- paste0(this.site, drop3.u[idC])
                                match.u <- tmp.u %in% frequent.sites
                                if (isalpha.l[idC]) {
                                    tmp.l <- paste0(this.site, drop3.l[idC])
                                    match.l <- tmp.l %in% frequent.sites
                                }
                            }
                            if (n.matches1 + n.matches2 > 0) {
                                if (match.u || match.l) {
                                    n.matches3 <- -1
                                    break
                                } else {
                                    next
                                }
                            }
                            if (match.u && !match.l) {
                                matches <- class.ids[site.strs == tmp.u]
                                suffix3 <- substr(ids.to.fix[idC],
                                                  nchar.this + 2,
                                                  max.nchar.ids)
                                sfxes <- substr(matches, nchar.this + 2,
                                                max.nchar.ids)
                                if (!(suffix3 %in% sfxes) &&
                                    nchar(suffix3) %in% unique(nchar(sfxes))) {
                                    close.match3[idC] <- tmp.u
                                    n.matches3 <- 1
                                } else {
                                    n.matches3 <- -1
                                    break
                                }
                            } else if(match.l && !match.u) {
                                matches <- class.ids[site.match == tmp.l]
                                suffix3 <- substr(ids.to.fix[idC],
                                                  nchar.this + 2,
                                                  max.nchar.ids)
                                sfxes <- substr(matches, nchar.this + 2,
                                                max.nchar.ids)
                                if (!((!auto.case && ignore.case &&
                                       tolower(suffix3) %in% tolower(sfxes)) ||
                                      ((auto.case || !ignore.case) &&
                                       suffix3 %in% sfxes) ||
                                      !(nchar(suffix3) %in%
                                        unique(nchar(sfxes))))) {
                                    close.match3[idC] <- tmp.l
                                    n.matches3 <- 1
                                } else {
                                    n.matches3 <- -1
                                    break
                                }
                            } else {
                                n.matches3 <- -1
                                break
                            }
                        } else if (n.matches1 + n.matches2 == 0) {
                            n.matches3 <- -1
                            break
                        }
                    }
                    if (n.matches3 < 0) {
                        next
                    }
                } else if (this.site %in% site.strs3) {
                    ## this.site is numeric
                    if (nchar.this > 1) {
                        charN <- substr(this.site, nchar.this, nchar.this)
                        add2.u <- typos.to.upper(charN)
                        add2.l <- typos.to.lower(charN)
                    }
                    if (nchar.this > 1 && grepl("[[:alpha:]]", add2.u)) {
                        tmp <- substr(this.site, 1, nchar.this - 1)
                        if (tmp %in% frequent.sites) {
                            if (n.matches1 > 0) {
                                next
                            }
                            matches <- class.ids[site.strs == tmp]
                            suffix.these <- substr(ids.to.fix, nchar.this + 1,
                                                   max.nchar.ids)
                            suffix.u <- paste0(add2.u, suffix.these)
                            sfxes <- substr(matches, nchar.this, max.nchar.ids)
                            if (!all(unique(nchar(suffix.u)) %in%
                                     unique(nchar(sfxes)))) {
                                next
                            }
                            if (grepl("[[:alpha:]]", add2.l)) {
                                suffix.l <- paste0(add2.l, suffix.these)
                                lower.ok <- TRUE
                            } else {
                                lower.ok <- FALSE
                            }
                            if (!auto.case && ignore.case) {
                                suffix.l <- tolower(suffix.u) # yes, .u
                                if (!any(suffix.l %in% tolower(sfxes))) {
                                    suffix2 <- suffix.l
                                    close.match2 <- tmp
                                    n.matches2 <- 1
                                } else {
                                    next
                                }
                            } else if (!(any(suffix.u %in% sfxes) ||
                                         (lower.ok &&
                                          any(suffix.l %in% sfxes)))) {
                                if (lower.ok) {
                                    sfxes1 <- substr(sfxes, 1, 1)
                                    n.upper <- length(grep("[^[:lower:]]",
                                                           sfxes1))
                                    n.lower <- length(grep("[^[:upper:]]",
                                                           sfxes1))
                                    if (n.lower > n.upper) {
                                        suffix2 <- suffix.l
                                    } else {
                                        suffix2 <- suffix.u
                                    }
                                } else {
                                    suffix2 <- suffix.u
                                }
                                close.match2 <- tmp
                                n.matches2 <- 1
                            } else {
                                next
                            }
                        }
                    }
                    ## drop fixed first character of tree/core
                    ## part, glue it to site part
                    charsN1 <- substr(ids.to.fix, nchar.this+1, nchar.this+1)
                    drop3 <- typos.to.numeric(charsN1, one.only=FALSE)
                    isnum <- grepl("\\d", drop3)
                    close.match3 <- character(n.fix)
                    for (idC in seq_len(n.fix)) {
                        if (isnum[idC]) {
                            tmp <- paste0(this.site, drop3[idC])
                            if (tmp %in% frequent.sites) {
                                if (n.matches1 + n.matches2 > 0) {
                                    n.matches3 <- -1
                                    break
                                }
                                matches <- class.ids[site.strs == tmp]
                                suffix3 <- substr(ids.to.fix[idC],
                                                  nchar.this+2,
                                                  max.nchar.ids)
                                sfxes <- substr(matches, nchar.this + 2,
                                                max.nchar.ids)
                                if (!((!auto.case && ignore.case &&
                                       tolower(suffix3) %in% tolower(sfxes)) ||
                                      ((auto.case || !ignore.case) &&
                                       suffix3 %in% sfxes) ||
                                      !(nchar(suffix3) %in%
                                        unique(nchar(sfxes))))) {
                                    close.match3[idC] <- tmp
                                    n.matches3 <- 1
                                } else {
                                    n.matches3 <- -1
                                    break
                                }
                            } else if (n.matches1 + n.matches2 == 0) {
                                n.matches3 <- -1
                                break
                            }
                        } else if (n.matches1 + n.matches2 == 0) {
                            n.matches3 <- -1
                            break
                        }
                    }
                    if (n.matches3 < 0) {
                        next
                    }
                }
                if (ignore.site.case) {
                    already.fixed <- c(already.fixed, these.sites)
                }
                if (n.matches1 == 1) {
                    part1 <- rep(close.match1, n.fix)
                    part2 <- substr(ids.to.fix, nchar.this + 1,
                                    max.nchar.ids)
                } else if (n.matches2 == 1) {
                    part1 <- rep(close.match2, n.fix)
                    part2 <- suffix2
                    warn.changed(substr(ids.to.fix,
                                        nchar.this + 1,
                                        max.nchar.ids), part2)
                } else {
                    part1 <- close.match3
                    part2 <- substr(ids.to.fix, nchar.this + 2,
                                    max.nchar.ids)
                    warn.changed(substr(ids.to.fix,
                                        nchar.this + 1,
                                        max.nchar.ids), part2)
                }
                fixed.ids <- paste0(part1, part2)
                class.ids[fix.idx] <- fixed.ids
                ids[class.idx2[fix.idx]] <- fixed.ids
                site.strs[fix.idx] <- part1
                fixed.freq <- sort(table(part1), decreasing = TRUE)
                fixed.sites <- names(fixed.freq)
                for (fixC in seq_along(fixed.sites)) {
                    if (fixed.freq[fixC] > 1) {
                        warning(sprintf(paste0(warn.times, warn.fmt),
                                        fixed.freq[fixC],
                                        this.site, fixed.sites[fixC]),
                                domain = NA)
                    } else {
                        warning(sprintf(warn.fmt,
                                        this.site, fixed.sites[fixC]),
                                domain = NA)
                    }
                }
                if (ignore.site.case) {
                    foo <- which(unique.sites %in% these.sites)
                } else {
                    foo <- which(unique.sites == this.site)
                }
                unique.sites <- unique.sites[-foo]
                site.freq <- site.freq[-foo]
                if (ignore.site.case) {
                    lower.sites <- tolower(unique.sites)
                    unique.lower <- unique(lower.sites)
                    tmp <- match(lower.sites, unique.lower)
                    site.freq.lower <- numeric(length(lower.sites))
                    names(site.freq.lower) <- unique.sites
                    for (i in seq_along(unique.lower)) {
                        tmp2 <- tmp == i
                        site.freq.lower[tmp2] <- sum(site.freq[tmp2])
                    }
                }
            }
            if (ignore.site.case) {
                site.strs <- tolower(site.strs)
            }
            unique.sites <- sort(unique(c(site.strs, site.str1)))
            site.vec[class.idx2] <- as.numeric(match(site.strs, unique.sites))
        } else {
            unique.sites <- site.str1
        }
        nchar.unique <- nchar(unique.sites)
        site.vec[class.idx1] <- as.numeric(match(site.str1, unique.sites))
        ## 2.8 Site ID for some of the remaining names in rwl
        for (nc in sort(unique(nchar.unique), decreasing=TRUE)) {
            na.site.vec <- which(is.na(site.vec))
            n.na <- length(na.site.vec)
            if (n.na == 0) {
                break
            }
            if (ignore.site.case) {
                na.site.strs <- tolower(substr(ids[na.site.vec], 1, nc))
            } else {
                na.site.strs <- substr(ids[na.site.vec], 1, nc)
            }
            site.vec[na.site.vec] <-
                as.numeric(match(na.site.strs, unique.sites))
        }
        nchar.unique <- nchar.unique + 1 # for extracting tree-core part
        foo <- which(!is.na(site.vec))
        ## Removes punctuation (max 1 char) and / or spaces
        ## between the site part and the tree/core part
        ids[foo] <- sub("^\\s*[[:punct:]]?\\s*", "",
                        substring(ids[foo], nchar.unique[site.vec[foo]],
                                  max.nchar.ids))
        ## 2.9 Site ID (first previously unused) for the last
        ##     remaining names.  Site part is empty.
        na.site.vec <- which(is.na(site.vec))
        n.na <- length(na.site.vec)
        if (n.na > 0) {
            if (n.na < n.cases) {
                site.vec[na.site.vec] <- max(site.vec, na.rm=TRUE) + 1
            } else {
                site.vec[na.site.vec] <- 1
            }
        }
        n.sites <- max(site.vec)
        stc.t.all <- rep(NA_real_, n.cases)
        ## 3. For each site...
        for (siteCounter in seq_len(n.sites)) {
            site.idx <- which(site.vec == siteCounter)
            n.in.site <- length(site.idx)
            ids.site <- ids[site.idx]
            nchar.ids <- nchar(ids.site)
            if (fix.typos && min(nchar.ids) == max(nchar.ids)) {
                ids.site <- typofix(ids.site)
            }
            if (!auto.case && ignore.case) {
                ids.site <- tolower(ids.site)
            }
            ## Find (guess) tree-core boundaries
            stc.t.vec <- rep(NA_real_, n.in.site)
            ## 3.1 Try splitting tree /core ID at different positions
            class.idx2 <- which(nchar.ids > 1)
            n.class <- length(class.idx2)
            if (n.class > 0) {
                nchar.ids.class <- nchar.ids[class.idx2]
                max.nchar.ids.class <- max(nchar.ids.class)
                ids.class <- ids.site[class.idx2]
                remaining.flag <- rep(TRUE, n.class)
                remaining.idx <- 1:n.class
                alpha.flag <- grepl(ALPHA.MISC, ids.class)
            } else {
                max.nchar.ids.class <- 0
            }
            numeric.trees <- FALSE
            for (stc.t in dec(max.nchar.ids.class - 1, 1)) {
                firstpart <- substr(ids.class, 1, stc.t)
                lastpart <- substr(ids.class,
                                   stc.t + 1, max.nchar.ids.class)
                ## Look for:
                ## - first part mostly digits, last part starts with
                ##   anything but a digit
                ## "Mostly" means that in addition to digits or
                ## alphabets, "_", "?" and "*" are accepted
                ## (perhaps used as markers for missing
                ## information).
                rem.first <- firstpart[remaining.idx]
                rem.last <- lastpart[remaining.idx]
                rem.all <- ids.class[remaining.idx]
                digits.first <- grepl(DIGIT.MISC, rem.first)
                temp.flag <- digits.first & grepl(START.NO.DIGIT,
                                                  rem.last)
                n.temp <- sum(temp.flag)
                num.notnum <- n.temp > 0
                if (fix.typos && num.notnum &&
                    n.temp < length(remaining.idx)) {
                    fixed.last <- typos.to.numeric(rem.last[temp.flag],
                                                   one.only=FALSE)
                    temp.first <- rem.first[temp.flag]
                    if (all(grepl(DIGIT.MISC, fixed.last)) &&
                        any(grepl(DIGIT.MISC,
                                  rem.all[!temp.flag &
                                          nchar.ids.class[remaining.idx] >
                                          stc.t])) &&
                        !any(temp.first %in% rem.all[!temp.flag])) {
                        foo <- paste0(temp.first, fixed.last)
                        ids.site[class.idx2[remaining.idx[temp.flag]]] <- foo
                        warn.changed(rem.all[temp.flag], foo)
                        num.notnum <- FALSE
                    }
                }
                if (num.notnum) {
                    match.idx <- remaining.idx[temp.flag]
                    remaining.flag[match.idx] <- FALSE
                    numeric.trees <- TRUE
                } else {
                    match.idx <- integer(0)
                    remaining.flag[remaining.idx[temp.flag]] <- FALSE
                }
                remaining.idx <- which(remaining.flag)
                n.remaining <- length(remaining.idx)
                ## Look for:
                ## - first part mostly alphabets, last part starts with
                ##   anything but an alphabet
                if (n.remaining > 0) {
                    rem.first <- firstpart[remaining.idx]
                    rem.last <- lastpart[remaining.idx]
                    alphas.first <- grepl(ALPHA.MISC, rem.first)
                    temp.flag <- alphas.first & grepl(START.NO.ALPHA,
                                                      rem.last)
                    temp.idx <- remaining.idx[temp.flag]
                    match.idx <- c(match.idx, temp.idx)
                    remaining.flag[temp.idx] <- FALSE
                    remaining.idx <- which(remaining.flag)
                    n.remaining <- length(remaining.idx)
                }
                if (n.remaining > 0 && (auto.case || !ignore.case)) {
                    ## In IDs that mostly contain alphabets
                    temp.idx <- remaining.idx[alpha.flag[remaining.idx]]
                    ## ...and have at least stc.t + 1 characters...
                    temp.idx <- temp.idx[nchar.ids.class[temp.idx] > stc.t]
                    ## ...look for first part all upper case,
                    ## last part all lower case. Or vice versa.
                    temp.first <- firstpart[temp.idx]
                    temp.last <- lastpart[temp.idx]
                    lower.first <- grepl(LOWER.MISC, temp.first)
                    upper.first <- grepl(UPPER.MISC, temp.first)
                    lower.last <- grepl(LOWER.MISC, temp.last)
                    upper.last <- grepl(UPPER.MISC, temp.last)
                    uplow <- upper.first & lower.last
                    lowup <- lower.first & upper.last
                    temp.flag <- uplow | lowup
                    temp.idx <- temp.idx[temp.flag]
                    match.idx <- c(match.idx, temp.idx)
                    remaining.flag[temp.idx] <- FALSE
                    remaining.idx <- which(remaining.flag)
                    n.remaining <- length(remaining.idx)
                }
                n.match <- length(match.idx)
                if (n.match > 0) {
                    ## Find matching series without a core ID
                    if (n.remaining > 0) {
                        nchar.flag <- nchar.ids.class[remaining.idx] == stc.t
                        ends.here <- remaining.idx[nchar.flag]
                        if (auto.case || ignore.case) {
                            new.matches <-
                                ends.here[tolower(firstpart[ends.here]) %in%
                                          tolower(firstpart[match.idx])]
                        } else {
                            new.matches <- ends.here[firstpart[ends.here] %in%
                                                     firstpart[match.idx]]
                        }
                        match.idx <- c(match.idx, new.matches)
                        remaining.flag[new.matches] <- FALSE
                        remaining.idx <- which(remaining.flag)
                        n.remaining <- length(remaining.idx)
                    }
                    stc.t.vec[class.idx2[match.idx]] <- stc.t
                }
                if (n.remaining == 0) {
                    break
                }
            }
            ## 3.2 Mostly digits ("_", "?" and "*" are accepted)
            na.flag <- is.na(stc.t.vec)
            class.idx3 <- which(na.flag)
            class.idx3 <- class.idx3[grep(DIGIT.MISC, ids.site[class.idx3])]
            class.flag3 <- rep(FALSE, n.in.site)
            class.flag3[class.idx3] <- TRUE
            n.class <- length(class.idx3)
            if (n.class > 0) {
                if (fix.typos) {
                    ## Try to fix the names so that new "mostly digits"
                    ## cases are found
                    new.matches <- which(na.flag & !class.flag3)
                    fixed.ids <- typos.to.numeric(ids.site[new.matches],
                                                  one.only = FALSE)
                    fix.ok <- grepl(DIGIT.MISC, fixed.ids)
                    ok.idx <- which(fix.ok)
                    n.ok <- length(ok.idx)
                    if (n.ok > 0) {
                        ## Try to avoid false positives:
                        ## for a fix to be valid, all the character
                        ## positions where a change was made must have
                        ## a DIGITS.MISC character across the site
                        tmp.ids <- c(fixed.ids, ids.site[!na.flag])
                        nchar.tmp.ids <- nchar(tmp.ids)
                        max.nchar.tmp <- max(nchar(fixed.ids[fix.ok]))
                        all.digit.misc <- logical(max.nchar.tmp)
                        for (i in seq_len(max.nchar.tmp)) {
                            all.digit.misc[i] <-
                                all(nchar.tmp.ids < i |
                                    grepl(DIGIT.MISC, substr(tmp.ids, i, i)))
                        }
                        split.orig <-
                            strsplit(ids.site[new.matches[fix.ok]], "")
                        split.fixed <- strsplit(fixed.ids[fix.ok], "")
                        for (i in seq_len(n.ok)) {
                            fixed.pos <-
                                which(split.orig[[i]] != split.fixed[[i]])
                            if (any(!all.digit.misc[fixed.pos])) {
                                fix.ok[ok.idx[i]] <- FALSE
                            }
                        }
                    }
                    new.matches <- new.matches[fix.ok]
                    if (length(new.matches) > 0) {
                        fixed.ids <- fixed.ids[fix.ok]
                        warn.changed(ids.site[new.matches], fixed.ids)
                        ids.site[new.matches] <- fixed.ids
                        class.flag3[new.matches] <- TRUE
                        class.idx3 <- which(class.flag3)
                        n.class <- length(class.idx3)
                    }
                }
                ids.class <- ids.site[class.idx3]
                foo <- nchar.ids[class.idx3]
                min.nchar.ids.class <- min(foo)
                max.nchar.ids.class <- max(foo)
                min.stc.t <- max.nchar.ids.class - min.nchar.ids.class + 1
                padded.ids <- min.nchar.ids.class != max.nchar.ids.class
                if (padded.ids) {
                    ## Pad with spaces on the left.
                    ids.class <-
                        str_pad(ids.class, width = max.nchar.ids.class,
                                side = "left", pad = " ")
                    ids.site[class.idx3] <- ids.class
                }
                numeric.ids <- suppressWarnings(as.numeric(ids.class))
                all.digits <- !is.na(numeric.ids)
                numeric.ids <- numeric.ids[all.digits]
                n.cases2 <- length(numeric.ids)
                ids2 <- ids.class[all.digits] ## only use "all digits" cases
                ii <- order(numeric.ids)
                ordered.ids <- numeric.ids[ii]
                if (numeric.trees || all(diff(ordered.ids) == 1)) {
                    ## a. previously found numeric trees followed by
                    ##    something else or
                    ## b. continuous numbering of series
                    ## => no reason to believe that the numeric IDs
                    ##    would contain both tree and core parts
                    stc.t <- max.nchar.ids.class
                } else {
                    ## Try to find a pattern: Cores numbered continuously
                    ord.char.ids <- ids2[ii]
                    stc.t <- NA
                    remember.zero <- FALSE
                    for (stcIter in dec(max.nchar.ids.class, min.stc.t + 1)) {
                        num.firstpart <- as.numeric(substr(ord.char.ids,
                                                           1, stcIter - 1))
                        if (0 %in% num.firstpart) {
                            break
                        }
                        num.lastpart <-
                            as.numeric(substr(ord.char.ids,
                                              stcIter, max.nchar.ids.class))
                        unique.nums <- sort(unique(num.lastpart))
                        diff.unums <- diff(unique.nums)
                        if (((!remember.zero ||
                              10^(max.nchar.ids.class - stcIter) %in%
                              unique.nums) &&
                             all(diff.unums == 1) &&
                             !(remember.zero <- FALSE)) ||
                            (unique.nums[1] == 0 &&
                             as.character(unique.nums[length(unique.nums)]) ==
                             paste0(rep("9",
                                        max.nchar.ids.class - stcIter + 1),
                                    collapse="") &&
                             all(diff.unums[-1] == 1) &&
                             (remember.zero <- TRUE))) {
                            if (remember.zero) {
                                next
                            }
                            bdary <- which(ord.char.ids[-1] !=
                                           ord.char.ids[-n.cases2])
                            n.bdary <- length(bdary)
                            if (n.bdary == 0) {
                                break
                            }
                            stc.t <- stcIter - 1
                        } else {
                            break
                        }
                    }
                }
                if (!is.na(stc.t)) {
                    stc.t.vec[class.idx3] <- stc.t
                } else if(padded.ids) {
                    stc.t.vec[class.idx3] <- max.nchar.ids.class
                }
            }
            ## 3.3 ID is empty or one character (not a digit)
            na.flag <- is.na(stc.t.vec)
            temp.flag <- na.flag & nchar.ids == 1
            stc.t.vec[temp.flag] <- 1
            class.flag0 <- na.flag & nchar.ids == 0
            stc.t.vec[class.flag0] <- 0
            ## 3.4 Tree and core parts separated by punctuation or spaces
            class.idx1 <- which(is.na(stc.t.vec))
            stc.t.temp <- regexpr("[^[:punct:][:space:]][[:punct:][:space:]].",
                                  ids.site[class.idx1])
            match.idx <- which(stc.t.temp != -1)
            class.idx1 <- class.idx1[match.idx]
            class.flag1 <- rep(FALSE, n.in.site)
            class.flag1[class.idx1] <- TRUE
            if (length(match.idx) > 0) {
                stc.t.temp <- stc.t.temp[match.idx]
                stc.t.vec[class.idx1] <- stc.t.temp
                ids.tmp <- ids.site[class.idx1]
                ids.site[class.idx1] <-
                    paste0(substring(ids.tmp, 1, stc.t.temp),
                           substring(ids.tmp, stc.t.temp + 2, max.nchar.ids))
            }
            na.flag <- is.na(stc.t.vec)
            remaining.idx <- which(na.flag)
            n.remaining <- length(remaining.idx)
            ## 3.5 Compare prefixes of remaining tree-core parts to
            ##     previously found tree parts
            if (n.remaining > 0 && n.remaining < n.in.site) {
                notna <- which(!na.flag)
                unique.trees <- unique(substring(ids.site[notna],
                                                 1, stc.t.vec[notna]))
                nchar.trees <- nchar(unique.trees)
                rem.ids <- ids.site[remaining.idx]
                nchar.rem <- nchar(rem.ids)
                for (nc in sort(unique(nchar.trees), decreasing = TRUE)) {
                    these.trees <- unique.trees[nchar.trees == nc]
                    idx.tmp <- which(nchar.rem > nc)
                    idx.tmp <- idx.tmp[substr(rem.ids[idx.tmp], 1, nc) %in%
                                       these.trees]
                    if (length(idx.tmp) > 0) {
                        stc.t.vec[remaining.idx[idx.tmp]] <- nc
                        remaining.idx <- remaining.idx[-idx.tmp]
                        n.remaining <- length(remaining.idx)
                        if (n.remaining == 0) {
                            break
                        }
                        rem.ids <- rem.ids[-idx.tmp]
                        nchar.rem <- nchar.rem[-idx.tmp]
                    }
                }
            }
            ## 3.6 Fallback solution
            if (n.remaining > 0) {
                if (use.cor) {
                    stc.t.temp <- cor.clust(rwl[site.idx[remaining.idx]],
                                            ids.site[remaining.idx])
                } else {
                    stc.t.temp <- default.stc.t(ids.site, stc.t.vec)
                }
                ## If any tree ID is zero, override stc.t.temp
                rem.ids <- ids.site[remaining.idx]
                while (stc.t.temp < max(nchar.ids[remaining.idx])) {
                    treepart <- substr(rem.ids, 1, stc.t.temp)
                    numtree <- suppressWarnings(as.numeric(treepart))
                    if (identical(any(numtree == 0), TRUE)) {
                        stc.t.temp <- stc.t.temp + 1
                    } else {
                        break
                    }
                }
                warn.fmt <- if (n.remaining == 1) {
                    gettext('uncertain tree-core scheme in %d name ("%s")',
                            domain = "R-dplR")
                } else if (n.remaining == 2) {
                    gettext('uncertain tree-core scheme in %d names ("%s", "%s")',
                            domain = "R-dplR")
                } else {
                    gettext('uncertain tree-core scheme in %d names ("%s", ..., "%s")',
                            domain = "R-dplR")
                }
                warn.fmt <- paste0(warn.fmt, ", ",
                                   gettext("using %d as size of tree part",
                                           domain = "R-dplR"))
                if (n.remaining == 1) {
                    warning(sprintf(warn.fmt, n.remaining, rem.ids[1],
                                    stc.t.temp), domain = NA)
                } else {
                    warning(sprintf(warn.fmt, n.remaining, rem.ids[1],
                                    rem.ids[n.remaining],
                                    stc.t.temp), domain = NA)
                }
                stc.t.vec[remaining.idx] <- stc.t.temp
            }
            stc.t.all[site.idx] <- stc.t.vec
            ids[site.idx] <- ids.site
        }
        ## 4.Tries to remove some unnecessary characters from the
        ## beginning of the tree part, so that no new collisions
        ## between sites are introduced.
        ## MAX.ITER limits how many trimming iterations can be done.
        ## Trimming is nice but not essential as it does not matter
        ## much if some tree IDs contain more digits / characters than
        ## necessary.  In conclusion, don't put too much effort on it.
        MAX.ITER <- 10
        idStartMax <- rep(NA_real_, n.sites)
        site.idx <- vector(mode = "list", length = n.sites)
        for (siteCounter in seq_len(n.sites)) {
            tmp.idx <- which(site.vec == siteCounter)
            stc.t.vec <- stc.t.all[tmp.idx]
            idStartMax[siteCounter] <-
                trim.from.start(ids[tmp.idx], stc.t.vec)
            site.idx[[siteCounter]] <- tmp.idx
        }
        idStart <- idStartMax
        counter <- 1
        while (!any(idStart == 0) && counter <= MAX.ITER + 2) {
            all.ones <- all(idStart == 1)
            all.max <- all(idStart == idStartMax)
            tree.vec <- rep(NA_real_, n.cases)
            core.vec <- rep(NA_real_, n.cases)
            tree.collisions <- FALSE
            for (siteCounter in seq_len(n.sites)) {
                tmp.idx <- site.idx[[siteCounter]]
                ids.site <- ids[tmp.idx]
                max.nchar.ids <- max(nchar(ids.site))
                ## 4.1 Try trimming from the beginning of the names.
                this.start <- idStart[siteCounter]
                stc.t.vec <- stc.t.all[tmp.idx]
                ## Tree and core strings
                tree.strs <- substring(ids.site, this.start, stc.t.vec)
                core.strs <- substring(ids.site, stc.t.vec + 1, max.nchar.ids)
                ## 4.2 Get tree and core IDs (includes fixing any
                ##     collisions).  Results from the last iteration
                ##     are kept.
                res <- tree.and.core.ids(tree.strs, core.strs, tree.vec,
                                         auto.case=auto.case, auto.trim=TRUE)
                tree.vec[tmp.idx] <- res$tree
                core.vec[tmp.idx] <- res$core
                tree.collisions <- tree.collisions || res$coll
                if (tree.collisions && !all.max) {
                    break
                }
            }
            if (!tree.collisions) {
                break
            }
            if (counter == MAX.ITER && !all.ones) {
                ## Second to last resort: no trimming
                idStart <- rep(1, n.sites)
            } else if(all.ones && any(idStartMax > 1)) {
                ## If even "no trimming" results in collisions,
                ## do full trimming per site
                idStart <- idStartMax
            } else if(counter > 1 && all(idStart == idStartMax)) {
                ## Avoid repeating from start
                break
            } else {
                ## idStart goes through some possible values
                idStart <- decrement(idStart, idStartMax)
            }
            counter <- counter + 1
        }
        if (!tree.collisions) {
            ## If no collisions, it might be one site after all...
            site.vec[] <- 1
        }
        ## Only use non-negative IDs in the output. Zero IDs are only
        ## possible if originally present in parts of 'ids'.
        tree.vec <- fix.negative(tree.vec)
    } else {
### 'auto.mode' is false, i.e. numeric 'stc' is present
        nchar.ids <- nchar(ids)
        if (any(nchar.ids < 1)) {
            stop("series names must be at least 1 character long")
        }
        site.chars <- c(1, stc[1])
        site.width <- max(0, site.chars[2] - site.chars[1] + 1)
        site.strs <- str_pad(substr(ids, site.chars[1], site.chars[2]),
                             width = site.width, side = "right", pad = " ")
        max.nchar.ids <- sum(stc)
        ## Optionally try to fix rare site strings
        if (fix.typos && site.chars[2] > site.chars[1]) {
            site.freq <- table(site.strs)
            unique.sites <- names(site.freq)
            if (ignore.site.case) {
                site.match <- tolower(site.strs)
                lower.sites <- tolower(unique.sites)
                unique.lower <- unique(lower.sites)
                tmp <- match(lower.sites, unique.lower)
                ## Sum of number of occurrences when case is ignored
                site.freq.lower <- numeric(length(lower.sites))
                for (i in seq_along(unique.lower)) {
                    tmp2 <- tmp == i
                    site.freq.lower[tmp2] <- sum(site.freq[tmp2])
                }
            } else {
                site.match <- site.strs
            }
            if (ignore.site.case) {
                max.freq <- max(site.freq.lower)
            } else {
                max.freq <- max(site.freq)
            }
            sites.to.fix <- unique.sites[site.freq <=
                                         floor(max.freq / typo.ratio)]
            warn.times <- gettext("%d times: ", domain = "R-dplR")
            warn.fmt <- gettext('changed site code "%s" to "%s"',
                                domain = "R-dplR")
        } else {
            sites.to.fix <- character(0)
        }
        for (this.site in sites.to.fix) {
            if (ignore.site.case) {
                frequent.sites <-
                    unique.sites[site.freq.lower >=
                                 typo.ratio * site.freq[this.site]]
                frequent.lower <- unique(tolower(frequent.sites))
            } else {
                frequent.sites <-
                    unique.sites[site.freq >=
                                 typo.ratio * site.freq[this.site]]
            }
            if (length(frequent.sites) == 0) {
                next
            }
            fix.idx <- which(site.strs == this.site)
            ids.to.fix <- ids[fix.idx]
            n.fix <- length(ids.to.fix)
            close.match1 <- amatches(this.site, frequent.sites,
                                     max.distance = 1,
                                     ignore.case = ignore.site.case)
            if (ignore.site.case) {
                close.match1 <- close.match1[tolower(close.match1) !=
                                             tolower(this.site)]
            }
            n.matches1 <- length(close.match1)
            if (n.matches1 > 1) {
                lookalikes <-
                    close.match1[vapply(close.match1, match.helper, FALSE,
                                        y=this.site)]
                if (ignore.site.case) {
                    lookalikes <- unique(tolower(lookalikes))
                }
                if (length(lookalikes) == 1) {
                    close.match1 <- lookalikes
                    n.matches1 <- 1
                } else {
                    next
                }
            } else if (n.matches1 == 1 && ignore.site.case) {
                close.match1 <- tolower(close.match1)
            }
            if (n.matches1 == 1) {
                matches <- ids[site.match == close.match1]
                part2 <- substr(ids.to.fix, site.width + 1, max.nchar.ids)
                sfxes <- substr(matches, site.width + 1, max.nchar.ids)
                if ((!auto.case && ignore.case &&
                     any(tolower(part2) %in% tolower(sfxes))) ||
                    ((auto.case || !ignore.case) && any(part2 %in% sfxes))) {
                    next
                }
            } else {
                next
            }
            ids[fix.idx] <- paste0(close.match1, part2)
            site.strs[fix.idx] <- close.match1
            if (n.fix > 1) {
                warning(sprintf(paste0(warn.times, warn.fmt), n.fix,
                                this.site, close.match1),
                        domain = NA)
            } else {
                warning(sprintf(warn.fmt, this.site, close.match1),
                        domain = NA)
            }
            foo <- which(unique.sites == this.site)
            unique.sites <- unique.sites[-foo]
            site.freq <- site.freq[-foo]
        }
        if (ignore.site.case) {
            site.strs <- tolower(site.strs)
        }
        unique.sites <- sort(unique(site.strs))
        n.sites <- length(unique.sites)
        if (n.sites > 1) {
            warning("there appears to be more than one site")
        }
        site.vec <- as.numeric(match(site.strs, unique.sites))
        tree.vec <- rep(NA_real_, n.cases)
        core.vec <- rep(NA_real_, n.cases)
        ids <- substr(ids, max(0, stc[1]) + 1, max.nchar.ids)
        tree.chars <- c(1, stc[2])
        core.chars <- c(tree.chars[2]+1, tree.chars[2] + stc[3])

        ## For each site, find tree and core IDs
        for (i in seq_len(n.sites)) {
            site.idx <- which(site.vec == i)
            these.ids <- ids[site.idx]
            nchar.these <- nchar(these.ids)
            if (fix.typos && min(nchar.these) == max(nchar.these)) {
                these.ids <- typofix(these.ids)
            }
            if (!auto.case && ignore.case) {
                these.ids <- tolower(these.ids)
            }
            tree.strs <- substr(these.ids, tree.chars[1], tree.chars[2])
            core.strs <- substr(these.ids, core.chars[1], core.chars[2])
            res <- tree.and.core.ids(tree.strs, core.strs, tree.vec,
                                     auto.case=auto.case, auto.trim=FALSE)
            tree.vec[site.idx] <- res$tree
            core.vec[site.idx] <- res$core
        }
        tree.vec <- fix.negative(tree.vec)
    }
    ## Only return site.vec if it contains some information.
    if (length(unique(site.vec)) == 1) {
        data.frame(tree=tree.vec, core=core.vec, row.names=names(rwl))
    } else {
        data.frame(tree=tree.vec, core=core.vec, site=site.vec,
                   row.names=names(rwl))
    }
}

context("multiple core functions of dplR")
test.bai.in <- function() {
    ## Test
    base.seq <- pi * seq(from=3, by=2, length.out=19)
    ones <- rep.int(1, 20)
    test_that("bai.in works with zero d2pith", {
        expect_equal(c(pi, base.seq), bai.in(data.frame(ones))[[1]])
    })
    test_that("bai.in works with nonzero d2pith", {
        expect_equal(c(base.seq, 41 * pi),
                     bai.in(data.frame(x1 = ones),
                            d2pith = data.frame(series="x1", d2pith=1))[[1]])
    })
}
test.bai.in()
test.bai.out <- function() {
    ## Test
    base.seq <- pi * seq(from=3, by=2, length.out=19)
    ones <- rep.int(1, 20)
    test_that("bai.out works with zero diam", {
        expect_equal(c(pi, base.seq), bai.out(data.frame(ones))[[1]])
    })
    test_that("bai.in works with nonzero diam", {
        expect_equal(c(base.seq, 41 * pi),
                     bai.out(data.frame(x1 = ones),
                             diam = data.frame(series="x1", diam=42))[[1]])
    })
}
test.bai.out()
test.ccf.series.rwl <- function() {
    ## Setup
    srs1 <- seq(from=1, to=2, length.out=500)
    names(srs1) <- seq_along(srs1)
    dat1 <- data.frame(srs1, srs1 + 0.05, srs1 + 0.1)
    ## perfect correlation at lag 0 (mean of dat1 is srs1 + constant)
    res1.1 <- ccf.series.rwl(rwl = dat1, series = srs1,
                             seg.length = 100, bin.floor = 100,
                             prewhiten = FALSE, biweight = TRUE,
                             make.plot = FALSE, floor.plus1 = FALSE)
    res1.2 <- ccf.series.rwl(rwl = dat1, series = srs1,
                             seg.length = 100, bin.floor = 100,
                             prewhiten = FALSE, biweight = FALSE,
                             make.plot = FALSE, floor.plus1 = TRUE)
    res1.3 <- ccf.series.rwl(rwl = dat1, series = srs1,
                             seg.length = 100, bin.floor = 100,
                             prewhiten = TRUE, biweight = FALSE,
                             make.plot = FALSE, floor.plus1 = TRUE)
    bins1.1 <- res1.1[["bins"]]
    bins1.2 <- res1.2[["bins"]]
    bins1.3 <- res1.3[["bins"]]
    nrow1.3 <- nrow(bins1.3)
    rnames1 <- rownames(res1.2[["ccf"]])

    srs2 <- sin(pi / 4 * seq_len(500)) + 1.5 # period is 8
    names(srs2) <- seq_along(srs2)
    dat2 <- data.frame(srs2)
    ## perfect correlation at lag 0 (the single column dat2 is a copy of srs2)
    res2 <- ccf.series.rwl(rwl = dat2, series = srs2,
                           seg.length = 250, bin.floor = 100,
                           prewhiten = FALSE, lag.max = 7,
                           make.plot = FALSE, floor.plus1 = TRUE)
    ccf2 <- res2[["ccf"]]
    bins2 <- res2[["bins"]]
    rnames2 <- rownames(ccf2)

    ## Test
    test_that("ccf.series.rwl bins are correct", {
        expect_equal(nrow(bins1.1), 7)
        expect_equal(nrow(bins1.2), 9)
        expect_equal(bins1.1[1, 1], 100)
        expect_equal(bins1.2[1, 1], 1)
        expect_equal(bins1.1[7, 2], 499)
        expect_equal(bins1.2[9, 2], 500)
        expect_equal(bins1.3[nrow1.3, 2], 500)
        expect_equal(nrow(bins2), 3)
        expect_equal(bins2[, 1], c(1, 126, 251))
        expect_equal(bins2[, 2], c(250, 375, 500))
    })
    test_that("lag 0 cor is 1 when series differ by a constant", {
        expect_equivalent(res1.1[["ccf"]]["lag.0", ], rep.int(1, 7))
        expect_equivalent(res1.2[["ccf"]]["lag.0", ], rep.int(1, 9))
        expect_equivalent(res1.3[["ccf"]]["lag.0", ], rep.int(1, nrow1.3))
    })
    test_that("ccf.series.rwl responds to lag.max", {
        expect_equal(length(rnames1), 11, info="default lag.max = 5")
        expect_equal(length(rnames2), 15)
    })
    test_that("lagged correlations with a sinusoid are correct", {
        expect_true(all(rnames2[apply(abs(ccf2), 2, which.min)] %in%
                        c("lag.-6", "lag.-2", "lag.2", "lag.6")),
                    info="phase difference of 1/4 or 3/4 cycles")
        expect_true(all(rnames2[apply(ccf2, 2, which.min)] %in%
                        c("lag.-4", "lag.4")),
                    info="phase difference of 1/2 cycles")
        expect_true(all(rnames2[apply(ccf2, 2, which.max)] == "lag.0"),
                    info="same phase")
    })
}
test.ccf.series.rwl()
test.combine.rwl <- function() {
    ## Setup
    v.1 <- 1 + runif(300)
    range.1 <- 51:400
    rnames.1 <- as.character(range.1)
    range.2 <- range.1 + 150
    rnames.2 <- as.character(range.2)
    range.3 <- range.1 + 350
    rnames.3 <- as.character(range.3)
    range.4 <- range.1 + 450
    rnames.4 <- as.character(range.4)
    df.1 <-
        structure(data.frame(col1 = c(v.1, rep.int(NA, 50)),
                             col2 = c(rep.int(NA, 25), v.1, rep.int(NA, 25)),
                             col3 = c(rep.int(NA, 50), v.1),
                             row.names = rnames.1),
                  class = c("rwl", "data.frame"))
    df.2 <- df.1
    rownames(df.2) <- rnames.2
    df.3 <- df.1
    rownames(df.3) <- rnames.3
    df.4 <- df.1
    rownames(df.4) <- rnames.4
    res.3 <- combine.rwl(list(df.1))
    res.4 <- combine.rwl(list(df.1, df.2, df.3, df.4))
    res.5 <- combine.rwl(df.1, df.1)
    res.6 <- combine.rwl(df.1, df.2)
    res.7 <- combine.rwl(df.1, df.3)
    res.8 <- combine.rwl(df.1, df.4)
    ## Test
    test_that("combine.rwl stops with nothing to combine", {
        expect_error(combine.rwl(list()),"nothing to combine",ignore.case=TRUE)
        expect_error(combine.rwl(df.1), "nothing to combine", ignore.case=TRUE)
    })
    test_that("combine.rwl works with a list of length one", {
        expect_equal(res.3, df.1)
    })
    test_that("combine.rwl works with multiple data.frames", {
        expect_equal(ncol(res.4), 12)
        expect_equal(res.4[1:350, 1:3], df.1)
        expect_equal(res.4[150+(1:350), 4:6], df.2)
        expect_equal(res.4[350+(1:350), 7:9], df.3)
        expect_equal(res.4[450+(1:350), 10:12], df.4)
    })
    test_that("combine.rwl works with identical data.frames", {
        ## ... but names will be duplicated (names are not tested)
        expect_equal(ncol(res.5), 6)
        expect_equal(res.5[1:3], df.1)
        expect_equal(res.5[4:6], df.1)
    })
    ## 6. ...have partially overlapping years
    test_that("combine.rwl works with partially overlapping years", {
        expect_equal(ncol(res.6), 6)
        expect_equal(nrow(res.6), 500)
        expect_equal(res.6[1:350, 1:3], df.1)
        expect_equal(res.6[150+(1:350), 4:6], df.2)
    })
    ## 7. ...have separate sets of years so that the result is continuous
    ## (y starts where x ends)
    test_that("combine.rwl works with separate, continuous, years", {
        expect_equal(ncol(res.7), 6)
        expect_equal(nrow(res.7), 700)
        expect_equal(res.7[1:350, 1:3], df.1)
        expect_equal(res.7[350+(1:350), 4:6], df.3)
    })
    ## 8. ...have separate sets of years so that the result is discontinuous
    test_that("combine.rwl works with separate, discontinuous, years", {
        expect_equal(ncol(res.8), 6)
        expect_equal(nrow(res.8), 800)
        expect_equal(res.8[1:350, 1:3], df.1)
        expect_equal(res.8[450+(1:350), 4:6], df.4)
    })
}
test.combine.rwl()
test.corr.rwl.seg <- function() {
    ## Setup
    srs1 <- rep.int(seq(from=0.5, to=1.5, length.out=50), 10)
    srs2 <- rev(srs1)
    srs3 <- srs1
    srs3[26:75] <- rev(srs3[26:75])
    srs4 <- srs1
    srs4[126:175] <- rev(srs4[126:175])
    srs4[326:425] <- rev(srs4[326:425])
    names(srs1) <- seq_along(srs1)
    dat1 <- data.frame(a=srs1, b=srs1, c=srs1, d=srs1, e=srs1, f=srs1, g=srs1)
    dat2 <- dat1
    dat2[1] <- srs2
    dat3 <- dat1
    dat3[1] <- srs3
    dat3[2] <- srs4
    res1 <- corr.rwl.seg(dat1, seg.length=50, bin.floor=100, make.plot=FALSE)
    res2 <- corr.rwl.seg(dat2, seg.length=50, bin.floor=100, make.plot=FALSE)
    res3 <- corr.rwl.seg(dat3, seg.length=100, bin.floor=100, pcrit=0.05,
                         make.plot=FALSE)
    res4 <- corr.rwl.seg(dat3, seg.length=100, bin.floor=100, pcrit=0.05,
                         prewhiten=FALSE, floor.plus1=TRUE, make.plot=FALSE)
    expected.cnames1 <- paste(res1[["bins"]][, 1], res1[["bins"]][, 2], sep=".")
    expected.cnames3 <- paste(res3[["bins"]][, 1], res3[["bins"]][, 2], sep=".")
    expected.cnames4 <- paste(res4[["bins"]][, 1], res4[["bins"]][, 2], sep=".")
    expected.rnames <- c("a", "b", "c", "d", "e", "f", "g")
    expected.corr1 <- array(1, dim(res1[["spearman.rho"]]),
                            dimnames=list(expected.rnames, expected.cnames1))
    expected.corr2 <- expected.corr1
    expected.corr2[1, ] <- -1
    expected.overall1 <- array(data=c(rep.int(1, 7), rep.int(0, 7)),
                               dim=c(7,2), dimnames=list(expected.rnames,
                                           c("rho", "p-val")))
    expected.overall2 <- expected.overall1
    expected.overall2["a", "rho"] <- -1
    expected.overall2["a", "p-val"] <- 1
    seg.names1 <- paste(seq(from=100, to=450, by=25),
                        seq(from=149, to=499, by=25), sep=".")
    expected.avg1 <- rep.int(1, length(seg.names1))
    names(expected.avg1) <- seg.names1
    expected.avg2 <- rep.int(5/7, length(seg.names1))
    names(expected.avg2) <- seg.names1
    expected.flags1 <- array(0, dim(res1[["p.val"]]),
                             dimnames=list(expected.rnames, expected.cnames1))
    expected.flags2 <- expected.flags1
    expected.flags3 <- array(0, dim(res3[["p.val"]]),
                             dimnames=list(expected.rnames, expected.cnames3))
    expected.flags4 <- array(0, dim(res4[["p.val"]]),
                             dimnames=list(expected.rnames, expected.cnames4))
    expected.flags2[1, ] <- 1
    expected.flags3[2, c("100.199", "300.399", "350.449")] <- 1
    expected.flags4[1, "1.100"] <- 1
    expected.flags4[2, c("101.200", "301.400", "351.450")] <- 1
    res1.flags <- array(0, dim(res1[["p.val"]]),
                        dimnames=dimnames(res1[["p.val"]]))
    res1.flags[res1[["p.val"]] >= 0.05] <- 1
    res2.flags <- array(0, dim(res2[["p.val"]]),
                        dimnames=dimnames(res2[["p.val"]]))
    res2.flags[res2[["p.val"]] >= 0.05] <- 1
    res3.flags <- array(0, dim(res3[["p.val"]]),
                        dimnames=dimnames(res3[["p.val"]]))
    res3.flags[res3[["p.val"]] >= 0.05] <- 1
    res4.flags <- array(0, dim(res4[["p.val"]]),
                        dimnames=dimnames(res4[["p.val"]]))
    res4.flags[res4[["p.val"]] >= 0.05] <- 1

    ## Test
    test_that("corr.rwl.seg bins are correct", {
        expect_true(all(res1[["bins"]][, 2] - res1[["bins"]][, 1] + 1 == 50))
        expect_equal(res1[["bins"]][1, 1], 100)
        expect_true(all(diff(res1[["bins"]][, 1]) == 25))
        expect_equal(res1[["bins"]][nrow(res1[["bins"]]), 1], 450)
        expect_equal(res2[["bins"]], res1[["bins"]])
        expect_true(all(res3[["bins"]][, 2] - res3[["bins"]][, 1] + 1 == 100))
        expect_equal(res3[["bins"]][1, 1], 100)
        expect_true(all(diff(res3[["bins"]][, 1]) == 50))
        expect_equal(res3[["bins"]][nrow(res3[["bins"]]), 1], 400)
        expect_true(all(res4[["bins"]][, 2] - res4[["bins"]][, 1] + 1 == 100))
        expect_equal(res4[["bins"]][1, 1], 1)
        expect_true(all(diff(res4[["bins"]][, 1]) == 50))
        expect_equal(res4[["bins"]][nrow(res4[["bins"]]), 1], 401)
    })
    test_that("corr.rwl.seg correlations (by bin) are correct", {
        expect_equal(res1[["spearman.rho"]], expected.corr1)
        expect_equal(res2[["spearman.rho"]], expected.corr2)
    })
    test_that("corr.rwl.seg correlations (overall) are correct", {
        expect_equal(res1[["overall"]], expected.overall1)
        expect_equal(res2[["overall"]], expected.overall2)
    })
    test_that("corr.rwl.seg correlations (average) are correct", {
        expect_equal(res1[["avg.seg.rho"]], expected.avg1)
        expect_equal(res2[["avg.seg.rho"]], expected.avg2)
    })
    test_that("corr.rwl.seg P-values are correct", {
        expect_equal(res1.flags, expected.flags1)
        expect_equal(res2.flags, expected.flags2)
        expect_equal(res3.flags, expected.flags3)
        expect_equal(res4.flags, expected.flags4)
    })
    test_that("corr.rwl.seg flags are correct", {
        expect_equal(length(res1[["flags"]]), 0)
        expect_equal(length(res2[["flags"]]), 1)
        expect_equal(length(res3[["flags"]]), 1)
        expect_equal(length(res4[["flags"]]), 2)
        expect_equal(res2[["flags"]][["a"]],
                          paste(seg.names1, collapse=", "))
        expect_equal(res3[["flags"]][["b"]], "100.199, 300.399, 350.449")
        expect_equal(res4[["flags"]][["a"]], "1.100")
        expect_equal(res4[["flags"]][["b"]], "101.200, 301.400, 351.450")
    })
}
test.corr.rwl.seg()
test.corr.series.seg <- function() {
    ## Setup
    srs1 <- rep.int(seq(from=0.5, to=1.5, length.out=50), 10)
    srs2 <- rev(srs1)
    srs3 <- srs1
    srs3[26:75] <- rev(srs3[26:75])
    srs3[326:425] <- rev(srs3[326:425])
    srs4 <- rep.int(seq(1, 2, length.out=50) + sin((1:50)*0.4), 10)
    names(srs1) <- seq_along(srs1)
    names(srs2) <- seq_along(srs2)
    names(srs3) <- seq_along(srs3)
    names(srs4) <- seq_along(srs4)
    dat <- data.frame(a=srs1, b=srs1, c=srs1, d=srs1, e=srs1, f=srs1, g=srs1)
    res1 <- corr.series.seg(rwl=dat, series=srs1, seg.length=50,
                            bin.floor=100, make.plot=FALSE)
    res2 <- corr.series.seg(rwl=dat, series=srs2, seg.length=50,
                            bin.floor=100, make.plot=FALSE)
    res3 <- corr.series.seg(rwl=dat, series=srs3, seg.length=100,
                            bin.floor=100, make.plot=FALSE)
    res4 <- corr.series.seg(rwl=dat, series=srs3, seg.length=100,
                            prewhiten=FALSE, bin.floor=100,
                            make.plot=FALSE, floor.plus1=TRUE)
    res5 <- corr.series.seg(rwl=dat, series=srs4, seg.length=50,
                            biweight=FALSE, prewhiten=FALSE,
                            bin.floor=100, make.plot=FALSE)
    res6 <- corr.series.seg(rwl=dat, series=srs4, seg.length=50,
                            biweight=FALSE, prewhiten=FALSE,
                            bin.floor=100, make.plot=FALSE, method="spearman")
    res6.2 <- corr.series.seg(rwl=dat, series=srs4, seg.length=50,
                              biweight=FALSE, prewhiten=FALSE,
                              bin.floor=50, make.plot=FALSE, method="spearman")
    res7 <- corr.series.seg(rwl=dat, series=srs4, seg.length=50,
                            biweight=FALSE, prewhiten=FALSE,
                            bin.floor=100, make.plot=FALSE, method="pearson")
    res8 <- corr.series.seg(rwl=dat, series=srs4, seg.length=50,
                            biweight=FALSE, prewhiten=FALSE,
                            bin.floor=100, make.plot=FALSE, method="kendall")
    res9 <- corr.series.seg(rwl=dat, series=srs4, seg.length=48,
                            biweight=FALSE, prewhiten=FALSE,
                            bin.floor=100, make.plot=FALSE, method="pearson")
    res10 <- corr.series.seg(rwl=dat, series=srs4, seg.length=100,
                             biweight=FALSE, prewhiten=FALSE,
                             bin.floor=100, make.plot=FALSE, method="pearson")
    res11 <- corr.series.seg(rwl=dat, series=srs4, seg.length=142,
                             biweight=FALSE, prewhiten=FALSE,
                             bin.floor=100, make.plot=FALSE, method="pearson")

    expected.cnames1 <- paste(res1[["bins"]][, 1], res1[["bins"]][, 2], sep=".")
    expected.cnames3 <- paste(res3[["bins"]][, 1], res3[["bins"]][, 2], sep=".")
    expected.cnames4 <- paste(res4[["bins"]][, 1], res4[["bins"]][, 2], sep=".")
    expected.corr1 <- rep.int(1, length(res1[["spearman.rho"]]))
    names(expected.corr1) <- expected.cnames1
    expected.corr2 <- rep.int(-1, length(res2[["spearman.rho"]]))
    names(expected.corr2) <- expected.cnames1
    expected.overall1 <- c(1, 0)
    names(expected.overall1) <- c("rho", "p-val")
    expected.overall2 <- c(-1, 1)
    names(expected.overall2) <- c("rho", "p-val")
    expected.flags1 <- rep.int(0, length(res1[["p.val"]]))
    names(expected.flags1) <- names(res1[["p.val"]])
    expected.flags2 <- rep.int(1, length(res2[["p.val"]]))
    names(expected.flags2) <- names(res2[["p.val"]])
    expected.flags3 <- rep.int(0, length(res3[["p.val"]]))
    names(expected.flags3) <- names(res3[["p.val"]])
    expected.flags4 <- rep.int(0, length(res4[["p.val"]]))
    names(expected.flags4) <- names(res4[["p.val"]])
    expected.flags3[c("300.399", "350.449")] <- 1
    expected.flags4[c("1.100", "301.400", "351.450")] <- 1
    res1.flags <- rep.int(0, length(res1[["p.val"]]))
    names(res1.flags) <- names(res1[["p.val"]])
    res1.flags[res1[["p.val"]] >= 0.05] <- 1
    res2.flags <- rep.int(0, length(res2[["p.val"]]))
    names(res2.flags) <- names(res2[["p.val"]])
    res2.flags[res2[["p.val"]] >= 0.05] <- 1
    res3.flags <- rep.int(0, length(res3[["p.val"]]))
    names(res3.flags) <- names(res3[["p.val"]])
    res3.flags[res3[["p.val"]] >= 0.05] <- 1
    res4.flags <- rep.int(0, length(res4[["p.val"]]))
    names(res4.flags) <- names(res4[["p.val"]])
    res4.flags[res4[["p.val"]] >= 0.05] <- 1
    range.moving.3 <- range(res3[["moving.rho"]][, "rho"], na.rm=TRUE)
    range.3 <- range(res3[["spearman.rho"]])

    ## Test
    test_that("corr.series.seg bins are correct", {
        expect_true(all(res1[["bins"]][, 2] - res1[["bins"]][, 1] + 1 == 50))
        expect_equal(res1[["bins"]][1, 1], 100)
        expect_true(all(diff(res1[["bins"]][, 1]) == 25))
        expect_equal(res1[["bins"]][nrow(res1[["bins"]]), 1], 450)
        expect_equal(res1[["bins"]], res2[["bins"]])
        expect_true(all(res3[["bins"]][, 2] - res3[["bins"]][, 1] + 1 == 100))
        expect_equal(res3[["bins"]][1, 1], 100)
        expect_true(all(diff(res3[["bins"]][, 1]) == 50))
        expect_equal(res3[["bins"]][nrow(res3[["bins"]]), 1], 400)
        expect_true(all(res4[["bins"]][, 2] - res4[["bins"]][, 1] + 1 == 100))
        expect_equal(res4[["bins"]][1, 1], 1)
        expect_true(all(diff(res4[["bins"]][, 1]) == 50))
        expect_equal(res4[["bins"]][nrow(res4[["bins"]]), 1], 401)
    })
    test_that("corr.series.seg correlations (by bin) are correct", {
        expect_equal(res1[["spearman.rho"]], expected.corr1)
        expect_equal(res2[["spearman.rho"]], expected.corr2)
    })
    test_that("corr.series.seg correlations (overall) are correct", {
        expect_equal(res1[["overall"]], expected.overall1)
        expect_equal(res2[["overall"]], expected.overall2)
    })
    test_that("corr.series.seg P-values are correct", {
        expect_equal(res1.flags, expected.flags1)
        expect_equal(res2.flags, expected.flags2)
        expect_equal(res3.flags, expected.flags3)
        expect_equal(res4.flags, expected.flags4)
    })
    test_that("corr.series.seg correlations (moving) are correct", {
        expect_equal(range(res1[["moving.rho"]][, "rho"], na.rm=TRUE), c(1, 1))
        expect_equal(range(res2[["moving.rho"]][, "rho"], na.rm=TRUE),c(-1,-1))
        expect_equal(range.moving.3,
                     c(min(range.moving.3[1], range.3[1]),
                       max(range.moving.3[2], range.3[2])))
        expect_equal(range(res4[["moving.rho"]][, "rho"], na.rm=TRUE),c(-1, 1))
    })
    test_that("default method is spearman", {
        tmpNames <- names(res5)
        expect_named(res6, tmpNames)
        for (i in seq_along(res5)) {
            expect_equal(res6[[i]], res5[[i]], info = tmpNames[i])
        }
    })
    test_that("correlation methods differ", {
        expect_false(isTRUE(all.equal(res6[["overall"]], res7[["overall"]])))
        expect_false(isTRUE(all.equal(res6[["overall"]], res8[["overall"]])))
        expect_false(isTRUE(all.equal(res7[["overall"]], res8[["overall"]])))
        expect_false(isTRUE(all.equal(res6[["moving.rho"]],
                                      res7[["moving.rho"]])))
        expect_false(isTRUE(all.equal(res6[["moving.rho"]],
                                      res8[["moving.rho"]])))
        expect_false(isTRUE(all.equal(res7[["moving.rho"]],
                                      res8[["moving.rho"]])))
        expect_false(isTRUE(all.equal(res6[["spearman.rho"]],
                                      res7[["spearman.rho"]])))
        expect_false(isTRUE(all.equal(res6[["spearman.rho"]],
                                      res8[["spearman.rho"]])))
        expect_false(isTRUE(all.equal(res7[["spearman.rho"]],
                                      res8[["spearman.rho"]])))
    })
    tmp7 <- as.vector(na.omit(res7[["moving.rho"]][, "rho"]))
    test_that("correlations are ok when segment length matches common cycle", {
        expect_equal(length(tmp7), 451)
        expect_equal(tmp7, rep.int(mean(tmp7), 451))
    })
    tmp9 <- na.omit(res9[["moving.rho"]][, "rho"])
    uniqueRho9 <- unique(tmp9)
    test_that("correlations are ok with segments shorter than common cycle", {
        expect_equal(length(tmp9), 453)
        expect_equal(length(uniqueRho9), 50)
    })
    tmp10 <- as.vector(na.omit(res10[["moving.rho"]][, "rho"]))
    test_that("correlations are ok when multiple cycles fit segment exactly", {
        expect_equal(length(tmp10), 401)
        expect_equal(tmp10, rep.int(mean(tmp10), 401))
    })
    tmp11 <- na.omit(res11[["moving.rho"]][, "rho"])
    uniqueRho11 <- unique(tmp11)
    test_that("correlations are ok with segments longer than common cycle", {
        expect_equal(length(tmp11), 359)
        expect_equal(length(uniqueRho11), 50)
    })
    test_that("bin.floor argument works", {
        expect_equal(length(res6.2[["spearman.rho"]]),
                     length(res6[["spearman.rho"]])+2)
        expect_equal(res6.2[["spearman.rho"]][-c(1, 2)],
                     res6[["spearman.rho"]])
    })
}
test.corr.series.seg()
test.caps <- function() {
    ## Setup
    n <- 100
    x <- seq_len(n)
    y <- x + 10 * sin(pi / 15 * x) + 5 * rnorm(n)
    lm.y <- lm(y ~ x)
    fitted.y <- fitted(lm.y)
    res.1 <- caps(y, f=0, nyrs=30)
    res.2 <- caps(y, f=0.9, nyrs=30)
    res.3 <- caps(y, f=0.9, nyrs=5)
    res.4 <- caps(y, f=1, nyrs=30)
    res.5 <- caps(x)
    error.1 <- sum((y - res.1)^2)
    error.2 <- sum((y - res.2)^2)
    error.3 <- sum((y - res.3)^2)
    ## Test
    test_that("caps handles special cases", {
        expect_equivalent(res.1, fitted.y)
        expect_equal(res.4, y)
        expect_equal(res.5, x)
    })
    test_that("smoother spline means more error", {
        expect_true(error.1 > error.2)
        expect_true(error.2 > error.3)
    })
    test_that("caps stops on bad parameters", {
        expect_error(caps(y, f=-1), "between 0 and 1")
        expect_error(caps(y, f=2), "between 0 and 1")
        expect_error(caps(y, nyrs=0), "'nyrs' must be a number greater than 0")
    })
}
test.caps()
test.gini.coef <- function() {
    ## Setup
    MAX.SIZE <- 1000
    NTIMES <- 10
    samp <- sample(seq.int(2, MAX.SIZE), max(0, min(NTIMES, MAX.SIZE - 1)))
    ## Test
    coefs <- vapply(samp,
                    function(x) {
                        foo <- numeric(x)
                        n <- sample(x - 1, 1)
                        nonzeros <- sample(x, n)
                        val <- runif(1, 1, 100)

                        foo[nonzeros[1]] <- val
                        a <- gini.coef(foo)

                        foo[nonzeros] <- val
                        b <- gini.coef(foo)

                        foo[] <- val
                        c <- gini.coef(foo)

                        c(a, b, c, n)
                    }, numeric(4))
    test_that("gini.coef handles special cases", {
        expect_equal(coefs[1, ], 1 - 1 / samp)
        expect_equal(coefs[2, ], 1 - coefs[4, ] / samp)
        expect_equal(coefs[3, ], numeric(length(samp)))
    })
}
test.gini.coef()
# test.glk <- function() {
#     ## Setup
#     seq.inc <- seq_len(10)
#     seq.dec <- seq.int(from = -1, to = -10)
#     seq.rand <- sample(x = seq.inc, size = 10, replace = FALSE)
#     seq.step <- rep(seq.rand, each = 2)
#     seq.step <- seq.step[-length(seq.step)]
#     glk.4col <- glk(data.frame(seq.rand, seq.rand, seq.rand, seq.rand))
#     ## Test
#     test_that("result of glk is correctly formatted", {
#         expect_equal(nrow(glk.4col), 4)
#         expect_equal(ncol(glk.4col), 4)
#         expect_true(all(glk.4col[upper.tri(x = glk.4col, diag = FALSE)] == 1))
#         expect_true(all(is.na(glk.4col[lower.tri(x = glk.4col, diag = TRUE)])))
#     })
#     test_that("cases without simultaneous zero diffs are ok", {
#         expect_equal(glk(data.frame(seq.inc, seq.inc + 1))[1, 2], 1,
#                      info="strictly monotonic sequences (both increasing)")
#         expect_equal(glk(data.frame(seq.inc, seq.dec))[1, 2], 0,
#                      info="strictly monotonic sequences (incr., decr.)")
#         expect_equal(glk(data.frame(seq.rand, seq.rand + 1))[1, 2], 1,
#                      info="signs of differences are the same")
#         expect_equal(glk(data.frame(seq.rand, -seq.rand))[1, 2], 0,
#                      info="signs of differences are opposite")
#         expect_equal(glk(data.frame(seq.rand,
#                                     rep.int(1, length(seq.rand))))[1, 2],
#                      0.5, info="one sequence is constant")
#     })
#     test_that("dplR >= 1.6.1: zero diffs are in full agreement", {
#         expect_equal(glk(data.frame(seq.step, -seq.step))[1, 2], 0.5,
#                      info="a zero difference in both series is full agreement")
#         expect_equal(glk(data.frame(seq.step, seq.step))[1, 2], 1,
#                      info="glk() is 1 when comparing any sequence with itself")
#         expect_equal(glk(data.frame(seq.step,
#                                     rep.int(1, length(seq.step))))[1, 2],
#                      0.75, info="halfway between 0.5 and 1")
#     })
# }
# test.glk()
test.hanning <- function() {
    ## Setup
    SAMP.SIZE <- 101
    FILTER.LENGTH <- c(7, 51)
    HALF.SIZE <- 50
    x.constant <- rep.int(42, SAMP.SIZE)
    x.impulse <- c(rep.int(0, HALF.SIZE), 1, rep.int(0, HALF.SIZE))
    for (filter.length in FILTER.LENGTH) {
        length.str <- paste0("filter length ", filter.length)
        not.na.length <- SAMP.SIZE - filter.length + 1
        y.constant <- hanning(x.constant, n=filter.length)
        y.impulse <- hanning(x.impulse, n=filter.length)
        not.na.constant <- which(!is.na(y.constant))
        ## Test
        test_that("number of NA values is correct", {
            expect_equal(length(not.na.constant), not.na.length,
                         info=length.str)
        })
        test_that("a constant series stays constant", {
            expect_equal(y.constant[not.na.constant],
                         rep.int(42, not.na.length), info=length.str)
        })
        test_that("unit impulse copies the filter coefficients", {
            expect_equal(sum(y.impulse, na.rm=TRUE), 1, info=length.str)
        })
        ## Needs more tests (?)
    }
    test_that("hanning stops on filter length n < 3", {
        expect_error(hanning(x.constant, n=2))
    })
}
test.hanning()
test.net <- function() {
    ## Setup
    seq.inc <- seq_len(10)
    seq.rand <- sample(x = seq.inc, size = 10, replace = FALSE)
    rowNames <- as.character(seq(from=100, length.out=length(seq.inc)))
    testFrame <- data.frame(seq.rand, seq.rand, seq.rand, seq.rand,
                            row.names = rowNames)
    net.testFrame <- net(testFrame)
    ## Test
    test_that("result of net is correctly formatted", {
        expect_is(net.testFrame, "list")
        expect_named(net.testFrame, c("all", "average"))
        expect_named(net.testFrame[["all"]], rowNames)
        expect_equivalent(net.testFrame[["all"]], c(NA_real_, rep.int(0, 9)))
        expect_equal(net.testFrame[["average"]], 0)
    })
    test_that("net returns correct results", {
        seq.dec <- seq.int(from = -1, to = -10)
        testFrame2 <- data.frame(seq.inc, seq.inc, seq.inc, seq.dec)
        exp1 <- c(NA_real_, rep.int(2.25, 9))
        exp2 <- rep.int(2, 10)
        exp3 <- c(NA_real_, rep.int(0.25, 9))
        expect_equal(net(testFrame2)[["all"]], exp1)
        expect_equal(net(testFrame2, weights=c(v=1, 0))[["all"]], exp2)
        expect_equal(net(testFrame2, weights=c(g=1, 0))[["all"]], exp3)
        testFrame3 <- testFrame2[c(1:5, 5, 6:10), ]
        row.names(testFrame3) <- NULL
        expect_equal(net(testFrame3)[["all"]], c(exp1[1:5], 3, exp1[6:10]))
        expect_equal(net(testFrame3, weights=c(v=1, 0))[["all"]],
                     c(exp2[1:5], 2, exp2[6:10]))
        expect_equal(net(testFrame3, weights=c(g=1, 0))[["all"]],
                     c(exp3[1:5], 1, exp3[6:10]))
    })
    test_that("input can be matrix or data.frame", {
        net.matrix <- net(as.matrix(testFrame))
        expect_equal(net.matrix[["all"]], net.testFrame[["all"]])
        expect_equal(net.matrix[["average"]], net.testFrame[["average"]])
    })
    test_that("invalid input and parameters fail", {
        expect_error(net(1:5), "matrix-like")
        expect_error(net(as.matrix(1:5)), "2 columns")
        expect_error(net(t(as.matrix(1:5))), "2 rows")
        expect_error(net(testFrame, weights = c(dontexist = 1, 0)), "unknown")
        expect_error(net(testFrame, weights = c(1, NA_real_)), "is.finite")
        expect_error(net(testFrame, weights = c(1, 1, 1)), "length")
        expect_error(net(testFrame, weights = c("a", "b")), "is.numeric")
    })
}
test.net()
test.read.ids <- function() {
    ## Setup
    site <- "abc"
    tree <- c(1, 2, 2, 2, 3, 3, 4, 5, 5, 5, 5, 6)
    core <- c(1, 1, 2, 3, 1, 2, 1, 1, 2, 3, 4, 1)
    ids1 <- paste(site, "0", tree, core, sep="")
    n.ids1 <- length(ids1)
    ids2 <- ids1[3:n.ids1]
    ids3 <- paste(site, "x", LETTERS[tree], letters[core], sep="")
    frame1 <- as.data.frame(matrix(data=1, nrow=1, ncol=length(ids1),
                                   dimnames=list("1", ids1)))
    frame2 <- as.data.frame(matrix(data=1, nrow=1, ncol=length(ids2),
                                   dimnames=list("1", ids2)))
    frame3 <- as.data.frame(matrix(data=1, nrow=1, ncol=length(ids3),
                                   dimnames=list("1", ids3)))
    frame4 <- as.data.frame(matrix(data=1, nrow=1, ncol=length(ids3),
                                   dimnames=list("1", rev(ids3))))
    res1 <- read.ids(rwl=frame1, stc=c(3, 2, 1))
    res2 <- read.ids(rwl=frame2, stc=c(3, 2, 1))
    res3 <- read.ids(rwl=frame3, stc=c(3, 2, 1))
    res4 <- read.ids(rwl=frame4, stc=c(3, 2, 1))
    ## Test
    test_that("read.ids works when tree and core are numbers", {
        expect_equal(row.names(res1), ids1)
        expect_equal(res1[["tree"]], tree)
        expect_equal(res1[["core"]], core)
    })
    test_that("read.ids does not change integer IDs", {
        expect_equal(row.names(res2), ids2)
        expect_equal(res2[["tree"]], tree[3:n.ids1])
        expect_equal(res2[["core"]], core[3:n.ids1])
    })
    test_that("read.ids converts alphabetic IDs to numbers", {
        expect_equal(row.names(res3), ids3)
        expect_equal(res3[["tree"]], tree)
        expect_equal(res3[["core"]], core)
    })
    test_that("ID mapping is invariant to order of columns", {
        expect_equal(res3, res4[seq(from=n.ids1, to=1), ])
    })
    ## TODO: Test autoread.ids()
}
test.read.ids()
test.rwi.stats <- function() {
    ## Setup
    v.1 <- 1 + runif(300)
    range.1 <- 51:400
    rnames.1 <- as.character(range.1)
    df.1 <- data.frame(col1 = c(v.1, rep.int(NA, 50)),
                       col2 = c(rep.int(NA, 25), v.1, rep.int(NA, 25)),
                       col3 = c(rep.int(NA, 50), v.1),
                       row.names = rnames.1)
    ## Test
    test_that("rwi.stats reports n", {
        expect_equal(rwi.stats(df.1, period="common")[["n"]], 3)
    })
    ## Needs more tests
}
test.rwi.stats()
test.sens1 <- function() {
    ## Setup
    SAMP.SIZE <- 1000
    ## Test
    test_that("sens1 of constant series is 0", {
        expect_equal(sens1(rep.int(42, SAMP.SIZE)), 0)
    })
    ## Needs more tests
}
test.sens1()
test.sens2 <- function() {
    ## Setup
    SAMP.SIZE <- 1000
    ## Test
    test_that("sens2 of constant series is 0", {
        expect_equal(sens2(rep.int(42, SAMP.SIZE)), 0)
    })
    ## Needs more tests
}
test.sens2()
test.tbrm <- function() {
    ## Setup
    SAMP.SIZE <- 1000
    half.32.52 <- c(rep.int(32, SAMP.SIZE / 2), rep.int(52, SAMP.SIZE / 2))
    outliers.in.42 <- c(rep.int(42, SAMP.SIZE / 2 + 1),
                        rep.int(-1e6, SAMP.SIZE / 4),
                        rep.int(1e6, SAMP.SIZE / 4))
    seq.odd <- seq_len(5)
    seq.even <- seq_len(6)
    ## Test
    test_that("tbrm handles empty input", {
        expect_equal(tbrm(NA), NaN)
        expect_equal(tbrm(numeric(0)), NaN)
    })
    test_that("tbrm handles constant input", {
        expect_equal(tbrm(rep.int(42, SAMP.SIZE)), 42)
    })
    test_that("two equally sized groups => all or nothing", {
        expect_equal(tbrm(half.32.52, C=1), 42) # C >= 1 (roughly)
        expect_equal(tbrm(half.32.52, C=0.5), NaN)
    })
    test_that("median abs deviation of 0 voids C", {
        expect_equal(tbrm(outliers.in.42, C=1e300), 42)
        expect_equal(tbrm(outliers.in.42, C=0), 42)
    })
    ## In the following, we see what happens when the median
    ## at first belongs and then does not belong to the set
    test_that("seq_len(x), odd or even x has an effect", {
        expect_equal(tbrm(seq.odd, C=0), mean(seq.odd))
        expect_equal(tbrm(seq.even, C=0), NaN)
    })
}
test.tbrm()

plot.rwl <- function(x, plot.type=c("seg","spag"),...){
  if (!inherits(x, "rwl")) {
    stop('use only with "rwl" objects')
  }
  switch(match.arg(plot.type),
         seg = seg.plot(x,...),
         spag = spag.plot(x,...))
}

series.rwl.plot <-
    function(rwl, series, series.yrs=as.numeric(names(series)),
             seg.length=100, bin.floor=100, n=NULL, prewhiten = TRUE,
             biweight=TRUE, floor.plus1 = FALSE) {

    ## Handle different types of 'series'
    tmp <- pick.rwl.series(rwl, series, series.yrs)
    rwl2 <- tmp[[1]]
    series2 <- tmp[[2]]
    series.yrs0 <- tmp[[3]][!is.na(series2)]

    ## run error checks
    qa.xdate(rwl2, seg.length, n, bin.floor)

    ## turn off warnings for this function
    ## The sig test for spearman's rho often produces warnings.
    w <- options(warn = -1)
    on.exit(options(w))

    seg.lag <- seg.length / 2

    mask <- !rowAlls(as.matrix(is.na(rwl2)))
    yrs0 <- as.numeric(row.names(rwl2))[mask]
    ## Normalize.
    tmp <- normalize.xdate(rwl2, series2, n, prewhiten, biweight)
    master <- tmp$master

    ## trim master so there are no NaN like dividing when
    ## only one series for instance.
    idx.good <- !is.nan(master)
    master <- master[idx.good]
    yrs <- as.numeric(names(master))

    series2 <- tmp$series
    series.yrs2 <- as.numeric(names(series2))
    ## trim series in case it was submitted stright from the rwl
    idx.good <- !is.na(series2)
    series.yrs2 <- series.yrs2[idx.good]
    series2 <- series2[idx.good]

    ## clip series to master dimensions
    series2 <- series2[series.yrs2 %in% yrs]
    series.yrs2 <- as.numeric(names(series2))
    ## clip master to series dimensions
    master <- master[yrs %in% series.yrs2]
    yrs <- as.numeric(names(master))

    if (is.null(bin.floor) || bin.floor == 0) {
        min.bin <- min(series.yrs2)
    } else if (floor.plus1) {
        min.bin <- ceiling((min(series.yrs2) - 1) / bin.floor) * bin.floor + 1
    } else {
        min.bin <- ceiling(min(series.yrs2) / bin.floor) * bin.floor
    }
    to <- max(series.yrs2) - seg.length - seg.lag + 1
    if (min.bin > to) {
        cat(gettextf("maximum year in (filtered) series: %d\n",
                     max(series.yrs2)))
        cat(gettextf("first bin begins: %d\n", min.bin))
        cat(gettext("cannot fit two segments (not enough years in the series)\n"))
        stop("shorten 'seg.length' or adjust 'bin.floor'")
    }
    bins <- seq(from=min.bin, to=to + seg.lag, by=seg.lag)
    bins <- cbind(bins, bins + (seg.length - 1), deparse.level=0)
    nbins <- nrow(bins)

    op <- par(no.readonly=TRUE)
    on.exit(par(op), add = TRUE)
    layout(matrix(c(1, 3, 2, 4), 2, 2), widths = c(1, 0.5),
           heights = c(1, 0.5))
    par(mar=c(4, 2, 2, 1) + 0.1, mgp=c(1.25, 0.25, 0), tcl=0.25)
    col.pal <- c("#E41A1C", "#377EB8", "#4DAF4A")
    ## plot 1
    plot(yrs, series2, type="n", ylim=c(0, max(series2, master, na.rm=TRUE)),
         ylab=gettext("RWI", domain="R-dplR"),
         xlab=gettext("Year", domain="R-dplR"), axes=FALSE)
    all.ticks <- c(bins[, 1], bins[c(nbins - 1, nbins), 2] + 1)
    abline(v=all.ticks, col="grey", lty="dotted")
    abline(h=1)
    odd.seq <- seq(from=1, to=nbins, by=2)
    even.seq <- seq(from=2, to=nbins, by=2)
    odd.ticks <- c(bins[odd.seq, 1], bins[odd.seq[length(odd.seq)], 2] + 1)
    even.ticks <- c(bins[even.seq, 1], bins[even.seq[length(even.seq)], 2] + 1)
    axis(1, at=odd.ticks)
    axis(3, at=even.ticks)
    axis(2)
    box()
    lines(yrs, series2, lwd=1.5, col=col.pal[1])
    lines(yrs, master, lwd=1.5, col=col.pal[2])
    legend(x = "bottomleft",
           legend = gettext(c("Series", "Master"),
           domain="R-dplR"),ncol = 2,
           col = c(col.pal[1], col.pal[2]), lty = "solid", lwd=1.5, bg="white")
    ## plot 2
    lm1 <- lm(master ~ series2)
    #tmp <- round(summary(lm1)$r.squared, 2)
    tmp <- round(cor(series2,master),2)
    plot(series2, master, type="p",
         ylab=gettext("Master", domain="R-dplR"),
         xlab=gettext("Series", domain="R-dplR"), pch=20,
         sub=bquote(r==.(tmp)))
    abline(coef = coef(lm1), lwd=2)

    ## plot 3
    
    # run corr.series.seg and stick the correlations in the boxes?
    tmp <- corr.series.seg(rwl, series, series.yrs = series.yrs,
                    seg.length = seg.length, bin.floor = bin.floor, n = n,
                    prewhiten = prewhiten, biweight = biweight, 
                    make.plot = FALSE, floor.plus1=floor.plus1)
    cors4boxes <- round(tmp[[1]],2)
    
    plot(yrs, series2, type="n", ylim=c(-1, 1), ylab="",
         xlab=gettext("Year", domain="R-dplR"),
         sub=gettextf("Segments: length=%d,lag=%d,bin.floor=%d",
         seg.length, seg.lag, bin.floor, domain="R-dplR"), axes=FALSE)
    abline(v=all.ticks, col="grey", lty="dotted")
    abline(h=0)
    axis(1, at=odd.ticks)
    axis(3, at=even.ticks)
    box()
    for (i in seq(1, nbins, by=2)) {
        
        xx <- bins[i, ]
        xmid <- mean(xx)
        xx <- c(xx, rev(xx))
        yy <- c(0, 0, 0.5, 0.5)
        polygon(xx, yy, col="grey90")
        text(x=xmid,y = 0.25,labels = cors4boxes[i],cex = 0.75)
    }
    for (i in seq(2, nbins, by=2)) {
        xx <- bins[i, ]
        xmid <- mean(xx)
        xx <- c(xx, rev(xx))
        yy <- c(0, 0, -0.5, -0.5)
        polygon(xx, yy, col="grey90")
        text(x=xmid,y = -0.25,labels = cors4boxes[i],cex = 0.75)
    }
    ## plot 4
    par(xpd = TRUE)
    plot(c(-1, 1), c(-2, 1), type="n", ylab="", xlab="", axes=FALSE)
    txt1 <- gettextf("Series:%d-%d", min(na.omit(series.yrs0)),
                     max(na.omit(series.yrs0)), domain="R-dplR")
    text(-1, 1, txt1, pos=4)
    txt2 <- gettextf("Master:%d-%d", min(yrs0, na.rm=TRUE),
                     max(yrs0, na.rm=TRUE), domain="R-dplR")
    text(-1, 0.5, txt2, pos=4)
    txt3 <- gettext("Detrended and Trimmed:", domain="R-dplR")
    text(-1, 0, txt3, pos=4)
    txt4 <- gettextf("%d-%d", min(yrs), max(yrs), domain="R-dplR")
    text(-1, -0.5, txt4, pos=4)
    txt5 <- gettext("Detrending Options:", domain="R-dplR")
    text(-1, -1, txt5, pos=4)
    if (is.null(n)) {
        txt6 <- paste0("Hanning=NULL,ar=", prewhiten)
    } else {
        txt6 <- paste0("Hanning=", n, ",ar=", prewhiten)
    }
    text(-1, -1.5, txt6, pos=4)

    list(series = series2, master = master)
}

plot.crs <- function(x,...){
  if (!inherits(x, "crs")) {
    stop('use only with "crs" objects putput from corr.rwl.seg()')
  }
  bins <- x$bins
  segs <- x$rwi
  pval <- x$p.val
  yrs <- as.numeric(rownames((x$rwi)))
  min.yr <- min(yrs)
  max.yr <- max(yrs)
  nseries <- ncol(segs)
  nbins <- nrow(bins)
  nyrs <- length(yrs)
  seq.series <- seq_len(nseries)
  
  
  seg.length <- x$seg.length
  seg.lag <- x$seg.lag
  pcrit <- x$pcrit
  label.cex <- x$label.cex
  
  
  extreme.year <- as.matrix(apply(segs, 2, yr.range, yr.vec = yrs))
  
  rsult <- sort.int(extreme.year[1, ], decreasing = FALSE, 
                    index.return = TRUE)
  neworder <- rsult$ix
  segs <- segs[, neworder, drop = FALSE]
  segs.mat <- t(extreme.year[, neworder])
  nsegs <- ncol(segs)
  op <- par(no.readonly = TRUE)
  on.exit(par(op), add = TRUE)
  col.pal <- c("#E41A1C", "#377EB8", "#4DAF4A")
  par(mar = c(4, 5, 4, 5) + 0.1, mgp = c(1.25, 0.25, 0), 
      tcl = 0.25)
  #dev.hold()
  #on.exit(dev.flush(), add = TRUE)
  plot(yrs, segs[, 1], type = "n", ylim = c(0.5, nsegs + 0.5), 
       axes = FALSE, ylab = "", xlab = gettext("Year"), 
       sub = gettextf("Segments: length=%d,lag=%d", seg.length, 
                      seg.lag, domain = "R-dplR"),...)
  iEven <- seq(from = 1, to = nseries, by = 2)
  rect(xleft = min.yr - 100, ybottom = iEven - 0.5, xright = max.yr + 
         100, ytop = iEven + 0.5, col = "grey90", border = NA)
  abline(v = c(bins[, 1], bins[c(nbins - 1, nbins), 2] + 
                 1), col = "grey", lty = "dotted")
  ax <- c(1, 3)
  for (odd.even in c(1, 2)) {
    this.seq <- seq(from = odd.even, to = nbins, by = 2)
    these.bins <- bins[this.seq, , drop = FALSE]
    com.segs <- matrix(NA, ncol = nseries, nrow = nyrs)
    flag.segs <- matrix(NA, ncol = nseries, nrow = nyrs)
    tmp <- pval[neworder, this.seq, drop = FALSE] > 
      pcrit
    for (i in seq.series) {
      for (j in seq_len(nrow(these.bins))) {
        mask <- yrs %in% seq(from = these.bins[j, 1], 
                             to = these.bins[j, 2])
        if (!is.na(tmp[i, j])) {
          com.segs[mask, i] <- 1
          if (tmp[i, j]) {
            flag.segs[mask, i] <- 1
          }
        }
      }
    }
    com.segs.mat <- t(apply(com.segs, 2, yr.range, yr.vec = yrs))
    guides.x.base <- c(these.bins[, 1], 
                       these.bins[length(this.seq), 2] + 1)
    axis(ax[odd.even], at = guides.x.base)
    if (odd.even == 1) {
      ytop <- seq.series
      ybottom <- ytop - 0.25
    }
    else {
      ybottom <- seq.series
      ytop <- ybottom + 0.25
    }
    rect(xleft = segs.mat[, 1], ybottom = ybottom, 
         xright = segs.mat[,2] + 1, 
         ytop = ytop, col = col.pal[3], border = NA)
    rect(xleft = com.segs.mat[, 1], ybottom = ybottom, 
         xright = com.segs.mat[, 2] + 1, ytop = ytop, 
         col = col.pal[2], border = NA)
    for (i in seq.series) {
      yb <- ybottom[i]
      yt <- ytop[i]
      flag.segs.mat <- yr.ranges(flag.segs[, i], yrs)
      if (nrow(flag.segs.mat) > 0) {
        rect(xleft = flag.segs.mat[, 1], ybottom = yb, 
             xright = flag.segs.mat[, 2] + 1, ytop = yt, 
             col = col.pal[1], border = NA)
      }
      guides.x <- guides.x.base[guides.x.base >= segs.mat[i,1]]
      guides.x <- guides.x[guides.x <= segs.mat[i,2]]
      if (length(guides.x) > 0) {
        segments(guides.x, yb, guides.x, yt, col = "white")
      }
    }
  }
  odd.seq <- seq(from = 1, to = nsegs, by = 2)
  even.seq <- seq(from = 2, to = nsegs, by = 2)
  cnames.segs <- colnames(segs)
  axis(2, at = odd.seq, labels = cnames.segs[odd.seq], 
       srt = 45, tick = FALSE, las = 2, cex.axis = label.cex)
  axis(4, at = even.seq, labels = cnames.segs[even.seq], 
       srt = 45, tick = FALSE, las = 2, cex.axis = label.cex)
  abline(h = seq.series, col = "white")
  box()
}

requireVersion <- function(package, ver) {
    requireNamespace(package, quietly = TRUE) &&
        packageVersion(package) >= ver
}

### Try to create directory named by tempdir() if it has gone missing
check.tempdir <- function() {
    td <- tempdir()
    if (!file.exists(td)) {
        dir.create(td, mode = "0700")
    }
}

### Checks that all arguments are TRUE or FALSE
check.flags <- function(...) {
    flag.bad <- vapply(list(...),
                       function(x) { !(identical(x, TRUE) ||
                                       identical(x, FALSE)) },
                       TRUE,
                       USE.NAMES = FALSE)
    if (any(flag.bad)) {
        offending <- vapply(match.call(expand.dots=TRUE)[c(FALSE, flag.bad)],
                            deparse, "")
        stop(gettextf("must be TRUE or FALSE: %s",
                      paste(sQuote(offending), collapse=", "),
                      domain="R-dplR"),
             domain = NA)
    }
}

### Function to check if x is equivalent to its integer
### representation. Note: Returns FALSE for values that fall outside
### the range of the integer type. The result has the same shape as x;
### at least vector and array x are supported.
is.int <- function(x) {
    suppressWarnings(y <- x == as.integer(x))
    y[is.na(y)] <- FALSE
    y
}

### Converts from "year and suffix" presentation to dplR internal
### years, where year 0 (e.g. as a row name) is actually year 1 BC
dplr.year <- function(year, suffix) {
    switch(toupper(suffix),
           AD = ifelse(year > 0, year, as.numeric(NA)),
           BC = ifelse(year > 0, 1-year, as.numeric(NA)),
           BP = ifelse(year > 0, 1950-year, as.numeric(NA)),
           as.numeric(NA))
}

### Prints the contents of a matrix row together with column labels.
### By default, doesn't print NA values.  If show.all.na is TRUE,
### reports all-NA rows as one NA.
row.print <- function(x, drop.na=TRUE, show.all.na=TRUE, collapse=", ") {
    if (drop.na) {
        not.na <- !is.na(x)
        if (any(not.na)) {
            paste(colnames(x)[not.na], x[not.na], sep=": ", collapse=collapse)
        } else if (show.all.na) {
            as.character(NA)
        }
    } else {
        paste(colnames(x), x, sep=": ", collapse=collapse)
    }
}

### Returns indices of rows in matrix X that match with pattern.
row.match <- function(X, pattern) {
    which(apply(X, 1,
                function(x) {
                    all(is.na(x) == is.na(pattern)) &&
                    all(x == pattern, na.rm = TRUE)
                }))
}

### Increasing sequence.
### The equivalent of the C loop 'for(i=from;i<=to;i++){}'
### can be achieved by writing 'for(i in inc(from,to)){}'.
### Note that for(i in from:to) fails to do the same if to < from.
inc <- function(from, to) {
    if (is.numeric(to) && is.numeric(from) && to >= from) {
        seq(from=from, to=to)
    } else {
        integer(length=0)
    }
}

### Decreasing sequence. See inc.
dec <- function(from, to) {
    if (is.numeric(to) && is.numeric(from) && to <= from) {
        seq(from=from, to=to)
    } else {
        integer(length=0)
    }
}

### AR function for chron, normalize1, normalize.xdate, ...
ar.func <- function(x, model = FALSE, ...) {
    y <- x
    idx.goody <- !is.na(y)
    ar1 <- ar(y[idx.goody], ...)
    y[idx.goody] <- ar1$resid+ar1$x.mean
    if (isTRUE(model)) {
        structure(y, model = ar1)
    } else {
        y
    }
}

### Range of years. Used in cms, rcs, rwl.stats, seg.plot, spag.plot, ...
yr.range <- function(x, yr.vec = as.numeric(names(x))) {
    na.flag <- is.na(x)
    if (all(na.flag)) {
        res <- rep(NA, 2)
        mode(res) <- mode(yr.vec)
        res
    } else {
        range(yr.vec[!na.flag])
    }
}

### Multiple ranges of years.
yr.ranges <- function(x, yr.vec = as.numeric(names(x))) {
    na.flag <- is.na(x)
    idx.good <- which(!na.flag)
    idx.bad <- which(na.flag)
    n <- length(x)
    res <- matrix(nrow=ceiling(n / 2), ncol=2)
    k <- 0
    while (length(idx.good) > 0) {
        first.good <- idx.good[1]
        idx.bad <- idx.bad[idx.bad > first.good]
        if (length(idx.bad) > 0) {
            first.bad <- idx.bad[1]
        } else {
            first.bad <- n + 1
        }
        idx.good <- idx.good[idx.good > first.bad]
        res[k <- k + 1, ] <- yr.vec[c(first.good, first.bad - 1)]
    }
    res[seq_len(k), , drop=FALSE]
}

### Used in cms, rcs, ...
sortByIndex <- function(x) {
    lowerBound <- which.min(is.na(x))
    c(x[lowerBound:length(x)], rep(NA, lowerBound - 1))
}

### Increment the given number (vector) x by one in the given base.
### Well, kind of: we count up to and including base (not base-1), and
### the smallest digit is one. Basically, we have a shift of one because
### of array indices starting from 1 instead of 0.  In case another
### digit is needed in the front, the result vector y grows.
count.base <- function(x, base) {
    n.x <- length(x)
    pos <- n.x
    y <- x
    y[pos] <- y[pos] + 1
    while (y[pos] == base + 1) {
        y[pos] <- 1
        if (pos == 1) {
            temp <- vector(mode="integer", length=n.x+1)
            temp[-1] <- y
            pos <- 2
            y <- temp
        }
        pos <- pos - 1
        y[pos] <- y[pos] + 1
    }
    y
}

### Compose a new name by attaching a suffix, which may partially
### replace the original name depending on the limit imposed on the
### length of names.
compose.name <- function(orig.name, alphabet, idx, limit) {
    idx.length <- length(idx)
    if (!is.null(limit) && idx.length > limit) {
        new.name <- ""
    } else {
        last.part <- paste(alphabet[idx], collapse="")
        if (is.null(limit)) {
            new.name <- paste0(orig.name, last.part)
        } else {
            new.name <- paste0(substr(orig.name, 1, limit - idx.length),
                               last.part)
        }
    }
    new.name
}

### Fix names so that they are unique and no longer than the given
### length.  A reasonable effort will be done in the search for a set of
### unique names, although some stones will be left unturned. The
### approach should be good enough for all but the most pathological
### cases. The output vector keeps the names of the input vector.
fix.names <- function(x, limit=NULL, mapping.fname="", mapping.append=FALSE,
                      basic.charset=TRUE) {
    fn <- mapping.fname
    if (!is.character(fn) || is.na(fn[1]) || Encoding(fn[1]) == "bytes") {
        fn <- ""
    } else {
        fn <- fn[1]
    }
    write.map <- FALSE
    n.x <- length(x)
    x.cut <- x
    rename.flag <- rep(FALSE, n.x)
    if (basic.charset) {
        bad.chars <- paste(c("[^",LETTERS,letters,0:9,"]"),collapse="")
        idx.bad <- grep(bad.chars, x.cut, perl=TRUE)
        if (length(idx.bad) > 0) {
            warning("characters outside a-z, A-Z, 0-9 present: renaming series")
            if (nzchar(fn)) {
                write.map <- TRUE
            }
            rename.flag[idx.bad] <- TRUE
            ## Remove inappropriate characters (replace with nothing)
            x.cut[idx.bad] <- gsub(bad.chars, "", x.cut[idx.bad],
                                   useBytes = !l10n_info()[["MBCS"]])
        }
    }
    if (!is.null(limit)) {
        over.limit <- nchar(x.cut) > limit
        if (any(over.limit)) {
            warning("some names are too long: renaming series")
            if (nzchar(fn)) {
                write.map <- TRUE
            }
            rename.flag[over.limit] <- TRUE
            x.cut[over.limit] <- substr(x.cut[over.limit], 1, limit)
        }
    }
    unique.cut <- unique(x.cut)
    n.unique <- length(unique.cut)
    ## Check if there are duplicate names after truncation and removal
    ## of inappropriate characters.  No duplicates => nothing to do
    ## beyond this point, except return the result.
    if (n.unique == n.x) {
        y <- x.cut
    } else {
        warning("duplicate names present: renaming series")
        if (nzchar(fn)) {
            write.map <- TRUE
        }

        y <- character(length=n.x)
        names(y) <- names(x)
        alphanumeric <- c(0:9, LETTERS, letters)
        n.an <- length(alphanumeric)
        ## First pass: Keep already unique names
        for (i in 1:n.unique) {
            idx.this <- which(x.cut %in% unique.cut[i])
            n.this <- length(idx.this)
            if (n.this == 1) {
                y[idx.this] <- x.cut[idx.this]
            }
        }

        if (!is.null(limit)) {
            x.cut <- substr(x.cut, 1, limit - 1)
        }
        x.cut[y != ""] <- NA
        unique.cut <- unique(x.cut) # may contain NA
        n.unique <- length(unique.cut)
        ## Second pass (exclude names that were set in the first pass):
        ## Make rest of the names unique
        for (i in 1:n.unique) {
            this.substr <- unique.cut[i]
            if (is.na(this.substr)) {# skip NA
                next
            }
            idx.this <- which(x.cut %in% this.substr)
            n.this <- length(idx.this)
            suffix.count <- 0
            for (j in 1:n.this){
                still.looking <- TRUE
                while (still.looking) {
                    suffix.count <- count.base(suffix.count, n.an)
                    proposed <-
                        compose.name(unique.cut[i],alphanumeric,suffix.count,limit)
                    if (!nzchar(proposed)) {
                        warning("could not remap a name: some series will be missing")
                        still.looking <- FALSE
                        ## F for Fail...
                        proposed <- paste0(unique.cut[i], "F")
                    } else if (!any(y %in% proposed)) {
                        still.looking <- FALSE
                    }
                }
                this.idx <- idx.this[j]
                y[this.idx] <- proposed
                rename.flag[this.idx] <- TRUE
            }
        }
    }
    if (write.map) {
        if (mapping.append && file.exists(fn)) {
            map.file <- file(fn, "a")
        } else {
            map.file <- file(fn, "w")
        }
        for (i in which(rename.flag)) {
            if (x[i] != y[i]) {
                cat(x[i], "\t", y[i], "\n", file=map.file, sep = "")
            }
        }
        close(map.file)
    }
    y
}

### Handle different types of 'series'.
###
### If series is a character or numeric vector of length 1, it is
### interpreted as a column index to rwl.  In this case, the
### corresponding column is also dropped from rwl.
###
### Returns list(rwl, series, series.yrs), where series is equipped
### with names indicating years.
###
### Intended to be used by ccf.series.rwl(), corr.series.seg(), ...
pick.rwl.series <- function(rwl, series, series.yrs) {
    if (length(series) == 1) {
        if (is.character(series)) {
            seriesIdx <- logical(ncol(rwl))
            seriesIdx[colnames(rwl) == series] <- TRUE
            nMatch <- sum(seriesIdx)
            if (nMatch == 0) {
                stop("'series' not found in 'rwl'")
            } else if (nMatch != 1) {
                stop("duplicate column names, multiple matches")
            }
            rwl2 <- rwl[, !seriesIdx, drop = FALSE]
            series2 <- rwl[, seriesIdx]
        } else if (is.numeric(series) && is.finite(series) &&
                   series >=1 && series < ncol(rwl) + 1) {
            rwl2 <- rwl[, -series, drop = FALSE]
            series2 <- rwl[, series]
        } else {
            stop("'series' of length 1 must be a column index to 'rwl'")
        }
        rNames <- rownames(rwl)
        names(series2) <- rNames
        series.yrs2 <- as.numeric(rNames)
    } else {
        rwl2 <- rwl
        series2 <- series
        names(series2) <- as.character(series.yrs)
        series.yrs2 <- series.yrs
    }
    list(rwl = rwl2, series = series2, series.yrs = series.yrs2)
}
# does the skeleton calculation
xskel.calc <- function(x,filt.weight=9,skel.thresh=3){
  x.dt <- hanning(x, filt.weight)
  n <- length(x)
  y <- rep(NA, n)
  ## calc rel growth
  n.diff <- n - 1
  idx <- 2:n.diff
  temp.diff <- diff(x)
  y[idx] <- rowMeans(cbind(temp.diff[-n.diff], -temp.diff[-1])) / x.dt[idx]
  y[y > 0] <- NA
  ## rescale from 0 to 10
  na.flag <- is.na(y)
  if(all(na.flag))
    y.range <- c(NA, NA)
  else
    y.range <- range(y[!na.flag])
  newrange <- c(10, 1)
  mult.scalar <-
    (newrange[2] - newrange[1]) / (y.range[2] - y.range[1])
  y <- newrange[1] + (y - y.range[1]) * mult.scalar
  y[y < skel.thresh] <- NA
  y <- ceiling(y)
  y
}

## Reorders vector x according to partial matching of its names to the
## names in Table.  This is designed to replicate argument matching in
## R function calls, which also means that it is possible to omit some
## or all names in x.  There is no equivalent of default values here,
## i.e. the lengths of the arguments must match.
vecMatched <- function(x, Table) {
    stopifnot(is.character(Table), !is.na(Table), nzchar(Table),
              length(x) == length(Table))
    xNames <- names(x)
    y <- as.vector(x)
    N <- length(Table)
    if (!is.null(xNames)) {
        matches <- pmatch(xNames, Table)
        isNA <- is.na(matches)
        nNA <- sum(isNA)
        if (nNA == 0) {
            y[matches] <- x
        } else {
            xNA <- xNames[isNA]
            flagBad <- is.na(xNA) | nzchar(xNA)
            if (any(flagBad)) {
                stop(gettextf("unknown element(s): %s",
                              paste(xNames[isNA][flagBad],collapse=", ")))
            }
            if (nNA < N) {
                notNA <- !isNA
                theMatch <- matches[notNA]
                y[theMatch] <- x[notNA]
                y[seq_len(N)[-theMatch]] <- x[isNA]
            }
        }
    }
    y
}

# Looks for internal NA in a series. Returns the position of internal NA via which
find.internal.na <- function(x) {
  x.na <- is.na(x)
  x.ok <- which(!x.na)
  n.ok <- length(x.ok)
  if (n.ok <= 1) {
    internal.na <- 0 # NA, NULL?
    return(internal.na)
  }
  
  first.ok <- x.ok[1]
  last.ok <- x.ok[n.ok]
  
  if (last.ok - first.ok + 1 > n.ok) {
    first.to.last <- first.ok:last.ok
    x.notok <- which(x.na)
    internal.na <- x.notok[x.notok %in% first.to.last]
  }
  else {
    internal.na <- 0 # NA, NULL?
  }
  internal.na    
}

capPOWT <- function (rwl, rescale = FALSE, return.power=FALSE) 
{
  # add a check for negative nums
  if(any(rwl <0, na.rm = TRUE)) {
    stop("'rwl' values cannot be negative")
  }
  
    if (!is.data.frame(rwl)) 
        stop("'rwl' must be a data.frame")
    if (!is.logical(rescale))
        stop("'rescale' must be either FALSE (the default) or TRUE")
  
    # used to set min numb to zeros.
    getprec <- function(rwl) {
        rwl.num <- as.numeric(as.matrix(rwl))
        rwl.num <- rwl.num[!is.na(rwl.num) & rwl.num != 0]
        if (length(rwl.num) == 0) {
            NA_real_
        }
        else {
            rwl.char <- format(rwl.num, scientific = FALSE)
            if (grepl(".", rwl.char[1], fixed = TRUE)) {
                maxdig <- nchar(sub("^[^.]*\\.", "", rwl.char[1]))
            }
            else {
                rm.trail.zeros <- sub("0*$", "", rwl.char)
                n.trail.zeros <- nchar(rwl.char) - nchar(rm.trail.zeros)
                maxdig <- -min(n.trail.zeros)
            }
            10^-maxdig
        }
    }
    fit.lm <- function(series) {
        n <- length(series)
        drop.1 <- series[-1]
        drop.n <- series[-n]
        runn.M <- (drop.1 + drop.n)/2
        runn.S <- abs(drop.1 - drop.n)
        runn.S[runn.S == 0] <- prec
        runn.M[runn.M == 0] <- prec
        mod <- lm.fit(cbind(1, log(runn.M)), log(runn.S))
        b <- mod[["coefficients"]][2]
        1 - b
    }
    transf <- function(x) {
        Xt <- x
        X.nna <- which(!is.na(x))
        X <- na.omit(x)
        p <- abs(fit.lm(X))
        
        X2 <- X^p
        Xt[X.nna] <- X2
        res <- list(Xt=Xt,p=p) 
        return(res)
    }
    prec <- getprec(rwl)
    
    xt <- lapply(rwl, FUN = function(x) transf(x)[['Xt']])
    p <- unlist(lapply(rwl, FUN = function(x) transf(x)[['p']]))
    if (rescale) {
        sds <- lapply(rwl, FUN = function(x) sd(x, na.rm = TRUE))
        means <- lapply(rwl, FUN = function(x) mean(x, na.rm = TRUE))
        rescale_ <- function(x, .sd, .mean) {
            scale(x) * .sd + .mean
        }
        xt <- mapply(rescale_, xt, sds, means)
    }
    res <- data.frame(xt, row.names = row.names(rwl), check.names = FALSE)
    class(res) <- c("rwl", "data.frame")
    
    if(return.power==TRUE){ #optional output of the power coefficient, default is FALSE
      names(p) <- names(rwl)
      res <- list(transformed.data=res,power=p)
    }
      
    return(res)
}

### Helper functions

### Computes the correlation coefficients between columns of x and y.
### Requires "limit" overlapping values in each pair.
cor.with.limit <- function(limit, x, y, method) {
  n.x <- ncol(x) # caller makes sure that n.x
  n.y <- ncol(y) # and n.y >= 1
  r.mat <- matrix(NA_real_, n.x, n.y)
  for (i in seq_len(n.x)) {
    this.x <- x[, i]
    good.x <- !is.na(this.x)
    for (j in seq_len(n.y)) {
      this.y <- y[, j]
      good.y <- !is.na(this.y)
      good.both <- which(good.x & good.y)
      n.good <- length(good.both)
      if (n.good >= limit && n.good > 0) {
        r.mat[i, j] <- cor(this.x[good.both], this.y[good.both],
                           method = method)
      }
    }
  }
  r.mat
}

### Computes the correlation coefficients between different columns of x.
cor.with.limit.upper <- function(limit, x, method) {
  n.x <- ncol(x) # caller makes sure that n.x >= 2
  r.vec <- rep.int(NA_real_, n.x * (n.x - 1) / 2)
  good.x <- !is.na(x)
  k <- 0
  for (i in seq_len(n.x - 1)) {
    good.i <- good.x[, i]
    for (j in (i + 1):n.x) {
      k <- k + 1
      good.both <- which(good.i & good.x[, j])
      if (length(good.both) >= limit) {
        r.vec[k] <- cor(x[good.both, i], x[good.both, j],
                        method = method)
      }
    }
  }
  r.vec
}

rwi.stats <- function(rwi, ids=NULL, period=c("max", "common"), 
                      method = c("spearman", "pearson", "kendall"),
                      ...) {
  args <- list(...)
  args[["rwi"]] <- rwi
  args[["ids"]] <- ids
  args[["period"]] <- period
  args[["method"]] <- method
  args[["running.window"]] <- FALSE
  do.call(rwi.stats.running, args)
}

### Main function, exported to user
rwi.stats.running <- function(rwi, ids=NULL, period=c("max", "common"),
                              method = c("spearman", "pearson", "kendall"),
                              prewhiten=FALSE,n=NULL,
                              running.window=TRUE,
                              window.length=min(50, nrow(rwi)),
                              window.overlap=floor(window.length / 2),
                              first.start=NULL,
                              min.corr.overlap=min(30, window.length),
                              round.decimals=3,
                              zero.is.missing=TRUE) {
  period2 <- match.arg(period)
  method2 <- match.arg(method)
  if (running.window) {
    if (window.length < 3) {
      stop("minimum 'window.length' is 3")
    }
    window.advance <- window.length - window.overlap
    if (window.advance < 1) {
      stop(gettextf("'window.overlap' is too large, max value is 'window.length'-1 (%d)",
                    window.length - 1))
    }
    if (window.length < min.corr.overlap) {
      stop("'window.length' is smaller than 'min.corr.overlap'")
    }
  }
  if(mean(unlist(rwi), na.rm=TRUE) <= 0) {
    stop("'rwi' has a non-positive grand mean (mean(unlist(rwi), na.rm=TRUE)), which will cause normalize1() to return incorrect rbar and eps values. This can happen when 'rwi' has been detrended using difference=TRUE, which produces first differences rather than ring widths. To fix this, pass a shifted version of 'rwi' directly into the function call, e.g. rwi.stats.running(rwi + 1, ...). Avoid overwriting your rwi object in your workspace as this can cause problems downstream.",
         call.=FALSE)
  }
  tmp <- normalize1(rwi, n, prewhiten)
  if(!all(tmp$idx.good)) {
    warning("after prewhitening, 'rwi' contains column(s) without at least four observations",
            call.=FALSE)
    cat(gettext("note that there is no error checking on column lengths if filtering is not performed\n",
                domain="R-dplR"))
  }
  rwi2 <- as.matrix(tmp$rwi.mat)
  n.cores <- ncol(rwi2)
  
  zero.flag <- rwi2 == 0
  if (any(zero.flag, na.rm=TRUE)) {
    if (!zero.is.missing) {
      warning("There are zeros in the data. Consider the option 'zero.is.missing'.")
    } else {
      rwi2[zero.flag] <- NA
    }
  }
  
  ## If 'ids' is NULL then assume one core per tree
  if (is.null(ids)) {
    ids3 <- data.frame(tree=seq_len(n.cores), core=rep.int(1, n.cores))
    rwi3 <- rwi2
  } else {
    ## Make error checks here
    if (!is.data.frame(ids) || !all(c("tree", "core") %in% names(ids))) {
      stop("'ids' must be a data.frame with columns 'tree' and 'core'")
    }
    if (!all(vapply(ids, is.numeric, TRUE))) {
      stop("'ids' must have numeric columns")
    }
    colnames.rwi <- colnames(rwi2)
    ## If all column names in 'rwi' are present in the set of row
    ## names in 'ids', arrange 'ids' to matching order
    rownames.ids <- row.names(ids)
    if (!is.null(rownames.ids) && all(colnames.rwi %in% rownames.ids)) {
      ids2 <- ids[colnames.rwi, c("tree", "core")]
    } else if (nrow(ids) == n.cores) {
      ids2 <- ids[c("tree", "core")]
    } else {
      stop("dimension problem: ", "'ncol(rwi)' != 'nrow(ids)'")
    }
    row.names(ids2) <- NULL
    unique.ids <- unique(ids2)
    n.unique <- nrow(unique.ids)
    if (n.unique < n.cores) {
      ## If more than one columns of 'rwi' share a tree/core ID pair,
      ## the columns are averaged and treated as one core
      ids3 <- unique.ids
      rwi3 <- matrix(data=as.numeric(NA), nrow=nrow(rwi2), ncol=n.unique,
                     dimnames=list(rownames(rwi2)))
      for (i in seq_len(n.unique)) {
        these.cols <- row.match(ids2, unique.ids[i, ])
        rwi3[, i] <-
          rowMeans(rwi2[, these.cols, drop=FALSE], na.rm=TRUE)
      }
      message("Series with matching tree/core IDs have been averaged")
    } else {
      ids3 <- ids2
      rwi3 <- rwi2
    }
  }
  rwiNotNA <- !is.na(rwi3)
  
  n.years <- nrow(rwi3)
  if (running.window && window.length > n.years) {
    stop("'window.length' is larger than the number of years in 'rwi'")
  }
  
  treeIds <- ids3$tree
  unique.trees <- unique(treeIds)
  n.trees <- length(unique.trees)
  if (n.trees < 2) {
    stop("at least 2 trees are needed")
  }
  cores.of.tree <- list()
  seq.tree <- seq_len(n.trees)
  for (i in seq.tree) {
    cores.of.tree[[i]] <- which(treeIds==unique.trees[i])
  }
  
  ## n.trees.by.year is recorded before setting rows with missing
  ## data to NA
  tree.any <- matrix(FALSE, n.years, n.trees)
  for (i in seq.tree) {
    tree.any[, i] <-
      rowAnys(rwiNotNA, cols = treeIds == unique.trees[i])
  }
  n.trees.by.year <- rowSums(tree.any)
  
  ## Easy way to force complete overlap of data
  if (period2 == "common") {
    bad.rows <- !rowAlls(rwiNotNA)
    rwi3[bad.rows, ] <- NA
    rwiNotNA[bad.rows, ] <- FALSE
    good.rows.flag <- !bad.rows
    period.common <- TRUE
  } else {
    good.rows.flag <- n.trees.by.year > 1
    period.common <- FALSE
  }
  good.rows <- which(good.rows.flag)
  
  if (length(good.rows) < min.corr.overlap) {
    stop("too few years with enough trees for correlation calculations")
  }
  
  if (running.window) {
    if (is.numeric(first.start)) {
      if (first.start < 1) {
        stop("'first.start' too small, must be >= 1")
      } else if (first.start > n.years - window.length + 1) {
        stop("'first.start' too large")
      }
      first.start2 <- first.start
    } else {
      ## Select locations of running windows by maximizing the
      ## number of data points (sum of number of series for each
      ## selected year), but don't count rows with less than two
      ## trees
      min.offset <-
        max(0, min(good.rows) - (window.length - min.corr.overlap) - 1)
      max.offset <-
        min(min.offset + window.advance - 1, n.years - window.length)
      offsets <- min.offset:max.offset
      n.offsets <- length(offsets)
      n.data <- rep.int(NA_real_, n.offsets)
      for (i in seq_len(n.offsets)) {
        offset <- offsets[i]
        n.windows.minusone <-
          (n.years - offset - window.length) %/% window.advance
        max.idx <-
          offset + window.length + n.windows.minusone * window.advance
        rowIdx <- seq(1 + offset, max.idx)
        n.data[i] <- sum(rwiNotNA[rowIdx[good.rows.flag[rowIdx]], ])
      }
      ## In case of a tie, choose large offset.
      ## In practice, this prefers recent years.
      first.start2 <-
        offsets[n.offsets - which.max(rev(n.data)) + 1] + 1
    }
    window.start <- seq(from = first.start2,
                        to = n.years - window.length + 1,
                        by = window.advance)
    window.length2 <- window.length
  } else {
    window.start <- 1
    window.length2 <- n.years
  }
  
  all.years <- as.numeric(rownames(rwi3))
  
  loop.body <- function(s.idx) {
    rbar.tot <- NA_real_
    rbar.wt <- NA_real_
    rbar.bt <- NA_real_
    
    ## Location of window
    start.year <- all.years[s.idx]
    e.idx <- s.idx + window.length2 - 1
    end.year <- all.years[e.idx]
    mid.year <- floor((start.year + end.year) / 2)
    year.idx <- s.idx:e.idx
    
    ## See p 138 in C&K
    ## Sum of all correlations among different cores (between trees)
    rsum.bt <- 0
    n.bt <- 0
    good.flag <- rep.int(FALSE, n.trees)
    for (i in seq_len(n.trees - 1)) {
      i.data <- rwi3[year.idx, cores.of.tree[[i]], drop=FALSE]
      for (j in (i + 1):n.trees) {
        j.data <- rwi3[year.idx, cores.of.tree[[j]], drop=FALSE]
        bt.r.mat <- cor.with.limit(min.corr.overlap, i.data, j.data,
                                   method=method2)
        bt.r.mat <- bt.r.mat[!is.na(bt.r.mat)]
        n.bt.temp <- length(bt.r.mat)
        if (n.bt.temp > 0) {
          rsum.bt <- rsum.bt + sum(bt.r.mat)
          n.bt <- n.bt + n.bt.temp
          good.flag[c(i, j)] <- TRUE
        }
      }
    }
    
    ## Sum of all correlations among different cores (within trees)
    good.trees <- which(good.flag)
    rsum.wt <- 0
    n.wt <- 0
    n.cores.tree <- rep.int(NA_real_, n.trees)
    for (i in good.trees) {
      these.cores <- cores.of.tree[[i]]
      if (length(these.cores)==1) { # make simple case fast
        n.cores.tree[i] <- 1
      } else {
        these.data <- rwi3[year.idx, these.cores, drop=FALSE]
        wt.r.vec <- cor.with.limit.upper(min.corr.overlap, these.data,
                                         method=method2)
        wt.r.vec <- wt.r.vec[!is.na(wt.r.vec)]
        n.wt.temp <- length(wt.r.vec)
        if (n.wt.temp > 0) {
          rsum.wt <- rsum.wt + sum(wt.r.vec)
          n.wt <- n.wt + n.wt.temp
          ## Solving c (> 0) in the formula n = 0.5 * c * (c-1)
          ## leads to c = 0.5 + sqrt(0.25+2*n)
          n.cores.tree[i] <- 0.5 + sqrt(0.25 + 2 * n.wt.temp)
        } else {
          n.cores.tree[i] <- 1
        }
      }
    }
    
    ## Mean correlations
    n.tot <- n.wt + n.bt
    if (n.tot > 0) {
      rbar.tot <- (rsum.wt + rsum.bt) / n.tot
    }
    if (n.wt > 0) {
      rbar.wt <- rsum.wt / n.wt
    }
    if (n.bt > 0) {
      rbar.bt <- rsum.bt / n.bt
    }
    
    coresPresent <- which(colAnys(rwiNotNA, rows = year.idx))
    treesPresent <- unique(treeIds[coresPresent])
    nCores <- length(coresPresent)
    nTrees <- length(treesPresent)
    # if (period.common) {
    #     ## If period is "common", we are only looking at the rows
    #     ## with no missing values (if any, so all or nothing).
    #     n <- nTrees
    # } else {
    #     ## Number of trees averaged over the years in the window.
    #     ## We keep this number separate of the correlation
    #     ## estimates, i.e. the data from some tree / year may
    #     ## contribute to n without taking part in the correlation
    #     ## estimates.
    #     n <- mean(n.trees.by.year[year.idx])
    # }
    
    # AGB May 2022
    # Setting n to be be nTrees for the EPS calc based on a comment from
    # S Klesse via github https://github.com/AndyBunn/dplR/issues/11
    # After consulting with him and Mikko we decided to revert back to this
    # value of n. Hence the code chunk commented out above.
    n<-length(good.trees)
    ## Expressed population signal
    if (n.wt == 0) {
      if (n.bt > 0) {
        c.eff <- 1
      } else {
        c.eff <- 0
      }
      rbar.eff <- rbar.bt
    } else {
      nCoresTree <- na.omit(n.cores.tree)
      uniqueNC <- unique(nCoresTree)
      ## The branches are equivalent but optimized for numerical
      ## precision in each situation
      if (length(uniqueNC) == 1) {
        c.eff <- uniqueNC
        rbar.eff <- rbar.bt / (rbar.wt + (1 - rbar.wt) / c.eff)
      } else {
        c.eff.rproc <- mean(1 / nCoresTree)
        c.eff <- 1 / c.eff.rproc # bookkeeping only
        rbar.eff <- rbar.bt / (rbar.wt + (1 - rbar.wt) * c.eff.rproc)
      }
    }
    ## EPS is on page 146 of C&K.
    ## In our interpretation of EPS, we use the average number of trees.
    eps <- n * rbar.eff / ((n - 1) * rbar.eff + 1)
    ## SNR is on page 109 of Cook and Pederson (2011).
    ## See help file for ref.
    snr <- n * rbar.eff / (1-rbar.eff)
    
    if (running.window) {
      out <-  c(start.year = start.year,
                mid.year = mid.year, end.year = end.year)
    } else {
      out <- numeric(0)
    }
    c(out,
      n.cores = nCores, n.trees = nTrees, n = n,
      n.tot = n.tot, n.wt = n.wt, n.bt = n.bt, rbar.tot = rbar.tot,
      rbar.wt = rbar.wt, rbar.bt = rbar.bt, c.eff = c.eff,
      rbar.eff = rbar.eff, eps = eps, snr = snr)
  }
  
  ## Iterate over all windows
  if (running.window &&
      !inherits(try(suppressWarnings(req.fe <-
                                     requireNamespace("foreach",
                                                      quietly=TRUE)),
                    silent = TRUE),
                "try-error") && req.fe) {
    
    exportFun <- c("<-", "+", "-", "floor", ":", "rep.int", "for",
                   "seq_len", "[", "[[", "cor.with.limit", "!",
                   "is.na", "length", "if", ">", "sum", "c",
                   "[<-", "which", "==", "cor.with.limit.upper",
                   "sqrt", "*", "/", "(", "{", "mean")
    
    compos.stats <-
      foreach::"%dopar%"(foreach::foreach(s.idx=window.start,
                                          .combine="rbind",
                                          .export=exportFun),
                         loop.body(s.idx))
  } else {
    compos.stats <- NULL
    for (s.idx in window.start) {
      compos.stats <- rbind(compos.stats, loop.body(s.idx))
    }
  }
  
  rownames(compos.stats) <- NULL
  if (is.numeric(round.decimals) && length(round.decimals) > 0 &&
      is.finite(round.decimals[1]) && round.decimals[1] >= 0) {
    data.frame(round(compos.stats, round.decimals[1]))
  } else {
    data.frame(compos.stats)
  }
}

`read.crn` <- function(fname, header=NULL, encoding = getOption("encoding"),
                       long = TRUE)
{
    ## Open the data file for reading
    con <- file(fname, encoding = encoding)
    on.exit(close(con))
    long2 <- isTRUE(long)
    if(is.null(header)){
        ## Try to determine if the file has a header. This is failable.
        ## Find out if an ITRDB header (3 lines) in file
        hdr1 <- readLines(con, n=1)
        if(length(hdr1) == 0)
            stop("file is empty")
        if(nchar(hdr1) < 10)
            stop("first line in the crn file ends before col 10")
        yrcheck <- suppressWarnings(as.numeric(substr(hdr1, 7, 10)))
        if(is.null(yrcheck) || length(yrcheck)!=1 || is.na(yrcheck) |
           yrcheck < -1e04 || yrcheck > 1e04) {
            cat(gettext("There appears to be a header in the crn file\n",
                        domain="R-dplR"))
            is.head <- TRUE
        }
        else {
            cat(gettext("There does not appear to be a header in the crn file\n",
                        domain="R-dplR"))
            is.head <- FALSE # No header lines
        }
    } else if(!is.logical(header) || length(header) != 1 || is.na(header)){
        stop("'header' must be NULL, TRUE or FALSE")
    } else{
        is.head <- header
    }
    if(is.head){
        ## Read 4th line - should be first data line
        dat1 <- readLines(con, n=4)
        if(length(dat1) < 4)
            stop("file has under 4 lines")
        dat1 <- dat1[4]
    } else{
        dat1 <- readLines(con, n=1)
        if(length(dat1) == 0)
            stop("file is empty")
    }
    if(nchar(dat1) < 10)
        stop("first data line ends before col 10")
    decade_pos <- 7L
    yrcheck <- as.numeric(substr(dat1, decade_pos, 10L))
    if(is.null(yrcheck) || length(yrcheck)!=1 || is.na(yrcheck) ||
       yrcheck < -1e04 || yrcheck > 1e04)
        stop(gettextf("cols %d-%d of first data line not a year",
                      decade_pos, 10L, domain="R-dplR"), domain = NA)
    if (long2) {
        year_now <- 1900 + as.POSIXlt(Sys.Date())$year
        if (yrcheck > year_now) {
            tmp_pos <-
                regexpr(" *-[[:digit:]]+$", substr(dat1, 2, 10))[[1L]] + 1L
            if (tmp_pos > 0L) {
                decade_pos <- tmp_pos
                message(gettextf("Using cols %d-%d for decade field",
                                 decade_pos, 10L, domain = "R-dplR"),
                        domain = NA)
            } else {
                warning(gettextf("year %d is in the future",
                                 yrcheck, domain = "R-dplR"),
                        domain = NA)
            }
        }
    }
    decade_fix <- decade_pos - 7L
    ## Look at last line to determine if Chronology Statistics are present
    ## if nchar <=63 then there is a stats line
    nlines <- length(readLines(con, n=-1))
    ## Read file
    skip.lines <- if (is.head) 3 else 0
    ## Do nothing. read.fwf closes (and destroys ?!?) the file connection
    on.exit()
    ## Get chron stats if needed
    chron.stats <- read.fwf(con, c(6 + decade_fix, 4 - decade_fix,
                                   6, 6, 6, 7, 9, 9, 10),
                            skip=nlines-1, strip.white=TRUE)
    ## If columns 3 in chron.stats is an integer then there is no
    ## statistics line
    if(is.numeric(chron.stats[[3]]) &&
       !is.int(as.numeric(chron.stats[[3]]))){
        names(chron.stats) <-
            c("SiteID", "nYears", "AC[1]", "StdDev", "MeanSens",
              "MeanRWI", "IndicesSum", "IndicesSS", "MaxSeries")
        cat(gettext("Embedded chronology statistics\n", domain="R-dplR"))
        print(chron.stats)
        n_dat <- nlines - skip.lines - 1
    } else {
        n_dat <- nlines - skip.lines
    }
    while (decade_fix >= -5L) {
        ## We need to create a new connection object.
        con <- file(fname, encoding = encoding)
        ## Really read file
        dat <- read.fwf(con, c(6 + decade_fix, 4 - decade_fix,
                               rep(c(4, 3), 10)),
                        skip = skip.lines, n = n_dat,
                        colClasses = c("character", rep("integer", 21)),
                        strip.white = TRUE)
        ## Remove any blank lines at the end of the file, for instance
        dat <- dat[!is.na(dat[[2]]), , drop=FALSE] # requires non-NA year

        series <- dat[[1]]
        series.ids <- unique(series)
        decade.yr <- dat[[2]]
        nseries <- length(series.ids)
        if (!long2 || nseries == 1L) {
            break
        } else {
            sign_table <- table(sign(diff(decade.yr)))
            if (length(sign_table) == 1L ||
                sum(sign_table) - max(sign_table) <= nseries - 1L) {
                break
            }
            decade_fix <- decade_fix - 1L
        }
    }
    decade_pos2 <- decade_fix + 7L
    if (decade_pos2 != decade_pos) {
        message(gettextf("Using cols %d-%d for decade field",
                         decade_pos2, 10L, domain = "R-dplR"),
                domain = NA)
    }
    
    cat(sprintf(ngettext(nseries,
                         "There is %d series\n",
                         "There are %d series\n",
                         domain="R-dplR"),
                nseries))
    series.index <- match(series, series.ids)
    min.year <- (min(decade.yr) %/% 10) * 10
    max.year <- ((max(decade.yr)+10) %/% 10) * 10
    span <- max.year - min.year + 1
    ncol.crn.mat <- nseries + 1
    crn.mat <- matrix(NA_real_, ncol=ncol.crn.mat, nrow=span)
    colnames(crn.mat) <- c(as.character(series.ids), "samp.depth")
    rownames(crn.mat) <- min.year:max.year
    ## RWI
    x <- as.matrix(dat[seq(from=3, to=21, by=2)])
    ## All sample depths
    y <- as.matrix(dat[seq(from=4, to=22, by=2)])
    for(i in seq_len(nseries)){
        idx <- which(series.index == i)
        for(j in idx) {
            yr <- (decade.yr[j] %/% 10) * 10
            row.seq <- seq(from = yr - min.year + 1, by = 1, length.out = 10)
            crn.mat[row.seq, i] <- x[j, ]
            if(i == 1) {
                crn.mat[row.seq, ncol.crn.mat] <- y[j, ]
            }
        }
    }
    ## Clean up NAs
    crn.mat[which(crn.mat[, -ncol.crn.mat] == 9990)] <- NA # column-major order
    crn.mat <-
        crn.mat[!rowAlls(is.na(crn.mat[, -ncol.crn.mat, drop=FALSE])),
                ,
                drop=FALSE]
    ## If samp depth is all 1 then dump it
    sd.one <- all(crn.mat[, ncol.crn.mat] == 1)
    if(sd.one) {
        save.names <- colnames(crn.mat)[-ncol.crn.mat]
        crn.mat <- crn.mat[, -ncol.crn.mat, drop=FALSE]
        crn.mat <- crn.mat / 1000
        crn.df <- as.data.frame(crn.mat)
        names(crn.df) <- save.names
        cat(gettext("All embedded sample depths are one...Dumping from matrix\n",
                    domain="R-dplR"))
    }
    else {
        seq.series <- seq_len(nseries)
        crn.mat[, seq.series] <- crn.mat[, seq.series] / 1000
        crn.df <- as.data.frame(crn.mat)
    }
    class(crn.df) <- c("crn", "data.frame")
    crn.df
}

!sbr spline: fits cubic spline curve to time series
!
!  derived from imsl library routines by edward r cook,
!  lamont-doherty geological observatory, and modified by
!  richard l holmes, tree-ring laboratory, university of arizona.
!
!  modified by r l holmes 10 jun 1988
!  modified by e r cook to full double precision 4 apr 1991
!
!*******************************************************************
!
! calculates a cubic spline function to model a time series.
! 'stiffness' of spline is specified by parameter 'stiffness'
!
! parameters:
!
! y: time series array to be modeled
! n: number of values in time series
! stiffness: length (stiffness) of spline to be used to model series
! pct: freq cutoff
! res: array of cubic spline function values computed
!
!*******************************************************************
!
! AGB Nov 2021 
! Modifying Ed Cook's spline subroutine to be a little less ugly
! This included removing all the implicit declarations and setting
! implicit none. Also removing archaic calls to dabs, dcos, dsqrt.
! Changed to follow f95 in terms of I/O variable decalration and
! c binding as well to play nicely with C and .Call
! Also aded back in pct as an arg to caps_f which mirrors the 
! ffcsaps function in dplR. This way they can be interchangable
!
!*******************************************************************

      subroutine caps_f(y,n,stiffness,pct,res) bind(C, name = "caps_f_")
      use, intrinsic :: iso_c_binding
      implicit none
      real(kind = c_double), intent(in)  :: y !input
      integer(kind = c_int), intent(in)  :: n,stiffness !input
      real(kind = c_double), intent(in)  :: pct !input
      real(kind = c_double), intent(out) :: res  !output
      integer, parameter :: dp = kind(1d0)
      real(dp) v,arg,rn,d1,d2,a,c1,c2,p,pi,sum,pp
      integer i,j,k,l,m,imncp1,i1,i2,iw,k1,kl,jm1,n1,nm2,nc,nc1,ncp1
      dimension a(9001,4),c1(4),c2(3),y(n),res(n)
      data c1 /1.d0,-4.d0,6.d0,-2.d0/
      data c2 /0.d0,.33333333333333d0,1.33333333333333d0/
      data pi / 3.1415926535897935d0 /
      if(n .lt. 4)then
        res(1)=-9998.
        return
      endif
      nm2=n-2
      v=dble(stiffness)
      arg=(2.d0*pi)/v
!      p=(6.d0*(cos(arg)-1.d0)**2.d0)/(cos(arg)+2.d0)
      pp=sngl(p) ! this doesn't do anything, right?
      p=((1.d0/(1.d0-pct)-1.d0)*6.d0*(cos(pi*2.d0/v)-1.d0)**2)/(cos(pi*2.d0/v)+2.d0)
!cc      p2=(6.d0*(cos(arg)-1.d0)**2.d0)/((2.d0**0.5d0-1.d0)*
!cc     *		(sqrt(arg)+2.d0))
      do i=1,nm2
         do j=1,3
            a(i,j)=c1(j)+p*c2(j)
            a(i,4)=dble(y(i))+c1(4)*dble(y(i+1))+dble(y(i+2))
         end do
      end do 
      a(1,1)=c2(1)
      a(1,2)=c2(1)
      a(2,1)=c2(1)
      nc=2
! begin ludapb
      rn=dble(1.d0/(dble(nm2)*16.d0))
      d1=1.d0
      d2=0.d0
      ncp1=nc+1
      if(nc .ne. 0)then
! initialize zero elements
        do i=1,nc
           do j=i,nc
              k=ncp1-j
              a(i,k)=0.d0
           end do
        end do 
! 'i' is row index of element being computed
! 'j' is column index of element being computed
! 'l' is row index of previously computed vector
!       being used to compute inner product
! 'm' is column index
      endif
      do i=1,nm2
         imncp1=i-ncp1
         i1=max0(1,1-imncp1)
         do j=i1,ncp1
            l=imncp1+j
            i2=ncp1-j
            sum=a(i,j)
            jm1=j-1
            if(jm1 .gt. 0)then
              do k=1,jm1
                 m=i2+k
                 sum=sum-a(i,k)*a(l,m)
              end do 
            endif
            if(j .eq. ncp1)then
              if(a(i,j)+sum*rn .le. a(i,j))then
                res(1)=-9999.
                return
              endif
              a(i,j)=1.d0/sqrt(sum)
! update determinant
              d1=d1*sum
   35         if(abs(d1) .gt. 1.d0)then
                d1=d1*.0625d0
                d2=d2+4.d0
                goto 35
              endif
   47         if(abs(d1) .le. .0625d0)then
                d1=d1*16.d0
                d2=d2-4.d0
                goto 47
                else
                goto 60
              endif
            endif
            a(i,j)=sum*a(l,ncp1)
   60       continue
         end do
      end do 
! end ludapb / begin luelpb
! solution ly = b
      nc1=nc+1
      iw=0
      l=0
      do i=1,nm2
        sum=a(i,4)
        if(nc .gt. 0)then
          if(iw .ne. 0)then
            l=l+1
            if(l .gt. nc) l=nc
              k=nc1-l
              kl=i-l
              do j=k,nc
                sum=sum-a(kl,4)*a(i,j)
                kl=kl+1
              end do 
            else
            if(sum .ne. 0.d0)iw=1
          endif
        endif
        a(i,4)=sum*a(i,nc1)
      end do
! solution ux = y
      a(nm2,4)=a(nm2,4)*a(nm2,nc1)
      n1=nm2+1
      do i=2,nm2
        k=n1-i
         sum=a(k,4)
         if(nc .gt. 0)then
           kl=k+1
           k1=min0(nm2,k+nc)
           l=1
           do j=kl,k1
             sum=sum-a(j,4)*a(j,nc1-l)
             l=l+1
           end do 
         endif
        a(k,4)=sum*a(k,nc1)
      end do 
! end luelpb
      do i=3,nm2
        res(i)=a(i-2,4)+c1(4)*a(i-1,4)+a(i,4)
      end do
      res(1)=a(1,4)
      res(2)=c1(4)*a(1,4)+a(2,4)
      res(n-1)=a(nm2-1,4)+c1(4)*a(nm2,4)
      res(n)=a(nm2,4)
      do i=1,n
        res(i)=y(i)-res(i)
      end do 
      return
      end subroutine caps_f

### Function for checking TRiDaS vocabulary.
tridas.vocabulary <-
    function(category = c("dating type", "measuring method", "shape",
             "location type", "variable", "unit", "remark",
             "dating suffix", "presence / absence",
             "complex presence / absence", "certainty"),
             idx = NA, term = NA, match.exact = FALSE){
        res <-
            switch(match.arg(category),
                   "dating type" = c("absolute", "dated with uncertainty",
                   "relative", "radiocarbon"),
                   "measuring method" = c("measuring platform",
                   "hand lens and graticule", "onscreen measuring",
                   "visual estimate"),
                   "shape" = c("whole section", "half section",
                   "third section", "quarter section",
                   "wedge where radius is smaller than circumference",
                   "wedge where radius equals the circumference",
                   "wedge where radius is bigger than the circumference",
                   "beam straightened on one side",
                   "squared beam from whole section",
                   "squared beam from half section",
                   "squared beam from quarter section",
                   "plank cut on one side", "radial plank through pith",
                   "radial plank up to pith",
                   paste("tangential plank not including pith with",
                         "breadth larger than a quarter section", sep=" "),
                   paste("plank not including pith with",
                         "breadth smaller than a quarter section", sep= " "),
                   "small part of section", "part of undetermined section",
                   "unknown", "other"),
                   "location type" = c("growth location",
                   "location of use (static)", "location of use (mobile)",
                   "current location", "manufacture location",
                   "find location"),
                   "variable" = c("ring width", "earlywood width",
                   "latewood width", "ring density", "earlywood density",
                   "latewood density", "maximum density", "latewood percent",
                   "vessel size"),
                   "unit" = c("micrometres", "1/100th millimetres",
                   "1/50th millimetres", "1/20th millimetres",
                   "1/10th millimetres", "millimetres", "centimetres",
                   "metres"),
                   "remark" = c("fire damage", "frost damage", "crack",
                   "false ring(s)", "compression wood", "tension wood",
                   "traumatic ducts", "unspecified injury", "single pinned",
                   "double pinned", "triple pinned", "missing ring",
                   "radius shift up", "radius shift down", "moon ring(s)",
                   "diffuse latewood", "density fluctuation",
                   "wide late wood", "wide early wood"),
                   "dating suffix" = c("AD", "BC", "BP"),
                   ## (complex) presence / absence options are in
                   ## order of "increasing information"
                   "presence / absence" = c("unknown", "absent", "present"),
                   "complex presence / absence" = c("unknown",
                   "not applicable", "absent", "incomplete", "complete"),
                   "certainty" = c("unknown", "exact", "approximately",
                   "after", "before")
                   )
        n.term <- length(term)
        if(!is.na(idx[1]))
            res[idx]
        else if(n.term > 1 || !is.na(term)){
            if(match.exact)
                any(res == term)
            else{
                res2 <- character(n.term)
                for(k in seq_len(n.term))
                    res2[k] <- match.arg(term[k], res)
                res2
            }
        }
        else
            res
    }

treeMean <- function(rwl, ids, na.rm=FALSE) {
    rwl2 <- as.matrix(rwl)
    if (!is.data.frame(ids) || !("tree" %in% names(ids))) {
        stop("'ids' must be a data.frame with column 'tree'")
    }
    colnames.rwl <- colnames(rwl2)
    trees <- as.matrix(ids["tree"])
    rownames.ids <- rownames(trees)
    ## If all column names in 'rwl' are present in the set of row
    ## names in 'ids', arrange 'ids' to matching order
    if (!is.null(rownames.ids) && !is.null(colnames.rwl) &&
        anyDuplicated(colnames.rwl) == 0 &&
        all(colnames.rwl %in% rownames.ids)) {
        trees <- trees[colnames.rwl, ]
    } else if (length(trees) == ncol(rwl2)) {
        trees <- as.vector(trees)
    } else {
        stop("dimension problem: ", "'ncol(rwl)' != 'nrow(ids)'")
    }
    uTrees <- unique(trees)
    if (any(is.na(uTrees))) {
        stop("missing tree IDs are not allowed")
    }
    matches <- match(trees, uTrees)
    res <- matrix(NA_real_, nrow=nrow(rwl2), ncol=length(uTrees))
    for (i in seq_along(uTrees)) {
      res[,i] <- rowMeans(rwl2[, matches == i, drop=FALSE], na.rm=na.rm)
    }
    res[is.nan(res)] <- NA_real_
    res <- as.data.frame(res, row.names = rownames(rwl2))
    names(res) <- uTrees
    class(res) <- c("rwl", "data.frame")
    res
}

### Exportable function
`write.tucson` <-
    function(rwl.df, fname, header=NULL, append=FALSE, prec=0.01,
             mapping.fname="", mapping.append=FALSE, long.names=FALSE,
             ...)
{
    line.term <- "\x0D\x0A" # CR+LF, ASCII carriage return and line feed
    if (!is.data.frame(rwl.df)) {
        stop("'rwl.df' must be a data.frame")
    }
    if (!is.numeric(prec) || length(prec) != 1 || is.na(prec) ||
        !(prec == 0.01 || prec == 0.001)) {
        stop("'prec' must equal 0.01 or 0.001")
    }
    header2 <- header
    if (append) {
        if (!file.exists(fname)) {
            stop(gettextf("file %s does not exist, cannot append", fname))
        }
        if (length(header2) > 0) {
            stop("bad idea to append with 'header'")
        }
    }
    if (length(header2) > 0) {
        if (!is.list(header2)) {
            stop("'header' must be a list")
        }
        header.names <-
            c("site.id", "site.name", "spp.code", "state.country",
              "spp", "elev", "lat", "long", "first.yr", "last.yr",
              "lead.invs", "comp.date")
        if (!all(header.names %in% names(header2))) {
            stop("'header' must be a list with the following names: ",
                 paste(dQuote(header.names), collapse = ", "))
        }
        ## Record #1: 1-6 Site ID, 10-61 Site Name, 62-65 Species
        ## Code, optional ID#s
        ## Record #2: 1-6 Site ID, 10-22 State/Country, 23-40 Species,
        ## 41-45 Elevation, 48-57 Lat-Long, 68-76 1st & last Year
        ## Note: lat-lons are in degrees and minutes, ddmm or dddmm
        ## Record #3: 1-6 Site ID, 10-72 Lead Investigator, 73-80
        ## comp. date
        header2 <- lapply(header2, as.character)
        site.id <- header2$site.id[1]
        site.name <- header2$site.name[1]
        spp.code <- header2$spp.code[1]
        state.country <- header2$state.country[1]
        spp <- header2$spp[1]
        elev <- header2$elev[1]
        lat <- header2$lat[1]
        long <- header2$long[1]
        lead.invs <- header2$lead.invs[1]
        comp.date <- header2$comp.date[1]
        lat.long <- if (isTRUE(nchar(long) > 5)) {
            paste0(lat, long)
        } else {
            paste(lat, long, sep=" ")
        }
        yrs <- paste(header2$first.yr[1], header2$last.yr[1], sep=" ")

        field.name <-
            c("site.id", "site.name", "spp.code", "state.country", "spp",
              "elev", "lat.long", "yrs", "lead.invs", "comp.date")
        field.width <- c(6, 52, 4, 13, 18, 5, 10, 9, 63, 8)
        for (i in seq_along(field.name)) {
            this.name <- field.name[i]
            this.width <- field.width[i]
            this.var <- get(this.name)
            this.nchar <- nchar(this.var)
            if (this.nchar > this.width) {
                assign(this.name, substr(this.var, 1, this.width))
            } else if (this.nchar < this.width) {
                assign(this.name, encodeString(this.var, width = this.width))
            }
        }

        hdr1 <- paste0(site.id, "   ", site.name, spp.code)
        hdr2 <- paste0(site.id, "   ", state.country, spp, " ", elev, "  ",
                       lat.long, "          ", yrs)
        hdr3 <- paste0(site.id, "   ", lead.invs, comp.date)
    }

    ## Loop through series and write each one
    nseries <- ncol(rwl.df)
    yrs.all <- as.numeric(row.names(rwl.df))
    col.names <- names(rwl.df)
    stopifnot(is.character(col.names), !is.na(col.names),
              Encoding(col.names) != "bytes")

    ## Sort years using increasing order, reorder rwl.df accordingly
    yrs.order <- sort.list(yrs.all)
    yrs.all <- yrs.all[yrs.order]
    rwl.df2 <- rwl.df[yrs.order, , drop=FALSE]

    first.year <- yrs.all[1]
    last.year <- yrs.all[length(yrs.all)]
    long.years <- FALSE
    if (first.year < -999) {
        long.years <- TRUE
        if (first.year < -9999) {
            stop("years earlier than -9999 (10000 BC) are not supported")
        }
    }
    if (last.year > 9999) {
        long.years <- TRUE
        if (last.year > 99999) {
            stop("years later than 99999 are not possible")
        }
    }

    ## The basic name.width is 7.
    name.width <- 7

    ## If we set exploit.short to TRUE:
    ## In the absence of long year numbers, it is possible to use a name
    ## that is one character longer.
    ## use.space adjusts the following:
    ## Do we use an extra space between the name and the decade
    ## (reduces maximum length of name by one)?

    ## Different interpretations exist...
    ## Setting long.names to FALSE will produce the same behavior
    ## as in old versions of the function.
    ## We offer the user only one bit of customization (i.e. two options),
    ## at this time anyway. Maybe the original idea of two customization
    ## options was too fine-grained.
    if (long.names) { # http://www.cybis.se/wiki/index.php?title=.rwl on 2010-04-21
        exploit.short <- TRUE  # limit is
        use.space <- FALSE     # 7 or 8 characters
    } else { # http://www.ncdc.noaa.gov/paleo/treeinfo.html on 2010-04-21
        exploit.short <- FALSE # limit is
        use.space <- TRUE      # 6 characters
    }

    if (exploit.short && !long.years) {
        name.width <- name.width + 1
    }
    if (use.space) {
        name.width <- name.width - 1
        opt.space <- " "
    } else {
        opt.space <- ""
    }
    name.width <- as.integer(name.width)
    year.width <-
        as.integer(12 - name.width - nchar(opt.space)) # year ends at col 12

    col.names <-
        fix.names(col.names, name.width, mapping.fname, mapping.append)

    if (append) {
        rwl.out <- file(fname, "a")
    } else {
        rwl.out <- file(fname, "w")
    }
    on.exit(close(rwl.out))
    if (length(header2) > 0) {
        cat(hdr1, line.term, file=rwl.out, sep="")
        cat(hdr2, line.term, file=rwl.out, sep="")
        cat(hdr3, line.term, file=rwl.out, sep="")
    }
    if (prec == 0.01) {
        na.str <- 9.99
        missing.str <- -9.99
        prec.rproc <- 100 # reciprocal of precision
    } else {
        na.str <- -9.999
        missing.str <- 0
        prec.rproc <- 1000
    }
    format.year <- sprintf("%%%d.0f", year.width)

    for (l in seq_len(nseries)) {
        series <- rwl.df2[[l]]
        idx <- !is.na(series)
        yrs <- yrs.all[idx]
        series <- series[idx]

        series <- c(series, na.str)
        yrs <- c(yrs, max(yrs) + 1)

        decades.vec <- yrs %/% 10 * 10
        ## Output for completely missing decades can be disabled by using
        ## the alternate definition of the "decades" list
        decades <- seq(from=min(decades.vec), to=max(decades.vec), by=10)
        ##     decades = unique(decades.vec)
        n.decades <- length(decades)

        ## 1--name.width
        rwl.df.name <- col.names[l]
        ## Pad to name.width
        rwl.df.name <- str_pad(rwl.df.name, name.width, side = "right")
        for (i in seq_len(n.decades)) {
            ## up to 4 numbers and a minus sign from long series
            dec <- decades[i]
            dec.idx <- decades.vec %in% dec
            dec.yrs <- yrs[dec.idx]
            dec.rwl <- series[dec.idx]

            ## Find negative values and mark as missing data, but
            ## allow the negative "end of series" marker when prec == 0.001
            neg.match <- dec.rwl < 0
            if (prec == 0.001 && i == n.decades) {
                neg.match[length(neg.match)] <- FALSE
            }
            dec.rwl[neg.match] <- missing.str

            ## Find missing data.
            if (n.decades == 1) {
                all.years <- dec.yrs[1]:dec.yrs[length(dec.yrs)]
            } else if (i == 1) {
                all.years <- dec.yrs[1]:(dec + 9)
            } else if (i == n.decades) {
                all.years <- dec:dec.yrs[length(dec.yrs)]
            } else {
                all.years <- dec:(dec + 9)
            }
            ## Mark missing data.
            if (length(all.years) > length(dec.yrs)) {
                missing.years <- setdiff(all.years, dec.yrs)
                dec.yrs <- c(dec.yrs, missing.years)
                dec.rwl <- c(dec.rwl,
                             rep(missing.str, times=length(missing.years)))
                dec.order <- sort.list(dec.yrs)
                dec.yrs <- dec.yrs[dec.order]
                dec.rwl <- dec.rwl[dec.order]
            }

            ## Pad to year.width (no leading zero)
            dec.year1 <- sprintf(format.year, dec.yrs[1])

            ## Convert millimeters to the desired precision
            dec.rwl <- round(dec.rwl * prec.rproc)

            ## Find and correct illegal uses of the stop marker
            if (prec == 0.01) {
                end.match <- dec.rwl == 999
                if (i == n.decades) {
                    end.match[length(end.match)] <- FALSE
                }
                dec.rwl[end.match] <-
                    sample(c(998, 1000), sum(end.match), replace=TRUE)
            }

            ## Pad to nchar 6 (no leading zero)
            dec.rwl <- sprintf("%6.0f", dec.rwl)

            cat(rwl.df.name, opt.space, dec.year1, dec.rwl, line.term,
                file = rwl.out, sep="")
        }
    }
    fname
}

combine.rwl <- function(x, y = NULL) {
    combinator <- function(x, y) {
        dim.x2 <- ncol(x)
        dim.y2 <- ncol(y)
        if(dim.x2 > 0 && dim.y2 > 0){
            dim2 <- dim.x2 + dim.y2
            years.x <- rownames(x)
            years.y <- rownames(y)
            min.x <- as.numeric(years.x[1])
            min.y <- as.numeric(years.y[1])
            max.x <- as.numeric(years.x[length(years.x)])
            max.y <- as.numeric(years.y[length(years.y)])
            min.year <- min(min.x, min.y)
            years <- min.year:max(max.x, max.y)
            new <- matrix(NA, nrow = length(years), ncol = dim2)
            new[(min.x-min.year+1):(max.x-min.year+1), seq_len(dim.x2)] <- x
            new[(min.y-min.year+1):(max.y-min.year+1), (dim.x2+1):dim2] <- y
            rownames(new) <- years
            colnames(new) <- c(colnames(x), colnames(y))
            new
        } else if(dim.y2 == 0) {
            x
        } else {
            y
        }
    }
    ## check, if x is a list with data.frames inside it
    ## if yes: forget about y, and loop through all items of x and apply
    ## combinator() one by one
    ## if no: just use combinator() with x and y
    if (is.list(x)) {
        n <- length(x)
        if (n > 0 && all(vapply(x, is.data.frame, TRUE))) {
            new.mat <- as.matrix(x[[1]])
            for (i in inc(2, n))
                new.mat <- combinator(new.mat, as.matrix(x[[i]]))
        } else if (is.data.frame(x) && is.data.frame(y)) {
            new.mat <- combinator(as.matrix(x), as.matrix(y))
        } else {
            stop("Nothing to combine here. Please supply data.frames formatted according to the data standards in dplR.")
        }
    } else {
        stop("Nothing to combine here. Please supply data.frames formatted according to the data standards in dplR.")
    }
    new.mat <- as.data.frame(new.mat)
    class(new.mat) <- c("rwl", "data.frame")
    new.mat
}

## Chronology stripping after Fowler & Boswijk 2003
## 1) all series are standardized using splines with 50% FC at 20 and
## 200 yrs
## 2) EPS of chronology with all series present
## 3) iterate through all series, calculating leave-one-out EPS
## 4) discard the single series (if any) whose removal increases EPS the most
## 5) iterate 2)-4) until no further increase in EPS is yielded
## 6) repeat 2)-5), but this time reinsert previously removed series
## 7) iterate through each year of the chronology and compare stripped
## and unstripped EPS for each year

## CAUTION: the function returns a data.frame of raw tree-ring widths,
## not a chronology

strip.rwl <- function(rwl, ids = NULL, verbose = FALSE, comp.plot = FALSE,
                      legacy.eps = FALSE) {

    if (!is.data.frame(rwl)) {
        stop("'rwl' must be a data.frame")
    }
    names.rwl <- names(rwl)
    if (is.null(names.rwl) || anyDuplicated(names.rwl) ||
        any(is.na(names.rwl))) {
        stop("'rwl' must have unique, non-NA names")
    }

    ids2 <- ids
    if (!is.null(ids)) {
        if (!is.data.frame(ids)) {
            stop("'ids' must be NULL or a data.frame")
        }
        rownames.ids <- row.names(ids)
        if (is.null(rownames.ids) || !all(names.rwl %in% rownames.ids)) {
            if (length(rwl) == nrow(ids)) {
                row.names(ids2) <- names(rwl)
            } else {
                stop("dimension problem: ", "'length(rwl)' != 'nrow(ids)'")
            }
        }
    }
    ids2 <- ids2[names.rwl, ]

    ## double detrend rwl
    rwl.d1 <- detrend(rwl, method = "Spline", nyrs = 20)
    rwl.d2 <- detrend(rwl.d1, method = "Spline", nyrs = 200)
    rwl.all <- rwl.d2

    eps.imp <- TRUE
    iter <- 1
    n <- length(rwl.all)

    ## This vector holds the name of the removed series for *each*
    ## iteration.  Names will be removed from the vector as series are
    ## added back in.
    removed <- rep(NA_character_, n)

    ## calc EPS for complete data.frame
    if (legacy.eps) {
        eps.all <- rwi.stats.legacy(rwl.all, ids2)$eps
    } else {
        eps.all <- rwi.stats(rwl.all, ids2)$eps
    }

    while (eps.imp) {
        eps.loo <- numeric(n)
        ## loop through series and flag series which *lower* EPS when
        ## included
        if (verbose) {
            cat("REMOVE -- Iteration ", iter, ": Initial EPS: ", eps.all,
                "\n", sep = "")
            cat("Leave-one-out EPS:\n")
        }
        names.all <- names(rwl.all)
        for (i in seq_len(n)) {
            rwl.loo <- rwl.all[-i]
            if (legacy.eps) {
                eps.loo[i] <-
                    rwi.stats.legacy(rwl.loo, ids2[names.all[-i], ])$eps
            } else {
                eps.loo[i] <- rwi.stats(rwl.loo, ids2)$eps
            }
            if (verbose) {
                cat(names.all[i], ": ", eps.loo[i], sep = "")
                if (eps.loo[i] > eps.all) {
                    cat(" *\n")
                } else {
                    cat("\n")
                }
            }
        }
        if (verbose) {
            cat("   ***\n")
        }
        if (any(eps.loo > eps.all)) {
            rm.idx <- which.max(eps.loo)
            rwl.all <- rwl.all[-rm.idx]
            eps.all.iter <- eps.loo[rm.idx]
            removed[iter] <- names.all[rm.idx]
            cat("REMOVE -- Iteration ", iter, ": leaving series ",
                removed[iter], " out.\n", sep = "")
            cat("EPS improved from ", eps.all, " to ", eps.all.iter,
                ".\n\n", sep = "")
            eps.all <- eps.all.iter
            iter <- iter + 1
            n <- n - 1
        } else {
            eps.imp <- FALSE
            cat("REMOVE -- Iteration ", iter,
                ": no improvement of EPS. Aborting...\n", sep = "")
        }
    }

    removed <- removed[seq_len(iter - 1)]

    if (iter > 1) {

        ## reinsert previously removed series (if there are any)
        eps.imp <- TRUE
        n <- iter - 1
        iter <- 1

        while (eps.imp) {
            eps.reins <- numeric(n)
            if (legacy.eps) {
                eps.init <-
                    rwi.stats.legacy(rwl.all, ids2[names(rwl.all), ])$eps
            } else {
                eps.init <- rwi.stats(rwl.all, ids2)$eps
            }
            if (verbose) {
                cat("REINSERT -- Initial EPS:", eps.init, "\n")
                cat("Back-in EPS:\n")
            }
            for (i in seq_len(n)) {
                series.name <- removed[i]
                rwl.reins <- cbind(rwl.all, rwl.d2[series.name])
                if (legacy.eps) {
                    eps.reins[i] <-
                        rwi.stats.legacy(rwl.reins,
                                         ids2[names(rwl.reins), ])$eps
                } else {
                    eps.reins[i] <- rwi.stats(rwl.reins, ids2)$eps
                }
                if (verbose) {
                    cat(series.name, ": ", eps.reins[i], sep = "")
                    if (eps.reins[i] > eps.init) {
                        cat(" *\n")
                    } else {
                        cat(" \n")
                    }
                }
            }
            if (any(eps.reins > eps.init)) {
                reins.idx <- which.max(eps.reins)
                recovered <- removed[reins.idx]
                rwl.all <- cbind(rwl.all, rwl.d2[recovered])
                removed <- removed[-reins.idx]
                if (legacy.eps) {
                    eps.all <-
                        rwi.stats.legacy(rwl.all, ids2[names(rwl.all), ])$eps
                } else {
                    eps.all <- rwi.stats(rwl.all, ids2)$eps
                }
                cat("REINSERT -- Iteration ", iter, ": Inserting series ",
                    recovered, ". EPS improved from ", eps.init,
                    " to ", eps.all, ".\n", sep = "")
                iter <- iter + 1
                n <- n - 1
            } else {
                eps.imp <- FALSE
                cat("REINSERT -- Iteration ", iter,
                    ": no improvement of EPS. Aborting...\n", sep = "")
            }
        }
    }
    ## if comp.plot == T, compare for each year stripped and unstripped
    ## chronologies by EPS
    n.removed <- length(removed)
    if (comp.plot && n.removed > 0) {
        yrs <- as.numeric(row.names(rwl.d2))
        nyrs <- length(yrs)
        comp.1 <- rep(NA_real_, nyrs)
        comp.2 <- rep(NA_real_, nyrs)
        notna.d2 <- !is.na(rwl.d2)
        notna.all <- !is.na(rwl.all)
        for (i in seq_len(nyrs)) {
            present.d2 <- which(notna.d2[i, ])
            present.str <- which(notna.all[i, ])
            if (legacy.eps) {
                if (length(present.d2) > 1) {
                    comp.1[i] <-
                        rwi.stats.legacy(rwl.d2[present.d2],
                                         ids2[present.d2, ])$eps
                }
                if (length(present.str) > 1) {
                    comp.2[i] <-
                        rwi.stats.legacy(rwl.all[present.str],
                                         ids2[present.str, ])$eps
                }
            } else {
                if (length(present.d2) > 1) {
                    comp.1[i] <- rwi.stats(rwl.d2[present.d2], ids2)$eps
                }
                if (length(present.str) > 1) {
                    comp.2[i] <- rwi.stats(rwl.all[present.str], ids2)$eps
                }
            }
        }
        diffs <- comp.2 - comp.1
        plot(yrs, diffs, col = ifelse(diffs >= 0, "blue", "red"),
             pch = "-", xlab = gettext("Year", domain="R-dplR"))
    }
    ## return *raw* series
    if (n.removed > 0) {
        rwl[-match(removed, names.rwl)]
    } else {
        rwl
    }
}

xdate.floater <- function(rwl, series, series.name = "Unknown", min.overlap = 50,
                          n = NULL, prewhiten = TRUE, biweight = TRUE,
                          method = c("spearman", "pearson", "kendall"),
                          make.plot = TRUE, return.rwl = TRUE, verbose = TRUE) {

  method2 <- match.arg(method)

  # Trim series in case it has NAs (e.g., submitted straight from an rwl)
  idx.good <- !is.na(series)
  series <- series[idx.good]
  nSeries <- length(series)

  ## Turn off warnings for this function.
  ## The significance test for Spearman's rho often produces warnings.
  w <- options(warn = -1)
  on.exit(options(w))

  ## Normalize
  tmp <- normalize.xdate(rwl, series, n, prewhiten, biweight)
  master <- tmp$master
  series2 <- tmp$series
  idx.good <- !is.na(series2)
  series2 <- series2[idx.good]

  ## Trim master so there are no NaN (e.g., from dividing when only one series)
  idx.good <- !is.nan(master)
  x <- master[idx.good]

  yrs <- as.numeric(names(x))
  y <- series2

  nx <- length(x)
  ny <- length(y)

  if (verbose) {
    cat("Original rwl years: ", min(time(rwl)), " to ", max(time(rwl)),
        " (", length(time(rwl)), ")\n", sep = "")
    cat("Detrended rwl years: ", min(yrs), " to ", max(yrs),
        " (", length(yrs), ")\n", sep = "")
    cat("Original series length:", nSeries, "\n")
    cat("Detrended series length:", ny, "\n")
    cat("Minimum overlap for search:", min.overlap, "\n")
  }

  if (min.overlap > ny) {
    stop("min.overlap must be less than series length after detrending")
  }

  minYrsOut <- numeric()
  maxYrsOut <- numeric()
  rOut <- numeric()
  pOut <- numeric()
  nOut <- numeric()

  # Crawl through backwards because start years on both the master and the
  # series can be affected by normalizing (e.g., Hanning, prewhiten).
  # End years are unaffected, so crawling backwards keeps the dating correct.
  crawl <- (nx + (ny - min.overlap)):(min.overlap)
  edgeCounter <- 0

  for (i in crawl) {
    if (i > nx) {
      xInd <- (i - ny + 1):nx
      yInd <- 1:(ny - (i - nx))
      tmp <- cor.test(x[xInd], y[yInd], method = method2,
                      alternative = "greater")
      rOut[i] <- tmp$estimate
      pOut[i] <- tmp$p.val
      # End date is the max of xInd plus the overlap off the edge
      maxYrsOut[i] <- max(yrs[xInd]) + ny - min.overlap + edgeCounter
      edgeCounter <- edgeCounter - 1
      minYrsOut[i] <- maxYrsOut[i] - nSeries + 1
      nOut[i] <- length(x[xInd])
    }
    if (i >= ny & i <= nx) {
      xInd <- (i - ny + 1):i
      tmp <- cor.test(x[xInd], y, method = method2, alternative = "greater")
      rOut[i] <- tmp$estimate
      pOut[i] <- tmp$p.val
      maxYrsOut[i] <- max(yrs[xInd])
      minYrsOut[i] <- max(yrs[xInd]) - nSeries + 1
      nOut[i] <- length(x[xInd])
    }
    if (i < ny) {
      xInd <- 1:i
      yInd <- xInd + ny - length(xInd)
      tmp <- cor.test(x[xInd], y[yInd], method = method2,
                      alternative = "greater")
      rOut[i] <- tmp$estimate
      pOut[i] <- tmp$p.val
      maxYrsOut[i] <- max(yrs[xInd])
      minYrsOut[i] <- max(yrs[xInd]) - nSeries + 1
      nOut[i] <- length(x[xInd])
    }
  }

  floaterCorStats <- data.frame(minYrsOut, maxYrsOut, rOut, pOut, nOut)
  names(floaterCorStats) <- c("first", "last", "r", "p", "n")
  mask <- rowSums(is.na(floaterCorStats)) == 0
  floaterCorStats <- floaterCorStats[mask, ]

  # Trim so end year can't be in the future
  thisYr <- as.numeric(format(Sys.time(), "%Y"))
  floaterCorStats <- floaterCorStats[floaterCorStats$last <= thisYr, ]

  # Best-fit dates
  rBest <- which.max(floaterCorStats$r)
  firstBest <- floaterCorStats$first[rBest]
  lastBest <- floaterCorStats$last[rBest]
  rBestVal <- floaterCorStats$r[rBest]
  pBestVal <- floaterCorStats$p[rBest]

  if (verbose) {
    cat("Years searched:", min(floaterCorStats$first), "to",
        max(floaterCorStats$last), "\n")
    cat("Highest overall correlation for series is", round(rBestVal, 2),
        "with dates", firstBest, "to", lastBest, "\n")
  }

  # Build rwl objects for best-fit position
  names(series) <- firstBest:lastBest
  rwlOut <- as.rwl(data.frame(series))
  names(rwlOut) <- series.name
  class(rwlOut) <- c("rwl", "data.frame")
  rwlCombined <- combine.rwl(rwl, rwlOut)

  res <- list(series.name    = series.name,
              floaterCorStats = floaterCorStats,
              rwlCombined    = rwlCombined,
              rwlOut         = rwlOut)
  class(res) <- c("floater", "list")

  if (make.plot) {
    plot(res)
  }

  if (!return.rwl) {
    res <- floaterCorStats
  }

  res
}


print.floater <- function(x, ...) {
  fcs <- x$floaterCorStats
  rBest <- which.max(fcs$r)
  cat("Floater: ", x$series.name, "\n", sep = "")
  cat("Years searched:", min(fcs$first), "to", max(fcs$last), "\n")
  cat("Best correlation:", round(fcs$r[rBest], 3),
      "with dates", fcs$first[rBest], "to", fcs$last[rBest], "\n")
  invisible(x)
}


plot.floater <- function(x, ...) {
  series.name     <- x$series.name
  floaterCorStats <- x$floaterCorStats
  rwlCombined     <- x$rwlCombined

  # Quantile envelope from the dated master (excluding the floater)
  rwlOrig <- rwlCombined
  rwlOrig[, series.name] <- NULL
  quantCor <- quantile(interseries.cor(rwlOrig)[, 1],
                       probs = c(0.05, 0.5, 0.95))

  op <- par(no.readonly = TRUE)
  on.exit(par(op))

  # Plot 1: segment plot with floater highlighted in green
  yr <- as.numeric(row.names(rwlCombined))
  first.year <- as.matrix(apply(rwlCombined, 2, yr.range, yr.vec = yr))[1, ]
  last.year  <- as.matrix(apply(rwlCombined, 2, yr.range, yr.vec = yr))[2, ]
  neworder <- order(first.year, decreasing = FALSE)
  segs <- rwlCombined[, neworder, drop = FALSE]
  n.col <- ncol(segs)
  seq.col <- seq_len(n.col)
  for (i in seq.col) {
    segs[[i]][!is.na(segs[[i]])] <- i
  }
  seg2col    <- which(names(segs) == series.name)
  segs.axis2 <- names(segs)
  segs.axis4 <- names(segs)
  segs.axis2[seq(2, n.col, by = 2)] <- NA
  segs.axis4[seq(1, n.col, by = 2)] <- NA

  par(mfcol = c(2, 1))
  par(mar = c(-0.1, 5, 2, 5) + 0.1, mgp = c(1.1, 0.1, 0), tcl = 0.5,
      xaxs = "i", yaxs = "i")
  plot(yr, segs[[1]], type = "n", ylim = c(0, n.col + 1),
       xlim = range(floaterCorStats$first, floaterCorStats$last),
       axes = FALSE, ylab = "", xlab = "")
  abline(h = seq.col, lwd = 1, col = "grey")
  grid(ny = NA)
  apply(segs, 2, lines, x = yr, lwd = 4, lend = 2, col = "grey60")
  abline(h = seq.col[[seg2col]], lwd = 1, col = "darkgreen")
  lines(x = yr, y = segs[[seg2col]], lwd = 4, lend = 2, col = "darkgreen")
  axis(2, at = seq.col, labels = segs.axis2, srt = 45, tick = FALSE, las = 2)
  axis(4, at = seq.col, labels = segs.axis4, srt = 45, tick = FALSE, las = 2)
  axis(3)
  box()

  # Plot 2: correlation by end year with interseries quantile envelope
  rBest     <- which.max(floaterCorStats$r)
  firstBest <- floaterCorStats$first[rBest]
  lastBest  <- floaterCorStats$last[rBest]
  rBestVal  <- floaterCorStats$r[rBest]

  sig <- qnorm(1 - 0.05 / 2) / sqrt(floaterCorStats$n)

  par(mar = c(2, 5, -0.1, 5) + 0.1, yaxs = "r")
  plot(floaterCorStats$last, floaterCorStats$r,
       type = "n", xlab = "Year", ylab = "End Year Cor.",
       xlim = range(floaterCorStats$first, floaterCorStats$last),
       ylim = range(quantCor, floaterCorStats$r), axes = FALSE)

  # Interseries correlation envelope (5th-95th percentile of master)
  xx <- c(min(floaterCorStats$first), max(floaterCorStats$last),
          max(floaterCorStats$last), min(floaterCorStats$first))
  yy <- c(quantCor[1], quantCor[1], quantCor[3], quantCor[3])
  polygon(xx, yy, border = NA, col = "lightblue")
  abline(h = quantCor[2], col = "darkblue")

  lines(floaterCorStats$last, sig, lty = "dashed")
  lines(floaterCorStats$last, floaterCorStats$r, col = "grey")
  abline(h = 0)

  points(lastBest,  rBestVal, col = "darkgreen", pch = 20)
  points(firstBest, rBestVal, col = "darkgreen", pch = 20)
  segments(x0 = firstBest, x1 = lastBest, y0 = rBestVal, y1 = rBestVal,
           lty = "dashed", col = "darkgreen")
  text(x = lastBest,  y = rBestVal, labels = lastBest,
       col = "darkgreen", adj = c(0, 1))
  text(x = firstBest, y = rBestVal, labels = firstBest,
       col = "darkgreen", adj = c(1, 1))
  axis(1); axis(2); box()

  invisible(x)
}

as.rwl <- function(x){
  if (!(is.data.frame(x) || is.matrix(x))) {
    stop("x must be a data.frame or matrix")
  }
  if (is.matrix(x) ||
      (!inherits(x, "rwl") && !identical(class(x), "data.frame"))) {
    x <- as.data.frame(x)
  }
  ## are rownames the time vector?
  row_names <- row.names(x)
  tmTest <- !is.null(row_names) && all(diff(as.numeric(row_names)) == 1)
  if (!tmTest) {
    stop("x must have time (years) in the rownames so that all(diff(as.numeric(row.names(x))) == 1)")
  }
  if (!inherits(x, "rwl")) {
    class(x) <- c("rwl", "data.frame")
  }
  x
}

sea <- function(x, key, lag = 5, resample = 1000) {
    if (!is.data.frame(x)) {
        stop("'x' must be a data.frame")
    }
    stopifnot(is.numeric(lag), length(lag) == 1, is.finite(lag),
              lag >= 0, round(lag) == lag)
    stopifnot(is.numeric(resample), length(resample) == 1, is.finite(resample),
              resample >= 1, round(resample) == resample)
    if (ncol(x) >= 1) {                 # remove samp.depth if present
        x.unscaled <- x[1]
    } else {
        stop("'x' must have at least one column")
    }
    rnames <- row.names(as.matrix(x.unscaled))
    if (is.null(rnames)) {
        stop("'x' must have non-automatic row.names")
    }
    rnames <- as.numeric(rnames)
    x.scaled <- data.frame(scale(x.unscaled))
    n <- length(key)
    m <- 2 * lag + 1
    se.table <- matrix(NA_real_, ncol = m, nrow = n)
    se.unscaled.table <- se.table
    yrs.base <- (-lag):(m - lag - 1)
    seq.n <- seq_len(n)
    for (i in seq.n) {
        yrs <- as.character(key[i] + yrs.base)
        se.table[i, ] <- x.scaled[yrs, ]
        se.unscaled.table[i, ] <- x.unscaled[yrs, ]
    }
    se <- colMeans(se.table, na.rm = TRUE)
    se.unscaled <- colMeans(se.unscaled.table, na.rm = TRUE)
    re.table <- matrix(NA_real_, ncol = m, nrow = resample)
    re.subtable <- matrix(NA_real_, ncol = m, nrow = n)
    l <- length(rnames)
    trim <- c(1:lag, (l - lag + 1):l)
    rnames.red <- rnames[-trim]
    for (k in seq_len(resample)) {
        rand.key <- sample(rnames.red, n, replace = TRUE)
        for (i in seq.n) {
            re.subtable[i, ] <-
                               x.scaled[as.character(rand.key[i] + yrs.base), ]
        }
        re.table[k, ] <- colMeans(re.subtable, na.rm = TRUE)
    }
    ## calculate significance for each (lagged) year
    ## compute confidence bands, too
    p <- rep(NA_real_, m)
    re_ecdfs <- apply(re.table, 2, ecdf)
    ci <- apply(re.table, 2, quantile,
               probs = c(0.025, 0.975, 0.005, 0.995)
               )

    if (resample < 1000) {
        warning("'resample' is lower than 1000, potentially leading to less accuracy in estimating p-values and confidence bands.")
    }
    
    for (i in seq_len(m)) {
        if (is.na(se[i])) {
            warning(gettextf("NA result at position %d. ", i),
                    "You could check whether 'key' years are in range.")
        } else {
            if (se[i] < 0) {         # superposed value < 0, it is
                                        # tested whether is
                                        # significantly LOWER than
                                        # random values
                p[i] <- re_ecdfs[[i]](se[i])
            } else {                        # ditto, but v.v.
                p[i] <- 1 - re_ecdfs[[i]](se[i])
            }
        }
    }
    data.frame(lag = c(-lag:lag),
               se,
               se.unscaled,
               p,
               ci.95.lower = ci[1,],
               ci.95.upper = ci[2,],
               ci.99.lower = ci[3,],
               ci.99.upper = ci[4,]
               )
}

read.compact <- function(fname)
{
    res <- .Call(dplR.rcompact, path.expand(fname))
    min.year <- res[[1]]
    max.year <- res[[2]]
    series.ids <- res[[3]]
    series.min <- res[[4]]
    series.max <- res[[5]]
    series.mplier <- res[[6]]
    rw.mat <- res[[7]]
    project.comments <- res[[8]]
    rownames(rw.mat) <- min.year:max.year
    nseries <- ncol(rw.mat)

    cat(sprintf(ngettext(nseries,
                         "There is %d series\n",
                         "There are %d series\n",
                         domain="R-dplR"),
                nseries))
    series.min.char <- format(series.min, scientific=FALSE, trim=TRUE)
    series.max.char <- format(series.max, scientific=FALSE, trim=TRUE)
    seq.series.char <- format(seq_len(nseries), scientific=FALSE, trim=TRUE)
    cat(paste0(format(seq.series.char, width=5), "\t",
               format(series.ids, width=8), "\t",
               format(series.min.char, width=5, justify="right"), "\t",
               format(series.max.char, width=5, justify="right"), "\t",
               format(series.mplier, scientific=FALSE,drop0trailing=TRUE),"\n"),
        sep="")
    if (length(project.comments) > 0) {
        cat(gettext("Comments:", domain="R-dplR"),
            paste(project.comments, collapse="\n"), sep="\n")
    }

    rw.df <- as.data.frame(rw.mat)
    names(rw.df) <- series.ids
    class(rw.df) <- c("rwl", "data.frame")
    rw.df
}

wavelet.plot <-
    function(wave.list,
             wavelet.levels = quantile(wave.list$Power, probs=(0:10)/10),
             add.coi = TRUE, add.sig = TRUE,
             x.lab = gettext("Time", domain = "R-dplR"),
             period.lab = gettext("Period", domain = "R-dplR"),
             crn.lab = gettext("RWI", domain = "R-dplR"),
             key.cols = rev(rainbow(length(wavelet.levels)-1)),
             key.lab = parse(text = paste0("\"",
                                           gettext("Power",
                                                   domain = "R-dplR"),
                                           "\"^2")),
             add.spline = FALSE, f = 0.5, nyrs = NULL,
             crn.col = "black", crn.lwd = 1,coi.col='black',
             crn.ylim = range(wave.list$y) * c(0.95, 1.05),
             side.by.side = FALSE,
             useRaster = FALSE, res = 150, reverse.y = FALSE, ...)
{

    ## Wavelet transform variables:
    y <- wave.list$y
    x <- wave.list$x
    period <- wave.list$period
    Signif <- wave.list$Signif
    coi <- wave.list$coi
    Power <- wave.list$Power
    siglvl <- wave.list$siglvl

    stopifnot(is.numeric(x), is.numeric(y), is.numeric(period),
              is.numeric(Signif), is.numeric(coi), is.numeric(Power),
              is.numeric(siglvl), is.logical(useRaster),
              length(useRaster) == 1,
              identical(side.by.side, TRUE) || identical(side.by.side, FALSE))
    stopifnot(is.numeric(wavelet.levels))
    n.x <- length(x)
    n.period <- length(period)
    dim.Power <- dim(Power)
    stopifnot(length(dim.Power) == 2, n.x == length(y), dim.Power[1] == n.x,
              dim.Power[2] == n.period, length(Signif) == n.period,
              length(coi) == n.x, length(siglvl) == 1, n.x >= 2, n.period >= 2)
    if (any(diff(x) <= 0) || any(diff(period) <= 0)) {
        stop("'wave.list$x' and 'wave.list$period' must be strictly ascending")
    }
    if (period[1] <= 0) {
        stop("'wave.list$period' must be positive")
    }

    coi[coi == 0] <- 1e-12

    ## Expand signif --> (length(wave.list$Scale))x(N) array
    Signif <- t(matrix(Signif, dim.Power[2], dim.Power[1]))
    ## Where ratio > 1, power is significant
    Signif <- Power / Signif

    ## Period is in years, period2 is in powers of 2
    period2 <- log2(period)
    ytick <- unique(trunc(period2)) # Unique integer powers of 2
    ytickv <- 2^ytick # Labels are in years

    ## coi is in years, coi2 in powers of 2
    coi2 <- log2(coi)
    coi2[coi2 < 0] <- 0
    coi2.yy <- c(coi2, rep(max(period2, na.rm=TRUE), n.x))
    coi2.yy[is.na(coi2.yy)] <- coi[2]
    yr.vec.xx <- c(x, rev(x))

    par.orig <- par(c("mar", "las", "mfrow", "mgp", "tcl"))
    on.exit(par(par.orig))
    nlevels <- length(wavelet.levels)
    seq.level <- seq_len(nlevels - 1)
    key.labs <- formatC(wavelet.levels, digits = 4, format = "f")
    asp <- NA
    xaxs <- "i"
    yaxs <- "i"
    las <- 1
    xlim <- range(x, finite=TRUE)
    ylim <- range(period2, finite=TRUE)
    if (isTRUE(reverse.y)) {
        ylim <- rev(ylim)
    }

    ## plot set up
    if (side.by.side) {
        layout(matrix(c(3, 2, 1), nrow=1, byrow=TRUE),
               widths=c(1, 1, 0.2))
        scale.xlim <- c(0, 1)
        scale.ylim <- c(1, nlevels)
        scale.side <- 4
        scale.xleft <- 0
        scale.ybottom <- seq.level
        scale.xright <- 1
        scale.ytop <- 2:nlevels
        mar1 <- c(3, 1, 3, 3)
        mar2 <- c(3, 3, 3, 3)
        mar3 <- c(3, 3, 3, 3)
    } else {
        layout(matrix(c(3, 2, 1), ncol=1, byrow=TRUE),
               heights=c(1, 1, 0.3))
        scale.xlim <- c(1, nlevels)
        scale.ylim <- c(0, 1)
        scale.side <- 1
        scale.xleft <- seq.level
        scale.ybottom <- 0
        scale.xright <- 2:nlevels
        scale.ytop <- 1
        mar1 <- c(3, 3, 0.1, 3)
        mar2 <- mar1
        mar3 <- c(0.1, 3, 3, 3)
    }
    ## plot 1: scale
    par(mar=mar1, tcl=0.5, mgp=c(1.5, 0.25, 0), las=las)
    dev.hold()
    on.exit(dev.flush(), add=TRUE)
    plot.new()
    plot.window(ylim=scale.ylim, xlim=scale.xlim,
                xaxs=xaxs, yaxs=yaxs, asp=asp)
    rect(scale.xleft, scale.ybottom, scale.xright, scale.ytop, col = key.cols)
    axis(scale.side, at=seq_along(wavelet.levels), labels=key.labs)
    ## add units
    if (side.by.side) {
        title(key.lab, cex.main=1)
    } else {
        title(sub=key.lab, cex.sub=1, line=1.5)
    }
    ## plot 2: contour-image
    par(mar=mar2, tcl=0.5, mgp=c(1.5, 0.25, 0))
    plot.new()

    plot.window(xlim, ylim, "", xaxs=xaxs, yaxs=yaxs, asp=asp, las=las)
    if (is.na(useRaster)) {
        useRaster2 <- names(dev.cur()) %in% c("pdf", "postscript")
    } else {
        useRaster2 <- useRaster
    }
    ## note replacement of .Internal(filledcontour(as.double(x),...)
    ## with .filled.contour() as of R-2.15.0
    cl <- quote(.filled.contour(as.double(x),
                                as.double(period2),
                                Power,
                                as.double(wavelet.levels),
                                key.cols))
    if (useRaster2) {
        args <- list(...)
        argNames <- names(args)
        if (is.null(argNames)) {
            args <- NULL
        } else {
            args <- args[!is.na(argNames) & nzchar(argNames)]
        }
        if (length(args) == 0L) {
            Call <- as.call(c(as.name("rasterPlot"),
                              alist(expr = cl, res = res, antialias = "none",
                                    interpolate = FALSE)))
        } else {
            Call <- as.call(c(as.name("rasterPlot"), args))
            Call <- as.list(match.call(rasterPlot, Call))
            anam <- names(Call[-1L])
            Call[["expr"]] <- quote(cl)
            Call[["res"]] <- quote(res)
            Call[["region"]] <- "plot"
            Call[["draw"]] <- TRUE
            if (!("antialias" %in% anam)) {
                Call[["antialias"]] <- "none"
            }
            if (!("interpolate" %in% anam)) {
                Call[["interpolate"]] <- FALSE
            }
            Call <- as.call(Call)
        }
        tryCatch(eval(Call),
                 error = function(e) {
                     message(as.character(e), appendLF = FALSE)
                     message("reverting to useRaster=FALSE")
                     eval(cl)
                 })
    } else {
        eval(cl)
    }
    if (isTRUE(add.sig)) {
        contour(x, period2, Signif, levels=1, labels=siglvl,
                drawlabels = FALSE, axes = FALSE,
                frame.plot = FALSE, add = TRUE,
                lwd = 2, col="black")
    }
    if (isTRUE(add.coi)) {
        polygon(yr.vec.xx, coi2.yy, density=c(10, 20),
                angle=c(-45, 45), col=coi.col)
    }
    axis(1)
    axis(2, at = ytick, labels = ytickv)
    if (side.by.side) {
        axis(3)
        axis(4, at = ytick, labels = ytickv)
    } else {
        axis(3, labels = NA)
        axis(4, at = ytick, labels = NA)
    }
    title(xlab = x.lab, ylab = period.lab)
    box()

    ## plot 3: chron
    par(mar = mar3, las=0)
    plot(x, y, type = "l", xlim, xaxs = xaxs, yaxs = yaxs,
         asp = asp, xlab = "", ylab = "", axes = FALSE, col = crn.col,
         lwd = crn.lwd, ylim = crn.ylim)
    if (add.spline) {
        spl <- y
        tmp <- na.omit(spl)
        if (is.null(nyrs)) {
            nyrs2 <- length(tmp) * 0.33
        } else {
            nyrs2 <- nyrs
        }
        #tmp <- ffcsaps(y = tmp, x = seq_along(tmp), nyrs = nyrs2, f = f)
        tmp <- caps(y = tmp, nyrs = nyrs2, f = f)
        spl[!is.na(spl)] <- tmp
        lines(x, spl, col = "red", lwd = 2)
    }
    axis(3)
    axis(4)
    if (side.by.side) {
        axis(1)
        axis(2)
        title(xlab = x.lab, ylab = crn.lab)
    } else {
        axis(1, labels = NA)
        axis(2, labels = NA)
        mtext(crn.lab, side=4, line=1.5, cex=0.75)
    }
    box()
    invisible()
}

ccf.series.rwl <- function(rwl, series,
                           series.yrs = as.numeric(names(series)),
                           seg.length = 50, bin.floor = 100, n = NULL,
                           prewhiten = TRUE, biweight = TRUE,
                           pcrit = 0.05, lag.max = 5, make.plot = TRUE,
                           floor.plus1 = FALSE, series.x = FALSE, ...) {

    series.x.txt <- ifelse(series.x,
                           "NB: With series.x = TRUE, postive lags indicate missing rings in series",
                           "NB: With series.x = FALSE (default), negative lags indicate missing rings in series")
    cat(series.x.txt)
    ## Handle different types of 'series'
    tmp <- pick.rwl.series(rwl, series, series.yrs)
    rwl2 <- tmp[[1]]
    series2 <- tmp[[2]]

    ## run error checks
    qa.xdate(rwl2, seg.length, n, bin.floor)
    if (lag.max > seg.length) {
        stop("'lag.max' > 'seg.length'")
    }
    seg.lag <- seg.length / 2

    ## Normalize.
    tmp <- normalize.xdate(rwl2, series2, n, prewhiten, biweight)
    master <- tmp$master

    ## trim master so there are no NaN like dividing when only one
    ## series for instance.
    idx.good <- !is.nan(master)
    master <- master[idx.good]
    yrs <- as.numeric(names(master))

    series2 <- tmp$series
    series.yrs2 <- as.numeric(names(series2))
    ## trim series in case it was submitted straight from the rwl
    idx.good <- !is.na(series2)
    series.yrs2 <- series.yrs2[idx.good]
    series2 <- series2[idx.good]

    ## clip series to master dimensions
    series2 <- series2[series.yrs2 %in% yrs]
    series.yrs2 <- as.numeric(names(series2))
    ## clip master to series dimensions
    master <- master[yrs %in% series.yrs2]
    yrs <- as.numeric(names(master))

    if (is.null(bin.floor) || bin.floor == 0) {
        min.bin <- min(series.yrs2)
    } else if(floor.plus1) {
        min.bin <- ceiling((min(series.yrs2) - 1) / bin.floor) * bin.floor + 1
    } else {
        min.bin <- ceiling(min(series.yrs2) / bin.floor) * bin.floor
    }
    to <- max(series.yrs2) - seg.length - seg.lag + 1
    if (min.bin > to) {
        cat(gettextf("maximum year in (filtered) series: %d\n",
                     max(series.yrs2), domain="R-dplR"))
        cat(gettextf("first bin begins: %d\n", min.bin, domain="R-dplR"))
        cat(gettext("cannot fit two segments (not enough years in the series)\n",
                    domain="R-dplR"))
        stop("shorten 'seg.length' or adjust 'bin.floor'")
    }
    bins <- seq(from=min.bin, to=to + seg.lag, by=seg.lag)
    bins <- cbind(bins, bins + (seg.length - 1), deparse.level=0)
    nbins <- nrow(bins)
    bin.names <- paste0(bins[, 1], ".", bins[, 2])

    ## structures for results
    lag.vec <- seq(from=-lag.max, to=lag.max, by=1)
    res.cor <- matrix(NA, length(lag.vec), nbins)
    rownames(res.cor) <- paste("lag", lag.vec, sep=".")
    colnames(res.cor) <- bin.names

    ## loop through bins
    for (j in seq_len(nbins)) {
        mask <- yrs%in%seq(from=bins[j, 1], to=bins[j, 2])
        ## cor is NA if there is not complete overlap
        if (!any(mask) ||
            any(is.na(series2[mask])) ||
            any(is.na(master[mask])) ||
            table(mask)[2] < seg.length) {
            bin.ccf <- NA
        }
        else {
            if(series.x){
              tmp <- ccf(x = series2[mask], y = master[mask], lag.max=lag.max,
                         plot=FALSE)
            }
          else {
            tmp <- ccf(x = master[mask], y = series2[mask], lag.max=lag.max,
                       plot=FALSE)
          }

            bin.ccf <- as.vector(tmp$acf)
        }
        res.cor[, j] <- bin.ccf
    }
    ## plot
    if (make.plot) {
        ccf.df <- data.frame(r = c(res.cor, recursive=TRUE),
                             bin = rep(colnames(res.cor),
                             each=length(lag.vec)),
                             lag = rep(lag.vec, nbins),
                             stringsAsFactors = FALSE)
        ## order bins so that lattice definitely keeps them in
        ## ascending order (i.e., no factor order funnies with long
        ## series)
        ccf.df$bin <- factor(ccf.df$bin,
                             levels = colnames(res.cor)[order(bins[, 1])])

        sig <- qnorm(1 - pcrit / 2) / sqrt(seg.length)
        sig <- c(-sig, sig)
        ccf.plot <-
            xyplot(r ~ lag | bin, data = ccf.df,
                   ylim = range(ccf.df$r, sig, na.rm=TRUE) * 1.1,
                   xlab = gettext("Lag", domain="R-dplR"),
                   ylab = gettext("Correlation", domain="R-dplR"),
                   col.line = NA,
                   cex = 1.25,
                   panel = function(x, y, ...) {
                       panel.abline(h=seq(from=-1, to=1, by=0.1),
                                    lty="solid", col="gray")
                       panel.abline(v=lag.vec, lty="solid", col="gray")
                       panel.abline(h=0, v=0, lwd=2)
                       panel.abline(h=sig, lwd=2, lty="dashed")
                       #col <- ifelse(y > 0, "#E41A1C", "#377EB8")
                       col <- ifelse(y > 0, "darkred", "darkblue")
                       bg <- ifelse(y > 0, "lightsalmon", "lightblue")
                       ## segments, dots for all r
                       #panel.segments(x1=x, y1=0, x2=x, y2=y, col=col, lwd=2)
                       #panel.dotplot(x, y, col = col, ...)
                       panel.segments(x1=x, y1=0, x2=x, y2=y,
                                      col=col, lwd=2)
                       panel.dotplot(x, y, col = col, fill=bg,
                                     pch=21,...)
                   }, ...)
        if(series.x) { ccf.plot <- update(ccf.plot, sub=series.x.txt) }
        else { ccf.plot <- update(ccf.plot, sub=series.x.txt) }
        trellis.par.set(strip.background = list(col = "transparent"),
                        warn = FALSE,
                        par.sub.text = list(font = 1, cex=0.75,
                                            just = "left",
                                            x = grid::unit(5, "mm")))
        print(ccf.plot)
    }
    res <- list(res.cor,bins)
    names(res) <- c("ccf", "bins")
    res
}

normalize.xdate <- function(rwl, series, n, prewhiten, biweight,
                            leave.one.out = FALSE) {
    loo <- isTRUE(leave.one.out)
    ## Run hanning filter over the data if n isn't NULL
    ## divide by mean if n is null
    if(is.null(n)){
        master.stats <- colMeans(rwl, na.rm=TRUE)
        master.df <- sweep(rwl, 2, master.stats, "/")
        if (!loo) {
            series.out <- series / mean(series, na.rm=TRUE)
        }
    } else {
        master.stats <- apply(rwl, 2, hanning, n)
        master.df <-  rwl / master.stats
        if (!loo) {
            series.out <- series / hanning(series, n)
        }
    }
    if (loo) {
        nseries <- ncol(rwl)
        ## Apply ar if prewhiten
        if(prewhiten){
            ## mark any columns without at least four observations
            goodCol <- colSums(!is.na(master.df)) > 3
            series.out <-  apply(master.df, 2, ar.func)
        } else {
            goodCol <- rep.int(TRUE, nseries)
            series.out <- master.df
        }
        master <- series.out
        if (!biweight) {
            for (i in seq_len(nseries)) {
                goodCol2 <- goodCol
                goodCol2[i] <- FALSE
                master[, i] <-
                    rowMeans(series.out[, goodCol2, drop = FALSE], na.rm=TRUE)
            }
        } else {
            for (i in seq_len(nseries)) {
                goodCol2 <- goodCol
                goodCol2[i] <- FALSE
                master[, i] <-
                    apply(series.out[, goodCol2, drop = FALSE], 1, tbrm, C = 9)
            }
        }
    } else {
        ## Apply ar if prewhiten
        if(prewhiten){
            ## drop any columns without at least four observations
            master.df <- master.df[, colSums(!is.na(master.df)) > 3, drop=FALSE]
            master.df <-  apply(master.df, 2, ar.func)
            series.out <- ar.func(series.out)
        }

        if (!biweight) master <- rowMeans(master.df, na.rm=TRUE)
        else master <- apply(master.df, 1, tbrm, C = 9)
    }
    list(master=master, series=series.out)
}

### This part of dplR was (arguably non-trivially) translated and
### adapted from public domain Fortran program REDFIT, version 3.8e
### (Michael Schulz and Manfred Mudelsee). The possibly non-free parts
### of REDFIT derived from Numerical Recipes were not used.
### http://www.geo.uni-bremen.de/geomod/staff/mschulz/
### Author of the dplR version is Mikko Korpela.
###
### Copyright (C) 2013-2015 Aalto University
###
### This program is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.
###
### This program is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### A copy of the GNU General Public License is available at
### http://www.r-project.org/Licenses/

## Comments have mostly been copied verbatim from the original version
## (a few typos were fixed). New comments are prefixed with "dplR:".

## Estimate red-noise background of an autospectrum, which is estimated from
## an unevenly spaced time series. In addition, the program corrects for the
## bias of Lomb-Scargle Fourier transform (correlation of Fourier components),
## which depends on the distribution of the sampling times t(i) along the
## time axis.
##
## Main Assumptions:
## -----------------
##     - The noise background can be approximated by an AR(1) process.
##     - The distribution of data points along the time axis is not
##       too clustered.
##     - The potential effect of harmonic signal components on the
##       estimation procedure is neglected.
##     - The time series has to be weakly stationary.
##
## The first-order autoregressive model, AR(1) model, which is used
## to describe the noise background in a time series x(t_i), reads
##
##
##            x(i) =  rho(i) * x(i-1)  +  eps(i)          (1)
##
##
## with                           t(i) - t(i-1)
##                rho(i) =  exp(- -------------)
##                                     tau
##
## and eps ~ NV(0, vareps). To ensure Var[red] = 1, we set
##
##                                2 * (t(i) - t(i-1))
##            vareps = 1 -  exp(- -------------------).
##                                       tau
##
## Stationarity of the generated AR(1) time series is assured by dropping
## the first N generated points.
##
##
## Computational Steps:
## --------------------
##
## 1. Estimate autospectrum Gxx of the unevenly spaced input time series in the
##    interval [0,fNyq], using the Lomb-Scargle Fourier transform in combination
##    with the Welch-Overlapped-Segment-Averaging (WOSA) procudure, as described
##    in Schulz and Stattegger (1997).
##
## 2. Estimate tau from the unevenly sampled time series using the time-
##    domain algorithm of Mudelsee (200?).
##
## 3. Determine the area under Gxx -> estimator of data variance ==> varx.
##
## 4. Repeat Nsim times
##    - create AR(1) time series (red) acc. to Eq. 1, using the observation
##      times of the input data, but each time with different eps(i)
##    - estimate autospectrum of red ==> Grr
##    - scale Grr such that area under the spectrum is identical to varx
##    - sum Grr ==> GrrSum
##
## 5. Determine arithmetic mean of GrrSum ==> GrrAvg.
##
## 6. Ensure that area under GrrAvg is identical to varx (account for rounding
##    errors).
##
## 7. Calculate theoretical AR(1) spectrum for the estimated tau ==> GRedth.
##
## 8. Scale GRedth such that area under the spectrum is identical to varx (this
##    step is required since the true noise variance of the data set is
##    unknown).
##
## 9. Estimate the frequency-dependent correction factor (corr) for the
##    Lomb-Scargle FT from the ratio between mean of the estimated AR(1) spectra
##    (GrrAvg) and the scaled theoretical AR(1) spectrum (GRedth).
##
## 10. Use correction factors to eliminate the bias in the estimated spectrum
##     Gxx ==> Gxxc.
##
## 11. Scale theorectical AR(1) spectrum for various significance levels.

## Notes:
## ------
## * A linear trend is subtracted from each WOSA segment.
##
## * tau is estimated separately for each WOSA segment and subsequently
##   averaged.
##
## * Default max. frequency = avg. Nyquist freq. (hifac = 1.0).
##
## dplR note: Authors of REDFIT
## Authors: Michael Schulz, MARUM and Faculty of Geosciences, Univ. Bremen
## -------- Klagenfurter Str., D-28334 Bremen
##          mschulz@marum.de
##          www.geo.uni-bremen.de/~mschulz
##
##          Manfred Mudelsee, Inst. of Meteorology, Univ. Leipzig
##          Stephanstr. 3, D-04103 Leipzig
##          Mudelsee@rz.uni-leipzig.de
##
## Reference: Schulz, M. and Mudelsee, M. (2002) REDFIT: Estimating
## ---------- red-noise spectra directly from unevenly spaced paleoclimatic
##            time series. Computers and Geosciences, 28, 421-426.


redfit <- function(x, t, tType = c("time", "age"), nsim = 1000, mctest = TRUE,
                   ofac = 4, hifac = 1, n50 = 3, rhopre = NULL,
                   p = c(0.10, 0.05, 0.02), iwin = 2,
                   txOrdered = FALSE, verbose = FALSE, seed = NULL,
                   maxTime = 10, nLimit = 10000) {
    cl <- match.call()
    if (!is.null(seed)) {
        set.seed(seed)
    }
    MIN_POINTS <- 3 # with 2 points, some windows have just 1 non-zero point
    WIN_NAMES <- c("rectangular", "welch i", "hanning",
                   "triangular", "blackman-harris")
    ## dplR: 21 is the lower limit of nsim where !anyDuplicated(c(idx80,
    ## idx90, idx95, idx99)) is TRUE.  (Also, none of the indices is
    ## 0.)  For more reliable results, a much greated value is
    ## recommended.
    NSIM_LIMIT <- 21
    ## dplR: Check
    tType2 <- match.arg(tType)
    tTime <- tType2 == "time"
    stopifnot(is.numeric(x))
    if (!is.null(rhopre)) {
        stopifnot(is.numeric(rhopre), length(rhopre) == 1, rhopre <= 1)
    }
    stopifnot(is.numeric(ofac), length(ofac) == 1, is.finite(ofac))
    if (ofac < 1) {
        stop("oversampling factor 'ofac' must be >= 1")
    }
    stopifnot(is.numeric(hifac), length(hifac) == 1, is.finite(hifac))
    if (hifac <= 0) {
        stop("'hifac' must be positive")
    }
    stopifnot(is.numeric(n50), length(n50) == 1, is.finite(n50), n50 >= 1,
              round(n50) == n50)
    stopifnot(is.numeric(nsim), length(nsim) == 1, is.finite(nsim), nsim >= 1,
              round(nsim) == nsim)
    if (length(p) > 0) {
        stopifnot(is.numeric(p) ||
                  (requireVersion("gmp", "0.5-5") && gmp::is.bigq(p)))
        stopifnot(p > 0, p < 1)
    }
    stopifnot(is.numeric(maxTime), length(maxTime) == 1, maxTime >= 0)
    stopifnot(is.numeric(nLimit), length(nLimit) == 1, nLimit >= 0,
              round(nLimit) == nLimit)
    stopifnot(identical(txOrdered, TRUE) || identical(txOrdered, FALSE))
    stopifnot(identical(verbose, TRUE) || identical(verbose, FALSE))
    stopifnot(identical(mctest, TRUE) || identical(mctest, FALSE))
    if (mctest && nsim < NSIM_LIMIT) {
        stop(gettextf("if 'mctest' is TRUE, 'nsim' must be at least %.0f",
                      NSIM_LIMIT, domain = "R-dplR"),
             domain = NA)
    }
    ## dplR: iwin can be a number or a string. iwin2 is a number %in% 0:4
    if (is.numeric(iwin)) {
        if (length(iwin) != 1 || !(iwin %in% 0:4)) {
            stop("numeric 'iwin' must be 0, 1, 2, 3 or 4")
        }
        iwin2 <- iwin
    } else if (is.character(iwin)) {
        iwin2 <- match.arg(tolower(iwin), WIN_NAMES)
        winvec <- 0:4
        names(winvec) <- WIN_NAMES
        iwin2 <- winvec[iwin2]
    } else {
        stop("'iwin' must be numeric or character")
    }
    if (is.double(x)) {
        x2 <- x
    } else {
        x2 <- as.numeric(x)
    }
    np <- as.numeric(length(x2))
    tGiven <- !missing(t)
    if (tGiven) {
        if (is.double(t)) {
            t2 <- t
        } else {
            t2 <- as.numeric(t)
        }
        if (length(t2) != np) {
            stop("lengths of 't' and 'x' must match")
        }
    } else {
        t2 <- as.numeric(seq_len(np))
    }
    naidx <- is.na(x2)
    if (tGiven) {
        naidx <- naidx | is.na(t2)
    }
    if (any(naidx)) {
        goodidx <- which(!naidx)
        t2 <- t2[goodidx]
        x2 <- x2[goodidx]
        nporig <- np
        np <- as.numeric(length(x2))
        nna <- nporig - np
        warning(sprintf(ngettext(nna,
                                 "%.0f NA value removed",
                                 "%.0f NA values removed",
                                 domain = "R-dplR"), nna), domain = NA)
    }
    if (np < MIN_POINTS) {
        stop(gettextf("too few points (%.0f), at least %.0f needed",
                      np, MIN_POINTS, domain = "R-dplR"), domain = NA)
    }
    duplT <- FALSE
    if (tGiven && !txOrdered) {
        idx <- order(t2)
        t2 <- t2[idx]
        x2 <- x2[idx]
        dupl <- duplicated(t2)
        if (any(dupl)) {
            duplT <- TRUE
            if (tTime) {
                warning("Duplicate times in 't', averaging data")
            } else {
                warning("Duplicate ages in 't', averaging data")
            }
            if (verbose) {
                if (tTime) {
                    cat(gettext("Number of duplicates by time,\n",
                                domain = "R-dplR"), file = stderr())
                } else {
                    cat(gettext("Number of duplicates by age,\n",
                                domain = "R-dplR"), file = stderr())
                }
                cat(gettext("'k' duplicates means 'k + 1' total observations:\n",
                            domain = "R-dplR"), file = stderr())
                dtable <- table(t2[dupl])
                if (tTime) {
                    dtable <- data.frame(time = as.numeric(names(dtable)),
                                         duplicates = as.vector(dtable))
                } else {
                    dtable <- data.frame(age = as.numeric(names(dtable)),
                                         duplicates = as.vector(dtable))
                }
                write.table(dtable, row.names = FALSE, file = stderr())
            }
            notdupl <- !dupl
            nunique <- sum(notdupl)
            xnew <- numeric(nunique)
            currentid <- 1
            currentstart <- 1
            for (k in 2:np) {
                if (notdupl[k]) {
                    xnew[currentid] <- mean(x2[currentstart:(k - 1)])
                    currentid <- currentid + 1
                    currentstart <- k
                }
            }
            if (currentid == nunique) {
                xnew[nunique] <- mean(x2[currentstart:np])
            }
            t2 <- t2[notdupl]
            x2 <- xnew
            np <- as.numeric(nunique)
            if (np < MIN_POINTS) {
                stop(gettextf("too few points (%.0f), at least %.0f needed",
                              np, MIN_POINTS, domain = "R-dplR"), domain = NA)
            }
        }
    }
    ## dplR: The rest of the function assumes that t2 is age, not time
    t2NoRev <- t2
    x2NoRev <- x2
    if (tTime) {
        t2 <- -rev(t2)
        x2 <- rev(x2)
    }
    if (tGiven) {
        difft <- diff(t2)
    } else {
        difft <- rep.int(1.0, np)
    }
    ## dplR: Setup
    params <- redfitSetdim(MIN_POINTS, t2, ofac, hifac, n50, verbose,
                           iwin = iwin2, nsim = nsim, mctest = mctest,
                           rhopre = rhopre, p = p)
    avgdt <- params[["avgdt"]]
    nseg <- params[["nseg"]]
    fnyq <- params[["fnyq"]]
    nfreq <- params[["nfreq"]]
    df <- params[["df"]]
    segskip <- params[["segskip"]]
    dn50 <- params[["n50"]]
    freq <- seq(from = 0, to = fnyq, length.out = nfreq)
    tr <- redfitTrig(t2, freq, nseg, dn50, segskip)
    ww <- matrix(NA_real_, nseg, dn50)
    for (i in as.numeric(seq_len(dn50))) {
        twk <- t2[.Call(dplR.seg50, i, nseg, segskip, np)]
        ww[, i] <- redfitWinwgt(twk, iwin2)
    }
    ## determine autospectrum of input data
    lmfitfun <-
        tryCatch(getExportedValue("stats", ".lm.fit"),
                 error = function(...) getExportedValue("stats", "lm.fit"))
    gxx <- .Call(dplR.spectr, t2, x2, np, ww, tr[[1]], tr[[2]], tr[[3]],
                 nseg, nfreq, avgdt, freq, dn50, segskip, lmfitfun)
    ## estimate data variance from autospectrum
    varx <- df * sum(gxx)
    ## dplR: estimate lag-1 autocorrelation coefficient unless prescribed
    if (is.null(rhopre) || rhopre < 0) {
        rho <- redfitGetrho(t2, x2, dn50, nseg, segskip, lmfitfun)
        ## make sure that tau is non-negative
        if (rho > 1) {
            warning(gettext("redfitGetrho returned rho = %f, forced to zero",
                            rho, domain = "R-dplR"),
                    domain = NA)
            rho <- 0
        }
    } else {
        rho <- rhopre
    }
    ## dplR: determine tau from rho.
    ## Avoids the rho -> tau -> rho mess of REDFIT.
    tau <- as.numeric(-avgdt / log(rho))

    ## Generate nsim AR(1) spectra
    if (mctest) {
        grr <- matrix(NA_real_, nfreq, nsim)
        for (i in seq_len(nsim)) {
            if (verbose && (i %% 50 == 0 || i == 1)) {
                cat("ISim = ", i, "\n", sep="")
            }
            ## setup AR(1) time series and estimate its spectrum
            grr[, i] <-
                .Call(dplR.spectr, t2, .Call(dplR.makear1, difft, np, tau), np,
                      ww, tr[[1]], tr[[2]], tr[[3]], nseg, nfreq, avgdt,
                      freq, dn50, segskip, lmfitfun)
            ## scale and sum red-noise spectra
            varr1 <- df * sum(grr[, i])
            grr[, i] <- varx / varr1 * grr[, i]
        }
        grrsum <- .rowSums(grr, nfreq, nsim)
    } else {
        grrsum <- numeric(nfreq)
        for (i in seq_len(nsim)) {
            if (verbose && (i %% 50 == 0 || i == 1)) {
                cat("ISim = ", i, "\n", sep="")
            }
            ## setup AR(1) time series and estimate its spectrum
            grr <- .Call(dplR.spectr, t2, .Call(dplR.makear1, difft, np, tau),
                         np, ww, tr[[1]], tr[[2]], tr[[3]], nseg, nfreq,
                         avgdt, freq, dn50, segskip, lmfitfun)
            ## scale and sum red-noise spectra
            varr1 <- df * sum(grr)
            grr <- varx / varr1 * grr
            grrsum <- grrsum + grr
        }
    }

    ## determine average red-noise spectrum; scale average again to
    ## make sure that roundoff errors do not affect the scaling
    grravg <- grrsum / nsim
    varr2 <- df * sum(grravg)
    grravg <- varx / varr2 * grravg
    rhosq <- rho * rho
    ## set theoretical spectrum (e.g., Mann and Lees, 1996, Eq. 4)
    ## make area equal to that of the input time series
    gredth <- (1 - rhosq) /
        (1 + rhosq - 2 * rho * cos(seq(from = 0, to = pi, length.out = nfreq)))
    varr3 <- df * sum(gredth)
    gredth <- varx / varr3 * gredth
    ## determine correction factor
    corr <- grravg / gredth
    invcorr <- gredth / grravg
    ## correct for bias in autospectrum
    gxxc <- gxx * invcorr

    ## red-noise false-alarm levels from percentiles of MC simulation
    if (mctest) {
        ## dplR: Sort the rows of grr. apply() turns the result
        ## around: the sorted rows are the columns of the result.
        grr <- apply(grr, 1, sort)
        ## set percentile indices
        idx80 <- floor(0.80 * nsim)
        idx90 <- floor(0.90 * nsim)
        idx95 <- floor(0.95 * nsim)
        idx99 <- floor(0.99 * nsim)
        ## find frequency-dependent percentile and apply bias correction
        ci80 <- grr[idx80, ] * invcorr
        ci90 <- grr[idx90, ] * invcorr
        ci95 <- grr[idx95, ] * invcorr
        ci99 <- grr[idx99, ] * invcorr
    } else {
        ci80 <- NULL
        ci90 <- NULL
        ci95 <- NULL
        ci99 <- NULL
    }

    if (iwin2 == 0 && ofac == 1 && dn50 == 1) {
        spectrcomp <- rep.int(0, nfreq)
        spectrcomp[gxxc - gredth >= 0] <- 1
        rcnt <- 1 + sum(diff(spectrcomp) != 0)

        ## dplR: Old formulas for rcritlo, rcrithi (REDFIT).
        ##
        ## ## test equality of theoretical AR1 and estimated spectrum
        ## ## using a runs test (Bendat and Piersol, 1986, p. 95). The
        ## ## empirical equations for calculating critical values for
        ## ## different significanes were derived from the tabulated
        ## ## critical values in B&P.
        ## sqrtHalfNfreq <- sqrt(nfreq %/% 2)
        ## ## REDFIT >= 3.8a (at least until 3.8e)
        ## 10-% level of significance
        ## rcritlo10 <- round((-0.62899892 + 1.0030933 * sqrtHalfNfreq)^2)
        ## rcrithi10 <- round(( 0.66522732 + 0.9944506 * sqrtHalfNfreq)^2)
        ## 5-% level of significance
        ## rcritlo5 <- round((-0.78161838 + 1.0069634 * sqrtHalfNfreq)^2)
        ## rcrithi5 <- round(( 0.75701059 + 0.9956021 * sqrtHalfNfreq)^2)
        ## 2-% level of significance
        ## rcritlo2 <- round((-0.92210867 + 1.0064993 * sqrtHalfNfreq)^2)
        ## rcrithi2 <- round(( 0.82670832 + 1.0014299 * sqrtHalfNfreq)^2)

        ## dplR: Updated formulas for rcritlo, rcrithi.
        ##
        ## The REDFIT formulas seem to be quite inexact.  For example,
        ## the width of the acceptance region increases, then
        ## decreases (*), and goes to <= 0 at nfreq >= 27144 (5 %
        ## significance).  Another example: rcritlo is 0 (impossible
        ## number of runs) at some small values of nfreq.
        ##
        ## (*) The increase and decrease are general trends, but there
        ## are local fluctuations against the trend.
        ##
        ## About the new formulas:
        ##
        ## Exact values were computed for nfreq <= NMAX (possibly
        ## depending on significance levels, see redfitTablecrit() for
        ## up-to-date values) at a selected few significance levels.
        ## For non-tabulated significance levels, the exact solution
        ## is computed if time permits and nfreq is not too large
        ## (maxTime, nLimit), or finally a normal approximation is
        ## used.
        ##
        ## The problem at hand (comparison of two spectra) is
        ## analogous to studying the number of runs of heads and tails
        ## with nfreq tosses of a fair coin (p == 0.5).
        ##
        ## The sequence of acceptance region widths is non-decreasing
        ## for both odd and even nfreq, individually.  However,
        ## because of the symmetric distribution, if rcritlo increases
        ## by 1 going from nfreq to nfreq + 1, rcrithi does not
        ## change, and the change in the width is -1.
        ##
        ## Reference: Bradley, J. V. (1968) Distribution-Free
        ## Statistical Tests. Prentice-Hall. p. 253--254, 259--263.
        tmp <- runcrit(nfreq, p, maxTime, nLimit)
        rcritlo <- tmp[[1]]
        rcrithi <- tmp[[2]]
        rcritexact <- tmp[[3]]
    } else {
        rcnt <- NULL
        rcritlo <- NULL
        rcrithi <- NULL
        rcritexact <- NULL
    }

    ## dplR: Elements of the list returned from this function:
    ##  varx      data variance estimated from spectrum
    ##  rho       average autocorrelation coefficient (estimated or prescribed)
    ##  tau       average tau, tau == -avgdt / log(rho)
    ##  rcnt      runs count, test of equality of theoretical and data spectrum
    ##  rcritlo   critical low value(s) for rcnt, one for each p
    ##  rcrithi   critical high value(s) for rcnt, one for each p
    ##  rcritexact   are the critical values (limits of acceptance region) exact?
    ##  freq      frequency vector
    ##  gxx       autospectrum of input data
    ##  gxxc      corrected autospectrum of input data
    ##  grravg    average AR(1) spectrum
    ##  gredth    theoretical AR(1) spectrum
    ##  corr      correction factor
    ##  ci80      80% false-alarm level from MC
    ##  ci90      90% false-alarm level from MC
    ##  ci95      95% false-alarm level from MC
    ##  ci99      99% false-alarm level from MC
    ##  call      how the function was called
    ##  params    parameters dependent on the command line arguments
    ##  vers      version of dplR containing the function
    ##  seed      if not NULL, value used for set.seed(seed)
    ##  t         t with duplicates removed (times or ages) or NULL
    ##  x         x averaged over duplicate values of t or NULL
    dplrNS <- tryCatch(getNamespace("dplR"), error = function(...) NULL)
    if (!is.null(dplrNS) && exists("redfit", dplrNS) &&
        identical(match.fun(as.list(cl)[[1]]), get("redfit", dplrNS))) {
        vers <- tryCatch(packageVersion("dplR"), error = function(...) NULL)
    } else {
        vers <- NULL
    }
    res <- list(varx = varx, rho = rho, tau = tau, rcnt = rcnt,
                rcritlo = rcritlo, rcrithi = rcrithi, rcritexact = rcritexact,
                freq = freq, gxx = gxx, gxxc = gxxc, grravg = grravg,
                gredth = gredth, corr = corr,
                ci80 = ci80, ci90 = ci90, ci95 = ci95, ci99 = ci99,
                call = cl, params = params, vers = vers, seed = seed,
                t = if (duplT) t2NoRev, x = if (duplT) x2NoRev)
    class(res) <- "redfit"
    res
}

## Determine 6dB bandwidth from OFAC corrected fundamental frequency.
## Note that the bandwidth for the Blackman-Harris taper is higher than
## reported by Harris (1978, cf. Nuttall, 1981)}
##
## window type (iwin)  0: Rectangular
##                     1: Welch 1
##                     2: Hanning
##                     3: Parzen (Triangular)
##                     4: Blackman-Harris 3-Term
redfitWinbw <- function(iwin, df, ofac, nseg) {

    ## dplR: bw has been computed with higher precision.  We also
    ## have results for short windows which shows the aliasing
    ## caused by sampling.  FFT() from package "fftw" and fft()
    ## were used.  Note that the results are for uniformly sampled
    ## windows.  For some reason, the asymptotic (approaching
    ## continuous time) bandwidths of the triangular and
    ## Blackman-Harris (defined by Nuttall) windows slightly
    ## differ from the bandwidths used by REDFIT (below, commented
    ## out): now 1.77 instead of 1.78 and 2.27 instead of 2.26,
    ## respectively.

    ## bw <- c(1.21, 1.59, 2.00, 1.78, 2.26)

    approxX <-
        switch(iwin + 1,
               ## 1 (Rectangular)
               c(2:49, 51, 52, 54, 55, 57, 60, 62, 65, 68, 72, 77,
                 82, 89, 99, 104),
               ## 2 (Welch)
               c(3:84, 86:89, 91, 92, 94, 95, 97, 99, 100, 102,
                 104, 106, 109, 111, 114, 117, 120, 123, 127, 131,
                 135, 140, 146, 152, 155, 489),
               ## 3 (Hanning)
               3,
               ## 4 (Triangular)
               c(3, 5:137, 171:219),
               ## 5 (Blackman-Harris)
               c(3:11, 13, 15, 18, 19))

    approxY <-
        switch(iwin + 1,
               ## 1 (Rectangular), results agree with exact formula
               ## given by Harris
               c(1.33333333333333, 1.258708,  1.2351995, 1.2247266,
                 1.2191411,        1.215808,  1.213658,  1.212190,
                 1.211143,         1.210370,  1.209784,  1.209327,
                 1.208966,         1.208674,  1.208436,  1.208238,
                 1.208073,         1.207933,  1.207813,  1.2077106,
                 1.20762,          1.207544,  1.207476,  1.207416,
                 1.2073622,        1.207315,  1.207272,  1.207234,
                 1.207200,         1.207168,  1.207140,  1.2071144,
                 1.207091,         1.2070694, 1.207050,  1.2070315,
                 1.207015,         1.206999,  1.2069849, 1.20697,
                 1.206959,         1.206948,  1.206937,  1.2069270,
                 1.206918,         1.206909,  1.206901,  1.20689,
                 1.2068788,        1.20687,   1.206860,  1.20685,
                 1.20684,          1.20683,   1.20682,   1.20681,
                 1.20680,          1.20679,   1.20678,   1.20677,
                 1.20676,          1.20675,   1.2067),
               ## 2 (Welch)
               c(2.00000000000000, 1.78680,   1.70821,   1.66954,
                 1.64744,          1.63354,   1.62421,   1.617630,
                 1.61281,          1.60918,   1.60636,   1.60414,
                 1.60236,          1.60090,   1.59969,   1.59869,
                 1.59784,          1.59711,   1.59649,   1.59595,
                 1.59548,          1.59506,   1.59470,   1.59438,
                 1.59409,          1.59383,   1.59360,   1.59339,
                 1.59321,          1.59304,   1.59288,   1.59274,
                 1.592608,         1.592489,  1.59238,   1.59228,
                 1.59219,          1.592099,  1.59202,   1.59194,
                 1.59188,          1.59181,   1.591750,  1.59169,
                 1.59164,          1.59159,   1.59154,   1.59150,
                 1.59146,          1.59142,   1.59138,   1.591349,
                 1.59132,          1.59129,   1.59126,   1.591228,
                 1.59120,          1.59118,   1.59115,   1.59113,
                 1.59111,          1.591087,  1.59107,   1.59105,
                 1.59103,          1.59101,   1.590996,  1.59098,
                 1.590965,         1.590951,  1.59094,   1.59092,
                 1.59091,          1.59090,   1.59089,   1.5908750,
                 1.59086,          1.59085,   1.59084,   1.59083,
                 1.590824,         1.59081,   1.59080,   1.59079,
                 1.59078,          1.59077,   1.59076,   1.59075,
                 1.59074,          1.59073,   1.59072,   1.59071,
                 1.59070,          1.59069,   1.59068,   1.59067,
                 1.59066,          1.59065,   1.59064,   1.59063,
                 1.59062,          1.59061,   1.59060,   1.59059,
                 1.59058,          1.59057,   1.59056,   1.59055,
                 1.5905,           1.5904),
               ## 3 (Hanning)
               2.000,
               ## 4 (Triangular), results agree with exact formula
               ## given by Harris (for even number of points)
               c(2.0000,   1.84975,   1.86328,  1.81083,  1.82157,
                 1.79521,  1.80317,   1.78740,  1.79341,  1.78294,
                 1.78760,  1.78015,   1.78385,  1.77829,  1.78130,
                 1.77699,  1.77948,   1.776045, 1.77814,  1.775335,
                 1.77712,  1.774789,  1.77633,  1.77436,  1.77570,
                 1.77402,  1.77519,   1.77374,  1.774780, 1.77351,
                 1.77444,  1.77332,   1.77415,  1.77316,  1.77391,
                 1.77302,  1.77370,   1.77290,  1.77352,  1.77280,
                 1.77337,  1.77271,   1.77323,  1.772635, 1.77311,
                 1.77257,  1.77301,   1.77251,  1.77292,  1.77245,
                 1.77284,  1.77241,   1.77276,  1.77236,  1.77270,
                 1.77232,  1.772636,  1.77229,  1.77258,  1.77226,
                 1.77253,  1.77223,   1.77249,  1.77220,  1.77245,
                 1.77218,  1.77241,   1.772158, 1.772376, 1.77214,
                 1.77234,  1.77212,   1.77232,  1.77210,  1.77229,
                 1.77209,  1.77226,   1.77207,  1.77224,  1.77206,
                 1.77222,  1.77205,   1.77220,  1.77203,  1.77218,
                 1.77202,  1.77216,   1.77201,  1.77215,  1.77200,
                 1.77213,  1.77199,   1.77212,  1.771985, 1.77210,
                 1.771977, 1.77209,   1.77197,  1.77208,  1.77196,
                 1.77207,  1.77196,   1.77206,  1.77195,  1.77205,
                 1.77194,  1.7720387, 1.77194,  1.77203,  1.77193,
                 1.772021, 1.77193,   1.77201,  1.77192,  1.77201,
                 1.771918, 1.77200,   1.77191,  1.77199,  1.77191,
                 1.771985, 1.77191,   1.77198,  1.77190,  1.77197,
                 1.77190,  1.77197,   1.771895, 1.77196,  1.77189,
                 1.77196,  1.77189,   1.77195,  1.7719,   1.771850,
                 1.77189,  1.77185,   1.77189,  1.771847, 1.77188,
                 1.77185,  1.77188,   1.77184,  1.77188,  1.77184,
                 1.77188,  1.77184,   1.77188,  1.77184,  1.77187,
                 1.77184,  1.77187,   1.77184,  1.77187,  1.771837,
                 1.77187,  1.771836,  1.77187,  1.771835, 1.77187,
                 1.77183,  1.77186,   1.77183,  1.77186,  1.77183,
                 1.77186,  1.77183,   1.77186,  1.77183,  1.77186,
                 1.77183,  1.77186,   1.77183,  1.77186,  1.77183,
                 1.77185,  1.771827,  1.77185,  1.77183,  1.77185,
                 1.77183,  1.77185,   1.7718),
               ## 5 (Blackman-Harris)
               c(1.9860952, 2.271338, 2.267059, 2.267229,
                 2.267416,  2.267511, 2.26755,  2.267572,
                 2.26758,   2.26757,  2.26756,  2.267552,
                 2.2675))

    ## df * ofac * bw[iwin + 1]
    df * ofac * approx(approxX, approxY, nseg,
                       method = "constant", rule = c(1, 2), f = 0)[["y"]]
}

## Effective number of degrees of freedom for the selected window
## and n50 overlapping segments (Harris, 1978).
## dplR: Computed more precise values for c50.
redfitGetdof <- function(iwin, n50) {
    ## dplR: Rectangular, Welch, Hanning, Triangular, Blackman-Harris
    ## c50 <- c(0.5, 0.34375, 1 / 6, 0.25, 0.0955489871755)
    ## c2 <- c50[iwin + 1]^2
    ## dplR: Precomputed squared c50. Note: (1/6)^2 == 1/36
    c2 <- c(0.25, 0.1181640625, 0.0277777777777778,
            0.0625, 0.00912960895026386)[iwin + 1]
    n50 / (0.5 + c2 - c2 / n50)
}

## dplR: print.redfit() is a separate function for printing the
## results of redfit(), with an output format very close to that in
## the original REDFIT.
print.redfit <- function(x, digits = NULL, csv.out = FALSE, do.table = FALSE,
                         prefix = "", row.names = FALSE, file = "", ...) {
    if (!inherits(x, "redfit")) {
        stop('use only with "redfit" objects')
    }
    stopifnot(identical(csv.out, TRUE) || identical(csv.out, FALSE))
    stopifnot(identical(do.table, TRUE) || identical(do.table, FALSE))

    ## dplR: Automatically adds prefix (for example "# " from REDFIT) and
    ## newline (if newline = TRUE) to output.
    precat <- function(..., newline = TRUE, sep = "") {
        cat(prefix)
        do.call("cat", c(alist(...), alist(sep = sep)))
        if (newline) {
            cat("\n")
        }
    }
    params <- x[["params"]]
    iwin <- params[["iwin"]]
    n50 <- params[["n50"]]
    mctest <- params[["mctest"]]
    gredth <- x[["gredth"]]

    ## scaling factors for red noise from chi^2 distribution
    dof <- redfitGetdof(iwin, n50)
    ## dplR: getchi2() in the original Fortran version uses upper tail
    ## probabilities. qchisq() uses lower tail probabilities unless
    ## lower.tail = FALSE.
    fac80 <- qchisq(0.80, dof) / dof
    fac90 <- qchisq(0.90, dof) / dof
    fac95 <- qchisq(0.95, dof) / dof
    fac99 <- qchisq(0.99, dof) / dof

    if (csv.out || do.table) {
        dframe <- c(x[c("freq", "gxx", "gxxc", "gredth", "grravg", "corr")],
                    list(gredth * fac80, gredth * fac90,
                         gredth * fac95, gredth * fac99))
        pct <- c("80", "90", "95", "99")
        names(dframe) <- c("Freq", "Gxx", "Gxx_corr", "Gred_th", "Gred_avg",
                           "CorrFac", paste0("Chi2_", pct, "pct"))
        if (mctest) {
            dframe <- c(dframe, x[paste0("ci", pct)])
            names(dframe)[11:14] <- paste0("MC_", pct, "pct")
        }
        dframe <- as.data.frame(dframe)
    }
    if (!csv.out) {
        ## dplR: print miscellaneous information AND if (do.table) print(dframe)
        nseg <- params[["nseg"]]
        ofac <- params[["ofac"]]
        rhopre <- params[["rhopre"]]

        ## critical false alarm level after Thomson (1990)
        ## dplR: modified from original REDFIT code to accommodate for
        ## lower / upper tail difference
        alphacrit <- (nseg - 1) / nseg
        faccrit <- qchisq(alphacrit, dof) / dof

        precat("redfit()", newline = FALSE)
        vers <- x[["vers"]]
        if (!is.null(vers)) {
            cat(" in dplR version ", as.character(vers), "\n", sep="")
        } else {
            cat("\n")
        }
        precat()
        gtxt <- gettext("Input:", domain = "R-dplR")
        precat(gtxt)
        precat(rep.int("-", nchar(gtxt, type = "width")))
        precat("ofac = ", format(ofac, digits = digits))
        precat("hifac = ", format(params[["hifac"]], digits = digits))
        precat("n50 = ", format(n50, digits = digits))
        precat("iwin = ", format(iwin, digits = digits))
        precat("nsim = ", format(params[["nsim"]], digits = digits))
        precat()
        gtxt <- gettext("Initial values:", domain = "R-dplR")
        precat(gtxt)
        precat(rep.int("-", nchar(gtxt, type = "width")))
        seed <- x[["seed"]]
        if (!is.null(seed)) {
            precat("seed = ", format(seed, digits = digits))
        }
        precat(gettextf("Data variance (from data spectrum) = %s",
                        format(x[["varx"]], digits = digits),
                        domain = "R-dplR"))
        precat(gettextf("Avg. dt = %s",
                        format(params[["avgdt"]], digits = digits),
                        domain = "R-dplR"))
        precat()
        gtxt <- gettext("Results:", domain = "R-dplR")
        precat(gtxt)
        precat(rep.int("-", nchar(gtxt, type = "width")))
        if (is.null(rhopre) || rhopre < 0) {
            precat(gettextf("Avg. autocorr. coeff., rho = %s",
                            format(x[["rho"]], digits = digits),
                            domain = "R-dplR"))
        } else {
            precat(gettextf("PRESCRIBED avg. autocorr. coeff., rho = %s",
                            format(rhopre, digits = digits),
                            domain = "R-dplR"))
        }
        precat(gettextf("Avg. tau = %s",
                        format(x[["tau"]], digits = digits),
                        domain = "R-dplR"))
        precat(gettextf("Degrees of freedom = %s",
                        format(dof, digits = digits),
                        domain = "R-dplR"))
        precat(gettextf("6-dB Bandwidth = %s",
                        format(redfitWinbw(iwin, params[["df"]], ofac, nseg),
                               digits = digits),
                        domain = "R-dplR"))
        precat(gettextf("Critical false-alarm level (Thomson, 1990) = %s",
                        format(alphacrit * 100, digits = digits),
                        domain = "R-dplR"))
        precat(gettextf("   ==> corresponding scaling factor for red noise = %s",
                        format(faccrit, digits = digits),
                        domain = "R-dplR"))
        precat()
        gtxt <- gettext("Equality of theoretical and data spectrum: Runs test",
                        domain = "R-dplR")
        precat(gtxt)
        precat(rep.int("-", nchar(gtxt, type = "width")))
        rcnt <- x[["rcnt"]]
        if (!is.null(rcnt)) {
            runP <- params[["p"]]
            nP <- length(runP)
            if (nP > 0) {
                gtxt <- gettextf("%s-%% acceptance region:",
                                 format(as.numeric(100 * (1 - runP)),
                                        digits = digits),
                                 domain = "R-dplR")
                nC <- nchar(gtxt[1], type = "width")
                rcritlo <- x[["rcritlo"]]
                rcrithi <- x[["rcrithi"]]
            }
            for (k in seq_len(nP)) {
                precat(gtxt[k], newline = FALSE)
                cat(" rcritlo = ", format(rcritlo[k], digits = digits), "\n",
                    sep = "")
                precat(rep.int(" ", nC), newline = FALSE)
                cat(" rcrithi = ", format(rcrithi[k], digits = digits), "\n",
                    sep = "")
                precat()
            }
            precat("r_test = ", format(rcnt, digits = digits))
        } else {
            if (iwin != 0) {
                precat(gettext("Test requires iwin = 0", domain = "R-dplR"))
            }
            if (ofac != 1) {
                precat(gettext("Test requires ofac = 1", domain = "R-dplR"))
            }
            if (n50 != 1) {
                precat(gettext("Test requires n50 = 1", domain = "R-dplR"))
            }
        }
        if (do.table) {
            precat()
            gtxt <- gettext("Data Columns:", domain = "R-dplR")
            precat(gtxt)
            precat(rep.int("-", nchar(gtxt, type = "width")))
            precat(gettext(" 1: Freq = frequency", domain = "R-dplR"))
            precat(gettext(" 2: Gxx = spectrum of input data",
                           domain = "R-dplR"))
            precat(gettext(" 3: Gxx_corr = bias-corrected spectrum of input data",
                           domain = "R-dplR"))
            precat(gettext(" 4: Gred_th = theoretical AR(1) spectrum",
                           domain = "R-dplR"))
            precat(gettext(" 5: Gred_avg = average spectrum of Nsim AR(1) time series (uncorrected)",
                           domain = "R-dplR"))
            precat(gettext(" 6: CorrFac = Gxx / Gxx_corr", domain = "R-dplR"))
            gtxt <-
                gettext("%.0f: Chi2_%.0fpct = %.0f%% false-alarm level (Chi^2)")
            precat(" ", sprintf(gtxt, 7, 80, 80))
            precat(" ", sprintf(gtxt, 8, 90, 90))
            precat(" ", sprintf(gtxt, 9, 95, 95))
            precat(sprintf(gtxt, 10, 99, 99))
            if (mctest) {
                gtxt <-
                    gettext("%.0f: MC_%.0fpct = %.0f%% false-alarm level (MC)")
                precat(sprintf(gtxt, 11, 80, 80))
                precat(sprintf(gtxt, 12, 90, 90))
                precat(sprintf(gtxt, 13, 95, 95))
                precat(sprintf(gtxt, 14, 99, 99))
            }
            print(dframe, digits = digits, row.names = row.names)
        }
    } else { # csv.out
        write.csv(dframe, file = file, row.names = row.names, ...)
    }
    invisible(x)
}

## dplR: Utility function.
redfitRunprobZ <- function(k, n) {
    invhalfpowM1 <- gmp::as.bigz(2)^(n - 1)
    if (k %% 2 == 0) {
        ## even number of runs
        r <- k / 2
        n1 <- seq(from = r, by = 1, to = n - r)
        nn1 <- length(n1)
        halfn1 <- nn1 %/% 2
        if (nn1 %% 2 == 1) {
            probsum <- gmp::chooseZ(n1[halfn1 + 1] - 1, r - 1)
            probsum <- probsum * probsum
        } else {
            probsum <- 0
        }
        lown1 <- n1[seq_len(halfn1)]
        if (length(lown1) > 0) {
            lown2 <- n - lown1
            probsum <- probsum + 2 * sum(gmp::chooseZ(lown1 - 1, r - 1) *
                                         gmp::chooseZ(lown2 - 1, r - 1))
        }
        probsum / invhalfpowM1
    } else if (k == 1) {
        ## one run
        gmp::as.bigq(2)^(1 - n)
    } else {
        ## odd number of runs
        r <- (k - 1) / 2
        n1 <- seq(from = r + 1, by = 1, to = n - r)
        n2 <- n - n1
        probsum <- sum(gmp::chooseZ(n1 - 1, r) * gmp::chooseZ(n2 - 1, r - 1))
        probsum / invhalfpowM1
    }
}

## dplR: Utility function.
redfitRuncsum <- function(n, crit, verbose = FALSE,
                          timelimit = Inf) {
    if (gmp::is.bigq(crit)) {
        halfcrit <- crit / 2
    } else {
        halfcrit <- gmp::as.bigq(crit, 2)
    }
    verbose2 <- isTRUE(verbose)
    nn <- length(n)
    csums <- vector(mode = "list", length = nn)
    for (j in seq_len(nn)) {
        thisn <- n[j]
        if (verbose2) {
            cat("n = ", thisn, " (", j, " / ", nn, ")\n", sep = "")
        }
        halfn <- thisn %/% 2
        oddn <- thisn %% 2
        complength <- halfn + oddn
        tmpcsums <- gmp::as.bigq(rep.int(NA_real_, complength))
        if (oddn == 1) {
            st <- system.time({
                csum <- (gmp::as.bigq(1) -
                         redfitRunprobZ(complength, thisn)) / 2
                tmpcsums[[complength]] <- csum
            })
        } else {
            st <- system.time({
                csum <- gmp::as.bigq(1, 2) - redfitRunprobZ(complength, thisn)
                tmpcsums[[complength]] <- csum
            })
        }
        if (st[3] > timelimit) {
            stop("timelimit exceeded")
        }
        if (csum > halfcrit) {
            finalk <- 1
            for (k in seq(from = complength - 1, by = -1,
                          length.out = max(0, complength - 2))) {
                csum <- csum - redfitRunprobZ(k, thisn)
                tmpcsums[[k]] <- csum
                if (csum <= halfcrit) {
                    finalk <- k
                    break
                }
            }
        } else {
            finalk <- complength
        }
        seqstart <- max(2, finalk)
        seqlength <- complength - seqstart + 1
        ## store n, drop 0 and NA
        csums[[j]] <- c(gmp::as.bigq(thisn),
                        tmpcsums[seq(from = seqstart, by = 1,
                                     length.out = seqlength)])
    }
    if (nn == 1) {
        csums <- csums[[1]]
    }
    csums
}

## dplR: Utility function.
## crit can be bigq or numeric
redfitCsumtocrit <- function(csums, crit, limits = FALSE) {
    if (is.list(csums)) {
        csums2 <- csums
    } else {
        csums2 <- list(csums)
    }
    nn <- length(csums2)
    ## Our own sorting function iSort can handle bigq
    tmp <- iSort(crit, decreasing = TRUE)
    halfcrit <- tmp[[1]] / 2
    ncrit <- length(halfcrit)
    if (limits) {
        lowcrit <- matrix(NA_real_, ncrit, nn)
        highcrit <- matrix(NA_real_, ncrit, nn)
        tmp2 <- as.character(as.numeric(rev(tmp[[1]])))
        rownames(lowcrit) <- tmp2
        rownames(highcrit) <- tmp2
    }
    noZeroCrit <- halfcrit[ncrit] != 0
    res <- matrix(NA_real_, 2 * ncrit, nn)
    rownames(res) <- as.character(as.numeric(c(rev(halfcrit), 1 - halfcrit)))
    for (j in seq_len(nn)) {
        Csums <- csums2[[j]]
        nthis <- length(Csums)
        n <- as.numeric(Csums[[1]])
        complength <- n %/% 2 + n %% 2
        if (complength == 1) {
            res[, j] <- rep(c(1, n), each = ncrit)
            if (limits) {
                lowcrit[, j] <- rep.int(0, ncrit)
                highcrit[, j] <- rep.int(0.5, ncrit)
            }
        } else {
            Csums <- c(Csums[seq(from = nthis, by = -1,
                                 length.out = nthis - 1)],
                       rep.int(NA_real_, complength - nthis),
                       Csums[[1]])
            lowaccept <- rep.int(NA_real_, ncrit)
            lowlow <- rep.int(NA_real_, ncrit)
            lowhigh <- rep.int(NA_real_, ncrit)
            allGood <- nthis == complength
            l <- 1
            for (k in seq_len(ncrit)) {
                thisHalfcrit <- halfcrit[k]
                l <- l - 1 + which.max(Csums[l:complength] <= thisHalfcrit)
                if (Csums[[l]] <= thisHalfcrit) {
                    lowaccept[k] <- complength + 1 - l
                    lowlow[k] <- as.numeric(Csums[[l]])
                    if (l > 1) {
                        lowhigh[k] <- as.numeric(Csums[[l - 1]])
                    } else {
                        lowhigh[k] <- 0.5
                    }
                } else if (allGood || thisHalfcrit == 0) {
                    lowaccept[k] <- 1
                    lowlow[k] <- 0
                    lowhigh[k] <- as.numeric(Csums[[complength - 1]])
                } else if (noZeroCrit) {
                    break
                }
            }
            highaccept <- n + 1 - lowaccept
            res[, j] <- c(rev(lowaccept), highaccept)
            if (limits) {
                lowcrit[, j] <- rev(lowlow)
                highcrit[, j] <- rev(lowhigh)
            }
        }
    }
    ## When limits = TRUE, we also return the limits pmin and pmax such
    ## that res holds when pmin < crit < pmax.
    if (limits) {
        list(drop(res),
             pmin = 2 * apply(lowcrit, 1, max),
             pmax = 2 * apply(highcrit, 1, min))
    } else {
        drop(res)
    }
}

## dplR: Normal approximation of the acceptance region of the number
## of runs test.
## p must be numeric. If limits is TRUE, length(p) must be 1.
redfitNormcrit <- function(n, p, limits = FALSE) {
    ## dplR: Empirical mean 'nMean' and standard deviation 'nSd' of
    ## the number of runs distribution, and code for computing them.
    ## First compute the distribution with runtableZ(), then use
    ## meanvar() to get mean and variance.
    ## library(gmp)
    ## runtableZ <- function(n) {
    ##     stopifnot(is.numeric(n), length(n) == 1,
    ##               is.finite(n), round(n) == n, n > 0)
    ##     halfn <- n %/% 2
    ##     oddn <- n %% 2
    ##     res <- numeric(n)
    ##     invhalfpowM1 <- as.bigz(2)^(n - 1)
    ##     ## Symmetric distribution. Compute first half only.
    ##     complength <- halfn + oddn
    ##     evenlength <- complength %/% 2
    ##     oddlength <- evenlength + complength %% 2 - 1
    ##     ## one run
    ##     res[1] <- 0.5^(n - 1)
    ##     ## odd number of runs (>= 3)
    ##     oddseq <- seq(from = 3, by = 2, length.out = oddlength)
    ##     for (k in oddseq) {
    ##         r <- (k - 1) / 2
    ##         n1 <- seq(from = r + 1, by = 1, to = n - r)
    ##         n2 <- n - n1
    ##         probsum <- sum(chooseZ(n1 - 1, r) * chooseZ(n2 - 1, r - 1))
    ##         res[k] <- as.numeric(probsum / invhalfpowM1)
    ##     }
    ##     ## even number of runs
    ##     evenseq <- seq(from = 2, by = 2, length.out = evenlength)
    ##     for (k in evenseq) {
    ##         r <- k / 2
    ##         n1 <- seq(from = r, by = 1, to = n - r)
    ##         nn1 <- length(n1)
    ##         halfn1 <- nn1 %/% 2
    ##         if (nn1 %% 2 == 1) {
    ##             probsum <- chooseZ(n1[halfn1 + 1] - 1, r - 1)
    ##             probsum <- probsum * probsum
    ##         } else {
    ##             probsum <- 0
    ##         }
    ##         leftn1 <- n1[seq_len(halfn1)]
    ##         if (length(leftn1) > 0) {
    ##             leftn2 <- n - leftn1
    ##             probsum <- probsum + 2 * sum(chooseZ(leftn1 - 1, r - 1) *
    ##                                          chooseZ(leftn2 - 1, r - 1))
    ##         }
    ##         res[k] <- as.numeric(probsum / invhalfpowM1)
    ##     }
    ##     ## Last half is mirror image of first half
    ##     res[seq(from = n, by = -1, length.out = halfn)] <-
    ##         res[seq_len(halfn)]
    ##     res / sum(res)
    ## }
    ## meanvar <- function(n, crit = 0.05) {
    ##     nn <- length(n)
    ##     res <- matrix(NA_real_, 2, nn)
    ##     for (k in seq_len(nn)) {
    ##         thisn <- n[k]
    ##         rtable <- runtableZ(thisn)
    ##         nseq <- 1:thisn
    ##         res[1, k] <- sum(rtable * nseq)
    ##         res[2, k] <- sum(rtable * (nseq - res[1, k])^2)
    ##     }
    ##     drop(res)
    ## }
    nMean <- 0.5 * n + 0.5
    nSd <- sqrt(0.25 * n - 0.25)
    halfP <- p / 2
    rcritlo <- floor(qnorm(halfP, mean = nMean, sd = nSd) + 0.5)
    lowGood <- rcritlo >= 1
    rcritlo <- pmax(1, rcritlo) # truncation
    rcrithi <- n + 1 - rcritlo  # symmetry
    ## When limits = TRUE, we also return the limits pmin and pmax such
    ## that rcritlo and rcrithi hold when pmin < p < pmax.
    if (isTRUE(limits)) {
        lowlow <- numeric(length(n))
        lowlow[!lowGood] <- 0
        lowlow[lowGood] <- pnorm(rcritlo[lowGood] - 0.5,
                                 mean = nMean[lowGood], sd = nSd[lowGood])
        lowhigh <- pnorm(rcritlo + 0.5, mean = nMean, sd = nSd)
        list(rbind(rcritlo, rcrithi, deparse.level=0),
             pmin = 2 * max(lowlow),
             pmax = 2 * min(lowhigh))
    } else {
        rbind(rcritlo, rcrithi, deparse.level=0)
    }
}

## Data needed for doing the correction in redfitTablecrit().  One
## list element for each correction table (range of p).  Mandatory
## elements in each list: method, nMax, pMin, pMax.  Correction method
## 1 needs "diffIdx" which contains the values of n (in increasing
## order!  e.g. which) where the approximation differs from the exact
## acceptance region.  The normal approximation seems to get better
## with increasing p.
##
## Functions redfitCsumtocrit(redfitRuncsum(...), limits = TRUE) and
## redfitNormcrit(..., limits = TRUE) can be used for obtaining exact
## and approximated acceptance regions ('rcritlo', 'rcrithi' in
## redfit()).  The correction tables can then easily be created by
## finding the positions where the approximation differs from the
## exact solution (and by how much).
##
runPrecomp <-
    structure(list(list(method = 1,
                        nMax = 20000,
                        pMin = 0.00999999574400225,
                        pMax = 0.0100000131426446,
                        diffIdx =

                        c(9, 27, 35, 40, 45, 50, 62, 74, 81, 88, 111, 120,
                          156, 176, 231, 255, 307, 320, 363, 378, 425, 457,
                          474, 491, 544, 658, 698, 719, 740, 761, 805, 849,
                          872, 895, 918, 966, 1116, 1195, 1222, 1305, 1479,
                          1539, 1570, 1632, 1663, 1727, 1892, 1960, 1995,
                          2030, 2065, 2100, 2172, 2394, 2628, 2708, 2872,
                          2956, 3214, 3484, 3576, 3764, 3860, 4105, 4205,
                          4409, 4460, 4617, 4670, 4723, 4831, 4939, 4994,
                          5049, 5104, 5216, 5328, 5385, 5442, 5558, 5674,
                          5733, 5851, 5910, 6030, 6091, 6274, 6335, 6522,
                          6585, 6648, 6839, 6904, 7099, 7164, 7296, 7363,
                          7497, 7564, 7700, 7836, 7905, 7974, 8114, 8254,
                          8325, 8396, 8467, 8611, 8755, 8828, 8901, 9122,
                          9197, 9497, 9649, 10034, 10190, 10269, 10506,
                          10666, 11152, 11650, 11818, 12158, 12330, 12852,
                          13386, 13566, 13657, 13748, 13839, 13930, 14114,
                          14579, 14767, 14862, 15052, 15147, 15339, 15921,
                          16216, 16315, 16614, 17220, 17527, 17630, 17733,
                          17941, 18149, 18254, 18359, 18464, 18676, 19318,
                          19643, 19752, 19861)

                        ),
                   list(method = 1,
                        nMax = 20000,
                        pMin = 0.0199999755461175,
                        pMax = 0.0200000250214551,
                        diffIdx =

                        c(8, 20, 28, 43, 55, 83, 99, 108, 117, 126, 146,
                          157, 168, 179, 268, 343, 392, 445, 540, 689, 712,
                          735, 758, 806, 831, 856, 881, 933, 1041, 1097,
                          1184, 1244, 1368, 1432, 1465, 1498, 1634, 1669,
                          1704, 1776, 1924, 2000, 2117, 2197, 2361, 2445,
                          2488, 2531, 2574, 2662, 2707, 2797, 3124, 3369,
                          3520, 3675, 3940, 4327, 4441, 4498, 4614, 4673,
                          4732, 4791, 4911, 5155, 5279, 5468, 5596, 5856,
                          5988, 6055, 6122, 6394, 6463, 6532, 6672, 6956,
                          7100, 7319, 7767, 7919, 8073, 8150, 8306, 8543,
                          9108, 9523, 9691, 9776, 9861, 10033, 10468, 11002,
                          11093, 11366, 11550, 11829, 12589, 12880, 13472,
                          13672, 13773, 13874, 13975, 14179, 14282, 14385,
                          14488, 14696, 15221, 15756, 16521, 16854, 17078,
                          17530, 17873, 18220, 19041, 19160, 19279)

                        ),
                   list(method = 1,
                        nMax = 20000,
                        pMin = 0.0499999631908032,
                        pMax = 0.050000015182738,
                        diffIdx =

                        c(18, 45, 68, 95, 139, 191, 268, 302, 397, 439, 552,
                          652, 847, 1002, 1101, 1709, 1838, 2063, 2110,
                          2157, 2763, 2926, 3325, 3504, 3626, 3750, 3876,
                          4004, 4134, 4333, 4816, 4887, 5031, 5550, 6095,
                          6500, 6749, 7613, 8624, 8719, 9202, 10414, 10941,
                          11048, 11591, 13415, 13772, 14500, 14623, 14871,
                          15627, 15883, 16012, 16141, 16796, 17195, 18007,
                          18559, 19119, 19975)

                        ),
                   list(method = 1,
                        nMax = 20000,
                        pMin = 0.0999999517957865,
                        pMax = 0.100000038090771,
                        diffIdx =

                        c(31, 38, 180, 214, 507, 1422, 1761, 2136, 2250,
                          3337, 4154, 4555, 5593, 6638, 7040, 7454, 8207,
                          8881, 8996, 9228, 10809, 11712, 12379, 12651,
                          13344, 13485, 14200, 15686, 15992, 16928)

                        )),
              methodDescriptions =
              "if n one of diffIdx: lower limit + 1, upper limit - 1")

## dplR: Logic of handling precomputed acceptance regions of the
## number of runs test.  Accesses correction data stored in
## runPrecomp.
##
## p can be bigq or numeric
redfitTablecrit <- function(n, p) {
    nP <- length(p)
    res <- matrix(NA_real_, 2, nP)
    ## Maximum n is for each correction table.
    tableN <- vapply(runPrecomp, function (x) x[["nMax"]], numeric(1))
    numP <- as.numeric(p)
    normCrit <- redfitNormcrit(n, numP)
    if (n > max(tableN)) {
        ## Early termination if n is too large
        return(list(res, normCrit))
    }
    ## The i:th correction table is valid for tableLowP[i] < p < tableHighP[i]
    tableLowP <- vapply(runPrecomp, function (x) x[["pMin"]], numeric(1))
    tableHighP <- vapply(runPrecomp, function (x) x[["pMax"]], numeric(1))
    ## Which correction method to use in case of each range of p.
    ## Currently there is only one method.
    tableMethod <- vapply(runPrecomp, function (x) x[["method"]], numeric(1))
    matchP <- rep.int(NA, nP)
    for (k in seq_len(nP)) {
        thisP <- numP[k]
        lowMatch <- tableLowP < thisP
        highMatch <- tableHighP > thisP
        bothMatch <- which(lowMatch & highMatch)
        if (length(bothMatch) > 0) {
            matchP[k] <- bothMatch[1]
        }
    }
    havePN <- !is.na(matchP) & n <= tableN[matchP]
    for (k in which(havePN)) {
        tPos <- matchP[k]
        res[, k] <-
            switch(tableMethod[tPos],
                   one = {
                       out <- normCrit[, k]
                       idx <- runPrecomp[[tPos]][["diffIdx"]]
                       whichN <- findInterval(n, idx)
                       if (whichN != 0 && idx[whichN] == n) {
                           out[1] <- out[1] + 1
                           out[2] <- out[2] - 1
                       }
                       out
                   },
                   c(NA_real_, NA_real_))
    }
    list(res, normCrit)
}

## dplR: Insertion sort.  Returns sorted vector and order.  Good for
## sorting any vector where the elements can be compared with `<` and
## `>`.  NAs are not allowed.
iSort <- function(x, decreasing = FALSE) {
    n <- length(x)
    ord <- seq_len(n)
    res <- x
    for (k in seq(from = 2, by = 1, length.out = n - 1)) {
        l <- k - 1
        thisVal <- res[[k]]
        thisOrd <- ord[k]
        if (isTRUE(decreasing)) {
            while (l >= 1 && thisVal > res[[l]]) {
                res[[l + 1]] <- res[[l]]
                ord[[l + 1]] <- ord[[l]]
                l <- l - 1
            }
        } else {
            while (l >= 1 && thisVal <  res[[l]]) {
                res[[l + 1]] <- res[[l]]
                ord[[l + 1]] <- ord[[l]]
                l <- l - 1
            }
        }
        res[[l + 1]] <- thisVal
        ord[[l + 1]] <- thisOrd
    }
    list(res, ord)
}

## dplR: Compute exact acceptance regions of the number of runs test.
## p can be bigq or numeric
redfitCompcrit <- function(n, p, maxTime) {
    minP <- min(p)
    nP <- length(p)
    csums <- tryCatch(redfitRuncsum(n, minP, FALSE, maxTime),
                      error = function(...) NULL)
    res <- matrix(NA_real_, 2, nP)
    if (!is.null(csums)) {
        tmp <- redfitCsumtocrit(csums, p)
        ## Our own sorting function iSort can handle bigq
        orderP <- iSort(p)[[2]]
        res[1, orderP] <- tmp[1:nP]
        res[2, orderP] <- tmp[(2 * nP):(nP + 1)]
    }
    res
}

## Fetch from table, compute exactly or approximate the acceptance
## region of the number of runs test (p == 0.5).  Exported function.
runcrit <- function(n, p = c(0.10, 0.05, 0.02), maxTime = 10, nLimit = 10000) {
    if (length(p) > 0) {
        stopifnot(is.numeric(p) ||
                  (requireVersion("gmp", "0.5-5") && gmp::is.bigq(p)))
        stopifnot(p > 0, p < 1)
    }
    stopifnot(is.numeric(n), length(n) == 1, n >= 1, round(n) == n)
    stopifnot(is.numeric(maxTime), length(maxTime) == 1, maxTime >= 0)
    stopifnot(is.numeric(nLimit), length(nLimit) == 1, nLimit >= 0,
              round(nLimit) == nLimit)
    ## Exact solution from precomputed tables (some values of p, n <=
    ## maxN, where maxN may depend on p)
    rcritlohi <- redfitTablecrit(n, p)
    normCrit <- rcritlohi[[2]]
    rcritlohi <- rcritlohi[[1]]
    colnames(rcritlohi) <- as.character(p)
    todo <- colSums(is.na(rcritlohi)) > 0
    if (any(todo)) {
        ## Exact solution by computation (n small enough)
        if (n < nLimit && requireVersion("gmp", "0.5-5")) {
            normCritMin <- normCrit[, which.min(p)]
            ## time allowed per "iteration"
            maxTime2 <- maxTime /
                ((normCritMin[2] - normCritMin[1] + 1) %/% 2 + n %% 2)
            rcritlohi[, todo] <- redfitCompcrit(n, p[todo], maxTime2)
            todo[todo] <- colSums(is.na(rcritlohi[, todo, drop = FALSE])) > 0
        }
        if (any(todo)) {
            ## Normal approximation
            rcritlohi[, todo] <- normCrit[, todo]
        }
    }
    list(rcritlo = rcritlohi[1, ], rcrithi = rcritlohi[2, ],
         rcritexact = !todo)
}

redfitSetdim <- function(min.nseg, t, ofac, hifac, n50, verbose, ...) {
    np <- length(t)
    ## dplR: Formula for nseg from the original Fortran version:
    ## Integer division (or truncation, or "floor").
    ## nseg <- (2 * np) %/% (n50 + 1)
    ## New version: rounding instead of truncation, order of operations changed.
    nseg <- round(np / (n50 + 1) * 2)       # points per segment
    if (nseg < min.nseg) {
        stop(gettextf("too few points per segment (%.0f), at least %.0f needed",
                      nseg, min.nseg, domain = "R-dplR"), domain = NA)
    }
    if (n50 == 1) {
        segskip <- 0
    } else {
        ## dplR: (ideal, not rounded) difference between starting indices of
        ## consecutive segments
        segskip <- (np - nseg) / (n50 - 1)
        if (segskip < 1) {
            stop("too many segments: overlap of more than nseg - 1 points")
        }
    }
    ## dplR: It seems that avgdt, fnyq, etc. were somewhat off in the
    ## original Fortran version because it would not use all of the
    ## data (t[np]) with some combinations of np and n50.
    avgdt <- (t[np] - t[1]) / (np - 1)          # avg. sampling interval
    fnyq <- hifac / (2 * avgdt)                 # average Nyquist freq.
    nfreq <- floor(hifac * ofac * nseg / 2 + 1) # f[1] == f0; f[nfreq] == fNyq
    df <- fnyq / (nfreq - 1)                    # freq. spacing
    if (verbose) {
        cat("    N = ", np, "\n", sep="")
        cat(" t[1] = ", t[1], "\n", sep="")
        cat(" t[N] = ", t[np], "\n", sep="")
        cat(" <dt> = ", avgdt, "\n", sep="")
        cat("Nfreq = ", nfreq, "\n", sep="")
        cat("\n")
    }
    ## dplR: ditched nout (nout == nfreq)
    res <- list(np = np, nseg = nseg, nfreq = nfreq, avgdt = avgdt, df = df,
                fnyq = fnyq, n50 = n50, ofac = ofac, hifac = hifac,
                segskip = segskip)
    args <- list(...)
    argnames <- names(args)
    for (k in which(nzchar(argnames))) {
        res[[argnames[k]]] <- args[[k]]
    }
    ## dplR: Convert integers (if any) to numeric
    for (k in seq_along(res)) {
        elem <- res[[k]]
        if (is.integer(elem)) {
            res[[k]] <- as.numeric(elem)
        }
    }
    res
}

redfitTrig <- function(t, freq, nseg, n50, segskip) {
    np <- as.numeric(length(t))
    tol1 <- 1.0e-4
    nfreqM1 <- length(freq) - 1
    tsin <- array(NA_real_, c(nseg, nfreqM1, n50))
    tcos <- array(NA_real_, c(nseg, nfreqM1, n50))
    wtau <- matrix(NA_real_, nfreqM1, n50)
    wfac <- 2 * pi # omega == 2*pi*f
    truncFreq <- trunc(freq)
    ## start segment loop
    for (j in as.numeric(seq_len(n50))) {
        tsamp <- t[.Call(dplR.seg50, j, nseg, segskip, np)]
        absTsamp <- abs(tsamp)
        tmpL <- trunc(absTsamp)
        tmpY <- absTsamp - tmpL
        ## start frequency loop
        ## dplR: In the original Fortran code, the variables ww (not used
        ## in this function), wtau, tsin and tcos have unused elements
        ## (one extra frequency).  The unused elements have now been
        ## dropped.
        for (k in seq_len(nfreqM1)) {
            ## calc. tau
            ## wrun <- wfac * freq[k + 1]
            ## arg2 <- wrun * tsamp
            ## dplR: Reorganized computation of arg2, range nicely limited.
            ## Change inspired by a note in the comments of REDFIT 3.8e:
            ## "bugfix: rare crahes due to overflow of trigonometric
            ##  arguments in subroutine TRIG (happens for "old"
            ##  high-resolution data, i.e. if t*omega gets larger than
            ##  approx 8e5). Workaround: change variable arg in SR
            ##  TRIG to double precision"
            tmpK <- truncFreq[k + 1]
            tmpX <- freq[k + 1] - tmpK
            tmpKY <- tmpK * tmpY
            tmpLX <- tmpL * tmpX
            tmpXY <- tmpX * tmpY
            arg2 <- (tmpKY - trunc(tmpKY) + (tmpLX - trunc(tmpLX)) + tmpXY) *
                wfac

            arg1 <- arg2 + arg2
            tc <- cos(arg1)
            ts <- sin(arg1)
            csum <- sum(tc)
            ssum <- sum(ts)
            sumtc <- sum(tsamp * tc)
            sumts <- sum(tsamp * ts)
            if (abs(ssum) > tol1 || abs(csum) > tol1) {
                watan <- atan2(ssum, csum)
            } else {
                watan <- atan2(-sumtc, sumts)
            }
            wtnew <- 0.5 * watan
            wtau[k, j] <- wtnew
            ## summations over the sample
            ## dplR: Summations can be found above, but these are not...
            arg2 <- arg2 - wtnew
            tcos[, k, j] <- cos(arg2)
            tsin[, k, j] <- sin(arg2)
        }
    }
    list(tsin = tsin, tcos = tcos, wtau = wtau)
}

## calc. normalized window weights
## window type (iwin)  0: Rectangular
##                     1: Welch 1
##                     2: Hanning
##                     3: Parzen (Triangular)
##                     4: Blackman-Harris 3-Term
## dplR: Fixed the Welch, Hann(ing), Triangular and Blacman-Harris (by
## Nuttall) windows to be DFT-even. The old definitions have been
## commented out.
redfitWinwgt <- function(t, iwin) {
    nseg <- length(t)
    ## useful factor for various windows
    ## fac1 <- nseg / 2 - 0.5
    ## fac2 <- 1 / (fac1 + 1)
    tlen <- t[nseg] - t[1]
    tlenFull <- nseg * tlen / (nseg - 1)
    tPeak <- t[nseg] - tlenFull / 2
    if (iwin == 0) {        # rectangle
        ww <- rep.int(1, nseg)
    } else if (iwin == 1) { # welch I
        ## ww <- (nseg / tlen * (t - t[1]) - fac1) * fac2
        ww <- abs(t - tPeak) / (t[nseg] - tPeak)
        ww <- 1 - ww * ww
    } else if (iwin == 2) { # hanning
        ## fac3 <- nseg - 1
        ## ww <- 1 - cos(2 * pi / fac3 * nseg / tlen * (t - t[1]))
        ww <- 1 - cos(2 * pi / nseg * (1 + (nseg - 1) / tlen * (t - t[1])))
    } else if (iwin == 3) { # triangular
        ## ww <- 1 - abs((nseg / tlen * (t - t[1]) - fac1) * fac2)
        ww <- 1 - abs(t - tPeak) / (t[nseg] - tPeak)
    } else {                # blackman-harris
        ## fac4 <- 2 * pi / (nseg - 1)
        ## jeff <- nseg / tlen * (t - t[1])
        fac4 <- 2 * pi / nseg
        jeff <- 1 + (nseg - 1) / tlen * (t - t[1])
        ww <- 0.4243801 - 0.4973406 * cos(fac4 * jeff) +
            0.0782793 * cos(fac4 * 2.0 * jeff)
    }
    ## determine scaling factor and scale window weights
    if (iwin != 0) {
        ww <- ww * sqrt(nseg / sum(ww * ww))
    }
    ww
}

## dplR: was gettau, converted to return rho only
redfitGetrho <- function(t, x, n50, nseg, segskip, lmfitfun) {
    np <- as.numeric(length(x))
    nseg2 <- as.numeric(nseg)
    segskip2 <- as.numeric(segskip)
    rhovec <- numeric(n50)
    twkM <- matrix(1, nseg2, 2)
    for (i in as.numeric(seq_len(n50))) {
        ## copy data of (i+1)'th segment into workspace
        iseg <- .Call(dplR.seg50, i, nseg2, segskip2, np)
        twk <- t[iseg]
        twkM[, 2] <- twk
        xwk <- x[iseg]
        ## detrend data
        xwk <- do.call(lmfitfun, list(twkM, xwk))[["residuals"]]
        ## estimate and sum rho for each segment
        rho <- redfitTauest(twk, xwk)
        ## bias correction for rho (Kendall & Stuart, 1967; Vol. 3))
        rhovec[i] <- (rho * (nseg2 - 1) + 1) / (nseg2 - 4)
    }
    ## average rho
    mean(rhovec)
}

## dplR: R version based on Mudelsee's code.
## dplR: Introduction copied from REDFIT (some variables removed).
##
## Manfred Mudelsee's code for tau estimation
## ----------------------------------------------------------------------
##  TAUEST: Routine for persistence estimation for unevenly spaced time series
## ----------------------------------------------------------------------
##        Main variables
##
##        t       :       time
##        x       :       time series value
##        np      :       number of points
##       dt       :       average spacing
##    scalt       :       scaling factor (time)
##      rho       :       in the case of equidistance, rho = autocorr. coeff.
##     mult       :       flag (multiple solution)
##     amin       :       estimated value of a = exp(-scalt/tau)
redfitTauest <- function(t, x) {
    np <- length(t)
    ## Correct time direction; assume that ages are input
    ## dplR: Correction of time direction is done by modifying this
    ## function and redfitMinls, not by explicitly reversing (and
    ## multiplying by one)
    ## tscal <- -rev(t)
    ## xscal <- rev(x)
    ## Scaling of x
    xscal <- x / sd(x)
    ## Scaling of t (=> start value of a = 1/e)
    dt <- (t[np] - t[1]) / (np - 1)
    ## dplR: rhoest() of REDFIT is now an "inline function" of two
    ## lines + comment line:
    ## Autocorrelation coefficient estimation (equidistant data)
    xscalMNP <- xscal[-np]
    rho <- sum(xscalMNP * xscal[-1]) / sum(xscalMNP * xscalMNP)
    if (rho <= 0) {
        rho <- 0.05
        warning("rho estimation: <= 0")
    } else if (rho > 1) {
        rho <- 0.95
        warning("rho estimation: > 1")
    }
    scalt <- -log(rho) / dt
    tscal <- t * scalt
    ## Estimation
    minRes <- redfitMinls(tscal, xscal)
    amin <- minRes[["amin"]]
    mult <- minRes[["nmu"]]
    warnings <- FALSE
    if (mult) {
        warning("estimation problem: LS function has > 1 minima")
        warnings <- TRUE
    }
    if (amin <= 0) {
        warning("estimation problem: a_min =< 0")
        warnings <- TRUE
    } else if (amin >= 1) {
        warning("estimation problem: a_min >= 1")
        warnings <- TRUE
    }
    if (!warnings) {
        ## determine tau
        tau <- -1 / (scalt * log(amin))
        ## determine rho, corresponding to tau
        exp(-dt / tau)
    } else {
        ## dplR: fail early
        stop("error in tau estimation")
    }
}

## dplR: Minimization of the built-in least-squares function lsfun
redfitMinls <- function(t, x) {
    ## Least-squares function
    lsfun <- function(a, difft, xM1, xMNP) {
        if (a > 0) {
            tmp <- xMNP - xM1 * a^difft
        } else if (a < 0) {
            tmp <- xMNP + xM1 * (-a)^difft
        } else {
            tmp <- xMNP
        }
        sum(tmp * tmp)
    }
    a_ar1 <- exp(-1) # 1 / e
    tol   <- 3e-8    # Brent's search, precision
    tol2  <- 1e-6    # multiple solutions, precision
    difft <- diff(t)
    np <- length(x)
    xM1 <- x[-1]
    xMNP <- x[-np]
    opt1 <- optimize(lsfun, c(-2, 2),     tol = tol, difft = difft,
                     xM1 = xM1, xMNP = xMNP)
    opt2 <- optimize(lsfun, c(a_ar1, 2),  tol = tol, difft = difft,
                     xM1 = xM1, xMNP = xMNP)
    opt3 <- optimize(lsfun, c(-2, a_ar1), tol = tol, difft = difft,
                     xM1 = xM1, xMNP = xMNP)
    a_ar11 <- opt1[["minimum"]]
    a_ar12 <- opt2[["minimum"]]
    a_ar13 <- opt3[["minimum"]]
    dum1 <- opt1[["objective"]]
    dum2 <- opt2[["objective"]]
    dum3 <- opt3[["objective"]]
    list(amin = c(a_ar11, a_ar12, a_ar13)[which.min(c(dum1, dum2, dum3))],
         nmu = ((abs(a_ar12 - a_ar11) > tol2 && abs(a_ar12 - a_ar1) > tol2) ||
                (abs(a_ar13 - a_ar11) > tol2 && abs(a_ar13 - a_ar1) > tol2)))
}

`ads` <- function(y,nyrs0=50,pos.slope=TRUE){
  
  # set up
  nobs <- length(y)
  nyrs <- 1:nobs + nyrs0 - 1
  # some error checks
  if (nobs < 3) {
    stop("there must be at least 3 data points")
  }
  if (nobs > 1e4) {
    stop("y shouldn't be longer than 1e4. ask for help. the f77 code will need to be recompiled. ")
  }
  if(!is.numeric(nyrs0) || length(nyrs0) != 1 || nyrs0 <= 1){
    stop("'nyrs0' must be an integer greater than 1")
  }
  # call the c that calls the f
  ySpl <-.Call(dplR.c_ads_f,
               y = as.double(y),
               n=as.integer(nobs),
               stiffness = as.integer(nyrs),
               res = as.double(rep(0,nobs)))
  if(length(ySpl)==1 && ySpl == 9999){
    stop("SBR matrix not positive definite")
  }
  
  # If the user wants to constrain a positive slope at the end of the series
  # But what do you do if all the diffs are positive? 
  # E.g. increasing slope for
  # whole spline? use first value?
  if(!pos.slope){
    # 1. Calc first diff of the existing spline (ySpl)
    ySplDiff <- c(0,diff(ySpl))
    # 2. Determine the last index where the spline changes slope
    ySplCutoff <- max(which(ySplDiff <= 0))
    # 3. Set the spline values from the cutoff to the end
    #    as the value at cuttoff.
    ySpl[ySplCutoff:nobs] <- ySpl[ySplCutoff]
    # 4. Rerun the ads on the splined data
    ySpl <-.Call(dplR.c_ads_f,
                 y = as.double(ySpl),
                 n=as.integer(nobs),
                 stiffness = as.integer(nyrs),
                 res = as.double(rep(0,nobs)))
  }
  return(ySpl)
}

# https://github.com/OpenDendro/dplR/pull/25
universalPOWT<-function (rwl, rescale = FALSE, return.power=FALSE)  
{
  # add a check for negative nums
  if(any(rwl <0, na.rm = TRUE)) {
    stop("'rwl' values cannot be negative")
  }
  
  if (!is.data.frame(rwl)) 
    stop("'rwl' must be a data.frame")
  getprec <- function(rwl) {
    rwl.num <- as.numeric(as.matrix(rwl))
    rwl.num <- rwl.num[!is.na(rwl.num) & rwl.num != 0]
    if (length(rwl.num) == 0) {
      NA_real_
    }
    else {
      rwl.char <- format(rwl.num, scientific = FALSE)
      if (grepl(".", rwl.char[1], fixed = TRUE)) {
        maxdig <- nchar(sub("^[^.]*\\.", "", rwl.char[1]))
      }
      else {
        rm.trail.zeros <- sub("0*$", "", rwl.char)
        n.trail.zeros <- nchar(rwl.char) - nchar(rm.trail.zeros)
        maxdig <- -min(n.trail.zeros)
      }
      10^-maxdig
    }
  }
  fit.lm <- function(rwl) {
    runn.M<-matrix(NA,ncol=ncol(rwl),nrow=nrow(rwl)-1)
    runn.S<-matrix(NA,ncol=ncol(rwl),nrow=nrow(rwl)-1)
    for(i in 1:ncol(rwl)){
      n <- length(rwl[,i])
      drop.1 <- rwl[,i][-1]
      drop.n <- rwl[,i][-n]
      runn.M[,i] <- (drop.1 + drop.n)/2
      runn.S[,i] <- abs(drop.1 - drop.n)
    }
    
    prec <- getprec(rwl)
    # AGB -- note typo corrected that let 0 get passed to log resulting in Inf
    #runn.S[runn.M == 0] <- prec
    runn.M[runn.M == 0] <- prec
    runn.S[runn.S == 0] <- prec
    
    
    df<-data.frame(run.M=c(log(runn.M)),
                   run.S=c(log(runn.S)),
                   year=rownames(rwl)[-1], 
                   ID=rep(colnames(rwl),each=nrow(rwl)-1))
    
    lmm<-lmer(run.S~run.M+(1|year),df,REML=FALSE)
    b <- fixef(lmm)[2]
    list(intercept=fixef(lmm)[1],power=1 - b,model=lmm,mean=ranef(lmm)$ID)
  }
  transf <- function(x,p) {
    Xt <- x
    X.nna <- which(!is.na(x))
    X <- na.omit(x)
    if(p <= 0){
      warning("Power estimate is <= 0 (p = ", round(p, 3), "). Using log transformation as a fallback.",
              call.=FALSE)
      X2 <- log(X)
    } else {
      X2 <- X^p
    }
    Xt[X.nna] <- X2
    Xt
  }
  
  p<-fit.lm(rwl)
  xt <-lapply(rwl, function(x)transf(x,p$power))
  
  if (rescale) {
    sds <- lapply(rwl, FUN = function(x) sd(x, na.rm = TRUE))
    means <- lapply(rwl, FUN = function(x) mean(x, na.rm = TRUE))
    rescale_ <- function(x, .sd, .mean) {
      scale(x) * .sd + .mean
    }
    xt <- mapply(rescale_, xt, sds, means)
  }
  
  res <- data.frame(xt, row.names = row.names(rwl), check.names = FALSE)
  class(res) <- c("rwl", "data.frame")
  
  if(return.power==TRUE){ #optional output of the power coefficient, default is FALSE
    names(p$power) <- NULL
    res <- list(transformed.data=res,power=p$power)
  }
  
  return(res)
}

### Helpers for vignettes.

## Return a string representing the given date(s) (default: current date)
## in the format used by \today in LaTeX.
## Example: latexDate("2013-12-06") returns "December 6, 2013"
latexDate <- function(x = Sys.Date(), ...) {
    ltDate <- as.POSIXlt(x, ...)
    sprintf("%s %d, %d",
            month.name[ltDate$mon + 1],
            ltDate$mday,
            1900 + ltDate$year)
}

## Usage: \Sexpr{latexify(string_produced_by_R_code)}
##
## It seems that Sweave needs doublebackslash = TRUE
## but knitr needs doublebackslash = FALSE.
latexify <- function(x, doublebackslash = TRUE, dashdash = TRUE,
                     quotes = c("straight", "curved"),
                     packages = c("fontenc", "textcomp")) {
    y <- as.character(x)
    ## Kludge for converting from "byte" to the current encoding
    ## in a way which preserves the hex notation.
    encBytes <- Encoding(y) == "bytes"
    if (any(encBytes)) {
        y[encBytes] <- captureOutput(cat(y[encBytes], sep = "\n"))
    }
    ## Convert strings to UTF-8 encoding, NFD (decomposed) form, for
    ## processing of accented characters. Doing this early to
    ## circumvent pecularities in gsub() (and nchar()) when working in
    ## the C locale.
    y <- stri_trans_nfd(y)
    Letters <- paste0(c(LETTERS, letters), collapse="")
    fontenc <- "fontenc" %in% packages
    textcomp <- "textcomp" %in% packages
    eurosym <- "eurosym" %in% packages
    straightQuotes <- match.arg(quotes) == "straight"
    ## Remove control characters (not spaces!)
    y <- gsub("(?![[:space:]])[[:cntrl:]]", "", y, perl=TRUE)
    ## Convert any sequence of whitespace to a single space.  This
    ## substitution must be done before control characters because
    ## newline belongs to both groups.
    y <- gsub("[[:space:]]+", " ", y)

    ## Handle LaTeX special characters in the ASCII range.
    ## Some substitutions are mandatory, others affect matters such as
    ## the rendering of the character in question (e.g. \textquote...)
    ## or line breaks (\slash).  Some substitutions are unnecessary if
    ## using a font encoding other than OT1 (\textbar, ...), but are
    ## performed every time nonetheless.  The dash (-) is given
    ## special treatment to disable the em and en dash ligatures.
    ## Source: Scott Pakin (2009) The Comprehensive LaTeX Symbol List.
    ## Accessible through "texdoc symbols".
    ## Particularly section 8.6 "ASCII and Latin 1 quick reference".
    ##
    ## The order of the elements in the list matters!
    ## First, { and } are replaced with \{ and \}, respectively.
    ## Then, \ is replaced with \textbackslash{},
    ## but not if followed by { or }.
    ## After that, the order does not matter.
    ##
    ## This starts the 'substitutions' list, which is finally
    ## processed close to the end of the function.
    substitutions <-
        list(c("([{}])", "\\\\\\1"),
             c("\\\\(?![{}])", "\\\\textbackslash{}"),
             c("\\^", "\\\\textasciicircum{}"),
             c("~", "\\\\textasciitilde{}"),
             c("<", "\\\\textless{}"),
             c(">", "\\\\textgreater{}"),
             c("\\|", "\\\\textbar{}"),
             c("([#$%&_])", "\\\\\\1"),
             if (isTRUE(dashdash)) {
                 c("-", "\\\\mbox{-}")
             },
             if (textcomp && straightQuotes) {
                 c("'", "\\\\textquotesingle{}")
             },
             if (textcomp && straightQuotes) {
                 c("`", "\\\\textasciigrave{}")
             },
             c('"', if (fontenc && straightQuotes) {
                 "\\\\textquotedbl{}"
             } else {
                 "\\\\textquotedblright{}"
             }),
             c("/", "\\\\slash{}"))
    substitutions <- substitutions[!vapply(substitutions, is.null, logical(1))]

    ## Treatment of non-ASCII characters follows.

    ## Digraphs and ligatures broken into their parts, except the
    ## Dutch digraphs IJ and ij which have their own commands.
    substitutions <-
        c(substitutions,
          list(c("\u0132", "\\\\IJ{}"),
               c("\u0133", "\\\\ij{}"),
               c("\u01f1", "DZ"),
               c("\u01f2", "Dz"),
               c("\u01f3", "dz"),
               c("\u01c4", "DZ\u030c"),
               c("\u01c5", "Dz\u030c"),
               c("\u01c6", "dz\u030c"),
               c("\u01c7", "LJ"),
               c("\u01c8", "Lj"),
               c("\u01c9", "lj"),
               c("\u01ca", "NJ"),
               c("\u01cb", "Nj"),
               c("\u01cc", "nj"),
               c("\ufb00", "ff"),
               c("\ufb01", "fi"),
               c("\ufb02", "fl"),
               c("\ufb03", "ffi"),
               c("\ufb04", "ffl"),
               c("\ufb05", "\u017ft"),
               c("\ufb06", "st")))

    ## Accents (diacritics) above the letter (drop i and j dots)
    above <- list(diaeresis   = c("\u0308", "\""),
                  acute       = c("\u0301", "'"),
                  dotabove    = c("\u0307", "."),
                  macron      = c("\u0304", "="),
                  circumflex  = c("\u0302", "^"),
                  grave       = c("\u0300", "`"),
                  tilde       = c("\u0303", "~"),
                  doubleacute = c("\u030b", "H"),
                  ringabove   = c("\u030a", "r"),
                  breve       = c("\u0306", "u"),
                  caron       = c("\u030c", "v"),
                  invbreve    = c("\u0311", "newtie")) # textcomp
    ## Accents that co-exist with i and j dots (mainly below the
    ## letter, but also the ligature tie)
    below <- list(macronbelow = c("\u0331", "b"),
                  cedilla     = c("\u0327", "c"),
                  dotbelow    = c("\u0323", "d"),
                  tie         = c("\u0361", "t"),
                  ogonek      = c("\u0328", "k")) # not in OT1 fontenc
    accents <- c(above, below)
    command <- paste0("\\\\[", Letters, "]+|\\\\.")
    combining <- paste0(vapply(accents, "[", character(1), 1),
                        collapse="")
    accPre <- paste0("(", command, "|.)({})?(?<![", combining, "])")
    accPost <- paste0("(?![", combining, "])")

    ## Accent above the letter
    aboveInCode <- vapply(above, "[", character(1), 1)
    ijPattern <- paste0("([ij])", aboveInCode, accPost)
    otherPattern <- paste0(accPre, aboveInCode, accPost)
    aboveOutCode <- vapply(above, "[", character(1), 2)
    ijReplacement <- paste0("\\\\", aboveOutCode, "{\\\\\\1}")
    otherReplacement <- paste0("\\\\", aboveOutCode, "{\\1}")
    ## for (accent in above) {
    ##     code <- accent[1]
    ##     replacement <- accent[2]
    ##     y <- gsub(paste0("([ij])", code, accPost),
    ##               paste0("\\\\", replacement, "{\\\\\\1}"), y, perl = TRUE)
    ##     y <- gsub(paste0(accPre, code, accPost),
    ##               paste0("\\\\", replacement, "{\\1}"), y, perl = TRUE)
    ## }

    ## Accent below the letter, and ligature tie
    belowInCode <- vapply(below, "[", character(1), 1)
    belowPattern <- paste0(accPre, belowInCode, accPost)
    belowOutCode <- vapply(below, "[", character(1), 2)
    belowReplacement <- paste0("\\\\", belowOutCode, "{\\1}")
    ## for (accent in below) {
    ##     code <- accent[1]
    ##     replacement <- accent[2]
    ##     y <- gsub(paste0(accPre, code, accPost),
    ##               paste0("\\\\", replacement, "{\\1}"), y, perl = TRUE)
    ## }

    ## Combining an enclosing circle with an accent seems to work
    circPre <- paste0("(", command, "({([^}]|\\\\})+})?|.)({})?")
    circPattern <- paste0(circPre, "\u20dd", accPost)
    circReplacement <- "\\\\textcircled{\\1}"

    substitutions <-
        c(substitutions,
          lapply(lapply(mapply(list, list(as.name("c")),
                               c(ijPattern, otherPattern,
                                 belowPattern, circPattern),
                               c(ijReplacement, otherReplacement,
                                 belowReplacement, circReplacement),
                               SIMPLIFY=FALSE), as.call), eval))

    ## The output of sQuote() and dQuote() may contain non-ASCII
    ## quoting characters.  If the input is ASCII, it may be a
    ## surprise to the user that an UTF-8 input encoding is then
    ## needed in LaTeX.  Converting the quotes to commands solves
    ## this problem.  The substitution list also contains
    ## non-ASCII letters and other unicode characters.
    substitutions <-
        c(substitutions,
          list(c("\u00a1", "\\\\textexclamdown{}"),
               c("\u00a3", "\\\\pounds{}"),
               c("\u00a7", "\\\\S{}"),
               c("\u00a9", "\\\\copyright{}"),
               c("\u00aa", "\\\\textordfeminine{}"),
               c("\u00ae", "\\\\textregistered{}"),
               c("\u00b6", "\\\\P{}"),
               c("\u00b7", "\\\\textperiodcentered{}"),
               c("\u00ba", "\\\\textordmasculine{}"),
               c("\u00bf", "\\\\textquestiondown{}"),
               c("\u2013", "\\\\textendash{}"),
               c("\u2014", "\\\\textemdash{}"),
               c("\u2018", "\\\\textquoteleft{}"),
               c("\u2019", "\\\\textquoteright{}"),
               c("\u201c", "\\\\textquotedblleft{}"),
               c("\u201d", "\\\\textquotedblright{}"),
               c("\u2020", "\\\\dag{}"),
               c("\u2021", "\\\\ddag{}"),
               c("\u2022", "\\\\textbullet{}"),
               c("\u2026", "\\\\dots{}"),
               c("\u2122", "\\\\texttrademark{}"),
               c("\u2423", "\\\\textvisiblespace{}"),
               c("\u00c6", "\\\\AE{}"),
               c("\u00e6", "\\\\ae{}"),
               c("\u0152", "\\\\OE{}"),
               c("\u0153", "\\\\oe{}"),
               c("\u00d8", "\\\\O{}"),
               c("\u00f8", "\\\\o{}"),
               c("\u0141", "\\\\L{}"),
               c("\u0142", "\\\\l{}"),
               ## U+1E9E Latin capital letter sharp s. Works with
               ## XeTeX and LuaTeX, provided that the character is
               ## present in the font. Otherwise \SS, which usually
               ## produces "SS".
               c("\u1e9e", "\\\\ifdefined\\\\XeTeXrevision\\\\iffontchar\\\\font\"1E9E\\\\symbol{\"1E9E}\\\\else\\\\SS\\\\fi\\\\else\\\\ifdefined\\\\directlua\\\\iffontchar\\\\font\"1E9E\\\\symbol{\"1E9E}\\\\else\\\\SS\\\\fi\\\\else\\\\SS\\\\fi\\\\fi{}"),
               c("\u00df", "\\\\ss{}"),
               ## U+017F Latin small letter long s
               c("\u017f", "\\\\ifdefined\\\\XeTeXrevision\\\\symbol{\"017F}\\\\else\\\\ifdefined\\\\directlua\\\\symbol{\"017F}\\\\else{\\\\fontencoding{TS1}\\\\selectfont s}\\\\fi\\\\fi{}")))
    ## Other non-ASCII letters, punctuation marks.
    ## These don't work with the OT1 font encoding.
    substitutions <-
        c(substitutions,
          list(c("\u00d0", "\\\\DH{}"),
               c("\u00f0", "\\\\dh{}"),
               c("\u0110", "\\\\DJ{}"),
               c("\u0111", "\\\\dj{}"),
               c("\u014a", "\\\\NG{}"),
               c("\u014b", "\\\\ng{}"),
               c("\u00de", "\\\\TH{}"),
               c("\u00fe", "\\\\th{}"),
               c("\u00ab", "\\\\guillemotleft{}"),
               c("\u00bb", "\\\\guillemotright{}"),
               c("\u201a", "\\\\quotesinglbase{}"),
               c("\u201e", "\\\\quotedblbase{}"),
               c("\u2039", "\\\\guilsinglleft{}"),
               c("\u203a", "\\\\guilsinglright{}")))
    ## Miscellaneous, arrows, delimiters, legal symbols, science
    ## and engineering, currencies, diacritics, text mode math.
    ## These require textcomp.
    substitutions <-
        c(substitutions,
          list(c("\u00a0", "~"),# no-break space (NBSP)
               c("\u00ad", "\\\\-"),# soft hyphen (SHY)
               c("\u200b", "\\\\hspace{0pt}"),# zero width space (ZWSP)
               c("\u2217", "\\\\textasteriskcentered{}"),
               c("\u2016", "\\\\textbardbl{}"),
               c("\u25ef", "\\\\textbigcircle{}"),
               c("\u2422", "\\\\textblank{}"),
               c("\u00a6", "\\\\textbrokenbar{}"),
               c("\u2052", "\\\\textdiscount{}"),
               c("\u212e", "\\\\textestimated{}"),
               c("\u203d", "\\\\textinterrobang{}"),
               c("\u2e18", "\\\\textinterrobangdown{}"),
               c("\u2116", "\\\\textnumero{}"),
               c("\u25e6", "\\\\textopenbullet{}"),
               c("\u2030", "\\\\textperthousand{}"),
               c("\u2031", "\\\\textpertenthousand{}"),
               c("\u211e", "\\\\textrecipe{}"),
               c("\u203b", "\\\\textreferencemark{}"),
               c("\u02f7", "\\\\texttildelow{}"),
               c("\u2190", "\\\\textleftarrow{}"),
               c("\u2191", "\\\\textuparrow{}"),
               c("\u2192", "\\\\textrightarrow{}"),
               c("\u2193", "\\\\textdownarrow{}"),
               c("\u3008", "\\\\textlangle{}"),
               c("\u3009", "\\\\textrangle{}"),
               c("\u301a", "\\\\textlbrackdbl{}"),
               c("\u301b", "\\\\textrbrackdbl{}"),
               c("\u2045", "\\\\textlquill{}"),
               c("\u2046", "\\\\textrquill{}"),
               c("\u2117", "\\\\textcircledP{}"),
               c("\u2120", "\\\\textservicemark{}"),
               c("\u2103", "\\\\textcelsius{}"),
               c("\u2127", "\\\\textmho{}"),
               c("\u00b5", "\\\\textmu{}"),
               c("\u03a9", "\\\\textohm{}"),
               c("\u0e3f", "\\\\textbaht{}"),
               c("\u00a2", "\\\\textcent{}"),
               c("\u20a1", "\\\\textcolonmonetary{}"),
               c("\u00a4", "\\\\textcurrency{}"),
               c("\u20ab", "\\\\textdong{}"),
               c("\u20ac", if (eurosym) "\\\\euro{}" else "\\\\texteuro{}"),
               c("\u20b2", "\\\\textguarani{}"),
               c("\u20a4", "\\\\textlira{}"),
               c("\u20a6", "\\\\textnaira{}"),
               c("\u20b1", "\\\\textpeso{}"),
               c("\u20a9", "\\\\textwon{}"),
               c("\u00a5", "\\\\textyen{}"),
               c("\u02dd", "\\\\textacutedbl{}"),
               c("\u00b4", "\\\\textasciiacute{}"),
               c("\u00b8", "\\\\c{}"),
               c("\u02d8", "\\\\textasciibreve{}"),
               c("\u02c7", "\\\\textasciicaron{}"),
               c("\u00a8", "\\\\textasciidieresis{}"),
               c("\u00af", "\\\\textasciimacron{}"),
               c("\u00b0", "\\\\textdegree{}"),
               c("\u00f7", "\\\\textdiv{}"),
               c("\u00bc", "\\\\textonequarter{}"),
               c("\u00bd", "\\\\textonehalf{}"),
               c("\u00be", "\\\\textthreequarters{}"),
               c("\u00d7", "\\\\texttimes{}"),
               c("\u00b1", "\\\\textpm{}"),
               c("\u00b9", "\\\\textonesuperior{}"),
               c("\u00b2", "\\\\texttwosuperior{}"),
               c("\u00b3", "\\\\textthreesuperior{}"),
               c("\u2044", "\\\\textfractionsolidus{}"),
               c("\u221a", "\\\\textsurd{}"),
               c("\u00ac", "\\\\textlnot{}"),
               c("\u2212", "\\\\textminus{}")))

    ## After a command, remove empty group (when at the end of the
    ## string or some suitable character follows) or replace it with a
    ## space (when a space does not follow).
    tmp <- paste0("(\\\\[", Letters, "]+){}")
    substitutions <-
        c(substitutions,
          list(c(paste0(tmp, "(?=$|[[:digit:],.?!;:\\\\}+*/-])"), "\\1"),
               c(paste0(tmp, "(?! )"), "\\1 ")))

    ## Apply the substitutions in the list
    for (subst in substitutions) {
        y <- gsub(subst[1], subst[2], y, perl = TRUE)
    }
    if (isTRUE(doublebackslash)) {
        y <- gsub("\\", "\\\\", y, fixed=TRUE)
    }
    ## Convert result to UTF-8 NFC encoding, although most non-ASCII
    ## content has probably been converted to LaTeX commands.
    stri_trans_nfc(y)
}

context("utility functions")
test.uuid.gen <- function() {
    ## Setup
    SAMP.SIZE <- 100
    ugen <- uuid.gen()
    uuids <- character(SAMP.SIZE)
    for (i in seq_len(SAMP.SIZE)) {
        uuids[i] <- ugen()
    }
    uuids.split <- strsplit(uuids, split="-", fixed=TRUE)
    unique.nchar <- unique(t(sapply(uuids.split, nchar)))
    unique.chars <-
        unique(strsplit(paste(sapply(uuids.split, paste, collapse=""),
                              collapse=""), split=character(0))[[1]])
    all.4 <- unique(substr(uuids, 15, 15))
    one.of.four <- unique(substr(uuids, 20, 20))
    ## Test
    test_that("unique IDs are unique", {
        expect_equal(SAMP.SIZE, length(unique(uuids)))
    })
    test_that("UUIDs have the right parts", {
        expect_true(all(nchar(uuids) == 36))
        expect_true(all(sapply(uuids.split, length) == 5))
        expect_true(nrow(unique.nchar) == 1 &&
                    all(as.vector(unique.nchar) == c(8, 4, 4, 4, 12)))
        expect_true(all(unique.chars %in%
                        c(as.character(0:9), letters[seq_len(6)])))
    })
    test_that("UUID fixed bits are correct", {
        expect_equal(all.4, "4")
        expect_true(all(one.of.four %in% c("8", "9", "a", "b")))
    })
}
test.uuid.gen()

## If 'testDocument' is TRUE, produces a test document to 'con', which may
## be a connection or a filename.
test.latexify <-
    function(testDocument=FALSE,
             con=tempfile(pattern = "latexify", fileext = ".tex"))
{
    ## First part of tests: basic character set

    ## Number of test strings (first part, including one "extra
    ## difficult" case and one empty string)
    SAMP.SIZE <- 50
    stopifnot(SAMP.SIZE >= 2)
    MIN.LENGTH <- 1
    MAX.LENGTH <- 1000
    MAX.NCHAR <- 20 # maximum length of a "word"
    ## All ASCII characters except NUL (0)
    characters <- rawToChar(as.raw(1:127), multiple = TRUE)
    ## LaTeX special characters.  Some of these must be converted to
    ## commands.  Others (e.g. single and double quote, *) are
    ## converted to commands or other characters for improved
    ## compatibility with other packages or to get a particular glyph
    ## (upright quote, centered asterisk) instead of the default
    ## (curved quote, asterisk located higher than the center of the
    ## line).  Some characters would work fine in font encodings other
    ## than OT1 (e.g. <, >, |), but are converted to commands anyway.
    ## The dash is included to prevent -- and --- turning into an n
    ## dash and an m dash, respectively.
    ##
    ## NOTE that the handling (what kind of treatment if any) of some
    ## characters (currently quotes) depends on the (default)
    ## arguments of latexify().
    specialChars <-
        c("{", "}", "\\", "#", "$", "%", "^", "&", "_", "~", "\"", "/", "'",
          "<", ">", "|", "`", "-")
    specialStr <- paste(specialChars, collapse="")
    ## latexify() is designed to convert any sequence of space
    ## characters to a single regular space
    spaceChars <- c("\t", "\n", "\v", "\f", "\r", " ")
    spaceStr <- paste(spaceChars, collapse="")
    ## latexify() is designed to drop control characters excluding spaces
    controlChars <- setdiff(rawToChar(as.raw(c(1:31, 127)), multiple = TRUE),
                            spaceChars)
    controlStr <- paste(controlChars, collapse="")
    ## Decide the length of each test string
    stringLengths <- sample(MIN.LENGTH:MAX.LENGTH, SAMP.SIZE - 2)
    nTotal <- sum(stringLengths)
    ## Create the test strings:
    ## * The last element is a "difficult case".
    ## * The other elements consist of a random sample of characters.
    strStop <- cumsum(stringLengths)
    strStart <- strStop - (stringLengths - 1)
    nSpaces <- round(0.2 * nTotal) # In addition to spaces in 'characters'
    testStrings <-
        c(substring(paste(sample(c(rep(characters,
                                       length.out = nTotal - nSpaces),
                                   rep.int(" ", nSpaces))),
                          collapse=""), strStart, strStop),
          paste(c(specialChars, " ",
                  rev(specialChars), " ",
                  unlist(lapply(lapply(specialChars,rep,3), c, " ")),
                  paste(specialChars, " \t")),
                collapse=""),
          "")

    ## Run latexify() on testStrings
    ltxDouble <- latexify(testStrings, doublebackslash=TRUE)
    ltxSingle <- latexify(testStrings, doublebackslash=FALSE)

    ## Tests
    test_that("doublebackslash works", {
        expect_equal(ltxDouble,
                     gsub("\\", "\\\\",
                          ltxSingle, fixed=TRUE, useBytes=TRUE))
    })
    test_that("no control characters are left", {
        expect_false(any(grepl(sprintf("[%s]", controlStr),
                               ltxSingle, useBytes=TRUE)))
    })
    test_that("space characters collapse", {
        expect_false(any(grepl(sprintf("[%s]{2,}", spaceStr),
                               ltxSingle, useBytes=TRUE)))
    })
    test_that("there are no line breaks", {
        expect_false(any(grepl("\\\\", ltxSingle, fixed=TRUE)))
    })
    Letters <- paste(c(LETTERS, letters), collapse="")
    textCommand <- sprintf("\\\\[%s]+(\\{([^}]|\\\\})+})?", Letters)
    commandTerminated <-
        paste(textCommand,
              "((?<=})|\\{\\}| +|(?=$|[[:digit:],.?!;:\\\\}+*/-]))", sep="")
    commandsAt <- gregexpr(commandTerminated, ltxSingle, perl=TRUE)
    test_that("commands are terminated", {
        expect_equal(lapply(gregexpr(textCommand, ltxSingle), as.vector),
                     lapply(commandsAt, as.vector))
    })
    escape <- sprintf("\\\\[^%s](\\{\\})?", Letters)
    escapesAt <- gregexpr(escape, ltxSingle)

    ## specialMap: record of special character -> command mapping
    specialMap <- vector(mode="list", length = SAMP.SIZE)
    ## Test that each test string was converted properly and prepare
    ## specialMap
    multiSpaceClass <- sprintf("[%s]+", spaceStr)
    controlClass <- sprintf("[%s]", controlStr)
    specialClass <- sprintf("[%s]", specialStr)
    ## Record an ASCII representation of each element in 'testStrings',
    inputDescription <-
        capture.output(invisible(vapply(testStrings, dput, "")))
    for (i in seq_len(SAMP.SIZE)) {
        thisInfo <- inputDescription[i]
        nChars <- nchar(ltxSingle[i])
        ## ltxChars: split ltxSingle[i] into strings representing one
        ## character each
        if (nChars > 0) {
            coms <- commandsAt[[i]]
            escs <- escapesAt[[i]]
            comLengths <- attr(coms, "match.length")
            escLengths <- attr(escs, "match.length")
            if (length(coms) == 1 && coms == -1) {
                coms <- numeric(0)
                comLengths <- numeric(0)
            }
            if (length(escs) == 1 && escs == -1) {
                escs <- numeric(0)
                escLengths <- numeric(0)
            }
            comsAndEscs <- c(coms, escs)
            idx <- order(comsAndEscs)
            comEsc <- comsAndEscs[idx]
            comEscLen <- c(comLengths, escLengths)[idx]
            nComEsc <- length(comEsc)
            charIdx <- numeric(nChars)
            prv <- 0
            prvIdx <- 0
            for (j in seq_along(comEsc)) {
                thisStart <- comEsc[j]
                nSingle <- thisStart - prvIdx - 1
                charIdx[seq(from = prvIdx + 1, length.out = nSingle)] <-
                    seq(from = prv + 1, length.out = nSingle)
                prv <- prv + nSingle + 1
                prvIdx <- thisStart + (comEscLen[j] - 1)
                charIdx[thisStart:prvIdx] <- prv
            }
            nSingle <- nChars - prvIdx
            charIdx[seq(from = prvIdx + 1, length.out = nSingle)] <-
                seq(from = prv + 1, length.out = nSingle)
            ## Each element of ltxChars should represent one character or
            ## a space between words
            strStart <- which(diff(c(0, charIdx)) > 0)
            strStop <- c(strStart[-1] - 1, nChars)
            ## Strip off empty group or spaces following a command
            ltxChars <- sub(sprintf("([%s])(\\{\\}| +)$", Letters), "\\1",
                            substring(ltxSingle[i], strStart, strStop))
        } else {
            ltxChars <- character(0)
        }
        ## stripChars: "Independently" do a part of what latexify()
        ## does, i.e. remove control characters and use single spaces
        ## only
        stripChars <-
            strsplit(gsub(multiSpaceClass,
                          " ",
                          gsub(controlClass,
                               "",
                               testStrings[i])),
                     "")[[1]]
        singleFlag <- nchar(ltxChars) == 1
        multiFlag <- !singleFlag
        specialMap[[i]] <- unique(cbind(stripChars[multiFlag],
                                        ltxChars[multiFlag]))
        ## Compare ltxChars and stripChars
        test_that("normal and special characters are treated correctly", {
            expect_equal(length(stripChars), length(ltxChars),
                         info=thisInfo, tolerance=0)
            expect_false(any(grepl(specialClass,
                                   ltxChars[singleFlag], useBytes=TRUE)),
                        info=thisInfo)
            expect_equal(ltxChars[singleFlag], stripChars[singleFlag],
                         info=thisInfo)
        })
    }
    ## specialMap becomes a combination of the unique rows across its
    ## elements
    specialMap <- do.call(rbind, specialMap)
    specialMap <- unique(specialMap)
    test_that("mapping of special characters is consistent", {
        expect_equal(nrow(specialMap), length(specialChars))
        expect_true(all(specialChars %in% specialMap[, 1]))
    })

    ## Second part of tests: extended character set

    ## A test for handling of different encodings in the input

    ## The following string must have a literal (escaped) backspace
    ## and "x" in front of every "special" character (byte), and each
    ## character code must consist of two hexadecimal digits.
    bytePrint <- "clich\\xe9\\x0ama\\xf1ana" # "\x0a" is a newline

    codeLoc <- as.vector(gregexpr("\\x", bytePrint, fixed=TRUE)[[1]])
    asSuch <- substring(bytePrint,
                        c(1, codeLoc + 4),
                        c(codeLoc - 1, nchar(bytePrint)))
    special <-
        rawToChar(as.raw(paste0("0x", substring(bytePrint,
                                                codeLoc + 2, codeLoc + 3))),
                  multiple = TRUE)
    latin1String <- paste0(asSuch, c(special, ""), collapse="")
    Encoding(latin1String) <- "latin1"
    byteString <- latin1String
    Encoding(byteString) <- "bytes"

    latinConverted <- latexify(latin1String, doublebackslash=FALSE)
    byteConverted <- latexify(byteString, doublebackslash=FALSE)
    test_that("latexify handles different encodings", {
        expect_equal(latinConverted, "clich\\'{e} ma\\~{n}ana")
        expect_equal(latexify(enc2utf8(latin1String), doublebackslash=FALSE),
                     latinConverted)
        expect_equal(tolower(byteConverted),# do hex codes print in lower case?
                     gsub("\\", "\\textbackslash ", bytePrint, fixed=TRUE))
    })

    ## A test for other than default quoting options
    quoteString <- "\"It's five o'clock\", he said."
    res1 <- latexify(quoteString, doublebackslash=FALSE)
    res2 <- latexify(quoteString, quotes="curved", doublebackslash=FALSE)
    res3 <- latexify(quoteString, packages=character(0), doublebackslash=FALSE)
    res4 <- latexify(quoteString, packages="fontenc", doublebackslash=FALSE)
    res5 <- latexify(quoteString, packages="textcomp", doublebackslash=FALSE)
    exp2 <- sub("\"", "\\\\textquotedblright ", quoteString)
    exp2 <- sub("\"", "\\\\textquotedblright", exp2)
    exp4 <- sub("\"", "\\\\textquotedbl ", quoteString)
    exp4 <- sub("\"", "\\\\textquotedbl", exp4)
    exp5 <- gsub("'", "\\\\textquotesingle ", exp2)
    exp1 <- gsub("'", "\\\\textquotesingle ", exp4)
    test_that("quoting options work", {
        expect_equal(res1, exp1, info="Default straight quotes")
        expect_equal(res2, exp2, info="Curved quotes")
        expect_equal(res3, res2, info="Fallback to curved quotes")
        expect_equal(res4, exp4, info="Fallback to curved single quotes")
        expect_equal(res5, exp5, info="Fallback to curved double quotes")
    })
    ## Check that non-ASCII quotes used by dQuote() and sQuote() are
    ## converted to LaTeX commands
    nestQuotes <- paste0("You said, \u201cHe said, ",
                         "\u2018Have a nice day.\u2019\u201d")
    nq <- latexify(nestQuotes, doublebackslash=FALSE)
    test_that("dQuote() and sQuote() are safe", {
        expect_equal(nq,
                     gsub("\\{\\}(?=\\\\)", "",
                          gsub("\u2018", "\\\\textquoteleft ",
                               gsub("\u2019", "\\\\textquoteright",
                                    gsub("\u201c", "\\\\textquotedblleft ",
                                         gsub("\u201d",
                                              "\\\\textquotedblright",
                                              nestQuotes)))),
                          perl=TRUE))
    })
    diaeresisD <- "o\u0308ljysa\u0308ilio\u0308"
    diasD <- latexify(diaeresisD, doublebackslash=FALSE)
    diaeresisC <- "\u00f6ljys\u00e4ili\u00f6"
    diasC <- latexify(diaeresisC, doublebackslash=FALSE)
    test_that("Unicode NFD and NFC are handled the same way", {
        expect_equal(diasD, diasC)
    })
    ## Strings containing practically every non-ASCII character that
    ## will be converted by latexify()
    breakWords <- rep.int(c("space", "vacation", "movie", "line", "break",
                            "rope", "period"), 2)
    allChars <-
        c("\u0132sselmeer is a lake, Berl\u0133n is Dutch for Berlin.",
          paste0("Other digraphs and ligatures: \u01f1, \u01f2, \u01f3, ",
                 "\u01c4, \u01c5, \u01c6, \u01c7, \u01c8, \u01c9, \u01ca, ",
                 "\u01cb, \u01cc, \ufb00, \ufb01, \ufb02, \ufb03, \ufb04, ",
                 "\ufb05, \ufb06"),
          paste0("\u017c\u00f3\u0142ty, p\u0113c, f\u00eate, ",
                 "vis-\u00e0-vis, pi\u00f1ata, Erd\u0151s, ",
                 "\u00c5ngstr\u00f6m, m\u0103r, t\u0159i, \u0203, t\u0331, ",
                 "fa\u00e7ade, \u1e0c, ",
                 "k\u0361p (there should be a ligature tie), t\u0105sa"),
          "Circled: a\u20dd, \u00f1\u20dd",
          "\u00a1Hola! \u00bfQu\u00e9 pasa? \u2e18Verdad\u203d",
          paste0("Money: \u00a3, \u20ac, \u00a2, \u00a5, \u00a4, \u0e3f, ",
                 "\u20a1, \u20ab, \u20b2, \u20a4, \u20a6, \u20b1, \u20a9"),
          paste0("No-Break\u00a0", breakWords, collapse=" "),
          paste0("Do-Break ", breakWords, collapse=" "),
          "visible\u2423space",
          paste0(rep.int("Zero\u200bWidth\u200bSpace\u200b", 10), collapse=""),
          paste0(rep.int(paste0("S\u00ado\u00adf\u00adt\u00ad",
                                "H\u00ady\u00adp\u00adh\u00ade\u00adn\u00ad",
                                "E\u00adv\u00ade\u00adr\u00ady\u00ad",
                                "w\u00adh\u00ade\u00adr\u00ade"), 8),
                 collapse="\u00ad"),
          "Legal \u00a7, \u00a9, \u00ae, \u00b6, \u2117, \u2120, \u2122",
          paste0("Letters \u00c6, \u00e6, \u0152, \u0153, \u00d8, \u00f8, ",
                 "\u0141, \u0142, \u1e9e, \u00df, \u017f, \u00d0, \u00f0, ",
                 "\u0110, \u0111, \u014a, \u014b, \u00de, \u00fe"),
          paste0("Quotes: \u00ab | \u00bb | \u201a | \u201e | \u2039 | \u203a",
                 " | \u2018 | \u2019 | \u201c | \u201d"),
          paste0("mho \u2127, ohm \u03a9, micro \u00b5, Celsius \u2103, ",
                 "degree \u00b0"),
          paste0("Division \u00f7, times \u00d7, fractions \u00bc, \u00bd, ",
                 "\u00be, plusminus \u00b1, root \u221a, minus \u2212, ",
                 "asterisk \u2217, not \u00ac, fraction solidus \u2044, ",
                 "superscript \u00b9, \u00b2, \u00b3"),
          paste0("Discount \u2052, estimated \u212e, \u2116 5, ",
                 "per mille \u2030, parts per ten thousand \u2031"),
          "Bullets: \u2022 Closed, \u25e6 open",
          paste0("Spacing diacritics: double acute \u02dd, acute \u00b4, ",
                 "cedilla \u00b8, breve \u02d8, ",
                 "caron \u02c7, diaeresis \u00a8, macron \u00af"),
          "Arrows: left \u2190, up \u2191, right \u2192, down \u2193",
          "Delimiters: \u3008, \u3009, \u301a, \u301b, \u2045, \u2046",
          paste0("Miscellaneous: feminine ordinal \u00aa, ",
                 "masculine ordinal \u00ba, middle dot \u00b7, ",
                 "daggers \u2020, \u2021, ellipsis \u2026, ",
                 "double vertical line \u2016, large circle \u25ef, ",
                 "blank symbol \u2422, broken bar \u00a6, recipe \u211e, ",
                 "reference mark \u203b, low tilde \u02f7, ",
                 "en dash \u2013, em dash \u2014"))
    ac <- latexify(allChars, doublebackslash=FALSE)
    test_that("a certain set of non-ASCII characters is converted", {
        expect_true(all(Encoding(ac) == "unknown"))
    })
    ## When used independently outside the test suite, the function
    ## can create a test document, but only in a UTF-8 locale.
    if (isTRUE(testDocument) && isTRUE(l10n_info()[["UTF-8"]])) {
        preamble <- c("\\documentclass[a4paper]{article}",
                      "\\usepackage{etoolbox}",
                      "\\usepackage{ifluatex}",
                      "\\usepackage{ifxetex}",
                      "\\usepackage{parskip}",
                      "\\usepackage{textcomp}",
                      "\\ifbool{luatex}{",
                      "\\usepackage{fontspec}",
                      "}{",
                      "\\ifbool{xetex}{",
                      "\\usepackage{fontspec}",
                      "}{",
                      "\\usepackage[T1]{fontenc}",
                      "\\usepackage{lmodern}",
                      "}}")
        id <- c(latin1String, byteString, rep.int(quoteString, 5),
                nestQuotes, diaeresisD, diaeresisC, allChars)
        extraInfo <- c(rep.int("", length(testStrings) + length(latin1String) +
                               length(byteString)),
                       paste0(" (", c("default", "curved", "no packages",
                                      "only fontenc", "only textcomp"), ")"),
                       rep.int("", length(nestQuotes)),
                       rep.int(" (Unicode NFD)", length(diaeresisD)),
                       rep.int(" (Unicode NFC)", length(diaeresisC)),
                       rep.int("", length(allChars)))

        ## Record an ASCII representation of each element in 'id',
        ## add to previous record of 'testStrings'
        inputDescription <- c(inputDescription,
                              capture.output(invisible(vapply(id, dput, ""))))

        allOutput <- c(ltxSingle, latinConverted, byteConverted, res1, res2,
                       res3, res4, res5, nq, diasD, diasC, ac)
        if (is.character(con)) {
            co <- file(con, open = "wt", encoding = "UTF-8")
            on.exit(close(co))
        } else {
            co <- con
            if (!isOpen(co)) {
                open(co, open = "wt")
                on.exit(close(co))
            }
        }
        writeLines(preamble, co)
        writeLines("\\begin{document}", co)
        writeLines("\\begin{enumerate}", co)
        writeLines(paste0("% ", inputDescription, extraInfo, "\n",
                          "\\item\\relax ", allOutput,
                          "% ", seq_along(allOutput)),
                   co, sep = "\n\n")
        writeLines("\\end{enumerate}", co)
        writeLines("\\end{document}", co)
        con
    }
}
test.latexify()

test.latexDate <- function() {
    SAMP.SIZE <- 100
    dates <- Sys.Date() + round(runif(SAMP.SIZE, min = -1000, max = 1000))
    latexDates <- latexDate(dates)
    test_that("length of output equals length of input", {
        expect_equal(length(dates), length(latexDates), tolerance=0)
    })
    splitDates <- strsplit(latexDates, ", ")
    test_that("year is at the end, separated by comma and space", {
        expect_equal(vapply(splitDates, length, numeric(1)),
                     rep.int(2, length(dates)), tolerance=0)
    })
    monthsDays <- vapply(splitDates, "[[", character(1), 1)
    splitDates2 <- strsplit(monthsDays, " ")
    test_that("month and day are separated by space", {
        expect_equal(vapply(splitDates2, length, numeric(1)),
                     rep.int(2, length(dates)), tolerance=0)
    })
    yearStr <- vapply(splitDates, "[[", character(1), 2)
    Months <- match(vapply(splitDates2, "[[", character(1), 1), month.name)
    monthStr <- sprintf("%02.0f", Months)
    Days <- suppressWarnings(as.numeric(vapply(splitDates2,
                                               "[[", character(1), 2)))
    dayStr <- sprintf("%02.0f", Days)
    test_that("output of latexDate(x) represents date x", {
        expect_equal(paste(yearStr, monthStr, dayStr, sep="-"),
                     as.character(dates))
    })
}
test.latexDate()

ffcsaps <- function(y, x=seq_along(y), nyrs=length(y)/2, f=0.5) {
  # for v 1.7.8
  #.Deprecated(new = "caps",msg = "ffcsaps is deprecated as of version 1.7.8. Please use caps instead. ffcsaps will be defunct as of version 1.7.9")
  deprecate_warn(when = "1.7.8", what = "ffcsaps()", with = "caps()")
  caps(y,nyrs,f)  # Forward arguments to caps()
  # for v 1.7.9
  #.Defunct(new = "caps",msg = "ffcsaps is defunct as of version 1.7.9. Please use caps instead.")
}
# ffcsaps <- function(y, x=seq_along(y), nyrs=length(y)/2, f=0.5) {
# ### support functions
#     ffppual <- function(breaks, c1, c2, c3, c4, x, left){
#         if (left){
#             ix <- order(x)
#             x2 <- x[ix]
#         } else{
#             x2 <- x
#         }
# 
#         n.breaks <- length(breaks)
#         if (left) {
#             ## index[i] is maximum of a and b:
#             ## a) number of elements in 'breaks[-n.breaks]' that are
#             ##    less than or equal to x2[i],
#             ## b) 1
#             index <- pmax(ffsorted(breaks[-n.breaks], x2), 1)
#         } else {
#             ## index[i] is:
#             ## 1 + number of elements in 'breaks[-1]' that are
#             ## less than x2[i]
#             index <- ffsorted2(breaks[-1], x2)
#         }
# 
#         x2 <- x2 - breaks[index]
#         v <- x2 * (x2 * (x2 * c1[index] + c2[index]) + c3[index]) + c4[index]
# 
#         if (left)
#             v[ix] <- v
#         v
#     }
# 
#     ffsorted <- function(meshsites, sites) {
#         index <- order(c(meshsites, sites))
#         which(index > length(meshsites)) - seq_along(sites)
#     }
# 
#     ffsorted2 <- function(meshsites, sites) {
#         index <- order(c(sites, meshsites))
#         which(index <= length(sites)) - seq(from=0, to=length(sites)-1)
#     }
# 
#     ## Creates a sparse matrix A of size n x n.
#     ## The columns of B are set to the diagonals of A so that column k
#     ## becomes the diagonal in position d[k] relative to the main
#     ## diagonal (zero d[k] is the main diagonal, positive d[k] is
#     ## above, negative is below the main diagonal).
#     ## A value on column j in A comes from row j in B.
#     ## This is similar in function to spdiags(B, d, n, n) in MATLAB.
#     spdiags <- function(B, d, n) {
#         n.d <- length(d)
#         A <- matrix(0, n.d * n, 3)
#         count <- 0
#         for(k in seq_len(n.d)){
#             this.diag <- d[k]
#             i <- inc(max(1, 1 - this.diag), min(n, n - this.diag)) # row
#             n.i <- length(i)
#             if(n.i > 0){
#                 j <- i + this.diag                                 # column
#                 row.idx <- seq(from=count+1, by=1, length.out=n.i)
#                 A[row.idx, 1] <- i
#                 A[row.idx, 2] <- j
#                 A[row.idx, 3] <- B[j, k]
#                 count <- count + n.i
#             }
#         }
#         A <- A[A[, 3] != 0, , drop=FALSE]
#         A[order(A[, 2], A[, 1]), , drop=FALSE]
#     }
# 
# ### start main function
# 
#     y2 <- as.numeric(y)
#     ## If as.numeric() does not signal an error, it is unlikely that
#     ## the result would not be numeric, but...
#     if(!is.numeric(y2)) stop("'y' must be coercible to a numeric vector")
#     x2 <- as.numeric(x)
#     if(!is.numeric(x2)) stop("'x' must be coercible to a numeric vector")
# 
#     n <- length(x2)
#     ## quick error check
#     if (n < 3) stop("there must be at least 3 data points")
#     if(!is.numeric(f) || length(f) != 1 || f < 0 || f > 1)
#         stop("'f' must be a number between 0 and 1")
#     if(!is.numeric(nyrs) || length(nyrs) != 1 || nyrs <= 1)
#         stop("'nyrs' must be a number greater than 1")
# 
#     ix <- order(x2)
#     zz1 <- n - 1
#     xi <- x2[ix]
#     zz2 <- n - 2
#     diff.xi <- diff(xi)
# 
#     ## quick error check
#     if (any(diff.xi == 0)) stop("the data abscissae must be distinct")
# 
#     yn <- length(y2)
# 
#     ## quick error check
#     if (n != yn)
#         stop("abscissa and ordinate vector must be of the same length")
# 
#     arg2 <- -1:1
#     odx <- 1 / diff.xi
#     R <- spdiags(cbind(c(diff.xi[-c(1, zz1)], 0),
#                        2 * (diff.xi[-1] + diff.xi[-zz1]),
#                        c(0, diff.xi[-c(1, 2)])),
#                  arg2, zz2)
#     R2 <- spdiags(cbind(c(odx[-zz1], 0, 0),
#                         c(0, -(odx[-1] + odx[-zz1]), 0),
#                         c(0, 0, odx[-1])),
#                   arg2, n)
#     R2[, 1] <- R2[, 1] - 1
#     forR <- Matrix(0, zz2, zz2, sparse = TRUE)
#     forR2 <- Matrix(0, zz2, n, sparse = TRUE)
#     forR[R[, 1:2, drop=FALSE]] <- R[, 3]
#     forR2[R2[, 1:2, drop=FALSE]] <- R2[, 3]
#     ## The following order of operations was tested to be relatively
#     ## accurate across a wide range of f and nyrs
#     p.inv <- (1 - f) * (cos(2 * pi / nyrs) + 2) /
#         (12 * (cos(2 * pi / nyrs) - 1) ^ 2) / f + 1
#     yi <- y2[ix]
#     p <- 1 / p.inv
#     mplier <- 6 - 6 / p.inv # slightly more accurate than 6*(1-1/p.inv)
#     ## forR*p is faster than forR/p.inv, and a quick test didn't
#     ## show any difference in the final spline
#     u <- as.numeric(solve(mplier * tcrossprod(forR2) + forR * p,
#                           diff(diff(yi) / diff.xi)))
#     yi <- yi - mplier * diff(c(0, diff(c(0, u, 0)) / diff.xi, 0))
#     test0 <- xi[-c(1, n)]
#     c3 <- c(0, u / p.inv, 0)
#     x3 <- c(test0, seq(from=xi[1], to=xi[n], length = 101))
#     cc.1 <- diff(c3) / diff.xi
#     cc.2 <- 3 * c3[-n]
#     cc.3 <- diff(yi) / diff.xi - diff.xi * (2 * c3[-n] + c3[-1])
#     cc.4 <- yi[-n]
#     to.sort <- c(test0, x3)
#     ix.final <- order(to.sort)
#     sorted.final <- to.sort[ix.final]
#     tmp <-
#         unique(data.frame(sorted.final,
#                           c(ffppual(xi, cc.1,cc.2,cc.3,cc.4, test0, FALSE),
#                             ffppual(xi, cc.1,cc.2,cc.3,cc.4, x3, TRUE))[ix.final]))
#     ## get spline on the right timescale - kludgy
#     tmp2 <- tmp
#     tmp2[[1]] <- round(tmp2[[1]], 5) # tries to deal with identical() issues
#     res <- tmp2[[2]][tmp2[[1]] %in% x2]
#     ## deals with identical() issues via linear approx
#     if(length(res) != n)
#         res <- approx(x=tmp[[1]], y=tmp[[2]], xout=x2, ties = "ordered")$y
#     res
# }
# 
# # ffcsaps <- function(y, x=seq_along(y), nyrs=length(y)/2, f=0.5) {
# # ### support functions
# #     ffppual <- function(breaks, c1, c2, c3, c4, x, left){
# #         if (left){
# #             ix <- order(x)
# #             x2 <- x[ix]
# #         } else{
# #             x2 <- x
# #         }
# # 
# #         n.breaks <- length(breaks)
# #         if (left) {
# #             ## index[i] is maximum of a and b:
# #             ## a) number of elements in 'breaks[-n.breaks]' that are
# #             ##    less than or equal to x2[i],
# #             ## b) 1
# #             index <- pmax(ffsorted(breaks[-n.breaks], x2), 1)
# #         } else {
# #             ## index[i] is:
# #             ## 1 + number of elements in 'breaks[-1]' that are
# #             ## less than x2[i]
# #             index <- ffsorted2(breaks[-1], x2)
# #         }
# # 
# #         x2 <- x2 - breaks[index]
# #         v <- x2 * (x2 * (x2 * c1[index] + c2[index]) + c3[index]) + c4[index]
# # 
# #         if (left)
# #             v[ix] <- v
# #         v
# #     }
# # 
# #     ffsorted <- function(meshsites, sites) {
# #         index <- order(c(meshsites, sites))
# #         which(index > length(meshsites)) - seq_along(sites)
# #     }
# # 
# #     ffsorted2 <- function(meshsites, sites) {
# #         index <- order(c(sites, meshsites))
# #         which(index <= length(sites)) - seq(from=0, to=length(sites)-1)
# #     }
# # 
# #     ## Creates a sparse matrix A of size n x n.
# #     ## The columns of B are set to the diagonals of A so that column k
# #     ## becomes the diagonal in position d[k] relative to the main
# #     ## diagonal (zero d[k] is the main diagonal, positive d[k] is
# #     ## above, negative is below the main diagonal).
# #     ## A value on column j in A comes from row j in B.
# #     ## This is similar in function to spdiags(B, d, n, n) in MATLAB.
# #     spdiags <- function(B, d, n) {
# #         n.d <- length(d)
# #         A <- matrix(0, n.d * n, 3)
# #         count <- 0
# #         for(k in seq_len(n.d)){
# #             this.diag <- d[k]
# #             i <- inc(max(1, 1 - this.diag), min(n, n - this.diag)) # row
# #             n.i <- length(i)
# #             if(n.i > 0){
# #                 j <- i + this.diag                                 # column
# #                 row.idx <- seq(from=count+1, by=1, length.out=n.i)
# #                 A[row.idx, 1] <- i
# #                 A[row.idx, 2] <- j
# #                 A[row.idx, 3] <- B[j, k]
# #                 count <- count + n.i
# #             }
# #         }
# #         A <- A[A[, 3] != 0, , drop=FALSE]
# #         A[order(A[, 2], A[, 1]), , drop=FALSE]
# #     }
# # 
# # ### start main function
# # 
# #     y2 <- as.numeric(y)
# #     ## If as.numeric() does not signal an error, it is unlikely that
# #     ## the result would not be numeric, but...
# #     if(!is.numeric(y2)) stop("'y' must be coercible to a numeric vector")
# #     x2 <- as.numeric(x)
# #     if(!is.numeric(x2)) stop("'x' must be coercible to a numeric vector")
# # 
# #     n <- length(x2)
# #     ## quick error check
# #     if (n < 3) stop("there must be at least 3 data points")
# #     if(!is.numeric(f) || length(f) != 1 || f < 0 || f > 1)
# #         stop("'f' must be a number between 0 and 1")
# #     if(!is.numeric(nyrs) || length(nyrs) != 1 || nyrs <= 1)
# #         stop("'nyrs' must be a number greater than 1")
# # 
# #     ix <- order(x2)
# #     zz1 <- n - 1
# #     xi <- x2[ix]
# #     zz2 <- n - 2
# #     diff.xi <- diff(xi)
# # 
# #     ## quick error check
# #     if (any(diff.xi == 0)) stop("the data abscissae must be distinct")
# # 
# #     yn <- length(y2)
# # 
# #     ## quick error check
# #     if (n != yn)
# #         stop("abscissa and ordinate vector must be of the same length")
# # 
# #     arg2 <- -1:1
# #     odx <- 1 / diff.xi
# #     R <- spdiags(cbind(c(diff.xi[-c(1, zz1)], 0),
# #                        2 * (diff.xi[-1] + diff.xi[-zz1]),
# #                        c(0, diff.xi[-c(1, 2)])),
# #                  arg2, zz2)
# #     R2 <- spdiags(cbind(c(odx[-zz1], 0, 0),
# #                         c(0, -(odx[-1] + odx[-zz1]), 0),
# #                         c(0, 0, odx[-1])),
# #                   arg2, n)
# #     R2[, 1] <- R2[, 1] - 1
# #     forR <- Matrix(0, zz2, zz2, sparse = TRUE)
# #     forR2 <- Matrix(0, zz2, n, sparse = TRUE)
# #     forR[R[, 1:2, drop=FALSE]] <- R[, 3]
# #     forR2[R2[, 1:2, drop=FALSE]] <- R2[, 3]
# #     ## The following order of operations was tested to be relatively
# #     ## accurate across a wide range of f and nyrs
# #     p.inv <- (1 - f) * (cos(2 * pi / nyrs) + 2) /
# #         (12 * (cos(2 * pi / nyrs) - 1) ^ 2) / f + 1
# #     yi <- y2[ix]
# #     p <- 1 / p.inv
# #     mplier <- 6 - 6 / p.inv # slightly more accurate than 6*(1-1/p.inv)
# #     ## forR*p is faster than forR/p.inv, and a quick test didn't
# #     ## show any difference in the final spline
# #     u <- as.numeric(solve(mplier * tcrossprod(forR2) + forR * p,
# #                           diff(diff(yi) / diff.xi)))
# #     yi <- yi - mplier * diff(c(0, diff(c(0, u, 0)) / diff.xi, 0))
# #     test0 <- xi[-c(1, n)]
# #     c3 <- c(0, u / p.inv, 0)
# #     x3 <- c(test0, seq(from=xi[1], to=xi[n], length = 101))
# #     cc.1 <- diff(c3) / diff.xi
# #     cc.2 <- 3 * c3[-n]
# #     cc.3 <- diff(yi) / diff.xi - diff.xi * (2 * c3[-n] + c3[-1])
# #     cc.4 <- yi[-n]
# #     to.sort <- c(test0, x3)
# #     ix.final <- order(to.sort)
# #     sorted.final <- to.sort[ix.final]
# #     tmp <-
# #         unique(data.frame(sorted.final,
# #                           c(ffppual(xi, cc.1,cc.2,cc.3,cc.4, test0, FALSE),
# #                             ffppual(xi, cc.1,cc.2,cc.3,cc.4, x3, TRUE))[ix.final]))
# #     ## get spline on the right timescale - kludgy
# #     tmp2 <- tmp
# #     tmp2[[1]] <- round(tmp2[[1]], 5) # tries to deal with identical() issues
# #     res <- tmp2[[2]][tmp2[[1]] %in% x2]
# #     ## deals with identical() issues via linear approx
# #     if(length(res) != n)
# #         res <- approx(x=tmp[[1]], y=tmp[[2]], xout=x2, ties = "ordered")$y
# #     res
# # }

!*******************************************************************
! calculates a cubic spline function to filter a time series.
! the 'stiffness' of spline is specified by parameter 'ls'
!
! n:  number of values in time series
! y:  time series array to be modeled
! res:  array of cubic spline function values computed
! stiffness: stiffness vector (50% frequency response) of spline
!
!*******************************************************************
!
! AGB Nov 2021 
! Modifying Ed Cook's spline2 subroutine to be a little less ugly
! This included removing all the implicit declarations and setting
! implicit none. Also removing archaic calls to dabs, dcos, dsqrt.
! Changed to follow f95 in terms of I/O variable decalration and
! c binding as well to play nicely with C and .Call
!
!*******************************************************************
      subroutine ads_f(y, n, stiffness, res) bind(C, name = "ads_f_")
      use, intrinsic :: iso_c_binding
      implicit none
      real(kind = c_double), intent(in)  :: y !input, nyrs
      integer(kind = c_int), intent(in)  :: n,stiffness !input
      real(kind = c_double), intent(out) :: res  !output
      integer, parameter :: dp = kind(1d0)
      real(dp) :: v,arg,rn,d1,d2,a,c1,c2,p,pct,pi,sum
      integer :: i,j,k,l,m,imncp1,i1,i2,iw,k1,kl,jm1,n1,nm2,nc,nc1,ncp1

      dimension a(9001,4),c1(4),c2(3),p(9001),stiffness(9001),res(n),y(n)
      data c1 /1.d0,-4.d0,6.d0,-2.d0/
      data c2 /0.d0,.33333333333333d0,1.33333333333333d0/
      data pct,pi / .50d0, 3.1415926535897935d0 /
      if(n .lt. 4)then
        res(1)=-9998. !too few points. bail out.
        return
      endif
	  
      nm2=n-2
! here is where the stiffness vector stiffness is converted into a p vector
! that the spline routine actually needs for calculating the spline
      do i=1,nm2
        v=dble(stiffness(i))                                   
        arg=(2.d0*pi)/v
        p(i)=(6.d0*(cos(arg)-1.d0)**2.d0)/(cos(arg)+2.d0)
      enddo

      do i=1,nm2
         do j=1,3
            a(i,j)=c1(j)+p(i)*c2(j)
            a(i,4)=dble(y(i))+c1(4)*dble(y(i+1))+dble(y(i+2))
         end do
      end do
      a(1,1)=c2(1)
      a(1,2)=c2(1)
      a(2,1)=c2(1)
      nc=2
! begin ludapb
      rn=dble(1.d0/(dble(nm2)*16.d0))
      d1=1.d0
      d2=0.d0
      ncp1=nc+1
      if(nc .ne. 0)then
! initialize zero elements
        do i=1,nc
           do j=i,nc
              k=ncp1-j
              a(i,k)=0.d0
           end do
        end do
! 'i' is row index of element being computed
! 'j' is column index of element being computed
! 'l' is row index of previously computed vector
!       being used to compute inner product
! 'm' is column index
      endif
      do i=1,nm2
         imncp1=i-ncp1
         i1=max0(1,1-imncp1)
         do j=i1,ncp1
            l=imncp1+j
            i2=ncp1-j
            sum=a(i,j)
            jm1=j-1
            if(jm1 .gt. 0)then
              do k=1,jm1
                 m=i2+k
                 sum=sum-a(i,k)*a(l,m)
              end do
            endif
            if(j .eq. ncp1)then
              if(a(i,j)+sum*rn .le. a(i,j))then
                res(1)=-9999.
                return
              endif
              a(i,j)=1.d0/sqrt(sum)
! update determinant
              d1=d1*sum
   35         if(abs(d1) .gt. 1.d0)then
                d1=d1*.0625d0
                d2=d2+4.d0
                goto 35
              endif
   47         if(abs(d1) .le. .0625d0)then
                d1=d1*16.d0
                d2=d2-4.d0
                goto 47
                else
                  goto 60
              endif
            endif
            a(i,j)=sum*a(l,ncp1)
   60       continue
         end do
      end do
! end ludapb / begin luelpb
! solution ly = b
      nc1=nc+1
      iw=0
      l=0
      do i=1,nm2
         sum=a(i,4)
         if(nc .gt. 0)then
           if(iw .ne. 0)then
             l=l+1
             if(l .gt. nc) l=nc
               k=nc1-l
               kl=i-l
               do j=k,nc
                 sum=sum-a(kl,4)*a(i,j)
                 kl=kl+1
               end do
               else
               if(sum .ne. 0.d0)iw=1
               endif
             endif
             a(i,4)=sum*a(i,nc1)
      end do
! solution ux = y
      a(nm2,4)=a(nm2,4)*a(nm2,nc1)
      n1=nm2+1
      do i=2,nm2
         k=n1-i
         sum=a(k,4)
         if(nc .gt. 0)then
           kl=k+1
           k1=min0(nm2,k+nc)
           l=1
           do j=kl,k1
             sum=sum-a(j,4)*a(j,nc1-l)
             l=l+1
           end do 
         endif
         a(k,4)=sum*a(k,nc1)
      end do
! end luelpb
      do i=3,nm2
         res(i)=a(i-2,4)+c1(4)*a(i-1,4)+a(i,4)
      end do
      res(1)=a(1,4)
      res(2)=c1(4)*a(1,4)+a(2,4)
      res(n-1)=a(nm2-1,4)+c1(4)*a(nm2,4)
      res(n)=a(nm2,4)
      do i=1,n
         res(i)=y(i)-res(i)
      end do
      return
      end subroutine ads_f

### Given a suitable wood completeness data.frame wc, creates a pith
### offset data.frame in the format required by rcs.
wc.to.po <- function(wc){
    n <- nrow(wc)
    pith <- wc$pith.presence
    if(is.null(pith))
        pith <- rep(as.character(NA), n)
    missing <- wc$n.missing.heartwood
    if(is.null(missing))
        missing <- rep(as.integer(NA), n)
    unmeasured <- wc$n.unmeasured.inner
    if(is.null(unmeasured))
        unmeasured <- rep(as.integer(NA), n)
    not.na <- which(pith %in% c("complete", "incomplete") |
                    !is.na(missing))
    pith.offset <- rep(as.integer(NA), n)
    pith.offset[not.na] <-
        as.integer(rowSums(cbind(missing[not.na], unmeasured[not.na]),
                           na.rm = TRUE) + 1)

    data.frame(series = row.names(wc),
               pith.offset)
}

### Utility function. Creates unique composite titles by pasting
### together titles in the four columns in 'titles'. Giving a matching
### numeric table as the optional parameter 'ids' may speed up the
### computation.  Uniqueness of titles is guaranteed by a call to
### fix.names.
create.composite.titles <- function(titles, ids=titles) {
    composite.titles <- character(length = dim(titles)[1])
    titles.used <- logical(4)
    rle.1 <- rle(ids[, 1])
    cs.1 <- cumsum(rle.1$lengths)
    n.bins <- length(cs.1)
    ## Use first level (tree) titles, if >1 unique exist
    titles.used[1] <- n.bins > 1
    if (titles.used[1]) {
        composite.titles <- paste0(composite.titles, titles[, 1])
    }
    for (i in seq_len(n.bins)) {
        n <- rle.1$lengths[i]
        idx.end <- cs.1[i]
        idx.start <- idx.end - n + 1
        rle.2 <- rle(ids[idx.start:idx.end, 2])
        cs.2 <- cumsum(rle.2$lengths) + idx.start - 1
        n.bins <- length(cs.2)
        ## Use second level (core) titles, if >1 unique exist
        titles.used[2] <- n.bins > 1
        if (titles.used[2]) {
            composite.titles[idx.start:idx.end] <-
                paste0(composite.titles[idx.start:idx.end],
                       titles[idx.start:idx.end, 2])
        }
        for (j in seq_len(n.bins)) {
            n <- rle.2$lengths[j]
            idx.end <- cs.2[j]
            idx.start <- idx.end - n + 1
            rle.3 <- rle(ids[idx.start:idx.end, 3])
            cs.3 <- cumsum(rle.3$lengths) + idx.start - 1
            n.bins <- length(cs.3)
            ## Use third level (radius) titles, if >1 unique exist
            titles.used[3] <- n.bins > 1
            if (titles.used[3]) {
                composite.titles[idx.start:idx.end] <-
                    paste0(composite.titles[idx.start:idx.end],
                           titles[idx.start:idx.end, 3])
            }
            previous.titles.not.used <- !any(titles.used[c(1, 2, 3)])
            for (k in seq_len(n.bins)) {
                n <- rle.3$lengths[k]
                idx.end <- cs.3[k]
                idx.start <- idx.end - n + 1
                rle.4 <- rle(ids[idx.start:idx.end, 4])
                cs.4 <- cumsum(rle.4$lengths) + idx.start - 1
                n.bins <- length(cs.4)
                ## Use fourth level (measurement) titles, if >1
                ## unique exist, or if 1st to 3rd level titles
                ## have not been used
                titles.used[4] <- n.bins > 1 || previous.titles.not.used
                if (titles.used[4]) {
                    composite.titles[idx.start:idx.end] <-
                        paste0(composite.titles[idx.start:idx.end],
                               titles[idx.start:idx.end, 4])
                }
            }
        }
    }
    composite.titles <- fix.names(composite.titles,
                                  basic.charset=FALSE)
}

### Utility function. Converts .site.id and .site.title lists to a
### data.frame with one column for each level of ids or titles.
site.info.to.df <- function(x, name.prefix=NULL) {
    x.length <- length(x)
    if (x.length > 0) {
        item.length <- vapply(x, length, 0) # assumed that all are > 0
        max.length <- max(item.length)
        one.na <- NA
        mode(one.na) <- mode(x[[1]]) # works for numeric and character
        y <- array(rep(one.na, x.length*max.length),
                   dim=c(x.length, max.length))
        for (i in seq_len(x.length)) {
            y[i, seq_len(item.length[i])] <- x[[i]]
        }
        y <- data.frame(y)
        if (!is.null(name.prefix)) {
            if (max.length == 1) {
                names(y) <- name.prefix
            } else {
                names(y) <- paste0(name.prefix, ".", seq_len(max.length))
            }
        }
    } else {
        y <- data.frame() # empty data.frame
    }
    y
}

### Utility function. Creates a hierarchy of [element, sample,
### radius, measurementSeries] (in dplR: [tree, core, radius,
### measurement]) IDs based on unique titles, disregarding the
### nesting structure of the XML elements in the TRiDaS file.
title.based.ids <- function(titles) {
    ## Assumes four columns in titles1 and titles2. The
    ## function could be more generic, but it would make the
    ## design more complicated.
    titles2 <- as.matrix(titles)
    ids <- array(as.numeric(NA), dim(titles2))
    colnames(ids) <- colnames(titles2)
    unique1 <- unique(titles2[, 1])
    ids[, 1] <- match(titles2[, 1], unique1)
    for (i1 in seq_along(unique1)) {
        idx.1 <- which(ids[, 1] == i1)
        unique2 <- unique(titles2[idx.1, 2])
        ids[idx.1, 2] <- match(titles2[idx.1, 2], unique2)
        for (i2 in seq_along(unique2)) {
            idx.2 <- idx.1[which(ids[idx.1, 2] == i2)]
            unique3 <- unique(titles2[idx.2, 3])
            ids[idx.2, 3] <- match(titles2[idx.2, 3], unique3)
            for (i3 in seq_along(unique3)) {
                idx.3 <- idx.2[which(ids[idx.2, 3] == i3)]
                unique4 <- unique(titles2[idx.3, 4])
                ids[idx.3, 4] <- match(titles2[idx.3, 4], unique4)
            }
        }
    }
    ids
}

### Utility function. Copies [tree, core, radius, measurement] IDs so
### that they match in each set of series with the same [identifier,
### domain] pair (copies IDs from the first member of the set to the
### rest).
identifier.based.ids <- function(ids, identifiers, domains) {
    ids.out <- ids
    unique.domains <- unique(domains)
    for (ud in unique.domains) {
        if (is.na(ud)) {
            mask <- which(is.na(domains))
        } else {
            mask <- which(domains == ud)
        }
        unique.identifiers <- unique(identifiers[mask])
        unique.identifiers <-
            unique.identifiers[!is.na(unique.identifiers)]
        for (ui in unique.identifiers) {
            idx <- mask[identifiers[mask] == ui]
            for (k in inc(2, length(idx))) {
                ids.out[idx[k], ] <- ids.out[idx[1], ]
            }
        }
    }
    ids.out
}

### Main function (exported)
read.tridas <- function(fname, ids.from.titles=FALSE,
                        ids.from.identifiers=TRUE, combine.series=TRUE,
                        trim.whitespace=TRUE, warn.units=TRUE) {

    ## Returns a list of handler functions usable by xmlEventParse
    handler.factory <- function() {
        ## Multipliers and divisors for converting values to millimetres,
        ## which is the internal format of dplR
        MULTIPLIERS <- c("centimetres"=10, "metres"=1000)
        DIVISORS <- c("micrometres"=1000, "1/100th millimetres"=100,
                      "1/50th millimetres"=50,
                      "1/20th millimetres"=20,
                      "1/10th millimetres"=10)
        MISSING.VALUE <- 0 # code used for missing values in dplR
        FIVE.COLS <- c("text", "lang", "normal", "normalId", "normalStd")
        SIX.COLS <- c(FIVE.COLS, "normalTridas")
        ID.ORDER <- c("tree", "core", "radius", "measurement")
        year.now <- as.numeric(format(Sys.Date(), "%Y"))

        ## Result variables, will be massaged and put in a list by get.results
        res.df <- res.ids <- res.titles <- res.wc <- list()
        res.unit <- character(0)
        res.project.title <- character(0)
        res.var <- array(character(0), dim=c(0, 6),
                         dimnames=list(NULL, SIX.COLS))
        res.undated.var <-
            array(character(0), dim=c(0, 6), dimnames=list(NULL, SIX.COLS))
        res.taxon <-
            array(character(0), dim=c(0, 5), dimnames=list(NULL, FIVE.COLS))
        res.undated.taxon <-
            array(character(0), dim=c(0, 5), dimnames=list(NULL, FIVE.COLS))
        res.derived.var <-
            array(character(0), dim=c(0, 6), dimnames=list(NULL, SIX.COLS))
        res.project.id <- numeric(0)
        res.site.id <- res.site.title <- list()
        res.undated$data <- res.undated <- list()
        res.undated$unit <- character(0)
        res.undated$ids <-
            array(numeric(0), dim=c(0, 4), dimnames=list(NULL, ID.ORDER))
        res.undated$titles <-
            array(character(0), dim=c(0, 4), dimnames=list(NULL, ID.ORDER))
        res.undated$project.id <- numeric(0)
        res.undated$project.title <- character(0)
        res.undated$site.id <- res.undated$site.title <- list()
        res.undated.pith.presence <- character(0)
        res.undated.heartwood.presence <- character(0)
        res.undated.sapwood.presence <- character(0)
        res.undated.bark.presence <- character(0)
        res.undated.last.ring.presence <- character(0)
        res.undated.last.ring.details <- character(0)
        res.undated.n.sapwood <- res.undated.n.missing.sapwood <- integer(0)
        res.undated.n.missing.heartwood <- integer(0)
        res.undated.n.unmeasured.inner <- integer(0)
        res.undated.n.unmeasured.outer <- integer(0)
        res.undated.missing.heartwood.foundation <- character(0)
        res.undated.missing.sapwood.foundation <- character(0)
        res.derived$link <- res.derived$data <- res.derived <- list()
        res.derived$id <- res.derived$project.id <- numeric(0)
        res.derived$project.title <- res.derived$title <- character(0)
        res.derived$unit <- character(0)
        res.derived$standardizing.method <- character(0)
        res.lab <- list()
        res.research <- list()

        ## Storage that will be combined and returned as part of the
        ## result list
        comments.text <- character(0)
        comments.project.id <- comments.tree.id <- numeric(0)
        comments.site.id <- comments.site.title <- list()
        comments.core.id <- comments.radius.id <- numeric(0)
        comments.measurement.id <- comments.derived.id <- numeric(0)
        comments.project.title <- comments.tree.title <- character(0)
        comments.core.title <- comments.radius.title <- character(0)
        comments.measurement.title <- comments.derived.title <- character(0)
        type.text <- character(0)
        type.lang <- character(0)
        type.normal <- character(0)
        type.normalId <- character(0)
        type.normalStd <- character(0)
        type.project.id <- type.tree.id <- numeric(0)
        type.site.id <- type.site.title <- list()
        type.core.id <- type.derived.id <- numeric(0)
        type.project.title <- type.tree.title <- character(0)
        type.core.title <- type.derived.title <- character(0)
        identifier.text <- identifier.domain <- character(0)
        identifier.project.id <- identifier.tree.id <- numeric(0)
        identifier.site.id <- identifier.site.title <- list()
        identifier.core.id <- identifier.radius.id <- numeric(0)
        identifier.measurement.id <- identifier.derived.id <- numeric(0)
        identifier.project.title <- identifier.tree.title <- character(0)
        identifier.core.title <- identifier.radius.title <- character(0)
        identifier.measurement.title <- character(0)
        identifier.derived.title <- character(0)
        remark.data.text <- remark.undated.text <- character(0)
        remark.data.frame <- remark.data.row <- remark.data.col <- numeric(0)
        remark.undated.series <- remark.undated.idx <- numeric(0)
        remark.derived.text <- character(0)
        remark.derived.series <- remark.derived.idx <- numeric(0)
        altitude.metres <- numeric(0)
        altitude.project.id <- altitude.tree.id <- numeric(0)
        altitude.project.title <- altitude.tree.title <- character(0)
        altitude.site.id <- altitude.site.title <- list()
        preferred.site.id <- preferred.site.title <- list()
        preferred.idRef <- preferred.xLink <- character(0)
        preferred.domain <- preferred.project.title <- character(0)
        preferred.tree.title <- preferred.identifier <- character(0)
        preferred.project.id <- preferred.tree.id <- numeric(0)

        ## Temporary storage for communication between the handler functions
        unit.converted <- FALSE
        entities <- character(0)
        text.buffer <- domain.text <- ""
        tag.stack <- character(10) # approx. depth of TRiDaS document tree
        last.closed <- as.character(NA)
        stack.pointer <- idx.project <- 0
        first.dplr <- firstYear.suffix <- last.dplr <- lastYear.suffix <- NULL
        project.title <- object.title <- element.title <- NULL
        sample.title <- radius.title <- series.title <- NULL
        these.ids <- these.titles <- value.vector <- count.vector <- NULL
        idx.object <- idx.element <- idx.sample <- idx.radius <- NULL
        idx.series <- idx.value <- idx.derived <- NULL
        taxon.lang <- taxon.normal <- taxon.normalId <- taxon.normalStd <- NULL
        variable.lang <- variable.normal <- variable.normalId <- NULL
        variable.normalStd <- variable.normalTridas <- NULL
        site.data <- ids.in.site <- titles.in.site <- taxon <- variable <- NULL
        site.taxon <- site.var <- site.unit <- NULL
        site.n.sapwood <- site.n.missing.sapwood <- NULL
        site.n.missing.heartwood <- NULL
        site.n.unmeasured.inner <- NULL
        site.n.unmeasured.outer <- NULL
        site.missing.heartwood.foundation <- NULL
        site.missing.sapwood.foundation <- NULL
        site.pith.presence <- NULL
        site.heartwood.presence <- NULL
        site.sapwood.presence <- NULL
        site.bark.presence <- NULL
        site.last.ring.presence <- NULL
        site.last.ring.details <- NULL
        names.simple <- c("site.bark.presence", "site.last.ring.presence")
        names.complex <- c("site.pith.presence", "site.heartwood.presence",
                           "site.sapwood.presence")
        names.sum <- c("n.unmeasured.inner", "n.missing.sapwood", "n.sapwood",
                       "n.missing.heartwood", "n.unmeasured.outer")
        names.paste.unique <- c("site.last.ring.details",
                                "site.missing.heartwood.foundation",
                                "site.missing.sapwood.foundation")
        ordered.simple <- tridas.vocabulary(category = "presence / absence")
        ordered.complex <-
            tridas.vocabulary(category = "complex presence / absence")

        first.year <- last.year <- unit <- stdizing.method <- NULL
        link.idRef <- link.xLink <- link.identifier <- NULL
        link.idRefs <- link.xLinks <- link.domains <- link.identifiers <- NULL
        unitless <- in.derived.values <- FALSE
        lab.name <- lab.names <- lab.acronym <- lab.acronyms <- NULL
        lab.identifier <- lab.identifiers <- lab.domains <- NULL
        lab.addressLine1 <- lab.addressLine1s <- NULL
        lab.addressLine2 <- lab.addressLine2s <- NULL
        lab.cityOrTown <- lab.cityOrTowns <- NULL
        lab.stateProvinceRegion <- lab.stateProvinceRegions <- NULL
        lab.postalCode <- lab.postalCodes <- NULL
        lab.country <- lab.countries <- NULL
        research.identifier <- research.identifiers <- NULL
        research.domains <- NULL
        research.description <- research.descriptions <- NULL
        remarks.handled <- undated.handled <- 0
        object.level <- NULL
        pith.presence <- heartwood.presence <- NULL
        sapwood.presence <- bark.presence <- NULL
        n.unmeasured.inner <- n.unmeasured.outer <- NULL
        n.missing.heartwood <- n.missing.sapwood <- n.sapwood <- NULL
        last.ring.presence <- NULL
        last.ring.details <- NULL
        missing.heartwood.foundation <- missing.sapwood.foundation <- NULL
        values.multiplier <- values.divisor <- values.n.remarks <- NULL
        remark.data.taxon <-
            array(character(0), dim=c(0, 5), dimnames=list(NULL, FIVE.COLS))
        remark.data.var <-
            array(character(0), dim=c(0, 6), dimnames=list(NULL, SIX.COLS))
        remark.data.unit <- character(0)

        ## Putting the function here enables references to it in the list
        if (trim.whitespace) {
            text.function <- function(content, ...) {
                text.buffer <<- gsub("[[:space:]]+", " ",
                                     paste0(text.buffer, content))
            }
        } else {
            text.function <- function(content, ...) {
                text.buffer <<- paste0(text.buffer, content)
            }
        }
        ## A function called from all end tag handlers.  Makes the
        ## code shorter but increases running time.
        end.element <- function(name, ...) {
            text.buffer <<- ""
            stack.pointer <<- stack.pointer - 1
            last.closed <<- name
        }

        ## Normally, measurement series are identified by their
        ## location in the document tree. We have user selectable
        ## options in read.tridas for completely (ids.from.titles) or
        ## partially (ids.from.identifiers) overriding the default ID
        ## numbering based on the titles and / or the optional
        ## identifiers of the series, respectively. The changed IDs
        ## also affect the labeling of some metadata (comments,
        ## identifier, preferred, altitude, type). The work is done by
        ## this function. There's no input/output: some variables
        ## outside the local scope are modified instead.
        if (ids.from.titles || ids.from.identifiers) {
            alternative.ids <- function() {
                n.undated <- length(res.undated$site.id)
                undated.in.site <- inc(undated.handled + 1, n.undated)
                titles.in.undated <-
                    res.undated$titles[undated.in.site, , drop=FALSE]
                ids.old <- rbind(ids.in.site,
                                 res.undated$ids[undated.in.site, ,
                                                 drop=FALSE])
                n.dated.in.site <- nrow(titles.in.site)
                n.all <- n.dated.in.site + nrow(titles.in.undated)
                if (ids.from.titles) {
                    converted.ids <-
                        title.based.ids(rbind(titles.in.site,
                                              titles.in.undated))
                } else {
                    converted.ids <- ids.old
                }
                if (ids.from.identifiers) {
                    md.in.site <- which(identifier.project.id == idx.project)
                    if (length(md.in.site) > 0) {
                        md.in.site <-
                            md.in.site[!is.na(identifier.tree.id[md.in.site])]
                    }
                    if (length(md.in.site) > 0) {
                        md.in.site <-
                            md.in.site[vapply(identifier.site.id[md.in.site],
                                              function(x) {
                                                  if (is.na(x[1])) {
                                                      -1
                                                  } else {
                                                      length(x)
                                                  }
                                              }, 0) == object.level]
                    }
                    if (length(md.in.site) > 0) {
                        md.in.site <-
                            md.in.site[vapply(identifier.site.id[md.in.site],
                                              function(x) {
                                                  all(x ==
                                                      idx.object[seq_len(object.level)])
                                              }, TRUE)]
                    }
                    if (length(md.in.site) > 0) {
                        temp.identifiers <- rep(as.character(NA), n.all)
                        temp.domains <- rep(as.character(NA), n.all)
                        for (md.idx in md.in.site) {
                            this.idvec <- c(identifier.tree.id[md.idx],
                                            identifier.core.id[md.idx],
                                            identifier.radius.id[md.idx],
                                            identifier.measurement.id[md.idx])
                            if (all(!is.na(this.idvec))) {
                                this.match <-
                                    row.match(ids.old[, ID.ORDER, drop=FALSE],
                                              this.idvec)
                                temp.identifiers[this.match] <-
                                    identifier.text[md.idx]
                                temp.domains[this.match] <-
                                    identifier.domain[md.idx]
                            }
                        }
                        converted.ids <-
                            identifier.based.ids(converted.ids,
                                                 temp.identifiers,
                                                 temp.domains)
                    }
                }
                ## Only need to remap ids in metadata if ids in data change
                if (ids.from.titles || length(md.in.site) > 0) {
                    ids.in.site <<-
                        converted.ids[seq_len(n.dated.in.site), , drop=FALSE]
                    res.undated$ids[undated.in.site, ] <<-
                        converted.ids[inc(n.dated.in.site + 1, n.all), ,
                                      drop=FALSE]
                    metadata.names <- c("comments", "identifier", "preferred",
                                        "altitude", "type")
                    metadata.levels <- c(4, 4, 1, 1, 2)
                    for (k in seq_along(metadata.names)) {
                        md.name <- metadata.names[k]
                        md.level <- metadata.levels[k]
                        md.project.id <-
                            get(paste0(md.name, ".project.id"))
                        md.site.id <- get(paste0(md.name, ".site.id"))
                        md.tree.id <- get(paste0(md.name, ".tree.id"))
                        if (md.level >= 2) {
                            md.core.id <-
                                get(paste0(md.name, ".core.id"))
                            if (md.level >= 3) {
                                md.radius.id <-
                                    get(paste0(md.name, ".radius.id"))
                                if (md.level >= 4) {
                                    md.measurement.id <-
                                        get(paste0(md.name,
                                                   ".measurement.id"))
                                }
                            }
                        }
                        md.in.site <- which(md.project.id == idx.project)
                        if (length(md.in.site) > 0) {
                            md.in.site <-
                                md.in.site[!is.na(md.tree.id[md.in.site])]
                        }
                        if (length(md.in.site) > 0) {
                            md.in.site <-
                                md.in.site[vapply(md.site.id[md.in.site],
                                                  function(x) {
                                                      if (is.na(x[1])) {
                                                          -1
                                                      } else {
                                                          length(x)
                                                      }
                                                  }, 0) == object.level]
                        }
                        if (length(md.in.site) > 0) {
                            md.in.site <-
                                md.in.site[vapply(md.site.id[md.in.site],
                                                  function(x) {
                                                      all(x == idx.object[seq_len(object.level)])
                                                  }, TRUE)]
                        }
                        for (md.idx in md.in.site) {
                            this.idvec <- md.tree.id[md.idx]
                            if (md.level >= 2) {
                                this.idvec <- c(this.idvec, md.core.id[md.idx])
                                if (md.level >= 3) {
                                    this.idvec <-
                                        c(this.idvec, md.radius.id[md.idx])
                                    if (md.level >= 4) {
                                        this.idvec <-
                                            c(this.idvec,
                                              md.measurement.id[md.idx])
                                    }
                                }
                            }
                            idx.notna <- which(!is.na(this.idvec))
                            this.idvec <- this.idvec[idx.notna]
                            this.match <-
                                row.match(ids.old[, ID.ORDER[idx.notna],
                                                  drop=FALSE],
                                          this.idvec)
                            match.vec <- converted.ids[this.match[1], ]
                            md.tree.id[md.idx] <- match.vec["tree"]
                            assign(paste0(md.name, ".tree.id"),
                                   md.tree.id,
                                   inherits=TRUE)
                            if (md.level >= 2) {
                                md.core.id[md.idx] <- match.vec["core"]
                                assign(paste0(md.name, ".core.id"),
                                       md.core.id,
                                       inherits=TRUE)
                                if (md.level >= 3) {
                                    md.radius.id[md.idx] <- match.vec["radius"]
                                    assign(paste0(md.name, ".radius.id"),
                                           md.radius.id,
                                           inherits=TRUE)
                                    if (md.level >= 4) {
                                        md.measurement.id[md.idx] <-
                                            match.vec["measurement"]
                                        assign(paste0(md.name,
                                                      ".measurement.id"),
                                               md.measurement.id,
                                               inherits=TRUE)
                                    }
                                }
                            }
                        }
                    }
                }
                undated.handled <<- n.undated
            }
        }

        ## Utility function. Combines measurementSeries with the same IDs.
        ## Also adjusts the metadata of any possible remarks related to
        ## the affected series.
        if (combine.series) {
            series.combiner <- function(i.i.s, t.i.s, df.ncol, f.y, l.y,
                                        remark.series, idx.unittaxvar) {
                unique.ids <- unique(i.i.s)
                n.unique <- nrow(unique.ids)
                if (n.unique < df.ncol) {
                    f.y2 <- f.y
                    l.y2 <- l.y
                    df.ncol2 <- n.unique
                    del.cols <- integer(0)
                    remark.series2 <- remark.series
                    for (k in seq_len(n.unique)) {
                        id <- unique.ids[k, , drop=FALSE]
                        idx.id <- row.match(i.i.s, id)
                        n.id <- length(idx.id)
                        if (n.id > 1) {
                            new.f.y <- f.y2[idx.id[1]]
                            new.l.y <- l.y2[idx.id[1]]
                            new.data <- site.data[[idx.unittaxvar[idx.id[1]]]]
                            del.cols <- c(del.cols, idx.id[2:n.id])
                            for (varname in names.simple) {
                                this.var <- get(varname)
                                this.subset <- this.var[idx.unittaxvar[idx.id]]
                                this.subset <- this.subset[!is.na(this.subset)]
                                if (length(this.subset) > 0) {
                                    found.labels <-
                                        which(ordered.simple %in% this.subset)
                                    if (length(found.labels) > 0) {
                                        max.label <-
                                            ordered.simple[max(found.labels)]
                                        this.var[idx.unittaxvar[idx.id[1]]] <-
                                            max.label
                                        assign(varname, this.var, inherits=TRUE)
                                    }
                                }
                            }
                            for (varname in names.complex) {
                                this.var <- get(varname)
                                this.subset <- this.var[idx.unittaxvar[idx.id]]
                                this.subset <- this.subset[!is.na(this.subset)]
                                if (length(this.subset) > 0) {
                                    found.labels <-
                                        which(ordered.complex %in% this.subset)
                                    if (length(found.labels) > 0) {
                                        max.label <-
                                            ordered.simple[max(found.labels)]
                                        this.var[idx.unittaxvar[idx.id[1]]] <-
                                            max.label
                                        assign(varname, this.var, inherits=TRUE)
                                    }
                                }
                            }
                            for (varname in names.sum) {
                                this.var <- get(varname)
                                this.subset <- this.var[idx.unittaxvar[idx.id]]
                                this.subset <- this.subset[!is.na(this.subset)]
                                if (length(this.subset) > 0) {
                                    this.var[idx.unittaxvar[idx.id[1]]] <-
                                        sum(this.subset)
                                    assign(varname, this.var, inherits=TRUE)
                                }
                            }
                            for (varname in names.paste.unique) {
                                this.var <- get(varname)
                                this.subset <- this.var[idx.unittaxvar[idx.id]]
                                this.subset <- this.subset[!is.na(this.subset)]
                                if (length(this.subset) > 0) {
                                    this.var[idx.unittaxvar[idx.id[1]]] <-
                                        paste(unique(this.subset),
                                              collapse=", ")
                                    assign(varname, this.var, inherits=TRUE)
                                }
                            }
                            for (l in 2:n.id) {
                                this.f.y <- f.y2[idx.id[l]]
                                this.l.y <- l.y2[idx.id[l]]
                                this.data <-
                                    site.data[[idx.unittaxvar[idx.id[l]]]]
                                n.this.data <- length(this.data)
                                n.new.data <- length(new.data)
                                no.l <- c(seq_len(l - 1), inc(l + 1, n.id))
                                if (this.l.y < new.f.y) {
                                    stretch <- new.f.y - this.l.y - 1
                                    new.data <- c(this.data,
                                                  rep(MISSING.VALUE, stretch),
                                                  new.data)
                                    idx.adjust <-
                                        remark.series2 %in% idx.unittaxvar[idx.id[no.l]]
                                    remark.data.row[idx.adjust] <-
                                        remark.data.row[idx.adjust] +
                                            n.this.data + stretch
                                } else if (this.f.y > new.l.y) {
                                    stretch <- this.f.y - new.l.y - 1
                                    new.data <- c(new.data,
                                                  rep(MISSING.VALUE, stretch),
                                                  this.data)
                                    idx.adjust <-
                                        remark.series2 == idx.unittaxvar[idx.id[l]]
                                    remark.data.row[idx.adjust] <-
                                        remark.data.row[idx.adjust] +
                                            n.new.data + stretch
                                } else if (this.f.y >= new.f.y &&
                                           this.l.y <= new.l.y) {
                                    idx.missing <-
                                        which(new.data %in% MISSING.VALUE)
                                    n.slack <- this.f.y - new.f.y
                                    idx.replacement <- idx.missing - n.slack
                                    do.replace <-
                                        which(idx.replacement >= 1 &
                                              idx.replacement <= n.this.data)
                                    new.data[idx.missing[do.replace]] <-
                                        this.data[idx.replacement[do.replace]]
                                    idx.adjust <-
                                        remark.series2 == idx.unittaxvar[idx.id[l]]
                                    remark.data.row[idx.adjust] <-
                                        remark.data.row[idx.adjust] + n.slack
                                } else if (this.f.y >= new.f.y) {
                                    n.exceeding <- this.l.y - new.l.y
                                    n.slack <- this.f.y - new.f.y
                                    first.exceeding <-
                                        n.this.data - n.exceeding + 1
                                    idx.missing <-
                                        which(new.data %in% MISSING.VALUE)
                                    idx.replacement <- idx.missing - n.slack
                                    do.replace <- which(idx.replacement >= 1)
                                    new.data[idx.missing[do.replace]] <-
                                        this.data[idx.replacement[do.replace]]
                                    new.data <-
                                        c(new.data,
                                          this.data[first.exceeding:n.this.data])
                                    idx.adjust <-
                                        remark.series2 == idx.unittaxvar[idx.id[l]]
                                    remark.data.row[idx.adjust] <-
                                        remark.data.row[idx.adjust] + n.slack
                                } else if (this.l.y <= new.l.y) {
                                    n.exceeding <- new.f.y - this.f.y
                                    idx.missing <-
                                        which(new.data %in% MISSING.VALUE)
                                    idx.replacement <- idx.missing + n.exceeding
                                    do.replace <-
                                        which(idx.replacement <= n.this.data)
                                    new.data[idx.missing[do.replace]] <-
                                        this.data[idx.replacement[do.replace]]
                                    new.data <-
                                        c(this.data[seq_len(n.exceeding)],
                                          new.data)
                                    idx.adjust <-
                                        remark.series2 %in% idx.unittaxvar[idx.id[no.l]]
                                    remark.data.row[idx.adjust] <-
                                        remark.data.row[idx.adjust] +
                                            n.exceeding
                                } else {
                                    n.exceeding.1 <- new.f.y - this.f.y
                                    first.exceeding.2 <-
                                        n.exceeding.1 + n.new.data + 1
                                    idx.missing <-
                                        which(new.data %in% MISSING.VALUE)
                                    idx.replacement <-
                                        idx.missing + n.exceeding.1
                                    new.data[idx.missing] <-
                                        this.data[idx.replacement]
                                    new.data <-
                                        c(this.data[seq_len(n.exceeding.1)],
                                          new.data,
                                          this.data[first.exceeding.2:n.this.data])
                                    idx.adjust <-
                                        remark.series2 %in% idx.unittaxvar[idx.id[no.l]]
                                    remark.data.row[idx.adjust] <-
                                        remark.data.row[idx.adjust] +
                                            n.exceeding.1
                                }
                                new.f.y <- min(new.f.y, this.f.y)
                                new.l.y <- max(new.l.y, this.l.y)
                            }
                            idx.adjust <-
                                remark.series2 %in% idx.unittaxvar[idx.id[2:n.id]]
                            remark.series2[idx.adjust] <-
                                idx.unittaxvar[idx.id[1]]
                            site.data[[idx.unittaxvar[idx.id[1]]]] <<- new.data
                            f.y2[idx.id[1]] <- new.f.y
                            l.y2[idx.id[1]] <- new.l.y
                        }
                    }
                    site.data <<- site.data[-del.cols]
                    f.y2 <- f.y2[-del.cols]
                    l.y2 <- l.y2[-del.cols]
                    t.i.s2 <- t.i.s[-del.cols, , drop=FALSE]
                    i.i.s2 <- i.i.s[-del.cols, , drop=FALSE]
                    idx.unittaxvar2 <- idx.unittaxvar[-del.cols]
                    for (varname in c(names.simple, names.complex,
                                      names.sum, names.paste.unique))
                        assign(varname, get(varname)[-del.cols], inherits=TRUE)
                    list(i.i.s2, t.i.s2, df.ncol2, f.y2, l.y2,
                         remark.series2, idx.unittaxvar2)
                } else {
                    list(i.i.s, t.i.s, df.ncol, f.y, l.y,
                         remark.series, idx.unittaxvar)
                }
            }
        }

        wc.names <-
            c("pith.presence", "heartwood.presence", "sapwood.presence",
              "last.ring.presence", "last.ring.details", "bark.presence",
              "n.sapwood", "n.missing.heartwood", "n.missing.sapwood",
              "missing.heartwood.foundation","missing.sapwood.foundation",
              "n.unmeasured.inner", "n.unmeasured.outer")

### The list of handler functions
        list(# Handlers for start tags
             bark = function(x, atts, ...) {
                 grandparent.element <- tag.stack[stack.pointer - 1]
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "bark"
                 if (!is.null(atts)) {
                     if (grandparent.element == "measurementSeries") {
                         bark.presence[2] <<- atts["presence"]
                     } else if (grandparent.element == "radius") {
                         bark.presence[1] <<- atts["presence"]
                     }
                 }
             },
             derivedSeries = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<-
                     "derivedSeries"
                 firstYear.suffix <<- lastYear.suffix <<- as.character(NA)
                 series.title <<- stdizing.method <<- as.character(NA)
                 idx.derived <<- idx.derived + 1
                 first.dplr <<- last.dplr <<- as.numeric(NA)
                 link.idRefs <<- character(0)
                 link.xLinks <<- character(0)
                 link.domains <<- character(0)
                 link.identifiers <<- character(0)
             },
             element = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "element"
                 element.title <<- as.character(NA)
                 taxon <<- taxon.lang <<- taxon.normal <<- as.character(NA)
                 taxon.normalId <<- taxon.normalStd <<- as.character(NA)
                 idx.element <<- idx.element + 1
                 idx.sample <<- idx.radius <<- idx.series <<- 0
                 ## Coming down the hierarchy of <object>s:
                 ## the case of an <object> with both <object>s and <element>s
                 if (last.closed == "object") {
                     object.level <<- object.level - 1
                 }
             },
             firstYear = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "firstYear"
                 firstYear.suffix <<- if (is.null(atts)) {
                     as.character(NA)
                 } else {
                     atts["suffix"]
                 }
             },
             heartwood = function(x, atts, ...) {
                 grandparent.element <- tag.stack[stack.pointer - 1]
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "heartwood"
                 if (!is.null(atts)) {
                     if (grandparent.element == "measurementSeries") {
                         heartwood.presence[2] <<- atts["presence"]
                     } else if (grandparent.element == "radius") {
                         heartwood.presence[1] <<- atts["presence"]
                     }
                 }
             },
             identifier = function(x, atts, ...){
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "identifier"
                 domain.text <<- if (is.null(atts)) {
                     as.character(NA)
                 } else {
                     atts["domain"]
                 }
             },
             ## A reference to an identifier in the same document
             idRef = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "idRef"
                 link.idRef <<- if (is.null(atts)) {
                     as.character(NA)
                 } else {
                     atts["ref"]
                 }
             },
             laboratory = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "laboratory"
                 domain.text <<- lab.identifier <<- as.character(NA)
                 lab.name <<- lab.acronym <<- as.character(NA)
                 lab.addressLine1 <<- lab.addressLine2 <<- as.character(NA)
                 lab.cityOrTown <<- lab.stateProvinceRegion <<- as.character(NA)
                 lab.postalCode <<- lab.country <<- as.character(NA)
             },
             lastRingUnderBark = function(x, atts, ...) {
                 grandparent.element <- tag.stack[stack.pointer - 1]
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<-
                     "lastRingUnderBark"
                 if (!is.null(atts)) {
                     if (grandparent.element == "measurementSeries") {
                         last.ring.presence[2] <<- atts["presence"]
                     } else if (grandparent.element == "radius") {
                         last.ring.presence[1] <<- atts["presence"]
                     }
                 }
             },
             lastYear = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "lastYear"
                 lastYear.suffix <<- if (is.null(atts)) {
                     as.character(NA)
                 } else {
                     atts["suffix"]
                 }
             },
             measurementSeries = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<-
                     "measurementSeries"
                 firstYear.suffix <<- lastYear.suffix <<- as.character(NA)
                 series.title <<- as.character(NA)
                 idx.series <<- idx.series + 1
                 these.ids <<- array(c(idx.element,
                                       idx.sample,
                                       idx.radius,
                                       idx.series), dim=c(1,4))
                 first.dplr <<- last.dplr <<- as.numeric(NA)
             },
             name = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "name"
                 lab.acronym <<- if (is.null(atts)) {
                     as.character(NA)
                 } else {
                     atts["acronym"]
                 }
             },
             object = function(...) {
                 parent.element <- tag.stack[stack.pointer]
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "object"
                 ## Deeper in the hierarchy of <object> elements
                 if (parent.element == "object" && last.closed != "object") {
                     object.level <<- object.level + 1
                     idx.object[object.level] <<- 1
                 } else {
                     idx.object[object.level] <<- idx.object[object.level] + 1
                 }
                 object.title[object.level] <<- as.character(NA)
                 idx.element <<- idx.sample <<- idx.radius <<- idx.series <<- 0
                 element.title <<- as.character(NA)
                 site.data <<- list()
                 site.n.sapwood <<- site.n.missing.sapwood <<- integer(0)
                 site.n.missing.heartwood <<- integer(0)
                 site.n.unmeasured.inner <<- integer(0)
                 site.n.unmeasured.outer <<- integer(0)
                 site.missing.heartwood.foundation <<- character(0)
                 site.missing.sapwood.foundation <<- character(0)
                 site.pith.presence <<- character(0)
                 site.heartwood.presence <<- character(0)
                 site.sapwood.presence <<- character(0)
                 site.bark.presence <<- character(0)
                 site.last.ring.presence <<- character(0)
                 site.last.ring.details <<- character(0)
                 titles.in.site <<-
                     array(character(0), dim=c(0,4),
                           dimnames=list(NULL, ID.ORDER))
                 site.taxon <<-
                     array(character(0), dim=c(0,5),
                           dimnames=list(NULL, FIVE.COLS))
                 site.var <<-
                     array(character(0), dim=c(0,6),
                           dimnames=list(NULL, SIX.COLS))
                 site.unit <<- character(0)
                 ids.in.site <<-
                     array(numeric(0), dim=c(0,4),
                           dimnames=list(NULL, ID.ORDER))
                 first.year <<- last.year <<- numeric(0)
             },
             pith = function(x, atts, ...) {
                 grandparent.element <- tag.stack[stack.pointer - 1]
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "pith"
                 if (!is.null(atts)) {
                     if (grandparent.element == "measurementSeries") {
                         pith.presence[2] <<- atts["presence"]
                     } else if (grandparent.element == "radius") {
                         pith.presence[1] <<- atts["presence"]
                     }
                 }
             },
             preferredSeries = function(x, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<-
                     "preferredSeries"
                 link.idRef <<- link.xLink <<- as.character(NA)
                 domain.text <<- link.identifier <<- as.character(NA)
             },
             project = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "project"
                 project.title <<- as.character(NA)
                 idx.project <<- idx.project + 1
                 idx.object <<- idx.element <<- idx.sample <<- 0
                 idx.radius <<- idx.series <<- idx.derived <<- 0
                 lab.domains <<- lab.identifiers <<- lab.names <<- character(0)
                 lab.acronyms <<- character(0)
                 lab.addressLine1s <<- lab.addressLine2s <<- character(0)
                 lab.cityOrTowns <<- lab.stateProvinceRegions <<- character(0)
                 lab.postalCodes <<- lab.countries <<- character(0)
                 research.domains <<- research.identifiers <<- character(0)
                 research.descriptions <<- character(0)
                 object.level <<- 1
             },
             radius = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "radius"
                 radius.title <<- as.character(NA)
                 idx.radius <<- idx.radius + 1
                 idx.series <<- 0
                 pith.presence <<- heartwood.presence <<- as.character(NA)
                 sapwood.presence <<- bark.presence <<- as.character(NA)
                 n.unmeasured.inner <<- n.unmeasured.outer <<- as.integer(NA)
                 n.missing.heartwood <<- n.missing.sapwood <<- as.integer(NA)
                 n.sapwood <<- as.integer(NA)
                 last.ring.presence <<- as.character(NA)
                 last.ring.details <<- as.character(NA)
                 missing.heartwood.foundation <<- as.character(NA)
                 missing.sapwood.foundation <<- as.character(NA)
             },
             research = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "research"
                 domain.text <<- research.identifier <<- as.character(NA)
                 research.description <<- as.character(NA)
             },
             sample = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "sample"
                 sample.title <<- as.character(NA)
                 idx.sample <<- idx.sample + 1
                 idx.radius <<- idx.series <<- 0
             },
             sapwood = function(x, atts, ...) {
                 grandparent.element <- tag.stack[stack.pointer - 1]
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "sapwood"
                 if (!is.null(atts)) {
                     if (grandparent.element == "measurementSeries") {
                         sapwood.presence[2] <<- atts["presence"]
                     } else if (grandparent.element == "radius") {
                         sapwood.presence[1] <<- atts["presence"]
                     }
                 }
             },
             series = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "series"
                 link.idRef <<- link.xLink <<- as.character(NA)
                 domain.text <<- link.identifier <<- as.character(NA)
             },
             taxon = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "taxon"
                 if (!is.null(atts)) {
                     for (attr in c("lang", "normal",
                                    "normalId", "normalStd")) {
                         attr.val <- atts[attr]
                         if (!is.null(attr.val)) {
                             assign(paste("taxon", attr, sep="."), attr.val,
                                    inherits = TRUE)
                         }
                     }
                 }
             },
             type = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "type"
                 if (!is.null(atts)) {
                     for (attr in c("lang", "normal",
                                    "normalId", "normalStd")) {
                         attr.val <- atts[attr]
                         if (!is.null(attr.val)) {
                             this.varname <- paste("type", attr, sep=".")
                             assign(this.varname,
                                    c(get(this.varname), attr.val),
                                    inherits = TRUE)
                         }
                     }
                 } else {
                     type.lang <<- c(type.lang, NA)
                     type.normal <<- c(type.normal, NA)
                     type.normalId <<- c(type.normalId, NA)
                     type.normalStd <<- c(type.normalStd, NA)
                 }
             },
             unit = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "unit"
                 if (!is.null(atts)) {
                     normal.unit <- atts["normalTridas"]
                     if (!is.na(normal.unit)) {
                         unit <<- normal.unit
                     } else {
                         unit <<- atts["normal"]
                     }
                 }
             },
             unitless = function(...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "unitless"
                 unitless <<- TRUE
             },
             value = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "value"
                 idx.value <<- idx.value + 1
                 no.atts <- is.null(atts)
                 if (no.atts) {
                     ## The value attribute is mandatory, but we just
                     ## quietly ignore this problem...
                     value.vector[idx.value] <<- MISSING.VALUE
                 } else {
                     this.val <- suppressWarnings(as.numeric(atts["value"]))
                     ## ...also here). Also suppresses the warning about
                     ## a string not interpretable as numeric.
                     if (is.na(this.val)) {
                         value.vector[idx.value] <<- MISSING.VALUE
                     } else {
                         value.vector[idx.value] <<- this.val
                     }
                 }
                 if (in.derived.values) {
                     count.vector[idx.value] <<- if (no.atts) {
                         as.numeric(NA)
                     } else {
                         as.numeric(atts["count"])
                     }
                 }
             },
             values = function(...) {
                 parent.element <- tag.stack[stack.pointer]
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "values"
                 if (is.na(first.dplr) || is.na(last.dplr)) {
                     ## An initial allocation of value.vector (and
                     ## count.vector).  If necessary, they will grow
                     ## automatically.  Both will later be trimmed to
                     ## proper length.
                     value.vector <<- double(length=1000)
                     if (parent.element == "derivedSeries") {
                         in.derived.values <<- TRUE
                         count.vector <<- double(length=1000)
                     }
                 } else if (last.dplr < first.dplr) {
                     stop(gettextf("in project %d, object %s, element %d, sample %d, radius %d, series %d: ",
                                   idx.project,
                                   paste(idx.object[seq_len(object.level)],
                                         collapse="."),
                                   idx.element,
                                   idx.sample,
                                   idx.radius,
                                   idx.series),
                          "lastYear < firstYear")
                 } else {
                     value.vector <<-
                         rep(as.numeric(NA),times=last.dplr - first.dplr + 1)
                     if (parent.element == "derivedSeries") {
                         in.derived.values <<- TRUE
                         count.vector <<- rep(as.numeric(NA),
                                              times=last.dplr - first.dplr + 1)
                     }
                 }
                 idx.value <<- 0
                 unit <<- variable <<- as.character(NA)
                 variable.lang <<- variable.normal <<- as.character(NA)
                 variable.normalId <<- as.character(NA)
                 variable.normalStd <<- as.character(NA)
                 variable.normalTridas <<- as.character(NA)
                 values.multiplier <<- values.divisor <<- as.numeric(NA)
                 values.n.remarks <<- 0
                 unitless <<- FALSE
                 unit.converted <<- FALSE
             },
             variable = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "variable"
                 if (!is.null(atts)) {
                     for (attr in c("lang", "normal", "normalId",
                                    "normalStd", "normalTridas")) {
                         attr.val <- atts[attr]
                         if (!is.null(attr.val)) {
                             assign(paste("variable", attr, sep="."), attr.val,
                                    inherits = TRUE)
                         }
                     }
                 }
             },
             ## A reference URI
             xLink = function(x, atts, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- "xLink"
                 link.xLink <<- if (is.null(atts)) {
                     as.character(NA)
                 } else {
                     atts["href"]
                 }
             },
             ## Other start tag
             .startElement = function(name, ...) {
                 tag.stack[stack.pointer <<- stack.pointer + 1] <<- name
             },
             ## (Internal general) entities declared ...
             .entityDeclaration = function(name, type, content, ...) {
                 entity.type <- names(type)
                 if (entity.type == "Internal_General") {
                     entities[name] <<- content
                 } else if (grep("^External", entity.type)) {
                     warning(gettextf("external entities are not supported: %s",
                                      name))
                 }
             },
             ## ... and used
             .getEntity = function(name, ...) {
                 if (is.na(ent <- entities[name])) {
                     warning(gettextf("unknown entity: %s", name))
                 }
                 ent
             },
### Handlers for end tags
             "/addressLine1" = function(...) {
                 lab.addressLine1 <<- text.buffer
                 end.element("addressLine1")
             },
             "/addressLine2" = function(...) {
                 lab.addressLine2 <<- text.buffer
                 end.element("addressLine2")
             },
             "/altitude" = function(...) {
                 new.length <- length(altitude.metres) + 1
                 altitude.metres <<-
                     c(altitude.metres, as.numeric(text.buffer))
                 altitude.project.id <<- c(altitude.project.id, idx.project)
                 altitude.project.title <<-
                     c(altitude.project.title, project.title)
                 seq.object <- seq_len(object.level)
                 altitude.site.id[[new.length]] <<- idx.object[seq.object]
                 altitude.site.title[[new.length]] <<- object.title[seq.object]
                 altitude.tree.id <<- c(altitude.tree.id, idx.element)
                 altitude.tree.title <<- c(altitude.tree.title, element.title)
                 end.element("altitude")
             },
             "/cityOrTown" = function(...) {
                 lab.cityOrTown <<- text.buffer
                 end.element("cityOrTown")
             },
             "/comments" = function(...) {
                 comments.text <<- c(comments.text, text.buffer)
                 comments.project.id <<- c(comments.project.id, idx.project)
                 comments.project.title <<- c(comments.project.title,
                                              project.title)
                 if (idx.derived > 0) {
                     comments.derived.id <<- c(comments.derived.id, idx.derived)
                     comments.derived.title <<- c(comments.derived.title,
                                                  series.title)
                     comments.site.id <<- c(comments.site.id, NA)
                     comments.site.title <<- c(comments.site.title, NA)
                     comments.tree.id <<- c(comments.tree.id, NA)
                     comments.tree.title <<- c(comments.tree.title, NA)
                     comments.core.id <<- c(comments.core.id, NA)
                     comments.core.title <<- c(comments.core.title, NA)
                     comments.radius.id <<- c(comments.radius.id, NA)
                     comments.radius.title <<- c(comments.radius.title, NA)
                     comments.measurement.id <<- c(comments.measurement.id, NA)
                     comments.measurement.title <<-
                         c(comments.measurement.title, NA)
                 } else {
                     comments.derived.id <<- c(comments.derived.id, NA)
                     comments.derived.title <<- c(comments.derived.title, NA)
                     if (idx.object[1] > 0) {
                         new.length <- length(comments.site.id) + 1
                         seq.object <- seq_len(object.level)
                         comments.site.id[[new.length]] <<-
                             idx.object[seq.object]
                         comments.site.title[[new.length]] <<-
                             object.title[seq.object]
                     } else {
                         comments.site.id <<- c(comments.site.id, NA)
                         comments.site.title <<- c(comments.site.title, NA)
                     }
                     if (idx.element > 0) {
                         comments.tree.id <<- c(comments.tree.id, idx.element)
                         comments.tree.title <<-
                             c(comments.tree.title, element.title)
                     } else {
                         comments.tree.id <<- c(comments.tree.id, NA)
                         comments.tree.title <<- c(comments.tree.title, NA)
                     }
                     if (idx.sample > 0) {
                         comments.core.id <<- c(comments.core.id, idx.sample)
                         comments.core.title <<-
                             c(comments.core.title, sample.title)
                     } else {
                         comments.core.id <<- c(comments.core.id, NA)
                         comments.core.title <<- c(comments.core.title, NA)
                     }
                     if (idx.radius > 0) {
                         comments.radius.id <<-
                             c(comments.radius.id, idx.radius)
                         comments.radius.title <<-
                             c(comments.radius.title, radius.title)
                     } else {
                         comments.radius.id <<- c(comments.radius.id, NA)
                         comments.radius.title <<-
                             c(comments.radius.title, NA)
                     }
                     if (idx.series > 0) {
                         comments.measurement.id <<-
                             c(comments.measurement.id, idx.series)
                         comments.measurement.title <<-
                             c(comments.measurement.title, series.title)
                     } else {
                         comments.measurement.id <<-
                             c(comments.measurement.id, NA)
                         comments.measurement.title <<-
                             c(comments.measurement.title, NA)
                     }
                 }
                 end.element("comments")
             },
             "/country" = function(...) {
                 lab.country <<- text.buffer
                 end.element("country")
             },
             "/derivedSeries" = function(...) {
                 if (length(link.idRefs) > 0) {
                     this.frame <-
                         data.frame(idRef = link.idRefs,
                                    xLink = link.xLinks,
                                    identifier = link.identifiers,
                                    domain = link.domains)
                     ## Remove unused columns
                     delete.idx <- which(apply(this.frame,
                                               2,
                                               function(x) all(is.na(x))))
                     this.frame[delete.idx] <-
                         rep(list(NULL), length(delete.idx))
                     res.derived$link[[length(res.derived$link) + 1]] <<-
                         this.frame
                 } else {
                     res.derived$link[[length(res.derived$link) + 1]] <<- NA
                 }
                 end.element("derivedSeries")
             },
             "/description" = function(...) {
                 research.description <<- text.buffer
                 end.element("description")
             },
             "/firstYear" = function(...) {
                 first.dplr <<- dplr.year(as.numeric(text.buffer),
                                          firstYear.suffix)
                 end.element("firstYear")
             },
             "/identifier" = function(...) {
                 parent.element <- tag.stack[stack.pointer - 1]
                 if (parent.element %in% c("series", "preferredSeries")) {
                     link.identifier <<- text.buffer
                 } else if (parent.element == "laboratory") {
                     lab.identifier <<- text.buffer
                 } else if (parent.element == "research") {
                     research.identifier <<- text.buffer
                 } else {
                     identifier.text <<- c(identifier.text, text.buffer)
                     identifier.domain <<- c(identifier.domain, domain.text)
                     identifier.project.id <<-
                         c(identifier.project.id, idx.project)
                     identifier.project.title <<-
                         c(identifier.project.title, project.title)
                     if (idx.derived > 0) {
                         identifier.derived.id <<-
                             c(identifier.derived.id, idx.derived)
                         identifier.derived.title <<-
                             c(identifier.derived.title, series.title)
                         identifier.site.id <<- c(identifier.site.id, NA)
                         identifier.site.title <<- c(identifier.site.title, NA)
                         identifier.tree.id <<- c(identifier.tree.id, NA)
                         identifier.tree.title <<- c(identifier.tree.title, NA)
                         identifier.core.id <<- c(identifier.core.id, NA)
                         identifier.core.title <<- c(identifier.core.title, NA)
                         identifier.radius.id <<- c(identifier.radius.id, NA)
                         identifier.radius.title <<-
                             c(identifier.radius.title, NA)
                         identifier.measurement.id <<-
                             c(identifier.measurement.id, NA)
                         identifier.measurement.title <<-
                             c(identifier.measurement.title, NA)
                     } else {
                         identifier.derived.id <<- c(identifier.derived.id, NA)
                         identifier.derived.title <<-
                             c(identifier.derived.title, NA)
                         if (idx.object[1] > 0) {
                             new.length <- length(identifier.site.id) + 1
                             seq.object <- seq_len(object.level)
                             identifier.site.id[[new.length]] <<-
                                 idx.object[seq.object]
                             identifier.site.title[[new.length]] <<-
                                 object.title[seq.object]
                         } else {
                             identifier.site.id <<- c(identifier.site.id, NA)
                             identifier.site.title <<-
                                 c(identifier.site.title, NA)
                         }
                         if (idx.element > 0) {
                             identifier.tree.id <<-
                                 c(identifier.tree.id, idx.element)
                             identifier.tree.title <<-
                                 c(identifier.tree.title, element.title)
                         } else {
                             identifier.tree.id <<- c(identifier.tree.id, NA)
                             identifier.tree.title <<-
                                 c(identifier.tree.title, NA)
                         }
                         if (idx.sample > 0) {
                             identifier.core.id <<-
                                 c(identifier.core.id, idx.sample)
                             identifier.core.title <<-
                                 c(identifier.core.title, idx.sample)
                         } else {
                             identifier.core.id <<- c(identifier.core.id, NA)
                             identifier.core.title <<-
                                 c(identifier.core.title, NA)
                         }
                         if (idx.radius > 0) {
                             identifier.radius.id <<-
                                 c(identifier.radius.id, idx.radius)
                             identifier.radius.title <<-
                                 c(identifier.radius.title, radius.title)
                         } else {
                             identifier.radius.id <<-
                                 c(identifier.radius.id, NA)
                             identifier.radius.title <<-
                                 c(identifier.radius.title, NA)
                         }
                         if (idx.series > 0) {
                             identifier.measurement.id <<-
                                 c(identifier.measurement.id, idx.series)
                             identifier.measurement.title <<-
                                 c(identifier.measurement.title, series.title)
                         } else {
                             identifier.measurement.id <<-
                                 c(identifier.measurement.id, NA)
                             identifier.measurement.title <<-
                                 c(identifier.measurement.title, NA)
                         }
                     }
                 }
                 end.element("identifier")
             },
             "/laboratory" = function(...) {
                 lab.domains <<- c(lab.domains, domain.text)
                 lab.identifiers <<- c(lab.identifiers, lab.identifier)
                 lab.names <<- c(lab.names, lab.name)
                 lab.acronyms <<- c(lab.acronyms, lab.acronym)
                 lab.addressLine1s <<- c(lab.addressLine1s, lab.addressLine1)
                 lab.addressLine2s <<- c(lab.addressLine2s, lab.addressLine2)
                 lab.cityOrTowns <<- c(lab.cityOrTowns, lab.cityOrTown)
                 lab.stateProvinceRegions <<-
                     c(lab.stateProvinceRegions, lab.stateProvinceRegion)
                 lab.postalCodes <<- c(lab.postalCodes, lab.postalCode)
                 lab.countries <<- c(lab.countries, lab.country)
                 end.element("laboratory")
             },
             "/lastRingUnderBark" = function(...) {
                 greatgrandparent.element <- tag.stack[stack.pointer - 3]
                 if (greatgrandparent.element == "measurementSeries") {
                     last.ring.details[2] <<- text.buffer
                 } else if (greatgrandparent.element == "radius") {
                     last.ring.details[1] <<- text.buffer
                 }
                 end.element("lastRingUnderBark")
             },
             "/lastYear" = function(...) {
                 last.dplr <<- dplr.year(as.numeric(text.buffer),
                                         lastYear.suffix)
                 end.element("lastYear")
             },
             "/measurementSeries" = function(...) {
                 for (this.name in wc.names) {
                     assign(this.name, get(this.name)[1], inherits = TRUE)
                 }
                 end.element("measurementSeries")
             },
             "/missingHeartwoodRingsToPith" = function(...) {
                 val <- as.integer(text.buffer)
                 if (val < 0) {
                     stop("negative missingHeartwoodRingsToPith")
                 }
                 greatgrandparent.element <- tag.stack[stack.pointer - 3]
                 if (greatgrandparent.element == "measurementSeries") {
                     n.missing.heartwood[2] <<- val
                 } else if (greatgrandparent.element == "radius") {
                     n.missing.heartwood[1] <<- val
                 }
                 end.element("missingHeartwoodRingsToPith")
             },
             "/missingHeartwoodRingsToPithFoundation" = function(...) {
                 greatgrandparent.element <- tag.stack[stack.pointer - 3]
                 if (greatgrandparent.element == "measurementSeries") {
                     missing.heartwood.foundation[2] <<- text.buffer
                 } else if (greatgrandparent.element == "radius") {
                     missing.heartwood.foundation[1] <<- text.buffer
                 }
                 end.element("missingHeartwoodRingsToPithFoundation")
             },
             "/missingSapwoodRingsToBark" = function(...) {
                 val <- as.integer(text.buffer)
                 if (val < 0) {
                     stop("negative missingSapwoodRingsToPith")
                 }
                 greatgrandparent.element <- tag.stack[stack.pointer - 3]
                 if (greatgrandparent.element == "measurementSeries") {
                     n.missing.sapwood[2] <<- val
                 } else if (greatgrandparent.element == "radius") {
                     n.missing.sapwood[1] <<- val
                 }
                 end.element("missingSapwoodRingsToBark")
             },
             "/missingSapwoodRingsToBarkFoundation" = function(...) {
                 greatgrandparent.element <- tag.stack[stack.pointer - 3]
                 if (greatgrandparent.element == "measurementSeries") {
                     missing.sapwood.foundation[2] <<- text.buffer
                 } else if (greatgrandparent.element == "radius") {
                     missing.sapwood.foundation[1] <<- text.buffer
                 }
                 end.element("missingSapwoodRingsToBarkFoundation")
             },
             "/name" = function(...) {
                 lab.name <<- text.buffer
                 end.element("name")
             },
             "/nrOfSapwoodRings" = function(...) {
                 val <- as.integer(text.buffer)
                 if (val < 0) {
                     stop("negative nrOfSapwoodRings")
                 }
                 greatgrandparent.element <- tag.stack[stack.pointer - 3]
                 if (greatgrandparent.element == "measurementSeries") {
                     n.sapwood[2] <<- val
                 } else if (greatgrandparent.element == "radius") {
                     n.sapwood[1] <<- val
                 }
                 end.element("nrOfSapwoodRings")
             },
             "/nrOfUnmeasuredInnerRings" = function(...) {
                 val <- as.integer(text.buffer)
                 if (val < 0) {
                     stop("negative nrOfUnmeasuredInnerRings")
                 }
                 grandparent.element <- tag.stack[stack.pointer - 2]
                 if (grandparent.element == "measurementSeries") {
                     n.unmeasured.inner[2] <<- val
                 } else if (grandparent.element == "radius") {
                     n.unmeasured.inner[1] <<- val
                 }
                 end.element("nrOfUnmeasuredInnerRings")
             },
             "/nrOfUnmeasuredOuterRings" = function(...) {
                 val <- as.integer(text.buffer)
                 if (val < 0) {
                     stop("negative nrOfUnmeasuredInnerRings")
                 }
                 grandparent.element <- tag.stack[stack.pointer - 2]
                 if (grandparent.element == "measurementSeries") {
                     n.unmeasured.outer[2] <<- val
                 } else if (grandparent.element == "radius") {
                     n.unmeasured.outer[1] <<- val
                 }
                 end.element("nrOfUnmeasuredOuterRings")
             },
             "/object" = function(...) {
                 ## Construct the data.frames belonging to the site
                 ## (object)
                 n.remark <- length(remark.data.text)
                 unique.unit <- unique(site.unit)

                 if (length(unique.unit) > 0 &&
                     (ids.from.titles || ids.from.identifiers)) {
                     alternative.ids()
                 }

                 ## Each combination of unit, ...
                 for (un in unique.unit) {
                     ## Which series in site.data have this unit?
                     idx.u <- which(site.unit %in% un)
                     ## Which remarks have this site and unit?
                     rematch.siteunit <-
                         which(remark.data.unit[inc(remarks.handled + 1,
                                                    n.remark)] %in% un) +
                                                        remarks.handled
                     ## ... taxon, and ...
                     unique.taxon <- unique(site.taxon[idx.u, , drop=FALSE])
                     for (i in seq_len(nrow(unique.taxon))) {
                         tax <- unique.taxon[i, , drop=FALSE]
                         ## Which series in site.data have this unit
                         ## and taxon?
                         idx.ut <-
                             idx.u[row.match(site.taxon[idx.u, ,
                                                        drop=FALSE], tax)]
                         ## Which remarks have this site, unit, and taxon?
                         idx.temp <-
                             row.match(remark.data.taxon[rematch.siteunit, ,
                                                         drop=FALSE], tax)
                         rematch.siteunittax <- rematch.siteunit[idx.temp]
                         ## ...variable gets a separate data.frame
                         unique.var <-
                             unique(site.var[idx.ut, , drop=FALSE])
                         for (j in seq_len(nrow(unique.var))) {
                             length.res <- length(res.df) + 1
                             var <- unique.var[j, , drop=FALSE]
                             ## Which series in site.data have this
                             ## unit, taxon, and variable?
                             idx.temp <-
                                 row.match(site.var[idx.ut, ,
                                                    drop=FALSE], var)
                             idx.utv <- idx.ut[idx.temp]
                             df.ncol <- length(idx.utv)
                             t.i.s <-
                                 titles.in.site[idx.utv, , drop=FALSE]
                             i.i.s <- ids.in.site[idx.utv, , drop=FALSE]
                             f.y <- first.year[idx.utv]
                             l.y <- last.year[idx.utv]
                             ## Which remarks have this site, taxon,
                             ## and variable?
                             idx.temp <-
                                 row.match(remark.data.var[rematch.siteunittax,
                                                           ,
                                                           drop=FALSE], var)
                             rematch.siteunittaxvar <-
                                 rematch.siteunittax[idx.temp]
                             remark.data.frame[rematch.siteunittaxvar] <<-
                                 length.res
                             ## For each matching remark, the index
                             ## number to the particular series in
                             ## site.data that the remark belongs to
                             remark.series <-
                                 remark.data.col[rematch.siteunittaxvar]
                             if (combine.series) {
                                 combi.results <-
                                     series.combiner(i.i.s, t.i.s, df.ncol,
                                                     f.y, l.y, remark.series,
                                                     idx.utv)
                                 i.i.s <- combi.results[[1]]
                                 t.i.s <- combi.results[[2]]
                                 df.ncol <- combi.results[[3]]
                                 f.y <- combi.results[[4]]
                                 l.y <- combi.results[[5]]
                                 remark.series <- combi.results[[6]]
                                 idx.utv <- combi.results[[7]]
                             }
                             df.first <- min(f.y)
                             df.last <- max(l.y)
                             df.nrow <- df.last - df.first + 1
                             this.df <-
                                 data.frame(array(as.numeric(NA),
                                                  dim=c(df.nrow, df.ncol)))
                             row.names(this.df) <- df.first:df.last
                             composite.titles <-
                                 create.composite.titles(t.i.s, i.i.s)
                             rownames(t.i.s) <- rownames(i.i.s) <-
                                 composite.titles
                             names(this.df) <- composite.titles
                             for (l in seq_len(df.ncol)) {
                                 first.idx <- f.y[l] - df.first + 1
                                 last.idx <- l.y[l] - df.first + 1
                                 this.df[first.idx:last.idx,l] <-
                                     site.data[[idx.utv[l]]]
                                 ## Adjusting the metadata of remarks
                                 idx.adjust <-
                                     rematch.siteunittaxvar[remark.series ==
                                                            idx.utv[l]]
                                 ## ... fixing the row numbers
                                 remark.data.row[idx.adjust] <<-
                                     remark.data.row[idx.adjust] + (first.idx - 1)
                                 ## ... fixing the col numbers
                                 remark.data.col[idx.adjust] <<- l
                             }
                             class(this.df) <- c("rwl", "data.frame")
                             res.df[[length.res]] <<- this.df
                             res.ids[[length.res]] <<- data.frame(i.i.s)
                             res.titles[[length.res]] <<- data.frame(t.i.s)
                             res.unit[length.res] <<- un
                             res.project.title <<-
                                 c(res.project.title, project.title)
                             res.project.id <<- c(res.project.id, idx.project)
                             seq.object <- seq_len(object.level)
                             res.site.title[[length.res]] <<-
                                 object.title[seq.object]
                             res.site.id[[length.res]] <<-
                                 idx.object[seq.object]
                             res.var <<- rbind(res.var, var)
                             res.taxon <<- rbind(res.taxon, tax)
                             this.wc <-
                                 data.frame(site.pith.presence[idx.utv],
                                            site.heartwood.presence[idx.utv],
                                            site.sapwood.presence[idx.utv],
                                            site.last.ring.presence[idx.utv],
                                            site.last.ring.details[idx.utv],
                                            site.bark.presence[idx.utv],
                                            site.n.sapwood[idx.utv],
                                            site.n.missing.heartwood[idx.utv],
                                            site.n.missing.sapwood[idx.utv],
                                            site.missing.heartwood.foundation[idx.utv],
                                            site.missing.sapwood.foundation[idx.utv],
                                            site.n.unmeasured.inner[idx.utv],
                                            site.n.unmeasured.outer[idx.utv])
                             names(this.wc) <- wc.names
                             row.names(this.wc) <- composite.titles
                             ## Remove unused columns
                             delete.idx <-
                                 which(apply(this.wc,
                                             2,
                                             function(x) all(is.na(x))))
                             this.wc[delete.idx] <- rep(list(NULL),
                                                        length(delete.idx))
                             res.wc[[length.res]] <<- this.wc
                         }
                     }
                 }
                 remarks.handled <<- n.remark
                 idx.element <<- idx.sample <<- idx.radius <<- idx.series <<- 0
                 site.data <<- list()
                 site.n.sapwood <<- site.n.missing.sapwood <<- integer(0)
                 site.n.missing.heartwood <<- integer(0)
                 site.n.unmeasured.inner <<- integer(0)
                 site.n.unmeasured.outer <<- integer(0)
                 site.missing.heartwood.foundation <<- character(0)
                 site.missing.sapwood.foundation <<- character(0)
                 site.pith.presence <<- character(0)
                 site.heartwood.presence <<- character(0)
                 site.sapwood.presence <<- character(0)
                 site.bark.presence <<- character(0)
                 site.last.ring.presence <<- character(0)
                 site.last.ring.details <<- character(0)
                 ids.in.site <<-
                     array(numeric(0), dim=c(0,4),
                           dimnames=list(NULL, ID.ORDER))
                 titles.in.site <<-
                     array(character(0), dim=c(0,4),
                           dimnames=list(NULL, ID.ORDER))
                 site.taxon <<-
                     array(character(0), dim=c(0,5),
                           dimnames=list(NULL, FIVE.COLS))
                 site.var <<-
                     array(character(0), dim=c(0,6),
                           dimnames=list(NULL, SIX.COLS))
                 site.unit <<- character(0)
                 first.year <<- last.year <<- numeric(0)
                 ## Coming down the hierarchy of <object>s:
                 ## the case of a container <object> without <element>s
                 if (last.closed == "object") {
                     object.level <<- object.level - 1
                 }
                 end.element("object")
             },
             "/postalCode" = function(...) {
                 lab.postalCode <<- text.buffer
                 end.element("postalCode")
             },
             "/preferredSeries" = function(...) {
                 if (idx.element < 1) {
                     preferred.tree.id <<- c(preferred.tree.id, NA)
                     preferred.tree.title <<- c(preferred.tree.title, NA)
                 } else {
                     preferred.tree.id <<- c(preferred.tree.id, idx.element)
                     preferred.tree.title <<-
                         c(preferred.tree.title, element.title)
                 }
                 preferred.idRef <<- c(preferred.idRef, link.idRef)
                 preferred.xLink <<- c(preferred.xLink, link.xLink)
                 preferred.identifier <<-
                     c(preferred.identifier, link.identifier)
                 preferred.domain <<- c(preferred.domain, domain.text)
                 preferred.project.id <<- c(preferred.project.id, idx.project)
                 preferred.project.title <<-
                     c(preferred.project.title, project.title)
                 length.res <- length(preferred.site.id) + 1
                 seq.object <- seq_len(object.level)
                 preferred.site.id[[length.res]] <<- idx.object[seq.object]
                 preferred.site.title[[length.res]] <<- object.title[seq.object]
                 end.element("preferredSeries")
             },
             "/project" = function(...) {
                 ## laboratory
                 if (length(lab.names) > 0) {
                     this.frame <-
                         data.frame(name = lab.names,
                                    acronym = lab.acronyms,
                                    identifier = lab.identifiers,
                                    domain = lab.domains,
                                    addressLine1 = lab.addressLine1s,
                                    addressLine2 = lab.addressLine2s,
                                    cityOrTown = lab.cityOrTowns,
                                    stateProvinceRegion = lab.stateProvinceRegions,
                                    postalCode = lab.postalCodes,
                                    country = lab.countries)
                     ## The name col is always kept, whether all names are NA
                     ## or not (name is a required element).
                     delete.idx <- which(apply(this.frame[, 2:ncol(this.frame),
                                                          drop=FALSE],
                                               2,
                                               function(x) all(is.na(x))))
                     this.frame[delete.idx + 1] <- rep(list(NULL),
                                                       length(delete.idx))
                     res.lab[[idx.project]] <<- this.frame
                 } else {
                     res.lab[[idx.project]] <<- NA
                 }
                 ## research
                 if (length(research.identifiers) > 0) {
                     res.research[[idx.project]] <<-
                         data.frame(identifier = research.identifiers,
                                    domain = research.domains,
                                    description = research.descriptions)
                 } else {
                     res.research[[idx.project]] <<- NA
                 }
                 end.element("project")
             },
             "/remark" = function(...) {
                 if (idx.derived > 0) {
                     remark.derived.text <<- c(remark.derived.text, text.buffer)
                     remark.derived.series <<- c(remark.derived.series,
                                                 length(res.derived$data) + 1)
                     remark.derived.idx <<- c(remark.derived.idx, idx.value)
                 } else if (is.na(first.dplr) && is.na(last.dplr)) {
                     remark.undated.text <<- c(remark.undated.text, text.buffer)
                     remark.undated.series <<- c(remark.undated.series,
                                                 length(res.undated$data) + 1)
                     remark.undated.idx <<- c(remark.undated.idx, idx.value)
                 } else {
                     remark.data.text <<- c(remark.data.text, text.buffer)
                     remark.data.frame <<- c(remark.data.frame, as.numeric(NA))
                     remark.data.taxon <<-
                         rbind(remark.data.taxon,
                               c(taxon, taxon.lang, taxon.normal,
                                 taxon.normalId, taxon.normalStd))
                     remark.data.var <<-
                         rbind(remark.data.var,
                               c(variable, variable.lang, variable.normal,
                                 variable.normalId, variable.normalStd,
                                 variable.normalTridas))
                     if (unitless) {
                         remark.data.unit <<- c(remark.data.unit, "unitless")
                     } else if (unit.converted) {
                         remark.data.unit <<- c(remark.data.unit, "millimetres")
                     } else {
                         remark.data.unit <<- c(remark.data.unit, unit)
                     }

                     ## Column information is not final at this point,
                     ## but is adjusted after the division of the site into
                     ## separate data.frames is clear.
                     remark.data.col <<-
                         c(remark.data.col, length(site.data) + 1)
                     ## Row information is not final at this point,
                     ## but is adjusted after the starting year
                     ## of the data frame is known.
                     ## NA values added in the middle will also affect these.
                     remark.data.row <<- c(remark.data.row, idx.value)
                 }
                 values.n.remarks <<- values.n.remarks + 1
                 end.element("remark")
             },
             "/research" = function(...) {
                 research.domains <<- c(research.domains, domain.text)
                 research.identifiers <<-
                     c(research.identifiers, research.identifier)
                 research.descriptions <<-
                     c(research.descriptions, research.description)
                 end.element("research")
             },
             "/series" = function(...) {
                 link.idRefs <<- c(link.idRefs, link.idRef)
                 link.xLinks <<- c(link.xLinks, link.xLink)
                 link.identifiers <<- c(link.identifiers, link.identifier)
                 link.domains <<- c(link.domains, domain.text)
                 end.element("series")
             },
             "/standardizingMethod" = function(...) {
                 stdizing.method <<- text.buffer
                 end.element("standardizingMethod")
             },
             "/stateProvinceRegion" = function(...) {
                 lab.stateProvinceRegion <<- text.buffer
                 end.element("stateProvinceRegion")
             },
             "/taxon" = function(...) {
                 if (nzchar(text.buffer)) {
                     taxon <<- text.buffer
                 }
                 end.element("taxon")
             },
             "/title" = function(...) {
                 parent.element <- tag.stack[stack.pointer - 1]
                 if (parent.element == "project") {
                     project.title <<- text.buffer
                 } else if (parent.element == "object") {
                     object.title[object.level] <<- text.buffer
                 } else if (parent.element == "element") {
                     element.title <<- text.buffer
                 } else if (parent.element == "sample") {
                     sample.title <<- text.buffer
                 } else if (parent.element == "radius") {
                     radius.title <<- text.buffer
                 } else if (parent.element == "measurementSeries") {
                     series.title <<- text.buffer
                     these.titles <<- array(c(element.title,
                                              sample.title,
                                              radius.title,
                                              series.title), dim=c(1,4))
                 } else if (parent.element == "derivedSeries") {
                     series.title <<- text.buffer
                 }
                 end.element("title")
             },
             "/type" = function(...) {
                 type.text <<- c(type.text, text.buffer)
                 type.project.id <<- c(type.project.id, idx.project)
                 type.project.title <<- c(type.project.title, project.title)
                 if (idx.derived > 0) {
                     type.derived.id <<- c(type.derived.id, idx.derived)
                     type.derived.title <<- c(type.derived.title, series.title)
                     type.site.id <<- c(type.site.id, NA)
                     type.site.title <<- c(type.site.title, NA)
                     type.tree.id <<- c(type.tree.id, NA)
                     type.tree.title <<- c(type.tree.title, NA)
                     type.core.id <<- c(type.core.id, NA)
                     type.core.title <<- c(type.core.title, NA)
                 } else {
                     type.derived.id <<- c(type.derived.id, NA)
                     type.derived.title <<- c(type.derived.title, NA)
                     if (idx.object[1] > 0) {
                         new.length <- length(type.site.id) + 1
                         seq.object <- seq_len(object.level)
                         type.site.id[[new.length]] <<- idx.object[seq.object]
                         type.site.title[[new.length]] <<-
                             object.title[seq.object]
                     } else {
                         type.site.id <<- c(type.site.id, NA)
                         type.site.title <<- c(type.site.title, NA)
                     }
                     if (idx.element > 0) {
                         type.tree.id <<- c(type.tree.id, idx.element)
                         type.tree.title <<- c(type.tree.title, element.title)
                     } else {
                         type.tree.id <<- c(type.tree.id, NA)
                         type.tree.title <<- c(type.tree.title, NA)
                     }
                     if (idx.sample > 0) {
                         type.core.id <<- c(type.core.id, idx.sample)
                         type.core.title <<- c(type.core.title, sample.title)
                     } else {
                         type.core.id <<- c(type.core.id, NA)
                         type.core.title <<- c(type.core.title, NA)
                     }
                 }
                 end.element("type")
             },
             "/unit" = function(...) {
                 if (is.na(unit)) {
                     unit <<- text.buffer
                 }
                 values.divisor <<- DIVISORS[unit]
                 if (is.na(values.divisor)) {
                     values.multiplier <<- MULTIPLIERS[unit]
                     if (!is.na(values.multiplier)) {
                         unit.converted <<- TRUE
                     }
                 } else {
                     unit.converted <<- TRUE
                 }
                 end.element("unit")
             },
             "/values" = function(...) {
                 if (idx.value > 0) {
                     parent.element <- tag.stack[stack.pointer - 1]
                     vector.size <- length(value.vector)
                     if (vector.size > idx.value) {
                         value.vector <<-
                             value.vector[-((idx.value + 1):vector.size)]
                     }
                     if (!unitless) {
                         ## We know some units that cannot get negative values
                         check.negative <- FALSE
                         if (unit == "millimetres") {
                             check.negative <- TRUE
                         } else if (!is.na(values.divisor)) {
                             value.vector <<- value.vector / values.divisor
                             check.negative <- TRUE
                         } else if (!is.na(values.multiplier)) {
                             value.vector <<- value.vector * values.multiplier
                             check.negative <- TRUE
                         }
                         if (check.negative) {
                             idx.negative <- which(value.vector < 0)
                             if (length(idx.negative) > 0) {
                                 value.vector[idx.negative] <<- MISSING.VALUE
                                 if (in.derived.values) {
                                     warning(gettextf("in project %d, derived series %d: ",
                                                      idx.project, idx.derived),
                                             "negative values interpreted as missing")
                                 } else {
                                     warning(gettextf("in project %d, object %s, element %d, sample %d, radius %d, series %d: ",
                                                      idx.project,
                                                      paste(idx.object[seq_len(object.level)],
                                                            collapse="."),
                                                      idx.element,
                                                      idx.sample,
                                                      idx.radius,
                                                      idx.series),
                                             "negative values interpreted as missing")
                                 }
                             }
                         }
                     }
                     if (parent.element == "measurementSeries") {
                         this.sapwood <- n.sapwood[length(n.sapwood)]
                         ## Knowing one of first and last year is
                         ## sufficient.  The other one is computed
                         ## from the number of values recorded.
                         if (!is.na(first.dplr)) {
                             last.computed <- first.dplr + idx.value - 1
                             last.year <<- c(last.year, last.computed)
                             first.year <<- c(first.year, first.dplr)
                             is.dated <- TRUE
                             if (!is.na(last.dplr)) {
                                 if (last.computed > last.dplr) {
                                     warning(gettextf("in project %d, object %s, element %d, sample %d, radius %d, series %d: ",
                                                      idx.project,
                                                      paste(idx.object[seq_len(object.level)],
                                                            collapse="."),
                                                      idx.element,
                                                      idx.sample,
                                                      idx.radius,
                                                      idx.series),
                                             gettextf("too many values (expected %d, got %d)",
                                                      idx.value + (last.dplr - last.computed),
                                                      idx.value))
                                 } else if (last.computed < last.dplr) {
                                     if (is.na(this.sapwood) ||
                                         this.sapwood >= idx.value ||
                                         this.sapwood == 0) {
                                         warning(gettextf("in project %d, object %s, element %d, sample %d, radius %d, series %d: ",
                                                          idx.project,
                                                          paste(idx.object[seq_len(object.level)],
                                                                collapse="."),
                                                          idx.element,
                                                          idx.sample,
                                                          idx.radius,
                                                          idx.series),
                                                 gettextf("too few values (expected %d, got %d)",
                                                          idx.value + (last.dplr - last.computed),
                                                          idx.value))
                                     } else {
                                         ## We quietly assume that
                                         ## rings are missing at the
                                         ## border of heartwood and
                                         ## sapwood and add symbols of
                                         ## missing value there.
                                         n.heartwood <- idx.value - this.sapwood
                                         n.missing <- last.dplr - last.computed
                                         value.vector <<-
                                             c(value.vector[seq_len(n.heartwood)],
                                               rep(MISSING.VALUE, n.missing),
                                               value.vector[(n.heartwood + 1):idx.value])
                                         last.year[length(last.year)] <<-
                                             last.dplr
                                         if (values.n.remarks > 0) {
                                             n.remarks <-
                                                 length(remark.data.text)
                                             idx.these <-
                                                 (n.remarks - values.n.remarks + 1):n.remarks
                                             idx.adjust <-
                                                 idx.these[remark.data.row[idx.these] > n.heartwood]
                                             remark.data.row[idx.adjust] <-
                                                 remark.data.row[idx.adjust] +
                                                     n.missing
                                         }
                                     }
                                 }
                             }
                         } else if (!is.na(last.dplr)) {
                             last.year <<- c(last.year, last.dplr)
                             first.year <<- c(first.year,
                                              last.dplr - idx.value + 1)
                             is.dated <- TRUE
                         } else {
                             is.dated <- FALSE
                         }
                         wc.minus.n.sapwood <- setdiff(wc.names, "n.sapwood")
                         if (is.dated) {
                             new.length <- length(site.data) + 1
                             site.taxon <<-
                                 rbind(site.taxon,
                                       c(taxon, taxon.lang, taxon.normal,
                                         taxon.normalId, taxon.normalStd))
                             site.var <<-
                                 rbind(site.var,
                                       c(variable, variable.lang,
                                         variable.normal, variable.normalId,
                                         variable.normalStd,
                                         variable.normalTridas))
                             if (unitless) {
                                 site.unit <<- c(site.unit, "unitless")
                             } else if (unit.converted) {
                                 site.unit <<- c(site.unit, "millimetres")
                             } else {
                                 site.unit <<- c(site.unit, unit)
                             }
                             site.data[[new.length]] <<- value.vector
                             ids.in.site <<- rbind(ids.in.site, these.ids)
                             titles.in.site <<-
                                 rbind(titles.in.site, these.titles)
                             site.n.sapwood <<- c(site.n.sapwood, this.sapwood)
                             for (attr in wc.minus.n.sapwood) {
                                 this.var <- get(attr)
                                 this.name <- paste("site", attr, sep=".")
                                 assign(this.name,
                                        c(get(this.name),
                                          this.var[length(this.var)]),
                                        inherits = TRUE)
                             }
                         } else {
                             ## Undated data, no matter which project,
                             ## site or variable, are stored in a
                             ## single list ($data). Also metadata are
                             ## stored in res.undated.
                             new.length <- length(res.undated$data) + 1
                             res.undated$data[[new.length]] <<- value.vector
                             res.undated$ids <<-
                                 rbind(res.undated$ids, these.ids)
                             res.undated$titles <<-
                                 rbind(res.undated$titles, these.titles)
                             res.undated$project.title <<-
                                 c(res.undated$project.title, project.title)
                             res.undated$project.id <<-
                                 c(res.undated$project.id, idx.project)
                             seq.object <- seq_len(object.level)
                             res.undated$site.title[[new.length]] <<-
                                 object.title[seq.object]
                             res.undated$site.id[[new.length]] <<-
                                 idx.object[seq.object]
                             if (unitless) {
                                 this.unit <- "unitless"
                             } else if (unit.converted) {
                                 this.unit <- "millimetres"
                             } else {
                                 this.unit <- unit
                             }
                             if (warn.units && this.unit != "millimetres") {
                                 if (this.unit == "unitless") {
                                     warning(gettextf("in undated series %d: ",
                                                      new.length),
                                             "unitless measurements present")
                                 } else {
                                     warning(gettextf("in undated series %d: ",
                                                      new.length),
                                             gettextf("strange unit \"%s\"",
                                                      this.unit))
                                 }
                             }
                             res.undated$unit <<- c(res.undated$unit, this.unit)

                             res.undated.taxon <<-
                                 rbind(res.undated.taxon,
                                       c(taxon, taxon.lang, taxon.normal,
                                         taxon.normalId, taxon.normalStd))
                             res.undated.var <<-
                                 rbind(res.undated.var,
                                       c(variable, variable.lang,
                                         variable.normal, variable.normalId,
                                         variable.normalStd,
                                         variable.normalTridas))

                             res.undated.n.sapwood <<- c(res.undated.n.sapwood,
                                                         this.sapwood)
                             for (attr in wc.minus.n.sapwood) {
                                 this.var <- get(attr)
                                 this.name <-
                                     paste("res.undated", attr, sep=".")
                                 assign(this.name,
                                        c(get(this.name),
                                          this.var[length(this.var)]),
                                        inherits = TRUE)
                             }
                         }
                     } else if (in.derived.values) {
                         idx.negative <-
                             which(count.vector[seq_along(value.vector)] < 0)
                         if (length(idx.negative) > 0) {
                             count.vector[idx.negative] <<- NA
                             warning(gettextf("in project %d, derived series %d: ",
                                              idx.project, idx.derived),
                                     "negative counts interpreted as missing")
                         }
                         if (!is.na(first.dplr)) {
                             this.last.year <- first.dplr + idx.value - 1
                             this.first.year <- first.dplr
                             if (!is.na(last.dplr)) {
                                 if (this.last.year > last.dplr) {
                                     warning(gettextf("in project %d, derived series %d: ",
                                                      idx.project,
                                                      idx.derived),
                                             gettextf("too many values (expected %d, got %d)",
                                                      idx.value + (last.dplr - this.last.year),
                                                      idx.value))
                                 } else if (this.last.year < last.dplr) {
                                     warning(gettextf("in project %d, derived series %d: ",
                                                      idx.project,
                                                      idx.derived),
                                             gettextf("too few values (expected %d, got %d)",
                                                      idx.value + (last.dplr - this.last.year),
                                                      idx.value))
                                 }
                             }
                             is.dated <- TRUE
                         } else if (!is.na(last.dplr)) {
                             this.last.year <- last.dplr
                             this.first.year <- last.dplr - idx.value + 1
                             is.dated <- TRUE
                         } else {
                             is.dated <- FALSE
                         }
                         series.frame <-
                             data.frame(series = value.vector,
                                        samp.depth = count.vector[seq_along(value.vector)])
                         if (is.dated) {
                             row.names(series.frame) <-
                                 as.character(this.first.year:this.last.year)
                             if (this.last.year > year.now) {
                                 warning(gettextf("in derived series %d: ",
                                                  idx.derived),
                                         "data from the future")
                             }
                         }
                         res.derived$data[[length(res.derived$data) + 1]] <<-
                             series.frame
                         res.derived$project.id <<-
                             c(res.derived$project.id, idx.project)
                         res.derived$project.title <<-
                             c(res.derived$project.title, project.title)
                         res.derived$id <<- c(res.derived$id, idx.derived)
                         res.derived$title <<-
                             c(res.derived$title, series.title)
                         res.derived.var <<-
                             rbind(res.derived.var,
                                   c(variable, variable.lang, variable.normal,
                                     variable.normalId, variable.normalStd,
                                     variable.normalTridas))
                         if (unitless) {
                             res.derived$unit <<-
                                 c(res.derived$unit, "unitless")
                         } else if (unit.converted) {
                             res.derived$unit <<-
                                 c(res.derived$unit, "millimetres")
                         } else {
                             res.derived$unit <<- c(res.derived$unit, unit)
                         }
                         res.derived$standardizing.method <<-
                             c(res.derived$standardizing.method,
                               stdizing.method)
                     }
                 }
                 in.derived.values <<- FALSE
                 end.element("values")
             },
             "/variable" = function(...) {
                 if (nzchar(text.buffer)) {
                     variable <<- text.buffer
                 }
                 end.element("variable")
             },
             ## Other end tag
             .endElement = end.element,
### Handlers for character data
             .text = text.function,
             .cdata = text.function,
### Returns the results of parsing in a convenient (?) list.
             get.results = function() {
                 res.all <- list()
                 ## res.all$measurements, $ids, $titles: the data from
                 ## <measurementSeries> and (the most) important metadata
                 df.size <- length(res.df)
                 if (df.size > 0) {
                     ## We don't want lists with length 1: just take
                     ## the contents
                     if (df.size == 1) {
                         res.all$measurements <- res.df[[1]]
                         res.all$ids <- res.ids[[1]]
                         res.all$titles <- res.titles[[1]]
                         if (ncol(res.wc[[1]]) > 0) {
                             res.all$wood.completeness <- res.wc[[1]]
                         }
                     } else {
                         res.all$measurements <- res.df
                         res.all$ids <- res.ids
                         res.all$titles <- res.titles
                         if (any(vapply(res.wc, ncol, 0) > 0)) {
                             res.all$wood.completeness <- res.wc
                         }
                     }
                     res.all$unit <- res.unit
                     res.all$project.id <- res.project.id
                     res.all$project.title <- res.project.title
                     res.all$site.id <- site.info.to.df(res.site.id, "site.id")
                     res.all$site.title <-
                         site.info.to.df(res.site.title, "site.title")
                     res.taxon <<- data.frame(res.taxon)
                     delete.idx <- which(vapply(res.taxon,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     if (length(delete.idx) == length(res.taxon)) {
                         delete.idx <- delete.idx[-1]
                     }
                     res.taxon[delete.idx] <<-
                         rep(list(NULL), length(delete.idx))
                     res.all$taxon <- res.taxon
                     res.var <<- data.frame(res.var)
                     delete.idx <- which(vapply(res.var,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     if (length(delete.idx) == length(res.var)) {
                         delete.idx <- delete.idx[-1]
                     }
                     res.var[delete.idx] <<- rep(list(NULL), length(delete.idx))
                     res.all$variable <- res.var

                     ## Print a summary of each data frame
                     if (df.size == 1) {
                         cat(gettext("'$measurements' is a data.frame\n",
                                     domain="R-dplR"))
                     } else {
                         cat(gettextf("there are %d data.frames in the '$measurements' list\n",
                                      df.size, domain="R-dplR"))
                     }
                     for (i in seq_len(df.size)) {
                         this.df <- res.df[[i]]
                         nseries <- ncol(this.df)
                         series.ids <- names(this.df)
                         rnames <- row.names(this.df)
                         min.year <- as.numeric(rnames[1])
                         max.year <- as.numeric(rnames[nrow(this.df)])
                         if (max.year > year.now) {
                             warning(gettextf("in data.frame %d: ", i),
                                     "data from the future")
                         }
                         not.na <- lapply(this.df, function(x) which(!is.na(x)))
                         series.min <- vapply(not.na, min, 0) + (min.year - 1)
                         series.max <- vapply(not.na, max, 0) + (min.year - 1)
                         not.na.title <- which(!is.na(res.all$site.title[i, ]))
                         title.level <-
                             max(not.na.title[length(not.na.title)], 1)
                         cat(gettextf("\ndata.frame #%d\n", i,
                                      domain="R-dplR"))
                         ## Note: all known units are converted to millimetres
                         this.unit <- res.all$unit[i]
                         if (warn.units && this.unit != "millimetres") {
                             if (this.unit == "unitless") {
                                 warning(gettextf("in data.frame %d: ", i),
                                         "unitless measurements present")
                             } else {
                                 warning(gettextf("in data.frame %d: ", i),
                                         gettextf("strange unit \"%s\"",
                                                  this.unit))
                             }
                         }
                         cat(gettext("* site: ", domain="R-dplR"),
                             paste(as.matrix(res.all$site.title[i,
                                                                seq_len(title.level)]),
                                   collapse=" / "),
                             "\n", sep="")
                         cat(gettext("* taxon: ", domain="R-dplR"))
                         cat("\n\t", row.print(res.all$taxon[i, , drop=FALSE],
                                               collapse="\n\t"), "\n", sep="")
                         cat(gettext("* variable: ", domain="R-dplR"))
                         cat("\n\t",
                             row.print(res.all$variable[i, , drop=FALSE],
                                       collapse="\n\t"), "\n", sep="")
                         cat(sprintf(ngettext(nseries,
                                              "There is %d series\n",
                                              "There are %d series\n",
                                              domain="R-dplR"),
                                     nseries))
                         cat(paste0(seq_len(nseries), "\t",
                                    series.ids, "\t",
                                    series.min, "\t",
                                    series.max, "\n"), sep="")
                     }
                 }
                 ## Undated data in <measurementSeries>
                 undated.size <- length(res.undated$data)
                 if (undated.size > 0) {
                     ## Say no to lists of length 1
                     if (undated.size == 1) {
                         res.undated$data <<- res.undated$data[[1]]
                     }
                     cat(sprintf(ngettext(undated.size,
                                          "There is %d undated series, returned in '$undated'\n",
                                          "There are %d undated series, returned in '$undated'\n",
                                          domain="R-dplR"),
                                 undated.size))
                     res.undated$ids <<- data.frame(res.undated$ids)
                     res.undated$titles <<- data.frame(res.undated$titles)
                     res.undated$site.id <<-
                         site.info.to.df(res.undated$site.id, "site.id")
                     res.undated$site.title <<-
                         site.info.to.df(res.undated$site.title, "site.title")
                     ## Create data.frames, remove unused elements
                     ## 1. Variable
                     res.undated.var <<- data.frame(res.undated.var)
                     delete.idx <- which(vapply(res.undated.var,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     if (length(delete.idx) == length(res.undated.var)) {
                         delete.idx <- delete.idx[-1]
                     }
                     res.undated.var[delete.idx] <<-
                         rep(list(NULL), length(delete.idx))
                     res.undated$variable <<- res.undated.var
                     ## 2. Taxon
                     res.undated.taxon <<- data.frame(res.undated.taxon)
                     delete.idx <- which(vapply(res.undated.taxon,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     if (length(delete.idx) == length(res.undated.taxon)) {
                         delete.idx <- delete.idx[-1]
                     }
                     res.undated.taxon[delete.idx] <<-
                         rep(list(NULL), length(delete.idx))
                     res.undated$taxon <<- res.undated.taxon
                     ## 3. Wood completeness
                     this.wc <-
                         data.frame(res.undated.pith.presence,
                                    res.undated.heartwood.presence,
                                    res.undated.sapwood.presence,
                                    res.undated.last.ring.presence,
                                    res.undated.last.ring.details,
                                    res.undated.bark.presence,
                                    res.undated.n.sapwood,
                                    res.undated.n.missing.heartwood,
                                    res.undated.n.missing.sapwood,
                                    res.undated.missing.heartwood.foundation,
                                    res.undated.missing.sapwood.foundation,
                                    res.undated.n.unmeasured.inner,
                                    res.undated.n.unmeasured.outer)
                     names(this.wc) <- wc.names
                     delete.idx <- which(apply(this.wc,
                                               2,
                                               function(x) all(is.na(x))))
                     this.wc[delete.idx] <- rep(list(NULL),
                                                length(delete.idx))
                     res.undated$wood.completeness <<- this.wc

                     ## Final preparations
                     delete.idx <- which(vapply(res.undated,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     if (length(delete.idx) > 0) {
                         res.undated <<- res.undated[-delete.idx]
                     }
                     res.all$undated <- res.undated
                 }
                 ## <derivedSeries>
                 ## Number of <values> within <derivedSeries>
                 derived.nvalues <- length(res.derived$data)
                 if (derived.nvalues > 0) {
                     if (derived.nvalues == 1) {
                         res.derived$data <<- res.derived$data[[1]]
                     }
                     cat(sprintf(ngettext(derived.nvalues,
                                          "There is %d derived series, returned in '$derived'\n",
                                          "There are %d derived series, returned in '$derived'\n",
                                          domain="R-dplR"),
                                 derived.nvalues))
                     if (all(vapply(res.derived$link, is.na, TRUE))) {
                         ## If there was no content in any of the linkSeries,
                         ## res.derived$link is removed
                         res.derived$link <<- NULL
                     } else if (length(res.derived$link) == 1) {
                         ## No lists of length 1
                         res.derived$link <<- res.derived$link[[1]]
                     }
                     ## Create a data.frame (Variable)
                     res.derived.var <<- data.frame(res.derived.var)
                     ## Remove unused elements
                     delete.idx <- which(vapply(res.derived.var,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     if (length(delete.idx) == length(res.derived.var)) {
                         delete.idx <- delete.idx[-1]
                     }
                     res.derived.var[delete.idx] <<-
                         rep(list(NULL), length(delete.idx))
                     res.derived$variable <<- res.derived.var
                     res.all$derived <- res.derived
                 }
                 ## The columns included in res.all$type depend on
                 ## what was observed in the input data
                 if (length(type.text) > 0) {
                     res.all$type <-
                         data.frame(text = type.text,
                                    lang = type.lang,
                                    normal = type.normal,
                                    normalId = type.normalId,
                                    normalStd = type.normalStd,
                                    project.id = type.project.id,
                                    site.info.to.df(type.site.id, "site.id"),
                                    tree.id = type.tree.id,
                                    core.id = type.core.id,
                                    derived.id = type.derived.id,
                                    project.title = type.project.title,
                                    site.info.to.df(type.site.title,
                                                    "site.title"),
                                    tree.title = type.tree.title,
                                    core.title = type.core.title,
                                    derived.title = type.derived.title)
                     ## Remove duplicated rows (titles are irrelevant)
                     no.title <- !grepl("title", names(res.all$type))
                     res.all$type <-
                         res.all$type[!duplicated(res.all$type[no.title]), ]
                     ## Remove unused columns
                     delete.idx <- which(vapply(res.all$type,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     res.all$type[delete.idx] <- rep(list(NULL),
                                                     length(delete.idx))
                 }
                 ## Similar treatment for res.all$comments
                 if (length(comments.text) > 0) {
                     res.all$comments <-
                         data.frame(text = comments.text,
                                    project.id = comments.project.id,
                                    site.info.to.df(comments.site.id,
                                                    "site.id"),
                                    tree.id = comments.tree.id,
                                    core.id = comments.core.id,
                                    radius.id = comments.radius.id,
                                    measurement.id = comments.measurement.id,
                                    derived.id = comments.derived.id,
                                    project.title = comments.project.title,
                                    site.info.to.df(comments.site.title,
                                                    "site.title"),
                                    tree.title = comments.tree.title,
                                    core.title = comments.core.title,
                                    radius.title = comments.radius.title,
                                    measurement.title = comments.measurement.title,
                                    derived.title = comments.derived.title)
                     ## Remove duplicated rows (titles are irrelevant)
                     no.title <- !grepl("title", names(res.all$comments))
                     res.all$comments <-
                         res.all$comments[!duplicated(res.all$comments[no.title]), ]
                     ## Remove unused columns
                     delete.idx <- which(vapply(res.all$comments,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     res.all$comments[delete.idx] <- rep(list(NULL),
                                                         length(delete.idx))
                 }
                 ## Similar treatment for res.all$identifier
                 if (length(identifier.text) > 0) {
                     res.all$identifier <-
                         data.frame(text = identifier.text,
                                    domain = identifier.domain,
                                    project.id = identifier.project.id,
                                    site.info.to.df(identifier.site.id,
                                                    "site.id"),
                                    tree.id = identifier.tree.id,
                                    core.id = identifier.core.id,
                                    radius.id = identifier.radius.id,
                                    measurement.id = identifier.measurement.id,
                                    derived.id = identifier.derived.id,
                                    project.title = identifier.project.title,
                                    site.info.to.df(identifier.site.title,
                                                    "site.title"),
                                    tree.title = identifier.tree.title,
                                    core.title = identifier.core.title,
                                    radius.title = identifier.radius.title,
                                    measurement.title = identifier.measurement.title,
                                    derived.title = identifier.derived.title)
                     ## Remove duplicated rows (titles are irrelevant)
                     no.title <- !grepl("title", names(res.all$identifier))
                     res.all$identifier <-
                         res.all$identifier[!duplicated(res.all$identifier[no.title]), ]
                     ## Remove unused columns
                     delete.idx <- which(vapply(res.all$identifier,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     res.all$identifier[delete.idx] <- rep(list(NULL),
                                                           length(delete.idx))
                 }
                 ## Remarks are split into 3 parts: measurements, undated,
                 ## derived. Empty parts are not included.
                 remark.all <- list()
                 if (length(remark.data.text) > 0) {
                     remark.all$measurements <-
                         data.frame(text = remark.data.text,
                                    frame = remark.data.frame,
                                    row = remark.data.row,
                                    col = remark.data.col)
                 }
                 if (length(remark.undated.text) > 0) {
                     remark.all$undated <-
                         data.frame(text = remark.undated.text,
                                    series = remark.undated.series,
                                    idx = remark.undated.idx)
                 }
                 if (length(remark.derived.text) > 0) {
                     remark.all$derived <-
                         data.frame(text = remark.derived.text,
                                    series = remark.derived.series,
                                    idx = remark.derived.idx)
                 }
                 if (length(remark.all) > 0) {
                     res.all$remark <- remark.all
                 }
                 ## res.all$laboratory and res.all$research are lists with
                 ## one item per project. The list structure is ditched if
                 ## there is only one project. If there are no <research> data,
                 ## res.all$research will be absent (<research> is optional).
                 if (idx.project > 0) {
                     if (idx.project == 1) {
                         res.all$laboratory <- res.lab[[1]]
                     } else {
                         res.all$laboratory <- res.lab
                     }
                     if (any(vapply(res.research,
                                    function(x) !is.na(x), TRUE))) {
                         if (idx.project == 1) {
                             res.all$research <- res.research[[1]]
                         } else {
                             res.all$research <- res.research
                         }
                     }
                 }
                 if (length(altitude.metres) > 0) {
                     res.all$altitude <-
                         data.frame(metres = altitude.metres,
                                    project.id = altitude.project.id,
                                    site.info.to.df(altitude.site.id,
                                                    "site.id"),
                                    tree.id = altitude.tree.id,
                                    project.title = altitude.project.title,
                                    site.info.to.df(altitude.site.title,
                                                    "site.title"),
                                    tree.title = altitude.tree.title)
                     ## Remove duplicated rows (titles are irrelevant)
                     no.title <- !grepl("title", names(res.all$altitude))
                     idx.temp <- !duplicated(res.all$altitude[no.title])
                     res.all$altitude <- res.all$altitude[idx.temp, ]
                 }
                 if (length(preferred.project.id) > 0) {
                     res.all$preferred <-
                         data.frame(idRef = preferred.idRef,
                                    xLink = preferred.xLink,
                                    identifier = preferred.identifier,
                                    domain = preferred.domain,
                                    project.id = preferred.project.id,
                                    site.info.to.df(preferred.site.id,
                                                    "site.id"),
                                    tree.id = preferred.tree.id,
                                    project.title = preferred.project.title,
                                    site.info.to.df(preferred.site.title,
                                                    "site.title"),
                                    tree.title = preferred.tree.title)
                     ## Remove duplicated rows (titles are irrelevant)
                     no.title <- !grepl("title", names(res.all$preferred))
                     idx.temp <- !duplicated(res.all$preferred[no.title])
                     res.all$preferred <- res.all$preferred[idx.temp, ]
                     ## Remove unused columns
                     delete.idx <- which(vapply(res.all$preferred,
                                                function(x) all(is.na(x)),
                                                TRUE))
                     res.all$preferred[delete.idx] <-
                         rep(list(NULL), length(delete.idx))
                 }
                 res.all
             }) # end of the list of handler functions
    } # end of the function 'handler.factory'

    h <- xmlEventParse(file = path.expand(fname),
                       handlers = handler.factory(),
                       ignoreBlanks = TRUE,
                       addContext = FALSE,
                       useTagName = TRUE,
                       asText = FALSE,
                       trim = TRUE,
                       isURL = FALSE,
                       saxVersion = 2,
                       validate = FALSE,
                       useDotNames = TRUE)
    h$get.results()
}

`summary.rwl` <- function(object,...){ rwl.stats(rwl=object) }

`rwl.stats` <-
    function(rwl)
{
    acf1 <- function(x){
        ar1 <- acf(x[!is.na(x)], lag.max=1, plot=FALSE)
        ar1$acf[2]
    }
    skew <- function(x){
        y <- x[!is.na(x)]
        sum((y-mean(y))^3) / (length(y)*sd(y)^3)
    }

    kurt <- function(x){
      y <- x[!is.na(x)]
      n <- length(y)
      y2 <- y - mean(y)
      n * sum(y2^4)/(sum(y2^2)^2) * (1 - 1/n)^2 - 3
    }

    
    yr <- as.numeric(row.names(rwl))
    series.stats <- data.frame(series=names(rwl))
    rwl2 <- as.matrix(rwl)
    the.range <- as.matrix(apply(rwl2, 2, yr.range, yr.vec=yr))
    series.stats$first <- the.range[1, ]
    series.stats$last <- the.range[2, ]
    series.stats$year <- series.stats$last - series.stats$first + 1
    series.stats$mean <- colMeans(rwl2, na.rm=TRUE)
    series.stats$median <- colMedians(rwl2, na.rm=TRUE)
    series.stats$stdev <- colSds(rwl2, na.rm=TRUE)
    series.stats$skew <- apply(rwl2, 2, skew)
    series.stats$kurtosis <- apply(rwl2, 2, kurt)
    #series.stats$sens1 <- apply(rwl2, 2, sens1)
    #series.stats$sens2 <- apply(rwl2, 2, sens2)
    series.stats$gini <- apply(rwl2, 2, gini.coef)
    series.stats$ar1 <- apply(rwl2, 2, acf1)
    seq.temp <- -seq_len(4)
    series.stats[, seq.temp] <- round(series.stats[, seq.temp], 3)

    series.stats
}

`chron.ci` <-
  function(x,biweight=TRUE,conf=0.95,R=100)
{
    samps <- rowSums(!is.na(x))
    if(any(samps<2)){
      stop("x contains row(s) with only one sample. Stopping.")
    } 
    yrs <- as.numeric(row.names(x))
    
    getMean <- function(aRow,ind){
      dat <- aRow[ind]
      mask <- is.na(dat)
      dat <-dat[!mask]
      mean(dat)
    }
    
    getTBRM <- function(aRow,ind){
      dat <- aRow[ind]
      mask <- is.na(dat)
      dat <-dat[!mask]
      tbrm(dat)
    }
    if(biweight){
      res <- do.call(rbind, apply(x, 1, FUN = function(x)
      {
        x1 <- boot.ci(boot(x, statistic = getTBRM, R = R),
                      type="norm",conf=conf)
        res <- data.frame(std = x1$t0, 
                          lowerCI = x1$normal[2], upperCI = x1$normal[3])
      }))
    }
    else{
      res <- do.call(rbind, apply(x, 1, FUN = function(x)
      {
        x1 <- boot.ci(boot(x, statistic = getMean, R = R),
                      type="norm",conf=conf)
        res <- data.frame(std = x1$t0, 
                          lowerCI = x1$normal[2], upperCI = x1$normal[3])
      }))
    }
    res <- data.frame(res,samp.depth=samps)
    rownames(res) <- yrs
    class(res) <- c("crn", "data.frame")
    comment(res) <- "ci"
    return(res)
}

interseries.cor <- function(rwl, n=NULL, prewhiten=TRUE, biweight=TRUE, 
                       method = c("spearman", "pearson", "kendall")) {
    method2 <- match.arg(method)
    nseries <- length(rwl)
    res.cor <- numeric(nseries)
    p.val <- numeric(nseries)
    rwl.mat <- as.matrix(rwl)
    tmp <- normalize.xdate(rwl=rwl.mat, n=n,
                           prewhiten=prewhiten, biweight=biweight,
                           leave.one.out = TRUE)
    series <- tmp[["series"]]
    master <- tmp[["master"]]
    for (i in seq_len(nseries)) {
        tmp2 <- cor.test(series[, i], master[, i],
                         method = method2, alternative = "greater")
        res.cor[i] <- tmp2[["estimate"]]
        p.val[i] <- tmp2[["p.value"]]
    }
    res <- data.frame(res.cor = res.cor, p.val = p.val, row.names = names(rwl))
    # change res.cor to r, rho, or tau based on method
    res
}

`detrend` <-
    function(rwl, y.name = names(rwl), make.plot = FALSE,
             method=c("Spline", "ModNegExp", "Mean", "Ar", "Friedman", 
                      "ModHugershoff", "AgeDepSpline"),
             nyrs = NULL, f = 0.5, pos.slope = FALSE,
             constrain.nls = c("never", "when.fail", "always"),
             verbose = FALSE, return.info = FALSE,
             wt, span = "cv", bass = 0, difference = FALSE)
{
    stopifnot(identical(make.plot, TRUE) || identical(make.plot, FALSE),
              identical(pos.slope, FALSE) || identical(pos.slope, TRUE),
              identical(verbose, TRUE) || identical(verbose, FALSE),
              identical(return.info, TRUE) || identical(return.info, FALSE))
    known.methods <- c("Spline", "ModNegExp", "Mean", "Ar", "Friedman", 
                       "ModHugershoff", "AgeDepSpline")
    constrain2 <- match.arg(constrain.nls)
    method2 <- match.arg(arg = method,
                         choices = known.methods,
                         several.ok = TRUE)
    if(!is.data.frame(rwl))
        stop("'rwl' must be a data.frame")
    rn <- row.names(rwl)

    detrend.args <- c(alist(rwl.i),
                      list(make.plot = make.plot, method = method2,
                           nyrs = nyrs, f = f, pos.slope = pos.slope,
                           constrain.nls = constrain2,
                           verbose = FALSE, return.info = return.info,
                           span = span, bass = bass, difference = difference))
    if (!missing(wt)) {
        detrend.args <- c(detrend.args, list(wt = wt))
    }
    if(!make.plot && !verbose &&
       ("Spline" %in% method2 || "ModNegExp" %in% method2 || "ModHugershoff" %in% method2) &&
       !inherits(try(suppressWarnings(req.it <-
                                      requireNamespace("iterators",
                                                       quietly=TRUE)),
                     silent = TRUE),
                 "try-error") && req.it &&
       !inherits(try(suppressWarnings(req.fe <-
                                      requireNamespace("foreach",
                                                       quietly=TRUE)),
                     silent = TRUE),
                 "try-error") && req.fe){
        it.rwl <- iterators::iter(rwl, by = "col")
        # get series name for each iteration in dopar as well
        it.names <- iterators::iter(names(rwl), by = "col")
        ## a way to get rid of "no visible binding" NOTE in R CMD check
        rwl.i <- y.name.i <- NULL

        exportFun <- c("names<-", "detrend.series")
        ## Use a dummy loop to suppress possible (non-)warning from
        ## initial call to %dopar% with a sequential backend...
        foo <- suppressWarnings(foreach::"%dopar%"(foreach::foreach(i=1), {}))
        ## ... but leave actual warnings on for the real loop.
        out <- foreach::"%dopar%"(foreach::foreach(rwl.i=it.rwl,
                                                   y.name.i=it.names,
                                                   .export=exportFun),
                              {
                                  names(rwl.i) <- rn
                                  # append the series name to detrend.args
                                  do.call(detrend.series, 
                                          c(detrend.args,list(y.name=y.name.i)))
                              })

        if (return.info) {
            modelCurves <- lapply(out, "[[", 2)
            modelStats <- lapply(out, "[[", 3)
            dataStats <- lapply(out, "[[", 4)
            out <- lapply(out, "[[", 1)
        }
    } else{
        n.series <- ncol(rwl)
        out <- vector(mode = "list", length = n.series)
        if (return.info) {
            modelCurves <- vector(mode = "list", length = n.series)
            modelStats <- vector(mode = "list", length = n.series)
            dataStats <- vector(mode = "list", length = n.series)
        }
        detrend.args[1] <- alist(rwl[[i]])
        detrend.args[["verbose"]] <- verbose
        # how to append detrend.args in the dopar call so that
        # name gets passed in? Important for detrend.series warnings
        # about neg values.
        detrend.args <- c(detrend.args, alist(y.name = y.name[i]))
    
        for (i in seq_len(n.series)) {
            fits <- do.call(detrend.series, detrend.args)
            if (return.info) {
                modelCurves[[i]] <- fits[[2]]
                modelStats[[i]] <- fits[[3]]
                dataStats[[i]] <- fits[[4]]
                fits <- fits[[1]]
            }
            if (is.data.frame(fits)) {
                row.names(fits) <- rn
            }
            out[[i]] <- fits
        }
    }
    series.names <- names(rwl)
    names(out) <- series.names
    if(length(method2) == 1){
        out <- data.frame(out, row.names = rn)
        names(out) <- y.name
        if(return.info){
          modelCurves <- data.frame(modelCurves, row.names = rn)
          names(modelCurves) <- y.name
        }
    }
    if (return.info) {
        names(modelStats) <- series.names
        names(dataStats) <- series.names
        list(series = out, curves = modelCurves, model.info = modelStats, data.info = dataStats)
    } else {
        out
    }
}

### Helper function. Finds the longest common prefix in cnames.
common.prefix <- function(cnames) {
    n.char <- nchar(cnames)
    n <- length(cnames)
    y <- "";
    for (i in dec(min(n.char), 1)) {
        first.prefix <- substr(cnames[1], 1, i)
        prefix.match <- TRUE
        for (j in inc(2, n)) {
            if (first.prefix != substr(cnames[j], 1, i)) {
                prefix.match <- FALSE
                break
            }
        }
        if (prefix.match) {
            y <- first.prefix
            break
        }
    }
    y
}

### Helper function. Checks if ids and titles are consistent, i.e.
### trees are 1-to-1 mapped, cores of any tree are 1-to-1, etc. Also
### checks if all tree-core-radius-measurement rows are unique as
### required.
consistent.ids.titles <- function(ids, titles) {
    unique.trees.1 <- unique(ids$tree)
    unique.trees.2 <- unique(titles$tree)
    if (length(unique.trees.1) != length(unique.trees.2)) {
        return(FALSE)
    }
    ## All levels: tree, ...
    for (tree in unique.trees.1) {
        idx.t <- which(ids$tree %in% tree)
        if (length(idx.t) > 1) {
            if (any(titles$tree[idx.t] != titles$tree[idx.t[1]])) {
                return(FALSE)
            }
            unique.cores.1 <- unique(ids$core[idx.t])
            unique.cores.2 <- unique(titles$core[idx.t])
            if (length(unique.cores.1) != length(unique.cores.2)) {
                return(FALSE)
            }
            ## core, ...
            for (core in unique.cores.1) {
                idx.c <- idx.t[ids$core[idx.t] %in% core]
                if (length(idx.c) > 1) {
                    if (any(titles$core[idx.c] != titles$core[idx.c[1]])) {
                        return(FALSE)
                    }
                    unique.radii.1 <- unique(ids$radius[idx.c])
                    unique.radii.2 <- unique(titles$radius[idx.c])
                    if (length(unique.radii.1) != length(unique.radii.2)) {
                        return(FALSE)
                    }
                    ## radius, ...
                    for (radius in unique.radii.1) {
                        idx.r <- idx.c[ids$radius[idx.c] %in% radius]
                        length.idx <- length(idx.r)
                        if (length.idx > 1) {
                            if (any(titles$radius[idx.r] !=
                                    titles$radius[idx.r[1]])) {
                                return(FALSE)
                            }
                            unique.mments.1 <- unique(ids$measurement[idx.r])
                            unique.mments.2 <- unique(titles$measurement[idx.r])
                            ## when all else is the same,
                            ## "measurement" must be unique
                            if (length(unique.mments.1) != length.idx ||
                                length(unique.mments.2) != length.idx) {
                                return(FALSE)
                            }
                        }
                    }
                }
            }
        }
    }
    return(TRUE)
}

### Helper function. Creates title hierarchies (tree, core, radius,
### measurement) out of data.frame column names.
### TODO: Some options to this function?
create.title.hierarchy <- function(cnames, ids) {
    n <- length(cnames)
    max.nchar <- max(nchar(cnames))
    out.t <- character(length = n)
    out.c <- character(length = n)
    out.r <- character(length = n)
    out.m <- character(length = n)
    unique.trees <- unique(ids$tree)
    t.names <- character(length = length(unique.trees))
    names(t.names) <- unique.trees
    t.map <- list()
    ## All levels: tree, ...
    for (tree in unique.trees) {
        idx.t <- which(ids$tree %in% tree)
        t.map[[tree]] <- idx.t
        if (length(idx.t) > 1) {
            cp <- common.prefix(cnames[idx.t])
            t.names[tree] <- cp
            chars.used <- length(cp)
            unique.cores <- unique(ids$core[idx.t])
            n.uc <- length(unique.cores)
            c.names <- character(length = n.uc)
            names(c.names) <- unique.cores
            c.map <- list()
            ## core, ...
            for (core in unique.cores) {
                idx.c <- idx.t[ids$core[idx.t] %in% core]
                c.map[[core]] <- idx.c
                if (length(idx.c) > 1) {
                    if (n.uc == 1) {
                        c.names[core] <- "1"
                        chars.used.2 <- chars.used
                    } else {
                        cp <- common.prefix(substr(cnames[idx.c],
                                                   chars.used+1,
                                                   max.nchar))
                        c.names[core] <- cp
                        chars.used.2 <- chars.used + length(cp)
                    }
                    unique.radii <- unique(ids$radius[idx.c])
                    n.ur <- length(unique.radii)
                    r.names <- character(length = n.ur)
                    names(r.names) <- unique.radii
                    r.map <- list()
                    ## radius, ...
                    for (radius in unique.radii) {
                        idx.r <- idx.c[ids$radius[idx.c] %in% radius]
                        r.map[[radius]] <- idx.r
                        if (length(idx.r) > 1) {
                            if (n.ur == 1) {
                                r.names[radius] <- "1"
                                chars.used.3 <- chars.used.2
                            } else {
                                cp <- common.prefix(substr(cnames[idx.r],
                                                           chars.used.2+1,
                                                           max.nchar))
                                r.names[radius] <- cp
                                chars.used.3 <- chars.used.2 + length(cp)
                            }
                            ## measurement
                            for (idx.m in idx.r) {
                                out.m[idx.m] <- substr(cnames[idx.m],
                                                       chars.used.3+1,
                                                       max.nchar)
                            }
                            suppressWarnings(out.m[idx.r] <-
                                             fix.names(out.m[idx.r],
                                                       basic.charset = FALSE))
                        } else {
                            if (n.ur == 1) {
                                r.names[radius] <- "1"
                            } else {
                                r.names[radius] <-
                                    substr(cnames[idx.r], chars.used.2+1,
                                           max.nchar)
                            }
                            out.m[idx.r] <- "1"
                        }
                    }
                    suppressWarnings(r.names <-
                                     fix.names(r.names,
                                               basic.charset = FALSE))
                    for (radius in unique.radii) {
                        out.r[r.map[[radius]]] <- r.names[radius]
                    }
                } else {
                    if (n.uc == 1) {
                        c.names[core] <- "1"
                    } else {
                        c.names[core] <-
                            substr(cnames[idx.c], chars.used+1, max.nchar)
                    }
                    out.r[idx.c] <- out.m[idx.c] <- "1"
                }
            }
            suppressWarnings(c.names <-
                             fix.names(c.names, basic.charset = FALSE))
            for (core in unique.cores) {
                out.c[c.map[[core]]] <- c.names[core]
            }
        } else {
            t.names[tree] <- cnames[idx.t]
            out.c[idx.t] <- out.r[idx.t] <- out.m[idx.t] <- "1"
        }
    }
    suppressWarnings(t.names <- fix.names(t.names, basic.charset = FALSE))
    for (tree in unique.trees) {
        out.t[t.map[[tree]]] <- t.names[tree]
    }
    data.frame(tree = out.t, core = out.c, radius = out.r, measurement = out.m)
}

### Utility function (exported) which, given a suitable pith offset
### data.frame po as used by rcs, creates a (partial) wood
### completeness data.frame, counting pith offset exceeding the value
### 1 as missing heartwood.
po.to.wc <- function(po) {
    data.frame(n.missing.heartwood = as.integer(po[[2]] - 1),
               row.names = po[[1]])
}

### Helper function for handling arguments to write.tridas.
expand.metadata <- function(md.in, crn, default.value="") {
    if (is.null(md.in)) {
        md.out <- lapply(crn, function(x) rep(default.value, length(x)))
    } else if (is.character(md.in)) {
        if (length(md.in) == 0) {
            md.out <- lapply(crn, function(x) rep(default.value, length(x)))
        } else {
            md.in2 <- rep(md.in, length.out=length(crn))
            md.out <- list()
            for (k in seq_along(crn)) {
                md.out[[k]] <- rep(md.in2[k], length(crn[[k]]))
            }
        }
    }
    md.out
}

### Main function
write.tridas <- function(rwl.df = NULL, fname, crn = NULL,
                         prec = NULL, # no rounding
                         ids = NULL, titles = NULL,
                         crn.types = NULL,
                         crn.titles = NULL,
                         crn.units = NULL,
                         tridas.measuring.method = NA,
                         other.measuring.method = "unknown",
                         sample.type = "core",
                         wood.completeness = NULL,
                         taxon = "",
                         tridas.variable = "ring width",
                         other.variable = NA,
                         project.info = list(
                         type = c("unknown"), # multiple types possible
                         description = NULL,
                         title = "",
                         category = "",
                         investigator = "",
                         period = ""
                         ),
                         lab.info = data.frame(
                         name = "",
                         acronym = NA,
                         identifier = NA,
                         domain = "",
                         addressLine1 = NA,
                         addressLine2 = NA,
                         cityOrTown = NA,
                         stateProvinceRegion = NA,
                         postalCode = NA,
                         country = NA
                         ),
                         research.info = data.frame(
                         identifier = NULL,
                         domain = NULL,
                         description = NULL
                         ),
                         site.info = list(
                         type = "unknown",
                         description = NULL,
                         title = ""
                         ),
                         random.identifiers = FALSE,
                         identifier.domain = lab.info$name[1],
                         ...) {

    if (!is.data.frame(lab.info) || nrow(lab.info) < 1) {
        stop("'lab.info' must be a data.frame with at least one row")
    }
    lab.names <- names(lab.info)
    if (!("name" %in% lab.names)) {
        stop("\"name\" is a required variable in 'lab.info'")
    }
    identifier.present <- "identifier" %in% lab.names
    if (identifier.present && !("domain" %in% lab.names)) {
        stop("\"domain\" is required together with \"identifier\" in 'lab.info'")
    }

    if (!is.data.frame(research.info) || nrow(research.info) < 1) {
        research.present <- FALSE
    } else {
        research.names <- names(research.info)
        if (!("identifier" %in% research.names)) {
            stop("\"identifier\" is a required variable in 'research.info'")
        }
        if (!("domain" %in% research.names)) {
            stop("\"domain\" is a required variable in 'research.info'")
        }
        if (!("description" %in% research.names)) {
            stop("\"description\" is a required variable in 'research.info'")
        }
        research.present <- TRUE
    }

    ## Insist that certain variables exist and are of character type.
    ## Has two modes of operation:
    ## 1. checking a simple (non-list) variable,
    ## 2. checking the contents of a list (including data.frame).
    ##
    ## must.exist is a list of character vectors with the following
    ## components:
    ## 1.    variable name (base.name)
    ## 2.    default value (default.value, "" if missing)
    ## For triggering mode 2, additionally:
    ## 3.-n. names of list members (component.name).
    ##       For each, insists that base.name[[component.name]] exists
    ##       and is of character type.
    check.char.vars <- function(must.exist) {
        for (var.specs in must.exist) {
            base.name <- var.specs[1]
            this.var <- get(base.name)
            specs.length <- length(var.specs)
            if (specs.length < 3) {
                if (!is.character(this.var)) {
                    if (specs.length == 1) {
                        default.value <- ""
                    } else {
                        default.value <- var.specs[2]
                    }
                    warning(gettextf("'%s' must be of type character - inserting \"%s\"",
                                     base.name, default.value))
                    assign(base.name, default.value, inherits = TRUE)
                }
            } else {
                default.value <- var.specs[2]
                if (!is.list(this.var)) {
                    warning(gettextf("'%s' must be a list. Creating one.",
                                     base.name))
                    this.var <- list()
                }
                for (component.name in var.specs[3:specs.length]) {
                    if (!is.character(this.var[[component.name]])) {
                        warning(gettextf("'%s$%s' must be of type character - inserting \"%s\"",
                                         base.name, component.name,
                                         default.value))
                        this.var[[component.name]] <- default.value
                    }
                }
                assign(base.name, this.var, inherits = TRUE)
            }
        }
    }

    check.char.vars(list(c("project.info", "",
                           "title", "category", "investigator", "period"),
                         c("project.info", "unknown",
                           "type")))
    if (random.identifiers) {
        check.char.vars(list("identifier.domain"))
    }

    if (!is.na(tridas.variable)) {
        tridas.variable2 <-
            tridas.vocabulary("variable", term=tridas.variable)
    } else {
        tridas.variable2 <- NA
    }

    address.order <- c("addressLine1",
                       "addressLine2",
                       "cityOrTown",
                       "stateProvinceRegion",
                       "postalCode",
                       "country")

    ## TODO: Enable identifiers given by the user
    if (random.identifiers) {
        ugen <- uuid.gen(paste0("dplR",
                                packageDescription("dplR", fields = "Version"),
                                fname))
    }

    ## <tridas>
    doc <- simpleXML(fname, root="tridas",
                     xml.ns="http://www.tridas.org/1.2.2")
    on.exit(doc$close()) # emits </tridas>, closes file
    ## Shortcuts to functions (speedup?)
    doc.addTag <- doc$addTag
    doc.addTag.nc <- doc$addTag.noCheck
    doc.closeTag <- doc$closeTag

    ## <project> (using only one project here)
    doc.addTag.nc("project", close = FALSE)
    doc.addTag("title", project.info$title[1])
    if (random.identifiers) {
        doc.addTag("identifier",
                   ugen(),
                   attrs = c(domain = identifier.domain))
    }
    for (t in project.info$type) {
        doc.addTag("type", t)
    }
    doc.addTag("description", project.info$description[1])
    acronym.present <- "acronym" %in% lab.names
    address.order <- address.order[address.order %in% lab.names]
    for (i in seq_len(nrow(lab.info))) {
        ## <laboratory>
        doc.addTag.nc("laboratory", close = FALSE)
        if (identifier.present) {
            this.identifier <- lab.info$identifier[i]
            if (!is.na(this.identifier) && nzchar(this.identifier)) {
                doc.addTag("identifier",
                           this.identifier,
                           attrs = c(domain = lab.info$domain[i]))
            }
        }
        if (acronym.present) {
            this.acronym <- lab.info$acronym[i]
            if (!is.na(this.acronym) && nzchar(this.acronym)) {
                doc.addTag("name", lab.info$name[i],
                           attrs = c(acronym = this.acronym))
            } else {
                doc.addTag("name", lab.info$name[i])
            }
        } else {
            doc.addTag("name", lab.info$name[i])
        }
        ## <address>
        doc.addTag.nc("address", close = FALSE)
        for (address.line in address.order) {
            address.text <- lab.info[[address.line]][i]
            if (!is.na(address.text) && nzchar(address.text)) {
                doc.addTag(address.line, address.text)
            }
        }
        doc.closeTag() # </address>
        doc.closeTag() # </laboratory>
    }
    doc.addTag("category", project.info$category[1])
    doc.addTag("investigator", project.info$investigator[1])
    doc.addTag("period", project.info$period[1])

    if (research.present) {
        for (i in seq_len(nrow(research.info))) {
            ## <research>
            doc.addTag.nc("research", close = FALSE)
            doc.addTag("identifier",
                       research.info$identifier[i],
                       attrs = c(domain = research.info$domain[i]))
            doc.addTag("description", research.info$description[i])
            doc.closeTag() # </research>
        }
    }

    ## Preprocessing -- things related to rwl.df / <object>
    if (!is.null(rwl.df)) {
        if (!is.data.frame(rwl.df)) {
            stop("'rwl.df' must be a data.frame")
        }
        check.char.vars(list(c("site.info", "unknown", "type"),
                             c("site.info", "", "title")))
        n.col <- ncol(rwl.df)
        cnames <- names(rwl.df)
        stopifnot(is.character(cnames), !is.na(cnames),
                  Encoding(cnames) != "bytes")

        ## If 'ids' is NULL then assume one core, radius and
        ## measurement per tree.  In case of missing columns (less
        ## than 4), make similar assumptions.  Only basic checks about
        ## the validity of the 'ids' argument are performed;
        ## undetected problems may lead to errors or malformed output.
        ids2 <- ids
        titles2 <- titles
        if (is.null(ids2)) {
            ones <- rep(1, n.col)
            ids2 <- data.frame(tree = seq_len(n.col),
                               core = ones,
                               radius = ones,
                               measurement = ones)
        } else if (is.data.frame(ids2) && nrow(ids2) == n.col) {
            ncol.ids <- ncol(ids2)
            if (ncol.ids == 2) {
                ones <- rep(1, n.col)
                if (!all(c("tree","core") %in% names(ids2))) {
                    stop("2-col 'ids' needs \"tree\" and \"core\" columns")
                }
                ids2 <- data.frame(ids2,
                                   radius = ones,
                                   measurement = ones)
            } else if(ncol.ids == 3) {
                if (!all(c("tree","core","radius") %in% names(ids2))) {
                    stop("3-col 'ids' needs \"tree\", \"core\", and \"radius\" columns")
                }
                ids2 <- data.frame(ids2,
                                   measurement = rep(1, n.col))
            } else if (ncol.ids == 4) {
                if (!all(c("tree","core","radius","measurement") %in%
                         names(ids2))) {
                    stop("4-col 'ids' needs \"tree\", \"core\", \"radius\", and \"measurement\" columns")
                }
            } else {
                stop("argument 'ids' is in wrong format (2, 3, or 4 columns required)")
            }
        } else {
            stop("argument 'ids' is not data.frame or has wrong number of rows")
        }
        if (!all(vapply(ids2, is.numeric, TRUE))) {
            stop("'ids' must have numeric columns")
        }

        if (is.null(titles2)) {
            titles2 <- create.title.hierarchy(cnames, ids2)
        } else if (is.data.frame(titles2) && nrow(titles2) == n.col) {
            if (ncol(titles2) != 4 ||
                !all(c("tree", "core", "radius", "measurement") %in%
                     names(ids2))) {
                stop("columns needed in 'titles': \"tree\", \"core\", \"radius\", and \"measurement\"")
            }
        } else {
            stop("argument 'titles' is not data.frame or has wrong number of rows")
        }
        if (!consistent.ids.titles(ids2, titles2)) {
            stop("'ids' and 'titles' not consistent or duplicates present")
        }

        if (!is.null(prec)) {
            if (prec == 0.001) {
                data.unit <- "micrometres"
                rwl.df2 <- round(rwl.df * 1000)
            } else if (prec == 0.01) {
                data.unit <- "1/100th millimetres"
                rwl.df2 <- round(rwl.df * 100)
            } else if (prec == 0.05) {
                data.unit <- "1/20th millimetres"
                rwl.df2 <- round(rwl.df * 20)
            } else if (prec == 0.1) {
                data.unit <- "1/10th millimetres"
                rwl.df2 <- round(rwl.df * 10)
            } else if (prec == 1) {
                data.unit <- "millimetres"
                rwl.df2 <- round(rwl.df)
            } else if (prec == 10) {
                data.unit <- "centimetres"
                rwl.df2 <- round(rwl.df / 10)
            } else if (prec == 100) {
                data.unit <- "centimetres" # decimetres not in TRiDaS units
                rwl.df2 <- round(rwl.df / 100) * 10
            } else if (prec == 1000) {
                data.unit <- "metres"
                rwl.df2 <- round(rwl.df / 1000)
            } else {
                warning("unknown 'prec' specified: no unit conversion or rounding done")
                data.unit <- "millimetres"
                rwl.df2 <- rwl.df
            }
        } else {
            data.unit <- "millimetres"
            rwl.df2 <- rwl.df
        }
        tridas.measuring.method2 <- tridas.measuring.method
        if (!all(is.na(tridas.measuring.method2))) {
            for (k in seq_along(tridas.measuring.method2)) {
                if (!is.na(this.mm <- tridas.measuring.method2[k])) {
                    tridas.measuring.method2[k] <-
                        tridas.vocabulary("measuring method", term=this.mm)
                }
            }
        }
        if (length(tridas.measuring.method2) != n.col) {
            tridas.measuring.method2 <-
                rep(tridas.measuring.method2, length.out = n.col)
        }
        check.char.vars(list(c("other.measuring.method", "unknown")))
        other.measuring.method2 <- other.measuring.method
        if (length(other.measuring.method2) != n.col) {
            other.measuring.method2 <-
                rep(other.measuring.method2, length.out = n.col)
        }
        check.char.vars(list(c("sample.type", "core")))
        if (length(sample.type) != n.col) {
            sample.type2 <- rep(sample.type, length.out = n.col)
        } else {
            sample.type2 <- sample.type
        }

        ## Check (and fix) wood.completeness
        wood.completeness2 <- wood.completeness
        if (!is.null(wood.completeness2)) {
            if (nrow(wood.completeness2) != n.col) {
                stop("'nrow(wood.completeness)' must be equal to 'ncol(rwl.df)'")
            }
            if (any(row.names(wood.completeness2) != cnames)) {
                stop("row names of 'wood.completeness' must match column names of 'rwl.df'")
            }
            names.wc <- names(wood.completeness2)
            wc <- TRUE
            names.complex <- c("pith.presence", "heartwood.presence",
                               "sapwood.presence")
            names.nonnegative <-
                c("n.unmeasured.inner", "n.missing.sapwood", "n.sapwood",
                  "n.missing.heartwood", "n.unmeasured.outer")
            ## Create missing columns
            for (nam in names.complex[!(names.complex %in% names.wc)]) {
                wood.completeness2[[nam]] <- rep("unknown", n.col)
            }
            ## Find invalid < 0 numbers
            for (nam in names.nonnegative[names.nonnegative %in% names.wc]) {
                temp <- na.omit(wood.completeness2[[nam]])
                if (any(!is.int(temp) | temp < 0)) {
                    stop(gettextf("some values in 'wood.completeness$%s' are invalid, i.e. not integer or < 0", nam))
                }
            }
            ## Replace NAs and consistence with complex vocabulary
            for (nam in names.complex) {
                wood.completeness2[[nam]][is.na(wood.completeness2[[nam]])] <-
                    "unknown"
                wood.completeness2[[nam]] <-
                    tridas.vocabulary("complex presence / absence",
                                      term = wood.completeness2[[nam]])
            }
            ## Check bark.presence
            if (!("bark.presence" %in% names.wc)) {
                wood.completeness2$bark.presence <- rep("unknown", n.col)
            }
            idx.bark.na <- which(is.na(wood.completeness2$bark.presence))
            if (length(idx.bark.na) > 0) {
                wood.completeness2$bark.presence[idx.bark.na] <- "unknown"
            }
            wood.completeness2$bark.presence <-
                tridas.vocabulary("presence / absence",
                                  term = wood.completeness2$bark.presence)
            ## Check last.ring.presence
            if ("last.ring.presence" %in% names.wc) {
                idx.notna <-
                    which(!is.na(wood.completeness2$last.ring.presence))
                wood.completeness2$last.ring.presence[idx.notna] <-
                    tridas.vocabulary("presence / absence",
                                      term = wood.completeness2$last.ring.presence[idx.notna])
                wc.lrp <- TRUE
                if ("last.ring.details" %in% names.wc) {
                    wc.lrd <- TRUE
                } else {
                    wc.lrd <- FALSE
                }
            } else {
                wc.lrp <- FALSE
            }
            ## Check availability of metadata in wood.completeness here
            ## in order to avoid repeating it later
            if ("n.missing.sapwood" %in% names.wc) {
                wc.nms <- TRUE
                if ("missing.sapwood.foundation" %in% names.wc) {
                    wc.msf <- TRUE
                } else {
                    wc.msf <- FALSE
                }
            } else {
                wc.nms <- FALSE
            }
            if ("n.sapwood" %in% names.wc) {
                wc.ns <- TRUE
            } else {
                wc.ns <- FALSE
            }
            if ("n.missing.heartwood" %in% names.wc) {
                wc.nmh <- TRUE
                if ("missing.heartwood.foundation" %in% names.wc) {
                    wc.mhf <- TRUE
                } else {
                    wc.mhf <- FALSE
                }
            } else {
                wc.nmh <- FALSE
            }
            if ("n.unmeasured.inner" %in% names.wc) {
                wc.nui <- TRUE
            } else {
                wc.nui <- FALSE
            }
            if ("n.unmeasured.outer" %in% names.wc) {
                wc.nuo <- TRUE
            } else {
                wc.nuo <- FALSE
            }
        } else {
            wc <- FALSE
        }
        ## End of rwl.df / <object> preprocessing
    }

    ## Preprocessing -- things related to crn / <derivedSeries>
    crn2 <- crn
    crn.types2 <- crn.types
    crn.units2 <- crn.units
    crn.titles2 <- crn.titles
    if (!is.null(crn2)) {
        if (is.data.frame(crn2)) {
            crn2 <- list(crn2)
        }
        if (!is.list(crn.types2)) {
            crn.types2 <- expand.metadata(crn.types2, crn2, "")
        } else {
            crn.types2 <- rep(crn.types2, length.out=length(crn2))
        }
        if (!is.list(crn.units2)) {
            crn.units2 <- expand.metadata(crn.units2, crn2, NA)
        } else {
            crn.units2 <- rep(crn.units2, length.out=length(crn2))
        }
        if (!is.null(crn.titles2)) {
            titles.present <- TRUE
            if (!is.list(crn.titles2)) {
                crn.titles2 <- list(crn.titles2)
            }
            crn.titles2 <- rep(crn.titles2, length.out=length(crn2))
        } else {
            titles.present <- FALSE
        }
        if (titles.present && length(crn2) != length(crn.titles2)) {
            titles.present <- FALSE
        }
    }

    ## Write to file -- things related to rwl.df / <object>
    if (!is.null(rwl.df)) {
        ## <object> (site)
        doc.addTag.nc("object", close = FALSE)
        doc.addTag("title", site.info$title[1])
        if (random.identifiers) {
            doc.addTag("identifier",
                       ugen(),
                       attrs = c(domain = identifier.domain))
        }
        doc.addTag("type", site.info$type[1])
        if (is.character(site.info$description)) {
            doc.addTag("description", site.info$description)
        }

        unique.trees <- unique(ids2$tree)
        yrs.all <- as.numeric(row.names(rwl.df2))

        for (tree in unique.trees) {
            idx.t <- which(ids2$tree %in% tree)
            ## <element> (trees)
            doc.addTag.nc("element", close = FALSE)
            doc.addTag("title", titles2$tree[idx.t[1]])
            if (random.identifiers) {
                doc.addTag("identifier",
                           ugen(),
                           attrs = c(domain = identifier.domain))
            }
            doc.addTag("taxon", taxon);
            unique.cores <- unique(ids2$core[idx.t])
            for (core in unique.cores) {
                idx.c <- idx.t[ids2$core[idx.t] %in% core]
                ## <sample> (core)
                doc.addTag.nc("sample", close = FALSE)
                doc.addTag("title", titles2$core[idx.c[1]])
                if (random.identifiers) {
                    doc.addTag("identifier",
                               ugen(),
                               attrs = c(domain = identifier.domain))
                }
                doc.addTag("type", sample.type2[idx.c[1]])
                unique.radii <- unique(ids2$radius[idx.c])
                for (radius in unique.radii) {
                    idx.r <- idx.c[ids2$radius[idx.c] %in% radius]
                    ## <radius>
                    doc.addTag.nc("radius", close = FALSE)
                    doc.addTag("title", titles2$radius[idx.r[1]])
                    if (random.identifiers) {
                        doc.addTag("identifier",
                                   ugen(),
                                   attrs = c(domain = identifier.domain))
                    }
                    for (idx.m in idx.r) {
                        ## <measurementSeries>
                        doc.addTag.nc("measurementSeries", close = FALSE)
                        doc.addTag("title", titles2$measurement[idx.r])
                        if (random.identifiers) {
                            doc.addTag("identifier",
                                       ugen(),
                                       attrs = c(domain = identifier.domain))
                        }
                        doc.addTag("comments", cnames[idx.m])
                        ## <woodCompleteness>
                        if (wc) {
                            doc.addTag.nc("woodCompleteness",
                                          close = FALSE)
                            ## <nrOfUnmeasuredInnerRings>
                            if (wc.nui &&
                                !is.na(this.val <-
                                       wood.completeness2$n.unmeasured.inner[idx.m])) {
                                doc.addTag.nc("nrOfUnmeasuredInnerRings",
                                              this.val)
                            }
                            ## <nrOfUnmeasuredOuterRings>
                            if (wc.nuo &&
                                !is.na(this.val <-
                                       wood.completeness2$n.unmeasured.outer[idx.m])) {
                                doc.addTag.nc("nrOfUnmeasuredOuterRings",
                                              this.val)
                            }
                            ## <pith/>
                            doc.addTag.nc("pith",
                                          attrs = c(presence = wood.completeness2$pith.presence[idx.m]))
                            ## <heartwood>
                            doc.addTag.nc("heartwood",
                                          attrs = c(presence = wood.completeness2$heartwood.presence[idx.m]),
                                          close = FALSE)
                            if (wc.nmh &&
                                !is.na(this.val <-
                                       wood.completeness2$n.missing.heartwood[idx.m])) {
                                doc.addTag.nc("missingHeartwoodRingsToPith",
                                              this.val)
                                if (wc.mhf &&
                                    !is.na(this.val <-
                                           wood.completeness2$missing.heartwood.foundation[idx.m])) {
                                    doc.addTag("missingHeartwoodRingsToPithFoundation",
                                               this.val)
                                }
                            }
                            doc.closeTag() # </heartwood>
                            ## <sapwood>
                            doc.addTag.nc("sapwood",
                                          attrs = c(presence = wood.completeness2$sapwood.presence[idx.m]),
                                          close = FALSE)
                            if (wc.ns &&
                                !is.na(this.val <-
                                       wood.completeness2$n.sapwood[idx.m])) {
                                doc.addTag.nc("nrOfSapwoodRings",
                                              this.val)
                            }
                            if (wc.lrp &&
                                !is.na(this.val <-
                                       wood.completeness2$last.ring.presence[idx.m])) {
                                if (wc.lrd &&
                                    !is.na(this.detail <-
                                           wood.completeness2$last.ring.details[idx.m])) {
                                    doc.addTag("lastRingUnderBark",
                                               this.detail,
                                               attrs = c(presence = this.val))
                                } else {
                                    doc.addTag("lastRingUnderBark",
                                               attrs = c(presence = this.val))
                                }
                            }
                            if (wc.nms &&
                                !is.na(this.val <-
                                       wood.completeness2$n.missing.sapwood[idx.m])) {
                                doc.addTag.nc("missingSapwoodRingsToBark",
                                              this.val)
                                if (wc.msf &&
                                    !is.na(this.val <-
                                           wood.completeness2$missing.sapwood.foundation[idx.m])) {
                                    doc.addTag("missingSapwoodRingsToBarkFoundation",
                                               this.val)
                                }
                            }
                            doc.closeTag() # </sapwood>
                            ## <bark>
                            doc.addTag.nc("bark",
                                          attrs = c(presence = wood.completeness2$bark.presence[idx.m]))
                            doc.closeTag() # </woodCompleteness>
                        }
                        if (!is.na(this.mm <-
                                   tridas.measuring.method2[idx.m])) {
                            doc.addTag.nc("measuringMethod", NULL,
                                          attrs = c(normalTridas=this.mm))
                        } else {
                            doc.addTag("measuringMethod",
                                       other.measuring.method2[idx.m])
                        }
                        ## <interpretation>
                        doc.addTag.nc("interpretation", close = FALSE)
                        series <- as.numeric(rwl.df2[[idx.m]])
                        idx <- !is.na(series)
                        series <- series[idx]
                        yrs <- yrs.all[idx]
                        min.year <- min(yrs)
                        max.year <- max(yrs)
                        if (min.year < 1) {
                            doc.addTag.nc("firstYear",
                                          1 - min.year,
                                          attrs = c(suffix = "BC"))
                        } else {
                            doc.addTag.nc("firstYear",
                                          min.year,
                                          attrs = c(suffix = "AD"))
                        }
                        if (max.year < 1) {
                            doc.addTag.nc("lastYear",
                                          1 - max.year,
                                          attrs = c(suffix = "BC"))
                        } else {
                            doc.addTag.nc("lastYear",
                                          max.year,
                                          attrs = c(suffix = "AD"))
                        }
                        doc.closeTag() # </interpretation>
                        ## <values>
                        doc.addTag.nc("values", close = FALSE)
                        if (!is.na(tridas.variable2)) {
                            doc.addTag.nc("variable", NULL,
                                          attrs = c(normalTridas = tridas.variable2))
                        } else {
                            doc.addTag("variable", other.variable)
                        }
                        doc.addTag.nc("unit", NULL,
                                      attrs = c(normalTridas = data.unit))
                        for (i in seq_along(series)) {
                            doc.addTag.nc("value", NULL,
                                          attrs = c(value = series[i]))
                        }
                        doc.closeTag() # </values>
                        doc.closeTag() # </measurementSeries>
                    }
                    doc.closeTag() # </radius>
                }
                doc.closeTag() # </sample>
            }
            doc.closeTag() # </element>
        }
        doc.closeTag() # </object>
    }

    ## Write to file -- things related to crn / <derivedSeries>
    if (!is.null(crn2)) {
        for (i in seq_along(crn2)) {
            this.frame <- crn2[[i]]
            yrs.all <- as.numeric(row.names(this.frame))
            crn.names <- names(this.frame)
            depth.idx <- grep("^samp[.]depth", crn.names)
            n.depth <- length(depth.idx)
            if (n.depth > 0) {
                depth.present <- TRUE
                series.idx <- setdiff(seq_along(crn.names), depth.idx)
                n.series <- length(series.idx)
                depth.idx <- rep(depth.idx, length.out = n.series)
            } else {
                depth.present <- FALSE
                n.series <- length(crn.names)
                series.idx <- seq_len(n.series)
            }
            this.typevec <- as.character(crn.types2[[i]])
            n.type <- length(this.typevec)
            if (n.type == 0) {
                this.typevec <- rep("", n.series)
            } else {
                this.typevec <- rep(this.typevec, length.out = n.series)
            }
            this.unitvec <- as.character(crn.units2[[i]])
            n.unit <- length(this.unitvec)
            if (n.unit == 0) {
                this.unitvec <- rep(NA, n.series)
            } else {
                this.unitvec <- rep(this.unitvec, length.out = n.series)
            }
            if (titles.present) {
                this.titlevec <- as.character(crn.titles2[[i]])
                n.title <- length(this.titlevec)
                if (n.title == 0) {
                    this.titlevec <- rep(NA, n.series)
                } else {
                    this.titlevec <- rep(this.titlevec, length.out = n.series)
                }
            }
            if (depth.present) {
                n.depth <- length(depth.idx)
            }
            for (j in seq_len(n.series)) {
                ## <derivedSeries>
                this.idx <- series.idx[j]
                series <- as.numeric(this.frame[[this.idx]])
                if (depth.present) {
                    samp.depth <- as.numeric(this.frame[[depth.idx[j]]])
                }
                doc.addTag.nc("derivedSeries", close = FALSE)
                this.crn.name <- crn.names[this.idx]
                if (titles.present) {
                    this.title <- this.titlevec[j]
                    if (is.na(this.title)) {
                        this.title <- this.crn.name
                        this.title.present <- FALSE
                    } else {
                        this.title.present <- TRUE
                    }
                } else {
                    this.title <- this.crn.name
                    this.title.present <- FALSE
                }
                doc.addTag("title", this.title)
                if (random.identifiers) {
                    doc.addTag("identifier",
                               ugen(),
                               attrs = c(domain = identifier.domain))
                }
                if (this.title.present && this.title != this.crn.name) {
                    doc.addTag("comments", this.crn.name)
                }
                doc.addTag("type", this.typevec[j])
                ## <linkSeries/>
                ## TODO: Add actual links, now just an empty element
                doc.addTag.nc("linkSeries")

                ## <interpretation>
                doc.addTag.nc("interpretation", close = FALSE)
                idx <- !is.na(series)
                series <- series[idx]
                yrs <- yrs.all[idx]
                min.year <- min(yrs)
                max.year <- max(yrs)
                if (min.year < 1) {
                    doc.addTag.nc("firstYear",
                                  1 - min.year,
                                  attrs = c(suffix = "BC"))
                } else {
                    doc.addTag.nc("firstYear",
                                  min.year,
                                  attrs = c(suffix = "AD"))
                }
                if (max.year < 1) {
                    doc.addTag.nc("lastYear",
                                  1 - max.year,
                                  attrs = c(suffix = "BC"))
                } else {
                    doc.addTag.nc("lastYear",
                                  max.year,
                                  attrs = c(suffix = "AD"))
                }
                doc.closeTag() # </interpretation>

                ## <values>
                doc.addTag.nc("values", close = FALSE)
                if (!is.na(tridas.variable2)) {
                    doc.addTag.nc("variable", NULL,
                                  attrs = c(normalTridas=tridas.variable2))
                } else {
                    doc.addTag("variable", other.variable)
                }
                this.unit <- this.unitvec[j]
                if (is.na(this.unit)) {
                    doc.addTag.nc("unitless", NULL)
                } else {
                    doc.addTag("unit", this.unit)
                }
                if (depth.present) {
                    for (i in seq_along(series)) {
                        doc.addTag.nc("value", NULL,
                                      attrs = c(count = samp.depth[i],
                                      value = series[i]))
                    }
                } else {
                    for (i in seq_along(series)) {
                        doc.addTag.nc("value", NULL,
                                      attrs = c(value = series[i]))
                    }
                }
                doc.closeTag() # </values>
                doc.closeTag() # </derivedSeries>
            }
        }
    }
    doc.closeTag() # </project>
    fname
}

`chron` <-
    function(x, biweight=TRUE, prewhiten=FALSE, ...)
{
    check.flags(biweight, prewhiten)
    samps <- rowSums(!is.na(x))
    if (!biweight) {
        std <- rowMeans(x, na.rm=TRUE)
    } else {
        std <- apply(x, 1, tbrm, C=9)
    }
    if (prewhiten) {
        x.ar <- apply(x, 2, ar.func, ...)
        if (!biweight) {
            res <- rowMeans(x.ar, na.rm=TRUE)
        } else {
            res <- apply(x.ar, 1, tbrm, C=9)
        }
        res[is.nan(res)] <- NA
        out <- data.frame(std, res, samps)
        names(out) <- c("std",
                        "res",
                        "samp.depth")
    } else {
        out <- data.frame(std, samps)
        names(out) <- c("std", "samp.depth")
    }
    row.names(out) <- row.names(x)
    class(out) <- c("crn", "data.frame")
    out
}

context("function chron")
test.chron <- function() {
    ## RNG Setup for srs1 (we use a _particular_ random series...)
    if (!exists(".Random.seed", 1)) {
        if (getRversion() <= "3.0.0") {
            runif(1)
        } else {
            set.seed(NULL)
        }
    }
    seed <- get(".Random.seed", 1)
    on.exit(assign(".Random.seed", seed, 1))
    RNGversion("2.15.0")
    set.seed(0)

    ## Other setup
    N <- 500
    srs1 <- pmax(rnorm(N, 1, 0.4), 0.1)
    dat1 <- data.frame(srs1 - 0.05, srs1, srs1 + 0.05)
    res1.1 <- chron(dat1, biweight = FALSE, prewhiten = TRUE)
    res1.2 <- chron(dat1, biweight = TRUE, prewhiten = FALSE)

    srs2 <- 0.5 * sin(pi / 50 * seq_len(N)) + 1 # period is 100
    sd2.1 <- sd(srs2)
    dat2 <- data.frame(srs2 - 0.1, srs2, srs2 + 0.1)
    res2 <- chron(dat2, biweight = FALSE, prewhiten = TRUE)
    sd2.2 <- sd(res2[["res"]], na.rm=TRUE)
    threes <- rep.int(3, N)

    ## Test
    test_that("chron works (no autocorrelation)", {
        expect_equal(res1.1[["samp.depth"]], threes)
        expect_equal(res1.2[["samp.depth"]], threes)
        expect_equal(res1.1[["std"]], srs1)
        expect_equal(res1.2[["std"]], srs1)
        expect_equal(res1.1[["res"]], srs1)
    })
    test_that("chron works (with autocorrelation)", {
        expect_equal(res2[["samp.depth"]], threes)
        expect_equal(res2[["std"]], srs2)
        expect_true(length(which(is.na(res2[["res"]]))) > 0)
        expect_true(sd2.2 < sd2.1)
    })
}
test.chron()

bai.out <- function(rwl, diam = NULL) {
  
  if(!is.data.frame(rwl))
    stop("'rwl' must be a data.frame")
  if(!is.null(diam)) {
    if(ncol(rwl) != nrow(diam))
      stop("dimension problem: ", "'ncol(rw)' != 'nrow(diam)'")
    if(!all(diam[, 1] == names(rwl))){
      print(data.frame(rwlNames=names(rwl),seriesID=diam[,1],test=diam[, 1] == names(rwl)))
      stop("series ids in 'diam' and 'rwl' do not match exactly.")
    }
    if(!is.data.frame(diam)){
      stop("'diam' must be a data.frame")
    }
    
    diam.vec <- diam[, 2]
  }
  
  out <- rwl
  ## vector of years
  n.vec <- seq_len(nrow(rwl))
  for(i in seq_len(ncol(rwl))){
    ## series to work with
    dat <- rwl[[i]]
    ## strip out data from NA
    dat2 <- na.omit(dat)
    ## get diameter if not given
    if(is.null(diam)) d <- sum(dat2)*2
    else d <- diam.vec[i]
    ## get ring area
    r0 <- d/2 - c(0, cumsum(rev(dat2)))
    bai <- -pi*rev(diff(r0*r0))
    ## find NA / not NA locations
    na <- attributes(dat2)$na.action
    no.na <- n.vec[!n.vec %in% na]
    ## write result
    out[no.na, i] <- bai
  }
  ## return result
  out
}

### Written by Mikko Korpela.
simpleXML <- function(fname, root="tridas", xml.ns=NULL,
                      line.term="\x0D\x0A", indent.step=2) {
### \x0D\x0A is CR+LF, ASCII carriage return and line feed.
### Another possible choice: \x0A, i.e. just the line feed.

    ## Takes _one_ string x, converts it to UTF-8 (if not UTF-8 already),
    ## and removes characters not allowed in XML.
    to.xml.utf8 <- function(x) {
        enc <- Encoding(x) # 3 possible values: "latin1", "UTF-8", or "unknown"
        ## If x not already UTF-8, we convert...
        if (enc == "unknown") {
            y <- iconv(x, from="", to="UTF-8")  # from encoding of the locale
        } else if (enc == "latin1") {
            y <- iconv(x, from=enc, to="UTF-8") # from latin1
        } else {
            y <- x
        }
        y.int <- utf8ToInt(y)
        ## Accepted characters are from XML 1.0 spec
        ## http://www.w3.org/TR/2008/REC-xml-20081126/
        good.chars <-
            (y.int == 0x9 |
             y.int == 0xA |
             y.int == 0xD |
             (y.int >= 0x20 & y.int <= 0xD7FF) |
             (y.int >= 0xE000 & y.int <= 0xFFFD) |
             (y.int >= 0x10000 & y.int <= 0x10FFFF))
        paste(unlist(strsplit(y, split="", fixed=TRUE))[good.chars],
              collapse = "")
    }

    close.tag <- function() {
        if (stack.pointer > 0) {
            indent <<- indent - indent.step
            cat(rep(" ", indent),
                "</", tag.stack[stack.pointer], ">", line.term,
                sep="", file=f)
            stack.pointer <<- stack.pointer - 1
        }
    }

    ## Escapes characters not allowed in the value of an attribute
    escape.attribute <- function(x) {
        gsub("\t", "&\t;",
             gsub("\n", "&\n;",
                  gsub("\r", "&\r;",
                       gsub('"', "&quot;",
                            gsub("<", "&lt;",
                                 gsub("&", "&amp;", x,
                                      fixed = TRUE, useBytes=FALSE),
                                 fixed = TRUE, useBytes=FALSE),
                            fixed = TRUE, useBytes=FALSE),
                       fixed = TRUE, useBytes=FALSE),
                  fixed = TRUE, useBytes=FALSE),
             fixed = TRUE, useBytes=FALSE)
    }

    ## Escapes characters not allowed in content
    escape.content <- function(x) {
        gsub("\r", "&\r;",
             gsub(">", "&gt;",
                  gsub("<", "&lt;",
                       gsub("&", "&amp;", x,
                            fixed = TRUE, useBytes=FALSE),
                       fixed = TRUE, useBytes=FALSE),
                  fixed = TRUE, useBytes=FALSE),
             fixed = TRUE, useBytes=FALSE)
    }

    f <- file(description = fname, open = "w", encoding = "UTF-8")
    tag.stack <- character(10)
    tag.stack[1] <- root
    stack.pointer <- 1
    cat('<?xml version="1.0" encoding="UTF-8"?>', line.term,
        '<', root, sep="", file=f)
    if (!is.null(xml.ns)) {
        cat(' xmlns="', xml.ns, '"', sep="", file=f)
    }
    cat('>', line.term, sep="", file=f)
    indent <- indent.step

    list(# This list of functions (function closures) is returned
         addTag = function(tag, content=NULL, attrs=character(0), close=TRUE) {
             cat(rep(" ", indent), "<", tag, sep="", file=f)
             for (attr in names(attrs)) {
                 cat(" ", attr, '="',
                     escape.attribute(to.xml.utf8(as.character(attrs[attr]))),
                     '"', sep="", file=f)
             }
             if (close) {
                 if (length(content) == 0) {
                     cat("/>", line.term, sep="", file=f)
                 } else {
                     cat(">",
                         escape.content(to.xml.utf8(paste(content,
                                                          collapse=" "))),
                         "</", tag, ">", line.term,
                         sep="", file=f)
                 }
             } else {
                 cat(">", line.term, sep="", file=f)
                 indent <<- indent + indent.step
                 tag.stack[stack.pointer <<- stack.pointer+1] <<- tag
             }
         },
         addTag.noCheck = function(tag, content=NULL, attrs=character(0),
         close=TRUE) {
             cat(rep(" ", indent), "<", tag, sep="", file=f)
             for (attr in names(attrs)) {
                 cat(" ", attr, '="', attrs[attr], '"', sep="", file=f)
             }
             if (close) {
                 if (length(content) == 0) {
                     cat("/>", line.term, sep="", file=f)
                 } else {
                     cat(">", content, "</", tag, ">", line.term,
                         sep="", file=f)
                 }
             } else {
                 cat(">", line.term, sep="", file=f)
                 indent <<- indent + indent.step
                 tag.stack[stack.pointer <<- stack.pointer+1] <<- tag
             }
         },
         closeTag = close.tag,
         close = function() {
             while (stack.pointer > 0) {
                 close.tag()
             }
             close(f)
         }
         )
}

`caps` <- function(y, nyrs=32, f= 0.5)
{
  y <- as.numeric(y)
  ## If as.numeric() does not signal an error, it is unlikely that
  ## the result would not be numeric, but...
  if(!is.numeric(y)) stop("'y' must be coercible to a numeric vector")
  
  nobs <- length(y)
  
  ## quick error check
  if (nobs < 3) stop("there must be at least 3 data points")
  
  if(!is.numeric(f) || length(f) != 1 || f < 0 || f > 1)
    stop("'f' must be a number between 0 and 1")
  
  if(!is.numeric(nyrs) || length(nyrs) != 1 || nyrs <= 0)
    stop("'nyrs' must be a number greater than 0")

  if(nyrs > nobs)
    warning("Spline stiffness (nyrs) is greater than the length of series (y)")
  # if nyrs is between 0 and 1 treat it as % spline.
  # e,g, nyrs = 0.667 would be a 2/3 spline
  if(nyrs <= 1 & nyrs > 0)
    nyrs <- nyrs * nobs
  
  # some error checks
  ySpl <-.Call(dplR.c_caps_f,
               y = as.double(y),
               n=as.integer(nobs),
               stiffness = as.integer(nyrs),
               pct = as.double(f),
               res = as.double(rep(0,nobs)))
  if(length(ySpl)==1 && ySpl == 9999){
    stop("SBR matrix not positive definite")
  }
  if(f==1){
    ySpl <- y
  }
  return(ySpl)
}

bakker <- function (rwl, ancillary) { 
  if (!is.data.frame(rwl)) 
    stop("'rwl' must be a data.frame")
  
  if (ncol(rwl) != nrow(ancillary)) 
    stop("dimension problem: ", "'ncol(rw)' != 'nrow(ancillary)'")
  
  if (ncol(ancillary) != 4) 
    stop("dimension problem: ", "'ncol(ancillary)' != 4")
  if (!all(ancillary[, 1] == names(rwl))) {
    print(data.frame(rwlNames = names(rwl), seriesID = ancillary[, 1], test = ancillary[, 1] == names(rwl)))
    stop("series ids in 'ancillary' and 'rwl' do not match exactly.")
  }
  if (!is.data.frame(ancillary)) {
    stop("'ancillary' must be a data.frame")
  }
  rwl_d2pith <- ancillary[, "d2pith"]
  rwl_diam <- ancillary[, "diam"]
  rwl_po<-ancillary[,"PO"]
  
  #filling in the missing rings towards the pith is recommended when there are multiple cores per tree. 
  #average multiple BAI time series of a tree, not raw ring width measurements.
  #Otherwise, one can end up with an artifact, i.e. one negative increment when two radii 
  #have different growth rates and the faster growing sample has more rings.
  #The filled rings will be deleted at the end.
  
  rwl_filled<-NULL
  for (i in 1:length(rwl)){
    if(rwl_d2pith[i] !=0){
      temp<-data.frame(rep(rwl_d2pith[i]/rwl_po[i],times=rwl_po[i])) #create a dummy time series with the number of estimated rings missing 
      temp2<-na.omit(rwl[,i])
      select_remove<-c(attr(temp2,"na.action"))
      if(!is.null(select_remove)){
        temp2<-data.frame(temp2,row.names=rownames(rwl)[-select_remove]) 
      }else{
        temp2<- data.frame(temp2,row.names=rownames(rwl))
      }
      rownames(temp)<-c((as.numeric(rownames(temp2))[1]-rwl_po[i]):(as.numeric(rownames(temp2))[1]-1)) 
      colnames(temp)<-colnames(temp2)
      rwl_filled[[i]]<-ts(rbind(temp,temp2),start=as.numeric(rownames(temp))[1]) #paste the dummy time series and ring width measurements
    } else { 
      temp3<-data.frame(rwl[,i],row.names=rownames(rwl))
      rwl_filled[[i]]<-ts(temp3,start=as.numeric(rownames(temp3))[1])
    }
    
    
  }
  names(rwl_filled)<-colnames(rwl)
  rwl_filled<-do.call(cbind,rwl_filled)
  rownames(rwl_filled)<-c(tsp(rwl_filled)[1]:tsp(rwl_filled)[2])
  
  cum.na <- function(x) { #cumulative sum that changes NAs to zeros. helper function
    x[which(is.na(x))] <- 0
    return(cumsum(x))
  }
  
  IP <- apply(rwl_filled, 2, sum,na.rm=T) #how long is your measured radius including missing distance to pith?
  IH <- rwl_filled
  G <- IH
  for(i in 1:ncol(rwl_filled)){IH[,i] <-cum.na(rwl_filled[,i])}  #cumulative sum
  for(i in 1:ncol(IH)){G[,i]<-(IP[i]-(IH[,i]))/IP[i]} #proportional cumulative sum, reversed
  
  DBHhist<-G
  colnames(DBHhist)<-colnames(rwl)
  for (i in 1:ncol(DBHhist)){
    DBHhist[,i]<-rwl_diam[i]-(G[,i] *rwl_diam[i]) #subtract proportional DBH from measured DBH, the final proportional historic DBH
  }
  
  for(i in 1:ncol(DBHhist)){  #clean zeros, fill with NAs
    temp<-which(is.na(rwl_filled[,i]))
    if(any(which(diff(temp)>1))){
      DBHhist[which(is.na(rwl_filled[,i]))[-which(diff(temp)>1)],i]<-NA
    }else{
      DBHhist[rev(which(is.na(rwl_filled[,i]))),i]<-NA 
    }
  }
  
  if(nrow(rwl_filled)-nrow(rwl)==0){
    DBHhist2<- DBHhist[1:nrow(DBHhist),]
    DBHhist2[is.na(rwl)]<-NA
  }else{
    DBHhist2<- DBHhist[-c(1:(nrow(rwl_filled)-nrow(rwl))),]
    DBHhist2[is.na(rwl)]<-NA
  }
  
  
  out <- DBHhist
  n.vec <- seq_len(nrow(DBHhist))
  for (i in seq_len(ncol(rwl))) {
    dat <- DBHhist[,i]
    dat2 <- na.omit(dat)
    
    
    r0 <- dat2/2*10 #if dbh in cm, if dbh in mm, then remove *10; /2 to convert to radius
    bai <- c(pi*(r0[1]*r0[1]),pi * (diff(r0 * r0)))
    na <- attributes(dat2)$na.action
    no.na <- n.vec[!n.vec %in% na]
    out[no.na, i] <-  bai
  }
  
  if(nrow(rwl_filled)-nrow(rwl)==0){
    out2<- out[1:nrow(out),]
    out2[is.na(rwl)]<-NA
  }else{
    out2<- out[-c(1:(nrow(rwl_filled)-nrow(rwl))),]
    out2[is.na(rwl)]<-NA #this removes all the filled dummy values from rwl_filled above
  }
  
  
  output<-list(DBHhist_raw=DBHhist2,baiBakker_raw=out2)
  #I termed it raw, because at some point back in time I wanted to expand the function
  #to fill the innermost rings of the sample with less rings towards the center with data from the other core(s).
  #Never got to it, though.
  
}

read.fh <- function(fname, BC_correction = FALSE) {
  inp <- readLines(fname, ok=TRUE, warn=FALSE)
  
  ## Get start and end positions of headers and data blocks
  header.begin <- grep("^HEADER:$", inp)
  header.end <- grep("^DATA:(Tree|Single)$", inp)
  n <- length(header.end)
  if(n == 0) {
    stop('file has no data in "Tree" or "Single" formats')
  }
  ## For each data block in one of the supported formats, find the
  ## corresponding header block
  header.taken <- logical(length(header.begin))
  for (i in seq_len(n)) {
    n.preceding <- sum(header.begin < header.end[i] - 1)
    if (n.preceding == 0 || header.taken[n.preceding]) {
      stop("invalid file: HEADER and DATA don't match")
    } else {
      header.taken[n.preceding] <- TRUE
    }
  }
  if (!all(header.taken)) {
    warning("more HEADER blocks than DATA blocks in supported formats")
  }
  ## For each data block in one of the supported formats, find the
  ## following header block (or end of file)
  data.end <- numeric(n)
  for (i in seq_len(n-1)) {
    tmp <- header.begin[header.begin > header.end[i]]
    data.end[i] <- tmp[1]
  }
  tmp <- header.begin[header.begin > header.end[n]]
  if (length(tmp) > 0) {
    data.end[n] <- tmp[1]
  } else {
    data.end[n] <- length(inp) + 1
  }
  ## Forget headers that are not used by the data blocks
  header.begin <- header.begin[header.taken]
  
  ## Get essential metadata from headers
  keycodes <- character(n)
  lengths <- numeric(n)
  end.years <- numeric(n)
  start.years <- numeric(n)
  multipliers <- rep(1, n)
  divisors <- rep(100, n)
  site.code <- rep(NA_character_, n)
  tree.vec <- rep(NA_real_, n)
  core.vec <- rep(NA_real_, n)
  radius.vec <- rep(NA_real_, n)
  stemdisk.vec <- rep(NA_real_, n)
  pith.offset <- rep(NA_real_, n)
  for (i in seq_len(n)) {
    this.header <- inp[(header.begin[i]+1):(header.end[i]-1)]
    ## get keycode (= series id)
    this.keycode <- sub("KeyCode=", "", fixed=TRUE,
                        x=grep("^KeyCode=|Keycode=", this.header, value=TRUE))
    if (length(this.keycode) != 1) {
      string2 <- gettext('number of "KeyCode" lines is not 1',
                         domain="R-dplR")
      stop(gettextf("in series %s: ", as.character(i), domain="R-dplR"),
           string2, domain=NA)
    } else {
      keycodes[i] <- this.keycode
    }
    ## get length
    this.length <- sub("Length=", "", fixed=TRUE,
                         x=grep("^Length=", this.header, value=TRUE))
    if (length(this.length) != 1) {
      string2 <- gettext('number of "Length" lines is not 1',
                         domain="R-dplR")
      stop(gettextf("in series %s: ", keycodes[i], domain="R-dplR"),
           string2, domain=NA)
    } else {
      lengths[i] <- as.numeric(this.length)
    }
    ## get end year
    this.end.year <- sub("DateEnd=", "", fixed=TRUE,
                         x=grep("^DateEnd=", this.header, value=TRUE))
    if (length(this.end.year) != 1) {
      string2 <- gettext('number of "DateEnd" lines is not 1',
                         domain="R-dplR")
      stop(gettextf("in series %s: ", keycodes[i], domain="R-dplR"),
           string2, domain=NA)
    } else {
      end.years[i] <- as.numeric(this.end.year)
    }
    ## get start year
    this.start.year <- sub("DateBegin=", "", fixed=TRUE,
                           x=grep("^DateBegin=", this.header, value=TRUE))
    if (length(this.start.year) != 1) {
      if(length(this.end.year) == 1) {
        start.years[i] <- end.years[i]- lengths[i] +1
      } else {
        string2 <- gettext('number of "DateBegin" lines is not 1',
                         domain="R-dplR")
        stop(gettextf("in series %s: ", keycodes[i], domain="R-dplR"),
           string2, domain=NA) }
      } else {
        start.years[i] <- as.numeric(this.start.year)
        # check for BC - AD dates and add not existing year 0
        if (start.years[i] <0 && end.years[i]>0 && BC_correction == TRUE) {
          start.years[i] <- start.years[i]+1 }
      }
    ## correct BC dates to +1, because rwl-format uses non-existing year 0 and this is not necessary in the FH-format
    if (start.years[i]<0 && end.years[i]<0 && BC_correction == TRUE) {
      start.years[i] <- start.years[i] + 1
      end.years[i] <- end.years[i] + 1
    }
    ## get unit (by default, divide by 100)
    this.unit <- sub("Unit=", "", fixed=TRUE,
                     x=grep("^Unit=", this.header, value=TRUE))
    if (length(this.unit) == 1) {
      this.unit <- sub("mm", "", this.unit, fixed=TRUE)
      div.loc <- regexpr("/", this.unit, fixed=TRUE)
      if (div.loc > 0) {
        multipliers[i] <- as.numeric(substr(this.unit, 1, div.loc-1))
        divisors[i] <- as.numeric(substr(this.unit, div.loc+1,
                                         nchar(this.unit)))
      } else {
        multipliers[i] <- as.numeric(this.unit)
        divisors[i] <- 1
      }
      if (is.na(multipliers[i]) || is.na(divisors[i])) {
        string2 <- gettext('cannot interpret "Unit" line',
                           domain="R-dplR")
        stop(gettextf("in series %s: ", keycodes[i], domain="R-dplR"),
             string2, domain=NA)
      }
    } else if (length(this.unit) > 1) {
      string2 <- gettext('number of "Unit" lines is > 1',
                         domain="R-dplR")
      stop(gettextf("in series %s: ", keycodes[i], domain="R-dplR"),
           string2, domain=NA)
    }
    ## get site code
    this.site <- sub("SiteCode=", "", fixed=TRUE,
                     x=grep("^SiteCode=", this.header, value=TRUE))
    if (length(this.site) == 1) {
      site.code[i] <- this.site
    }
    ## get tree number
    this.tree <- sub("TreeNo=", "", fixed=TRUE,
                     x=grep("^TreeNo=", this.header, value=TRUE))
    if (length(this.tree) == 1) {
      tmp <- suppressWarnings(as.numeric(this.tree))
      if (identical(tmp, round(tmp))) {
        tree.vec[i] <- tmp
      }
    }
    ## get core number
    this.core <- sub("CoreNo=", "", fixed=TRUE,
                     x=grep("^CoreNo=", this.header, value=TRUE))
    if (length(this.core) == 1) {
      tmp <- suppressWarnings(as.numeric(this.core))
      if (identical(tmp, round(tmp))) {
        core.vec[i] <- tmp
      }
    }
    ## get radius number
    this.radius <- sub("RadiusNo=", "", fixed=TRUE,
                       x=grep("^RadiusNo=", this.header, value=TRUE))
    if (length(this.radius) == 1) {
      tmp <- suppressWarnings(as.numeric(this.radius))
      if (identical(tmp, round(tmp))) {
        radius.vec[i] <- tmp
      }
    }
    ## get stem disk number
    this.stemdisk <- sub("StemDiskNo=", "", fixed=TRUE,
                         x=grep("^StemDiskNo=", this.header, value=TRUE))
    if (length(this.stemdisk) == 1) {
      tmp <- suppressWarnings(as.numeric(this.stemdisk))
      if (identical(tmp, round(tmp))) {
        stemdisk.vec[i] <- tmp
      }
    }
    ## get pith offset (missing rings before start of series)
    this.missing <-
      sub("MissingRingsBefore=", "", fixed=TRUE,
          x=grep("^MissingRingsBefore=", this.header, value=TRUE))
    if (length(this.missing) == 1) {
      tmp <- suppressWarnings(as.numeric(this.missing))
      if (identical(tmp, round(tmp)) && tmp >= 0) {
        pith.offset[i] <- tmp + 1
      }
    }
  }
  
  ## calculate time span for data.frame
  min.year <- min(start.years)
  r.off <- min.year - 1
  max.year <- max(end.years)
  span <- min.year:max.year
  dendro.matrix <- matrix(NA, ncol = n, nrow = length(span))
  colnames(dendro.matrix) <- keycodes
  rownames(dendro.matrix) <- span
  ## get rid of comments (if any)
  strip.comment <- function(x) {
    strsplit(x, ";")[[1]][1]
  }
  for (i in seq_len(n)) { # loop through data blocks
    portion.start <- header.end[i] + 1
    portion.end <- data.end[i] - 1
    n.expected <- end.years[i] - start.years[i] + 1
    if (n.expected < 0) {
      stop(gettext("Series ", keycodes[i], " has an end date before the start date"))
    }
    if (portion.end < portion.start) {
      stop(gettextf("in series %s: ", keycodes[i], domain="R-dplR"),
           gettextf("too few values (expected %d, got %d)",
                    n.expected, 0, domain="R-dplR"), domain=NA)
    }
    portion <- inp[portion.start:portion.end]
    if (nchar(portion[1]) < 60 ||
        grepl(";", portion[1], fixed=TRUE)) { # data is in column format
      data <- as.numeric(vapply(portion, strip.comment, "foo"))
    } else { # data is in block format
      data <- numeric(length(portion) * 10)
      for (j in seq_along(portion)) {
        row.fwf <- substring(portion[j],
                             seq(from=1, by=6, length=10),
                             seq(from=6, by=6, length=10))
        row.numeric <- as.numeric(row.fwf)
        data[(j * 10 - 9):(j * 10)] <- row.numeric
      }
      ## Remove trailing zeros
      zeros <- which(data == 0)
      if (length(zeros) > 0) {
        nonzeros <- setdiff(zeros[1]:length(data), zeros)
        if (length(nonzeros) > 0) {
          zeros <- zeros[zeros > max(nonzeros)]
          if (length(zeros) > 0) {
            data <- data[-zeros]
          }
        } else {
          data <- data[-zeros]
        }
      }
    }
    data <- data * multipliers[i] / divisors[i]
    n.true <- length(data)
    if (n.true == n.expected) {
      ## write data into matrix
      dendro.matrix[(start.years[i]-r.off):(end.years[i]-r.off), i] <-
        data
    } else if (n.true < n.expected) {
      stop(gettextf("in series %s: ", keycodes[i], domain="R-dplR"),
           gettextf("too few values (expected %d, got %d)",
                    n.expected, n.true, domain="R-dplR"), domain=NA)
    } else if (all(is.na(data[(n.expected+1):n.true]))) {
      dendro.matrix[(start.years[i]-r.off):(end.years[i]-r.off), i] <-
        data[seq_len(n.expected)]
    } else {
      stop(gettextf("in series %s: ", keycodes[i], domain="R-dplR"),
           gettextf("too many values (expected %d, got %d)",
                    n.expected, n.true, domain="R-dplR"), domain=NA)
    }
  }
  cat(sprintf(ngettext(n,
                       "There is %d series\n",
                       "There are %d series\n",
                       domain="R-dplR"),
              n))
  start.years.char <- format(start.years, scientific=FALSE, trim=TRUE)
  end.years.char <- format(end.years, scientific=FALSE, trim=TRUE)
  seq.series.char <- format(seq_len(n), scientific=FALSE, trim=TRUE)
  cat(paste0(format(seq.series.char, width=5), "\t",
             format(keycodes, width=8), "\t",
             format(start.years.char, width=5, justify="right"), "\t",
             format(end.years.char, width=5, justify="right"), "\t",
             format(multipliers/divisors,
                    scientific=FALSE, drop0trailing=TRUE),"\n"), sep="")
  rwl <- as.data.frame(dendro.matrix) # return data.frame
  ## Create data.frame for site, tree, core, radius, stem disk IDs
  all.have.treeID <- !any(is.na(tree.vec))
  na.core <- is.na(core.vec)
  all.have.coreID <- !any(na.core)
  ## Try to find implicit core IDs (tree ID occurs once)
  if (all.have.treeID && !all.have.coreID) {
    foo <- table(tree.vec)
    measured.once <- as.numeric(names(foo)[foo == 1])
    core.vec[na.core & tree.vec %in% measured.once] <- 1
    all.have.coreID <- !any(is.na(core.vec))
  }
  ## Only include "ids" data.frame if all tree and core IDs are known
  if (all.have.treeID && all.have.coreID) {
    unique.sites <- unique(site.code)
    n.unique <- length(unique.sites)
    if (n.unique > 1) {
      site.vec <- match(site.code, unique.sites)
      tree.vec2 <- complex(n, NA_real_, NA_real_)
      total.dupl <- 0
      for (i in seq_len(n.unique)) {
        idx <- which(site.vec == i)
        ut <- unique(tree.vec[idx])
        for (this.tree in ut) {
          idx2 <- idx[tree.vec[idx] == this.tree]
          if (this.tree %in% tree.vec2) {
            tree.vec2[idx2] <- 1i * (total.dupl + 1)
            total.dupl <- total.dupl + 1
          } else {
            tree.vec2[idx2] <- this.tree
          }
        }
      }
      if (total.dupl > 0) {
        dont.change <- Im(tree.vec2) == 0
        existing <- unique(Re(tree.vec2[dont.change]))
        max.existing <- max(existing)
        if (max.existing < 1) {
          free.ids <- 1:total.dupl
        } else {
          free.ids <- which(!(1:max.existing %in% existing))
          free.ids <-
            c(free.ids,
              seq(from=max.existing+1, by=1,
                  length.out=max(0, total.dupl-length(free.ids))))
        }
        tree.vec2[!dont.change] <-
          free.ids[Im(tree.vec2[!dont.change])]
      }
      tree.vec2 <- Re(tree.vec2)
      adf <- data.frame(tree=tree.vec2, core=core.vec, site=site.vec,
                        row.names=keycodes)
    } else {
      adf <- data.frame(tree=tree.vec, core=core.vec, row.names=keycodes)
    }
    if (any(!is.na(radius.vec))) {
      adf <- cbind(adf, radius=radius.vec)
    }
    if (any(!is.na(stemdisk.vec))) {
      adf <- cbind(adf, stemDisk=stemdisk.vec)
    }
    attr(rwl, "ids") <- adf
    cat(gettext('Tree and core IDs were found. See attribute "ids".\n',
                domain="R-dplR"))
  }
  ## Include pith offset data.frame if some pith offsets are known
  na.po <- is.na(pith.offset)
  if (any(!na.po)) {
    attr(rwl, "po") <- data.frame(series=keycodes, pith.offset=pith.offset)
    if (any(na.po)) {
      cat(gettext('Pith offsets were found (some missing values). See attribute "po".\n',
                  domain="R-dplR"))
    } else {
      cat(gettext('Pith offsets were found (no missing values). See attribute "po".\n',
                  domain="R-dplR"))
    }
  }
  class(rwl) <- c("rwl", "data.frame")
  rwl
}

cms <- function(rwl, po, c.hat.t=FALSE, c.hat.i=FALSE) {
    ## support func
    biologicalTrend <- function(tt, theDat.2){
        tt.p1 <- tt + 1
        sqrt.tt <- sqrt(tt)
        sqrt.tt.p1 <- sqrt(tt.p1)
        err4 <- theDat.2 * theDat.2 * ( tt + tt.p1 + 2 * sqrt.tt * sqrt.tt.p1 )
        med <- median(err4) # export for each series?
        err6 <- sqrt(med) * ( sqrt.tt.p1 - sqrt.tt )
        list(indices=err6, c.val=med)
    }
### main func
    if(!is.data.frame(rwl))
        stop("'rwl' must be a data.frame")
    if(!is.data.frame(po))
        stop("'po' must be a data.frame")
    rwl2 <- rwl
    n.col <- ncol(rwl2)
    if(n.col != nrow(po))
        stop("dimension problem: ", "'ncol(rw)' != 'nrow(po)'")
    col.names <- names(rwl2)
    if(!all(sort(po[[1]]) == sort(col.names)))
        stop("series ids in 'po' and 'rwl' do not match")
    rownames(rwl2) <- rownames(rwl2) # guard against NULL names funniness
    n.row <- nrow(rwl2)

    ## divide each series by c curve and restore to cal years
    rwi <- rwl2
    yrs <- time(rwi)
    c.vec <- rep(as.numeric(NA), n.col)
    names(c.vec) <- col.names
    if(c.hat.t){
        c.curve.mat <- matrix(NA, ncol=n.col, nrow=n.row + max(po[[2]]))
        colnames(c.curve.mat) <- col.names
    }
    for(i in seq_len(n.col)){
        the.po <- po[[2]][po[[1]] %in% col.names[i]]
        this.series <- rwl2[[i]]
        series.yrs <- yr.range(this.series, yr.vec=yrs)
        this.series <- sortByIndex(this.series)
        no.na <- which(!is.na(this.series))
        if(length(no.na) > 0){
            series.no.na <- this.series[no.na]
            tmp <- biologicalTrend(no.na+(the.po-1), series.no.na)
            c.vec[i] <- tmp[[2]]
            c.curve <- tmp[[1]]
            if(c.hat.t)
                c.curve.mat[(the.po + 1):(the.po + length(c.curve)), i] <-
                    c.curve
            first <- series.yrs[1]
            last <- series.yrs[2]
            rwi[[i]][yrs %in% first:last] <- series.no.na / c.curve
        }
    }
    ## export options
    if(c.hat.t) {
        if(c.hat.i)
            list(rwi=rwi, c.hat.t=data.frame(c.curve.mat), c.hat.i=c.vec)
        else
            list(rwi=rwi, c.hat.t=data.frame(c.curve.mat))
    } else {
        if(c.hat.i)
            list(rwi=rwi, c.hat.i=c.vec)
        else
            rwi
    }
}

read.rwl <-
  function(fname,
           format=c("auto", "tucson", "compact", "tridas", "heidelberg", "csv"),
           ...)
  {
    switch(match.arg(format),
           auto = {
             cat(gettext("Attempting to automatically detect format.\n",
                         domain="R-dplR"))
             f <- file(fname,"r")
             l1 <- readLines(f, n=1)
             if(length(l1) == 0){
               close(f)
               stop("file is empty")
             }
             ## A rough test for a compact format file
             if(grepl("[1-9][0-9]*\\([1-9][0-9]*F[1-9][0-9]*\\.0\\)~ *$",
                      l1)){
               cat(gettext("Detected a DPL compact format file.\n",
                           domain="R-dplR"))
               close(f)
               read.compact(fname, ...)
             } 
             else if(grepl("^HEADER:$", l1)){ # Heidelberg test
               cat(gettext("Detected a Heidelberg format file.\n",
                           domain="R-dplR"))
               close(f)
               read.fh(fname, ...)
             } 
             else { # two tests for tridas
               ## A rough test for a TRiDaS file
               if(grepl("<tridas>", l1)){
                 cat(gettext("Detected a TRiDaS file.\n",
                             domain="R-dplR"))
                 close(f)
                 read.tridas(fname, ...)
               } 
               else {
                 ## <tridas> may be preceded by an XML
                 ## declaration, comments, etc. Therefore, if
                 ## the first line did not contain <tridas>, we
                 ## read a "reasonable number" of additional
                 ## lines, and try to detect <tridas> in those.
                 more.lines <- readLines(f, n=20)
                 close(f)
                 if(any(grepl("<tridas>", more.lines))){
                   cat(gettext("Detected a TRiDaS file.\n",
                               domain="R-dplR"))
                   read.tridas(fname, ...)
                 } 
                 else if(any(grepl(",", more.lines))){
                   cat(gettext("Detected a csv file.\n",
                               domain="R-dplR"))
                   csv2rwl(fname, ...)
                 }
                 else { # Settle for tucson if none of the above pan out
                   cat(gettext("Assuming a Tucson format file.\n",
                               domain="R-dplR"))
                   read.tucson(fname, ...)
                 }
               }
             }
           },
           tucson = read.tucson(fname, ...),
           compact = read.compact(fname, ...),
           tridas = read.tridas(fname, ...),
           heidelberg = read.fh(fname, ...),
           csv = csv2rwl(fname, ...))
  }

corr.series.seg <- function(rwl, series, series.yrs=as.numeric(names(series)),
                            seg.length=50, bin.floor=100, n=NULL,
                            prewhiten = TRUE, biweight=TRUE,
                            method = c("spearman", "pearson", "kendall"),
                            pcrit=0.05, make.plot = TRUE,
                            floor.plus1 = FALSE, ...) {

    method2 <- match.arg(method)

    ## Handle different types of 'series'
    tmp <- pick.rwl.series(rwl, series, series.yrs)
    rwl2 <- tmp[[1]]
    series2 <- tmp[[2]]

    ## run error checks
    qa.xdate(rwl2, seg.length, n, bin.floor)

    ## turn off warnings for this function
    ## The sig test for spearman's rho often produces warnings.
    w <- options(warn = -1)
    on.exit(options(w))

    seg.lag <- seg.length / 2

    ## Normalize.
    tmp <- normalize.xdate(rwl2, series2, n, prewhiten, biweight)
    master <- tmp$master

    ## trim master so there are no NaN like dividing when
    ## only one series for instance.
    idx.good <- !is.nan(master)
    master <- master[idx.good]
    yrs <- as.numeric(names(master))

    series2 <- tmp$series
    series.yrs2 <- as.numeric(names(series2))
    ## trim series in case it was submitted stright from the rwl
    idx.good <- !is.na(series2)
    series.yrs2 <- series.yrs2[idx.good]
    series2 <- series2[idx.good]

    ## clip series to master dimensions
    series2 <- series2[series.yrs2 %in% yrs]
    series.yrs2 <- as.numeric(names(series2))
    ## clip master to series dimensions
    master <- master[yrs %in% series.yrs2]
    yrs <- as.numeric(names(master))
    nyrs <- length(series.yrs2)

    if (nyrs < seg.length) {
        stop("number of overlapping years is less than 'seg.length'")
    }
    if (is.null(bin.floor) || bin.floor == 0) {
        min.bin <- min(series.yrs2)
    } else if (floor.plus1) {
        min.bin <- ceiling((min(series.yrs2) - 1) / bin.floor) * bin.floor + 1
    } else {
        min.bin <- ceiling(min(series.yrs2) / bin.floor) * bin.floor
    }
    to <- max(series.yrs2) - seg.length - seg.lag + 1
    if (min.bin > to) {
        cat(gettextf("maximum year in (filtered) series: %d\n",
                     max(series.yrs2)))
        cat(gettextf("first bin begins: %d\n", min.bin))
        cat(gettext("cannot fit two segments (not enough years in the series)\n"))
        stop("shorten 'seg.length' or adjust 'bin.floor'")
    }
    bins <- seq(from=min.bin, to=to + seg.lag, by=seg.lag)
    bins <- cbind(bins, bins + (seg.length - 1), deparse.level=0)
    nbins <- nrow(bins)
    bin.names <- paste0(bins[, 1], ".", bins[, 2])
    ## structures for results
    res.cor <- rep(NA, nbins)
    names(res.cor) <- bin.names

    res.pval <- rep(NA, nbins)
    names(res.pval) <- bin.names

    overall.cor <- rep(NA, 2)
    names(overall.cor) <- c("rho", "p-val")

    segavg.cor <- rep(NA, nbins)
    names(segavg.cor) <- bin.names

    ## moving correlation
    res.mcor <- matrix(NA, nyrs, 2)
    colnames(res.mcor) <- c("rho", "p.val")
    rownames(res.mcor) <- series.yrs2

    ## loop through bins
    for (j in seq_len(nbins)) {
        mask <- yrs %in% seq(from=bins[j, 1], to=bins[j, 2])
        ## cor is NA if there is not complete overlap
        if (!any(mask) ||
            any(is.na(series2[mask])) ||
            any(is.na(master[mask]))) {
            bin.cor <- NA
            bin.pval <- NA
        } else {
            tmp <- cor.test(series2[mask], master[mask], method = method2,
                            alternative = "greater")
            bin.cor <- tmp$estimate
            bin.pval <- tmp$p.val
        }
        res.cor[j] <- bin.cor
        res.pval[j] <- bin.pval
    }
    ## overall correlation
    tmp <- cor.test(series2, master, method = method2,
                    alternative = "greater")
    overall.cor[1] <- tmp$estimate
    overall.cor[2] <- tmp$p.val

    ## moving correlation
    for (i in seq_len(nyrs - seg.length + 1)) {
        mask <- i:(i + seg.length - 1)
        tmp <- cor.test(series2[mask], master[mask],
                        method = method2, alternative = "greater")
        res.mcor[i + seg.lag, 1] <- tmp$estimate
        res.mcor[i + seg.lag, 2] <- tmp$p.val
    }
    ## plot
    if (make.plot) {
        mcor.tmp <- na.omit(res.mcor)
        yrs.tmp <- as.numeric(rownames(mcor.tmp))
        mcor.tmp <- mcor.tmp[, 1]
        n.below <- ceiling(max(0, min.bin - min(yrs.tmp)) / seg.lag)
        start.below <- seq(from=min.bin - n.below * seg.lag, by=seg.lag,
                           length.out=n.below)
        ticks <- c(start.below, bins[, 1], bins[c(nbins - 1, nbins), 2] + 1)
        nticks <- length(ticks)

        par(mar=c(4, 2, 2, 1) + 0.1, mgp=c(1.25, 0.25, 0), tcl=0.25)
        sig <- qnorm(1 - pcrit / 2) / sqrt(seg.length)
        plot(yrs.tmp, mcor.tmp, type="l",
             xlim=c(ticks[1], ticks[nticks]),
             ylim=range(mcor.tmp, sig, na.rm=TRUE),
             ylab=gettext("Correlation", domain="R-dplR"),
             xlab=gettext("Year", domain="R-dplR"),
             sub=gettextf("Segments: length=%d,lag=%d", seg.length, seg.lag,
             domain="R-dplR"),
             axes=FALSE, ...)
        abline(v=ticks, col="grey", lty="dotted")
        axis(1, at=ticks[seq(from=1, to=nticks, by=2)])
        axis(3, at=ticks[seq(from=2, to=nticks, by=2)])
        axis(2)
        box()
        ## lines bins
        for (i in seq_len(nbins)) {
            xx <- c(bins[i, ], recursive=TRUE)
            yy <- c(res.cor[i], res.cor[i])
            lines(xx, yy, lwd=2)
        }
        lines(yrs.tmp, mcor.tmp, lwd=1.5)
        abline(h=sig, lty="dashed")
    }
    res <- list(res.cor, res.pval, overall.cor, bins, res.mcor)
    names(res) <- c("spearman.rho", "p.val", "overall", "bins", "moving.rho")
    res
}

xskel.plot <- function(rwl,series,series.yrs = as.numeric(names(series)),
         win.start, win.end=win.start+100, n = NULL, prewhiten = TRUE,
         biweight = TRUE) {

  ## Handle different types of 'series'
  tmp <- pick.rwl.series(rwl, series, series.yrs)
  rwl2 <- tmp[[1]]
  series2 <- tmp[[2]]

  master.yrs <- as.numeric(rownames(rwl2))
  series.yrs2 <- as.numeric(names(series2))
  yrs <- seq(from=win.start,to=win.end)
  nyrs <- length(yrs)

  if(nyrs > 101){
    warning("These plots get crowded with windows longer than 100 years.")
  }
  ## check window overlap with master and series yrs
  if (!all(yrs %in% series.yrs2)) {
      cat(gettextf("Window Years: %d-%d", min(yrs), max(yrs),
                   domain = "R-dplR"),
          " & ",
          gettextf("Series Years: %d-%d", min(series.yrs2), max(series.yrs2),
                   domain = "R-dplR"),
          "\n", sep="")
      stop("Fix window overlap")
  }
  if (!all(yrs %in% master.yrs)) {
      cat(gettextf("Window Years: %d-%d", min(yrs), max(yrs),
                   domain = "R-dplR"),
          " & ",
          gettextf("Master Years: %d-%d", min(master.yrs), max(master.yrs),
                   domain = "R-dplR"),
          "\n", sep="")
      stop("Fix window overlap")
  }

  ## normalize.
  names(series2) <- series.yrs2
  tmp <- normalize.xdate(rwl2, series2, n, prewhiten, biweight)

  ## master
  master <- tmp$master
  master.yrs <- as.numeric(names(master))
  master <- master[master.yrs%in%yrs]
  master.yrs <- as.numeric(names(master))
  ## series
  series2 <- tmp$series
  series.yrs2 <- as.numeric(names(series2))
  series2 <- series2[series.yrs2%in%yrs]
  series.yrs2 <- as.numeric(names(series2))


  ## skeleton
  master.skel <- cbind(master.yrs,xskel.calc(master))
  master.skel <- master.skel[master.skel[,1]%in%yrs,]
  master.yrs.sig <- master.skel[!is.na(master.skel[,2]),1]
  series.skel <- cbind(series.yrs2,xskel.calc(series2))
  series.skel <- series.skel[series.skel[,1]%in%yrs,]
  series.yrs.sig <- series.skel[!is.na(series.skel[,2]),1]

  ## cor and skel agreement
  overall.r <- round(cor(series2,master),3)
  overall.agree <- sum(series.yrs.sig%in%master.yrs.sig)/length(master.yrs.sig)
  overall.agree <- round(overall.agree*100,1)

  ## build page for plotting
  grid.newpage()
  fontsize <- 12
  textJust <- "center"
  col1light <- "lightgreen"
  col1dark <- "darkgreen"
  ## 1.0 a bounding box for margins
  bnd.vp <- plotViewport(margins=rep(0.5,4), # 1/2 line margin
                         name = "bnd.vp",
                         gp = gpar(fontsize = fontsize))
  ## go from bottom up.

  ## 4.1 bounding box for skeleton plot. 55% of area
  skel.bnd.vp <- viewport(x = 0, y = 0, width = 1, height = 0.95,
                          just = c("left", "bottom"), name = "skel.bnd.vp")
  ## 4.2 plotting region for skeleton plot. 2 lines left and right.
  ## 3 lines on top and bottom
  skel.region.vp <- plotViewport(margins=c(3,2,3,2),
                                 xscale=c(min(yrs)-0.5,max(yrs)+0.5),
                                 yscale=c(-10,10),
                                 name = "skel.region.vp")
  ## 5.0 a box for overall text. 5%
  overall.txt.vp <- viewport(x = 0, y = 0.95, width = 1, height = 0.05,
                             just = c("left", "bottom"),
                             name = "overall.txt.vp")

  ## actual plotting
  dev.hold()
  on.exit(dev.flush())
  pushViewport(bnd.vp) # inside margins
  pushViewport(skel.bnd.vp) # inside skel
  pushViewport(skel.region.vp) # inside margins
  grid.rect(gp = gpar(col=col1light, lwd=1))
  grid.grill(h = unit(seq(-10, 10, by=1), "native"),
             v = unit(yrs-0.5, "native"),
             gp = gpar(col=col1light, lineend = "square",
                       linejoin = "round"))
  ## rw plot
  grid.rect(x = yrs, y = 0, width = 1, height = 2 * master,
            hjust = 0.5, vjust = 1, default.units = "native",
            gp=gpar(fill=col1light,col=col1dark))
  grid.rect(x = yrs, y = 0, width = 1, height = 2 * series2,
            hjust = 0.5, vjust = 0, default.units = "native",
            gp=gpar(fill=col1light,col=col1dark))

  ## master
  grid.segments(x0=master.yrs.sig,y0=0,
                x1=master.yrs.sig,y1=-10,
                default.units="native",
                gp=gpar(lwd=1,col='black',lineend="butt"))
  grid.segments(x0=master.skel[,1],y0=0,
                x1=master.skel[,1],y1=master.skel[,2]*-1,
                default.units="native",
                gp=gpar(lwd=5,col='black',lineend="butt"))
  ## series
  grid.segments(x0=series.yrs.sig,y0=0,
                x1=series.yrs.sig,y1=10,
                default.units="native",
                gp=gpar(lwd=1,col='black',lineend="butt"))
  grid.segments(x0=series.skel[,1],y0=0,
                x1=series.skel[,1],y1=series.skel[,2],
                default.units="native",
                gp=gpar(lwd=5,col='black',lineend="butt"))

  ## text
  grid.text(master.yrs.sig, x=unit(master.yrs.sig,"native"),
            y = unit(0, "npc"), rot = 90,just="right")
  grid.text(series.yrs.sig, x=unit(series.yrs.sig,"native"),
            y = unit(1, "npc"), rot = 90,just="left")
  grid.text(gettext("Master", domain="R-dplR"),x=unit(0,"npc"),
            y=unit(0,"npc"),hjust = 0,vjust = 0,rot=90)
  grid.text(gettext("Series", domain="R-dplR"),x=unit(0,"npc"),
            y=unit(1,"npc"),hjust=1,vjust=0,rot=90)

  popViewport(2) # back to bnd
  pushViewport(overall.txt.vp) # description
  periodPattern <- gettext("Period: %d-%d", domain = "R-dplR")
  agreePattern <- gettext("Skeleton Agreement %s%%", domain = "R-dplR")
  tmp.txt <- substitute(period * ", " * r[lag0] == corr * ", " * agree,
                        list(period = sprintf(periodPattern,
                             min(yrs), max(yrs)),
                             corr = overall.r,
                             agree = sprintf(agreePattern, overall.agree)))
  grid.text(tmp.txt,y=unit(0.5,"npc"),x=unit(0.5,"npc"),
            just = textJust)
  popViewport(2)

}

`sgc` <- 
  function(x, overlap = 50, prob = TRUE) 
{
  # checks class rwl and correct overlap
  if(!("rwl" %in% class(x))) {
    warning("'x' is not class rwl")
  }
  if(any(length(overlap)!=1 | !is.numeric(overlap) | 
         overlap%%1!=0 | overlap < 3)){
    stop("'overlap' should be a single integer >=3")
  }
  if(overlap < 50) {
    warning("The minimum number of overlap is lower than 50. This might lead to statistically insignificant matches.")
  }
  if (!is.logical(prob)) {
    stop("'prob' must be either TRUE (the default) or FALSE")
  }
  # function starts here
  n <- dim(x)[2]
  sgc_mat <- matrix(NA_real_, nrow = n, ncol = n)
  ssgc_mat <- matrix(NA_real_, nrow = n, ncol = n)
  overlap_n <- matrix(NA_real_, nrow = n, ncol = n)
  rownames(sgc_mat) <- colnames(sgc_mat) <- names(x)
  rownames(ssgc_mat) <- colnames(ssgc_mat) <- names(x)
  rownames(overlap_n) <- colnames(overlap_n) <- names(x)
  treering_sign <- apply(x, 2, diff)
  treering_sign <- sign(treering_sign)
  for (i in 1:n) {
    treering_GC <- abs(treering_sign[,i]-treering_sign)
    # determine overlap (the number of overlapping growth changes)
    treering_GCol <- colSums(!is.na(treering_GC))
    treering_GCol[treering_GCol==0] <- NA
    treering_GCol[treering_GCol<overlap] <- NA
    # semi synchronous growth changes
    treering_GC1 <- colSums(treering_GC==1,na.rm=TRUE)
    # synchronous growth changes
    treering_GC0 <- colSums(treering_GC==0,na.rm=TRUE)
    SGC_values <- treering_GC0 / treering_GCol
    SSGC_values <- treering_GC1 / treering_GCol
    sgc_mat[i,] <- SGC_values
    ssgc_mat[i,] <- SSGC_values
    overlap_n[i,] <- treering_GCol
  }
  if (prob == TRUE) {
    s_mat <- 1 /(2*sqrt(overlap_n))
    z_mat <- (sgc_mat-0.5)/s_mat
    z_normcdf <- apply(z_mat, 2, function(z) pnorm(z, mean=0, sd=1))
    p_mat <- 2*(1-z_normcdf) 
    sgc_list <- list(sgc_mat = sgc_mat,
                     ssgc_mat = ssgc_mat,
                     overlap = overlap_n,
                     p_mat = p_mat)
  } else {
    sgc_list <- list(sgc = sgc_mat,
                     ssgc = ssgc_mat,
                     overlap = overlap_n)
  }
  sgc_list
}

`seg.plot` <-
    function(rwl, ...)
{
    if (!is.data.frame(rwl)) {
        stop("'rwl' must be a data.frame")
    }
    yr <- as.numeric(row.names(rwl))
    first.year <- as.matrix(apply(rwl, 2, yr.range, yr.vec=yr))[1, ]
    last.year <- as.matrix(apply(rwl, 2, yr.range, yr.vec=yr))[2, ]
    neworder <- order(first.year, decreasing=FALSE)
    segs <- rwl[, neworder, drop=FALSE]
    n.col <- ncol(segs)
    seq.col <- seq_len(n.col)
    for (i in seq.col) {
        segs[[i]][!is.na(segs[[i]])] <- i
    }
    segs.axis2 <- names(segs)
    segs.axis4 <- names(segs)
    segs.axis2[seq(2,n.col,by=2)] <- NA
    segs.axis4[seq(1,n.col,by=2)] <- NA
    op <- par(no.readonly=TRUE) # Save par
    on.exit(par(op))            # Reset par on exit
    par(mar=c(2, 5, 2, 5) + 0.1, mgp=c(1.1, 0.1, 0), tcl=0.5,
        xaxs="i",yaxs="i")
    plot(yr, segs[[1]], type="n", ylim=c(0, n.col+1), axes=FALSE,
         ylab="", xlab=gettext("Year", domain="R-dplR"), ...)
    abline(h=seq.col,lwd=1,col="grey")
    grid(ny = NA)
    apply(segs, 2, lines, x=yr, lwd=4,lend=2)
    axis(2, at=seq.col, labels=segs.axis2, srt=45, tick=FALSE, las=2)
    axis(4, at=seq.col, labels=segs.axis4, srt=45, tick=FALSE, las=2)
    axis(1)
    axis(3)
    box()
}

pass.filt <- function(y, W, type = c("low", "high", "stop", "pass"),
                      method = c("Butterworth", "ChebyshevI"),
                      n = 4, Rp = 1) {
  if (any(is.na(y))) stop("y contains NA")
  
  ## check W's length
  type2 <- match.arg(type)
  nW <- length(W)
  if (type2 == "low"  && nW != 1) stop("length(W) > 1")
  if (type2 == "high" && nW != 1) stop("length(W) > 1")
  if (type2 == "stop" && nW != 2) stop("length(W) != 2")
  if (type2 == "pass" && nW != 2) stop("length(W) != 2")
  
  
  ## if W is in period (>1) then convert to f
  if (any(W>1)) {
    f <- 1/W
    p <- W
  } else {
    p <- 1/W
    f <- W
  }

  ## sort f in case it's passed in backwards
  f <- sort(f)

  method2 <- match.arg(method)

  if (method2 == "ChebyshevI"){
    filt <- signal::cheby1(n = n, W = f*2, type = type2, Rp = Rp, plane = "z")
  }
  else {
    filt <- signal::butter(n = n, W = f*2, type = type2, plane = "z")
  }

  ## remove mean
  yAvg <- mean(y)
  y <- y - yAvg

  ## pad the data to twice the max period
  pad <- max(p) * 2
  ny <- length(y)
  ## pad the data
  yPad <- c(y[pad:1], y, y[ny:(ny-pad)])
  ## run the filter
  yFilt <- signal::filtfilt(filt, yPad)
  ## unpad the filtered data
  yFilt <- yFilt[(pad+1):(ny+pad)]
  ## return with mean added back in
  yFilt + yAvg
}

`ssf` <- function(rwl, 
                  method="Spline", 
                  nyrs = NULL,
                  difference = FALSE,
                  max.iterations = 25, 
                  mad.threshold = 5e-4,
                  recode.zeros = FALSE,
                  return.info = FALSE, 
                  verbose = TRUE)
{
  
  if(max.iterations > 25){
    warning("Having to set max.iterations > 25 in order to meet a stopping criteria  is generally a sign that the data are not ideal for signal free detrending.")
  }
  
  if(mad.threshold < 0.0001 | mad.threshold > 0.001){
    warning("The stopping criteria, mad.threshold,  is outside the recommended range of 0.0001 to 0.001.")
  }
  
  # error msgs for later
  negCurveMsg <- gettext("[1] The signal free detrending curve has values <= 0. See help (?ssf).",
                         domain = "R-dplR")
  
  maxIterMsg <- gettext("[2] Reached maximum iterations and stopping criteria are not satisfied. See help (?ssf).",
                        domain = "R-dplR")

  crn0Msg <- gettext("[3] The initial chronology contains at least one row (year) with a zero, creating div0 problems. See help (?ssf).",
                 domain = "R-dplR")
  
  input0Msg <- gettext("[4] Input data contain at least one row (year) with all zero values, creating div0 problems. See help (?ssf).",
                       domain = "R-dplR")
  
  zeroColMsg <- gettext("[5] Input data contain at least one series with all zero values. See help (?ssf).",
                        domain = "R-dplR")
  inputNAmsg <- gettext("[6] Input data contain at least one row (year) with all NA values, creating div0 problems. See help (?ssf).",
                       domain = "R-dplR")
  
  # make a copy of rwl just in case we change it.
  dat <- rwl

  # check class of rwl
  if(!any(class(dat) %in% "rwl")) {
    warning("Input data needs to be class \"rwl\". Attempting to coerce.")
    dat <- as.rwl(dat)
  }
  
  # recode zeros to 0.001 if asked.
  if(recode.zeros){dat[dat==0] <- 0.001}
  
  # error checks
  
  # Look for any rows where all the values are NA -- unconnected floaters
  if(any(rowSums(is.na(dat)) == ncol(dat))){
    stop(inputNAmsg)
  }
  # Can't have all zeros across the board for a year. This is
  # a conservative check but if there are zeros for a year, the chron can eval to zero
  # which causes headaches with div0.
  zeroRowCheck <- apply(dat,1,function(x){sum(x,na.rm=TRUE)==0})
  if(any(zeroRowCheck)){
    stop(input0Msg)
  }
  # Heck look for zeros in series too. Never know what kind of silliness users come up with.
  zeroColCheck <- apply(dat,2,function(x){sum(x,na.rm=TRUE)==0})
  if(any(zeroColCheck)){
    stop(zeroColMsg)
  }
  
  
  # get some detrending options
  method2 <- match.arg(arg = method,
                       choices = c("Spline", "AgeDepSpline"),
                       several.ok = FALSE)
  
  # useful vars
  yrs <- time(dat)
  seriesNames <- names(dat)
  
  nSeries <- dim(dat)[2]
  nYrs <- dim(dat)[1]
  medianAbsDiff <- 1 #init only
  datSummary <- rwl.stats(dat)
  medianSegLength <- median(datSummary$year)
  
  if(method2 == "AgeDepSpline"){
    infoList <- list(method=method2, 
                     nyrs = nyrs, 
                     pos.slope = TRUE,
                     max.iterations = max.iterations, 
                     mad.threshold = mad.threshold)
  }
  else {
    infoList <- list(method=method2, 
                     nyrs = nyrs, 
                     max.iterations = max.iterations, 
                     mad.threshold = mad.threshold)
  }
  
  # Make some storage objects
  # These are arrays of [nYrs,nSeries,max.iterations]
  # Array to hold the SF measurements
  sfRW_Array <- array(NA,dim=c(nYrs,nSeries,max.iterations))
  # Array to hold the rescaled SF measurements
  sfRWRescaled_Array <- array(NA,dim=c(nYrs,nSeries,max.iterations))
  # Array to hold the rescaled SF curves
  sfRWRescaledCurves_Array <- array(NA,dim=c(nYrs,nSeries,max.iterations))
  # Array to hold the SF RWI
  sfRWI_Array <- array(NA,dim=c(nYrs,nSeries,max.iterations))
  # Array (2d though) to hold the SF Crn
  sfCrn_Mat <- array(NA,dim=c(nYrs,max.iterations))
  # Array (2d though) to hold the HF Crn
  hfCrn_Mat <- array(NA,dim=c(nYrs,max.iterations))
  # Vector for storing median absolute difference (mad)
  MAD_Vec <- numeric(max.iterations-1)
  # Array (2d though) to hold the differences between the kth
  # and the kth-1 high freq chronology residuals
  hfCrnResids_Mat <- matrix(NA,nrow = nYrs,ncol=max.iterations-1)
  
  # Let's do it. First, here is a simplish detrending function modified from
  # detrend.series(). The issue with using detrend() is that negative values are
  # not allowed for the detrend funcs. Maybe they should be (e.g., z-scored 
  # data) but they aren't as of right now. So here is a simplified detrend function.
  getCurve <- function(y,method=method2,
                       nyrs=NULL,
                       pos.slope=TRUE){
    
    
    ## Remove NA from the data (they will be reinserted later)
    good.y <- which(!is.na(y))
    if(length(good.y) == 0) {
      stop("all values are 'NA'")
    } else if(any(diff(good.y) != 1)) {
      stop("'NA's are not allowed in the middle of the series")
    }
    y2 <- y[good.y]
    nY2 <- length(y2)
    ## Recode any zero values to 0.001 to avoid div0
    y2[y2 == 0] <- 0.001
    
    
    # Age Dep Spl  
    if("AgeDepSpline" %in% method2){
      ## Age dep smoothing spline with nyrs (50 default) as the init stiffness
      ## are NULL
      if(is.null(nyrs))
        nyrs2 <- 50
      else
        nyrs2 <- nyrs
      
      Curve <- ads(y=y2, nyrs0=nyrs2, pos.slope = TRUE)
      # Put NA back in
      Curve2 <- rep(NA, length(y))
      Curve2[good.y] <- Curve
      
    }  
    
    # CAPS  
    if("Spline" %in% method2){
      if(is.null(nyrs))
        nyrs2 <- length(y2) * 0.6667
      else if(nyrs > 0 & nyrs < 1) {
        nyrs2 <- length(y2) * nyrs
      }
      else
        nyrs2 <- nyrs
      
      Curve <- caps(y=y2, nyrs=nyrs2)
      # Put NA back in
      Curve2 <- rep(NA, length(y))
      Curve2[good.y] <- Curve
    }
    
    
    return(Curve2)
  }
  
  # STEP 1 - GET AN INITIAL CHRONOLOGY
  # fit curves
  datCurves <- apply(dat,2,getCurve,
                     method=method2,
                     nyrs=nyrs,
                     pos.slope=TRUE)
  rownames(datCurves) <- time(dat)
  
  if(any(datCurves <= 0,na.rm = TRUE)){
    stop(negCurveMsg)
  }
  
  
  # get RWI
  if(difference){ datRWI <- dat - datCurves }
  else { datRWI <- dat / datCurves }
  # and initial chron at iter0
  iter0Crn <- chron(datRWI,biweight = TRUE)
  # Check for zeros in the chronology. This can happen in VERY sensitive
  # chrons with years that mostly zeros if the chron is built with tukey's
  # biweight robust mean (e.g., co021). This causes problems with div0 later on
  # so if there are any zeros in the chron, switch straight mean which should
  # head off any zeros in the chron unless the data themseleves are bunk.
  # e.g., UT024.
  if(any(iter0Crn[,1]==0)){
    iter0Crn <- chron(datRWI,biweight = FALSE)
  }
  # Additional check. If there are still zeros it should mean that the OG data were passed in with zeros.
  if(any(iter0Crn[,1]==0)){
    stop(crn0Msg)
  }
  
  datSampDepth <- iter0Crn$samp.depth # for later
  normalizedSampleDepth <- sqrt(datSampDepth-1)/sqrt(max(datSampDepth-1)) # for later
  iter0Crn <- iter0Crn[,1] # just keep the chron
  
  # STEP 2 - Divide each series of measurements by the chronology
  # NB: This can produce some very very funky values when iter0Crn is near zero.
  # E.g., in co021 row 615 has a tbrm RWI of 0.0044 which makes for some huge SF
  if(difference){ sfRW_Array[,,1] <- as.matrix(dat - iter0Crn) }
  else { sfRW_Array[,,1] <- as.matrix(dat/iter0Crn) }
  
  # STEP 3 - Rescale to the original mean
  colMeansMatdatSF <- matrix(colMeans(sfRW_Array[,,1],na.rm = TRUE),
                             nrow = nrow(sfRW_Array[,,1]),
                             ncol = ncol(sfRW_Array[,,1]),
                             byrow = TRUE)
  colMeansMatdat <- matrix(colMeans(dat,na.rm = TRUE),
                           nrow = nrow(dat),
                           ncol = ncol(dat),
                           byrow = TRUE)
  sfRWRescaled_Array[,,1] <- (sfRW_Array[,,1] - colMeansMatdatSF) + colMeansMatdat
  
  # STEP 4 - Replace signal-free measurements with original measurements when samp depth is 1
  sfRWRescaled_Array[datSampDepth==1,,1] <- as.matrix(dat[datSampDepth==1,])
  
  # STEP 5 - Fit curves to signal free measurements
  sfRWRescaledCurves_Array[,,1] <- apply(sfRWRescaled_Array[,,1],2,getCurve,
                                         method=method2,
                                         nyrs=nyrs,
                                         pos.slope=TRUE)
  
  if(any(sfRWRescaledCurves_Array[,,1] <= 0,na.rm = TRUE)){
    stop(negCurveMsg)
  }
  
  # STEP 6 - divide original measurements by curve obtained from signal free measurements fitting
  if(difference){ sfRWI_Array[,,1] <- as.matrix(dat - sfRWRescaledCurves_Array[,,1]) }
  else { sfRWI_Array[,,1] <- as.matrix(dat/sfRWRescaledCurves_Array[,,1]) }
  
  # STEP 7 - create 1st signal-free chronology
  sfCrn_Mat[,1] <- chron(sfRWI_Array[,,1],biweight = TRUE)[,1]
  # Check for zeros in the chronology. This can happen in VERY sensitive
  # chrons with years that mostly zeros if the chron is built with tukey's
  # biweight robust mean (e.g., co021). This causes problems with div0 later on
  # so if there are any zeros in the chron, switch straight mean which should
  # head off any zeros in the chron unless the data themseleves are bunk
  if(any(sfCrn_Mat[,1]==0)){
    sfCrn_Mat[,1] <- chron(sfRWI_Array[,,],biweight = FALSE)[,1]
  }
  
  # And calc the high freq crn that will be used to determine MAD stopping crit
  hfCrn_Mat[,1] <- sfCrn_Mat[,1] - caps(sfCrn_Mat[,1], #empty?
                                        nyrs = floor(medianSegLength))
  
  
  # STEP 8 - Repeat (2) through (7) until the MAD threshold
  # is reached or we hit maxIter
  if(verbose){
    cat("Data read. Running ssf with \n") 
    print(unlist(infoList))
    cat("\nFirst iteration done.\n")
  }
  
  iterationNumber <- 2 # Start on 2 b/c we did one above
  
  while(medianAbsDiff > mad.threshold){
    k = iterationNumber
    
    # STEP 2 - Divide each series of measurements by the last SF chronology
    
    if(difference){ sfRW_Array[,,k] = as.matrix(dat - sfCrn_Mat[,k-1]) }
    else { sfRW_Array[,,k] = as.matrix(dat/sfCrn_Mat[,k-1]) }
    
    
    # STEP 3 - Rescale to the original mean
    colMeansMatdatSF <- matrix(colMeans(sfRW_Array[,,k],na.rm = TRUE),
                               nrow = nrow(sfRW_Array[,,k]),
                               ncol = ncol(sfRW_Array[,,k]),
                               byrow = TRUE)
    tmp <- (sfRW_Array[,,k] - colMeansMatdatSF) + colMeansMatdat
    
    # can get a nan if unlucky. set to? zero?
    tmp[is.nan(tmp)] <- 0
    sfRWRescaled_Array[,,k] <- tmp
    # STEP 4 - Replace signal-free measurements with original measurements when sample depth is one
    sfRWRescaled_Array[datSampDepth==1,,k] <- as.matrix(dat[datSampDepth==1,])
    
    # STEP 5 - fit curves to signal free measurements
    sfRWRescaledCurves_Array[,,k] <- apply(sfRWRescaled_Array[,,k],2,getCurve,
                                           method=method2,
                                           nyrs=nyrs,
                                           pos.slope=TRUE)
    
    if(any(sfRWRescaledCurves_Array[,,k] <= 0,na.rm = TRUE)){
      stop(negCurveMsg)
    }
    
    # STEP 6 - divide original measurements by curve obtained from signal free curves
    if(difference){ sfRWI_Array[,,k] <- as.matrix(dat - sfRWRescaledCurves_Array[,,k]) }
    else { sfRWI_Array[,,k] <- as.matrix(dat/sfRWRescaledCurves_Array[,,k]) }
    
    # STEP 7 - create kth signal-free chronology
    sfCrn_Mat[,k] <- chron(sfRWI_Array[,,k],biweight = TRUE)[,1]
    # Check for zeros in the chronology. This can happen in VERY sensitive
    # chrons with years that mostly zeros if the chron is built with tukey's
    # biweight robust mean (e.g., co021). This causes problems with div0 later on
    # so if there are any zeros in the chron, switch straight mean which should
    # head off any zeros in the chron unless the data themseleves are bunk
    if(any(sfCrn_Mat[,k]==0)){
      sfCrn_Mat[,k] <- chron(sfRWI_Array[,,k],biweight = FALSE)[,1]
    }
    
    # Now look at diffs in fit using median abs diff in the high freq resids
    # This is the (high freq) resids from the current iter minus the resids from prior iter
    hfCrn_Mat[,k] <- sfCrn_Mat[,k] - caps(sfCrn_Mat[,k], #empty?
                                          nyrs = floor(medianSegLength))
    
    hfCrnResids_Mat[,k-1] <- hfCrn_Mat[,k] - hfCrn_Mat[,k-1]
    # calculate the median absolute differences weighted by the normalized sample depth
    medianAbsDiff <- median(abs(hfCrn_Mat[,k]*normalizedSampleDepth - hfCrn_Mat[,k-1]*normalizedSampleDepth))
    
    MAD_Vec[k-1] <- medianAbsDiff
    if(verbose){
      cat("Iteration: ", k, " Median Abs Diff: ",round(medianAbsDiff,5),
          " (",round(mad.threshold/medianAbsDiff * 100,5),"% of threshold)\n",
          sep = "")
    }
    
    if(iterationNumber==max.iterations & medianAbsDiff > mad.threshold){
      stop(maxIterMsg)
    }
    iterationNumber <- iterationNumber + 1
  }
  # Remove empty NAs from output that aren't needed anymore.
  # Trim the SF measurements
  sfRW_Array <- sfRW_Array[,,1:k]
  # Trim the rescaled SF measurements
  sfRWRescaled_Array <- sfRWRescaled_Array[,,1:k]
  # Trim the rescaled SF curves
  sfRWRescaledCurves_Array <- sfRWRescaledCurves_Array[,,1:k]
  # Trim the SF RWI
  sfRWI_Array <- sfRWI_Array[,,1:k]
  # Trim the SF crn
  sfCrn_Mat <- sfCrn_Mat[,1:k]
  # Trim the HF crn
  hfCrn_Mat <- hfCrn_Mat[,1:k]
  # Trim the hfCrnResids
  hfCrnResids_Mat <- hfCrnResids_Mat[,1:(k-1)]
  
  # Trim the differences
  MAD_Vec <- MAD_Vec[1:(k-1)]
  
  ### return final crn in class(crn)
  finalCrn <- data.frame(sfc=sfCrn_Mat[,k],samp.depth=datSampDepth)
  row.names(finalCrn) <- row.names(dat)
  class(finalCrn) <- c("crn", "data.frame")
  
  if(verbose){ 
    cat("Simple Signal Free Chronology Complete \n")
  }
  
  
  
  if(return.info){
    # add the original data to the arrays and mats
    # the original data
    #dat 
    #nIter <- dim(sfRW_Array)[3]
    arrayDims <- dim(sfRW_Array)
    arrayDims[3] <- arrayDims[3] + 1
    
    # get final outputs setup
    #sfRW_Array
    sfRW_Array2 <- array(data = NA,dim=arrayDims)
    sfRW_Array2[,,1] <- as.matrix(dat)
    sfRW_Array2[,,2:arrayDims[3]] <- sfRW_Array
    dimnames(sfRW_Array2) <- list(yrs,seriesNames,0:k)
    
    #sfRWRescaled_Array
    sfRWRescaled_Array2 <- array(data = NA,dim=arrayDims)
    sfRWRescaled_Array2[,,1] <- as.matrix(dat)
    sfRWRescaled_Array2[,,2:arrayDims[3]] <- sfRWRescaled_Array
    dimnames(sfRWRescaled_Array2) <- list(yrs,seriesNames,0:k)
    
    #sfRWRescaledCurves_Array
    sfRWRescaledCurves_Array2 <- array(data = NA,dim=arrayDims)
    sfRWRescaledCurves_Array2[,,1] <- as.matrix(datCurves)
    sfRWRescaledCurves_Array2[,,2:arrayDims[3]] <- sfRWRescaledCurves_Array
    dimnames(sfRWRescaledCurves_Array2) <- list(yrs,seriesNames,0:k)
    
    #sfRWI_Array
    sfRWI_Array2 <- array(data = NA,dim=arrayDims)
    sfRWI_Array2[,,1] <- as.matrix(datRWI)
    sfRWI_Array2[,,2:arrayDims[3]] <- sfRWI_Array
    dimnames(sfRWI_Array2) <- list(yrs,seriesNames,0:k)
    
    sfCrn_Mat2 <- cbind(iter0Crn,sfCrn_Mat)
    rownames(sfCrn_Mat2) <- yrs
    colnames(sfCrn_Mat2) <- 0:k
    
    #hfCrn_Mat
    #iter0
    tmp <- iter0Crn - caps(iter0Crn,
                           nyrs = floor(medianSegLength))
    hfCrn_Mat2 <- cbind(tmp,hfCrn_Mat)
    rownames(hfCrn_Mat2) <- yrs
    colnames(hfCrn_Mat2) <- 0:k
    
    
    res <- list(infoList = infoList,
                k = k,
                ssfCrn = finalCrn,
                # The SF measurements
                sfRW_Array = sfRW_Array2,
                # The rescaled SF measurements
                sfRWRescaled_Array = sfRWRescaled_Array2,
                # The rescaled SF curves
                sfRWRescaledCurves_Array = sfRWRescaledCurves_Array2,
                # The SF RWI
                sfRWI_Array = sfRWI_Array2,
                # The SF crn
                sfCrn_Mat = sfCrn_Mat2,
                # The high freq chronology
                hfCrn_Mat = hfCrn_Mat2,
                # The high freq chronology residuals
                hfCrnResids_Mat = hfCrnResids_Mat,
                # The median abs diff
                MAD_Vec = MAD_Vec)
    comment(res) <- "ssfLong"
    class(res) <- c("crn","data.frame")
    return(res)
  }
  else{
    
    return(finalCrn)
  }
}

rcs <- function(rwl, po = NULL, nyrs = NULL, f = 0.5, biweight = TRUE,
                ratios = TRUE, rc.out = FALSE, make.plot = TRUE,
                method = c("caps", "ads"), min.n = NULL,
                pos.slope = TRUE, ...) {
  
  method <- match.arg(method)
  
  if (is.null(po)) {
    po <- data.frame(series = names(rwl),
                     pith.offset = rep(1L, ncol(rwl)),
                     stringsAsFactors = FALSE)
  }
  
  sortByIndex <- function(x) {
    nas <- is.na(x)
    c(x[!nas], x[nas])
  }
  
  n.col <- length(rwl)
  col.names <- names(rwl)
  seq.cols <- seq_len(n.col)
  
  rwl2 <- rwl
  rownames(rwl2) <- rownames(rwl2)
  
  max.series.length <- max(colSums(!is.na(rwl)))
  max.series.length.plus.po <- max.series.length + max(po[, 2])
  
  rwl.ord <- apply(rwl2, 2, sortByIndex)
  
  rwca <- matrix(NA,
                 ncol = n.col,
                 nrow = max.series.length.plus.po)
  
  for (i in seq.cols) {
    yrs2pith <- po[po[, 1] %in% col.names[i], 2]
    series.length <- sum(!is.na(rwl[[i]]))
    rwca[yrs2pith:(yrs2pith + series.length - 1), i] <- rwl.ord[1:series.length, i]
  }
  
  if (biweight) {
    ca.m <- apply(rwca, 1, tbrm, C = 9)
  } else {
    ca.m <- rowMeans(rwca, na.rm = TRUE)
  }
  
  # sample depth by cambial age
  ca.n <- rowSums(!is.na(rwca))
  
  # truncate RC tail where sample depth falls below min.n
  if (!is.null(min.n)) {
    last.valid <- max(which(ca.n >= min.n))
    ca.m[seq_along(ca.m) > last.valid] <- NA
  }
  
  if (method == "caps") {
    nyrs2 <- if (is.null(nyrs)) floor(length(na.omit(ca.m)) * 0.1) else nyrs
    tmp <- caps(y = na.omit(ca.m), nyrs = nyrs2, f = f)
  } else {
    nyrs2 <- if (is.null(nyrs)) 50 else nyrs
    tmp <- ads(y = na.omit(ca.m), nyrs0 = nyrs2, pos.slope = pos.slope)
  }
  
  rc <- rep(NA, nrow(rwca))
  rc[!is.na(ca.m)] <- tmp
  
  if (ratios) {
    rwica <- rwca / rc
  } else {
    rwica <- rwca - rc
  }
  
  rwi <- rwl2
  yrs <- as.numeric(row.names(rwi))
  for (i in seq.cols) {
    yrs2pith <- po[po[, 1] %in% col.names[i], 2]
    series.length <- sum(!is.na(rwl[[i]]))
    series.yrs <- yr.range(rwl2[[i]], yr.vec = yrs)
    first <- series.yrs[1]
    last <- series.yrs[2]
    rwi[[i]][yrs %in% first:last] <- rwica[yrs2pith:(yrs2pith + series.length - 1), i]
  }
  
  if (make.plot) {
    par(mar = c(4, 4, 0.5, 0.5) + 0.1, mgp = c(1.25, 0.25, 0), tcl = 0.25)
    plot(rwca[, 1], ylim = range(rwca, na.rm = TRUE), type = "n", ylab = "mm",
         xlab = gettext("Cambial Age (Years)", domain = "R-dplR"), ...)
    for (i in seq.cols) {
      lines(rwca[, i], col = adjustcolor("grey20", alpha.f = min(0.15, 5/n.col)))
    }
    lines(ca.m, lwd = 1, col = "grey30")
    lines(rc, lwd = 2.5, col = "steelblue")
    legend("topright",
           legend = c("Series", "Mean", "Regional Curve"),
           col = c(adjustcolor("grey20", alpha.f = 0.5), "grey30", "steelblue"),
           lwd = c(1, 1, 2.5),
           bty = "n")
  }
  
  if (rc.out) {
    list(rwi = rwi, rc = rc)
  } else {
    rwi
  }
}

normalize1 <- function(rwl, n, prewhiten){
    rwl.mat <- as.matrix(rwl)
    ## Run hanning filter over the data if n isn't NULL
    ## divide by mean if n is null
    if(is.null(n)){
        master.stats <- colMeans(rwl.mat, na.rm=TRUE)
        master.mat <- sweep(rwl.mat, 2, master.stats, "/")
    } else {
        #master.stats <- apply(rwl.mat, 2, hanning, n)
        ## 15-dec-2022 AGB found a bug where a div0 was resulting in Nan. So recoding zeros.
        ## Apply hanning
        ## Recode any zero values to 0.001 -- here or after prewhiten? Same div0 bug should apply?
        master.stats <- apply(rwl.mat, 2, function(x){
          x2 <- hanning(x,n=n)
          x2[x2==0] <- 0.001
          x2
        })
        master.mat <- rwl.mat / master.stats
    }
    ## Apply ar if prewhiten
    if(prewhiten){
        ## take note of, ignore later, any columns without at least
        ## four observations
        idx.good <- colSums(!is.na(master.mat)) > 3
        master.mat <- apply(master.mat, 2, ar.func)
    } else {
        idx.good <- rep(TRUE, ncol(master.mat))
    }
    list(rwi.mat=master.mat, idx.good=idx.good)
}

`i.detrend` <- function(rwl, y.name=names(rwl), nyrs = NULL, f = 0.5,
                        pos.slope = FALSE)
{
    if(!is.data.frame(rwl))
        stop("'rwl' must be a data.frame")
    out <- rwl
    n.col <- ncol(rwl)
    fmt <- gettext("Detrend series %d of %d\n", domain="R-dplR")
    for(i in seq_len(n.col)){
        cat(sprintf(fmt, i, n.col))
        fits <- i.detrend.series(rwl[[i]], y.name=y.name[i], nyrs = nyrs,
                                 f = f, pos.slope = pos.slope)
        out[, i] <- fits
    }
    out
}

\encoding{UTF-8}
\name{morlet}
\alias{morlet}
\title{ Perform a Continuous Morlet Wavelet Transform }
\description{
  This function performs a continuous wavelet transform on a time series.
}
\usage{
morlet(y1, x1 = seq_along(y1), p2 = NULL, dj = 0.25, siglvl = 0.95)
}
\arguments{
  \item{y1}{\code{numeric} vector.  Series to be transformed.}
  \item{x1}{\code{numeric}.  A vector of values giving the years for the
    plot.  Must be the same length as \code{length(\var{y1})}.}
  \item{p2}{\code{numeric}.  The number of power of two to be computed
    for the wavelet transform.  Calculated from length of \code{\var{y1}}
    if \code{NULL}. }
  \item{dj}{\code{numeric}.  Sub-octaves per octave calculated.}
  \item{siglvl}{\code{numeric}.  Level for the significance test.}
}
\details{
  This performs a continuous wavelet transform of a time series.  This
  function is typically invoked with \code{\link{wavelet.plot}}.
}
\value{
  A \code{list} containing:
    \item{y}{\code{numeric}.  The original time series.}
    \item{x}{\code{numeric}.  The time values.}
    \item{wave}{\code{complex}.  The wavelet transform.}
    \item{coi}{\code{numeric}.  The cone of influence.}
    \item{period}{\code{numeric}.  The period.}
    \item{Scale}{\code{numeric}.  The scale.}
    \item{Signif}{\code{numeric}.  The significant values.}
    \item{Power}{\code{numeric}.  The squared power.}
}
\references{
  Torrence, C. and Compo, G. P. (1998) A practical guide to wavelet
  analysis.  \emph{Bulletin of the American Meteorological Society},
  \bold{79}(1), 61\enc{–}{--}78.
}

\note{ This is a port of Torrence\enc{’}{'}s \acronym{IDL} code,
which can be accessed through the
\href{https://web.archive.org/web/20141216211516/https://paos.colorado.edu/research/wavelets/software.html}{Internet Archive Wayback Machine}.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{wavelet.plot}} }
\examples{library(utils)
data(ca533)
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
ca533.crn <- chron(ca533.rwi, prewhiten = FALSE)
Years <- time(ca533.crn)
CAMstd <- ca533.crn[, 1]
out.wave <- morlet(y1 = CAMstd, x1 = Years, dj = 0.1, siglvl = 0.99)
}
\keyword{ hplot }

\encoding{UTF-8}
\name{corr.series.seg}
\alias{corr.series.seg}
\title{ Compute Correlation between a Series and a Master Chronology }
\description{
  Compute correlation between a tree-ring series and a master chronology by
  segment.
}
\usage{
corr.series.seg(rwl, series, series.yrs = as.numeric(names(series)),
                seg.length = 50, bin.floor = 100, n = NULL,
                prewhiten = TRUE, biweight = TRUE, 
                method = c("spearman", "pearson","kendall"),
                pcrit = 0.05,
                make.plot = TRUE, floor.plus1 = FALSE, \dots)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}}. }
  \item{series}{ a \code{numeric} or \code{character} vector.  Usually a
    tree-ring series.  If the length of the value is 1, the
    corresponding column of \code{\var{rwl}} is selected (by name or
    position) as the series and ignored when building the master
    chronology.  Otherwise, the value must be \code{numeric}. }
  \item{series.yrs}{ a \code{numeric} vector giving the years of
    \code{\var{series}}.  Defaults to
    \code{as.numeric(names(\var{series}))}.  Ignored if
    \code{\var{series}} is an index to a column of \code{\var{rwl}}. }
  \item{seg.length}{ an even integral value giving length of segments in
    years (e.g., 20, 50, 100 years). }
  \item{bin.floor}{ a non-negative integral value giving the base for
    locating the first segment (e.g., 1600, 1700, 1800
    \acronym{AD}).  Typically 0, 10, 50, 100, etc. }
  \item{n}{ \code{NULL} or an integral value giving the filter length
    for the \code{\link{hanning}} filter used for removal of low
    frequency variation. }
  \item{prewhiten}{ \code{logical} flag.  If \code{TRUE} each series is
    whitened using \code{\link{ar}}. }
  \item{biweight}{ \code{logical} flag.  If \code{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}}. }
  \item{method}{Can be either \code{"pearson"}, \code{"kendall"}, or
    \code{"spearman"} which indicates the correlation coefficient to be
    used.  Defaults to \code{"spearman"}.  See \code{\link{cor.test}}. }
  \item{pcrit}{ a number between 0 and 1 giving the critical value for
    the correlation test. }
  \item{make.plot}{ \code{logical} flag indicating whether to make a
    plot. }
  \item{floor.plus1}{ \code{logical} flag.  If \code{TRUE}, one year is
    added to the base location of the first segment (e.g., 1601, 1701,
    1801 \acronym{AD}). }
  \item{\dots}{ other arguments passed to plot. }
}
\details{

  This function calculates the correlation between a tree-ring series and a
  master chronology built from a rwl object.  Correlations are done by
  segment (see below) and with a moving correlation with length equal to
  the \code{\var{seg.length}}.  The function is typically invoked to
  produce a plot.

}
\value{
  A \code{list} containing matrices \code{\var{bins}},
  \code{\var{moving.rho}}, and vectors \code{\var{spearman.rho}},
  \code{\var{p.val}}, and \code{\var{overall}}.

  Matrix \code{\var{bins}} contains the years encapsulated by each bin
  (segments).  Matrix \code{\var{moving.rho}} contains the moving
  correlation and p-value for a moving average equal to
  \code{\var{seg.length}}.  Vector \code{\var{spearman.rho}} contains
  the correlations by bin and \code{\var{p.val}} contains
  the p-values.  Vector \code{\var{overall}} contains the average
  correlation and p-value.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{
  \code{\link{corr.series.seg}}, \code{\link{skel.plot}},
  \code{\link{series.rwl.plot}}, \code{\link{ccf.series.rwl}}
}
\examples{library(utils)
data(co021)
dat <- co021
## Create a missing ring by deleting a year of growth in a random series
flagged <- dat$"641143"
flagged <- c(NA, flagged[-325])
names(flagged) <- rownames(dat)
dat$"641143" <- NULL
seg.100 <- corr.series.seg(rwl = dat, series = flagged,
                           seg.length = 100, biweight = FALSE)
\dontrun{
flagged2 <- co021$"641143"
names(flagged2) <- rownames(dat)
seg.100.1 <- corr.series.seg(rwl=dat, seg.length=100, biweight=FALSE,
                             series = flagged2)
## Select series by name or column position
seg.100.2 <- corr.series.seg(rwl=co021, seg.length=100, biweight=FALSE,
                             series = "641143")
seg.100.3 <- corr.series.seg(rwl=co021, seg.length=100, biweight=FALSE,
                             series = which(colnames(co021) == "641143"))
identical(seg.100.1, seg.100.2) # TRUE
identical(seg.100.2, seg.100.3) # TRUE
}
}
\keyword{ manip }

\encoding{UTF-8}
\name{detrend.series}
\alias{detrend.series}
\title{ Detrend a Ring-Width Series }
\description{
  Detrend a tree-ring series by one of two methods, a smoothing spline or
  a statistical model.  The series and fits are plotted by default.
}
\usage{
detrend.series(y, y.name = "", make.plot = TRUE,
               method = c("Spline", "ModNegExp", "Mean", 
               "Ar", "Friedman", "ModHugershoff","AgeDepSpline"),
               nyrs = NULL, f = 0.5, pos.slope = FALSE,
               constrain.nls = c("never", "when.fail", "always"),
               verbose = FALSE, return.info = FALSE,
               wt, span = "cv", bass = 0, difference = FALSE)
}
\arguments{

  \item{y}{ a \code{numeric} vector.  Usually a tree-ring series. }

  \item{y.name}{ an optional \code{character} vector to name the series
    for plotting purposes. }

  \item{make.plot}{ a \code{logical} flag.  Makes plots of the raw data
    and detrended data if \code{TRUE}. }

  \item{method}{ a \code{character} vector to determine the detrending
    methods.  See details below.  Possible values are all subsets of\cr
    \code{c("Spline", "ModNegExp", "Mean", "Ar", "Friedman", "ModHugershoff",
    "AgeDepSpline")}.
    Defaults to using all the available methods.}

  \item{nyrs}{ a number controlling the smoothness of the
    fitted curve in methods. See Details.}
  
  \item{f}{ a number between 0 and 1 giving the frequency response or
    wavelength cutoff in method \code{"Spline"}.  Defaults to 0.5. 
    See \code{\link{caps}}.}
  
  \item{pos.slope}{ a \code{logical} flag.  Will allow for a positive
    slope to be used in method \code{"ModNegExp"} or \code{"AgeDepSpline"}.  
    If \code{FALSE} the
    line will be horizontal. }

  \item{constrain.nls}{ a \code{character} string which controls
    the constraints of the \code{"ModNegExp"} model and and the 
    \code{"ModHugershoff"} model which are fit using nonlinear least-squares 
    via \code{\link{nls}}.  The value is an
    answer to the question: When should the parameters of the \code{\link{nls}}
    function be constrained?  The options are
    \code{"never"}: do not constrain (the default), \code{"when.fail"}:
    only compute the constrained solution if the unconstrained fit
    contains other than positive values, and \code{"always"}: return the
    constrained solution, even if the unconstrained one would have been
    valid.  See \sQuote{Details}. }
  
  \item{verbose}{ a \code{logical} flag.  Write out details to the screen? }

  \item{return.info}{ a \code{logical} flag.  If \code{TRUE}, details
    about models and data will be added to the return value.  See
    \sQuote{Value}. }
  
  \item{wt}{ a \code{numeric} vector of case weights for method
    \code{"Friedman"}. The default means equals weights. }

  \item{span}{ a \code{numeric} value controlling method
    \code{"Friedman"}, or \code{"cv"} (default) for automatic choice by
    cross-validation. See \code{\link{supsmu}}. }

  \item{bass}{ a \code{numeric} value controlling the smoothness of the
    fitted curve in method \code{"Friedman"}. See \code{\link{supsmu}}. }

  \item{difference}{ a \code{logical} flag.  Compute residuals by subtraction if TRUE, otherwise use division. }

}
\details{
  This detrends and standardizes a tree-ring series.  The detrending is the estimation and removal of the tree\enc{’}{'}s natural biological growth trend.  The default standardization is done by dividing each series by the growth trend to produce units in the dimensionless ring-width index (\acronym{RWI}).  If \code{difference} is TRUE, the index is calculated by subtraction. Values of zero (typically missing rings) in \code{y} are set to 0.001. 
  
  There are currently seven methods available for
  detrending although more are certainly possible.  The methods
  implemented are an age-dependent spline via \code{\link{ads}} 
  (\code{\var{method} = "AgeDepSpline"}), the residuals of an AR model
  (\code{\var{method} = "Ar"}), Friedman's Super Smoother 
  (\code{\var{method} = "Friedman"}), a simple horizontal line
  (\code{\var{method} = "Mean"}),  or a modified Hugershoff
  curve (\code{\var{method} = "ModHugershoff"}), a modified negative exponential
  curve (\code{\var{method} = "ModNegExp"}), and a smoothing spline via \code{\link{caps}} (\code{\var{method} = "Spline"}).

  The \code{"AgeDepSpline"} approach uses an age-dependent spline with an initial
  stiffness of 50 (\code{\var{nyrs}=50}). See \code{\link{ads}}. If some of the fitted 
  values are not positive then method \code{"Mean"} is used. However, this is
  extremely unlikely.

  The \code{"Ar"} approach is also known as prewhitening where the detrended 
  series is the residuals of an \code{\link{ar}} model divided by the
  mean of those residuals to yield a series with white noise and a mean of one.
  This method removes all but the high frequency variation in the series
  and should only be used as such.

  The \code{"Friedman"} approach uses Friedman\enc{’}{'}s \sQuote{super
  smoother} as implemented in \code{\link{supsmu}}.  The parameters
  \code{\var{wt}}, \code{\var{span}} and \code{\var{bass}} can be
  adjusted, but \code{\var{periodic}} is always set to \code{FALSE}. If some of 
  the fitted values are not positive then method \code{"Mean"} is used.

  The \code{"Mean"} approach fits a horizontal line using the mean of
  the series.  This method is the fallback solution in cases where the
  \code{"Spline"} or the linear fit (also a fallback solution itself)
  contains zeros or negative values, which would lead to invalid
  ring-width indices.
  
  The \code{"ModHugershoff"} approach attempts to fit a Hugershoff
  model of biological growth of the form \eqn{f(t) = a t^b e^{-g t} + d}{f(t)
  = a t^b exp(-g t) + d}, where the argument of the function is time, using
  \code{\link{nls}}.  See Fritts (2001) for details about the
  parameters.  Option \code{\var{constrain.nls}} gives a
  possibility to constrain the parameters of the modified negative
  exponential function.  If the constraints are enabled, the nonlinear
  optimization algorithm is instructed to keep the parameters in the
  following ranges: \eqn{a \ge 0}{a >= 0}, \eqn{b \ge 0}{b >= 0} and
  \eqn{d \ge 0}{d >= 0}.  The default is to not constrain the parameters
  (\code{\var{constrain.nls} = "never"}) for  \code{\link{nls}} but
  warn the user when the parameters go out of range.
  
  If a suitable nonlinear model cannot be fit
  (function is non-decreasing or some values are not positive) then a
  linear model is fit.  That linear model can have a positive slope
  unless \code{\var{pos.slope}} is \code{FALSE} in which case method
  \code{"Mean"} is used.

  The \code{"ModNegExp"} approach attempts to fit a classic nonlinear
  model of biological growth of the form \eqn{f(t) = a e^{b t} + k}{f(t)
  = a exp(b t) + k}, where the argument of the function is time, using
  \code{\link{nls}}.  See Fritts (2001) for details about the
  parameters.  Option \code{\var{constrain.nls}} gives a
  possibility to constrain the parameters of the modified negative
  exponential function.  If the constraints are enabled, the nonlinear
  optimization algorithm is instructed to keep the parameters in the
  following ranges: \eqn{a \ge 0}{a >= 0}, \eqn{b \le 0}{b <= 0} and
  \eqn{k \ge 0}{k >= 0}.  The default is to not constrain the parameters
  (\code{\var{constrain.nls} = "never"}) for  \code{\link{nls}} but
  warn the user when the parameters go out of range.
  
  If a suitable nonlinear model cannot be fit
  (function is non-decreasing or some values are not positive) then a
  linear model is fit.  That linear model can have a positive slope
  unless \code{\var{pos.slope}} is \code{FALSE} in which case method
  \code{"Mean"} is used.

  The \code{"Spline"} approach uses a spline where the frequency
  response is 0.50 at a wavelength of 0.67 * \dQuote{series length in
  years}, unless specified differently using \code{\var{nyrs}} and
  \code{\var{f}} in the function \code{\link{caps}}. If some of the fitted 
  values are not positive then method \code{"Mean"} is used.

  These methods are chosen because they are commonly used in
  dendrochronology.  There is a rich literature on detrending
  and many researchers are particularly skeptical of the use of the 
  classic nonlinear model of biological growth (\eqn{f(t) = a e^{b t} + k}{f(t)
  = a exp(b t) + k}) for detrending.  It is, of course, up to the 
  user to determine the best detrending method for their data.

  Note that the user receives a warning if a series has negative values in the  
  fitted curve. This happens fairly commonly with with the \sQuote{Ar} method 
  on high-order data. It happens less often with method \sQuote{Spline} but
  isn't unheard of (see \sQuote{Examples}). If this happens, users should look
  carefully at their data before continuing. Automating detrending and not 
  evaluating each series individually is folly. Remember, frustration over 
  detrending is the number one cause of dendros going to live as hermits in 
  the tallgrass prairie, where there are no trees to worry about.  

  See the references below for further details on detrending. It's a dark art.
}
\value{

  If several methods are used, returns a \code{data.frame} containing
  the detrended series (\code{\var{y}}) according to the methods used.
  The columns are named and ordered to match \code{\var{method}}.  If
  only one method is selected, returns a vector.

  If \code{\var{return.info}} is \code{TRUE}, the return value is a
  \code{list} with four parts:

  \item{series}{ the main result described above (\code{data.frame} or
    vector) }

  \item{curves}{ the curve or line used to detrend \code{series}. Either a \code{data.frame} or vector. }

  \item{model.info}{ Information about the models corresponding to each
    output series.  Whereas \code{\var{series}} may return a vector,
    \code{\var{model.info}} is always a list where each top level
    element corresponds to one selected method.  Also these elements are
    named and arranged according to the methods selected.  Each element
    is a list with some of the following sub-elements, depending on
    which detrending method was actually used:

    \describe{

      \item{n.zeros}{ See \code{"data.info"} below.  Always present. }

      \item{zero.years}{ See \code{"data.info"}.  Always present. }

      \item{method}{ The method actually used for detrending.  One of
        \code{"Mean"}, \code{"Line"}, \code{"ModNegExp"},
        \code{"Spline"} or \code{"Ar"}.  Always present. }

      \item{mean}{ Mean of the input series, missing values removed.
        Only for method \code{"Mean"}. }

      \item{coefs}{ Coefficients of the model.  Methods \code{"Line"}
        and \code{"ModNegExp"}.}

      \item{formula}{ The \code{"ModNegExp"} \code{\link{formula}}. }

      \item{is.constrained}{ A \code{logical} flag indicating whether
        the parameters of the \code{"ModNegExp"} model were
        constrained.  Only interesting when argument
        \code{\var{constrain.nls}} is set to \code{"when.fail"}. }

      \item{nyrs}{ The value of \code{\var{nyrs}} used for
        \code{\link{caps}}. For methods \code{"Spline"} and
         \code{"AgeDepSpline"}. }

      \item{order}{ The order of the autoregressive model, selected by
        AIC (Akaike information criterion).  Only for method
        \code{"Ar"}. }

      \item{ar}{ The autoregressive coefficients used by method
        \code{"Ar"}.  A \code{numeric} vector ordered by increasing
        lag. }

    }
    
  }


  \item{data.info}{ Information about the input series: number
    (\code{"n.zeros"}) and location (\code{"zero.years"}) of zero
    values.  If the locations are in a \code{character} vector, they are
    years.  Otherwise they are indices to the input series. }
    
  \item{dirtDog}{ A logical flag indicating whether the requested method resulted in negative values for the curve fit, what Cook's ARSTAN called a Dirty Dog. }


}
\references{
  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Fritts, H. C. (2001) \emph{Tree Rings and Climate}.
  Blackburn.  \acronym{ISBN-13}: 978-1-930665-39-2.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela.  A bug fix
  related to negative output values is based on work by Alice Cecile. }
\seealso{ \code{\link{detrend}} }
\examples{library(stats)
library(utils)
## Use series CAM011 from the Campito data set
data(ca533)
series <- ca533[, "CAM011"]
names(series) <- rownames(ca533)
# defaults to all six methods
series.rwi <- detrend.series(y = series, y.name = "CAM011", verbose=TRUE)
# see plot with three methods
series.rwi <- detrend.series(y = series, y.name = "CAM011",
                             method=c("Spline", "ModNegExp","Friedman"),
                             difference=TRUE)
# see plot with two methods
# interesting to note difference from ~200 to 250 years 
# in terms of what happens to low frequency growth
series.rwi <- detrend.series(y = series, y.name = "CAM011",
                             method=c("Spline", "ModNegExp"))
# see plot with just one method and change the spline
# stiffness to 50 years (which is not neccesarily a good choice!)
series.rwi <- detrend.series(y = series, y.name = "CAM011",
                             method="Spline",nyrs=50)
                             
# note that method "Ar" doesn't get plotted in first panel
# since this approach doesn't approximate a growth curve.
series.rwi <- detrend.series(y = series, y.name = "CAM011",
                             method="Ar")
                             
# note the difference between ModNegExp and ModHugershoff at the 
# start of the series. Also notice how curves, etc. are returned
# via return.info
data(co021)
series <- co021[, 4]
names(series) <- rownames(co021)
series.rwi <- detrend.series(y = series, y.name = names(co021)[4],
                             method=c("ModNegExp", "ModHugershoff"),
                             verbose = TRUE, return.info = TRUE, 
                             make.plot = TRUE)
                             
# A dirty dog.
# In the case of method=="Spline" the function carries-on
# and applies method=="Mean" as an alternative. 
data(nm046)
series <- nm046[,8]
names(series) <- rownames(nm046)
series.rwi <- detrend.series(y = series, y.name = names(nm046)[8],
                             method="Spline",
                             make.plot = FALSE)

}
\keyword{ manip }

\encoding{UTF-8}
\name{ssf}
\alias{ssf}
\title{ Simple Signal Free Standardization }
\description{
  A simple implementation of the signal-free chronology
}
\usage{
ssf(rwl, 
    method="Spline", 
    nyrs = NULL, 
    difference = FALSE,
    max.iterations = 25, 
    mad.threshold = 5e-4,
    recode.zeros = FALSE,
    return.info = FALSE, 
    verbose = TRUE)
}
\arguments{

  \item{rwl}{ a \code{rwl} object with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}}. See \sQuote{Details}. }

  \item{method}{ a \code{character} vector to determine the detrending
    method.  See \sQuote{Details} below.  Possible values are \cr \code{c("Spline", "AgeDepSpline")}.}
    
  \item{nyrs}{ a number controlling the smoothness of the
    fitted curve in methods. See \sQuote{Details} in \code{\link{detrend.series}}.}
  
  \item{difference}{ a \code{logical} flag.  Compute residuals by subtraction 
  if TRUE, otherwise use division. See \sQuote{Details} 
  in \code{\link{detrend.series}} }

  \item{max.iterations}{ a \code{numeric}. The maximum number of iterations to be used in the fitting process. See \sQuote{Details}. }

  \item{mad.threshold}{ a \code{numeric}. The median absolute difference threshold used for a stopping criteria. See \sQuote{Details}. }

  \item{recode.zeros}{ a \code{logical} flag.  Replace values of zero in \code{rwl} with 0.001 if TRUE. }

  \item{return.info}{ a \code{logical} flag.  If \code{TRUE}, details
    about models and data will be added to the return value.  See
    \sQuote{Value}. }
    
  \item{verbose}{ a \code{logical} flag.  Write out details to the screen? }
  
}

\details{
This function creates a simple signal-free chronology that loosely follows the procedures laid out on p75 of Melvin and Briffa (2008). This function is a lighter version of that procedure and users who want more control and refinement should look to the CRUST program decried in Melvin and Briffa (2014). These steps are described in more detail in \href{https://opendendro.github.io/dplR-workshop/}{Learning to Love R}.
  \enumerate{ 
    \item Detrend each series using the selected method, calculate RWI by division (or subtraction), and create an initial mean-value chronology.
    \item Create signal-free measurements by dividing (or subtracting) each series of measurements by the chronology. If \code{return.info} is invoked these are returned in \code{sfRW_Array}.
    \item Rescale the signal-free measurements to their original mean. If \code{return.info} is invoked these are returned in \code{sfRWRescaled_Array}.
    \item If the sample depth is one, replace signal-free measurements with original measurements.
    \item Fit curves to signal free measurements.If \code{return.info} is invoked these are returned in \code{sfRWRescaledCurves_Array}.
    \item Get new growth indicies by dividing (or subtracting) the original measurements by curves in the last step. If \code{return.info} is invoked these are returned in \code{sfRWI_Array}.
    \item Create a mean-value chronology using the indicies from the prior step. If \code{return.info} is invoked these are returned in \code{sfCrn_Mat}.
    \item Repeat steps two through seven up to \code{maxIter} or until the \code{madThreshold} is reached. The stopping criteria is determined using the absolute difference between filtered chronologies generated in interation k and k-1. This is done with the residuals of a high-pass filter on the chronology using a cubic smoothing spline (\code{\link{caps}}) with the stiffness set as the median of the segment lengths of series contributing to the chronology. The stopping threshold is calculated as the median absolute difference of the kth and kth-1 chronologies weighted by the normalized sample depth. If \code{return.info} is invoked the residual chronologies are returned in \code{hfCrnResids_Mat} and the median absolute differences are returns in \code{MAD_Vec}.
}
  
    The input object (\code{rwl}) should be of \code{class} \code{rwl}. If it not, the function will attempt to coerce it using \code{\link{as.rwl}} and a warning will be issued.
  
  See the references below for further details on detrending. It's a dark art.
}
\value{
    An object of of class \code{crn} and \code{data.frame} with the signal-free chronology and the sample depth. The years are stored as row numbers.
    
    If \code{return.info} is \code{TRUE} a \code{list} containing ouptput by iteration (k): 
    
    \item{infoList}{ a \code{list} with information on the arguments used in the function call. } 
    
    \item{k}{a \code{numeric} containing the number of iterations. }
    
    \item{ssfCrn}{ the signal-free chronology as above. } 

    \item{sfRW_Array}{ an \code{array} of years by series by iteration (0:k) that holds the signal free measurements. The values in \code{sfRW_Array[,,1]} are what is used at the start of the iterative process (iteration 0) which here are the original data.}     

    \item{sfRWRescaled_Array}{ an \code{array} of years by series by iteration (0:k) that holds the rescaled signal free measurements. The values in \code{sfRWRescaled_Array[,,1]} are what is used at the start of the iterative process (iteration 0) which here are the original data.}     

    \item{sfRWRescaledCurves_Array}{ an \code{array} of years by series by iteration (0:k) that holds the rescaled signal free curve fits. The values in \code{sfRWRescaledCurves_Array[,,1]} are the curves used at the start of the iterative process (iteration 0).}     

    \item{sfRWI_Array}{ an \code{array} of years by series by iteration (0:k) that holds the detrended signal free measurements. The values in \code{sfRWI_Array[,,1]} are the RWI values at the start of the iterative process (iteration 0).}     

    \item{sfCrn_Mat}{ a \code{matrix} of years by iteration (0:k) that holds the calculated chronology. The values in \code{sfCrn_Mat[,1]} is the initial chronology from the start of the iterative process (iteration 0).}     

    \item{hfCrn_Mat}{ a \code{matrix} of years by iteration (0:k) that holds the calculated high-frequency chronology. The values in \code{hfCrn_Mat[,1]} is the initial high-frequency chronology from the start of the iterative process (iteration 0).}     

    \item{hfCrnResids_Mat}{ a \code{matrix} of years by k-1 that holds the differences between the kth and the kth-1 high frequency chronology residuals. }   
    
    \item{MAD_out}{ a \code{vector} containing the median absolute difference between iteration k and k-1. }
    

}
\references{

Melvin, TM, Briffa, KR (2008) A 'signal-free' approach to dendroclimatic standardisation. Dendrochronologia 26: 71–86 doi: 10.1016/j.dendro.2007.12.001

Melvin T. M. and Briffa K.R. (2014a) CRUST: Software for the implementation of Regional Chronology Standardisation: Part 1. Signal-Free RCS. Dendrochronologia 32, 7-20, doi: 10.1016/j.dendro.2013.06.002

Melvin T. M. and Briffa K.R. (2014b) CRUST: Software for the implementation of Regional Chronology Standardisation: Part 2. Further RCS options and recommendations. Dendrochronologia 32, 343-356, doi: 10.1016/j.dendro.2014.07.008

}
\author{ Ed Cook provided Fortran code that was ported to R by Andy Bunn. }
\seealso{ \code{\link{detrend}}. \code{\link{chron}} }
\examples{library(stats)
data(wa082)
wa082_SSF <- ssf(wa082)
plot(wa082_SSF,add.spline=TRUE,nyrs=20)

wa082_SSF_Full <- ssf(wa082,method = "AgeDepSpline", 
  difference = TRUE, return.info = TRUE)
plot(wa082_SSF_Full$ssfCrn,add.spline=TRUE,nyrs=20)

}
\keyword{ manip }

\encoding{UTF-8}
\name{combine.rwl}

\alias{combine.rwl}

\title{ Combine Tree-Ring Data Sets }

\description{ This function combines any number of \code{data.frames} of
  tree-ring data into one \code{data.frame}.  }

\usage{
combine.rwl(x, y = NULL)
}

\arguments{
  \item{x}{ either a \code{data.frame} to be combined with
    \code{\var{y}}, or a \code{list} of \code{data.frame}s to be
    combined. } \item{y}{a \code{data.frame} to be combined with
    \code{data.frame} \code{\var{x}}.}
}

\details{ The sequence of years in each \code{data.frame} must be
  increasing and continuous.  The output produced by the function
  also fulfills this condition.  If the input is differently formatted,
  the result will be wrong.}

\value{An object of class \code{c("rwl", "data.frame")} with the series in 
  columns and the years as
  rows.  The keycodes are the column names and the years are the row
  names. }

\author{ Christian Zang.  Patched by Mikko Korpela. }

\examples{library(utils)
data(ca533)
data(co021)
combi1 <- combine.rwl(list(ca533, co021))
## or alternatively for data.frames to combine
combi2 <- combine.rwl(ca533, co021)
identical(combi1, combi2) # TRUE
}

\keyword{ manip }

\encoding{UTF-8}
\name{seg.plot}
\alias{seg.plot}
\title{ Segment Plot }
\description{
  Makes a segment plot of tree-ring data.
}
\usage{
seg.plot(rwl, \dots)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}}. }
  \item{\dots}{ arguments to be passed to plot. }
}
\details{
  This makes a simple plot of the length of each series in a tree-ring
  data set.
}
\value{
  None.  This function is invoked for its side effect, which is to
  produce a plot.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{spag.plot}} }
\examples{library(utils)
data(co021)
seg.plot(co021)
}
\keyword{ hplot }

\encoding{UTF-8}
\name{latexify}
\alias{latexify}
\title{
  Convert Character Strings for Use with LaTeX
}
\description{
  Some characters cannot be entered directly into a LaTeX document.
  This function converts the input \code{character} vector to a form
  suitable for inclusion in a LaTeX document in text mode.  It can be
  used together with \samp{\Sexpr} in vignettes.
}
\usage{
latexify(x, doublebackslash = TRUE, dashdash = TRUE,
         quotes = c("straight", "curved"),
         packages = c("fontenc", "textcomp"))
}
\arguments{
  \item{x}{ a \code{character} vector }
  \item{doublebackslash}{ a \code{logical} flag.  If \code{TRUE},
    backslashes in the output are doubled.  It seems that Sweave needs
    \code{TRUE} and knitr \code{FALSE}. }
  \item{dashdash}{ a \code{logical} flag.  If \code{TRUE} (the default),
    consecutive dashes (\dQuote{-}) in the input will be rendered as
    separate characters.  If \code{FALSE}, they will not be given any
    special treatment, which will usually mean that two dashes are
    rendered as an en dash and three dashes make an em dash. }
  \item{quotes}{ a \code{character} string specifying how single and
    double quotes (\acronym{ASCII} codes 39 and 34) and stand-alone
    grave accents (\acronym{ASCII} code 96) in the input are treated.
    The default is to use straight quotes and the proper symbol for the
    grave accent.  The other option is to use curved right side
    (closing) quotes, and let LaTeX convert the grave accent to opening
    curved quotes.  Straight double quotes are not available in the
    default OT1 font encoding of LaTeX.  Straight single quotes and the
    grave accent symbol require the \dQuote{textcomp} package.  See
    \code{\var{packages}}. }
  \item{packages}{ a \code{character} vector specifying the LaTeX
    packages allowed.  The use of some symbols in LaTeX requires
    commands or characters made available by an add-on package.  If a
    package required for a given character is not marked as available, a
    fallback solution is silently used.  For example, curved quotes may
    be used instead of straight quotes.  The supported packages are
    \code{"eurosym"} (not used by default), \code{"fontenc"} and
    \code{"textcomp"}.  Including \code{"fontenc"} in the vector means
    that some other encoding than OT1 is going to be used.  Note that
    straight quotes are an exception in the sense that a reasonable
    substitute (curved quotes) is available.  In many other cases,
    \code{"textcomp"} and \code{"fontenc"} are silently assumed. }
}
\details{

  The function is intended for use with unformatted inline text.
  Newlines, tabs and other whitespace characters (\code{"[:space:]"} in
  \link{regex}) are converted to spaces.  Control characters
  (\code{"[:cntrl:]"}) that are not whitespace are removed.  Other more
  or less special characters in the \acronym{ASCII} set are \sQuote{\{},
  \sQuote{\}}, \sQuote{\\}, \sQuote{#}, \sQuote{$}, \sQuote{\%},
  \sQuote{^}, \sQuote{&}, \sQuote{_}, \sQuote{~}, double quote,
  \sQuote{/}, single quote, \sQuote{<}, \sQuote{>}, \sQuote{|}, grave
  and \sQuote{-}.  They are converted to the corresponding LaTeX
  commands.  Some of the conversions are affected by user options,
  e.g. \code{\var{dashdash}}.

  Before applying the substitutions described above, input elements with
  \code{\link{Encoding}} set to \code{"bytes"} are printed and the
  output is stored using \code{\link[R.utils]{captureOutput}}.  The result of
  this intermediate stage is \acronym{ASCII} text where some characters
  are shown as their byte codes using a hexadecimal pair prefixed with
  \code{"\\x"}.  This set includes tabs, newlines and control
  characters.  The substitutions are then applied to the intermediate
  result.

  The quoting functions \code{\link{sQuote}} and \code{\link{dQuote}}
  may use non-\acronym{ASCII} quote characters, depending on the locale.
  Also these quotes are converted to LaTeX commands.  This means that
  the quoting functions are safe to use with any LaTeX input encoding.
  Similarly, some other non-\acronym{ASCII} characters, e.g. letters,
  currency symbols, punctuation marks and diacritics, are converted to
  commands.

  Adding \code{"eurosym"} to \code{\var{packages}} enables the use of the
  euro sign as provided by the \code{"eurosym"} package (\samp{\euro}).

  The result is converted to UTF-8 encoding, Normalization Form C (NFC).
  Note that this function will not add any non-\acronym{ASCII}
  characters that were not already present in the input.  On the
  contrary, some non-\acronym{ASCII} characters, e.g. all characters in
  the \code{"latin1"} (\acronym{ISO}-8859-1) \code{\link{Encoding}}
  (character set), are removed when converted to LaTeX commands.  Any
  remaining non-\acronym{ASCII} character has a good chance of working
  when the document is processed with XeTeX or LuaTeX, but the Unicode
  support available with pdfTeX is limited.

  Assuming that \samp{pdflatex} is used for compilation, suggested
  package loading commands in the document preamble are: \preformatted{\usepackage[T1]{fontenc}    \% no '"' in OT1 font encoding
\usepackage{textcomp}       \% some symbols e.g. straight single quote
\usepackage[utf8]{inputenx} \% UTF-8 input encoding
\input{ix-utf8enc.dfu}      \% more supported characters}

}
\value{
  A \code{character} vector
}
\references{
  INRIA.  Tralics: a LaTeX to XML translator, HTML documentation of all
  TeX commands.  \url{https://www-sop.inria.fr/marelle/tralics/}.

  Levitt, N., Persch, C., and Unicode, Inc. (2013) GNOME Character Map,
  application version 3.10.1.

  Mittelbach, F., Goossens, M., Braams, J., Carlisle, D., and
  Rowley, C. (2004) \emph{The LaTeX Companion}.  Addison-Wesley,
  second edition.  \acronym{ISBN-13}: 978-0-201-36299-2.

  Pakin, S. (2009) The Comprehensive LaTeX Symbol
  List.  \url{https://www.ctan.org/tex-archive/info/symbols/comprehensive}.

  The Unicode Consortium.  The Unicode Standard.
  \url{https://home.unicode.org/}.
}
\author{
  Mikko Korpela
}
\examples{
x1 <- "clich\xe9\nma\xf1ana"
Encoding(x1) <- "latin1"
x1
x2 <- x1
Encoding(x2) <- "bytes"
x2
x3 <- enc2utf8(x1)
testStrings <-
    c("different     kinds\nof\tspace",
      "control\a characters \ftoo",
      "{braces} and \\\\backslash",
      '#various$ \%other^ &characters_ ~escaped"/coded',
      x1,
      x2,
      x3)
latexStrings <- latexify(testStrings, doublebackslash = FALSE)
## All should be "unknown"
Encoding(latexStrings)
cat(latexStrings, sep="\n")
## Input encoding does not matter
identical(latexStrings[5], latexStrings[7])
}
\keyword{ utilities }

\encoding{UTF-8}
\name{treeMean}
\alias{treeMean}
\title{Calculate mean across cores in a tree}
\description{
  This function calculates the mean value for each tree in a rwl or rwi object.
}
\usage{
treeMean(rwl, ids, na.rm=FALSE)
}
\arguments{
  \item{rwl}{a \code{data.frame} of ring widths with
    \code{rownames(\var{rwl})} containing years and \code{colnames(rwl)}
    containing each series \acronym{ID} such as produced by
    \code{\link{read.rwl}}}
  \item{ids}{ a \code{data.frame} with column one named
    \code{"tree"} giving a \code{numeric} \acronym{ID} for each tree and
    column two named \code{"core"} giving a \code{numeric} \acronym{ID}
    for each core. }
  \item{na.rm}{ \code{logical} passed to \code{\link{rowMeans}}. Should 
    missing values be removed?}    
}
\details{
  This function averages together multiple cores to give a mean value of growth.
  It is very common in dendrochronology to take more than one core per tree. In
  those cases it is occasionally desirable to have an average of the cores.
  This function merely loops through the \code{rwl} object and calculates the
  \code{\link{rowMeans}} for each tree. If \code{na.rm=TRUE} trees with >1 
  sample will be averaged only over the period where the samples overlap. 
  If FALSE the output can vary in the number of samples. See examples.
}
\value{
  An object of class \code{c("rwl", "data.frame")} with the mean annual value 
  for each tree.
}

\author{ Andy Bunn. Patched and improved by Mikko Korpela. }

\seealso{ \code{\link{read.rwl}}, \code{\link{read.ids}}
}
\examples{
data(gp.rwl)
gp.ids <- read.ids(gp.rwl, stc = c(0, 2, 1))

gp.treeMean <- treeMean(gp.rwl, gp.ids)
gp.treeMean2 <- treeMean(gp.rwl, gp.ids, na.rm=TRUE)

# look at an example of a single tree with different averaging periods
tree40 <- data.frame(gp.rwl[, c("40A","40B")],
                     gp.treeMean[, "40", drop=FALSE],
                     gp.treeMean2[, "40", drop=FALSE])
names(tree40) <- c("coreA", "coreB", "treeMean1", "treeMean2")
head(tree40,50)

data(ca533)
ca533.treeMean <- treeMean(ca533, autoread.ids(ca533))
# plot using S3method for class "rwl"
plot(ca533.treeMean,plot.type="spag")

}
\keyword{ manip }

\encoding{UTF-8}
\name{gp.dbh}
\docType{data}
\alias{gp.dbh}
\title{ Ponderosa Pine Stem Diameters and Bark Thickness (\code{\link{gp.rwl}}) }
\description{
  This data set gives the diameter at breast height for each series that
  matches the series in \code{\link{gp.rwl}} \enc{–}{-} a data set of ponderosa
  pine (\emph{Pinus ponderosa}) from the Gus Pearson Natural Area
  (\acronym{GPNA}) in northern Arizona, \acronym{USA}.  Data are further
  described by Biondi and Qeadan (2008) and references therein.
}
\usage{data(gp.dbh)}
\format{
  A \code{data.frame} containing series \acronym{ID}s in column 1
  (\code{\var{series}}), tree diameter (in cm) at breast height
  (\code{\var{dbh}}), and the bark thickness (in cm).  This can be used
  together with the ring widths to calculate the area of each ring.
}
\source{
  DendroLab, University of Nevada Reno, \acronym{USA}.
  \url{https://dendrolaborg.wordpress.com/}
}
\references{
  Biondi, F. and Qeadan, F. (2008) A theory-driven approach to tree-ring
  standardization: Defining the biological trend from expected basal area
  increment.  \emph{Tree-Ring Research}, \bold{64}(2), 81\enc{–}{--}96.
}
\keyword{datasets}

\encoding{UTF-8}
\name{pass.filt}
\alias{pass.filt}
\title{ Low-pass, high-pass, band-pass, and stop-pass filtering }
\description{
  Applies low-pass, high-pass, band-pass, or stop-pass filtering to \code{\var{y}} with frequencies (or periods) supplied by the user.
}
\usage{
pass.filt(y, W, type = c("low", "high", "stop", "pass"),
          method = c("Butterworth", "ChebyshevI"),
          n = 4, Rp = 1)
}
\arguments{
  \item{y}{ a \code{numeric} vector, typically a tree-ring series. }
  \item{W}{ a \code{numeric} vector giving frequency or period of the
    filter. See details. }
  \item{type}{ a \code{character} giving the type of filter. Values can
    be "low", "high", "stop", or "pass" for low-pass, high-pass,
    band-pass, or stop-pass filters. Defaults to "low". }
  \item{method}{ a \code{character} specifying indicating whether to use
    a Butterworth (default) or a type I Chebyshev filter.}
  \item{n}{ a \code{numeric} value giving the order of the
    filter. Larger numbers create steeper fall off.}
  \item{Rp}{ a \code{numeric} value giving the dB for the passband ripple. }
}

\details{
  This function applies either a Butterworth or a Chebyshev type I filter of order \code{\var{n}} to a signal and is nothing more than a wrapper for  functions in the \code{signal} package. The filters are designed via \code{\link[signal]{butter}} and \code{\link[signal]{cheby1}}. The filter is applied via \code{\link[signal]{filtfilt}}. 
  
  The input data (\code{y}) has the mean value subtracted and is then padded via reflection at the start and the end to a distance of twice the maximum period. The padded data and the filter are passed to \code{\link[signal]{filtfilt}} after which the data are unpadded and returned afer the mean is added back. 
  
  The argumement \code{\var{W}} can be given in either frequency between 0 and 0.5 or, for convenience, period (minimum value of 2). For low-pass and high-pass filters, \code{\var{W}} must have a length of one. For low-pass and high-pass filters \code{\var{W}} must be a two-element vector (\code{c(low, high)}) specifying the lower and upper boundaries of the filter. 
  
  Because this is just a wrapper for casual use with tree-ring data the frequencies and periods assume a sampling frequency of one. Users are encouraged to build their own filters using the \code{signal} package.

}

\value{ 
  A filtered vector. 
}

\author{
  Andy Bunn. Patched and improved by Mikko Korpela.
}

\seealso{ \code{\link{hanning}}, \code{\link{detrend}} }

\examples{
data("co021")
x <- na.omit(co021[, 1])

# 20-year low-pass filter -- note freq is passed in
bSm <- pass.filt(x, W=0.05, type="low", method="Butterworth")
cSm <- pass.filt(x, W=0.05, type="low", method="ChebyshevI")
plot(x, type="l", col="grey")
lines(bSm, col="red")
lines(cSm, col="blue")

# 20-year high-pass filter -- note period is passed in
bSm <- pass.filt(x, W=20, type="high")
plot(x, type="l", col="grey")
lines(bSm, col="red")

# 20 to 100-year band-pass filter -- note freqs are passed in
bSm <- pass.filt(x, W=c(0.01, 0.05), type="pass")
cSm <- pass.filt(x, W=c(0.01, 0.05), type="pass", method="ChebyshevI")
plot(x, type="l", col="grey")
lines(bSm, col="red")
lines(cSm, col="blue")

# 20 to 100-year stop-pass filter -- note periods are passed in
cSm <- pass.filt(x, W=c(20, 100), type="stop", method="ChebyshevI")
plot(x, type="l", col="grey")
lines(cSm, col="red")
}
\keyword{ smooth }

\encoding{UTF-8}
\name{i.detrend.series}
\alias{i.detrend.series}
\title{ Interactively Detrend a Ring-Width Series }
\description{
  Interactively detrend a tree-ring series by one of three methods, a
  smoothing spline, a linear model, or the mean.  This is a wrapper for
  \code{\link{detrend.series}}.
}
\usage{
i.detrend.series(y, y.name = NULL, nyrs = NULL, f = 0.5,
                 pos.slope = FALSE)
}
\arguments{
  \item{y}{ a \code{numeric} vector.  Usually a tree-ring series. }
  \item{y.name}{ an optional \code{character} vector to name the series
    for plotting purposes. }
  \item{nyrs}{ a number giving the rigidity of the smoothing spline,
    defaults to 0.67 of series length if \code{\var{nyrs}} is
    \code{NULL}. }
  \item{f}{ a number between 0 and 1 giving the frequency response or
    wavelength cutoff.  Defaults to 0.5. }
  \item{pos.slope}{ a \code{logical} flag.  Will allow for a positive
    slope to be used in method \code{"ModNegExp"}.  If \code{FALSE} the
    line will be horizontal. }
}
\details{
  This function allows a user to choose a detrending method based on a
  plot that is produced by \code{\link{detrend.series}} for which it is
  essentially a wrapper.  The user enters their choice of detrended
  method via keyboard at a prompt.  See \code{\link{detrend.series}} for
  examples and details on the detrending methods.
}
\value{
  A vector containing the detrended series (\code{\var{y}}) according to
  the method used with names set to \code{colnames(\var{y})}.  These are
  typically years.  A plot is also produced and the user chooses a method
  through keyboard input.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{detrend.series}} }
\keyword{ iplot }

\encoding{UTF-8}
\name{gp.rwl}
\docType{data}
\alias{gp.rwl}
\title{ Ponderosa Pine Ring Widths from Gus Pearson Natural Area }
\description{
  This data set includes ring-width measurements for ponderosa pine
  (\emph{Pinus ponderosa}) increment cores collected at the Gus Pearson
  Natural Area (\acronym{GPNA}) in northern Arizona,
  \acronym{USA}.  There are 58 series from 29 trees (2 cores per
  tree).  Data are further described by Biondi and Qeadan (2008) and
  references therein.
}
\usage{data(gp.rwl)}
\format{
  A \code{data.frame} containing 58 ring-width series in columns and 421
  years in rows.
}
\source{
  DendroLab, University of Nevada Reno, \acronym{USA}.
  \url{https://dendrolaborg.wordpress.com/}
}
\references{
  Biondi, F. and Qeadan, F. (2008) A theory-driven approach to tree-ring
  standardization: Defining the biological trend from expected basal area
  increment.  \emph{Tree-Ring Research}, \bold{64}(2), 81\enc{–}{--}96.
}
\keyword{datasets}

\encoding{UTF-8}
\name{common.interval}
\alias{common.interval}
\title{Common Interval}
\description{
  This function finds the common interval on a set of tree-ring widths
  such as that produced by \code{\link{read.rwl}}.
}
\usage{
common.interval(rwl, type=c("series", "years", "both"),
                make.plot=TRUE)
}
\arguments{
  \item{rwl}{a \code{data.frame} of ring widths with
    \code{rownames(\var{x})} containing years and \code{colnames(\var{x})}
    containing each series \acronym{ID} such as produced by
    \code{\link{read.rwl}}}
  \item{type}{a \code{character} string of
    \code{"series"}, \code{"years"}, or \code{"both"}.  Argument
    matching is performed. }
  \item{make.plot}{a \code{logical} indicating if a plot should be drawn }
}
\details{

  This trims an \code{rwl} object to a common interval that maximizes
  the number of series (\code{type="series"}), the number of years
  (\code{type="years"}), or a compromise between the two
  (\code{type="both"}).  A modified \code{\link{seg.plot}} can be drawn
  as well.
  
}
\value{
  A \code{data.frame} with \code{colnames(\var{x})} and
  \code{rownames(\var{x})}.
}
\author{ Filipe Campelo, Andy Bunn and Mikko Korpela }
\seealso{ \code{\link{seg.plot}} }

\examples{
library(utils)
data(co021)
co021.s <- common.interval(co021, type="series", make.plot=TRUE)
co021.y <- common.interval(co021, type="years", make.plot=TRUE)
co021.b <- common.interval(co021, type="both", make.plot=TRUE)

dim(co021)
dim.s <- dim(co021.s)
dim.s       # the highest number of series
prod(dim.s) #   (33 series x 288 years = 9504)
dim.y <- dim(co021.y)
dim.y       # the highest number of years
prod(dim.y) #   (27 series x 458 years = 12366)
dim.b <- dim(co021.b)
dim.b       # compromise solution
prod(dim.b) #   (28 series x 435 years = 12180)
}
\keyword{ manip }

\encoding{UTF-8}
\name{rwl.report}
\alias{rwl.report}
\title{Do some reporting on a RWL object}
\description{
  This function generates a small report on a \code{rwl} (or \code{rwi}) object
  that gives the user some basic information on the data including the number of
  series, the span of the data, the mean interseries correlation, the number of
  missing rings (zeros), internal \code{NA} values, and rings that are very small,
  or very large.
}
\usage{
rwl.report(rwl,small.thresh=NA,big.thresh=NA)
}
\arguments{
  \item{rwl}{a \code{data.frame} of ring widths with
    \code{rownames(\var{x})} containing years and \code{colnames(x)}
    containing each series \acronym{ID} such as produced by
    \code{\link{read.rwl}}}
  \item{small.thresh}{a \code{numeric} value for the threshold value that will cause
    small rings to be listed. If values is \code{NA} this will be omitted.}
  \item{big.thresh}{a \code{numeric} value for the threshold value that will cause
    large rings to be listed. If values is \code{NA} this will be omitted.}
}
\details{
  This generates information about a \code{rwl} object including the number of series, the mean length of all the series, the first year, last year, the mean first-order  autocorrelation (via \code{\link{summary.rwl}}), the mean interseries  correlation (via \code{\link{interseries.cor}}), the years where a series has a missing ring (zero), internal NA, very small ring, very large rings, etc.  

  This output of this function is not typically meant for the user to access but has a \code{print} method for the user.
}
\value{
  A \code{list} with elements containing descriptive information on the \code{rwl} object. Specifically:
  \item{small.thresh}{a \code{numeric} value passed in.}
  \item{big.thresh}{a \code{numeric} value passed in.}
  \item{nSeries}{a \code{numeric} value with the number of series.}
  \item{n}{a \code{numeric} value with the total number of rings.}
  \item{meanSegLength}{a \code{numeric} with the mean segment length.}
  \item{firstYear}{a \code{numeric} with the first year.}
  \item{lastYear}{a \code{numeric} with the last year.}
  \item{meanAR1}{a \code{numeric} with the mean AR1 value of all series.}
  \item{sdAR1}{a \code{numeric} with the standard deviation of the AR1 values of all series.}
  \item{unconnected}{a \code{logical} indicating if there are rows (years) with all NA values.}
  \item{unconnectedYrs}{a \code{numeric} vector with unconnected years.}
  \item{nZeros}{a \code{numeric} with number of zeros in the data. }
  \item{zeros}{ a \code{list} containing series and years with zeros or a \code{numeric} of 0.}
  \item{allZeroYears}{a \code{numeric} indicating rows (years) with all zero values. }
  \item{consecutiveZeros}{ a \code{list} containing series and years with consecutive zeros or a \code{numeric} of 0. }
  \item{meanInterSeriesCor}{a \code{numeric} with the mean interseries correlation.}
  \item{sdInterSeriesCor}{a \code{numeric} with the standard deviation of the interseries correlations.}
  \item{internalNAs}{a \code{list} containing series and years with internal NA values or a \code{numeric} of 0.}
  \item{smallRings}{a \code{list} containing series and years with small rings or a \code{numeric} of 0.}
  \item{bigRings}{a \code{list} containing series and years with large rings or a \code{numeric} of 0.}

}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{read.rwl}}, \code{\link{summary.rwl}},
          \code{\link{interseries.cor}}
}
\examples{
data("gp.rwl")
rwl.report(rwl = gp.rwl)
# list very small (smallest 1pct) of rings as well
one.pct <- quantile(gp.rwl[gp.rwl != 0], na.rm=TRUE, probs=0.01)
rwl.report(rwl = gp.rwl, small.thresh = one.pct)
}
\keyword{ manip }

\encoding{UTF-8}
\name{xdate.floater}
\alias{xdate.floater}
\alias{print.floater}
\alias{plot.floater}
\title{Crossdate an undated series}
\description{
  Pulls an undated (or misdated) series through a dated \code{rwl} object
  in order to establish possible dates for the series.
}
\usage{
xdate.floater(rwl, series, series.name = "Unknown", min.overlap = 50,
              n = NULL, prewhiten = TRUE, biweight = TRUE,
              method = c("spearman", "pearson", "kendall"),
              make.plot = TRUE, return.rwl = TRUE, verbose = TRUE)

\method{print}{floater}(x, \dots)

\method{plot}{floater}(x, \dots)
}
\arguments{
  \item{rwl}{a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}}.}
  \item{series}{a numeric vector of ring widths, e.g. a single column
    from an \code{rwl} object.}
  \item{series.name}{a \code{character} string giving a name for the
    undated series. Defaults to \code{"Unknown"}.}
  \item{min.overlap}{a positive integer giving the minimum number of
    years of overlap required between the series and the master
    chronology at each search position. Defaults to \code{50}.}
  \item{n}{\code{NULL} or a positive integer giving the filter length
    for the \code{\link{hanning}} filter used for removal of low
    frequency variation prior to correlation. \code{NULL} skips the
    filter.}
  \item{prewhiten}{logical flag. If \code{TRUE} (the default) each
    series is whitened using \code{\link{ar}} prior to correlation.}
  \item{biweight}{logical flag. If \code{TRUE} (the default) a robust
    mean is used when building the master chronology via
    \code{\link{tbrm}}.}
  \item{method}{the correlation coefficient to use. One of
    \code{"spearman"} (default), \code{"pearson"}, or \code{"kendall"}.
    See \code{\link{cor.test}}.}
  \item{make.plot}{logical flag. If \code{TRUE} (the default) a
    two-panel plot is produced showing the segments and the correlation
    surface. See Details.}
  \item{return.rwl}{logical flag. If \code{TRUE} (the default) a
    \code{"floater"} list object is returned. If \code{FALSE} only the
    correlation statistics \code{data.frame} is returned.}
  \item{verbose}{logical flag. If \code{TRUE} (the default) a brief
    summary is printed to the console.}
  \item{x}{a \code{"floater"} object produced by
    \code{xdate.floater}.}
  \item{\dots}{additional arguments — currently ignored.}
}
\details{
  The undated series is slid along the master chronology built from
  \code{rwl} (using the leave-one-out principle) and the correlation
  between the series and the master is computed at each position with
  sufficient overlap. The position with the highest correlation gives
  the proposed dates.

  Both series and master are optionally prewhitened and/or smoothed
  with a Hanning filter before correlation, consistent with the
  approach used in \code{\link{corr.series.seg}}.

  The two-panel plot (produced when \code{make.plot = TRUE}, or by
  calling \code{plot()} on the returned object) shows: (1) a segment
  plot of the master series with the floater at its best-fit position
  highlighted in green; and (2) the correlation at each end-year
  searched, overlaid with a dashed significance line and a light-blue
  polygon showing the 5th--95th percentile band of the typical
  interseries correlation in the master chronology. The dark-blue
  horizontal line shows the median interseries correlation. The
  best-fit position is marked with green points and a dashed segment.
}
\value{
  If \code{return.rwl = TRUE} (the default), an object of class
  \code{"floater"} is returned. This is a named list with the
  following elements:
  \item{series.name}{the name given to the undated series.}
  \item{floaterCorStats}{a \code{data.frame} with columns \code{first},
    \code{last}, \code{r}, \code{p}, and \code{n} giving the start year,
    end year, correlation, p-value, and sample size at each search
    position.}
  \item{rwlCombined}{an \code{rwl} object containing the dated master
    series combined with the undated series placed at its best-fit
    dates.}
  \item{rwlOut}{an \code{rwl} object containing only the undated series
    at its best-fit dates.}

  If \code{return.rwl = FALSE}, only the \code{floaterCorStats}
  \code{data.frame} is returned.
}
\note{
  This function is experimental and may change in future releases.
  Users should always verify proposed dates visually and against the
  physical wood.

  The choice of \code{min.overlap} affects both which search positions
  are evaluated and the reliability of the correlations at those
  positions. A short \code{min.overlap} allows the series to be placed
  near the edges of the master chronology where little overlap exists,
  producing correlations based on few observations that may be
  spuriously high. A very long \code{min.overlap} restricts the search
  to positions where the series is well inside the master, which is
  statistically conservative but can cause the function to miss the
  correct dates entirely if the true position lies near an edge, or to
  produce wrong dates if the best correlation within the constrained
  search window happens to be a false match. As a rule,
  \code{min.overlap} should be long enough to produce a stable
  correlation (50 years is a common starting point for annual
  tree-ring data) but no longer than roughly half the length of the
  undated series. The dashed significance line in the plot is
  calculated from the overlap length at each position and provides a
  useful guide: positions with short overlaps will have high
  significance thresholds and should be interpreted with caution.
}
\author{Andy Bunn. Patched and improved by Mikko Korpela.}
\seealso{
  \code{\link{corr.series.seg}}, \code{\link{ccf.series.rwl}},
  \code{\link{skel.plot}}, \code{\link{series.rwl.plot}}
}
\examples{
library(utils)
data(co021)
summary(co021)

# Remove a series and try to recover its dates
foo <- co021[, "645232"]
bar <- co021
bar$"645232" <- NULL
out <- xdate.floater(bar, foo, min.overlap = 50, series.name = "645232")
print(out)

# A longer series. With min.overlap = 100 the correct dates are recovered.
# With min.overlap = 200 the search window is so constrained that the true
# position is excluded and the function returns the best match within the
# remaining positions, which is a false fit. Compare the two results.
foo <- co021[, "646118"]
bar <- co021
bar$"646118" <- NULL
out <- xdate.floater(bar, foo, min.overlap = 100, series.name = "646118")
out <- xdate.floater(bar, foo, min.overlap = 200, series.name = "646118")
}
\keyword{manip}

\encoding{UTF-8}
\name{zof.anc}
\docType{data}
\alias{zof.anc}
\title{ Ancillary Data Corresponding to \code{\link{zof.rwl}} }
\description{
  This data set gives the pith offsets, distance to pith, and diameter that match the ring widths for \code{\link{zof.rwl}} \enc{–}{-} a data set of European Beech (\emph{Fagus sylvatica}) increment cores collected at the Zofingen in Switzerland.
}
\usage{data(zof.anc)}
\format{
  A \code{data.frame} containing four columns. Column one gives the series ID. Column two (\code{PO}) gives the number of rings estimated to be missing to the pith. Column three (\code{d2pith}) gives the estimated distance to the pith (mm). Column four (\code{diam}) gives the diameter at breast height (DBH) in cm.
}
\source{
  Contributed by Stefan Klesse
}
\references{
  Klesse, S., Babst, F., Lienert, S., Spahni, R., Joos, F., Bouriaud, O., Carrer, M., Filippo, A.D., Poulter, B., Trotsiuk, V., Wilson, R., Frank, D.C. (2018) A combined tree ring and vegetation model assessment of European forest growth sensitivity to interannual climate variability. \emph{Global Biogeochemical Cycles} 32, 1226 -- 1240.
}
\keyword{datasets}

\encoding{UTF-8}
\name{powt}
\alias{powt}
\title{Power Transformation of Tree-Ring Data}
\description{
  Power transformation of tree-ring width.
}
\usage{
powt(rwl, method = "universal", rescale = FALSE,
     return.power=FALSE)
}
\arguments{
  \item{rwl}{a \code{data.frame}, typically of raw tree-ring widths series, such 
  as that produced by \code{\link{read.rwl}} or \code{\link{read.fh}}}
    
  \item{method}{ a \code{character} vector to determine the transform
    method.  See details below.  Possible values are either "universal" or 
    "cook." Pattern matching is used.}

  \item{rescale}{ \code{logical} flag.  If \code{TRUE} then each transformed 
  series is rescaled to have the orginal mean and standard deviation of the 
  input data. See details. }

  \item{return.power}{ \code{logical} flag.  If \code{TRUE} then the power 
  estimate(s) is returned. }
}
\value{
  Either an object of class \code{c("rwl", "data.frame")} containing the 
  power transformed ring width series with the series in columns and the years 
  as rows or in the case of a single series, a possibly named vector of the same. 
  With class \code{rwl}, the series \acronym{ID}s are the column names and the 
  years are the row names.
  
  If \code{return.power=TRUE} the returned 
  object is a \code{list} containing the power transformed data and a 
  \code{numeric} with the power estimate(s) used to transform the data.
  
}
\details{
  In dendrochronology, ring width series are sometimes power transformed to 
  address heteroscedasticity. 
  
  The classic procedure used by \code{method="cook"} 
  is a variance stabilization technique implemented after
  Cook & Peters (1997): for each series a linear model is fitted on the
  logs of level and spread, where level is defined as the local mean
  \eqn{M_t = \left(R_t + R_{t-1}\right)/2}{M_t = (R_t + R_(t-1))/2} with
  ring widths R, and spread S is the local standard deviation defined as
  \eqn{S_t = \left|R_t - R_{t-1}\right|}{S_t = |R_t - R_(t-1)|}.  The
  regression coefficient b from a linear model  
  \eqn{\log S = k + b \log M}{log S = k + b log M} is then used for the 
  power transform \eqn{\star{R}_t = R_t^{1-b}}{R*_t = R_t^(1-b)}.

  The procedure above is modified with \code{method="universal"} where all samples 
  are used simultaneously in a linear mixed-effects model with time (year)
  as a random effect: \code{lmer(log S ~ log M + (1|year)}. This "universal" or 
  "signal free" approach accounts for the common year effect across all of the 
  series in \code{rwl} and should address that not every year has the same 
  change in environmental conditions to the previous year. 
  
  The \code{rescale} argument will return the series with a mean and standard
  deviation that matches the input data. While this is a common convention,
  users should note that this can produce negative values which can be confusing
  if thought of as "ring widths."

}
\references{
  Cook, E. R. and Peters, K. (1997) Calculating unbiased
  tree-ring indices for the study of climatic and environmental
  change.  \emph{The Holocene}, \bold{7}(3), 361\enc{–}{--}370.
}

\author{
  Christian Zang implemented the Cook and Peters method. Stefan Klesse conceived
  and wrote the universal method. Patched and improved by Mikko Korpela and 
  Andy Bunn.
}

\examples{library(utils)
data(zof.rwl)
powtUniversal <- powt(zof.rwl, method = "universal")
powtCook <- powt(zof.rwl, method = "cook")

op <- par(no.readonly = TRUE)
par(mfcol = c(1, 3))
hist(summary(zof.rwl)$skew,
     breaks = seq(-2.25,2.25,by=0.25),
     main="Raw Data",xlab="Skew")
hist(summary(powtUniversal)$skew,
     breaks = seq(-2.25,2.25,by=0.25),
     main="Universal POWT",xlab="Skew")
hist(summary(powtCook)$skew,
     breaks = seq(-2.25,2.25,by=0.25),
     main="Cook POWT",xlab="Skew")
par(op) # restore graphical parameters
}
\keyword{manip}

\encoding{UTF-8}
\name{write.tridas}
\alias{write.tridas}

\title{ Write Tree Ring Data Standard (TRiDaS) file }

\description{ This function writes measured or derived (standardized,
  averaged) series of values to a TRiDaS format file.  Some metadata are
  also supported. }

\usage{
write.tridas(rwl.df = NULL, fname, crn = NULL, prec = NULL, ids = NULL,
    titles = NULL, crn.types = NULL, crn.titles = NULL,
    crn.units = NULL, tridas.measuring.method = NA,
    other.measuring.method = "unknown", sample.type = "core",
    wood.completeness = NULL, taxon = "",
    tridas.variable = "ring width", other.variable = NA,
    project.info = list(type = c("unknown"), description = NULL,
                        title = "", category = "", investigator = "",
                        period = ""),
    lab.info = data.frame(name = "", acronym = NA, identifier = NA,
                          domain = "", addressLine1 = NA,
                          addressLine2 = NA, cityOrTown = NA,
                          stateProvinceRegion = NA, postalCode = NA,
                          country = NA),
    research.info = data.frame(identifier = NULL, domain = NULL,
                               description = NULL),
    site.info = list(type = "unknown", description = NULL, title = ""),
    random.identifiers = FALSE, identifier.domain = lab.info$name[1],
    \dots)
}

\arguments{

  \item{rwl.df}{ \code{data.frame} containing tree-ring ring widths in
    millimetres with the series in columns and the years as rows.  The
    series \acronym{ID}s are the column names and the years are the row
    names.  This type of \code{data.frame} is produced by
    \code{\link{read.tucson}}, \code{\link{read.compact}} and
    \code{\link{read.tridas}}.  Defaults to \code{NULL}: no measurement
    series are written. }

  \item{fname}{ \code{character} vector giving the file name of the rwl
    file }

  \item{crn}{ \code{data.frame} or a list of \code{data.frame}s
    containing tree-ring chronologies.  Accepts \code{data.frame}s of
    the type produced by \code{\link{chron}}.  Additionally, allows
    several chronologies per \code{data.frame}.  Any column of the
    \code{data.frame}(s) with a name starting with \code{"samp.depth"}
    is interpreted as a sample depth.  The rest of the columns are
    interpreted as chronologies whose titles are determined from the
    column names (optionally from parameter \code{\var{crn.titles}}).
    Chronology columns and sample depth columns are paired in order so
    that the first chronology gets the first sample depth column, second
    chronology gets the second set of sample depths, etc.  If there are
    less sample depth columns than chronologies, the sample depths are
    recycled.  Defaults to \code{NULL}: no chronologies are written. }

  \item{prec}{ optional \code{numeric} indicating the rounding precision
    of the output file when writing the data contained in
    \code{\var{rwl.df}}.  Defaults to \code{NULL}: no rounding is done
    and the measurements are written in (non-integer) millimetres.
    Possible \code{numeric} values are 0.001, 0.01, 0.05, 0.1, 1, 10,
    100 and 1000, which cause the data to be transformed to micrometres,
    1/100th millimetres, 1/20th millimetres, 1/10 millimetres,
    (millimetres), centimetres, decimetres or metres, respectively, and
    then rounded to integral values.  Data rounded to decimetres are
    written in centimetres (values always ending in zero).  Otherwise,
    the matching unit is used in the file. }

  \item{ids}{ optional \code{data.frame} with column one named
    \code{"tree"} giving the \code{numeric} \acronym{ID} of the tree,
    column two named \code{"core"} giving the \code{numeric}
    \acronym{ID} of the core, optional column three named
    \code{"radius"} giving the \code{numeric} \acronym{ID} of the
    radius, and optional column four named \code{"measurement"} giving
    the \code{numeric} \acronym{ID} of the measurement.  If column
    \code{"measurement"} exists, column \code{"radius"} must also exist.
    Defaults to one core, radius and measurement per tree:\preformatted{
  data.frame(tree=1:n.col, core=rep(1,n.col),
             radius=rep(1,n.col), measurement=rep(1,n.col))}
    where \code{\var{n.col}} is the number of columns in
    \code{\var{rwl.df}}. }

  \item{titles}{ optional \code{data.frame} with column one named
    \code{"tree"} giving the title of the tree, column two named
    \code{"core"} giving the title of the core, column three named
    \code{"radius"} giving the title of the radius, and column four
    named \code{"measurement"} giving the title of the measurement.  By
    default, \code{\var{titles}} is \code{NULL}, and the title hierarchy
    is automatically created out of the column names of
    \code{\var{rwl.df}}, taking \code{\var{ids}} into account. }

  \item{crn.types}{ \code{character} vector or a \code{list} of
    \code{character} vectors giving the types of the derived series in
    \code{\var{crn}}.  A single vector is interpreted as one type per
    \code{data.frame} in \code{\var{crn}}, recycled if not long enough.
    A list of vectors is interpreted as one list element per
    \code{data.frame}.  In this case, the list is recycled to the
    correct length.  After that, the vectors inside the list are
    recycled to match the number of derived series in each
    \code{data.frame} of \code{\var{crn}}.  The default is to write
    empty \samp{<type>} elements. }

  \item{crn.titles}{ optional \code{character} vector or a \code{list}
    of \code{character} vectors giving the titles of the derived series
    in \code{\var{crn}}.  The interpretation is the same as with
    \code{\var{crn.types}}, except that the default is to derive the
    titles from the column names of \code{\var{crn}}.  Also \code{NA}
    means that the column name is used.  }

  \item{crn.units}{ optional \code{character} vector or a \code{list} of
    \code{character} vectors giving the units of the derived series in
    \code{\var{crn}}.  The interpretation is the same as with
    \code{\var{crn.types}}, except that the default is to mark the
    series as \samp{<unitless>}.  Also \code{NA} means
    \samp{<unitless>}.  }

  \item{tridas.measuring.method}{ \code{character} vector giving the
    measuring method used to acquire each series of \code{\var{rwl.df}}.
    Partial matching is used to replace these with the complete terms in
    \code{\link{tridas.vocabulary}}.  If the vector is shorter than the
    number of columns in \code{\var{rwl.df}}, it is recycled to the
    correct length.  The default is to use the information in
    \code{\var{other.measuring.method}} instead.  Also, \code{NA} in any
    position of the vector means that the measuring method information
    for that series is looked up in \code{\var{other.measuring.method}}.
    }

  \item{other.measuring.method}{ \code{character} vector giving the
    measuring method used to acquire each series of \code{\var{rwl.df}}.
    In contrast to \code{\var{tridas.measuring.method}}, these are
    free-form strings in English.  If the vector is shorter than the
    number of columns in \code{\var{rwl.df}}, it is recycled to the
    correct length.  The default value is \code{"unknown"}.  }

  \item{sample.type}{ optional \code{character} vector giving the type
    of the samples, corresponding to \code{"core"} in \code{\var{ids}}.
    The length of the vector, however, must match the number of columns
    in \code{\var{rwl.df}}, or it is recycled to the correct length.  If
    there are several measurements per sample, some elements of
    \code{\var{sample.type}} are redundant.  The default is to use
    \code{"core"} for all series.  }

  \item{wood.completeness}{ optional \code{data.frame} giving wood
    completeness information for the measurement series in
    \code{\var{rwl.df}}.  The number of rows must match the number of
    columns in \code{\var{rwl.df}}.  The columns are expected to be a
    subset of the following (descriptions almost directly quoted from
    TRiDaS specification):

    \describe{

      \item{n.unmeasured.inner}{Field for recording whether there are
        any rings at the inner (i.e. towards pith) edge of the sample
        that have not been measured.  Typically used to note when rings
        are too damaged to measure.  Non-negative integral value.}

      \item{n.unmeasured.outer}{Field for recording whether there are
        any rings at the outer (i.e. towards bark) edge of the sample
        that have not been measured.  Typically used to note when rings
        are too damaged to measure.  Non-negative integral value.}

      \item{pith.presence}{Whether the pith is present or absent.
        Each element must be a partial match with the contents of
        category \code{"complex presence / absence"} in
        \code{\link{tridas.vocabulary}}.}

      \item{heartwood.presence}{Whether the outer (youngest) heartwood
        is present and if so whether it is complete.

        Category \code{"complex presence / absence"} in
        \code{\link{tridas.vocabulary}}.}

      \item{n.missing.heartwood}{Estimated number of missing heartwood
        rings to the pith.  Non-negative integral value.}

      \item{missing.heartwood.foundation}{Description of the way the
        estimation of how many heartwood rings are missing was made and
        what the certainty is.  Free-form string.}

      \item{sapwood.presence}{Whether the sapwood is present or not.

        Category \code{"complex presence / absence"}.}

      \item{n.sapwood}{Number of sapwood rings measured.  Non-negative
        integral value.}

      \item{last.ring.presence}{Last ring under the bark is present or
        absent.

        Category \code{"presence / absence"}.}

      \item{last.ring.details}{If the last ring under the bark is
        present, include information about the completeness of this ring
        and/or season of felling.  Free-form string.}

      \item{n.missing.sapwood}{Estimated number of missing sapwood rings
        to the bark.  Non-negative integral value.}

      \item{missing.sapwood.foundation}{Description of the way the
        estimation of how many sapwood rings are missing was made and
        what the certainty is.  Free-form string.}

      \item{bark.presence}{Bark is present or absent.

        Category \code{"presence / absence"} in
        \code{\link{tridas.vocabulary}}.}

    }
  }

  \item{taxon}{ \code{character} string.  The most detailed taxonomic
    name known for this element (species, genus, family etc).
    Preferably from the \href{https://www.catalogueoflife.org/}{Catalogue
      of Life} controlled vocabulary.  The same string is used for all of
    \code{\var{rwl.df}}.  The default value is an empty string, but a
    proper value should really be given.  }

  \item{tridas.variable}{ \code{character} string.  Measured variable
    (ring width, earlywood, latewood etc) taken from the TRiDaS
    controlled vocabulary (\code{\link{tridas.vocabulary}}, category
    \code{"variable"}).  The same string is used for all of
    \code{\var{rwl.df}}.  Defaults to \code{"ring width"}.  }

  \item{other.variable}{ \code{character} string.  Measured variable as
    a free-form string.  The same string is used for all of
    \code{\var{rwl.df}}.  This is only used if
    \code{\var{tridas.variable}} is \code{NA}.  }

  \item{project.info}{ \code{list} containing information about the
    project.  Elements are the following (includes quotes from the
    TRiDaS specification):

    \describe{

      \item{type}{\code{character} vector.  The type(s) of the project.
        Defaults to \code{"unknown"}.}

      \item{description}{\code{character} string.  A description of the
        project.  Defaults to \code{NULL}: no description.}

      \item{title}{\code{character} string.  The title of the project.
        Defaults to an empty string.}

      \item{category}{\code{character} string.  Category of research
        this project falls into.  Defaults to an empty string.}

      \item{investigator}{\code{character} string.  Principal
        investigator of this project.  Defaults to an empty string.}

      \item{period}{\code{character} string.  When the
        dendrochronological project took place.  Could consist of a
        start- and end-date.  If unknown it should be estimated.
        Defaults to an empty string.}

    }
  }

  \item{lab.info}{ \code{data.frame}. Information about the
    dendrochronological research laboratories where this work was done.
    One row per laboratory.  Defaults to one laboratory with an empty
    name and no other information.  The columns are expected to be a
    subset of the following:

    \describe{

      \item{name}{Name of the laboratory}

      \item{acronym}{Optional acronym of the laboratory}

      \item{identifier}{Optional identifier of the laboratory}

      \item{domain}{The domain which the identifier of the laboratory is
        applicable to.  Could be the \acronym{URL} of the
        organization\enc{’}{'}s server or the name of the organization
        as long as it is not ambiguous.}

      \item{addressLine1}{First address line}

      \item{addressLine2}{Second address line}

      \item{cityOrTown}{City or town}

      \item{stateProvinceRegion}{State, province or region}

      \item{postalCode}{Postal code.  Beware of ignored leading zeros
        if these are given in \code{numeric} or \code{integer} type
        values.  It is always safe to use \code{character} values.}

      \item{country}{Country}

    }
  }

  \item{research.info}{ optional \code{data.frame} with information
    about the systems in which the research project is
    registered.  Columns are the following:

    \describe{

      \item{identifier}{Identifier}

      \item{domain}{Domain which the identifier is applicable to}

      \item{description}{General description}

    }
  }

  \item{site.info}{ \code{list} containing information about the site
    (\samp{<object>}).  Elements are the following, and all are
    \code{character} strings:

    \describe{

      \item{type}{Type of the site.  Defaults to \code{"unknown"}.}

      \item{description}{Description.  Defaults to no description.}

      \item{title}{Title of the site.  Defaults to an empty string.}

    }
  }

  \item{random.identifiers}{ \code{logical} flag.  If \code{TRUE},
    unique random identifiers are created with \code{\link{uuid.gen}}
    and attached to each \samp{<project>} (one in the file),
    \samp{object} (site, one in the file), \samp{<element>} (tree),
    \samp{<sample>} (core), \samp{<radius>}, \samp{<measurementSeries>}
    (measurement) and \samp{<derivedSeries>} element in the resulting
    TRiDaS file.  }

  \item{identifier.domain}{ \code{character} string.  The domain which
    the random identifiers are applicable to.  Could be the
    \acronym{URL} of the organization\enc{’}{'}s server or the name of the
    organization as long as it is not ambiguous.  Defaults to the name
    of the first laboratory in \code{\var{lab.info}}.  }

  \item{\dots}{ Unknown arguments are accepted but not used. }

}

\details{
  The Tree Ring Data Standard (TRiDaS) is described in Jansma et. al
  (2010).  
}

\value{ \code{\var{fname}} }

\references{
  Jansma, E., Brewer, P. W., and Zandhuis, I. (2010) TRiDaS 1.1: The
  tree-ring data standard.  \emph{Dendrochronologia}, \bold{28}(2),
  99--130.

}

\author{ Mikko Korpela }

\note{ This is an early version of the function.  Bugs are likely to
  exist, and parameters are subject to change.  }

\seealso{ \code{\link{write.rwl}}, \code{\link{write.tucson}},
  \code{\link{write.compact}}, \code{\link{write.crn}},
  \code{\link{read.tridas}} }

\examples{library(utils)
\dontrun{
## Write raw ring widths
data(co021)
fname1 <- write.tridas(rwl.df = co021,
    fname = tempfile(fileext=".xml"), prec = 0.01,
    site.info = list(title = "Schulman old tree no. 1, Mesa Verde",
                     type = "unknown"),
    taxon = "Pseudotsuga menziesii var. menziesii (Mirb.) Franco",
    project.info = list(investigator = "E. Schulman",
                        title = "", category = "",
                        period = "", type = "unknown"))
print(fname1) # tempfile used for output

## Write mean value chronology of detrended ring widths
data(ca533)
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
ca533.crn <- chron(ca533.rwi, prewhiten = TRUE)
fname2 <- write.tridas(crn = ca533.crn,
    fname = tempfile(fileext=".xml"),
    taxon = "Pinus longaeva D.K. Bailey",
    project.info =
        list(investigator = "Donald A. Graybill, V.C. LaMarche, Jr.",
             title = "Campito Mountain", category = "",
             period = "", type = "unknown"))
print(fname2) # tempfile used for output

unlink(c(fname1, fname2)) # remove the files
}
}
\keyword{ IO }

\encoding{UTF-8}
\name{cana157}
\docType{data}
\alias{cana157}
\title{Twisted Tree Heartrot Hill Standard Chronology}
\description{
  This data set gives the standard chronology for white spruce
  \emph{Picea glauca} at Twisted Tree Heartrot Hill in Yukon,
  Canada.  Data set was created using \code{\link{read.crn}} and saved to
  an .rda file using \code{\link{save}}.
}
\usage{data(cana157)}
\format{
  A \code{data.frame} containing the standard chronology in column one
  and the sample depth in column two.  There are 463 years
  (1530\enc{–}{--}1992) in the rows.
}
\source{International tree-ring data bank, Accessed on 20-April-2021 at
  \url{https://www.ncei.noaa.gov/pub/data/paleo/treering/chronologies/northamerica/canada/cana157.crn}
}
\references{
  Jacoby, G., D\enc{’}{'}Arrigo, R. and Buckley, B. (1992) Twisted Tree
  Heartrot Hill Data Set.  \acronym{IGBP} \acronym{PAGES}/World Data
  Center for Paleoclimatology Data Contribution Series 1992-CANA157.CRN.
  \acronym{NOAA}/\acronym{NCDC} Paleoclimatology Program, Boulder,
  Colorado, \acronym{USA}.
}
\keyword{datasets}

\encoding{UTF-8}
\name{insert.ring}
\alias{insert.ring}
\alias{delete.ring}
\title{ Edit a Ring-Width Series }
\description{
  Insert or delete rings from a ring-width series
}
\usage{
insert.ring(rw.vec,rw.vec.yrs=as.numeric(names(rw.vec)),
            year,ring.value=mean(rw.vec,na.rm=TRUE),
            fix.last=TRUE,fix.length=TRUE)
delete.ring(rw.vec,rw.vec.yrs=as.numeric(names(rw.vec)),
            year,fix.last=TRUE,fix.length=TRUE)
}
\arguments{
  \item{rw.vec}{ a vector of data }
  \item{rw.vec.yrs}{ the years for \code{rw.vec} as \code{names} }
  \item{year}{ the year to add or delete }
  \item{ring.value}{ the value to add }
  \item{fix.last}{ logical.  If TRUE the last year of the series
    is fixed and the first year changes.}
  \item{fix.length}{ logical.  If TRUE the length of the output will be the 
    length of the input.}

}
\details{
  Simple editing of ring widths.
}
\value{
  A named vector.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{dplR}}
}
\examples{library(utils)
data(gp.rwl)
dat <- gp.rwl
# insert a value of zero for the year 1950 in series 50A
# fix the last year of growth and maintain the length of the series
tmp <- insert.ring(rw.vec=dat$"50A",rw.vec.yrs = time(dat),
                   year=1950,ring.value=0,fix.length = TRUE)
# with fix.length=TRUE this can be merged back into the rwl object:
data.frame(dat$"50A",tmp)
dat$"50A" <- tmp

# note that if fix.last = FALSE and fix.length = FALSE inserting a ring causes the
# ending year of the series to be pushed forward and the length of the output to
# be longer than the original series.
tmp <- insert.ring(rw.vec=dat$"50A",rw.vec.yrs = time(dat),
                   year=1950,ring.value=0, fix.last = FALSE, 
                   fix.length = FALSE)
# with fix.length=FALSE this can't be merged back into the rwl object the
# same way as above
tail(tmp)
length(tmp)
nrow(dat)

# the same logic applies to deleting rings.
dat <- gp.rwl
# delete the year 1950 in series 50A
# fix the last year of growth and maintain the length of the series
tmp <- delete.ring(rw.vec=dat$"50A",rw.vec.yrs = time(dat),
                   year=1950,fix.last = TRUE, fix.length = TRUE)
# with fix.length=TRUE this can be merged back into the rwl object:
data.frame(dat$"50A",tmp)
dat$"50A" <- tmp

# note that if fix.last = FALSE and fix.length = FALSE inserting a ring causes the
# ending year of the series to be pushed forward and the length of the output to
# be longer than the original series.
tmp <- delete.ring(rw.vec=dat$"50A", rw.vec.yrs = time(dat),
                   year=1950, fix.last = FALSE, 
                   fix.length = FALSE)
# with fix.length=FALSE this can't be merged back into the rwl object the
# same way as above
tail(tmp)
length(tmp)
nrow(dat)
}
\keyword{ manip }

\encoding{UTF-8}
\name{sss}
\alias{sss}
\title{ Subsample Signal Strength }
\description{
  Calculate subsample signal strength on a
  \code{data.frame} of (usually) ring-width indices.
}
\usage{
sss(rwi, ids = NULL)
}
\arguments{

  \item{rwi}{ a \code{data.frame} with detrended and standardized ring
    width indices as columns and years as rows such as that produced by
    \code{\link{detrend}}. }

  \item{ids}{ an optional \code{data.frame} with column one named
    \code{"tree"} giving a \code{numeric} \acronym{ID} for each tree and
    column two named \code{"core"} giving a \code{numeric} \acronym{ID}
    for each core.  Defaults to one core per tree as\cr
    \code{data.frame(tree=1:ncol(\var{rwi}), core=rep(1, ncol(\var{rwi})))}. }
}
\details{
  This calculates subsample signal strength (sss) following equation 3.50 in 
  Cook and Kairiukstis (1990) but using notation from Buras (2017) because 
  writing the prime unicode symbol seems too difficult. The function 
  calls \code{\link{rwi.stats}} and passes it the arguments \code{ids} 
  and \code{prewhiten}. 
  
  To make better use of variation in growth within and between series, an 
  appropriate mask (parameter \code{\var{ids}}) should be provided that 
  identifies each series with a tree as it is common for dendrochronologists 
  to take more than one core per tree.  The function \code{\link{read.ids}} is 
  helpful for creating a mask based on the series \acronym{ID}.

  Subsample signal strength is calculated as \eqn{\frac{n[1+(N-1)\bar{r}]}{N[1+(n-1)\bar{r}]}}{n*(1+(N-1)*rbar) / N*(1+(n-1)*rbar)} 
  where \code{n} and \code{N} are the number of cores or trees in the 
  subsample and sample respectively and \code{rbar} is mean interseries 
  correlation. If there is only one core per tree \code{n} is the sample 
  depth in a given year (\code{rowSums(!is.na(rwi))}), \code{N} is the 
  number of cores (\code{n.cores} as given by \code{\link{rwi.stats}}), 
  and \code{rbar} is the mean interseries correlation between all series 
  (\code{r.bt} as given by \code{\link{rwi.stats}}). If there are multiple 
  cores per tree \code{n} is the number of trees present in a given year, 
  \code{N} is the number of trees (\code{n.trees} as given by 
  \code{\link{rwi.stats}}), and \code{rbar} is the effective mean interseries 
  correlation (\code{r.eff} as given by \code{\link{rwi.stats}}).

  Readers interested in the differences between subsample signal strength and 
  the more commonly used (running) expressed population signal should look at 
  Buras (2017) on the frequently mis-used citation of Wigley et al. (1984) 
  of the expressed population signal threshold EPS=0.85 as well 
  as Cook and Pederson (2011) for a more general approach to categorizing 
  variability in tree-ring data.

}

\value{ A \code{numeric} containing the subsample signal strength that is 
the same as number if rows of\code{rwi}. 
}
\references{

  Buras, A. (2017) A comment on the Expressed Population Signal. 
  Dendrochronologia 44:130-132. 
  
  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Cook, E. R. and Pederson, N. (2011) Uncertainty, Emergence, and
  Statistics in Dendrochronology.  In Hughes, M. K., Swetnam, T. W., and
  Diaz, H. F., editors, \emph{Dendroclimatology: Progress and
  Prospects}, pages 77\enc{–}{--}112.  Springer.  \acronym{ISBN-13}:
  978-1-4020-4010-8.
  
  Wigley, T. M., Briffa, K. R. and Jones, P. D. (1984) On the average 
  value of correlated time series, with applications in dendroclimatology 
  and hydrometeorology. Journal of Applied Meteorology and Climatology 
  23: 201-213.

}

\author{Andy Bunn. Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{rwi.stats}}, \code{\link{read.ids}}  } 
\examples{
data(ca533)
ca533.rwi <- detrend(ca533,method="Spline")
# assuming 1 core / tree
ca533.sss <- sss(ca533.rwi)

ca533.ids <- autoread.ids(ca533)
# done properly with >=1 core / tree as per the ids
ca533.sss2 <- sss(ca533.rwi,ca533.ids)

yr <- time(ca533)
plot(yr,ca533.sss,type="l",ylim=c(0.4,1),
     col="darkblue",lwd=2,xlab="Year",ylab="SSS")
lines(yr,ca533.sss2,lty="dashed",
      col="darkgreen",lwd=2)

# Plot the chronology showing a potential cutoff year based on SSS
# (using sss2 with the correct series IDs to get >=1 core / tree as per the ids)
ca533.crn <- chron(ca533.rwi)
def.par <- par(no.readonly=TRUE)
par(mar = c(2, 2, 2, 2), mgp = c(1.1, 0.1, 0), tcl = 0.25, xaxs='i')
plot(yr, ca533.crn[, 1], type = "n", xlab = "Year",
     ylab = "RWI", axes=FALSE)
cutoff <- max(yr[ca533.sss2 < 0.85])
xx <- c(500, 500, cutoff, cutoff)
yy <- c(-1, 3, 3, -1)
polygon(xx, yy, col = "grey80")
abline(h = 1, lwd = 1.5)
lines(yr, ca533.crn[, 1], col = "grey50")
lines(yr, caps(ca533.crn[, 1], nyrs = 32), col = "red", lwd = 2)
axis(1); axis(2); axis(3);
par(new = TRUE)
## Add SSS
plot(yr, ca533.sss2, type = "l", xlab = "", ylab = "",
     axes = FALSE, col = "blue")
abline(h=0.85,col="blue",lty="dashed")
axis(4, at = pretty(ca533.sss2))
mtext("SSS", side = 4, line = 1.1, lwd=1.5)
box()
par(def.par)

}
\keyword{ misc }

\encoding{UTF-8}
\name{hanning}
\alias{hanning}
\title{ Hanning Filter }
\description{
  Applies a Hanning filter of length \code{\var{n}} to \code{\var{x}}.
}
\usage{
hanning(x, n = 7)
}
\arguments{
  \item{x}{ a vector }
  \item{n}{ length of the Hanning filter, defaults to 7 }
}
\details{
  This applies a low frequency Hanning (a.k.a. Hann) filter to
  \code{\var{x}} with weight set to \code{\var{n}}.
}
\value{
  A filtered vector.
}
\references{
  Oppenheim, A. V., Schafer, R. W., and Buck, J. R. (1999) \emph{
    Discrete-Time Signal Processing}.  Prentice-Hall, 2nd edition.
  \acronym{ISBN-13}: 978-0-13-754920-7.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link[stats]{filter} } }

\examples{library(graphics)
library(utils)
data(ca533)
yrs <- time(ca533)
y <- ca533[, 1]
not.na <- !is.na(y)
yrs <- yrs[not.na]
y <- y[not.na]
plot(yrs, y, xlab = "Years", ylab = "Series1 (mm)",
     type = "l", col = "grey")
lines(yrs, hanning(y, n = 9), col = "red", lwd = 2)
lines(yrs, hanning(y, n = 21), col = "blue", lwd = 2)
legend("topright", c("Series", "n=9", "n=21"),
       fill=c("grey", "red", "blue"))
}
\keyword{ ts }

\encoding{UTF-8}
\name{net}
\alias{net}
\title{
  Calculate NET
}
\description{
  Computes the \eqn{\mathit{NET}}{NET} parameter for a set of tree-ring
  records or other time-series data.
}
\usage{
net(x, weights = c(v = 1, g = 1))
}
\arguments{
  \item{x}{
    A \code{matrix} or \code{data.frame} with at least two rows and two
    columns containing \code{numeric} data.  The rows should represent a
    sequence of sampling points with uniform intervals (e.g. a range of
    years), but this is not checked.  Each column is a time-series
    spanning either the whole time range or a part of it.
  }
  \item{weights}{
    A \code{numeric} vector with two elements.  Normally, variation
    (\code{"v"}) and \enc{Gegenläufigkeit}{Gegenlaeufigkeit}
    (\code{"g"}) contribute to NET with equal weight.  It is possible to
    use different weights by setting them here.  The names of the vector
    are matched to \code{c("v", "g")} (see \sQuote{Examples}).  If no
    names are given, the first element is the weight of variation.
  }
}
\details{

  This function computes the \eqn{\mathit{NET}}{NET} parameter (Esper et
  al., 2001).  The overall \eqn{\mathit{NET}}{NET} is an average of all
  (non-\code{NA}) yearly values \eqn{\mathit{NET_j}}{NET[j]}, which are
  computed as follows:

  \deqn{\mathit{NET_j}=v_j+(1-G_j)}{NET[j] = v[j] + (1-G[j])}

  The yearly variation \eqn{v_j}{v[j]} is the standard deviation of the
  measurements of a single year divided by their mean.
  \enc{Gegenläufigkeit}{Gegenlaeufigkeit} \eqn{1-G_j}{1-G[j]} is based
  on one definition of \enc{Gleichläufigkeit}{Gleichlaeufigkeit}
  \eqn{G_j}{G[j]}, similar to but not the same as what \code{\link{glk}}
  computes.  Particularly, in the formula used by this function (Esper
  et al., 2001), simultaneous zero differences in two series are not
  counted as a synchronous change.

  The weights of \eqn{v_j}{v[j]} and \eqn{1-G_j}{1-G[j]} in the sum can
  be adjusted with the argument \code{\var{weights}} (see above).  As a
  rather extreme example, it is possible to isolate variation or
  \enc{Gegenläufigkeit}{Gegenlaeufigkeit} by setting one of the weights
  to zero (see \sQuote{Examples}).

}
\value{

  A \code{list} with the following components, in the same order as
  described here:
  
  \item{all }{a \code{numeric} vector containing
    \eqn{\mathit{NET_j}}{NET[j]}.  Row names of \code{\var{x}} (if any)
    are copied here. }

  \item{average }{a \code{numeric} value \eqn{\mathit{NET}}{NET}, the
    average of the \code{"all"} vector (\code{NA} values removed). }
  
}
\references{

  Esper, J., Neuwirth, B., and Treydte, K. (2001) A new parameter to
  evaluate temporal signal strength of tree-ring chronologies.
  \emph{Dendrochronologia}, \bold{19}(1), 93\enc{–}{--}102.
  
}
\author{
  Mikko Korpela
}
\examples{library(utils)
data(ca533)
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
ca533.net <- net(ca533.rwi)
tail(ca533.net$all)
ca533.net$average
\dontrun{
## Isolate the components of NET
ca533.v <- net(ca533.rwi, weights=c(v=1,0))
ca533.g <- net(ca533.rwi, weights=c(g=1,0))
}
}
\keyword{ ts }

\encoding{UTF-8}
\name{sens1}
\alias{sens1}
\title{ Calculate Mean Sensitivity  }
\description{
  This function calculates mean sensitivity of a detrended ring-width
  series.
}
\usage{
  sens1(x)
}
\arguments{
  \item{x}{ a \code{numeric} vector containing the series }
}
\details{
  This calculates mean sensitivity according to Eq. 1 in Biondi and
  Qeadan (2008).  This is the standard measure of sensitivity in
  dendrochronology and is typically calculated on detrended series.
  However, note that mean sensitivity is not a robust statistic and 
  should rarely, if ever, be used (Bunn et al. 2013).

}
\value{ the mean sensitivity.
}
\references{

  Biondi, F. and Qeadan, F. (2008) Inequality in Paleorecords.
  \emph{Ecology}, \bold{89}(4), 1056\enc{–}{--}1067.

  Bunn, A. G., Jansma, E., Korpela, M., Westfall, R. D., and Baldwin,
  J. (2013) Using simulations and data to evaluate mean sensitivity
  (\eqn{\zeta}{zeta}) as a useful statistic in dendrochronology.
  \emph{Dendrochronologia}, \bold{31}(3), 250\enc{–}{--}254.
  
}
\author{ Mikko Korpela, based on original by Andy Bunn }
\seealso{ \code{\link{sens2}}, \code{\link{rwl.stats}} }
\examples{library(utils)
data(ca533)
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
sens1(ca533.rwi[, 1])
}
\keyword{ univar }

\encoding{UTF-8}
\name{bai.out}
\alias{bai.out}
\title{ Basal Area Increment (Outside In) }
\description{
  Convert multiple ring-width series to basal area increment (i.e., ring
  area) going from the bark to the pith.
}
\usage{
bai.out(rwl, diam = NULL)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}} }
  \item{diam}{ an optional \code{data.frame} containing two variables.
    If present, variable one (\code{\var{series}} in the example below)
    gives the series \acronym{ID} as either \code{character}s or
    \code{factor}s.  These must exactly match
    \code{colnames(\var{rwl})}.  Variable two (\code{\var{diam}} in the
    example below) gives the diameter of the tree (in mm) up to the
    outermost ring (e.g., the diameter of the tree where the core was
    taken minus the thickness of the bark).  If \code{\var{diam}} is
    \code{NULL} then the diameter is taken as twice the sum of the
    widths for each series (column) in \code{\var{rwl}}. }
}
\details{
  This converts ring-width series (mm) to ring-area series (mm squared)
  (aka basal area increments) based on the diameter of the tree and the
  width of each ring moving towards the pith of the tree.  It is related
  to \code{\link{bai.in}}, which calculates each ring area starting from
  the inside of the tree and working outward.  Both methods assume a
  circular cross section (Biondi 1999).  See the references below for
  further details.
}
\value{
  A \code{data.frame} containing the ring areas for each series with
  column names, row names and dimensions of \code{\var{rwl}}.
}
\references{
  Biondi, F. (1999) Comparing tree-ring chronologies and repeated timber
  inventories as forest monitoring tools.  \emph{Ecological
    Applications}, \bold{9}(1), 216\enc{–}{--}227.

  Biondi, F. and Qeadan, F. (2008) A theory-driven approach to tree-ring
  standardization: Defining the biological trend from expected basal area
  increment.  \emph{Tree-Ring Research}, \bold{64}(2), 81\enc{–}{--}96.
}
\note{ DendroLab website: \url{https://dendrolaborg.wordpress.com/} }
\author{ Code by Andy Bunn based on work from DendroLab, University of
  Nevada Reno, \acronym{USA}.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{bai.in}} }
\examples{
library(graphics)
library(utils)
\dontrun{
library(stats)
## Toy
n <- 100
## Make three fake tree-ring series to show that these funcs work on rwl objects
base.series <- 0.75 + exp(-0.2 * 1:n)
rwl <- data.frame(x1 = base.series + abs(rnorm(n, 0, 0.05)),
                  x2 = base.series + abs(rnorm(n, 0, 0.05)),
                  x3 = base.series + abs(rnorm(n, 0, 0.05)))

## The inside out method
foo <- bai.in(rwl = rwl)
## The outside in method
bar <- bai.out(rwl = rwl)

## Identical
head(bar)
head(foo)
}
## Use gp data
data(gp.rwl)
data(gp.dbh)
## dbh (minus the bark) from cm to mm 
gp.dbh2 <- gp.dbh[, 1:2]
gp.dbh2[, 2] <- (gp.dbh[, 2] - gp.dbh[, 3]) * 10
bar <- bai.out(rwl = gp.rwl, diam = gp.dbh2)
bar.crn <- chron(bar)
yrs <- time(bar.crn)
plot(yrs, bar.crn[, 1], type = "n",
     xlab = "Year", ylab = expression(mm^2))
lines(yrs, bar.crn[, 1], col = "grey", lty = "dashed")
lines(yrs, caps(bar.crn[, 1], nyrs = 32), col = "red", lwd = 2)
}
\keyword{ manip }

\encoding{UTF-8}
\name{chron}
\alias{chron}
\title{Build Mean Value Chronology}
\description{
  This function builds a mean value chronology, typically from a
  \code{data.frame} of detrended ring widths as produced by
  \code{\link{detrend}}.
}
\usage{
chron(x, biweight = TRUE, prewhiten = FALSE, \dots)
}
\arguments{
  \item{x}{a \code{data.frame} of (usually detrended) ring widths with
    \code{rownames(\var{x})} containing years and \code{colnames(x)}
    containing each series \acronym{ID} such as produced by
    \code{\link{read.rwl}}}
  \item{biweight}{\code{logical} flag.  If \acronym{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}.}}
  \item{prewhiten}{\code{logical} flag.  If \acronym{TRUE} each series is
    whitened using \code{\link{ar}} prior to averaging.}
  \item{\dots}{Arguments passed to \code{\link{ar}} when
    \code{prewhiten} is \code{TRUE}.  For example, use \code{aic}
    and \code{order.max} to control the order of the AR model.}
}
\details{
  This either averages the rows of the \code{data.frame} using a mean or
  a robust mean (the so-called standard chronology) or can do so from
  the residuals of an AR process (the residual chronology).
  
  Note that the residual chronology in this function will return different
  values than the residual chronology from \code{\link{chron.ars}} which uses
  a slightly different method for determining AR order.
}
\value{
  An object of of class \code{crn} and \code{data.frame} with the standard chronology, residual chronology (if prewhitening was performed), and the sample depth. The years are stored as row numbers.
}
\references{
  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Fritts, H. C. (2001) \emph{Tree Rings and Climate}.  Blackburn.
  \acronym{ISBN-13}: 978-1-930665-39-2.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{read.rwl}}, \code{\link{detrend}},
  \code{\link{ar}}, \code{\link{crn.plot}}
}
\examples{library(graphics)
library(utils)
data(ca533)
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
ca533.crn <- chron(ca533.rwi)
plot(ca533.crn,xlab="Year",ylab="RWI")
## With residual chron
ca533.crn2 <- chron(ca533.rwi, prewhiten = TRUE)
plot(ca533.crn2,xlab="Year",ylab="RWI")
}
\keyword{ manip }

\encoding{UTF-8}
\name{rwl.stats}
\alias{rwl.stats}
\alias{summary.rwl}
\title{ Calculate Descriptive Summary Statistics on Ring-Width Series }
\description{
  This function calculates descriptive statistics on a \code{rwl} object
  of raw or detrended ring-width series.
}
\usage{
rwl.stats(rwl)

\method{summary}{rwl}(object, ...)

}
\arguments{

  \item{rwl, object}{ a \code{rwl} object with (usually) raw ring-width series
  as columns and years as rows such as that produced by
  \code{\link{read.rwl}}.  It is sometimes desirable to run this on
  detrended (e.g., rwi) data. }

  \item{\dots}{ Additional arguments from the generic function.  These
    are silently ignored. }

}

\details{

  This calculates a variety of descriptive statistics commonly used in
  dendrochronology (see below).  Users unfamiliar with these should see
  Cook and Kairiukstis (1990) and Fritts (2001) for further details.

  The \code{\link{summary}} method for class \code{"rwl"} is a wrapper
  for \code{rwl.stats}.
  
}

\value{
  A \code{data.frame} containing descriptive stats on each
  \code{"series"}.  These are the first and last year of the series as
  well as the length of the series (\code{"first"}, \code{"last"},
  \code{"year"}).  The mean, median, standard deviation are given
  (\code{"mean"}, \code{"median"}, \code{"stdev"}) as are the skewness,
  the excess kurtosis (calculated as Pearson’s kurtosis minus 3), the Gini 
  coefficient, and first order
  autocorrelation (\code{"skew"}, \code{"kurtosis"}, \code{"\link{gini.coef}"}, 
  \code{"ar1"}).
  
  Note that prior to version 1.6.8, two measures of sensitivity were also included. However mean sensitivity is not a robust statistic that should rarely, if ever, be used (Bunn et al. 2013). Those sensitivity functions (\code{"\link{sens1}"} and \code{"\link{sens2}"}) are still available for continuity. Users should consider the coef of variation in lieu of mean sensitivity.
}
\references{

  Bunn, A. G., Jansma, E., Korpela, M., Westfall, R. D., and Baldwin,
  J. (2013) Using simulations and data to evaluate mean sensitivity
  (\eqn{\zeta}{zeta}) as a useful statistic in dendrochronology.
  \emph{Dendrochronologia}, \bold{31}(3), 250\enc{–}{--}254.

  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Fritts, H. C. (2001) \emph{Tree Rings and Climate}.  Blackburn.
  \acronym{ISBN-13}: 978-1-930665-39-2.

}
\author{ Andy Bunn.  Slightly improved by Mikko Korpela. }
\seealso{ \code{\link{rwi.stats}}, \code{\link{read.rwl}} }
\examples{library(utils)
data(ca533)
rwl.stats(ca533)
summary(ca533)
}
\keyword{ misc }

\encoding{UTF-8}
\name{caps}
\alias{caps}
\title{ Cook and Peters Smoothing Spline with User-Specified Rigidity and Frequency Cutoff }
\description{
  Applies a smoothing spline to \code{\var{y}} with rigidity determined
  by two parameters: frequency response \code{\var{f}} at a wavelength
  of \code{\var{nyrs}} years.
}
\usage{
caps(y, nyrs = 32, f = 0.5)
}
\arguments{
  \item{y}{ a \code{numeric} vector, typically a tree-ring series. }
  \item{nyrs}{ a number greater than zero, affecting the rigidity of the spline.  If \code{nyrs} is between zero and one it will be treated as percentage of the length of the series. E.g., a value of 0.667 would 0.667 * length(y) or what is often called a 2/3 spline in the tree-ring literature. When \code{\var{f}} is kept constant, a larger \code{\var{nyrs}} produces a more rigid spline.}
  \item{f}{ a number between 0 and 1 giving the frequency response at a wavelength of \code{\var{nyrs}} years.  When \code{\var{nyrs}} is kept constant, a smaller \code{\var{f}} produces a more rigid spline: At one extreme, \code{\var{f} = 0} causes the function to return the least-squares straight line fit to the data.  At the other extreme, as \code{\var{f}} approaches 1 the result approaches the natural spline, i.e. the function outputs \code{\var{y}}. The default value is 0.5 and shouldn't be changed without a good reason.}
}
\details{
  This applies the classic smoothing spline from Cook and Peters (1981). The rigidity of the spline has a frequency response of 50\% at a wavelength of \code{\var{nyrs}}. The references, of course, have more information.
  
  This function was introduced to \code{\var{dplR}} in version 1.7.3 and replaces the now defunct \code{\var{ffcsaps}}. Where \code{\var{ffcsaps}} was written entirely in R, \code{\var{caps}} is a wrapper for a Fortran subroutine from Ed Cook's ARSTAN program that is thousands of times faster.
  
  The default value of \code{\var{nyrs}} was changed to 32 from 2/3 length of \code{\var{y}} in version 1.7.8 based on a suggestion from Klesse (2021).
  
  Note: \code{\var{caps}} returns \code{NA} if there are any \code{NA} values in \code{y}. See examples.
}
\value{
  A filtered vector.
}
\references{
  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.
    
  Cook, E. R. and Peters, K. (1981) The Smoothing Spline: A New Approach to Standardizing Forest Interior Tree-Ring Width Series for Dendroclimatic Studies. Tree-Ring Bulletin, 41, 45-53.

  Klesse, S. (2021) Critical Note on the Application of the "Two-Third"" Spline. Dendrochronologia Volume 65: 125786
}
\author{
  Fortran code provided by Ed Cook and adapted for dplR by Andy Bunn.
}

\seealso{ \code{\link{ads}} }

\examples{library(graphics)
library(utils)

## Use first series from the Mesa Verde data set
data(co021)
series <- co021[, 1]
series <- series[!is.na(series)]
plot(series, type = "l", ylab = "Ring Width (mm)", col = "grey")
lines(caps(series, nyrs = 10), col = "red", lwd = 2)
lines(caps(series, nyrs = 100), col = "green", lwd = 2)
# A 2/3 spline, the default, 462.6667 yrs here
lines(caps(series), col = "blue", lwd = 2)
legend("topright",
       c("Series", "nyrs=10", "nyrs=100",
         paste("Default nyrs (", floor(length(series) * 2/3), ")", sep="")),
       fill=c("grey", "red", "green", "blue"))


## Note behavior when NA is encountered and
## take appropriate measures as demonstrated above
y <- c(NA,NA,rnorm(100))
caps(y)

}
\keyword{ smooth }

\encoding{UTF-8}
\name{fill.internal.NA}
\alias{fill.internal.NA}
\title{Fill Internal NA}
\description{
  This function fills internal \code{NA} values (i.e., those with numeric data
  above and below a small data gap) in each column of a
  \code{data.frame} such as a data set of ring widths as produced by
  \code{\link{read.rwl}}.
}
\usage{
fill.internal.NA(x, fill = c("Mean", "Spline", "Linear"))
}
\arguments{
  \item{x}{a \code{data.frame} of ring widths with
    \code{row.names(\var{x})} containing years and \code{names(\var{x})}
    containing each series \acronym{ID} such as produced by
    \code{\link{read.rwl}}}
  \item{fill}{a \code{numeric} value (e.g., \code{0}) or a \code{character} string of
    \code{"Mean"}, \code{"Spline"}, or \code{"Linear"}.  Defaults to \code{"Mean"}.}
}
\details{
  There are occasionally data gaps within a tree-ring series.  Some of
  the functions in dplR will fail
  when an internal \code{NA} is encountered (e.g. \code{\link{caps}}).  This
  function fills internal \code{NA} values with either a given numeric value
  (e.g., \code{0}) or through crude imputation.  The latter can be calculated as 
  the mean of the series (\code{fill="Mean"}) or calculated by fitting a cubic spline
  (\code{fill="Spline"}) using the \code{\link{spline}} function or calculated
  by linear approximation (\code{fill="Linear"}) using the function
  \code{\link{approx}}.

  Editorial: Having internal \code{NA} in a tree-ring series is
  often bad practice and filling those values should be done with
  caution.  For instance, some users code missing rings as \code{NA}
  instead of \code{0}.  And missing values (i.e., \code{NA}) are
  sometimes present in maximum latewood density data when the rings are
  small.  A common, but not recommended, practice is to leave stretches
  of \code{NA} values in places where it has been impossible to
  accurately measure rings (perhaps because of a break in the core).  It
  is often better to treat that core as two separate series (e.g., "01A"
  and "01B" rather than have internal \code{NA} values.  As with all
  processing, the analyst should make a decision based on their
  experience with the wood and not rely on software to make a choice for
  them!
}
\value{
  A \code{data.frame} with \code{colnames(\var{x})} and
  \code{rownames(\var{x})}.  Internal \code{NA}s
  filled as above.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{spline}}, \code{\link{approx}}
}
\examples{
library(graphics)
library(stats)
foo <- data.frame(x1=c(rnorm(5), NA, NA, rnorm(3)),
                  x2=c(rnorm(10)),
                  x3=c(NA, NA, rnorm(3), NA, rnorm(4)),
                  x4=c(NA, NA, rnorm(3), NA, rnorm(3), NA),
                  x5=c(NA, NA, rnorm(8)),
                  x6=c(NA, rnorm(9)),
                  x7=c(NA, rnorm(5), NA, rnorm(3)),
                  x8=c(rnorm(8), NA, NA),
                  x9=c(rnorm(5), NA, rnorm(3), NA))
row.names(foo) <- 1901:1910
class(foo) <- c("rwl","data.frame")
fill.internal.NA(foo, fill=0)

bar <- fill.internal.NA(foo, fill="Spline")
baz <- fill.internal.NA(foo, fill="Linear")

## note differences in method "Spline" vs. "Linear"
yrs <- time(foo)
plot(yrs, foo$x7, type="b", lwd=3)
lines(yrs, bar$x7, col="red", lwd=2)
lines(yrs, baz$x7, col="green", lwd=1)

}
\keyword{ manip }

\encoding{UTF-8}
\name{gp.po}
\docType{data}
\alias{gp.po}
\title{ Ponderosa Pine Pith Offsets Corresponding to \code{\link{gp.rwl}} }
\description{
  This data set gives the pith offsets that match the ring widths for
  \code{\link{gp.rwl}} \enc{–}{-} a data set of ponderosa pine (\emph{Pinus
  ponderosa}) from the Gus Pearson Natural Area (\acronym{GPNA}) in
  northern Arizona, \acronym{USA}.  Data are further described by Biondi
  and Qeadan (2008) and references therein.
}
\usage{data(gp.po)}
\format{
  A \code{data.frame} containing series \acronym{ID}s in column 1
  (\code{\var{series}}) and the number of years between the beginning of
  that series in \code{\link{gp.rwl}} and the pith of the tree
  (\code{\var{pith.offset}}).  This can be used together with the ring
  widths to calculate the cambial age of each ring.
}
\source{
  DendroLab, University of Nevada Reno, \acronym{USA}.
  \url{https://dendrolaborg.wordpress.com/}
}
\references{
  Biondi, F. and Qeadan, F. (2008) A theory-driven approach to tree-ring
  standardization: Defining the biological trend from expected basal area
  increment.  \emph{Tree-Ring Research}, \bold{64}(2), 81\enc{–}{--}96.
}
\keyword{datasets}

\encoding{UTF-8}
\name{read.tridas}
\alias{read.tridas}

\title{ Read Tree Ring Data Standard (TRiDaS) File }

\description{ This function reads in a TRiDaS format \acronym{XML} file.
  Measurements, derived series and various kinds of metadata are
  supported. }

\usage{
read.tridas(fname, ids.from.titles = FALSE,
            ids.from.identifiers = TRUE, combine.series = TRUE,
            trim.whitespace = TRUE, warn.units = TRUE)
}

\arguments{

  \item{fname}{ \code{character} vector giving the file name of the
    TRiDaS file. }

  \item{ids.from.titles}{ \code{logical} flag indicating whether to
    override the (tree, core, radius, measurement) structure imposed by
    the element hierarchy (element, sample, radius, measurementSeries)
    of the file.  If \code{TRUE}, measurement series will be rearranged
    by matching titles in the file at the aforementioned four levels of
    the hierarchy.  Defaults to \code{FALSE}, i.e. the element hierarchy
    of the file will be used. }

  \item{ids.from.identifiers}{ \code{logical} flag indicating whether to
    (partially) override the element hierarchy of the file.  If
    \code{TRUE}, measurement series will be grouped according to
    matching identifiers at the measurementSeries level, where
    identifiers are available.  The changes caused by this option are
    applied on top of the structure imposed by the file or computed from
    matching titles if \code{\var{ids.from.titles} == TRUE}.  Defaults
    to \code{TRUE}. }

  \item{combine.series}{ \code{logical} flag indicating whether to
    combine two or more measurement series with the same set of (tree,
    core, radius, measurement) \acronym{ID} numbers.  Each set of
    combined measurement series will be represented by one column of a
    resulting \code{data.frame}.  Overlapping years of combined series do not
    produce a warning.  If several data points are available for a given
    year, the function chooses one in a rather arbitrary manner.  This
    option can only have effect when \code{\var{ids.from.titles} ||
      \var{ids.from.identifiers}}. }

  \item{trim.whitespace}{ \code{logical} flag indicating whether to
    replace repeated white spaces in the text content of the file with
    only one space.  Defaults to \code{TRUE}, i.e. excess white space
    will be trimmed from the text. }

  \item{warn.units}{ \code{logical} flag indicating whether to warn
    about unitless measurements and \dQuote{strange} units.  The
    function expects measurements in units that can be converted to
    millimetres.  Defaults to \code{TRUE}: warnings will be given.  For
    example, density measurements will trigger warnings, which can be
    disabled by setting this option to \code{FALSE}. }
  
}

\details{

  The Tree Ring Data Standard (TRiDaS) is described in Jansma et. al
  (2010).  

  The parameters used for rearranging (\code{\var{ids.from.titles}},
  \code{\var{ids.from.identifiers}}) and combining
  (\code{\var{combine.series}}) measurement series only affect the four
  lowest levels of document structure: element, sample, radius,
  measurementSeries.  Series are not reorganized or combined at the
  upper structural levels (project, object).

}

\value{
  A list with a variable number of components according to the contents
  of the input file.  The possible list components are:

  \item{measurements}{A \code{data.frame} or a list of \code{data.frame}s with
    the series in columns and the years as rows.  Contains measurements
    (\samp{<measurementSeries>}) with known years.  The series
    \acronym{ID}s are the column names and the years are the row
    names.  The series \acronym{ID}s are derived from \samp{<title>}
    elements in the input file.  Each unique combination of
    \samp{<project>}, \samp{<object>}, \samp{<unit>}, \samp{<taxon>},
    and \samp{<variable>} gets a separate \code{data.frame}. }

  \item{ids}{A \code{data.frame} or a list of \code{data.frame}s with columns
    named \code{"tree"}, \code{"core"}, \code{"radius"}, and
    \code{"measurement"}, together giving a unique \code{numeric} \acronym{ID}
    for each column of the \code{data.frame}(s) in
    \code{\var{measurements}}.

    If \code{!\var{combine.series} && (\var{ids.from.titles} || \var{ids.from.identifiers})}, some rows may be non-unique. }
  
  \item{titles}{A \code{data.frame} or a list of \code{data.frame}s with
    columns named \code{"tree"}, \code{"core"}, \code{"radius"}, and
    \code{"measurement"}, containing the \samp{<title>} hierarchy of
    each column of the \code{data.frame}(s) in \code{\var{measurements}}. }
  
  \item{wood.completeness}{A \code{data.frame} or a list of \code{data.frame}s
    containing wood completeness information.  Column names are a subset
    of the following, almost self-explanatory set:
    \code{"pith.presence"}, \code{"heartwood.presence"},
    \code{"sapwood.presence"},\cr
    \code{"last.ring.presence"}, \code{"last.ring.details"},
    \code{"bark.presence"},\cr
    \code{"n.sapwood"}, \code{"n.missing.heartwood"},
    \code{"n.missing.sapwood"},\cr
    \code{"missing.heartwood.foundation"},
    \code{"missing.sapwood.foundation"},\cr
    \code{"n.unmeasured.inner"}, \code{"n.unmeasured.outer"}. }
  
  \item{unit}{A \code{character} vector giving the unit of the
    measurements.  Length equals the number of \code{data.frame}s in
    \code{\var{measurements}}. }
  
  \item{project.id}{A \code{numeric} vector giving the project
    \acronym{ID}, i.e. the position of the corresponding
    \samp{<project>} element), of the measurements in each
    \code{data.frame} in \code{\var{measurements}}.  Length equals the
    number of \code{data.frame}s. }
  
  \item{project.title}{A \code{character} vector giving the title of the
    project of each \code{data.frame} in \code{\var{measurements}}.
    Length equals the number of \code{data.frame}s. }

  \item{site.id}{A \code{data.frame} giving the site \acronym{ID}
    (position of \samp{<object>} element(s) within a \samp{<project>})
    of each \code{data.frame} in \code{\var{measurements}}.  May have
    several columns to reflect the possibly nested \samp{<object>}
    elements. }
  
  \item{site.title}{A \code{data.frame} giving the site
    (\samp{<object>}) title of each \code{data.frame} in
    \code{\var{measurements}}.  May have several columns to reflect the
    possibly nested \samp{<object>} elements. }
  
  \item{taxon}{A \code{data.frame} showing the taxonomic name for each
    \code{data.frame} in \code{\var{measurements}}.  Contains some of
    the following columns: \code{"text"}, \code{"lang"},
    \code{"normal"}, \code{"normalId"}, \code{"normalStd"}.  The first
    two are a free-form name and its language, and the rest are related
    to a normalized name. }
  
  \item{variable}{A \code{data.frame} showing the measured variable of
    each \code{data.frame} in \code{\var{measurements}}.  Contains some
    of the following columns: \code{"text"}, \code{"lang"},
    \code{"normal"}, \code{"normalId"}, \code{"normalStd"},
    \code{"normalTridas"}.  The first two are a free-form name and its
    language, and the rest are related to a normalized name. }
  
  \item{undated}{A \code{list} of measurements with unknown years,
    together with metadata.  Elements are a subset of the following:

    \describe{

      \item{data}{A \code{numeric} vector or a list of such vectors
        containing measurement series }

      \item{unit}{A \code{character} vector giving the unit of the
        measurements.  Length equals the number of measurement series in
        \code{\var{undated}$\var{data}} }

      \item{ids}{A \code{data.frame} with columns named \code{"tree"},
        \code{"core"}, \code{"radius"}, and \code{"measurement"},
        together giving a \code{numeric} \acronym{ID} for each measurement
        series in \code{\var{undated}$\var{data}}.  The rows are
        guaranteed to be unique only when comparing measurement series
        with the same \code{\var{project.id}} and \code{\var{site.id}},
        but not if \code{\var{ids.from.titles} ||
          \var{ids.from.identifiers}}. }

      \item{titles}{A \code{data.frame} with columns named
        \code{"tree"}, \code{"core"}, \code{"radius"}, and
        \code{"measurement"}, containing the \samp{<title>} hierarchy of
        each measurement series in \code{\var{undated}$\var{data}} }
      
      \item{project.id}{A \code{numeric} vector giving the project
        \acronym{ID} of each measurement series in
        \code{\var{undated}$\var{data}} }
      
      \item{project.title}{A \code{character} vector giving the project
        title of each measurement series in
        \code{\var{undated}$\var{data}} }
      
      \item{site.id}{A \code{data.frame} giving the site \acronym{ID} of
        each measurement series in \code{\var{undated}$\var{data}} }

      \item{site.title}{A \code{data.frame} giving the site title of
        each measurement series in \code{\var{undated}$\var{data}} }

      \item{variable}{A \code{data.frame} containing the variable of
        each measurement series in \code{\var{undated}$\var{data}} }

      \item{taxon}{A \code{data.frame} containing taxonomic names of
        each measurement series in \code{\var{undated}$\var{data}} }

      \item{wood.completeness}{A \code{data.frame} containing wood
        completeness information of each measurement series in
        \code{\var{undated}$\var{data}} }
      
    }
  }

  \item{derived}{A \code{list} of calculated series of values, together
    with metadata.  Elements are a subset of the following:

    \describe{

      \item{data}{A \code{numeric} vector or a list of such vectors
        containing calculated series of values.}

      \item{link}{A \code{list} of \code{data.frame}s, one for each
        series in \code{\var{derived}$\var{data}}, giving links to the
        measurements used to form the corresponding derived series.
        Each \code{data.frame} has a subset of the following columns:
        \code{"idRef"} (reference to a series in the same file),
        \code{"xLink"} (\acronym{URI}), \code{"identifier"}, and
        \code{"domain"} (identifier and its domain, not necessarily in
        the same file).}

      \item{project.id}{A \code{numeric} vector giving the project
        \acronym{ID} of each derived series in
        \code{\var{derived}$\var{data}}}

      \item{id}{A \code{numeric} vector giving the \acronym{ID} (order
        of appearance in the project) of each derived series in
        \code{\var{derived}$\var{data}}}

      \item{title}{A \code{character} vector giving the title of each
        derived series in \code{\var{derived}$\var{data}}}

      \item{project.title}{A \code{character} vector giving the project
        title of each derived series in \code{\var{derived}$\var{data}}}

      \item{unit}{A \code{character} vector giving the unit of the
        derived series.  Length equals the number of series in
        \code{\var{derived}$\var{data}}.}

      \item{standardizing.method}{A \code{character} vector giving the
        standardizing method of the derived series.  Length equals the
        number of series in \code{\var{derived}$\var{data}}.}

      \item{variable}{A \code{data.frame} containing the variable of
        each series in \code{\var{derived}$\var{data}}}
      
    }
  }

  \item{type}{A \code{data.frame} containing the type of various
    entities, and metadata related to each \samp{type} element.
    Contents are \code{NA} where the metadata is not applicable (e.g.,
    no \code{\var{tree.id}} when the \samp{type} element refers to a
    project).  Columns are a subset of the following:

    \describe{

      \item{text}{The text of the \samp{type} element}

      \item{lang}{The language of the text}

      \item{normal}{The normalized name of the type}

      \item{normalId}{The \acronym{ID} value of the type in the standard
        dictionary}

      \item{normalStd}{The name of the standard}

      \item{project.id}{The \acronym{ID} of the project}

      \item{site.id}{One or more columns with this prefix, depending on
        the maximum depth of the \samp{<object>} hierarchy.  Gives the
        \acronym{ID} of the site where the \samp{<type>} element
        appeared.}

      \item{tree.id}{The \acronym{ID} of the tree}

      \item{core.id}{The \acronym{ID} of the core}

      \item{derived.id}{The \acronym{ID} of the derived series}

      \item{project.title}{The title of the project}

      \item{site.title}{One or more columns with this prefix, depending
        on the maximum depth of the \samp{<object>} hierarchy.  Gives the
        title of the site where the \samp{<type>} element appeared.}

      \item{tree.title}{The title of the tree}

      \item{core.title}{The title of the core}

      \item{derived.title}{The title of the derived series}

    }
  }

  \item{comments}{A \code{data.frame} containing comments to various
    entities, and metadata related to each \samp{comments} element.
    Contents are \code{NA} where the metadata is not applicable.  Columns
    are a subset of the following:

    \describe{

      \item{text}{The text of the \samp{comments} element}

      \item{project.id}{The \acronym{ID} of the project}

      \item{site.id}{One or more columns with this prefix, depending on
        the maximum depth of the \samp{<object>} hierarchy.  Gives the
        \acronym{ID} of the site.}

      \item{tree.id}{The \acronym{ID} of the tree}

      \item{core.id}{The \acronym{ID} of the core}

      \item{radius.id}{The \acronym{ID} of the radius}

      \item{measurement.id}{The \acronym{ID} of the measurement series}

      \item{derived.id}{The \acronym{ID} of the derived series}

      \item{project.title}{The title of the project}

      \item{site.title}{One or more columns with this prefix, depending
        on the maximum depth of the \samp{<object>} hierarchy.  Gives the
        title of the site.}

      \item{tree.title}{The title of the tree}

      \item{core.title}{The title of the core}

      \item{radius.title}{The title of the radius}

      \item{measurement.title}{The title of the measurement series}

      \item{derived.title}{The title of the derived series}

    }
  }

  \item{identifier}{A \code{data.frame} containing identifiers of
    various entities, and metadata related to each \samp{identifier}
    element.  Contents are \code{NA} where the metadata is not
    applicable.  Columns are a subset of the following:

    \describe{

      \item{text}{The text of the \samp{identifier} element}

      \item{domain}{The domain which the identifier is applicable to}

      \item{project.id}{The \acronym{ID} of the project}

      \item{site.id}{One or more columns with this prefix, depending on
        the maximum depth of the \samp{<object>} hierarchy.  Gives the
        \acronym{ID} of the site.}

      \item{tree.id}{The \acronym{ID} of the tree}

      \item{core.id}{The \acronym{ID} of the core}

      \item{radius.id}{The \acronym{ID} of the radius}

      \item{measurement.id}{The \acronym{ID} of the measurement series}

      \item{derived.id}{The \acronym{ID} of the derived series}

      \item{project.title}{The title of the project}

      \item{site.title}{One or more columns with this prefix, depending
        on the maximum depth of the \samp{<object>} hierarchy.  Gives the
        title of the site.}

      \item{tree.title}{The title of the tree}

      \item{core.title}{The title of the core}

      \item{radius.title}{The title of the radius}

      \item{measurement.title}{The title of the measurement series}

      \item{derived.title}{The title of the derived series}

    }
  }

  \item{remark}{A \code{list} of remarks concerning individual measured
    or derived values, with some of the following items:

    \describe{

      \item{measurements}{Remarks related to measurements with a known
        year.  A \code{data.frame} with the following columns:

        \describe{

          \item{text}{The remark}

          \item{frame}{Index to a \code{data.frame} in
            \code{\var{measurements}}}

          \item{row}{Index to a row of the \code{data.frame}}

          \item{col}{Index to a column of the \code{data.frame}}

        }
      }
      
      \item{undated}{Remarks related to measurements without a known
        year.  A \code{data.frame} with the following columns:

        \describe{

          \item{text}{The remark}

          \item{series}{Index to a series in
            \code{\var{undated}$\var{data}}}

          \item{idx}{Index to a value in the series}

        }
      }

      \item{derived}{Remarks related to derived values.  A
        \code{data.frame} with the following columns:

        \describe{

          \item{text}{The remark}

          \item{series}{Index to a series in
            \code{\var{derived}$\var{data}}}

          \item{idx}{Index to a value in the series}

        }
      }
    }
  }

  \item{laboratory}{A \code{data.frame} or a \code{list} of
    \code{data.frame}s with one item per project.  Each
    \code{data.frame} contains information about the research
    laboratories involved in the project.  Columns are a subset of the
    following:

    \describe{

      \item{name}{Name of the laboratory}

      \item{acronym}{Acronym of the name}

      \item{identifier}{Identifier}

      \item{domain}{Domain which the identifier is applicable to}

      \item{addressLine1}{Address}

      \item{addressLine2}{Another address line}

      \item{cityOrTown}{City or town}

      \item{stateProvinceRegion}{State, province or region}

      \item{postalCode}{Postal code}

      \item{country}{Country}

    }
  }

  \item{research}{A \code{data.frame} or a \code{list} of
    \code{data.frame}s with one item per project.  Each
    \code{data.frame} contains information about the systems in which
    the research project is registered.  Columns are the following:

    \describe{

      \item{identifier}{Identifier}

      \item{domain}{Domain which the identifier is applicable to}

      \item{description}{General description}

    }
  }

  \item{altitude}{A \code{data.frame} containing the altitude of
    trees.  Columns are the following:

    \describe{

      \item{metres}{The altitude in metres}

      \item{project.id}{The \acronym{ID} of the project}

      \item{site.id}{One or more columns with this prefix, depending on
        the maximum depth of the \samp{<object>} hierarchy.  Gives the
        \acronym{ID} of the site.}

      \item{tree.id}{The \acronym{ID} of the tree}

      \item{project.title}{The title of the project}

      \item{site.title}{One or more columns with this prefix, depending
        on the maximum depth of the \samp{<object>} hierarchy.  Gives the
        title of the site.}

      \item{tree.title}{The title of the tree}

    }
  }

  \item{preferred}{A \code{data.frame} containing links to preferred
    measurement series.  Columns are a subset of the following:

    \describe{

      \item{idRef}{Reference to a series in the same file}

      \item{xLink}{\acronym{URI}}

      \item{identifier}{Identifier of a series not necessarily in the
        same file}

      \item{domain}{Domain which the identifier is applicable to}

      \item{project.id}{The \acronym{ID} of the project}

      \item{site.id}{One or more columns with this prefix, depending on
        the maximum depth of the \samp{<object>} hierarchy.  Gives the
        \acronym{ID} of the site.}

      \item{tree.id}{The \acronym{ID} of the tree}

      \item{project.title}{The title of the project}

      \item{site.title}{One or more columns with this prefix, depending on
        the maximum depth of the \samp{<object>} hierarchy.  Gives
        the title of the site.}

      \item{tree.title}{The title of the tree}

    }
  }
}

\references{
  Jansma, E., Brewer, P. W., and Zandhuis, I. (2010) TRiDaS 1.1: The
  tree-ring data standard.  \emph{Dendrochronologia}, \bold{28}(2),
  99--130.

}

\author{ Mikko Korpela }

\note{ This is an early version of the function.  Bugs are likely to
  exist, and parameters and return values are subject to change.  Not
  all metadata defined in the TRiDaS specification is supported --
  unsupported elements are quietly ignored. }

\seealso{ \code{\link{read.rwl}}, \code{\link{read.tucson}},
  \code{\link{read.compact}}, \code{\link{read.fh}},
  \code{\link{write.tridas}} }

\keyword{ IO }

\encoding{UTF-8}
\name{wavelet.plot}
\alias{wavelet.plot}
\title{ Plot a Continuous Wavelet Transform }
\description{
  This function creates a \code{filled.contour} plot of a continuous
  wavelet transform as output from \code{\link{morlet}}.
}
\usage{
wavelet.plot(wave.list,
             wavelet.levels = quantile(wave.list$Power,
                                       probs = (0:10)/10),
             add.coi = TRUE, add.sig = TRUE,
             x.lab = gettext("Time", domain = "R-dplR"),
             period.lab = gettext("Period", domain = "R-dplR"),
             crn.lab = gettext("RWI", domain = "R-dplR"),
             key.cols = rev(rainbow(length(wavelet.levels)-1)),
             key.lab = parse(text=paste0("\"",
                                         gettext("Power",
                                                 domain="R-dplR"),
                                         "\"^2")),
             add.spline = FALSE, f = 0.5, nyrs = NULL,
             crn.col = "black", crn.lwd = 1,coi.col='black',
             crn.ylim = range(wave.list$y) * c(0.95, 1.05),
             side.by.side = FALSE,
             useRaster = FALSE, res = 150, reverse.y = FALSE, \dots)
}
\arguments{
  \item{wave.list}{A \code{list}.  Output from \code{\link{morlet}}.}
  \item{wavelet.levels}{A \code{numeric} vector.  Values for levels of
    the filled contours for the wavelet plot.}
  \item{add.coi}{A \code{logical} flag.  Add cone of influence?}
  \item{add.sig}{A \code{logical} flag.  Add contour lines for significance?}
  \item{x.lab}{X-axis label.}
  \item{period.lab}{Y-axis label for the wavelet plot.}
  \item{crn.lab}{Y-axis label for the time-series plot.}
  \item{key.cols}{A vector of colors for the wavelets and the key.}
  \item{key.lab}{Label for key.}
  \item{add.spline}{A \code{logical} flag.  Add a spline to the
    time-series plot using \code{\link{caps}}?. }
  \item{nyrs}{ A number giving the rigidity of the smoothing spline,
    defaults to 0.33 of series length if \code{\var{nyrs}} is
    \code{NULL}. }
  \item{f}{ A number between 0 and 1 giving the frequency response or
    wavelength cutoff for the smoothing spline.  Defaults to 0.5. }
  \item{crn.col}{Line color for the time-series plot.}
  \item{crn.lwd}{Line width for the time-series plot.}
  \item{coi.col}{Color for the COI if \code{\var{add.coi}} is \code{TRUE}.}
  \item{crn.ylim}{Axis limits for the time-series plot.}
  \item{side.by.side}{A \code{logical} flag.  Plots will be in one row if
    \code{TRUE}.}
  \item{useRaster}{A \code{logical} flag.  If \code{TRUE}, the filled
    contours are drawn as a raster image.  Other parts of the plot are
    not affected.  \code{useRaster=TRUE} can be especially useful when a
    \code{pdf} device is used: the size and complexity of the
    \acronym{PDF} file will probably be greatly reduced.  Setting this
    to \code{TRUE} has negative effects when used with a bitmap
    device such as \code{png}.  If \code{NA}, plotting of a raster image
    will be attempted if and only if the name of the graphics device is
    \code{"pdf"} or \code{"postscript"}.  The default is \code{FALSE}:
    draw directly to the graphics device without using an intermediate
    raster image.  }
  \item{res}{A \code{numeric} vector of length 1.  The resolution
    (pixels per inch) of the filled contours when a raster image is
    used.  See \code{\var{useRaster}}.}
  \item{reverse.y}{A \code{logical} flag.  If \code{TRUE}, the Y-axis
    will be reversed, i.e. period increasing towards the bottom.  The
    default is \code{FALSE}. }
  \item{\dots}{Arguments passed to \code{\link{rasterPlot}}.  Only
    relevant when the filled contours are drawn as a raster image.  See
    \code{\var{useRaster}}.
  }
}
\details{
  This produces a plot of a continuous wavelet transform and plots the
  original time series.  Contours are added for significance and a cone of
  influence polygon can be added as well.  Anything within the cone of
  influence should not be interpreted.

  The time series can be plotted with a smoothing spline as well.
}
\value{
  None.  This function is invoked for its side effect, which is to produce a
  plot.
}
\references{
  Torrence, C. and Compo, G. P. (1998) A practical guide to wavelet
  analysis.  \emph{Bulletin of the American Meteorological Society},
  \bold{79}(1), 61\enc{–}{--}78.
}
\note{ The function \code{morlet} is a port of Torrence\enc{’}{'}s
  \acronym{IDL} code, which can be accessed through the
\href{https://web.archive.org/web/20141216211516/https://paos.colorado.edu/research/wavelets/software.html}{Internet Archive Wayback Machine}.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{morlet}}, \code{\link{caps}}}
\examples{library(stats)
library(utils)
data(ca533)
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
ca533.crn <- chron(ca533.rwi, prewhiten = FALSE)
Years <- time(ca533.crn)
CAMstd <- ca533.crn[, 1]
out.wave <- morlet(y1 = CAMstd, x1 = Years, p2 = 9, dj = 0.1,
                   siglvl = 0.99)
wavelet.plot(out.wave, useRaster = NA)
\dontrun{
# Alternative palette with better separation of colors
# via: rev(RColorBrewer::brewer.pal(10, "Spectral"))
specCols <- c("#5E4FA2", "#3288BD", "#66C2A5", "#ABDDA4", "#E6F598", 
              "#FEE08B", "#FDAE61", "#F46D43", "#D53E4F", "#9E0142")
wavelet.plot(out.wave, key.cols=specCols,useRaster = NA)

# fewer colors
levs <- quantile(out.wave$Power, probs = c(0, 0.5, 0.75, 0.9, 0.99))
wavelet.plot(out.wave, wavelet.levels = levs, add.sig = FALSE,
             key.cols = c("#FFFFFF", "#ABDDA4", "#FDAE61", "#D7191C"), useRaster = NA)
             }
}
\keyword{ hplot }

\encoding{UTF-8}
\name{sea}

\alias{sea}

\title{ Superposed Epoch Analysis }

\description{
  This function calculates the significance of the departure from the
  mean for a given set of key event years and lagged years.
}

\usage{
sea(x, key, lag = 5, resample = 1000)
}

\arguments{
  \item{x}{ a chronology \code{data.frame} of ring-width indices (such
    as produced by \link{chron})}
  \item{key}{ a vector specifying the key event years for the superposed
    epoch}
  \item{lag}{ an integral value defining the number of lagged years }
  \item{resample}{ an integral value specifying the number of bootstrap
    sample for calculation of confidence intervals}
}

\details{
  Superposed epoch analysis (SEA) is used to test the significance of a mean
  tree growth response to certain events (such as droughts).  Departures
  from the mean \acronym{RWI} values for the specified years prior to
  each event year, the event year, and the specified years immediately
  after each event are averaged to a superposed epoch.  To determine if
  \acronym{RWI} for these years was significantly different from
  randomly selected sets of \code{\var{lag}+1} other years, bootstrap
  resampling is used to randomly select sets of \code{\var{lag}+1} years
  from the data set and to estimate significances for the departures
  from the mean \acronym{RWI}.
  
  SEA computation is based on scaled \acronym{RWI} values, and
  95\%-confidence intervals are computed for the scaled values for each
  year in the superposed epoch.
}

\value{
  A \code{data.frame} with
  \item{lag}{the lagged years,}
  \item{se}{the superposed epoch, i.e. the scaled mean \acronym{RWI} for
    the event years,}
  \item{se.unscaled}{the unscaled superposed epoch, i.e. the mean
    \acronym{RWI} for the event years,}
  \item{p}{significance of the departure from the chrono\enc{’}{'}s mean
    \acronym{RWI},}
  \item{ci.95.lower}{lower 95\% confidence band,}
  \item{ci.95.upper}{upper 95\% confidence band,}
  \item{ci.99.lower}{lower 99\% confidence band,}
  \item{ci.99.upper}{upper 99\% confidence band.}
}

\author{ Christian Zang.  Patched and improved by Mikko Korpela. }

\references{

  Lough, J. M. and Fritts, H. C. (1987) An assessment of the possible
  effects of volcanic eruptions on North American climate using
  tree-ring data, 1602 to 1900 \acronym{AD}.  \emph{Climatic Change},
  \bold{10}(3), 219\enc{–}{--}239.
  
}

\examples{library(graphics)
library(utils)
data(cana157)
event.years <- c(1631, 1742, 1845)
cana157.sea <- sea(cana157, event.years)
foo <- cana157.sea$se.unscaled
names(foo) <- cana157.sea$lag
barplot(foo, col = ifelse(cana157.sea$p < 0.05, "grey30", "grey75"), 
        ylab = "RWI", xlab = "Superposed Epoch")
}

\keyword{ ts }

\encoding{UTF-8}
\name{print.rwl.report}
\alias{print.rwl.report}
\title{Do some reporting on a RWL object}
\description{
  This function prints the results of \code{\link{rwl.report}}
}
\usage{
\method{print}{rwl.report}(x, ...)
}
\arguments{
  \item{x}{a \code{list} from \code{\link{rwl.report}}}
  \item{\dots}{not implemented}
}
\details{
  This function formats the \code{list} from \code{\link{rwl.report}} for the 
  user to have a summary report of the number of series, the mean length
  of all the series, the first year, last year, the mean first-order 
  autocorrelation (via \code{\link{summary.rwl}}), the mean interseries 
  correlation (via \code{\link{interseries.cor}}), the years where a series has 
  a missing ring (zero), internal NA, or a very small ring (<0.005). 
}
\value{
  Invisible
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{rwl.report}}, \code{\link{summary.rwl}},
          \code{\link{interseries.cor}}
}
\examples{
data("gp.rwl")
rwl.report(gp.rwl)
foo <- gp.rwl
foo[177,1] <- NA 
foo[177:180,3] <- NA 
foo[185,4] <- 0.001 
rwl.report(foo)
}
\keyword{ manip }

\encoding{UTF-8}
\name{ca533}
\docType{data}
\alias{ca533}
\title{Campito Mountain Tree Ring Widths}
\description{
  This data set gives the raw ring widths for bristlecone pine
  \emph{Pinus longaeva} at Campito Mountain in California,
  \acronym{USA}.  There are 34 series.  Data set was created using
  \code{\link{read.rwl}} and saved to an .rda file using
  \code{\link{save}}.
}
\usage{data(ca533)}
\format{
  A \code{data.frame} containing 34 tree-ring series in columns and 1358
  years in rows.
}
\source{
  International tree-ring data bank, Accessed on 20-April-2021 at
  \url{https://www.ncei.noaa.gov/pub/data/paleo/treering/measurements/northamerica/usa/ca533.rwl}
}
\references{
  Graybill, D. A. and LaMarche, Jr., V. C. (1983) Campito Mountain Data
  Set.  \acronym{IGBP} \acronym{PAGES}/World Data Center for
  Paleoclimatology Data Contribution Series 1983-CA533.RWL.
  \acronym{NOAA}/\acronym{NCDC} Paleoclimatology Program, Boulder,
  Colorado, \acronym{USA}.
}
\keyword{datasets}

\encoding{UTF-8}
\name{ccf.series.rwl}
\alias{ccf.series.rwl}
\title{ Cross-Correlation between a Series and a Master Chronology }
\description{
  Computes cross-correlations between a tree-ring series and a master
  chronology built from a rwl object at user-specified lags and segments.
}
\usage{
ccf.series.rwl(rwl, series, series.yrs = as.numeric(names(series)),
               seg.length = 50, bin.floor = 100, n = NULL,
               prewhiten = TRUE, biweight = TRUE, pcrit = 0.05,
               lag.max = 5, make.plot = TRUE,
               floor.plus1 = FALSE, series.x = FALSE, \dots)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as rows
    such as that produced by \code{\link{read.rwl}}. }
  \item{series}{ a \code{numeric} or \code{character} vector.  Usually a
    tree-ring series.  If the length of the value is 1, the
    corresponding column of \code{\var{rwl}} is selected (by name or
    position) as the series and ignored when building the master
    chronology.  Otherwise, the value must be \code{numeric}. }
  \item{series.yrs}{ a \code{numeric} vector giving the years of
    \code{\var{series}}.  Defaults to
    \code{as.numeric(names(\var{series}))}.  Ignored if
    \code{\var{series}} is an index to a column of \code{\var{rwl}}. }
  \item{seg.length}{ an even integral value giving length of segments in
    years (e.g., 20, 50, 100 years). }
  \item{bin.floor}{ a non-negative integral value giving the base for
    locating the first segment (e.g., 1600, 1700, 1800
    \acronym{AD}).  Typically 0, 10, 50, 100, etc. }
  \item{n}{ \code{NULL} or an integral value giving the filter length
    for the \code{\link{hanning}} filter used for removal of low
    frequency variation. }
  \item{prewhiten}{ \code{logical} flag.  If \code{TRUE} each series is
    whitened using \code{\link{ar}}. }
  \item{biweight}{ \code{logical} flag.  If \code{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}}. }
  \item{pcrit}{ a number between 0 and 1 giving the critical value for the
    correlation test. }
  \item{lag.max}{ an integral value giving the maximum lag at which to
    calculate the \code{\link{ccf}}. }
  \item{make.plot}{ \code{logical} flag indicating whether to make a
    plot. }
  \item{floor.plus1}{ \code{logical} flag.  If \code{TRUE}, one year is
    added to the base location of the first segment (e.g., 1601, 1701,
    1801 \acronym{AD}). }
  \item{series.x}{ \code{logical} flag indicating whether to make the \code{series}
    the \code{x} argument to \code{\link{ccf}}. See Details. }
  \item{\dots}{ other arguments passed to plot. }
}
\details{
  This function calculates the cross-correlation function between a
  tree-ring series and a master chronology built from \code{\var{rwl}}
  looking at correlations lagged positively and negatively using
  \code{\link{ccf}} at overlapping segments set by
  \code{\var{seg.length}}.  For instance, with \code{\var{lag.max}} set
  to 5, cross-correlations would be calculated at for each segment with
  the master lagged at \code{\var{k} = -5:5} years.
  
  The cross correlations are calculated calling
  \code{\link{ccf}} as \cr
  \code{ccf(x=master, y=series, lag.max=lag.max, plot=FALSE)} if \code{series.x} is 
  \code{FALSE} and as \code{ccf(x=series, y=master, lag.max=lag.max, plot=FALSE)} if 
  \code{series.x} is \code{TRUE}.  This argument was introduced in dplR version 1.7.0.
  Different users have different expectations about how missing or extra rings are 
  notated. If \code{series.x = FALSE} the behavior will be like COFECHA where a missing
  ring in a series produces a negative lag in the plot rather than a positive lag.

  Correlations are calculated for the first segment, then the
  second segment and so on.  Correlations are only calculated for segments with
  complete overlap with the master chronology.

  Each series (including those in the \code{\var{rwl}} object) is
  optionally detrended as the residuals from a \code{\link{hanning}}
  filter with weight \code{\var{n}}.  The filter is not applied if
  \code{\var{n}} is \code{NULL}.  Detrending can also be done via
  prewhitening where the residuals of an \code{\link{ar}} model are
  added to each series mean.  This is the default.  The master chronology
  is computed as the mean of the \code{\var{rwl}} object using
  \code{\link{tbrm}} if \code{\var{biweight}} is \code{TRUE} and
  \code{rowMeans} if not.  Note that detrending typically changes the
  length of the series. E.g., a \code{\link{hanning}} filter will
  shorten the series on either end by \code{floor(\var{n}/2)}.  The
  prewhitening default will change the series length based on the
  \code{\link{ar}} model fit.  The effects of detrending can be seen with
  \code{\link{series.rwl.plot}}.

}
\value{
  A \code{list} containing matrices \code{\var{ccf}} and
  \code{\var{bins}}.  Matrix \code{\var{ccf}} contains the correlations
  between the series and the master chronology at the lags window given
  by \code{\var{lag.max}}.  Matrix \code{\var{bins}} contains the years
  encapsulated by each bin.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{corr.rwl.seg}}, \code{\link{corr.series.seg}},
  \code{\link{skel.plot}}, \code{\link{series.rwl.plot}}
}
\references{ Bunn, A. G. (2010) Statistical and visual crossdating 
  in R using the dplR library.  \emph{Dendrochronologia},
  \bold{28}(4), 251\enc{–}{--}258.
}
\examples{
library(utils)
data(co021)
dat <- co021
## Create a missing ring by deleting a year of growth in a random series
flagged <- dat$"641143"
flagged <- c(NA, flagged[-325])
names(flagged) <- rownames(dat)
dat$"641143" <- NULL
ccf.100 <- ccf.series.rwl(rwl = dat, series = flagged, seg.length = 100)
\dontrun{
flagged2 <- co021$"641143"
names(flagged2) <- rownames(dat)
ccf.100.1 <- ccf.series.rwl(rwl = dat, seg.length = 100,
                            series = flagged2)
## Select series by name or column position
ccf.100.2 <- ccf.series.rwl(rwl = co021, seg.length = 100,
                            series = "641143")
ccf.100.3 <- ccf.series.rwl(rwl = co021, seg.length = 100,
                            series = which(colnames(co021) == "641143"))
identical(ccf.100.1, ccf.100.2) # TRUE
identical(ccf.100.2, ccf.100.3) # TRUE
}
}
\keyword{ manip }

\encoding{UTF-8}
\name{chron.ars}
\alias{chron.ars}
\title{Build ARSTAN Chronology}
\description{
  This function builds three varieties of the mean-value chronology, including 
  the ARSTAN chronology, typically from a
  \code{data.frame} of detrended ring widths as produced by
  \code{\link{detrend}}.
}
\usage{
chron.ars(x, biweight=TRUE, maxLag=10, firstAICmin=TRUE, 
  verbose=TRUE, prewhitenMethod=c("ar.yw","arima.CSS-ML"))
}
\arguments{
  \item{x}{a \code{data.frame} of (usually detrended) ring widths with
    \code{rownames(\var{x})} containing years and \code{colnames(x)}
    containing each series \acronym{ID} such as produced by
    \code{\link{read.rwl}}}
  \item{biweight}{\code{logical} flag.  If \acronym{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}.}}
  \item{maxLag}{an \code{integer} giving the maximum lag to consider in
    the pooled AR modeling. If the AIC does not reach a local minimum
    within this range when using \code{firstAICmin = TRUE}, the function
    stops with an error. Increase \code{maxLag} if this occurs.}
  \item{firstAICmin}{\code{logical} flag.  If \acronym{TRUE} the final
    AR order is selected using the first local AIC minimum; otherwise
    the order is selected by the overall minimum across all lags up to
    \code{maxLag}. See Details.}
  \item{verbose}{\code{logical} flag.  If \acronym{TRUE} the function
    prints information from the AR modeling to the screen.}
  \item{prewhitenMethod}{a \code{character} vector to determine the AR
    model fitting. See details below.  Possible values are either
    \code{"ar.yw"} or \code{"arima.CSS-ML"}. Can be abbreviated.
    Defaults to \code{"ar.yw"}.}
}
\details{
  This produces three mean-value chronologies: standard, residual, and
  ARSTAN. Users unfamiliar with the concept behind the ARSTAN method
  should look to Cook (1985) for background and inspiration.

  The standard chronology is the (biweight) mean value across rows and
  identical to \code{\link{chron}}.

  The residual chronology is the prewhitened chronology as described by
  Cook (1985) and uses multivariate autoregressive modeling to determine
  the order of the AR process. It is important to note that the residual
  chronology produced here is different from the simple residual
  chronology produced by \code{\link{chron}}, which returns the
  residuals of an AR process using a naive call to \code{\link{ar}}. In
  practice the results will be similar. For more on the residual
  chronology in this function, see pp. 153-154 in Cook's 1985
  dissertation.

  The ARSTAN chronology builds on the residual chronology but returns a
  re-whitened chronology where the pooled AR coefficients from the
  multivariate autoregressive modeling are reintroduced. See references
  for details.

  The order of the AR model is selected from the pooled AR coefficients
  by AIC using either the first local AIC minimum
  (\code{firstAICmin = TRUE}, the default) or the overall minimum across
  all lags up to \code{maxLag} (\code{firstAICmin = FALSE}). When
  \code{firstAICmin = TRUE} and the AIC decreases monotonically across
  all tested lags without reaching a local minimum, the function stops
  with an error. This can occur when \code{maxLag} is too small for the
  data or when the series have persistence structures that are not
  well-suited to this approach (e.g., non-stationary or very long-memory
  series). In such cases, consider increasing \code{maxLag} or reviewing
  the data preparation steps described in Cook (1985). When
  \code{firstAICmin = FALSE} the function always selects an order and
  does not stop in this way.

  If the selected AR order is zero (no detectable common autocorrelation
  among the series), no prewhitening or post-reddening is applied and
  all three returned chronologies are identical to the standard
  chronology.

  Once the AR order is determined an AR(p) model is fit to each series
  using either \code{\link{ar}} via the Yule-Walker method or by
  \code{\link{arima}} via conditional-sum-of-squares to find starting
  values, then maximum likelihood. It is possible that the model will
  not converge in which case a warning is produced. The AR fitting is
  determined via \code{prewhitenMethod} and defaults to using
  \code{\link{ar}}.
}
\value{
  A \code{data.frame} of class \code{"crn"} with the following columns:

  \item{std}{the standard chronology: the (biweight) mean ring-width
    index across all series for each year.}
  \item{res}{the residual chronology: the standard chronology after
    multivariate autoregressive prewhitening to remove common
    persistence.}
  \item{ars}{the ARSTAN chronology: the residual chronology with the
    pooled autoregression reintroduced. When the selected AR order is
    zero, \code{std}, \code{res}, and \code{ars} are identical.}
  \item{samp.depth}{the number of series with non-missing values in
    each year.}

  Row names are the years taken from \code{rownames(x)}.
}
\references{
  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Cook, E. R. (1985). A Time Series Analysis Approach to Tree Ring 
    Standardization. PhD thesis, The University of Arizona.
}
\author{ Andy Bunn with contributions from Kevin Achukaitis and Ed Cook.
         Much of the function is a port of Cook's FORTRAN code. }
\seealso{ \code{\link{chron}}, \code{\link{crn.plot}},
          \code{\link{ar}}, \code{\link{arima}} }

\examples{library(graphics)
library(utils)
data(co021)
co021.rwi <- detrend(rwl = co021, method = "AgeDepSpline")
co021.crn <- chron.ars(co021.rwi)
plot(co021.crn,xlab="Year",ylab="RWI",add.spline=TRUE,nyrs=20)
cor(co021.crn)
}
\keyword{ manip }

\encoding{UTF-8}
\name{redfit}
\alias{redfit}
\alias{runcrit}
\title{
  Red-Noise Spectra of Time-Series
}
\description{
  Estimate red-noise spectra from a possibly unevenly spaced time-series.
}
\usage{
redfit(x, t, tType = c("time", "age"), nsim = 1000, mctest = TRUE,
       ofac = 4, hifac = 1, n50 = 3, rhopre = NULL,
       p = c(0.10, 0.05, 0.02), iwin = 2,
       txOrdered = FALSE, verbose = FALSE, seed = NULL,
       maxTime = 10, nLimit = 10000)

runcrit(n, p = c(0.10, 0.05, 0.02), maxTime = 10, nLimit = 10000)
}
\arguments{

  \item{x}{ a \code{numeric} vector representing a possibly unevenly
    spaced time-series.  }

  \item{t}{ a \code{numeric} vector of times or ages corresponding to
    \code{\var{x}}.  If missing, this is set to an evenly spaced
    increasing integer sequence (1, 2, \ldots) along \code{\var{x}}.
    See \code{\var{txOrdered}}. }

  \item{tType}{ a \code{character} string indicating the type of the
    \code{\var{t}} vector: either times or ages.  }

  \item{nsim}{ a \code{numeric} value giving the number of simulated AR1
    spectra to compute.  }

  \item{mctest}{ a \code{logical} flag.  If \code{TRUE}, performs a
    Monte Carlo test for computing red noise false-alarm levels.  In
    that case, the result list will contain non-\code{NULL} elements
    \code{"ci80"}, \code{"ci90"}, \code{"ci95"} and \code{"ci99"}.  }

  \item{ofac}{ oversampling factor for Lomb-Scargle Fourier transform.
    A \code{numeric} value. }

  \item{hifac}{ maximum frequency to analyze relative to the Nyquist
    frequency.  A \code{numeric} value. }

  \item{n50}{ number of segments.  The segments overlap by about 50
    percent. }

  \item{rhopre}{ a \code{numeric} value giving the prescribed
    autocorrelation coefficient.  If \code{NULL} or negative, the
    autocorrelation coefficient will be estimated from the data. }

  \item{p}{ a \code{numeric} or \code{\link[gmp:bigrational]{bigq}} (if
    package \code{"gmp"} is installed) vector of significance levels for
    a statistical test considering the number of runs in a sequence.
    See \sQuote{Details}. }
  
  \item{iwin}{ the type of window used for scaling the values of each
    segment of data.  A \code{numeric} value or one of
    \code{"rectangular"}, \code{"welch i"}, \code{"hanning"},
    \code{"triangular"} and \code{"blackman-harris"}.  Integers
    \code{0:4} correspond to the character values, in this order.  The
    default \code{\var{iwin} = 2} corresponds to the \code{"hanning"}
    window. }

  \item{txOrdered}{ a \code{logical} flag.  If \code{TRUE}, it is
    assumed that \code{\var{t}} is in ascending order without
    duplicates.  If \code{FALSE} (the default), \code{\var{t}} will be
    sorted and \code{\var{x}} reordered accordingly.  Values of
    \code{\var{x}} at identical values of \code{\var{t}} are averaged.
    If duplicates are found, the non-duplicated \code{\var{t}} and
    averaged \code{\var{x}} will be included in the return value of the
    function.  }

  \item{verbose}{ a \code{logical} flag.  If \code{TRUE}, the function
    prints some information while operating.  }

  \item{seed}{ a value to be given to \code{set.seed} at the start of
    the function.  The value is also recorded in the list returned.  If
    not \code{NULL}, this can be used for reproducible results.  }

  \item{maxTime}{ a \code{numeric} value giving the approximate maximum
    time in seconds to use for computing the exact acceptance regions of
    the number of runs test.  See also \code{\var{nLimit}}. }

  \item{nLimit}{ a \code{numeric} value giving the maximum
    \code{\var{n}} for which \code{runcrit} will try to compute an exact
    solution to the acceptance regions of the number of runs test.
    Precomputed solutions may exist for larger \code{\var{n}}).  This
    limit is in place because a part of the exact solution always needs
    to be computed in order to roughly estimate the total time and
    whether it would exceed \code{\var{maxTime}}.  If
    \code{\var{nLimit}} is very large, \code{\var{maxTime}} may be
    (greatly) exceeded while computing the aforementioned part of the
    exact solution. }

  \item{n}{ an integral value giving the length of the sequence in the
    number of runs test. }

}

\details{

  Function \code{redfit} computes the spectrum of a possibly unevenly
  sampled time-series by using the Lomb-Scargle Fourier transform.  The
  spectrum is bias-corrected using spectra computed from simulated AR1
  series and the theoretical AR1 spectrum.

  The function duplicates the functionality of program REDFIT by Schulz
  and Mudelsee.  See the manual of that program for more information.
  The results of this function should be very close to REDFIT.  However,
  some changes have been made:

  \itemize{

    \item More precision is used in some constants and computations.

    \item All the data are used: the last segment always contains the
    last pair of (\var{t}, \var{x}).  There may be small differences
    between \code{redfit} and REDFIT with respect to the number of
    points per segment and the overlap of consecutive segments.

    \item The critical values of the runs test (see the description of
    \code{runcrit} below) differ between \code{redfit} and REDFIT.  The
    approximate equations in REDFIT produce values quite far off from
    the exact values when the number of frequencies is large.

    \item The user can select the significance levels of the runs test.

    \item Most of the window functions have been adjusted.

    \item 6 dB bandwidths have been computed for discrete-time windows.

  }

  Function \code{runcrit} computes the limits of the acceptance region
  of a number of runs test: assuming a sequence of \code{\var{n}} i.i.d.
  discrete random variables with two possible values \var{a} and \var{b}
  of equal probability (0.5), we are examining the distribution of the
  number of runs.  A run is an uninterrupted sequence of only \var{a} or
  only \var{b}.  The minimum number of runs is 1 (a sequence with only
  \var{a} or only \var{b}) while the maximum number is \code{\var{n}}
  (alternating \var{a} and \var{b}).  See Bradley, p. 253\enc{–}{--}254,
  259\enc{–}{--}263.  The function is also called from \code{redfit};
  see \code{\var{rcnt}} in \sQuote{Value} for the interpretation.  In
  this case the arguments \code{\var{p}}, \code{\var{maxTime}} and
  \code{\var{nLimit}} are passed from \code{redfit} to \code{runcrit},
  and \code{\var{n}} is the number of output frequencies.

  The results of \code{runcrit} have been essentially precomputed for
  some values of \code{\var{p}} and \code{\var{n}}.  If a precomputed
  result is not found and \code{\var{n}} is not too large
  (\code{\var{nLimit}}, \code{\var{maxTime}}), the exact results are
  computed on-demand.  Otherwise, or if package \code{"gmp"} is not
  installed, the normal distribution is used for approximation.

}

\value{

  Function \code{runcrit} returns a \code{list} containing
  \code{\var{rcritlo}}, \code{\var{rcrithi}} and \code{\var{rcritexact}}
  (see below).  Function \code{redfit} returns a \code{list} with the
  following elements:

  \item{varx }{ variance of \code{\var{x}} estimated from its spectrum.
    A \code{numeric} value. }

  \item{rho }{ average autocorrelation coefficient, either estimated
    from the data or prescribed (\code{\var{rhopre}}).  A \code{numeric}
    value. }

  \item{tau }{ the time scale of an AR1 process corresponding to
    \code{\var{rho}}.  A \code{numeric} value.  }

  \item{rcnt }{ a \code{numeric} value giving the number of runs to be
    used for a statistical test studying the difference between a
    theoretical AR1 spectrum and the bias-corrected spectrum estimated
    from the data.  Null hypothesis: the two spectra agree, i.e. the
    probability of either being larger than the other is 0.5 at every
    point.  Requires that \code{\var{iwin} == 0} (\code{"rectangular"}),
    \code{\var{ofac} == 1} and \code{\var{n50} == 1}.  Otherwise the
    test is not performed and \var{rcnt} is \code{NULL}.  }

  \item{rcritlo }{ a \code{numeric} vector of critical low values for
    the number of runs, i.e. the lowest value for accepting the null
    hypothesis at each level of significance \code{\var{p}}.  When
    returned from \code{redfit}, \code{NULL} when \code{\var{rcnt}} is
    \code{NULL}. }

  \item{rcrithi }{ a \code{numeric} vector of critical high values for
    the number of runs, i.e. the highest value for accepting the null
    hypothesis at each level of significance \code{\var{p}}.  When
    returned from \code{redfit}, \code{NULL} when \code{\var{rcnt}} is
    \code{NULL}. }
  
  \item{rcritexact }{ a \code{logical} vector specifying whether each
    pair of \code{\var{rcritlo}} and \code{\var{rcrithi}} are exact
    values (\code{TRUE}) or approximated from a normal distribution
    (\code{FALSE}).  When returned from \code{redfit}, \code{NULL} when
    \code{\var{rcnt}} is \code{NULL}. }

  \item{freq }{ the frequencies used.  A \code{numeric} vector.  The
    other numeric vectors have the same length, i.e. one value for each
    frequency studied. }

  \item{gxx }{ estimated spectrum of the data (\var{t}, \var{x}).  A
    \code{numeric} vector.  }

  \item{gxxc }{ red noise corrected spectrum of the data.  A
    \code{numeric} vector.  }

  \item{grravg }{ average AR1 spectrum over \code{\var{nsim}}
    simulations.  A \code{numeric} vector. }

  \item{gredth }{ theoretical AR1 spectrum.  A \code{numeric} vector.  }

  \item{corr }{ a \code{numeric} vector containing the by-frequency
    correction: \code{\var{gxxc}} equals \code{\var{gxx}} divided by
    \code{\var{corr}} (or multiplied by the inverse correction).  Also
    used for computing \code{\var{ci80}}, \code{\var{ci90}},
    \code{\var{ci95}} and \code{\var{ci99}}. }

  \item{ci80 }{ a \code{numeric} vector containing the bias-corrected
    80th percentile (by frequency) red noise spectrum.  Only if
    \code{\var{mctest}} is \code{TRUE}. }

  \item{ci90 }{ a \code{numeric} vector containing the 90th percentile
    red noise spectrum. }

  \item{ci95 }{ 95th percentile red noise spectrum. }

  \item{ci99 }{ 99th percentile red noise spectrum. }

  \item{call}{the call to the function.  See \code{\link{match.call}}.}

  \item{params }{

    A \code{list} with the following items:
    \describe{

      \item{np }{ number of points in the data.  }
      
      \item{nseg }{ number of points in each segment.  }
      
      \item{nfreq }{ number of frequencies in the results.  }
      
      \item{avgdt }{ average sampling interval.  }
      
      \item{df }{ difference between consecutive frequencies.  }
      
      \item{fnyq }{ average Nyquist frequency.  }
      
      \item{n50 }{ value of the \code{\var{n50}} argument.  }
      
      \item{ofac }{ value of the \code{\var{ofac}} argument.  }
      
      \item{hifac }{ value of the \code{\var{hifac}} argument.  }
      
      \item{segskip }{ the ideal, non-rounded difference between
        starting indices of consecutive segments.  }
      
      \item{iwin }{ value of the \code{\var{iwin}} argument.  If a
        \code{character} value was used, this contains the corresponding
        number used internally. }
      
      \item{nsim }{ value of the \code{\var{nsim}} argument.  }
      
      \item{mctest }{ value of the \code{\var{mctest}} argument.  }

      \item{rhopre }{ value of the \code{\var{rhopre}} argument.  }
      
    }
  }

  \item{vers}{version of dplR.  See \code{\link{packageVersion}}.}

  \item{seed }{ value of the \code{\var{seed}} argument.  }

  \item{t }{ if duplicated values of \code{\var{t}} are given, the
    non-duplicated \code{numeric} time or age vector (see
    \code{\var{tType}}) is returned here.  Otherwise \code{NULL}.  See
    \code{\var{txOrdered}}. }

  \item{x }{ if duplicated values of \code{\var{t}} are given, the
    averaged \code{numeric} data vector is returned here.  Otherwise
    \code{NULL}. }

}
\references{

  Function \code{redfit} is based on the Fortran program
  \href{https://www.marum.de/Prof.-Dr.-michael-schulz.html}{REDFIT}
  (version 3.8e), which is in the public domain.

  Bradley, J. V. (1968) \emph{Distribution-Free Statistical
    Tests}.  Prentice-Hall.
  
  Schulz, M. and Mudelsee, M. (2002) REDFIT: estimating red-noise
  spectra directly from unevenly spaced paleoclimatic time series.
  \emph{Computers & Geosciences}, \bold{28}(3), 421\enc{–}{--}426.

}
\author{
  Mikko Korpela.  Examples by Andy Bunn.
}
\seealso{
  \code{\link{print.redfit}}
}
\examples{
# Create a simulated tree-ring width series that has a red-noise
# background ar1=phi and sd=sigma and an embedded signal with 
# a period of 10 and an amplitude of have the rednoise sd.
library(graphics)
library(stats)
runif(1)
rs <- .Random.seed
set.seed(123)
nyrs <- 500
yrs <- 1:nyrs

# Here is an ar1 time series with a mean of 2mm,
# an ar1 of phi, and sd of sigma
phi <- 0.7
sigma <- 0.3
sigma0 <- sqrt((1 - phi^2) * sigma^2)
x <- arima.sim(list(ar = phi), n = nyrs, sd = sigma0) + 2

# Here is a sine wave at f=0.1 to add in with an amplitude
# equal to half the sd of the red noise background
per <- 10
amp <- sigma0 / 2
wav <- amp * sin(2 * pi / per * yrs)

# Add them together so we have signal and noise
x <- x + wav

# Here is the redfit spec
redf.x <- redfit(x, nsim = 500)

# Acceptance region of number of runs test
# (not useful with default arguments of redfit())
runcrit(length(redf.x[["freq"]]))

op <- par(no.readonly = TRUE) # Save to reset on exit
par(tcl = 0.5, mar = rep(2.2, 4), mgp = c(1.1, 0.1, 0))

plot(redf.x[["freq"]], redf.x[["gxxc"]],
     ylim = range(redf.x[["ci99"]], redf.x[["gxxc"]]),
     type = "n", ylab = "Spectrum", xlab = "Frequency (1/yr)",
     axes = FALSE)
grid()
lines(redf.x[["freq"]], redf.x[["gxxc"]], col = "#1B9E77")
lines(redf.x[["freq"]], redf.x[["ci99"]], col = "#D95F02")
lines(redf.x[["freq"]], redf.x[["ci95"]], col = "#7570B3")
lines(redf.x[["freq"]], redf.x[["ci90"]], col = "#E7298A")
freqs <- pretty(redf.x[["freq"]])
pers <- round(1 / freqs, 2)
axis(1, at = freqs, labels = TRUE)
axis(3, at = freqs, labels = pers)
mtext(text = "Period (yr)", side = 3, line = 1.1)
axis(2); axis(4)
legend("topright", c("x", "CI99", "CI95", "CI90"), lwd = 2,
       col = c("#1B9E77", "#D95F02", "#7570B3", "#E7298A"),
       bg = "white")
box()

\dontrun{
# Second example with tree-ring data
# Note the long-term low-freq signal in the data. E.g.,
# crn.plot(cana157)

library(utils)
data(cana157)
yrs <- time(cana157)
x <- cana157[, 1]
redf.x <- redfit(x, nsim = 1000)

plot(redf.x[["freq"]], redf.x[["gxxc"]],
     ylim = range(redf.x[["ci99"]], redf.x[["gxxc"]]),
     type = "n", ylab = "Spectrum", xlab = "Frequency (1/yr)",
     axes = FALSE)
grid()
lines(redf.x[["freq"]], redf.x[["gxxc"]], col = "#1B9E77")
lines(redf.x[["freq"]], redf.x[["ci99"]], col = "#D95F02")
lines(redf.x[["freq"]], redf.x[["ci95"]], col = "#7570B3")
lines(redf.x[["freq"]], redf.x[["ci90"]], col = "#E7298A")
freqs <- pretty(redf.x[["freq"]])
pers <- round(1 / freqs, 2)
axis(1, at = freqs, labels = TRUE)
axis(3, at = freqs, labels = pers)
mtext(text = "Period (yr)", side = 3, line = 1.1)
axis(2); axis(4)
legend("topright", c("x", "CI99", "CI95", "CI90"), lwd = 2,
       col = c("#1B9E77", "#D95F02", "#7570B3", "#E7298A"),
       bg = "white")
box()
par(op)
}
.Random.seed <- rs
}
\keyword{ ts }
\keyword{ htest }

\encoding{UTF-8}
\name{zof.rwl}
\docType{data}
\alias{zof.rwl}
\title{ European Beech Ring Widths from Zofingen, Switzerland }
\description{
  This data set includes ring-width measurements for European Beech (\emph{Fagus sylvatica}) increment cores collected at the Zofingen in Switzerland.  There are 61 series.
}
\usage{data(zof.rwl)}
\format{
  A \code{data.frame} containing 61 ring-width series in columns and 156
  years in rows.
}
\source{
  Contributed by Stefan Klesse
}
\references{
  Klesse, S., Babst, F., Lienert, S., Spahni, R., Joos, F., Bouriaud, O., Carrer, M., Filippo, A.D., Poulter, B., Trotsiuk, V., Wilson, R., Frank, D.C. (2018) A combined tree ring and vegetation model assessment of European forest growth sensitivity to interannual climate variability. \emph{Global Biogeochemical Cycles} 32, 1226 -- 1240.
}
\keyword{datasets}

\encoding{UTF-8}
\name{anos1}
\docType{data}
\alias{anos1}
\title{Rothenburg Tree Ring Widths}
\description{
  This data set gives the raw ring widths for Norway spruce \emph{Picea
  abies} at Rothenburg ob der Tauber, Bavaria, Germany.  There are 20
  series from 10 trees.  Data set was created using
  \code{\link{read.rwl}} and saved to an .rda file using
  \code{\link[base]{save}}.
}
\usage{data(anos1)}
\format{A \code{data.frame} containing 20 tree-ring series from 10 trees
  in columns and 98 years in rows.  The correct stc mask for use with
  \code{\link{read.ids}} is \code{c(5, 2, 1)}.}
\references{
  Zang, C. (2010) \emph{Growth reaction of temperate forest tree species
  to summer drought \enc{–}{--} a multispecies tree-ring network
  approach}.  Ph.D. thesis, Technische \enc{Universität}{Universitaet}
  \enc{München}{Muenchen}.
}
\keyword{datasets}

\encoding{UTF-8}
\name{skel.plot}
\alias{skel.plot}
\title{ Skeleton Plot }
\description{
  Automatically generates a skeleton plot of tree-ring data.
}
\usage{
skel.plot(rw.vec, yr.vec = NULL, sname = "", filt.weight = 9,
          dat.out = FALSE, master = FALSE, plot = TRUE)
}
\arguments{
  \item{rw.vec}{ a \code{numeric} vector of a tree-ring chronology or
    series }
  \item{yr.vec}{ optional \code{numeric} vector giving years for the
    tree-ring data }
  \item{sname}{ an optional \code{character} string giving the
    \acronym{ID} for the data.  The width of the string, often identical
    to the number of characters, must be less than 8. }
  \item{filt.weight}{ filter length for the Hanning filter, defaults to 9 }
  \item{dat.out}{ \code{logical} flag indicating whether to return a
    \code{data.frame} containing the data. }
  \item{master}{ \code{logical} flag indicating whether to make the plot
    with the segments inverted to ease pattern matching.  If \code{TRUE}
    the segments will be plotted from the top down and the labels on the
    x axes will be on the bottom. }
  \item{plot}{ \code{logical} flag indicating whether to make a plot. }
}
\details{
  This makes a skeleton plot \enc{–}{-} a plot that gives the relative growth for
  year \code{\var{t}} relative to years \code{\var{t}-1} and
  \code{\var{t}+1}.  Note that this plot is a standard plot in
  dendrochronology and typically made by hand for visually cross-dating
  series.  This type of plot might be confusing to those not accustomed
  to visual cross-dating.  See references for more information.  The
  implementation is based on Meko\enc{’}{'}s (2002) skeleton plotting approach.

  The skeleton plot is made by calculating departures from high
  frequency growth for each year by comparing year \code{\var{t}} to the
  surrounding three years
  (\code{\var{t}-1},\code{\var{t}},\code{\var{t}+1}).  Low frequency
  variation is removed using a \code{\link{hanning}} filter.  Relative
  growth is scaled from one to ten but only values greater than three
  are plotted.  This function\enc{’}{'}s primary effect is to create plot with
  absolute units that can be printed and compared to other plots.  Here,
  anomalous growth is plotted on a 2mm grid and up to 120 years are
  plotted on a single row with a maximum of 7 rows (840 years).  These
  plots are designed to be plotted on standard paper using an
  appropriate \code{device}, e.g., \code{postscript} with defaults or to
  \code{pdf} with plot width and height to accommodate a landscape plot,
  e.g., \code{\var{width} = 10}, \code{\var{height} = 7.5},
  \code{\var{paper} = "USr"}.  These plots are designed to be printable
  and cut into strips to align long series.  Statistical cross-dating is
  possible if the data are output but more easily done using the functions
  \code{\link{xskel.plot}} and \code{\link{xskel.ccf.plot}}.

}
\value{
  This function is invoked primarily for its side effect, which is to
  produce a plot.  If \code{\var{dat.out}} is \code{TRUE} then a
  \code{data.frame} is returned with the years and height of the
  skeleton plot segments as columns.
}
\references{

  Stokes, M. A. and Smiley, T. L. (1968) \emph{An Introduction to
    Tree-Ring Dating}.  The University of Arizona Press.
  \acronym{ISBN-13}: 978-0-8165-1680-3.
  
  Sheppard, P. R. (2002) Crossdating Tree Rings Using Skeleton Plotting.
  \url{https://www.ltrr.arizona.edu/skeletonplot/introcrossdate.htm}.
  
  Meko, D. (2002) Tree-Ring MATLAB Toolbox.
  \url{https://www.mathworks.com/matlabcentral/}.
  
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{  \code{\link{Devices}}, \code{\link{hanning}},
           \code{\link{xskel.plot}}, \code{\link{xskel.ccf.plot}}}

\examples{library(utils)
data(co021)
x <- co021[,33]
x.yrs <- time(co021)
x.name <- colnames(co021)[33]
## On a raw ring width series - undated
skel.plot(x)
## On a raw ring width series - dated with names
skel.plot(x, yr.vec = x.yrs, sname = x.name, master = TRUE)
\dontrun{
library(grDevices)
## Try cross-dating
y <- co021[, 11]
y.yrs <- time(co021)
y.name <- colnames(co021)[11]

## send to postscript - 3 pages total
fname1 <- tempfile(fileext=".ps")
print(fname1) # tempfile used for PS output
postscript(fname1)
## "Master series" with correct calendar dates
skel.plot(x, yr.vec = x.yrs, sname = x.name, master = TRUE)
## Undated series, try to align with last plot
skel.plot(y)
## Here's the answer...
skel.plot(y, yr.vec = y.yrs, sname = y.name)
dev.off()

unlink(fname1) # remove the PS file

## alternatively send to pdf
fname2 <- tempfile(fileext=".pdf")
print(fname2) # tempfile used for PDF output
pdf(fname2, width = 10, height = 7.5, paper = "USr")
skel.plot(x, yr.vec = x.yrs, sname = x.name, master = TRUE)
skel.plot(y)
skel.plot(y, yr.vec = y.yrs, sname = y.name)
dev.off()

unlink(fname2) # remove the PDF file
}
}
\keyword{ hplot }

\encoding{UTF-8}
\name{plot.crs}
\alias{plot.crs}
\title{
  Plotting crs Objects
}
\description{
  Plots objects returned from \code{\link{corr.rwl.seg}}.
}
\usage{
\method{plot}{crs}(x, ...)
}
\arguments{

  \item{x}{ An object of class \code{"crs"}.  }

  \item{\dots}{ Additional arguments passed to \code{\link{plot}}  }
  
}
\value{
  None.  A plot is produced.
}
\author{
  Andy Bunn
}
\seealso{
  \code{\link{corr.rwl.seg}}
}
\examples{
library(graphics)
data(co021)
foo <- corr.rwl.seg(co021, make.plot = FALSE)
plot(foo)
}
\keyword{ hplot }

\encoding{UTF-8}
\name{interseries.cor}
\alias{interseries.cor}
\title{ Individual Series Correlation Against a Master Chronology }
\description{
  This function calculates the correlation between a series and a master
  chronology.
}
\usage{
  interseries.cor(rwl,n=NULL,prewhiten=TRUE,biweight=TRUE,
    method = c("spearman", "pearson", "kendall"))
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}}. }
  \item{n}{ \code{NULL} or an integral value giving the filter length
    for the \code{\link{hanning}} filter used for removal of low
    frequency variation. }
  \item{prewhiten}{ \code{logical} flag.  If \code{TRUE} each series is
    whitened using \code{\link{ar}}. }
  \item{biweight}{ \code{logical} flag.  If \code{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}}.}
  \item{method}{Can be either \code{"pearson"}, \code{"kendall"}, or
    \code{"spearman"} which indicates the correlation coefficient to be
    used.  Defaults to \code{"spearman"}.  See \code{\link{cor.test}}. }
}
\details{
  This function calculates correlation serially between each tree-ring
  series and a master chronology built from all the other series in the
  \code{\var{rwl}} object (leave-one-out principle).

  Each series in the rwl object is optionally
  detrended as the residuals from a \code{\link{hanning}} filter with
  weight \code{\var{n}}.  The filter is not applied if \code{\var{n}} is
  \code{NULL}.  Detrending can also be done via prewhitening where the
  residuals of an \code{\link{ar}} model are added to each series
  mean.  This is the default.  The master chronology is computed as the
  mean of the \code{\var{rwl}} object using \code{\link{tbrm}} if
  \code{\var{biweight}} is \code{TRUE} and \code{rowMeans} if not.  Note
  that detrending can change the length of the series.  E.g., a
  \code{\link{hanning}} filter will shorten the series on either end by
  \code{floor(\var{n}/2)}.  The prewhitening default will change the
  series length based on the \code{\link{ar}} model fit.  The effects of
  detrending can be seen with \code{\link{series.rwl.plot}}.

  This function produces the same output of the \code{\var{overall}} portion of 
  \code{\link{corr.rwl.seg}}.  The mean correlation value given is sometimes 
  referred to as the \dQuote{overall interseries correlation} or the \dQuote{COFECHA 
  interseries correlation}.  This output differs from the \code{rbar} 
  statistics given by \code{\link{rwi.stats}} in that \code{rbar} is 
  the average pairwise correlation between series where this is the 
  correlation between a series and a master chronology.
}
\value{ a \code{data.frame} with correlation values and p-values given from
\code{\link{cor.test}}
}
\author{ Andy Bunn, patched and improved by Mikko Korpela  }
\seealso{ \code{\link{rwl.stats}}, \code{\link{rwi.stats}} }
\examples{library(utils)
data(gp.rwl)
foo <- interseries.cor(gp.rwl)
# compare to: 
# corr.rwl.seg(rwl=gp.rwl,make.plot=FALSE)$overall
# using pearson's r
foo <- interseries.cor(gp.rwl,method="pearson")

# two measures of interseries correlation
# compare interseries.cor to rbar from rwi.stats
gp.ids <- read.ids(gp.rwl, stc = c(0, 2, 1))
bar <- rwi.stats(gp.rwl, gp.ids, prewhiten=TRUE)
bar$rbar.eff
mean(foo[,1])

}
\keyword{ manip }

\encoding{UTF-8}
\name{chron.stabilized}
\alias{chron.stabilized}
\title{Build Mean Value Chronology with Stabilized Variance}
\description{
  This function builds a variance stabilized mean-value chronology, typically from a
  \code{data.frame} of detrended ring widths as produced by
  \code{\link{detrend}}.
}
\usage{
chron.stabilized(x, winLength, biweight = TRUE, running.rbar = FALSE)
}
\arguments{
  \item{x}{a \code{data.frame} of ring widths with
    \code{rownames(\var{x})} containing years and \code{colnames(x)}
    containing each series \acronym{ID} such as produced by
    \code{\link{read.rwl}}}
  \item{winLength}{a odd \code{integer} specifying the window length.}
  \item{biweight}{\code{logical} flag.  If \acronym{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}.}}
  \item{running.rbar}{\code{logical} flag.  If \acronym{TRUE} the running interseries correlation is returned as well. }
}
\details{
  The variance of a mean chronology depends on the variance of the individual samples, the number of series averaged together, and their interseries correlation (Wigley et al. 1984). As the number of series commonly decreases towards the beginning of a chronology averaging introduces changes in variance that are a solely an effect of changes in sample depth.
  
  Additionally, time-dependent changes in interseries correlation can cause artificial variance changes of the final mean chronology. The function \code{chron.stabilized} accounts for both temporal changes in the interseries correlation and sample depth to produce a mean value chronology with stabilized variance.
  
  The basic correction centers around the use of the effective independent sample size, \code{Neff}, which considers sample replication and mean interseries correlation between the samples at every time. This is defined as: \code{Neff = n(t) / 1+(n(t)-1)rbar(t)}

where \code{n(t)} is the number of series at time \code{t}, and \code{rbar} is the interseries correlation (see \code{\link{interseries.cor}}). Multiplication of the mean time series with the square root of \code{Neff} at every time \code{t} theoretically results in variance that is independent of sample size. In the limiting cases, when the \code{rbar} is zero or unity, \code{Neff} obtains values of the true sample size and unity, respectively.


}
\value{
  An object of of class \code{crn} and \code{data.frame} with the variance stabilized chronology, running interseries correlation (`if \code{running.rbar=TRUE}), and the sample depth.
}
\references{
  Frank, D, Esper, J, Cook, E, (2006) \emph{On variance adjustments in tree-ring chronology development}. Tree rings in archaeology, climatology and ecology, TRACE 4, 56–66
  
  Frank, D, Esper, J, Cook, E, (2007) \emph{Adjustment for proxy number and coherence in a large-scale temperature reconstruction}. Geophysical Research Letters 34
  
  Wigley, T, Briffa K, Jones P (1984) \emph{On the Average Value of Correlated Time Series, with Applications in Dendroclimatology and Hydrometeorology}. J. Climate Appl. Meteor., 23, 201–213
  }
\author{ Original code by David Frank and adapted for dplR by Stefan Klesse. Patched and improved by Andy Bunn. 
}
\seealso{ \code{\link{chron}} 
}
\examples{library(graphics)
library(utils)
data(co021)
co021.rwi <- detrend(co021,method = "Spline")
co021.crn <- chron(co021.rwi)
co021.crn2 <- chron.stabilized(co021.rwi,
                                winLength=101,
                                biweight = TRUE,
                                running.rbar = FALSE)
yrs <- time(co021)
plot(yrs,co021.crn$std,type="l",col="grey")
lines(yrs,co021.crn2$adj.crn,col="red")
}
\keyword{ manip }

\encoding{UTF-8}
\name{latexDate}
\alias{latexDate}
\title{
  Date Conversion to Character in LaTeX Format
}
\description{
  This is a simple convenience function that returns a date in the
  format used by \samp{\today} in LaTeX.  A possible use case is fixing
  the date shown in a vignette at weaving time.
}
\usage{
latexDate(x = Sys.Date(), ...)
}
\arguments{
  \item{x}{ any object for which an \code{as.POSIXlt} method exists.
    Defaults to the current date. }
  \item{\dots}{ other arguments to \code{as.POSIXlt} }
}
\value{
  A \code{character} vector
}
\author{
  Mikko Korpela
}
\examples{
latexDate()                              # today
latexDate(Sys.Date() + 5)                # today + 5 days
latexDate(c("2013-12-06", "2014-09-19")) # fixed dates
## [1] "December 6, 2013"   "September 19, 2014"
latexDate(5*60*60*24, origin=Sys.Date()) # today + 5 days
}
\keyword{ utilities }

\encoding{UTF-8}
\name{rasterPlot}
\alias{rasterPlot}
\title{
  Add Raster Elements to Plot
}
\description{
  This function takes plotting commands and uses a temporary
  bitmap graphics device to capture their output.  The
  resulting raster image is drawn in the plot or figure region of the
  active high-level plot.  A new plot is started if one does not exist.
}
\usage{
rasterPlot(expr, res = 150, region = c("plot", "figure"), antialias,
           bg = "transparent", interpolate = TRUE, draw = TRUE,
           Cairo = FALSE, \dots)
}
\arguments{
  \item{expr}{
    Low-level plotting commands (\code{lines}, \code{points},
    \code{text}, \ldots) representing elements to be added to the
    current plot.  A \code{\link{call}} or an \code{\link{expression}}.
  }
  \item{res}{
    Resolution in points per inch (ppi).  A numeric value.  Suggested
    values for different types of display media are given in
    \code{\link{compactPDF}}.  For example, the default 150 ppi
    corresponds to \dQuote{ebook} quality.
  }
  \item{region}{
    The function can draw in the \code{"plot"} region or the
    \code{"figure"} region which also includes \code{"mar"} margins (see
    \code{\link{par}}).  If the drawing commands in \code{\var{expr}}
    contain no material for the margins, the default \code{"plot"} is
    optimal.  Plotting in the outer margins (\code{"oma"} in
    \code{\link{par}}) is not supported.
  }
  \item{antialias}{
    Antialiasing argument passed to \code{\link{png}}.  The default
    (missing argument) is probably good for line plots but
    \code{"none"} is preferred for images in which color signifies value
    of data.  Unused if a \code{\link[Cairo]{Cairo}} device is used instead of
    \code{png}.
  }
  \item{bg}{
    Background color of the raster plot, an argument passed to
    the bitmap device.  If the default \code{"transparent"} does not
    work, try \code{"white"} or another color.  Note that a
    non-transparent background will mask any previous content in the
    figure or plot region, depending on the value of \code{\var{region}}.
  }
  \item{interpolate}{
    Argument passed to \code{\link{rasterImage}}.  A \code{logical}
    flag.  The default is \code{TRUE}: use linear interpolation.
    Analogously to \code{\var{antialias}}, \code{FALSE} is preferred
    when color maps to value.
  }
  \item{draw}{
    A \code{logical} flag.  Draw the results (\code{TRUE}, the default)
    or return an image object (\code{FALSE})?
  }
  \item{Cairo}{
    A \code{logical} flag.  \code{TRUE} for preferring a
    \code{\link[Cairo]{Cairo}} to \code{\link{png}} as the bitmap device,
    \code{FALSE} (the default) for the opposite.  If the preferred
    device cannot be used, the other one will be tried.
  }
  \item{\dots}{
    Other arguments to \code{\link{png}} or \code{\link[Cairo]{Cairo}}.
  }
}
\details{
  The appropriate graphical parameters of the current graphics device
  are copied to the temporary bitmap device.  Therefore the
  appearance of the raster contents should be almost the same as when
  directly drawn.

  The call or expression \code{\var{expr}} is evaluated in the
  environment of the caller.

  It is possible that the raster contents will maintain a constant size
  when the graphics device is resized.  If resizing works, however, the
  image may become distorted.  For example, circle symbols will turn
  into ellipses if the width to height ratio is not maintained (see
  \sQuote{Examples}).  This is in contrast to a standard plot in a
  display graphics device, e.g. \code{\link{x11}}, where text and
  symbols maintain their size when the device is resized.
}
\value{
  If \code{\var{draw}} is \code{TRUE}, there is no return value.  The
  function is used for the side effects.

  If \code{\var{draw}} is \code{FALSE}, an object of class
  \code{"nativeRaster"} is returned.  The object can be used as input
  for \code{\link{rasterImage}} or \code{\link{grid.raster}}.  See
  \code{\link[png]{readPNG}}.  If no bitmap device is available
  (see \sQuote{Note}), \code{NULL} is returned.
}
\author{
  Mikko Korpela
}
\note{

  \itemize{

    \item{The graphics device used for the output must have support for
      including raster images.  See \code{"rasterImage"} in
      \code{\link{dev.capabilities}}.}

    \item{The \R build must have a functional \code{\link{png}} device,
      which requires one of the following \code{\link{capabilities}}:
      \code{"png"}, \code{"aqua"} or \code{"cairo"}.  Alternatively, a
      \code{\link[Cairo]{Cairo}} device from package Cairo must be
      available with \code{\link[Cairo]{Cairo.capabilities}}
      \code{"raster"} or \code{"png"}.}

  }

  If either of these requirements is not met, at least one
  \code{\link{message}} is generated and the function reverts to regular
  plotting.  The \code{\var{bg}} argument is then handled by drawing a
  filled rectangle.  Also \code{\var{region}} is honored, but the other
  settings do not apply.
}
\examples{library(graphics)
library(stats)

## Picture with various graphical elements
x <- 1:100
y0 <- quote(sin(pi * x / 20) + x / 100 + rnorm(100, 0, 0.2))
y <- eval(y0)
ylab <- deparse(y0)
spl <- smooth.spline(y)
plot(x, y, type = "n", axes = FALSE, ylab = ylab)
usr <- par("usr")
xrange <- usr[2] - usr[1]
xsize <- xrange * 0.4
nsteps <- 8
xmar <- xsize / 20
yrange <- usr[4] - usr[3]
ysize <- yrange / 20
ymar <- 0.5 * ysize
X <- seq(usr[1] + xmar, by = xsize / nsteps, length.out = nsteps + 1)
xleft <- X[-(nsteps + 1)]
xright <- X[-1]
pin <- par("pin")
maxrad <- xsize / 3 * min(1, pin[2] / pin[1])
nrad <- 16
minrad <- maxrad / nrad
Rad <- seq(maxrad, by = (minrad - maxrad) / (nrad - 1), length.out=nrad)
xmar2 <- xmar + maxrad
ymar2 <- (xmar2 / xrange) * pin[1] / pin[2] * yrange
expr <- quote({
    rect(xleft, usr[4] - 1.5 * ysize, xright, usr[4] - ymar,
         col = rainbow(8), border = NA)
    symbols(rep(usr[2] - xmar2, nrad), rep(usr[3] + ymar2, nrad),
            circles = Rad, inches = FALSE, add = TRUE, fg = NA,
            bg = gray.colors(nrad + 1, 1, 0)[-1])
    points(y)
    lines(spl)
})
rasterPlot(expr, res = 50)
box()
axis(1)
axis(2)

## The same picture with higher resolution but no antialiasing
plot(y, type = "n", axes = FALSE, ann = FALSE)
## No content in margin, but region = "figure" and bg = "white"
## paints margin white
rasterPlot(expr, antialias = "none", interpolate = FALSE,
           region = "figure", bg = "white")
## Draw box, axes, labels
parnew <- par(new = TRUE)
plot(x, y, type = "n", ylab = ylab)
par(parnew)

## Draw plot(1:5) with adjusted margins and additional axes.  Some parts
## are drawn with rasterPlot, others normally.  Resize to see stretching.
op <- par(no.readonly = TRUE)
par(mar = c(5.1, 4.1, 2.1, 2.1))
plot(1:5, type = "n", axes = FALSE, ann = FALSE)
expr2 <- quote({
    points(c(2, 4), c(2, 4))
    axis(2)
    axis(3)
})
rasterPlot(expr2, region = "figure", bg = "white")
points(c(1, 3, 5), c(1, 3, 5))
box()
axis(1)
axis(4)
title(xlab = "Index", ylab = "1:5")
par(op)
}
\keyword{ aplot }
\keyword{ utilities }

\encoding{UTF-8}
\name{tridas.vocabulary}
\alias{tridas.vocabulary}
\title{
  Browse and Check Standard TRiDaS Vocabulary
}
\description{
  This function can be used to browse the TRiDaS vocabulary by category.
}
\usage{
tridas.vocabulary(category = c("dating type", "measuring method",
                      "shape", "location type", "variable", "unit",
                      "remark", "dating suffix", "presence / absence",
                      "complex presence / absence", "certainty"),
                  idx = NA, term = NA, match.exact = FALSE)
}
\arguments{
  \item{category}{Vocabulary category as a \code{character} vector of
    length one.  One of \code{"dating type"}, \code{"measuring method"},
    \code{"shape"}, \code{"location type"}, \code{"variable"},
    \code{"unit"}, \code{"remark"}, \code{"dating suffix"}, 
    \code{"presence / absence"}, \code{"complex presence / absence"},
    \code{"certainty"}.  Partial matches are allowed.}
  \item{idx}{A \code{numeric} vector.  Index to the \code{character}
    vector containing the vocabulary of the given category.}
  \item{term}{A \code{character} vector.  One or more (partial) terms to
    look for in the given category.}
  \item{match.exact}{A \code{logical} value.  If \code{TRUE}, partial
    matching of \code{\var{term}} is not used.  Defaults to
    \code{FALSE}.}
}
\details{
  The Tree Ring Data Standard (TRiDaS) is described in Jansma et. al
  (2010).  

  The function has four usage modes:
  \enumerate{

    \item When \code{\var{idx}} is given, returns item number
    \code{\var{idx}} in the given \code{\var{category}}.  There may be
    several numbers in \code{\var{idx}}, in which case multiple items
    are returned.

    \item When \code{\var{term}} contains one or more items and
    \code{\var{match.exact}} is \code{TRUE}, checks whether any of the
    terms is an exact match in the given \code{\var{category}}

    \item When \code{\var{term}} contains one or more items and
    \code{\var{match.exact}} is \code{FALSE}, expands partial matches of
    the terms in the vocabulary of the given \code{\var{category}}

    \item When only \code{\var{category}} is given, returns the complete
    vocabulary in the given \code{\var{category}}

  }
}
\value{

  \item{In mode 1}{A \code{character} vector, same length as in
    \code{\var{idx}}}

  \item{In mode 2}{A \code{logical} value}

  \item{In mode 3}{A \code{character} vector, same length as in
    \code{\var{term}}}

  \item{In mode 4}{A \code{character} vector}

}
\references{
  Jansma, E., Brewer, P. W., and Zandhuis, I. (2010) TRiDaS 1.1: The
  tree-ring data standard.  \emph{Dendrochronologia}, \bold{28}(2),
  99--130.
}
\author{
  Mikko Korpela
}

\seealso{
  \code{\link{read.tridas}}, \code{\link{write.tridas}}
}
\examples{## Show all entries in category "measuring method"
tridas.vocabulary(category = "measuring")

## Show item number one in category "complex presence / absence"
tridas.vocabulary(category = "complex", idx = 1)

## Check whether "half section" exists in category "shape"
tridas.vocabulary(category = "shape", term = "half section",
                  match.exact = TRUE)

## Return unabbreviated matches to several queries in category "remark"
tridas.vocabulary(category = "remark",
                  term = c("trauma", "fire", "diffuse"))
}
\keyword{ utilities }

\encoding{UTF-8}
\name{write.compact}
\alias{write.compact}
\title{ Write \acronym{DPL} Compact Format Ring Width File }
\description{
  This function writes a chronology to a \acronym{DPL} compact format
  file.
}
\usage{
write.compact(rwl.df, fname, append = FALSE, prec = 0.01,
              mapping.fname = "", mapping.append = FALSE, \dots)
}
\arguments{
  \item{rwl.df}{ a \code{data.frame} containing tree-ring ring widths
    with the series in columns and the years as rows.  The series
    \acronym{ID}s are the column names and the years are the row
    names.  This type of \code{data.frame} is produced by
    \code{\link{read.rwl}} and \code{\link{read.compact}}. }  
  \item{fname}{ a \code{character} vector giving the file name of the
    rwl file. }
  \item{append}{ \code{logical} flag indicating whether to append this
    chronology to an existing file. }
  \item{prec}{ \code{numeric} indicating the precision of the output
    file.  This must be equal to either 0.01 or 0.001 (units are in
    mm). }
  \item{mapping.fname}{ a \code{character} vector of length one giving
    the file name of an optional output file showing the mapping between
    input and output series \acronym{ID}s.  The mapping is only printed
    for those \acronym{ID}s that are transformed.  An empty name (the
    default) disables output. }
  \item{mapping.append}{ \code{logical} flag indicating whether to
    append the description of the altered series \acronym{ID}s to an
    existing file.  The default is to create a new file. }
  \item{\dots}{ Unknown arguments are accepted but not used. }
}
\details{
  The output should be readable by the Dendrochronology Program Library
  (\acronym{DPL}) as a compact format file.

  In series \acronym{ID}s, letters of the English alphabet and numbers
  are allowed.  Other characters will be removed.  The length of the
  \acronym{ID}s is limited to about 50 characters, depending on the
  length of the other items to be placed on the header lines of the
  output file.  Longer \acronym{ID}s will be truncated.  Also any
  duplicate \acronym{ID}s will be automatically edited so that only
  unique \acronym{ID}s exist.  If series \acronym{ID}s are changed, one
  or more warnings are shown.  In that case, the user may wish to print a
  list of the renamings (see Arguments).
}
\value{
  \code{\var{fname}}
}
\author{ Mikko Korpela, based on write.tucson by Andy Bunn }
\seealso{ \code{\link{write.rwl}}, \code{\link{write.tucson}},
  \code{\link{write.tridas}}, \code{\link{read.compact}}
}
\examples{library(utils)
data(co021)
fname <- write.compact(rwl.df = co021,
                       fname = tempfile(fileext=".rwl"),
                       append = FALSE, prec = 0.001)
print(fname) # tempfile used for output

unlink(fname) # remove the file
}
\keyword{ IO }

\encoding{UTF-8}
\name{read.fh}
\alias{read.fh}
\title{ Read Heidelberg Format Ring Width File }
\description{
  This function reads in a Heidelberg (block or column) format file of
  ring widths (.fh).
}
\usage{
read.fh(fname, BC_correction = FALSE)
}
\arguments{
  \item{fname}{ a \code{character} vector giving the file name of the fh
    file. }
  \item{BC_correction}{ a \code{logical}. If TRUE the correction moves BC-dates one year forward. }
}
\details{
  This reads in a fh-file with ring widths in blocks (decadal format) or
  in columns (e.g., as with comment flags) as used by TSAP program.
  Chronologies or half-chronos in fh-format are not supported.
}
\value{
  An object of class \code{c("rwl", "data.frame")} with the series in
  columns and the years as rows.  The keycodes are the column names and
  the years are the row names.  Depending on metadata available in the
  input file, the following attributes may be present in the
  \code{data.frame}:

  \item{ids}{A \code{data.frame} identifying the series.  Always contains
    columns \code{"tree"} and \code{"core"}, may contain columns
    \code{"site"}, \code{"radius"}, and \code{"stemDisk"}.  All
    columns are \code{numeric}.}
  \item{po}{A \code{data.frame} containing pith offsets, formatted for
    use in \code{\link{rcs}}.}
}
\author{ Christian Zang.  New features and patches by Mikko Korpela and Ronald Visser. }
\seealso{ \code{\link{read.rwl}} }

\references{ Rinn, F. (2003) \emph{TSAP-Win User Reference Manual.}
  Rinntech, Heidelberg.  \url{https://rinntech.info/products/tsap-win/}.
}
\keyword{ IO }

\encoding{UTF-8}
\name{write.tucson}
\alias{write.tucson}
\title{ Write Tucson Format Chronology File }
\description{
  This function writes a chronology to a Tucson (decadal) format file.
}
\usage{
write.tucson(rwl.df, fname, header = NULL, append = FALSE,
             prec = 0.01, mapping.fname = "", mapping.append = FALSE,
             long.names = FALSE, \dots)
}
\arguments{
  \item{rwl.df}{ a \code{data.frame} containing tree-ring ring widths
    with the series in columns and the years as rows.  The series
    \acronym{ID}s are the column names and the years are the row
    names.  This type of \code{data.frame} is produced by
    \code{\link{read.rwl}} and \code{\link{read.compact}}. }
  \item{fname}{ a \code{character} vector giving the file name of the
    rwl file. }
  \item{header}{ a \code{list} giving information for the header of the
    file.  If \code{NULL} then no header information will be written. }
  \item{append}{ \code{logical} flag indicating whether to append this
    chronology to an existing file.  The default is to create a new
    file. }
  \item{prec}{ \code{numeric} indicating the precision of the output
    file.  This must be equal to either 0.01 or 0.001 (units are in
    mm). }
  \item{mapping.fname}{ a \code{character} vector of length one giving
    the file name of an optional output file showing the mapping between
    input and output series \acronym{ID}s.  The mapping is only printed
    for those \acronym{ID}s that are transformed.  An empty name (the
    default) disables output. }
  \item{mapping.append}{ \code{logical} flag indicating whether to
    append the description of the altered series \acronym{ID}s to an
    existing file.  The default is to create a new file. }
  \item{long.names}{ \code{logical} flag indicating whether to allow
    long series \acronym{ID}s (7 or 8 characters) to be written to the
    output.  The default is to only allow 6 characters. }
  \item{\dots}{ Unknown arguments are accepted but not used. }
}
\details{
  This writes a standard rwl file as defined according to the standards
  of the \acronym{ITRDB} at
  \url{https://www.ncei.noaa.gov/pub/data/paleo/treering/treeinfo.txt}.  This is the
  decadal or Tucson format.  It is an \acronym{ASCII} file and machine
  readable by the standard dendrochronology programs.  Header information
  for the rwl can be written according to the International Tree Ring
  Data Bank (\acronym{ITRDB}) standard.  The header standard is not very
  reliable however and should be thought of as experimental here.  Do not
  try to write headers using dplR to submit to the \acronym{ITRDB}.  When
  submitting to the \acronym{ITRDB}, you can enter the metadata via
  their website.  If you insist however, the header information is given
  as a \code{list} and must be formatted with the following:
  \tabular{lrrr}{
    \emph{Description} \tab \emph{Name} \tab \emph{Class} \tab \emph{Max Width} \cr
    Site \acronym{ID} \tab \code{\var{site.id}} \tab \code{character} \tab 5 \cr
    Site Name \tab \code{\var{site.name}} \tab \code{character} \tab 52 \cr
    Species Code \tab \code{\var{spp.code}} \tab \code{character} \tab 4 \cr
    State or Country \tab \code{\var{state.country}} \tab \code{character} \tab 13 \cr
    Species \tab \code{\var{spp}} \tab \code{character} \tab 18 \cr
    Elevation \tab \code{\var{elev}} \tab \code{character} \tab 5 \cr
    Latitude \tab \code{\var{lat}} \tab \code{character} or \code{numeric} \tab 5 \cr
    Longitude \tab \code{\var{long}} \tab \code{character} or \code{numeric} \tab 5 \cr
    First Year \tab \code{\var{first.yr}} \tab \code{character} or \code{numeric} \tab 4 \cr
    Last Year \tab \code{\var{last.yr}} \tab \code{character} or \code{numeric} \tab 4 \cr
    Lead Investigator \tab \code{\var{lead.invs}} \tab \code{character} \tab 63 \cr
    Completion Date \tab \code{\var{comp.date}} \tab \code{character} \tab 8 \cr
  }
  See examples for a correctly formatted header list.  If the width of
  the fields is less than the max width, then the fields will be padded
  to the right length when written.  Note that \code{\var{lat}} and
  \code{\var{long}} are really \code{\var{lat} * 100} or
  \code{\var{long} * 100} and given as integral values.  E.g., 37 degrees
  30 minutes would be given as 3750.

  Series can be appended to the bottom of an existing file with a second
  call to \code{write.tucson}.  The output from this file is suitable for
  publication on the \acronym{ITRDB}.

  The function is capable of altering excessively long and/or duplicate
  series \acronym{ID}s to fit the Tucson specification.  Additionally,
  characters other than numbers or English letters will be removed.  If
  series \acronym{ID}s are changed, one or more warnings are shown.  In
  that case, the user may wish to print a list of the renamings (see
  Arguments).

  Setting \code{\var{long.names} = TRUE} allows series \acronym{ID}s to
  be 8 characters long, or 7 in case there are year numbers using 5
  characters.  Note that in the latter case the limit of 7 characters
  applies to all \acronym{ID}s, not just the one corresponding to the
  series with long year numbers.  The default (\code{\var{long.names} =
  FALSE}) is to allow 6 characters.  Long \acronym{ID}s may cause
  incompatibility with other software.
}
\value{
  \code{\var{fname}}
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{write.crn}}, \code{\link{read.tucson}},
  \code{\link{write.rwl}}, \code{\link{write.compact}},
  \code{\link{write.tridas}}
}
\examples{library(utils)
data(co021)
co021.hdr <- list(site.id = "CO021",
                  site.name = "SCHULMAN OLD TREE NO. 1, MESA VERDE",
                  spp.code = "PSME", state.country = "COLORADO",
                  spp = "DOUGLAS FIR", elev = "2103M", lat = 3712,
                  long = -10830, first.yr = 1400, last.yr = 1963,
                  lead.invs = "E. SCHULMAN", comp.date = "")
fname <- write.tucson(rwl.df = co021, fname = tempfile(fileext=".rwl"),
                      header = co021.hdr, append = FALSE, prec = 0.001)
print(fname) # tempfile used for output

unlink(fname) # remove the file
}
\keyword{ IO }

\encoding{UTF-8}
\name{dplR-defunct}
\alias{dplR-defunct}
\alias{plotRings}
\alias{ffcsaps}
\title{
  Defunct functions in dplR
}
\description{
  These are functions no longer included in dplR
}
\details{
  \tabular{ll}{
    Function: \tab plotRings\cr
    Function: \tab ffcsaps, use \code{\link{caps}} instead\cr
  }
}
\keyword{ help }

\encoding{UTF-8}
\name{po.to.wc}
\alias{po.to.wc}
\title{
  Convert Pith Offset to Wood Completeness
}
\description{
  This function creates a partial wood completeness data structure based
  on pith offset data.
}
\usage{
po.to.wc(po)
}
\arguments{
  \item{po}{
    A \code{data.frame} containing two variables.  Variable one
    (\code{\var{series}}) gives the series \acronym{ID} as either
    \code{character}s or \code{factor}s.  Variable two
    (\code{\var{pith.offset}}) contains integral values and gives the
    years from the beginning of the core to the pith (or center) of the
    tree.  The minimum value is 1.
  }
}
\details{
  Uses \code{\var{pith.offset} - 1} as the number of missing heartwood
  rings.
}
\value{
  A \code{data.frame} containing one variable of wood completeness data:
  \code{\var{n.missing.heartwood}} (\code{integer} type).  This can be
  used as input to \code{\link{write.tridas}}.
}
\author{
  Mikko Korpela
}
\seealso{
  \code{\link{wc.to.po}}, \code{\link{rcs}}, \code{\link{write.tridas}}
}
\examples{
\dontrun{
library(utils)
data(gp.po)
all(wc.to.po(po.to.wc(gp.po)) == gp.po)
}
}
\keyword{ manip }

\encoding{UTF-8}
\name{time.rwl}
\alias{time.rwl}
\alias{time.crn}
\alias{time<-}
\alias{time<-.rwl}
\alias{time<-.crn}

\title{
  Retrieve or set the time values for rwl and crn objects
}
\description{
  Retrieve or set the time values for rwl and crn objects.
}
\usage{
\method{time}{rwl}(x, \dots)
\method{time}{crn}(x, \dots)
time(x) <- value

}
\arguments{

  \item{x}{ An object of class \code{"rwl"} or an aboject of class \code{"crn"}.  }

  \item{\dots}{ Not used.  }
  
  \item{value}{ A \code{numeric} vector to assign the time. }

}

\value{
  A \code{numeric} vector of time (typically in years) for the object. This is done via \code{as.numeric(rownames(x))} but has been asked for by users so many times that it is being included as a function.
}
\author{
  Andy Bunn
}
\seealso{
  \code{\link{read.rwl}}
  \code{\link{read.crn}}
}
\examples{
library(utils)
data(co021)
# extract years
co021.yrs <- time(co021)
# set years -- silly example
time(co021) <- co021.yrs+100
}
\keyword{ manip }

\encoding{UTF-8}
\name{chron.ci}
\alias{chron.ci}
\title{Build Mean Value Chronology with Confidence Intervals}
\description{
  This function builds a mean value chronology with bootstrapped confidence 
  intervals, typically from a \code{data.frame} of detrended ring widths as produced by
  \code{\link{detrend}}.
}
\usage{
chron.ci(x, biweight=TRUE, conf=0.95, R=100)
}
\arguments{
  \item{x}{a \code{data.frame} of (usually detrended) ring widths with
    \code{rownames(\var{x})} containing years and \code{colnames(x)}
    containing each series \acronym{ID} such as produced by
    \code{\link{read.rwl}}}
  \item{biweight}{\code{logical} flag.  If \acronym{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}.}}
  \item{conf}{\code{numeric}. A scalar for the confidence level.}
  \item{R}{\code{integer}. The number of bootstrap replicates. }
}
\details{
  This either averages the rows of the \code{data.frame} using a mean or
  a robust mean (the so-called standard chronology) and calculates boostrapped confidence intervals using the normal approximation. The function will fail if there are any rows in \code{x} that contain only one sample and in practice there should be several samples in a row. One of the guiding principles of bootstrapping is that the population is to the sample as the sample is to the bootstrap samples.
}
\value{
  An object of of class \code{data.frame} with the standard chronology, the upper and lower confidence interval, and the sample depth. The years are stored as row numbers.
}

\author{ Andy Bunn.}
\seealso{ \code{\link{read.rwl}}, \code{\link{detrend}},
  \code{\link[boot]{boot}}, \code{\link[boot]{boot.ci}}
}
\examples{library(graphics)
library(utils)
data(wa082)
# truncate to a sample depth of five
wa082Trunc <- wa082[rowSums(!is.na(wa082))>4,]
# detrend
wa082RWI <- detrend(wa082Trunc, method="AgeDepSpline")
# bootstrap the chronology and
wa082Crn <- chron.ci(wa082RWI, biweight = TRUE, R = 100, conf = 0.99)
head(wa082Crn)
# plot (this is so much easier in ggplot!)
wa082Crn$yrs <- time(wa082Crn)
xx <- c(wa082Crn$yrs,rev(wa082Crn$yrs))
yy <- c(wa082Crn$lowerCI,rev(wa082Crn$upperCI))
plot(wa082Crn$yrs,wa082Crn$std,type="n",ylim=range(yy),
     ylab="RWI",xlab="Year",main="Chronology with CI")
polygon(x=xx,y=yy,col = "grey",border = NA)
lines(wa082Crn$yrs,wa082Crn$std)
}
\keyword{ manip }

\encoding{UTF-8}
\name{series.rwl.plot}
\alias{series.rwl.plot}
\title{ Plot Series and a Master }
\description{
  Plots a tree-ring series with a master chronology and displays
  their fit, segments, and detrending options in support of the
  cross-dating functions.
}
\usage{
series.rwl.plot(rwl, series, series.yrs = as.numeric(names(series)),
                seg.length = 100, bin.floor = 100, n = NULL,
                prewhiten = TRUE, biweight = TRUE, floor.plus1 = FALSE)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as rows
    such as that produced by \code{\link{read.rwl}}. }
  \item{series}{ a \code{numeric} or \code{character} vector.  Usually a
    tree-ring series.  If the length of the value is 1, the
    corresponding column of \code{\var{rwl}} is selected (by name or
    position) as the series and ignored when building the master
    chronology.  Otherwise, the value must be \code{numeric}. }
  \item{series.yrs}{ a \code{numeric} vector giving the years of
    \code{\var{series}}.  Defaults to
    \code{as.numeric(names(\var{series}))}.  Ignored if
    \code{\var{series}} is an index to a column of \code{\var{rwl}}. }
  \item{seg.length}{ an even integral value giving length of segments in
    years (e.g., 20, 50, 100 years). }
  \item{bin.floor}{ a non-negative integral value giving the base for
    locating the first segment (e.g., 1600, 1700, 1800
    \acronym{AD}).  Typically 0, 10, 50, 100, etc. }
  \item{n}{ \code{NULL} or an integral value giving the filter length
    for the \code{\link{hanning}} filter used for removal of low
    frequency variation. }
  \item{prewhiten}{ \code{logical} flag.  If \code{TRUE} each series is
    whitened using \code{\link{ar}}. }
  \item{biweight}{ \code{logical} flag.  If \code{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}}. }
  \item{floor.plus1}{ \code{logical} flag.  If \code{TRUE}, one year is
    added to the base location of the first segment (e.g., 1601, 1701,
    1801 \acronym{AD}). }
}
\details{
  The function is typically invoked to produce four plots showing the
  effect of the detrending options \code{\var{n}} and
  \code{\var{prewhiten}} and the binning options \code{\var{seg.length}}
  and \code{\var{bin.floor}}.

  \describe{
    \item{Plot 1}{Time series plot of the filtered series and the
      master}
    \item{Plot 2}{Scatterplot of series vs. master}
    \item{Plot 3}{Segments that would be used in the other cross-dating
      functions (e.g., \code{\link{corr.series.seg}})}
    \item{Plot 4}{Text giving the detrending options and the time span
      of the raw and filtered series and master}
  }

  The series and master are returned as well.

  See help pages for  \code{\link{corr.rwl.seg}},
  \code{\link{corr.series.seg}}, and  \code{\link{ccf.series.rwl}} for
  more information on these arguments.
}
\value{
  A \code{list} containing the filtered vectors \code{\var{series}} and
  \code{\var{master}}.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{
  \code{\link{corr.rwl.seg}}, \code{\link{corr.series.seg}},
  \code{\link{ccf.series.rwl}}
}
\examples{library(utils)
data(co021)
foo <- series.rwl.plot(rwl = co021, series = "646244", seg.length = 100,
                       n = 5)
## note effect of n on first year in the series
foo <- series.rwl.plot(rwl = co021, series = "646244", seg.length = 100,
                       n = 13, prewhiten = FALSE)
bar <- series.rwl.plot(rwl = co021, series = "646244", seg.length = 100,
                       n = 7, prewhiten = FALSE)
head(foo$series)
head(bar$series)
}
\keyword{ manip }

\encoding{UTF-8}
\name{csv2rwl}
\alias{csv2rwl}
\title{ Read Ring Width File from CSV }
\description{
  This function reads in a file of ring widths (.rwl) from a text file with comma separated values (csv).
}
\usage{
csv2rwl(fname,...)
}
\arguments{

  \item{fname}{ a \code{character} vector giving the file name of the
    csv file. }

  \item{\dots}{ other arguments passed to \code{\link{read.table}}. }
}
\details{
This is a simple wrapper to \code{\link{read.table}} that reads in a text file with ring-width data in "spreadsheet" format. I.e., with series in columns and the the years as rows. The first column should contain the years and each subsequent column should contain a tree-ring series. The series names should be in the first row of the file. The default for \code{\link{NA}} values are empty cells or as the character string \code{"NA"} but can also be set using the \code{na.strings} argument passed to \code{\link{read.table}}. E.g.,:
\tabular{lllll}{
Year \tab Ser1A \tab Ser1B \tab Ser2A \tab Ser2B\cr
1901 \tab NA \tab 0.45 \tab 0.43 \tab 0.24\cr
1902 \tab NA \tab 0.05 \tab  0.00 \tab 0.07\cr
1903 \tab 0.17 \tab 0.46 \tab 0.03 \tab 0.21\cr
1904 \tab 0.28 \tab 0.21 \tab 0.54 \tab 0.41\cr
1905 \tab 0.29 \tab 0.85 \tab 0.17 \tab 0.76\cr
1906 \tab 0.56 \tab 0.64 \tab 0.56 \tab 0.31\cr
1907 \tab 1.12 \tab 1.06 \tab 0.99 \tab 0.83\cr
etc...
}

Note that this is a rudimentary convenience function that isn't doing anything sophisticated. It reads in a file, assigns the years to the row names and sets the class of the object to \code{c("rwl","data.frame")} which allows \code{dplR} to recognize it. 

Although arguments can be passed to \code{\link{read.table}}, this is not designed to be a  flexible function. As is the philosophy with \code{dplR}, modifying the code is easy should the user wish to read in a different style of text file (e.g., tab delimited). Typing \code{csv2rwl} at the \code{R} prompt will get the user started. 

}
\value{
  Function returns an object of class
  \code{c("rwl", "data.frame")} with the series in columns and the years
  as rows.  The series \acronym{ID}s are the column names and the years
  are the row names.
}
\author{ Andy Bunn }
\seealso{ \code{\link{read.rwl}}, \code{\link{read.table}} }
\examples{
library(utils)
data(ca533)
# write out a rwl file in a format that csv2rwl will understand
tm <- time(ca533)
foo <- data.frame(tm,ca533)
# this is the temp file where foo will be written
tmpName <- tempfile()
write.csv(foo,file=tmpName,row.names=FALSE)
# read it back in using csv2rwl
bar <- csv2rwl(tmpName)
# check to see if identical
identical(ca533,bar)
# delete temp file
unlink(tmpName)
}
\keyword{ IO }

\encoding{UTF-8}
\name{print.redfit}
\alias{print.redfit}
\title{
  Printing Redfit Results
}
\description{
  Print information contained in or derived from a redfit object.
}
\usage{
\method{print}{redfit}(x, digits = NULL, csv.out = FALSE, do.table = FALSE,
      prefix = "", row.names = FALSE, file = "", ...)
}
\arguments{

  \item{x}{ An object of class \code{"redfit"}.  }

  \item{digits}{ Specifies the desired number of significant digits in
    the output.  The argument is passed to \code{\link{format}} and
    \code{\link{print.data.frame}}.  A positive integral value or
    \code{NULL}.  If \code{NULL}, the value in \code{options("digits")}
    is used. }

  \item{csv.out}{ A \code{logical} flag.  If \code{TRUE}, writes a
    large, comma-separated table of information.  The table contains one
    row for each frequency.  If \code{FALSE}, writes a few summary
    numbers instead.  See \sQuote{Details}. }

  \item{do.table}{ A \code{logical} flag.  If \code{TRUE}, the large
    information table is also printed when \code{csv.out} is
    \code{FALSE}, although not in a comma-separated format but with
    \code{print.data.frame}.  }

  \item{prefix}{ A prefix to be used on every output line except the
    large information table.  REDFIT (see \code{References}) uses
    \code{"# "}.  }

  \item{row.names}{ A \code{logical} flag enabling or disabling
    automatic row names from the large information table, or a
    \code{character} vector of names.  In any case, the table will
    contain frequency as the first column. }

  \item{file}{ A writable connection or a character string naming a
    file.  Used for setting the output destination when
    \code{\var{csv.out}} is \code{TRUE}.  The default is to write the
    comma-separated table to the console.  }

  \item{\dots}{ Arguments to \code{write.csv}.  Used when
    \code{\var{csv.out}} is \code{TRUE}.  }
  
}
\value{
  Invisibly returns \code{\var{x}}.
}
\references{

  This function is based on the Fortran program
  \href{https://www.marum.de/Prof.-Dr.-michael-schulz.html}{REDFIT},
  which is in the public domain.

  Schulz, M. and Mudelsee, M. (2002) REDFIT: estimating red-noise
  spectra directly from unevenly spaced paleoclimatic time series.
  \emph{Computers & Geosciences}, \bold{28}(3), 421\enc{–}{--}426.

}
\author{
  Mikko Korpela
}
\seealso{
  \code{\link{redfit}}
}
\examples{library(utils)
data(ca533)
tm <- time(ca533)
x <- ca533[[1]]
idx <- which(!is.na(x))
redf <- redfit(x[idx], tm[idx], "time",
               nsim = 100, iwin = 0, ofac = 1, n50 = 1)
print(redf)
fname <- tempfile(fileext=".csv")
print(fname) # tempfile used for output
print(redf, csv.out = TRUE, file = fname)
redftable <- read.csv(fname)
unlink(fname) # remove the file
}
\keyword{ print }

\encoding{UTF-8}
\name{ads}
\alias{ads}
\title{Age-Dependent Spline}
\description{
  Applies an age-dependent smoothing spline to \code{\var{y}}.
}

\usage{
ads(y, nyrs0 = 50, pos.slope = TRUE)
}
\arguments{
  \item{y}{ a \code{numeric} vector, typically a tree-ring series. }
  \item{nyrs0}{ a number greater than one, affecting the rigidity of the
    initial spline.  A larger \code{\var{nyrs0}} produces a more 
    rigid spline.}
  \item{pos.slope}{ a \code{logical} flag.  Will allow for a positive
    slope at the end of fitted value. If \code{FALSE} the
    line will be horizontal. See details. }
}
\details{
This implements the age-dependent smoothing spline similar to that described by Melvin (2004). In this implementation a cubic smoothing spline (\code{\link{caps}}) is fit to \code{\var{y}} with an initial stiffness of \code{\var{nyrs0}}. For each succesive measurement, the siffness is incremented by that ring index. This results in a spline is \code{\var{nyrs0}} flexible at the start of the series and grows progressively stiffer. This can help capture the initial fast growth of a juvinielle tree with a flexible spline that then progresses to a stiffer spline that can better model the constant growth commonly found in mature trees. In its details, the cubic smoothing spline follows the Cook and Peters (1981) spline with a 50\% frequency cutoff. See Cook and Kairiukstis (1990) for more information.

The default setting for \code{\var{nyrs0}} is 50 years which is approprite for trees with a classic growth model but a value of 10 or 20 might be more appropriate for a Hugershoff-like initial increase in growth. Cook (pers comm) suggests a value of 20 for RCS.

If \code{\var{pos.slope}} is FALSE, the function will attempt to prevent a positive slope at the end of the series. In some cases when \code{\var{ads}} is used for detrending, a positive slope can be considered biologically unlikely. In those cases, the user can constrain the positive slope in the spline. This works by calculating the spline and taking the first difference. Then the function finds the last (outside) index where the spline changes slope and fixes the spline values from the that point to the end. Finally the spline is rerun along this constrained curve. See examples for details. The wisdom of constraining the slope in this manner depends very much on expert knowledge of the system.
}

\value{
  A filtered vector.
}
\references{
  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of Dendrochronology: Applications in the Environmental Sciences}. Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.
  
  Melvin, T. M. (2004) \emph{Historical Growth Rates and Changing Climatic Sensitivity of Boreal Conifers}. PhD Thesis, Climatic Research Unit, School of Environmental Sciences, University of East Anglia.
  
  Cook, E. R. and Peters, K. (1981) The Smoothing Spline: A New Approach to Standardizing Forest Interior Tree-Ring Width Series for Dendroclimatic Studies. Tree-Ring Bulletin, 41, 45-53.

}
\author{
  Fortran code provided by Ed Cook. Ported and adapted for dplR by Andy Bunn.
}

\seealso{ \code{\link{caps}}, \code{\link{detrend}} }

\examples{
# fit a curve
data(co021)
aSeries <- na.omit(co021$`641114`)
plot(aSeries,type="l",col="grey50")
lines(ads(y = aSeries),col="blue",lwd=2)

# show an artificial series with a Hugershoff-like curve.
a <- 0.5
b <- 1
g <- 0.1
d <- 0.25

n <- 300
x <- 1:n
y <- I(a*x^b*exp(-g*x)+d)
# add some noise
y <- y + runif(n=length(y),min = 0,max = 0.5)
# Plot with two different splines.
plot(y,type="l",col="grey50")
lines(ads(y,50),col="darkgreen",lwd=2) # bad
lines(ads(y,10),col="darkblue",lwd=2) # good

# now repeat with a positive slope to constrain
y <- I(a*x^b*exp(-g*x)+d)
y[251:300] <- y[251:300] + seq(0,0.25,length.out=50)
y <- y + runif(n=length(y),min = 0,max = 0.5)
plot(y,type="l",col="grey50")
lines(ads(y,10),col="darkgreen",lwd=2) #bad?
lines(ads(y,10,pos.slope=FALSE),col="darkgreen",lwd=2,lty="dashed")
}
\keyword{ smooth }

\encoding{UTF-8}
\name{write.crn}
\alias{write.crn}
\title{ Write Tucson Format Chronology File }
\description{
  This function writes a chronology to a Tucson (decadal) format file.
}
\usage{
write.crn(crn, fname, header = NULL, append = FALSE)
}
\arguments{
  \item{crn}{ a \code{data.frame} containing a tree-ring chronology with
    two columns of the type produced by \code{\link{chron}}.  The first
    column contains the mean value chronology, the second column gives
    the sample depth.  Years for the chronology are determined from the
    row names.  The chronology \acronym{ID} is determined from the first
    column name.}
  \item{fname}{ a \code{character} vector giving the file name of the
    crn file. }
  \item{header}{ a \code{list} giving information for the header of the
    file.  If \code{NULL} then no header information will be written. }
  \item{append}{ \code{logical} flag indicating whether to append this
    chronology to an existing file. }
}
\details{
  This writes a standard crn file as defined according to the standards
  of the \acronym{ITRDB} at
  \url{https://www.ncei.noaa.gov/pub/data/paleo/treering/treeinfo.txt}.  This is the
  decadal or Tucson format.  It is an \acronym{ASCII} file and machine
  readable by the standard dendrochronology programs.  Header information
  for the chronology can be written according to the International Tree
  Ring Data Bank (\acronym{ITRDB}) standard.  The header standard is not
  very reliable however and should be thought of as experimental
  here.  Do not try to write headers using dplR to submit to the
  \acronym{ITRDB}.  When submitting to the \acronym{ITRDB}, you can enter
  the metadata via their website.  If you insist however, the header
  information is given as a \code{list} and must be formatted with the
  following:
  \tabular{lrrr}{
    \emph{Description} \tab \emph{Name} \tab \emph{Class} \tab \emph{Max Width} \cr
    Site \acronym{ID} \tab \code{\var{site.id}} \tab \code{character} \tab 6 \cr
    Site Name \tab \code{\var{site.name}} \tab \code{character} \tab 52 \cr
    Species Code \tab \code{\var{spp.code}} \tab \code{character} \tab 4 \cr
    State or Country \tab \code{\var{state.country}} \tab \code{character} \tab 13 \cr
    Species \tab \code{\var{spp}} \tab \code{character} \tab 18 \cr
    Elevation \tab \code{\var{elev}} \tab \code{character} \tab 5 \cr
    Latitude \tab \code{\var{lat}} \tab \code{character} or \code{numeric} \tab 5 \cr
    Longitude \tab \code{\var{long}} \tab \code{character} or \code{numeric} \tab 5 \cr
    First Year \tab \code{\var{first.yr}} \tab \code{character} or \code{numeric} \tab 4 \cr
    Last Year \tab \code{\var{last.yr}} \tab \code{character} or \code{numeric} \tab 4 \cr
    Lead Investigator \tab \code{\var{lead.invs}} \tab \code{character} \tab 63 \cr
    Completion Date \tab \code{\var{comp.date}} \tab \code{character} \tab 8 \cr
  }
  See examples for a correctly formatted header list.  If the width of
  the fields is less than the max width, then the fields will be padded
  to the right length when written.  Note that \code{\var{lat}} and
  \code{\var{long}} are really \code{\var{lat}*100} or
  \code{\var{long}*100} and given as integral values.  E.g., 37 degrees
  30 minutes would be given as 3750.

  This function takes a single chronology with sample depth as
  input.  This means that it will fail if given output from
  \code{\link{chron}} where \code{\var{prewhiten} == TRUE}.  However,
  more than one chronology can be appended to the bottom of an existing
  file (e.g., standard and residual) with a second call to
  \code{\link{write.crn}}.  However, the \acronym{ITRDB} recommends
  saving and publishing only one chronology per file.  The examples
  section shows how to circumvent this.  The output from this function
  might be suitable for publication on the \acronym{ITRDB} although the
  header writing is clunky (see above) and rwl files are much better
  than crn files in terms of usefulness on the \acronym{ITRDB}.
}
\value{
  \code{\var{fname}}
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{chron}}, \code{\link{read.crn}}
}
\examples{library(utils)
data(ca533)
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
ca533.crn <- chron(ca533.rwi)
fname1 <- write.crn(ca533.crn, tempfile(fileext=".crn"))
print(fname1) # tempfile used for output

## Put the standard and residual chronologies in a single file
## with ITRDB header info on top. Not recommended.
ca533.crn <- chron(ca533.rwi, prewhiten = TRUE)
ca533.hdr <- list(site.id = "CAM", site.name = "Campito Mountain",
                  spp.code = "PILO", state.country = "California",
                  spp = "Bristlecone Pine", elev = "3400M", lat = 3730,
                  long = -11813, first.yr = 626, last.yr = 1983,
                  lead.invs = "Donald A. Graybill, V.C. LaMarche, Jr.",
                  comp.date = "Nov1983")
fname2 <- write.crn(ca533.crn[, -2], tempfile(fileext=".crn"),
                    header = ca533.hdr)
write.crn(ca533.crn[, -1], fname2, append = TRUE)
print(fname2) # tempfile used for output

unlink(c(fname1, fname2)) # remove the files
}
\keyword{ IO }

\encoding{UTF-8}
\name{write.rwl}
\alias{write.rwl}
\title{ Write Chronology File }
\description{
  This function writes a chronology to a file in one of the available
  formats.
}
\usage{
write.rwl(rwl.df, fname, format = c("tucson", "compact", "tridas"), \dots)
}
\arguments{

  \item{rwl.df}{ a \code{data.frame} containing tree-ring ring widths
    with the series in columns and the years as rows.  The series
    \acronym{ID}s are the column names and the years are the row
    names.  This type of \code{data.frame} is produced by
    \code{\link{read.rwl}}. }

  \item{fname}{ a \code{character} vector giving the file name of the
    rwl file. }

  \item{format}{ a \code{character} vector giving the format.  This must
    be \code{"tucson"}, \code{"compact"}, or \code{"tridas"}.  Tucson
    format is the default. }

  \item{\dots}{ arguments specific to the function implementing the
    operation for the chosen format. }

}
\details{
  This is a simple wrapper to the functions actually implementing the
  write operation.
}
\value{
  \code{\var{fname}}
}
\author{ Mikko Korpela }
\seealso{ \code{\link{write.crn}}, \code{\link{write.tucson}},
  \code{\link{write.compact}}, \code{\link{write.tridas}},
  \code{\link{read.rwl}}
}
\examples{library(utils)
data(co021)
co021.hdr <- list(site.id = "CO021",
                  site.name = "SCHULMAN OLD TREE NO. 1, MESA VERDE",
                  spp.code = "PSME", state.country = "COLORADO",
                  spp = "DOUGLAS FIR", elev = 2103, lat = 3712,
                  long = -10830, first.yr = 1400, last.yr = 1963,
                  lead.invs = "E. SCHULMAN", comp.date = "")
fname <- write.rwl(rwl.df = co021, fname = tempfile(fileext=".rwl"),
                   format = "tucson", header = co021.hdr,
                   append = FALSE, prec = 0.001)
print(fname) # tempfile used for output

unlink(fname) # remove the file
}
\keyword{ IO }

\encoding{UTF-8}
\name{spag.plot}
\alias{spag.plot}
\title{ Spaghetti Plot }
\description{
  Makes a spaghetti plot of tree-ring data.
}
\usage{
spag.plot(rwl, zfac = 1, useRaster = FALSE, res = 150, \dots)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}}. }
  \item{zfac}{ a multiplier for \code{\var{rwl}} to enhance clarity when
    plotting. }
  \item{useRaster}{A \code{logical} flag.  If \code{TRUE}, the tree-ring
    series are drawn as a raster image.  Other parts of the plot are
    not affected.  Other choices are \code{NA} (automatic) and
    \code{FALSE} (use vector graphics, the default).  See
    \code{\link{wavelet.plot}}.  }
  \item{res}{A \code{numeric} vector of length 1.  The resolution
    (pixels per inch) of the tree-ring series when a raster image is
    used. }
  \item{\dots}{ arguments to be passed to \code{\link{lines}}. }
}
\details{
  This makes a simple plot of each series in a tree-ring data set.  Each
  series is centered first by subtracting the column mean using
  \code{\link{scale}}.  The plot can be grossly tuned with
  \code{\var{zfac}} which is a multiplier to \code{\var{rwl}} before
  plotting and centering.
}
\value{
  None.  This function is invoked for its side effect, which is to
  produce a plot.
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{seg.plot}} }
\examples{library(utils)
data(co021)
plot(co021,plot.type = "spag")
spag.plot(co021, zfac = 2)
}
\keyword{ hplot }

\encoding{UTF-8}
\name{detrend}
\alias{detrend}
\title{ Detrend Multiple Ring-Width Series Simultaneously }
\description{
  This is a wrapper for \code{\link{detrend.series}} to detrend many
  ring-width series at once.
}
\usage{
detrend(rwl, y.name = names(rwl), make.plot = FALSE,
        method = c("Spline", "ModNegExp", "Mean", "Ar", "Friedman",
                   "ModHugershoff", "AgeDepSpline"),
        nyrs = NULL, f = 0.5, pos.slope = FALSE,
        constrain.nls = c("never", "when.fail", "always"),
        verbose = FALSE, return.info = FALSE,
        wt, span = "cv", bass = 0, difference = FALSE)
}
\arguments{

  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}} }

  \item{y.name}{ a \code{character} vector of
    \code{length(ncol(\var{rwl}))} that gives the \acronym{ID} of each
    series.  Defaults to the column names of \code{\var{rwl}}. }

  \item{make.plot}{ a \code{logical} flag.  Makes plots of the raw data
    and detrended data if \code{TRUE}.  See details below. }
  
  \item{method}{ a \code{character} vector to determine the detrending
    methods.  See details below.  Possible values are all subsets of\cr
    \code{c("Spline", "ModNegExp", "Mean", "Ar", "Friedman", "ModHugershoff",
    "AgeDepSpline")}.
    Defaults to using all the available methods.}
  
  \item{nyrs}{ a number giving the rigidity of the smoothing spline,
    defaults to 0.67 of series length if \code{\var{nyrs}} is
    \code{NULL}. }
  
  \item{f}{ a number between 0 and 1 giving the frequency response or
    wavelength cutoff.  Defaults to 0.5. }

  \item{pos.slope}{ a \code{logical} flag.  Will allow for a positive
    slope to be used in method \code{"ModNegExp"} and \code{"ModHugershoff"}.  
    If \code{FALSE} the
    line will be horizontal. }
  
  \item{constrain.nls}{ a \code{character} string which controls
    the constraints of the \code{"ModNegExp"} model and the 
    \code{"ModHugershoff"}.  See 
    \code{\link{detrend.series}} for further details. }
    
  \item{verbose}{ \code{logical}.  Write out details? }

  \item{return.info}{ a \code{logical} flag.  If \code{TRUE}, details
    about models and data will be added to the return value.  See
    \sQuote{Value}. }

  \item{wt}{ a \code{numeric} vector of case weights for method
    \code{"Friedman"}. The default means equals weights. }

  \item{span}{ a \code{numeric} value controlling method
    \code{"Friedman"}, or \code{"cv"} (default) for automatic choice by
    cross-validation. See \code{\link{supsmu}}. }

  \item{bass}{ a \code{numeric} value controlling the smoothness of the
    fitted curve in method \code{"Friedman"}. See \code{\link{supsmu}}. }

  \item{difference}{ a \code{logical} flag.  Compute residuals by substraction if TRUE, otherwise use division. }

}
\details{
  See \code{\link{detrend.series}} for details on detrending
  methods.  Setting \code{\var{make.plot} = TRUE} will cause plots of
  each series to be produced.  These could be saved using
  \code{\link{Devices}} if desired.
}
\value{
  If one detrending method is used, a \code{data.frame} containing the
  dimensionless detrended ring widths with column names, row names and
  dimensions of \code{\var{rwl}}.  If more methods are used, a list with
  \code{ncol(\var{rwl})} elements each containing a \code{data.frame}
  with the detrended ring widths in each column.

  If \code{\var{return.info}} is \code{TRUE}, the return value is a
  \code{list} with four parts:

  \item{series}{ the main result described above (\code{data.frame} or
    list of data.frames) }

  \item{curves}{ the curve or line used to detrend \code{series}. Either a \code{data.frame} or a list of data.frames. }

  \item{model.info}{ Information about the models corresponding to each
    output series.  A \code{list} with one element for each column of
    \code{\var{rwl}}.  See \code{\link{detrend.series}} (\sQuote{Value},
    \var{model.info}) for a description of the contents. }

  \item{data.info}{ Information about the input series.  A \code{list}
    with one element for each column of \code{\var{rwl}}.  See
    \code{\link{detrend.series}} (\sQuote{Value}, \var{data.info}). }

}
\note{
  This function uses the \code{\link[foreach]{foreach}} looping
  construct with the \code{\link[foreach:foreach]{\%dopar\%}} operator.
  For parallel computing and a potential speedup, a parallel backend
  must be registered before running the function.  If
  \code{\var{verbose}} is \code{TRUE}, parallel computation is disabled.
}
\author{ Andy Bunn.  Improved by Mikko Korpela. }
\seealso{ \code{\link{detrend.series}} }
\examples{library(utils)
data(ca533)
## Detrend using modified exponential decay. Returns a data.frame
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
## Detrend using splines and compute
## residuals via subtraction
ca533.rwi <- detrend(rwl = ca533, method = "Spline",
                     difference = TRUE)

## Detrend using modified Hugershoff curve and return info on the model
## fits. Returns a list with: series, curves, modelinfo and data.info
data(co021)
co021.rwi <- detrend(rwl = co021, method = "ModHugershoff",
                     return.info=TRUE)

\dontrun{
library(grDevices)
## Detrend using all methods. Returns a list
ca533.rwi <- detrend(rwl = ca533)
## Save a pdf of all series
fname <- tempfile(fileext=".pdf")
print(fname) # tempfile used for output
pdf(fname)
ca533.rwi <- detrend(rwl = ca533, method = c("Spline", "ModNegExp"),
                     make.plot = TRUE)
dev.off()

unlink(fname) # remove the file
}
}
\keyword{ manip }

\encoding{UTF-8}
\name{gini.coef}
\alias{gini.coef}
\title{ Calculate the Gini Coefficient   }
\description{
  This function calculates the Gini coefficient on raw or detrended
  ring-width series.
}
\usage{
gini.coef(x)
}
\arguments{
  \item{x}{ a \code{numeric} vector }
}
\details{
  This calculates the Gini coefficient of inequality which is used as an
  all-lag measure of diversity in tree-ring records \enc{–}{-} typically detrended
  series.  Lower values indicate lower diversity.  The use of the Gini
  coefficient in dendrochronology is described by Biondi and Qeadan (2008).
  See Handcock and Morris (1999) for more information.  
}
\value{ the Gini coefficient.
}
\references{
  Biondi, F. and Qeadan, F. (2008) Inequality in Paleorecords.
  \emph{Ecology}, \bold{89}(4), 1056\enc{–}{--}1067.

  Handcock, M. S. and Morris, M. (1999) \emph{Relative Distribution
    Methods in the Social Sciences}.  Springer.  \acronym{ISBN}:
  0-387-98778-9.
}
\author{ Mikko Korpela, based on original by Andy Bunn }
\seealso{ \code{\link{rwl.stats}} }
\examples{
library(utils)
data(ca533)
ca533.rwi <- detrend(rwl = ca533, method = "ModNegExp")
ca533.crn <- chron(ca533.rwi)
gini.coef(ca533.crn)
}
\keyword{ univar }

\encoding{UTF-8}
\name{rwi.stats.running}
\alias{rwi.stats.running}
\alias{rwi.stats}
\alias{rwi.stats.legacy}
\title{ (Running Window) Statistics on Detrended Ring-Width Series }
\description{
  These functions calculate descriptive statistics on a
  \code{data.frame} of (usually) ring-width indices.  The statistics are
  optionally computed in a running window with adjustable length and
  overlap.  The data can be filtered so that the comparisons are made to
  on just high-frequency data.
}
\usage{
rwi.stats.running(rwi, ids = NULL, period = c("max", "common"),
                  method = c("spearman", "pearson","kendall"),
                  prewhiten=FALSE,n=NULL,
                  running.window = TRUE,
                  window.length = min(50, nrow(rwi)),
                  window.overlap = floor(window.length / 2),
                  first.start = NULL,
                  min.corr.overlap = min(30, window.length),
                  round.decimals = 3,
                  zero.is.missing = TRUE)

rwi.stats(rwi, ids=NULL, period=c("max", "common"), 
          method = c("spearman", "pearson","kendall"), \dots)

rwi.stats.legacy(rwi, ids=NULL, period=c("max", "common"))
}
\arguments{

  \item{rwi}{ a \code{data.frame} with detrended and standardized ring
    width indices as columns and years as rows such as that produced by
    \code{\link{detrend}}. }

  \item{ids}{ an optional \code{data.frame} with column one named
    \code{"tree"} giving a \code{numeric} \acronym{ID} for each tree and
    column two named \code{"core"} giving a \code{numeric} \acronym{ID}
    for each core.  Defaults to one core per tree as\cr
    \code{data.frame(tree=1:ncol(\var{rwi}), core=rep(1, ncol(\var{rwi})))}. }
  \item{period}{ a \code{character} string, either \code{"common"} or
    \code{"max"} indicating whether correlations should be limited to
    complete observations over the period common to all cores (i.e. rows
    common to all samples) or the maximum pairwise overlap.  Defaults
    to \code{"max"}. }
  \item{method}{Can be either \code{"pearson"}, \code{"kendall"}, or
    \code{"spearman"} which indicates the correlation coefficient to be
    used.  Defaults to \code{"spearman"}.  See \code{\link{cor}}. }
  \item{n}{ \code{NULL} or an integral value giving the filter length
    for the \code{\link{hanning}} filter used for removal of low
    frequency variation. }
  \item{prewhiten}{ \code{logical} flag.  If \code{TRUE} each series is
    whitened using \code{\link{ar}}. }
  \item{running.window}{ \code{logical} flag indicating whether to use a
    running window (\code{TRUE}, the default) or to ignore the other
    window parameters and effectively use one window covering all years
    (\code{FALSE}). }
  \item{window.length}{ \code{numeric} value indicating the length of
    the running window in years.  The default is 50 years or the number
    of years (rows) in \code{\var{rwi}}, whichever is smaller. }
  \item{window.overlap}{ \code{numeric} value indicating the overlap of
    consecutive window positions, i.e. the number of common years.  The
    default is half of the window length, rounded down. }
  \item{first.start}{ an optional \code{numeric} value setting the
    position of the first window.  Must be a value between \code{1} and
    \code{\var{n.years}-\var{window.length}+1}, where
    \code{\var{n.years}} is the number of years in \code{\var{rwi}}.  The
    default value \code{NULL} lets the function make the decision using
    some heuristic rules. }
  \item{min.corr.overlap}{ \code{numeric} value setting the minimum
    number of common years in any pair of ring-width series required for
    their correlation to be included in the calculations.  Smaller
    overlaps are considered to yield unreliable correlation values which
    are ignored.  Defaults to the minimum of 30 and the length of the
    window.  One way to lift the restriction and include all correlations
    is to set \code{\var{min.corr.overlap} = 0}. }
  \item{round.decimals}{ non-negative integer \code{numeric} value
    setting the desired number of decimal places in the results.  Use
    \code{NA}, \code{NULL} or a negative number for no rounding. }
  \item{zero.is.missing}{ \code{logical} flag indicating whether to
    treat zeros as missing values (\code{TRUE}, the default) or to
    include them in computation (\code{FALSE}).  }
  \item{\dots}{ arguments passed on to \code{rwi.stats.running} }

}
\details{
  This calculates a variety of descriptive statistics commonly used in
  dendrochronology.

  The function \code{rwi.stats} is a wrapper that calls
  \code{rwi.stats.running} with \code{\var{running.window} = FALSE}.
  The results may differ from those prior to dplR 1.5.3, where the
  former \code{rwi.stats} (now renamed to \code{rwi.stats.legacy}) was
  replaced with a call to \code{rwi.stats.running}.

  For correctly calculating the statistics on within and between series
  variability, an appropriate mask (parameter \code{\var{ids}}) must be
  provided that identifies each series with a tree as it is common for
  dendrochronologists to take more than one core per tree.  The function
  \code{\link{read.ids}} is helpful for creating a mask based on the
  series \acronym{ID}.

  If \code{\var{ids}} has duplicate tree/core combinations, the
  corresponding series are averaged before any statistics are computed.
  The value of the parameter \code{\var{zero.is.missing}} is relevant in
  the averaging: \code{TRUE} ensures that zeros don\enc{’}{'}t
  contribute to the average.  The default value of
  \code{\var{zero.is.missing}} is \code{TRUE}.  The default prior to
  dplR 1.5.3 was \code{FALSE}.  If the parameter is set to \code{FALSE},
  the user will be warned in case zeros are present.  Duplicate
  tree/core combinations are not detected by \code{rwi.stats.legacy}.

  Row names of \code{\var{ids}} may be used for matching the
  \acronym{ID}s with series in \code{\var{rwi}}.  In this case, the
  number of rows in \code{\var{ids}} is allowed to exceed the number of
  series.  If some names of \code{\var{rwi}} are missing from the row
  names of \code{\var{ids}}, the rows of \code{\var{ids}} are assumed to
  be in the same order as the columns of \code{\var{rwi}}, and the
  dimensions must match.  The latter is also the way that
  \code{rwi.stats.legacy} handles \code{\var{ids}}, i.e. names are
  ignored and dimensions must match.

  Note that \code{period = "common"} can produce \code{NaN} for many of
  the stats if there is no common overlap period among the cores.  This
  happens especially in chronologies with floating subfossil samples
  (e.g., \code{\link{ca533}}).

  Some of the statistics are specific to dendrochronology (e.g., the
  effective number of cores or the expressed population signal).  Users
  unfamiliar with these should see Cook and Kairiukstis (1990) and
  Fritts (2001) for further details for computational details on the
  output.  The signal-to-noise ratio is calculated following Cook and 
  Pederson (2011).
  
  Please note that Buras (2017) cautions against the use of expressed 
  population signal (EPS) for evaluating the climate-reconstruction potential 
  of a tree-ring sample. Instead, he recommends the use of subsample signal 
  strength (\code{\link{sss}}) for evaluating the loss of predictive power 
  back in time when sample replication drops.

  If desired, the \code{\var{rwi}} can be filtered in the same manner
  as the family of cross-dating functions using \code{\var{prewhiten}} and
  \code{\var{n}}.  See the help page for \code{\link{corr.rwl.seg}} for
  more details.
  

}
\value{ A \code{data.frame} containing the following columns (each row
  corresponds to one position of the window):

  \item{start.year}{the first year in the window.  Not returned if
    \code{running.window} is \code{FALSE} or called as
    \code{rwi.stats} or \code{rwi.stats.legacy}.}

  \item{mid.year}{the middle year in the window, rounded down.  Not
    returned if \code{running.window} is \code{FALSE} or called as
    \code{rwi.stats} or \code{rwi.stats.legacy}.}
  
  \item{end.year}{the last year in the window.  Not returned if
    \code{running.window} is \code{FALSE} or called as
    \code{rwi.stats} or \code{rwi.stats.legacy}.}

  \item{n.cores}{the number of cores}

  \item{n.trees}{the number of trees}
  
  \item{n}{the average number of trees (for each year, a tree needs at
    least one non-\code{NA} core in order to be counted).  Not returned
    in the results of \code{rwi.stats.legacy}}
  
  \item{n.tot}{total number of correlations calculated as
    \code{\var{n.wt} + \var{n.bt}}.

    Equal to \code{\var{n.cores} * (\var{n.cores}-1)/2} if there is
    overlap between all samples }

  \item{n.wt}{number of within-tree correlations computed}

  \item{n.bt}{number of between-tree correlations computed}

  \item{rbar.tot}{the mean of all the correlations between different 
      cores}

  \item{rbar.wt}{the mean of the correlations between series from the
    same tree over all trees}

  \item{rbar.bt}{the mean interseries correlation between all series
    from different trees}

  \item{c.eff}{the effective number of cores (takes into account the
    number of within-tree correlations in each tree)}

  \item{rbar.eff}{the effective signal calculated as
    \code{\var{rbar.bt} / (\var{rbar.wt} + (1-\var{rbar.wt}) / \var{c.eff}) }}

  \item{eps}{the expressed population signal calculated using the average
    number of trees as\cr
    \code{\var{n} * \var{rbar.eff} / ((\var{n} - 1) * \var{rbar.eff} + 1)} }

  \item{snr}{the signal to noise ratio calculated using the average
    number of trees as\cr
    \code{\var{n} * \var{rbar.eff} / (1-\var{rbar.eff})} }
}
\references{

  Buras, A. (2017) A comment on the Expressed Population Signal. Dendrochronologia 44:130-132. 

  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Cook, E. R. and Pederson, N. (2011) Uncertainty, Emergence, and
  Statistics in Dendrochronology.  In Hughes, M. K., Swetnam, T. W., and
  Diaz, H. F., editors, \emph{Dendroclimatology: Progress and
  Prospects}, pages 77\enc{–}{--}112.  Springer.  \acronym{ISBN-13}:
  978-1-4020-4010-8.

  Fritts, H. C. (2001) \emph{Tree Rings and Climate}.  Blackburn.
    \acronym{ISBN-13}: 978-1-930665-39-2.
  
}
\note{
  This function uses the \code{\link[foreach]{foreach}} looping
  construct with the \code{\link[foreach:foreach]{\%dopar\%}} operator.
  For parallel computing and a potential speedup, a parallel backend
  must be registered before running the function.
}
\author{ Mikko Korpela, based on \code{\link{rwi.stats.legacy}} by Andy
  Bunn }
\seealso{ \code{\link{detrend}}, \code{\link{cor}},
  \code{\link{read.ids}}, \code{\link{rwi.stats}},
  \code{\link{corr.rwl.seg}}  } 
\examples{library(utils)
data(gp.rwl)
data(gp.po)
gp.rwi <- cms(rwl = gp.rwl, po = gp.po)
gp.ids <- read.ids(gp.rwl, stc = c(0, 2, 1))
# On a running window
rwi.stats.running(gp.rwi, gp.ids)
## With no running window (i.e. running.window = FALSE)
rwi.stats(gp.rwi, gp.ids)
## Restrict to common overlap (in this case 1899 to 1987)
rwi.stats(gp.rwi, gp.ids, period="common")
rwi.stats.legacy(gp.rwi, gp.ids) # rwi.stats prior to dplR 1.5.3

}
\keyword{ misc }

\encoding{UTF-8}
\name{read.compact}
\alias{read.compact}
\title{ Read \acronym{DPL} Compact Format Ring Width File }
\description{  
  This function reads in a \acronym{DPL} compact format file of ring
  widths.
}
\usage{
read.compact(fname)
}
\arguments{
  \item{fname}{ a \code{character} vector giving the file name of the
    rwl file. }
}
\details{
  This function should be able to read files written by the
  Dendrochronology Program Library (\acronym{DPL}) in its compact
  format.
}
\value{
  An object of class \code{c("rwl", "data.frame")} with the series in
  columns and the years as rows.  The series \acronym{ID}s are the
  column names and the years are the row names.
}
\author{ Mikko Korpela }
\seealso{ \code{\link{read.rwl}}, \code{\link{read.tucson}},
  \code{\link{read.tridas}}, \code{\link{read.fh}},
  \code{\link{write.compact}}}
  
\examples{
\dontrun{
data(co021)
fname <- write.compact(rwl.df = co021,
                       fname = tempfile(fileext=".rwl"),
                       append = FALSE, prec = 0.001)
foo <- read.compact(fname)
str(foo)
str(co021)
all.equal(foo,co021)
unlink(fname)
}
}
\keyword{ IO }

\encoding{UTF-8}
\name{read.tucson}
\alias{read.tucson}
\title{ Read Tucson Format Ring Width File }
\description{
  This function reads in a Tucson (decadal) format file of ring widths (.rwl).
}
\usage{
read.tucson(fname, header = NULL, long = FALSE,
            encoding = getOption("encoding"), 
            edge.zeros = TRUE, verbose = TRUE)
}
\arguments{
  \item{fname}{ a \code{character} vector giving the file name of the
    rwl file. }
  \item{header}{ \code{logical} flag indicating whether the file has a
    header.  If \code{NULL} then the function will attempt to determine
    if a header exists. }
  \item{long}{ \code{logical} flag indicating whether dates in file span
    0 \acronym{CE} and therefore use negative numbers.  If \code{TRUE}
    only the first 7 characters can be used for series \acronym{ID}s.
    If \code{FALSE} then series \acronym{ID}s can be up to 8
    characters. }
  \item{encoding}{ the name of the encoding to be used when reading the
    rwl file.  Usually the default value will work, but an rwl file
    written in a non-default encoding may crash the function.  In that
    case, identifying the encoding and specifying it here should fix the
    problem.  Examples of popular encodings available on many systems
    are \code{"ASCII"}, \code{"UTF-8"}, and \code{"latin1"} alias
    \code{"ISO-8859-1"}.  See the help of \code{\link{file}}. }
  \item{edge.zeros}{ \code{logical} flag indicating whether leading or
    trailing zeros in series will be preserved (when the flag is
    \code{TRUE}, the default) or discarded, i.e. marked as \code{NA}
    (when \code{FALSE}). }
  \item{verbose}{ \code{logical} flag, print info on data. }
}
\details{
  This reads in a standard rwl file as defined according to the
  standards of the \acronym{ITRDB} at
  \url{https://www.ncei.noaa.gov/pub/data/paleo/treering/treeinfo.txt}.  Despite the
  standards at the \acronym{ITRDB}, this occasionally fails due to
  formatting problems.
}
\value{
  An object of class \code{c("rwl", "data.frame")} with the series in
  columns and the years as rows.  The series \acronym{ID}s are the
  column names and the years are the row names.
}
\author{ Andy Bunn.  Patched and greatly improved by Mikko Korpela. }
\seealso{ \code{\link{read.rwl}}, \code{\link{read.compact}},
  \code{\link{read.tridas}}, \code{\link{read.fh}},
  \code{\link{write.tucson}} }
\keyword{ IO }

\encoding{UTF-8}
\name{read.ids}
\alias{read.ids}
\alias{autoread.ids}
\title{ Read Site-Tree-Core \acronym{ID}s }
\description{
  These functions try to read site, tree, and core \acronym{ID}s from a
  rwl \code{data.frame}.
}
\usage{
read.ids(rwl, stc = c(3, 2, 3), ignore.site.case = FALSE,
         ignore.case = FALSE, fix.typos = FALSE, typo.ratio = 5,
         use.cor = TRUE)

autoread.ids(rwl, ignore.site.case = TRUE, ignore.case = "auto",
             fix.typos = TRUE, typo.ratio = 5, use.cor = TRUE)
}
\arguments{

  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}} or
    \code{\link{ca533}} }

  \item{stc}{ a vector of three integral values or character string
    "auto".  The numbers indicate the number of characters to split the
    site code (\code{\var{stc}[1]}), the tree \acronym{ID}s
    (\code{\var{stc}[2]}), and the core \acronym{ID}s
    (\code{\var{stc}[3]}).  Defaults to \code{c(3, 2, 3)}.  If "auto",
    tries to automatically determine the split locations.  See Details
    for further information. }

  \item{use.cor}{ a \code{logical} flag.  If \code{TRUE} and
    \code{\var{stc} is "auto"}, correlation clustering may be used for
    determining the length of the tree and core parts.  See Details. }

  The following parameters affect the handling of suspected typing
  errors.  Some have different default values in \code{read.ids} and
  \code{autoread.ids}.

  \item{ignore.site.case}{ a \code{logical} flag.  If \code{TRUE}, the
    function does not distinguish between upper case and lower case
    letters in the site part of the series names. }

  \item{ignore.case}{ a \code{logical} flag or "auto".  If \code{TRUE},
    the function does not distinguish between upper case and lower case
    letters in the tree / core part of the series names.  The default in
    \code{read.ids} is \code{FALSE}, i.e. the difference matters.  The
    default in \code{read.ids} is \code{"auto"}, which means that the
    function tries to be smart with respect to case sensitivity.  In
    \code{"auto"} mode, the function generally ignores case differences,
    unless doing so would result in additional duplicate combinations of
    tree and core \acronym{ID}s.  Also, when in \code{"auto"} mode and
    \code{\var{stc}} is \code{"auto"}, case sensitivity is used in
    highly heuristic ways when deciding the boundary between the site
    part and the tree part in uncertain cases. }

  \item{fix.typos}{ a \code{logical} flag.  If \code{TRUE}, the function
    will try to detect and fix typing errors. }

  \item{typo.ratio}{ a \code{numeric} value larger than \code{1},
    affecting the eagerness of the function to fix typing errors.  The
    default is 5.  See Details. }
}
\details{

  Because dendrochronologists often take more than one core per tree, it
  is occasionally useful to calculate within vs. between tree variance.
  The International Tree Ring Data Bank (\acronym{ITRDB}) allows the
  first eight characters in an rwl file for series \acronym{ID}s but
  these are often shorter.  Typically the creators of rwl files use a
  logical labeling method that can allow the user to determine the tree
  and core \acronym{ID} from the label.

  Argument \code{\var{stc}} tells how each series separate into site,
  tree, and core \acronym{ID}s.  For instance a series code might be
  \code{"ABC011"} indicating site \code{"ABC"}, tree 1, core 1.  If this
  format is consistent then the \code{\var{stc}} mask would be
  \code{c(3, 2, 3)} allowing up to three characters for the core
  \acronym{ID} (i.e., pad to the right).  If it is not possible to
  define the scheme (and often it is not possible to machine read
  \acronym{ID}s), then the output \code{data.frame} can be built
  manually.  See Value for format.

  The function \code{autoread.ids} is a wrapper to \code{read.ids} with
  \code{stc="auto"}, i.e. automatic detection of the site / tree / core
  scheme, and different default values of some parameters.  In automatic
  mode, the names in the same \code{\var{rwl}} can even follow different
  site / tree / core schemes.  As there are numerous possible encoding
  schemes for naming measurement series, the function cannot always
  produce the correct result.

  With \code{stc="auto"}, the site part can be one of the following.
  \itemize{

    \item In names mostly consisting of numbers, the longest common
    prefix is the site part

    \item Alphanumeric site part ending with alphabet, when followed by
    numbers and alphabets

    \item Alphabetic site part (quite complicated actual
    definition).  Setting \code{\var{ignore.case}} to \code{"auto"}
    allows the function to try to guess when a case change in the middle
    of a sequence of alphabets signifies a boundary between the site
    part and the tree part.

    \item The characters before the first sequence of space /
    punctuation characters in a name that contains at least two such
    sequences } These descriptions are somewhat general, and the details
    can be found in regular expressions inside the function.  If a name
    does not match any of the descriptions, it is matched against a
    previously found site part, starting from the longest.

  The following \acronym{ID} schemes are detected and supported in the
  tree / core part.  The detection is done per site. \itemize{

    \item Numbers in tree part, core part starts with something else

    \item Alphabets in tree part, core part starts with something else

    \item Alphabets, either tree part all lower case and core part all
    upper case or vice versa.  For this to work,
    \code{\var{ignore.case}} must be set to \code{"auto"} or
    \code{FALSE}.

    \item All digits.  In this case, the number of characters belonging
    to the tree and core parts is detected with one of the following
    methods. \itemize{

      \item If numeric tree parts were found before, it is assumed that
      the core part is missing (one core per tree).

      \item It the series are numbered continuously, one core per tree
      is assumed.
      
      \item Otherwise, try to find a core part as the suffix so that the
      cores are numbered continuously.

    } If none of the above fits, the tree / core split of the all-digit
    names will be decided with the methods described further down the
    list, or finally with the fallback mechanism.

    \item The combined tree / core part is empty or one character.
    In this case, the core part is assumed to be missing.

    \item Tree and core parts separated by a punctuation or white space
    character

  } If the split of a tree / core part cannot be found with any of the
  methods described above, the prefix of the string is matched against a
  previously found tree part, starting from the longest.  The fallback
  mechanism for the still undecided tree / core parts is one of the
  following.  The first one is used if \code{\var{use.cor}} is
  \code{TRUE}, number two if it is \code{FALSE}.

  \enumerate{

    \item Pairwise correlation coefficients are computed between all
    remaining series.  Pairs of series with above median correlation
    are flagged as similar, and the other pairs are flagged as
    dissimilar.  Each possible number of characters (minimum 1) is
    considered for the share of the tree \acronym{ID}.  The
    corresponding unique would-be tree \acronym{ID}s determine a set of
    clusterings where one cluster is formed by all the measurement
    series of a single tree.  For each clustering (allocation of
    characters), an agreement score is computed.  The agreement score is
    defined as the sum of the number of similar pairs with matching
    cluster number and the number of dissimilar pairs with non-matching
    cluster number.  The number of characters with the maximum agreement
    is chosen.

    \item If the majority of the names in the site use \var{k}
    characters for the tree part, that number is chosen.  Otherwise, one
    core per tree is assumed.  Parameter \code{\var{typo.ratio}} has a
    double meaning as it also defines what is meant by majority here: at
    least \code{\var{typo.ratio} / (\var{typo.ratio} + 1) *
    \var{n.tot}}, where \var{n.tot} is the number of names in the site.
    
  } In both fallback mechanisms, the number of characters allocated for
  the tree part will be increased until all trees have a non-zero
  \acronym{ID} or there are no more characters.

  Suspected typing errors will be fixed by the function if
  \code{\var{fix.typos}} is \code{TRUE}.  The parameter
  \code{\var{typo.ratio}} affects the eagerness to fix typos, i.e. the
  number of counterexamples required to declare a typo.  The following
  main typo fixing mechanisms are implemented: \describe{

    \item{ Site \acronym{ID}s. }{ If a rare site string resembles an at
      least \code{\var{typo.ratio}} times more frequent alternative, and
      if fixing it would not create any name collisions, make the fix.
      The alternative string must be unique, or if there is more than
      one alternative, it is enough if only one of them is a look-alike
      string.  Any kind of substitution in one character place is
      allowed if the alternative string has the same length as the
      original string.  The alternative string can be one character
      longer or one character shorter than the original string, but only
      if it involves interpreting one digit as the look-alike alphabet
      or vice versa.  There are requirements to how long a site string
      must be in order to be eligible for replacement / typo fixing,
      i.e. cannot be shortened to zero length, cannot change the only
      character of a site string.  The parameters
      \code{\var{ignore.case}} and \code{\var{ignore.site.case}} have
      some effect on this typo fixing mechanism. }

    \item{ Tree and core \acronym{ID}s. }{ If all tree / core parts of a
      site have the same length, each character position is inspected
      individually.  If the characters in the \var{i}:th position are
      predominantly digits (alphabets), any alphabets (digits) are
      changed to the corresponding look-alike digit (alphabet) if there
      is one.  The look-alike groups are \{0, O, o\}, \{1, I, i\}, \{5,
      S, s\} and \{6, G\}.  The parameter \code{\var{typo.ratio}}
      determines the decision threshold of interpreting the type of each
      character position as alphabet (digit): the ratio of alphabets
      (digits) to the total number of characters must be at least
      \code{\var{typo.ratio} / (\var{typo.ratio} + 1)}.  If a name
      differs from the majority type in more than one character
      position, it is not fixed.  Also, no fixes are performed if any of
      them would cause a possible monotonic order of numeric prefixes to
      break. }
    
  }

  The function attempts to convert the tree and core substrings to
  integral values.  When this succeeds, the converted values are copied
  to the output without modification.  When non-integral substrings are
  observed, each unique tree is assigned a unique integral value.  The
  same applies to cores within a tree, but there are some subtleties
  with respect to the handling of duplicates.  Substrings are sorted
  before assigning the \code{numeric} \acronym{ID}s.

  The order of columns in \code{\var{rwl}}, in most cases, does not
  affect the tree and core \acronym{ID}s assigned to each series.
  
}
\value{

  A \code{data.frame} with column one named \code{"tree"} giving an
  \acronym{ID} for each tree and column two named \code{"core"} giving
  an \acronym{ID} for each core.  The original series \acronym{ID}s are
  copied from rwl as rownames.  The order of the rows in the output
  matches the order of the series in \code{\var{rwl}}.  If more than one
  site is detected, an additional third column named \code{"site"} will
  contain a site \acronym{ID}.  All columns have integral valued
  \code{numeric} values.
  
}
\author{ Andy Bunn (original version) and Mikko Korpela (patches,
  \code{\var{stc}="auto"}, \code{\var{fix.typos}}, etc.). }
\seealso{ \code{\link{rwi.stats}}, \code{\link{read.rwl}} }
\examples{library(utils)
data(ca533)
read.ids(ca533, stc = c(3, 2, 3))
autoread.ids(ca533)
}
\keyword{ misc }

\encoding{UTF-8}
\name{i.detrend}
\alias{i.detrend}
\title{ Interactively Detrend Multiple Ring-Width Series }
\description{
  Interactively detrend multiple tree-ring series by one of two methods, a
  smoothing spline or a statistical model.  This is a wrapper for
  \code{\link{detrend.series}}.
}
\usage{
i.detrend(rwl, y.name = names(rwl), nyrs = NULL, f = 0.5,
          pos.slope = FALSE)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}} or
    \code{\link{ca533}} }
  \item{y.name}{ a \code{character} vector of length
    \code{ncol(\var{rwl})} that gives the \acronym{ID} of each
    series.  Defaults to the column names of \code{\var{rwl}}. }
  \item{nyrs}{ a number giving the rigidity of the smoothing spline,
    defaults to 0.67 of series length if \code{\var{nyrs}} is
    \code{NULL}. }
  \item{f}{ a number between 0 and 1 giving the frequency response or
    wavelength cutoff.  Defaults to 0.5. }
  \item{pos.slope}{ a \code{logical} flag.  Will allow for a positive
    slope to be used in method \code{"ModNegExp"}.  If \code{FALSE} the
    line will be horizontal. }
}
\details{
  This function allows a user to choose detrending curves based on plots
  that are produced by \code{\link{detrend.series}} for which it is
  essentially a wrapper.  The user enters their choice of detrended
  method via keyboard at a prompt for each ring width series in
  \code{\var{rwl}}.  See \code{\link{detrend.series}} for examples and
  details on the detrending methods.
}
\value{
  A \code{data.frame} containing each detrended series according to the
  method used as columns and rownames set to
  \code{colnames(\var{y})}.  These are typically years.  Plots are also
  produced as the user chooses the detrending methods through keyboard
  input.
}
\author{ Andy Bunn }
\seealso{ \code{\link{detrend.series}} }
\keyword{ iplot }

\encoding{UTF-8}
\name{co021}
\docType{data}
\alias{co021}
\title{Schulman Old Tree No. 1, Mesa Verde}
\description{
  This data set gives the raw ring widths for Douglas fir
  \emph{Pseudotsuga menziesii} at Mesa Verde in Colorado, \acronym{USA}.
  There are 35 series.  Data set was created using \code{\link{read.rwl}}
  and saved to an .rda file using \code{\link{save}}.
}
\usage{data(co021)}
\format{
  A \code{data.frame} containing 35 tree-ring series in columns and 788
  years in rows.
}
\source{
  International tree-ring data bank, Accessed on 20-April-2021 at
  \url{https://www.ncei.noaa.gov/pub/data/paleo/treering/measurements/northamerica/usa/co021.rwl}
}
\references{
  Schulman, E. (1963) Schulman Old Tree No. 1 Data Set.  \acronym{IGBP}
  \acronym{PAGES}/World Data Center for Paleoclimatology Data
  Contribution Series 1983-CO021.RWL.  \acronym{NOAA}/\acronym{NCDC}
  Paleoclimatology Program, Boulder, Colorado, \acronym{USA}.
}
\keyword{datasets}

\encoding{UTF-8}
\name{corr.rwl.seg}
\alias{corr.rwl.seg}
\title{ Compute Correlations between Series }
\description{
  Computes the correlation between each tree-ring series in a rwl object.
}
\usage{
corr.rwl.seg(rwl, seg.length = 50, bin.floor = 100, n = NULL,
             prewhiten = TRUE, pcrit = 0.05, biweight = TRUE,
             method = c("spearman", "pearson","kendall"),
             make.plot = TRUE, label.cex = 1, floor.plus1 = FALSE,
             master = NULL,
             master.yrs = as.numeric(if (is.null(dim(master))) {
                              names(master)
                          } else {
                              rownames(master)
                          }),
             \dots)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}}. }
  \item{seg.length}{ an even integral value giving length of segments in
    years (e.g., 20, 50, 100 years). }
  \item{bin.floor}{ a non-negative integral value giving the base for
    locating the first segment (e.g., 1600, 1700, 1800
    \acronym{AD}).  Typically 0, 10, 50, 100, etc. }
  \item{n}{ \code{NULL} or an integral value giving the filter length
    for the \code{\link{hanning}} filter used for removal of low
    frequency variation. }
  \item{prewhiten}{ \code{logical} flag.  If \code{TRUE} each series is
    whitened using \code{\link{ar}}. }
  \item{pcrit}{ a number between 0 and 1 giving the critical value for
    the correlation test. }
  \item{biweight}{ \code{logical} flag.  If \code{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}}. }
  \item{method}{Can be either \code{"pearson"}, \code{"kendall"}, or
    \code{"spearman"} which indicates the correlation coefficient to be
    used.  Defaults to \code{"spearman"}.  See \code{\link{cor.test}}. }
  \item{make.plot}{ \code{logical flag} indicating whether to make a
    plot. }
  \item{label.cex}{ \code{numeric} scalar for the series labels on the
    plot.  Passed to \code{\var{axis.cex}} in \code{\link{axis}}. }
  \item{floor.plus1}{ \code{logical} flag.  If \code{TRUE}, one year is
    added to the base location of the first segment (e.g., 1601, 1701,
    1801 \acronym{AD}). }
  \item{master}{ \code{NULL}, a \code{numeric} \code{vector} or a
    \code{matrix}-like object of \code{numeric} values, including a
    \code{data.frame}.  If \code{NULL}, a number of master chronologies,
    one for each series in \code{\var{rwl}}, is built from
    \code{\var{rwl}} using the leave-one-out principle.  If a
    \code{vector}, the function uses this as the master chronology.  If
    a \code{matrix} or \code{data.frame}, this object is used for
    building the master chronology (no leave-one-out). }
  \item{master.yrs}{ a \code{numeric} vector giving the years of
    \code{\var{series}}.  Defaults to \code{names} or \code{rownames} of 
    \code{\var{master}} coerced to \code{numeric} type. }
  \item{\dots}{ other arguments passed to plot. }
}
\details{

  This function calculates correlation serially between each tree-ring
  series and a master chronology built from all the other series in the
  \code{\var{rwl}} object (leave-one-out principle).  Optionally, the
  user may give a \code{\var{master}} chronology (a \code{vector}) as an
  argument.  In the latter case, the same master chronology is used for
  all the series in the \code{\var{rwl}} object.  The user can also
  choose to give a \code{\var{master}} \code{data.frame} (series as
  columns, years as rows), from which a single master chronology is
  built.

  Correlations are done for each segment of the series where segments
  are lagged by half the segment length (e.g., 100-year segments would
  be overlapped by 50-years).  The first segment is placed according to
  \code{\var{bin.floor}}.  The minimum bin year is calculated as
  \code{ceiling(\var{min.yr}/\var{bin.floor})*\var{bin.floor}} where
  \code{\var{min.yr}} is the first year in either the \code{\var{rwl}}
  object or the user-specified \code{\var{master}} chronology, whichever
  is smaller.  For example if the first year is 626 and
  \code{\var{bin.floor}} is 100 then the first bin would start in 700.
  If \code{\var{bin.floor}} is 10 then the first bin would start in 630.

  Correlations are calculated for the first segment, then the second
  segment and so on.  Correlations are only calculated for segments with
  complete overlap with the master chronology.  For now, correlations are
  Spearman\enc{’}{'}s rho calculated via \code{\link{cor.test}} using
  \code{method = "spearman"}.

  Each series (including those in the rwl object) is optionally
  detrended as the residuals from a \code{\link{hanning}} filter with
  weight \code{\var{n}}.  The filter is not applied if \code{\var{n}} is
  \code{NULL}.  Detrending can also be done via prewhitening where the
  residuals of an \code{\link{ar}} model are added to each series
  mean.  This is the default.  The master chronology is computed as the
  mean of the \code{\var{rwl}} object using \code{\link{tbrm}} if
  \code{\var{biweight}} is \code{TRUE} and \code{rowMeans} if not.  Note
  that detrending can change the length of the series.  E.g., a
  \code{\link{hanning}} filter will shorten the series on either end by
  \code{floor(\var{n}/2)}.  The prewhitening default will change the
  series length based on the \code{\link{ar}} model fit.  The effects of
  detrending can be seen with \code{\link{series.rwl.plot}}.

  The function is typically invoked to produce a plot where each segment
  for each series is colored by its correlation to the master
  chronology.  Green segments are those that do not overlap completely
  with the width of the bin.  Blue segments are those that correlate
  above the user-specified critical value.  Red segments are those that
  correlate below the user-specified critical value and might indicate a
  dating problem.
}
\value{
  A \code{list} containing matrices \code{\var{spearman.rho}},
  \code{\var{p.val}}, \code{\var{overall}}, \code{\var{bins}},  
  \code{\var{rwi}}, vector \code{\var{avg.seg.rho}},
  numeric \code{\var{seg.lag}}, \code{\var{seg.length}}, \code{\var{pcrit}}, 
  \code{\var{label.cex}}.  An additional \code{character}
  \code{\var{flags}} is also returned if any segments fall below the
  critical value.  Matrix \code{\var{spearman.rho}} contains the
  correlations for each series by bin.  Matrix \code{\var{p.val}}
  contains the p-values on the correlation for each series by
  bin.  Matrix \code{\var{overall}} contains the average correlation and
  p-value for each series.  Matrix \code{\var{bins}} contains the years
  encapsulated by each bin.  The vector \code{\var{avg.seg.rho}}
  contains the average correlation for each bin. Matrix \code{\var{rwi}}
  contains the detrended rwl data, the numerics \code{\var{seg.lag}}, 
  \code{\var{seg.length}}, \code{\var{pcrit}}, \code{\var{label.cex}} 
  are from the oroginal call and used to pass into \code{\var{plot.crs}}. 
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{
  \code{\link{corr.series.seg}}, \code{\link{skel.plot}},
  \code{\link{series.rwl.plot}}, \code{\link{ccf.series.rwl}},
  \code{\link{plot.crs}}
}
\examples{library(utils)
data(co021)
crs <- corr.rwl.seg(co021, seg.length = 100, label.cex = 1.25)
names(crs)
## Average correlation and p-value for the first few series
head(crs$overall)
## Average correlation for each bin
crs$avg.seg.rho
}
\keyword{ manip }

\encoding{UTF-8}
\name{uuid.gen}
\alias{uuid.gen}
\title{
  \acronym{UUID} Generator
}
\description{
  Initializes and returns a generator of universally unique identifiers.
  Use the returned function repeatedly for creating one or more
  \acronym{UUID}s, one per function call.
}
\usage{
uuid.gen(more.state = "")
}
\arguments{
  \item{more.state}{
    A \code{character} string for altering the state of the generator
  }
}
\details{
  This function returns a function (closure) which generates
  \acronym{UUID}s.  The state of that anonymous function is set when
  \code{uuid.gen} is called.  The state consists of the following:
  \itemize{
    \item System and user information (\code{\link{Sys.info}})
    \item \R version (\code{\link{R.version}})
    \item Platform information (\code{\link{.Platform}})
    \item Working directory
    \item Process \acronym{ID} of the \R session
    \item Time when \code{uuid.gen} was called (precision of seconds or
    finer)
    \item The text in parameter \code{\var{more.state}}
  }

  The Pseudo Random Number Generator of \R (see
  \code{\link{.Random.seed}}) is used in the generation of
  \acronym{UUID}s.  No initialization of the \acronym{PRNG} is done.
  Tampering with the state of the \R \acronym{PRNG} while using a given
  \acronym{UUID} generator causes a risk of non-unique identifiers.
  Particularly, setting the state of the \acronym{PRNG} to the same
  value before two calls to the \acronym{UUID} generator guarantees two
  identical identifiers.  If two \acronym{UUID} generators have a
  different state, it is \emph{not} a problem to have the \acronym{PRNG}
  going through or starting from the same state with both generators.

  The user is responsible for selecting a \acronym{PRNG} with a
  reasonable number of randomness.  Usually, this doesn\enc{’}{'}t require any
  action.  For example, any \acronym{PRNG} algorithm available in \R
  works fine.  However, the uniqueness of \acronym{UUID}s can be
  destroyed by using a bad user-supplied \acronym{PRNG}.

  The \acronym{UUID}s produced by \code{uuid.gen} generators are Version
  4 (random) with 122 random bits and 6 fixed bits.  The \acronym{UUID}
  is presented as a \code{character} string of 32 hexadecimal digits and
  4 hyphens:

  \samp{xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx}

  where \var{x} is any hexadecimal digit and \var{y} is one of
  \code{"8"}, \code{"9"}, \code{"a"}, or \code{"b"}.  Each \var{x} and
  \var{y} in the example is an independent variables (for all practical
  purposes); subscripts are omitted for clarity.  The \acronym{UUID}
  generator gets 32 hex digits from the \acronym{MD5} message digest
  algorithm by feeding it a string consisting of the constant generator
  state and 5 (pseudo) random numbers.  After that, the 6 bits are fixed
  and the hyphens are added to form the final \acronym{UUID}.
}
\value{
  A parameterless function which returns a single \acronym{UUID}
  (\code{character} string)
}
\references{
  Leach, P., Mealling, M., and Salz, R. (2005) A Universally Unique
  IDentifier (\acronym{UUID}) \acronym{URN} namespace.  \acronym{RFC}
  4122, RFC Editor.  \url{https://www.rfc-editor.org/rfc/rfc4122.txt}.
}
\author{
  Mikko Korpela
}
\seealso{
  \code{\link[digest]{digest}}, \code{\link{Random}}
}
\examples{
## Normal use
ug <- uuid.gen()
uuids <- character(100)
for(i in 1:100){
  uuids[i] <- ug()
}
length(unique(uuids)) == 100 # TRUE, UUIDs are unique with high probability

## Do NOT do the following unless you want non-unique IDs
rs <- .Random.seed
set.seed(0L)
id1 <- ug()
set.seed(0L)
id2 <- ug()
id1 != id2 # FALSE, The UUIDs are the same
.Random.seed <- rs

## Strange usage pattern, but will probably produce unique IDs
ug1 <- uuid.gen("1")
set.seed(0L)
id1 <- ug1()
ug2 <- uuid.gen("2")
set.seed(0L)
id2 <- ug2()
id1 != id2 # TRUE, The UUIDs are different with high probability
.Random.seed <- rs
}
\keyword{ utilities }

\encoding{UTF-8}
\name{plot.rwl}
\alias{plot.rwl}
\title{
  Plotting Rwl Objects
}
\description{
  Plots rwl objects.
}
\usage{
\method{plot}{rwl}(x, plot.type=c("seg","spag"), ...)
}
\arguments{

  \item{x}{ An object of class \code{"rwl"}.  }

  \item{plot.type}{ Character.  Type "seg" calls \code{\link{seg.plot}} 
  while "spag" calls \code{\link{spag.plot}} }

  \item{\dots}{ Additional arguments for each \code{type}  }
  
}
\value{
  None.  A plot is produced.
}
\author{
  Andy Bunn
}
\seealso{
  \code{\link{read.rwl}}
}
\examples{library(graphics)
library(utils)
data(co021)
plot(co021, plot.type="seg")
plot(co021, plot.type="spag")
plot(co021, plot.type="spag", zfac=2)
}
\keyword{ hplot }

\encoding{UTF-8}
\name{cms}
\alias{cms}
\title{ C-Method Standardization }
\description{
  Detrend multiple ring-width series simultaneously using the C-method.
}
\usage{
cms(rwl, po, c.hat.t = FALSE, c.hat.i = FALSE)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}} }
  \item{po}{ a \code{data.frame} containing two variables.  Variable one
    (\code{\var{series}} in the example below) gives the series
    \acronym{ID} as either \code{character}s or \code{factor}s.  These
    must exactly match \code{colnames(\var{rwl})}.  Variable two
    (\code{\var{pith.offset}} in the example below) must be integral
    values and give the years from the beginning of the core to the pith
    (or center) of the tree. }
  \item{c.hat.t}{ a \code{logical} indicating whether to export the
    C-curves for each tree by biological age. }
  \item{c.hat.i}{ a \code{logical} indicating whether to export the
    expected ring widths for each series. }
}
\details{
  This method detrends and standardizes tree-ring series by calculating
  a growth curve based on constant annual basal area increment.  The
  method is based on the \dQuote{assumption that constant growth is
  expressed by a constant basal area increment distributed over a
  growing surface} (Biondi and Qeadan 2008).  The detrending is the
  estimation and removal of the tree\enc{’}{'}s natural biological
  growth trend.  The standardization is done by dividing each series by
  the growth trend to produce units in the dimensionless ring-width
  index (\acronym{RWI}).

  This attempts to remove the low frequency variability that is due to
  biological or stand effects.

  See the reference below for further details.
}
\value{
  A \code{data.frame} containing the dimensionless and detrended
  ring-width indices with column names, row names and dimensions of
  \code{\var{rwl}} if \code{\var{c.hat.t}} is \code{FALSE} and
  \code{\var{c.hat.i}} is \code{FALSE}.
  
  Otherwise a \code{list} of length 2 or 3 containing the \acronym{RWI}
  \code{data.frame}, a \code{data.frame} containing the C-curves for
  each tree (\code{\var{c.hat.t}}), and/or a vector containing the
  C-values for each tree (\code{\var{c.hat.i}}) depending on the output
  flags.  See Eq. 12 in Biondi and Qeadan (2008) for more detail on
  \code{\var{c.hat.t}}, and \code{\var{c.hat.i}}.
}
\references{
  Biondi, F. and Qeadan, F. (2008) A theory-driven approach to tree-ring
  standardization: Defining the biological trend from expected basal area
  increment.  \emph{Tree-Ring Research}, \bold{64}(2), 81\enc{–}{--}96.
}
\note{ DendroLab website: \url{https://dendrolaborg.wordpress.com/} }
\author{
  Code provided by DendroLab based on programming by F. Qeadan and
  F. Biondi, University of Nevada Reno, \acronym{USA} and adapted for
  dplR by Andy Bunn.  Patched and improved by Mikko Korpela.
}
\seealso{ \code{\link{detrend}}, \code{\link{chron}}, \code{\link{rcs}} }
\examples{library(graphics)
library(utils)
data(gp.rwl)
data(gp.po)
gp.rwi <- cms(rwl = gp.rwl, po = gp.po)
gp.crn <- chron(gp.rwi)
crn.plot(gp.crn, add.spline = TRUE)
## c.hat
gp.rwi <- cms(rwl = gp.rwl, po = gp.po, c.hat.t = TRUE, c.hat.i = TRUE)
dotchart(gp.rwi$c.hat.i, ylab = "Series", xlab = expression(hat(c)[i]))
tmp <- gp.rwi$c.hat.t
plot(tmp[, 1], type = "n", ylim = range(tmp, na.rm = TRUE),
     xlab = "Cambial Age", ylab = expression(hat(c)[t]))
apply(tmp, 2, lines)
}
\keyword{ manip }

\encoding{UTF-8}
\name{read.rwl}
\alias{read.rwl}
\title{ Read Ring Width File }
\description{
  This function reads in a file of ring widths (.rwl) in one of the
  available formats.
}
\usage{
read.rwl(fname,
         format = c("auto", "tucson", "compact", "tridas", "heidelberg", "csv"),
         \dots)
}
\arguments{

  \item{fname}{ a \code{character} vector giving the file name of the
    rwl file. }

  \item{format}{ a \code{character} vector giving the format.  This must
    be \code{"auto"} (rough automatic detection), \code{"tucson"},
    \code{"compact"}, \code{"tridas"}, \code{"heidelberg"} or \code{"csv"}.  Automatic
    format detection is the default but failable. }

  \item{\dots}{ arguments specific to the function implementing the
    operation for the chosen format. }
  
}
\details{
  This is a simple wrapper to the functions actually implementing the
  read operation.
}
\value{
  If a \code{"tucson"}, \code{"compact"}, \code{"heidelberg"}, \code{"csv"} file is
  read (even through \code{"auto"}), returns an object of class
  \code{c("rwl", "data.frame")} with the series in columns and the years
  as rows.  The series \acronym{ID}s are the column names and the years
  are the row names.
  
  If a \code{"tridas"} file is read (even through \code{"auto"}),
  returns a list of results.  See \code{\link{read.tridas}} for more
  information.
}
\author{ Mikko Korpela }
\seealso{ \code{\link{read.tucson}}, \code{\link{read.compact}},
  \code{\link{read.tridas}}, \code{\link{read.fh}}, \code{\link{csv2rwl}}, 
  \code{\link{write.rwl}} }
\keyword{ IO }

\encoding{UTF-8}
\name{bakker}
\alias{bakker}

\title{
  Basal Area Increment (Bakker)
}
\description{
  Convert multiple ring-width series to basal area increment (i.e., ring area) following the proportional method of Bakker (2005).
}
\usage{
  bakker(rwl, ancillary)
}

\arguments{
  \item{rwl}{a \code{data.frame} with series as columns and years as rows such as that produced by \code{read.rwl}}
  \item{ancillary}{A \code{data.frame} containing four columns. Column one (series in the example below) gives the series ID as either characters or factors. These must exactly match colnames(rwl). Column two (\code{PO}) gives the number of rings estimated to be missing to the pith. Column three (\code{d2pith}) gives the estimated distance to the pith (mm). Column four (\code{diam}) gives the diameter at breast height (DBH) in cm.}
}
\details{
  This converts ring-width series (mm) to ring-area series (mm squared) (aka basal area increments) based on the diameter of the tree, the missing distance to the pith and the missing number of rings to the pith, following the proportional method for reconstructing historical tree diameters by Bakker (2005).
  It prevents \code{\link{bai.out}} transformations from producing negative increments when the sum of all ring widths in a series is larger than DBH/2. It prevents \code{\link{bai.in}} transformations from producing too small values when the sum of all ring widths in a series is smaller than DBH/2.
  
}
\value{
  A list containing the following objects:
    \item{DBHhist_raw}{\code{data.frame} with the reconstructed diameter for each series with column names, row names and dimensions of rwl.}
  \item{baiBakker_raw}{\code{data.frame} with the basal area increments for each series with column names, row names and dimensions of rwl.}
  
}
\references{
  Bakker, J.D., 2005. A new, proportional method for reconstructing historical tree diameters. Canadian Journal of Forest Research 35, 2515–2520. https://doi.org/10.1139/x05-136
  
}
\author{
  Code by Stefan Klesse. Adapted for dplR by Andy Bunn.
}

\examples{
  
data(zof.rwl)
data(zof.anc)
zof.bakker <- bakker(rwl = zof.rwl,ancillary = zof.anc)
zof.bai <- zof.bakker$baiBakker_raw

# first series bai
yrs <- time(zof.rwl)
plot(yrs,zof.bai[,1],type="l",
     xlab="Year",
     ylab=expression(BAI~(mm^2)),
     main = colnames(zof.bai)[1])
  
}

\encoding{UTF-8}
\name{plot.crn}
\alias{plot.crn}
\alias{crn.plot}
\title{ Plot a Tree-Ring Chronology }
\description{
  This function makes a default plot of a tree-ring chronology from a
  \code{data.frame} of the type produced by \code{\link{chron}}, 
  \code{\link{chron.ars}}, \code{\link{chron.stabilized}}, \code{\link{ssf}}.
}
\usage{
\method{plot}{crn}(x, add.spline = FALSE, nyrs = NULL, ...)

}
\arguments{
  \item{x}{ a \code{data.frame} of \code{class{crn}}.e.g., as produced by
    \code{\link{chron}}, \code{\link{chron.ars}}, \code{\link{chron.stabilized}}, 
    \code{\link{ssf}}. }
  \item{add.spline}{ a \code{logical} flag.  Will add a line with a
    smoothing spline using \code{\link{caps} } }
  \item{nyrs}{ a number giving the rigidity of the smoothing spline.
    Defaults to 1/3 times the length of the first chronology if
    \code{\var{nyrs}} is \code{NULL} }
  \item{\dots}{ Additional arguments to pass to \code{\link{plot}}. }
}
\details{
  This makes a crude plot of one or more tree-ring chronologies.
}
\value{
  None.  Invoked for side effect (plot).
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{chron}}
}
\examples{library(graphics)
data(wa082)
# Truncate the RW data to a sample depth at least 5
wa082Trunc <- wa082[rowSums(!is.na(wa082))>4,]
# Detrend with age-dependent spline
wa082RWI <- detrend(wa082Trunc,method = "AgeDep")
# make several chronologies
wa082CRN1 <- chron(wa082RWI)
wa082CRN2 <- chron.stabilized(wa082RWI,
                              winLength=51,
                              biweight = TRUE,
                              running.rbar = TRUE)
wa082CRN3 <- chron.ars(wa082RWI)
wa082CRN4 <- ssf(wa082Trunc)

# and plot
plot.crn(wa082CRN1,add.spline = TRUE,nyrs=20)
plot.crn(wa082CRN2,add.spline = TRUE,nyrs=20)
plot(wa082CRN3,add.spline = TRUE,nyrs=20)
plot(wa082CRN4,add.spline = TRUE,nyrs=20)

# a custom crn
foo <- data.frame(wa082CRN1,sfc=wa082CRN4$sfc)
foo <- foo[,c(1,3,2)]
class(foo) <- c("crn","data.frame")
plot.crn(foo,add.spline = TRUE,nyrs=20)

}
\keyword{ hplot }

\encoding{UTF-8}
\name{wc.to.po}
\alias{wc.to.po}
\title{
  Convert Wood Completeness to Pith Offset
}
\description{
  This function creates a pith offset data structure based on wood
  completeness data.
}
\usage{
wc.to.po(wc)
}
\arguments{
  \item{wc}{
    A \code{data.frame} containing wood completeness data, as returned
    by \code{\link{read.tridas}}.
  }
}
\details{
  Computes the sum of the variables \code{\var{n.missing.heartwood}} and
  \code{\var{n.unmeasured.inner}} in \code{\var{wc}}.
}
\value{
  A \code{data.frame} containing two variables.  Variable one
  (\code{\var{series}}) gives the series \acronym{ID} as either
  \code{character}s or \code{factor}s.  These match
  \code{rownames(\var{wc})}.  Variable two (\code{\var{pith.offset}}) is
  of \code{integer} type and gives the years from the beginning of the
  core to the pith (or center) of the tree.  The minimum value is 1.
}
\author{
  Mikko Korpela
}
\seealso{
  \code{\link{po.to.wc}}, \code{\link{rcs}}, \code{\link{read.tridas}}
}
\examples{library(utils)
data(gp.po)
all(wc.to.po(po.to.wc(gp.po)) == gp.po)
}
\keyword{ manip }

\encoding{UTF-8}
\name{tbrm}
\alias{tbrm}
\title{ Calculate Tukey's Biweight Robust Mean }
\description{
  This calculates a robust average that is unaffected by outliers.
}
\usage{
tbrm(x, C = 9)
}
\arguments{
  \item{x}{ a \code{numeric} vector }
  \item{C}{ a constant.  \code{\var{C}} is preassigned a value of 9
    according to the Cook reference below but other values are
    possible. }
}
\details{
  This is a one step computation that follows the Affy whitepaper below,
  see page 22.  This function is called by \code{\link{chron}} to
  calculate a robust mean.  \code{\var{C}} determines the point at which
  outliers are given a weight of 0 and therefore do not contribute to
  the calculation of the mean.  \code{\var{C} = 9} sets values roughly
  +/-6 standard deviations to 0.  \code{\var{C} = 6} is also used in
  tree-ring chronology development.  Cook and Kairiukstis (1990) have
  further details.

  An exact summation algorithm (Shewchuk 1997) is used.  When some
  assumptions about the rounding of floating point numbers and
  conservative compiler optimizations hold, summation error is
  completely avoided.  Whether the assumptions hold depends on the
  platform, i.e. compiler and \acronym{CPU}.
}
\value{
  A \code{numeric} mean.
}
\references{

  Statistical Algorithms Description Document, 2002, Affymetrix.

  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences.}
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Mosteller, F. and Tukey, J. W. (1977) \emph{Data Analysis and
  Regression: a second course in statistics}.  Addison-Wesley.
  \acronym{ISBN-13}: 978-0-201-04854-4.

  Shewchuk, J. R. (1997) Adaptive precision floating-point arithmetic
  and fast robust geometric predicates.  \emph{Discrete and
  Computational Geometry}, \bold{18}(3), 305\enc{–}{--}363.

}
\author{ Mikko Korpela }
\seealso{ \code{\link{chron}} }
\examples{
library(stats)
library(utils)
foo <- rnorm(100)
tbrm(foo)
mean(foo)

## Compare
data(co021)
co021.rwi <- detrend(co021, method = "ModNegExp")
crn1 <- apply(co021.rwi, 1, tbrm)
crn2 <- chron(co021.rwi)
cor(crn1, crn2[, 1])
}
\keyword{ robust }
\keyword{ univar }

\encoding{UTF-8}
\name{glk}

\alias{glk}
\alias{glk.legacy}

\title{ Calculate \enc{Gleichläufigkeit}{Gleichlaeufigkeit} }

\description{
  This function calculates the \enc{Gleichläufigkeit}{Gleichlaeufigkeit} and related measures for a given set of tree-ring records.
}

\usage{
  glk(x, overlap = 50, prob = TRUE)
  glk.legacy(x)
}

\arguments{
  \item{x}{ a \code{data.frame} of tree-ring data with records in
    columns, and years as rows.}
  \item{overlap}{ integer value with minimal length of overlapping growth changes (compared number of tree rings - 1). Comparisons with less overlap are not compared.}
  \item{prob}{if \code{TRUE} then the probability of exceedence of the \enc{Gleichläufigkeit}{Gleichlaeufigkeit} will be calculated}
}

\details{
  \enc{Gleichläufigkeit}{Gleichlaeufigkeit} is a classical agreement test based on sign tests (Eckstein and Bauch, 1969). This function implements \enc{Gleichläufigkeit}{Gleichlaeufigkeit} as the pairwise comparison of all records in data set. This vectorized implementation is faster than the previous version and follows the original definition (Huber 1942), instead of the incorrect interpretation that has been used in the past (Schweingruber 1988, see Buras/Wilmking 2015 for the correction). 
  
  The probability of exceedence (p) for the \enc{Gleichläufigkeit}{Gleichlaeufigkeit} expresses the chance that the \enc{Gleichläufigkeit}{Gleichlaeufigkeit} is incorrect. The observed value of the \enc{Gleichläufigkeit}{Gleichlaeufigkeit} is converted to a z-score and based on the standard normal curve the probability of exceedence is calculated. The result is a matrix of all p-values (Jansma 1995, 60-61, see also Visser 2020).
  
  Note that prior to dplR version 1.7.2, \code{glk} did not have the \code{overlap} or \code{prob} and returned a \code{matrix} with just the \enc{Gleichläufigkeit}{Gleichlaeufigkeit} for all possible combinations of records. That function can still be accessed via \code{glk.legacy}.
  
}

\value{
  The funtions returns a named \code{list} of two or three matrices  (p_mat is optional if \code{prob = TRUE}):
    \enumerate{
      \item glk_mat: \code{matrix} with \enc{Gleichläufigkeit}{Gleichlaeufigkeit with all possible combinations of records}
      \item overlap: \code{matrix} with number of overlapping growth changes.This is the number of overlapping years minus one.
      \item p_mat: \code{matrix} of all probabilities of exceedence for all observed \enc{Gleichläufigkeit}{Gleichlaeufigkeit} values.
    }
  
  The matrices can be extracted from the list by selecting the name or index number. If two curves have less than 3 years of overlap, \enc{Gleichläufigkeit}{Gleichlaeufigkeit} cannot be computed, and \code{NA} is returned.
  To calculate the global glk of the dataset \code{mean(x$glk_mat, na.rm = TRUE)}.
}

\author{ Christian Zang.  Patched and improved by Mikko Korpela.
  Improved by Allan Buras. Further improved and expanded by Ronald Visser and Andy Bunn}

\references{
  
  Buras, A. and Wilmking, M. (2015) Correcting the calculation of Gleichläufigkeit, \emph{Dendrochronologia} \bold{34}, 29-30. DOI: https://doi.org/10.1016/j.dendro.2015.03.003
  
  Eckstein, D. and Bauch, J. (1969) Beitrag zur Rationalisierung eines dendrochronologischen Verfahrens und zur Analyse seiner Aussagesicherheit. \emph{Forstwissenschaftliches Centralblatt}, \bold{88}(1), 230-250.
  
  Huber, B. (1943) Über die Sicherheit jahrringchronologischer Datierung. \emph{Holz als Roh- und Werkstoff} \bold{6}, 263-268. DOI: https://doi.org/10.1007/BF02603303
  
  Jansma, E., 1995. \emph{RemembeRINGs; The development and application of local and regional tree-ring chronologies of oak for the purposes of archaeological and historical research in the Netherlands}, Nederlandse Archeologische Rapporten 19, Rijksdienst voor het Oudheidkundig Bodemonderzoek, Amersfoort
  
  Schweingruber, F. H. (1988) \emph{Tree rings: basics and applications of dendrochronology}, Kluwer Academic Publishers, Dordrecht, Netherlands, 276 p.
  
  Visser, R.M. (2020) On the similarity of tree-ring patterns: Assessing the influence of semi-synchronous growth changes on the \enc{Gleichläufigkeit}{Gleichlaeufigkeit} for big tree-ring data sets,\emph{Archaeometry}, \bold{63}, 204-215 DOI: https://doi.org/10.1111/arcm.12600
}

\seealso{
  \code{\link{sgc}} \code{sgc} is an alternative for \code{glk})
}

\examples{library(utils)
  data(ca533)
  ca533.glklist <- glk(ca533)
  ca533.glk_mat <- ca533.glklist$glk_mat
  # calculating the mean GLK including self-similarities
  mean(ca533.glk_mat, na.rm = TRUE) 
  # calculating the mean GLK excluding self-similarities
  mean(ca533.glk_mat[upper.tri(ca533.glk_mat)], na.rm = TRUE) 

}

\keyword{ ts }

\encoding{UTF-8}
\name{as.rwl}
\alias{as.rwl}
\title{ as.rwl }
\description{
  Attempts to turn its argument into a rwl object.
}
\usage{
as.rwl(x)
}
\arguments{
  \item{x}{ a \code{data.frame} or \code{matrix} with series as columns and years as rows }
}
\details{
  This tries to coerce \code{x} into class \code{c("rwl","data.frame")}. Failable.
}
\value{
  An object of class \code{c("rwl", "data.frame")} with the series in
  columns and the years as rows.  The series \acronym{ID}s are the
  column names and the years are the row names.
}

\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\examples{
library(graphics)
library(stats)
library(utils)
## Toy
n <- 100
## Make a data.frame that is tree-ring like
base.series <- 0.75 + exp(-0.2 * 1:n)
foo <- data.frame(x1 = base.series + abs(rnorm(n, 0, 0.25)),
                  x2 = base.series + abs(rnorm(n, 0, 0.25)),
                  x3 = base.series + abs(rnorm(n, 0, 0.25)),
                  x4 = base.series + abs(rnorm(n, 0, 0.25)),
                  x5 = base.series + abs(rnorm(n, 0, 0.25)),
                  x6 = base.series + abs(rnorm(n, 0, 0.25)))
# coerce to rwl and use plot and summary methods
foo <- as.rwl(foo)
class(foo)
plot(foo, plot.type="spag")
summary(foo)
}
\keyword{ manip }

\encoding{UTF-8}
\name{pointer}

\alias{pointer}

\title{ Calculates Pointer Years from a Group of Ring-Width Series }

\description{

  This function calculates pointer years on a \code{data.frame} of
  ring-width series using the Becker algorithm.  The pointer years are
  computed with adjustable thresholds of relative radial growth
  variation and number of series displaying similar growth pattern
  (i.e. positive or negative variations).
  
}

\usage{
pointer(rwl, rgv.thresh = 10, nseries.thresh = 75, round.decimals = 2)
}

\arguments{

  \item{rwl}{ a \code{data.frame} with ring-width series as columns and
    years as rows such as that produced by \code{\link{read.rwl}}. }
  
  \item{rgv.thresh}{ a \code{numeric} giving the minimum absolute
    relative radial growth variation (in percentage) above which the
    growth change from the year \emph{t-1} to \emph{t} is considered as
    significant.  Must be > 0.  Values > 100 are possible but highly
    unusual.  See references.  Defaults to 10. }
  
  \item{nseries.thresh}{ a \code{numeric} giving the minimum percentage
    of series displaying significant relative radial growth variations
    of a same sign above which the year \code{\var{t}} is considered as
    a pointer year.  Positive significant variations will results in a
    positive pointer year, negative ones in a negative pointer year.
    This number ranges from 1 to 100.  Defaults to 75. }

  \item{round.decimals}{ an \code{integer} indicating the number of
    decimal places to be used for outputs.  This number must be
    positive.  Defaults to 2. }

}

\details{

  This calculates pointer years from ring-width series for each year
  \code{\var{t}} of the time period covered by the series using the
  Becker algorithm.  This algorithm is based on, first, the calculation
  of the individual relative radial growth variation by comparison of
  ring-width of year \code{\var{t}} to that of year \code{\var{t}-1} for
  each series, and second, the inter-series comparison of both sign and
  magnitude of these variations.
  
  For example, if \code{\var{rgv.thresh}} and
  \code{\var{nseries.thresh}} are set at 10 and 75 respectively, pointer
  years will be defined as those years when at least 75\% of the series
  present an absolute relative radial growth variation higher than 10\%.
  
  Users unfamiliar with the Becker algorithm should refer to Becker et
  al. (1994) and \enc{Mérian}{Merian} and Lebourgeois (2011) for further
  details.
  
}

\value{

  A \code{data.frame} containing the following columns (each row
  corresponds to one position of the window):

  \item{Year}{ Considered year (t).}

  \item{Nb.series}{ Number of available series.}

  \item{Perc.pos}{ Percentage of series displaying a significant
    positive radial growth variation.}

  \item{Perc.neg}{ Percentage of series displaying a significant
    negative radial growth variation.}

  \item{Nature}{ Number indicating whether the year is a positive
    pointer year (1), a negative pointer year (-1) or a regular year
    (0).}

  \item{RGV_mean}{ Mean radial growth variations over the available
    series.}

  \item{RGV_sd}{ Standard deviation of the radial growth variations over
    the available series. }

}

\references{ 

  Becker, M., Nieminen, T. M., and \enc{Gérémia}{Geremia}, F. (1994)
  Short-term variations and long-term changes in oak productivity in
  northeastern France \enc{–}{--} the role of climate and atmospheric
  CO2.  \emph{Annals of Forest Science}, \bold{51}(5),
  477\enc{–}{--}492.

  \enc{Mérian}{Merian}, P. and Lebourgeois, F. (2011) Size-mediated
  climate\enc{–}{--}growth relationships in temperate forests: A multi-species
  analysis.  \emph{Forest Ecology and Management},
  \bold{261}(8), 1382\enc{–}{--}1391.
  
}

\author{ Pierre \enc{Mérian}{Merian}.  Improved by Mikko Korpela and Andy
  Bunn. }

\seealso{ \code{\link{skel.plot}} }

\examples{## Pointer years calculation on ring-width series. Returns a data.frame.
library(utils)
data(gp.rwl)
py <- pointer(rwl=gp.rwl, rgv.thresh=10, nseries.thresh=75,
              round.decimals=2)
tail(py)
}

\encoding{UTF-8}
\name{dplR-package}
\alias{dplR-package}
\alias{dplR}
\docType{package}
\title{
  Dendrochronology Program Library in \R
}
\description{
  This package contains functions for performing some standard tree-ring
  analyses.
}
\details{
\tabular{ll}{
Package: \tab dplR\cr
Type: \tab Package\cr
License: \tab \acronym{GPL} (>= 2)\cr
}
\emph{Main Functions}
  \describe{
    \item{\code{\link{read.rwl}}}{ reads rwl files }
    \item{\code{\link{detrend}}}{ detrends raw ring widths }
    \item{\code{\link{chron}}}{ builds chronologies }
    \item{\code{\link{corr.rwl.seg}}}{ crossdating function }
  }
}
\author{
  Andy Bunn \email{andy.bunn@wwu.edu} with major additions from Mikko
  Korpela and other significant contributions from Franco Biondi, Filipe
  Campelo, Pierre \enc{Mérian}{Merian}, Fares Qeadan and Christian
  Zang.  Function \code{\link{redfit}} is an improved translation of
  program REDFIT which is original work of Manfred Mudelsee and Michael
  Schulz.  Jacob Cecile contributed a bug fix to
  \code{\link{detrend.series}}.  Allan Buras came up with the revised
  formula of \code{\link{glk}} in dplR >= 1.6.1.
}
\references{
  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Fritts, H. C. (2001) \emph{Tree Rings and Climate}.
  Blackburn.  \acronym{ISBN-13}: 978-1-930665-39-2.
}
\keyword{ package }

\encoding{UTF-8}
\name{read.crn}
\alias{read.crn}
\title{ Read Tucson Format Chronology File }
\description{
  This function reads in a Tucson (decadal) format file of tree-ring
  chronologies (.crn).
}
\usage{
read.crn(fname, header = NULL, encoding = getOption("encoding"),
         long = TRUE)
}
\arguments{
  \item{fname}{ a \code{character} vector giving the file name of the
    crn file. }
  \item{header}{ \code{logical} flag indicating whether the file has a
    header.  If \code{NULL} then the function will attempt to determine
    if a header exists }
  \item{encoding}{the name of the encoding to be used when reading the
    crn file.  Usually the default value will work, but a crn file
    written in a non-default encoding may crash the function.  In that
    case, identifying the encoding and specifying it here should fix the
    problem.  Examples of popular encodings available on many systems
    are \code{"ASCII"}, \code{"UTF-8"}, and \code{"latin1"} alias
    \code{"ISO-8859-1"}.  See the help of \code{\link{file}}.}
  \item{long}{ \code{logical} flag indicating whether to automatically
    detect when an input file uses more than 4 characters for the decade.
    If \code{FALSE}, the function assumes 6 characters are used for the
    site \acronym{ID} and 4 characters for the decade, which is the
    standard. If \code{TRUE} (the default), long records may work. }
}
\details{
  This reads in a standard crn file as defined according to the
  standards of the \acronym{ITRDB} at
  \url{https://www.ncei.noaa.gov/pub/data/paleo/treering/treeinfo.txt}.  Despite the
  standards at the \acronym{ITRDB}, this occasionally fails due to
  formatting problems.
}
\value{
  A \code{data.frame} with each chronology in columns and the years as
  rows.  The chronology \acronym{ID}s are the column names and the years
  are the row names.  If the file includes sample depth that is included
  as the last column (\code{\var{samp.depth}}).  The output class is
  class "crn" and "data.frame"
}
\author{ Andy Bunn.  Patched and improved by Mikko Korpela. }
\keyword{ IO }

\encoding{UTF-8}
\name{bai.in}
\alias{bai.in}
\title{ Basal Area Increment (Inside Out) }
\description{
  Convert multiple ring-width series to basal area increment (i.e., ring
  area) going from the pith to the bark.
}
\usage{
bai.in(rwl, d2pith = NULL)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}} }
  \item{d2pith}{ an optional \code{data.frame} containing two variables.  If
    present, then variable one (\code{series} in the example below)
    gives the series \acronym{ID} as either \code{characters} or
    \code{factors}.  These must exactly match
    \code{colnames(\var{rwl})}.  Variable two (\code{\var{d2pith}} in
    the example below) gives the distance from the innermost measured
    ring to the pith of the tree in mm.  If \code{\var{d2pith}} is
    \code{NULL} then the distance to pith is assumed to be zero for each
    series (column) in \code{\var{rwl}}. }
}
\details{
  This converts ring-width series (mm) to ring-area series (mm squared)
  (aka basal area increments) based on the distance between the
  innermost measured ring and the pith of the tree.  It is related to
  \code{\link{bai.out}}, which calculates each ring area starting from
  the outside of the tree and working inward.  Both methods assume a
  circular cross section (Biondi 1999).  See the references below for
  further details.
}
\value{
  A \code{data.frame} containing the ring areas for each series with
  column names, row names and dimensions of \code{\var{rwl}}.
}
\references{
  Biondi, F. (1999) Comparing tree-ring chronologies and repeated timber
  inventories as forest monitoring tools.  \emph{Ecological
    Applications}, \bold{9}(1), 216\enc{–}{--}227.

  Biondi, F. and Qeadan, F. (2008) A theory-driven approach to tree-ring
  standardization: Defining the biological trend from expected basal area
  increment.  \emph{Tree-Ring Research}, \bold{64}(2), 81\enc{–}{--}96.
}
\note{ DendroLab website: \url{https://dendrolaborg.wordpress.com/} }
\author{ Code by Andy Bunn based on work from DendroLab, University of
  Nevada Reno, \acronym{USA}.  Patched and improved by Mikko Korpela. }
\seealso{ \code{\link{bai.out}} }
\examples{
library(graphics)
library(stats)
library(utils)
## Toy
n <- 100
## Make three fake tree-ring series to show that these funcs work on rwl objects
base.series <- 0.75 + exp(-0.2 * 1:n)
rwl <- data.frame(x1 = base.series + abs(rnorm(n, 0, 0.05)),
                  x2 = base.series + abs(rnorm(n, 0, 0.05)),
                  x3 = base.series + abs(rnorm(n, 0, 0.05)))

## The inside out method
foo <- bai.in(rwl = rwl)
## The outside in method
bar <- bai.out(rwl = rwl)

## Identical
head(bar)
head(foo)

## Use gp data
data(gp.rwl)
data(gp.d2pith)
foo <- bai.in(rwl = gp.rwl, d2pith = gp.d2pith)
foo.crn <- chron(foo)
yrs <- time(foo.crn)
plot(yrs, foo.crn[, 1], type = "n",
     xlab = "Year", ylab = expression(mm^2))
lines(yrs, foo.crn[, 1], col = "grey", lty = "dashed")
lines(yrs, caps(foo.crn[, 1], nyrs = 32), col = "red", lwd = 2)
}
\keyword{ manip }

\encoding{UTF-8}
\name{rcs}
\alias{rcs}
\title{ Regional Curve Standardization }
\description{  
  Detrend multiple ring-width series simultaneously using a regional
  curve.
}
\usage{
rcs(rwl, po = NULL, nyrs = NULL, f = 0.5, biweight = TRUE, ratios = TRUE,
    rc.out = FALSE, make.plot = TRUE, method = c("caps", "ads"),
    min.n = NULL, pos.slope = TRUE, \dots)
}
\arguments{
  \item{rwl}{ a \code{data.frame} with series as columns and years as
    rows such as that produced by \code{\link{read.rwl}} }
  \item{po}{ a \code{data.frame} containing two variables.  Variable one
    (\code{\var{series}} in the example below) gives the series
    \acronym{ID} as either \code{character}s or \code{factor}s.  These
    must exactly match \code{colnames(\var{rwl})}.  Variable two
    (\code{\var{pith.offset}} in the example below) must be integral
    values and give the years from the beginning of the core to the pith
    (or center) of the tree.  The minimum value is 1.  If \code{NULL}
    (the default) a pith offset of 1 is assumed for all series, meaning
    each series is treated as beginning at cambial age 1.  This is
    almost certainly unrealistic for most datasets but is provided as a
    convenience when pith offset data are unavailable. }
  \item{nyrs}{ a number giving the rigidity of the smoothing spline.
    For \code{method = "caps"} this defaults to 0.1 of the length of
    the regional curve if \code{NULL}.  For \code{method = "ads"} this
    sets the starting stiffness (\code{nyrs0}) and defaults to 50 if
    \code{NULL}. }
  \item{f}{ a number between 0 and 1 giving the frequency response or
    wavelength cutoff.  Defaults to 0.5. }
  \item{biweight}{ \code{logical} flag.  If \code{TRUE} then a robust
    mean is calculated using \code{\link{tbrm}}. }
  \item{ratios}{ \code{logical} flag.  If \code{TRUE} (the default) then
    indices are calculated by division, if \code{FALSE} indices are
    calculated by subtraction. }
  \item{rc.out}{ \code{logical} flag.  Returns the regional curve along
    with the ring-width indices if \code{TRUE}. }
  \item{make.plot}{ \code{logical} flag.  Makes a plot of the raw
    ring-width series by cambial age, the biweight mean, and the regional
    curve if \code{TRUE}. }
  \item{method}{ a \code{character} string giving the smoothing method
    to apply to the mean regional curve.  Either \code{"caps"} (the
    default, a cubic smoothing spline via \code{\link{caps}}) or
    \code{"ads"} (an age-dependent spline via \code{\link{ads}}). }
  \item{min.n}{ an optional integer giving the minimum sample depth
    required at a given cambial age for that age to be included when
    fitting the regional curve.  Ages where fewer than \code{\var{min.n}}
    series contribute are excluded from the tail of the curve, and
    indices for rings falling beyond the truncated curve are returned as
    \code{NA}.  If \code{NULL} (the default) the full curve is used. }
  \item{pos.slope}{ a \code{logical} flag passed to \code{\link{ads}}.
    If \code{TRUE} (the default) a positive slope is allowed at the end
    of the fitted curve.  Ignored when \code{\var{method} = "caps"}. }
  \item{\dots}{ other arguments passed to
    \code{\link[graphics]{plot}}. }
}
\details{

  This method detrends and standardizes tree-ring series by calculating
  an age-related growth curve specific to the \code{\var{rwl}}.  The
  detrending is the estimation and removal of the tree\enc{’}{'}s natural
  biological growth trend.  The standardization is done by either
  dividing each series by the growth trend or subtracting the growth
  trend from each series to produce units in the dimensionless
  ring-width index (\acronym{RWI}).  The option to produce indices by
  subtraction is intended to be used on series that have been subject to
  variance stabilization (e.g., using \code{\link{powt}}).

  Two smoothers are available for fitting the regional curve.  The
  default (\code{method = "caps"}) uses a cubic smoothing spline where
  the frequency response is 0.50 at a wavelength of 10 percent of the
  maximum cambial age unless specified differently via \code{\var{nyrs}}
  and \code{\var{f}} (see \code{\link{caps}}).  The alternative
  (\code{method = "ads"}) uses an age-dependent spline whose stiffness
  increases with cambial age, which may better preserve low-frequency
  variability in the biological trend (see \code{\link{ads}}).

  The \code{\var{min.n}} argument can be used to truncate the regional
  curve where sample depth becomes thin, preventing poorly replicated
  tail ages from influencing the fit.

  This attempts to remove the low frequency variability that is due to
  biological or stand effects.  See the references below for further
  details on detrending in general, and Biondi and Qeadan (2008) for an
  explanation of \acronym{RCS}.

}
\value{

  A \code{data.frame} containing the dimensionless and detrended
  ring-width indices with column names, row names and dimensions of
  \code{\var{rwl}}.  If \code{\var{rc.out}} is \code{TRUE} then a
  \code{list} will be returned with a \code{data.frame} containing the
  detrended ring widths as above and a \code{vector} containing the
  regional curve.

}
\references{
  Biondi, F. and Qeadan, F. (2008) A theory-driven approach to tree-ring
  standardization: Defining the biological trend from expected basal area
  increment.  \emph{Tree-Ring Research}, \bold{64}(2), 81\enc{–}{--}96.

  Cook, E. R. and Kairiukstis, L. A., editors (1990) \emph{Methods of
    Dendrochronology: Applications in the Environmental Sciences}.
    Springer.  \acronym{ISBN-13}: 978-0-7923-0586-6.

  Fritts, H. C. (2001) \emph{Tree Rings and Climate}.
  Blackburn.  \acronym{ISBN-13}: 978-1-930665-39-2.
}
\author{
  Code provided by DendroLab based on programming by F. Qeadan and
  F. Biondi, University of Nevada Reno, \acronym{USA} and adapted for
  dplR by Andy Bunn.  Patched and improved by Mikko Korpela.  Extended
  with \code{method}, \code{min.n}, and \code{pos.slope} arguments by
  Andy Bunn.
}
\seealso{ \code{\link{detrend}}, \code{\link{chron}}, \code{\link{cms}},
  \code{\link{caps}}, \code{\link{ads}} }
\examples{library(utils)
data(gp.rwl)
data(gp.po)

## Basic use: caps smoother (default), return indices only
gp.rwi <- rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
              make.plot = FALSE)
str(gp.rwi)

## Return the regional curve alongside the indices
gp.out <- rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
              rc.out = TRUE, make.plot = FALSE)
str(gp.out)

## Plot the regional curve with a title
rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
    rc.out = FALSE, make.plot = TRUE, main = "Regional Curve")

## Use the age-dependent spline smoother instead of caps
gp.rwi.ads <- rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
                  method = "ads", rc.out = FALSE, make.plot = FALSE)

## Truncate the regional curve where fewer than 5 series contribute.
## Indices for rings beyond the truncated curve are returned as NA.
gp.out.trunc <- rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
                    method = "ads", min.n = 5,
                    rc.out = TRUE, make.plot = FALSE)

## Compare the two smoothers visually
op <- par(no.readonly = TRUE)
par(mfrow = c(1, 2))
rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
    method = "caps", make.plot = TRUE, main = "caps")
rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
    method = "ads", make.plot = TRUE, main = "ads")
par(op)

## Compare caps and ads indices for one series
gp.caps <- rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
               method = "caps", rc.out = FALSE, make.plot = FALSE)
gp.ads  <- rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
               method = "ads",  rc.out = FALSE, make.plot = FALSE)
cor(gp.caps[[1]], gp.ads[[1]], use = "complete.obs")

## Use subtraction rather than division (e.g. after variance stabilization)
gp.rwi.sub <- rcs(rwl = gp.rwl, po = gp.po, biweight = TRUE,
                  ratios = FALSE, make.plot = FALSE)
}
\keyword{ manip }

\encoding{UTF-8}
\name{strip.rwl}
\alias{strip.rwl}
\title{Chronology Stripping by \acronym{EPS}}
\description{
  \acronym{EPS}-based chronology stripping after Fowler & Boswijk 2003.
}
\usage{
strip.rwl(rwl, ids = NULL, verbose = FALSE, comp.plot = FALSE,
          legacy.eps = FALSE)
}
\arguments{

  \item{rwl}{a \code{data.frame} of raw tree-ring widths series, such as
    that produced by \code{\link{read.rwl}} or \code{\link{read.fh}}}

  \item{ids}{ an optional \code{data.frame} with column one named
    \code{"tree"} giving a \code{numeric} \acronym{ID} for each tree and
    column two named \code{"core"} giving a \code{numeric} \acronym{ID}
    for each core.  This is passed on to \code{\link{rwi.stats}}.  See
    its manual for the meaning of the default value \code{NULL} and more
    information. }

  \item{verbose}{\code{logical} flag, indicating if the \acronym{EPS}
    calculated at each step and other details should be printed on the
    console during the chronology stripping process}

  \item{comp.plot}{\code{logical} flag, indicating if a diagnostic plot
    with year-wise stripped and unstripped \acronym{EPS} should be drawn
    (see details below)}

  \item{legacy.eps}{\code{logical} flag, indicating if the \acronym{EPS}
    will be calculated with \code{rwi.stats} (\code{FALSE}, the default)
    or \code{rwi.stats.legacy} (\code{TRUE})}

}
\value{

  The functions returns a \code{data.frame} of raw tree-ring widths,
  where series that do not contribute to an overall improvement in
  \acronym{EPS} are left out.
  
}
\details{

  The \acronym{EPS}-based chronology stripping is implemented after
  Fowler & Boswijk 2003: First, all series are standardized using a
  double detrending procedure with splines and frequency cutoffs of 50\%
  at 20 and 200 years.  Then, \acronym{EPS} is calculated for the
  chronology including all (remaining) series.  In each iteration, the
  algorithm calculates leave-one-out \acronym{EPS} values, and the
  series whose removal increases overall \acronym{EPS} the most is
  discarded.  This is repeated until no further increase in
  \acronym{EPS} is gained by discarding a single series.  The procedure
  is then repeated in the opposite direction, i.e., the reinsertion of
  each previously removed series into the \code{data.frame} is
  considered.  In each iteration, the series (if any) whose reinsertion
  increases \acronym{EPS} the most is reinserted.  As a last step,
  \acronym{EPS} is calculated for each year of the stripped and original
  chronology including all series.  If \code{comp.plot} is set to
  \code{TRUE}, a diagnostic plot is shown for the year-wise comparison.

  When verbose output is chosen, the \acronym{EPS} values for all
  leave-one-out (or back-in) chronologies are reported.  If discarding
  or re-inserting a single series leads to an improvement in
  \acronym{EPS}, this series is marked with an asterisk.

}
\references{

  Fowler, A. and Boswijk, G. (2003) Chronology stripping as a tool for
  enhancing the statistical quality of tree-ring
  chronologies. \emph{Tree-Ring Research}, \bold{59}(2), 53\enc{–}{--}62.
  
}
\seealso{
  \code{\link{rwi.stats}}
}
\examples{
library(utils)
data(anos1)
anos1.ids <- read.ids(anos1, stc = c(4, 3, 1))
srwl <- strip.rwl(anos1, ids = anos1.ids, verbose = TRUE)
tail(srwl)
}
\author{
  Christian Zang.  Patched and improved by Mikko Korpela.
}
\keyword{ manip }
